//
//  fair_counterexample.h
//  
//
//  Created by Ivaylo Dobrikov on 02/11/15.
//
//

#ifndef fair_counterexample_h
#define fair_counterexample_h

#include <stdio.h>
#include "ltlc.h"

/* transition_id: the id of the transition in the search graph
   op: the term reference of the transition operation 
 */
typedef struct {
    unsigned int transition_id;
    SP_term_ref op_id;
    // is an empty list when none of the fairness arguments are matched
//    SP_term_ref fairness_op_term;
} itemtransition;

/* an item for the counterexample search.
   atom: the atom from the tarjan search
   executed_transition: the transition by which the item is reached
   wf_enabled: the array of enabled transitions i.r.t. the weak fairness constraints
   sf_enabled: the array of enabled transitions i.r.t. the strong fairness constraints
*/
typedef struct {
    ltlatom *atom;
    /* the transition from which the item is reached
       incl. saved the SP_term_ref for the incoming transition */
    itemtransition executed_transition;
//    itemtransition *wf_enabled;
//    itemtransition *sf_enabled;
    unsigned int wf_size;
    bitfield *wf_executed;
    unsigned int sf_size;
    bitfield *sf_executed;
} fairloopitem;

typedef struct {
    ltlarray stack;
    unsigned int stacksize;
    unsigned int arraysize;
    ltlarray nodes;
    unsigned int number_of_nodes;
} loop_data;

void get_fair_loop(ltl_formula *f, found_scc_type *found, modelitem **result, int *resultlength, int *loop_length);
#endif /* fair_counterexample_h */
