//
//  loop_fairness.c
//  
//
//  Created by Ivaylo Dobrikov on 24/09/14.
//
//

#include "loop_fairness.h"


SP_term_ref create_loop_nodes_list(ltlarray *bf_items, unsigned int solution, int size) {
    loopitem *item;
    unsigned int index = solution;
    
    ltlatom *atom;
    ltlnode *node;
    SP_term_ref ref,list,tail;
    int i;
    tail = SP_new_term_ref();
    list = tail;
    if (size > 0) {
        item = (loopitem*) get_array_element(bf_items, solution);
        index = item->previous;
        for (i=0; i<size; i++) {
            item = (loopitem*) get_array_element(bf_items,index);
            atom = item->atom;
            node = atom->node;
            if (node->type == int_node) {
                ref = SP_new_term_ref();
                SP_put_integer(ref, node->id.id);
            } else {
                ref = node->id.ref;
            }
            list = SP_new_term_ref();
            SP_cons_list(list,ref,tail);
            tail = list;
            index = item->previous;
        }
    } else {
        fprintf(stderr, "No atom in the loop (create_loop_nodes_list).");
        exit(1);
    }
    return list;
}

int check_loop_strong_fairness1(ltl_formula *f, ltlatom **loop_atoms, int loop_size, SP_term_ref loop_node_ids, SP_term_ref sf_assumption, SP_pred_ref trans_ids_predicate) {

    int i,msize;
    int sf_loop_result = 1;
    SP_integer* matching_transitions;
    int sfargs_size;
    SP_term_ref *sf_arguments = prolog_list_to_array(sf_assumption,&sfargs_size);

    for (i=0; i<sfargs_size; i++) {
        matching_transitions = get_matching_transitions(trans_ids_predicate,sf_arguments[i],loop_node_ids,&msize,f,0);
        
        // when we test a single path for strong fairness, we just need to exercise the nodes
        // and the transitions in the loop
        if (    msize >0
            && !check_if_loop_contains_transitions(f,loop_atoms,loop_size,matching_transitions,msize)
            && enabled_in_some_node1(f,loop_atoms,loop_size,matching_transitions,msize)) {
            
            sf_loop_result = 0;
            ltl_free(matching_transitions);
            break;
        }
        if (msize>0) ltl_free(matching_transitions);
    }
    ltl_free(sf_arguments);
    return sf_loop_result;

}

int check_loop_strong_fairness(ltl_formula *f, ltlatom **loop_atoms, int loop_size, SP_term_ref loop_node_ids, SP_pred_ref trans_ids_predicate) {
    
    int sf_size,sf_args_size,i,j,msize;
    int  result = 0;
    SP_term_ref *sf_assumptions = prolog_list_to_array(f->strong_fairness_arguments,&sf_size);
    SP_term_ref *sf_arguments;
    SP_integer *matching_transitions;

    int sf_result = 1;
    if (sf_size>0) {
        for (i=0; i<sf_size; i++) {
            sf_arguments = prolog_list_to_array(sf_assumptions[i],&sf_args_size);
            for (j=0; j<sf_args_size; j++) {
                matching_transitions = get_matching_transitions(trans_ids_predicate,sf_arguments[j],loop_node_ids,&msize,f,0);
                if (    msize >0
                    && !check_if_loop_contains_transitions(f,loop_atoms,loop_size,matching_transitions,msize)
                    && enabled_in_some_node1(f, loop_atoms, loop_size, matching_transitions, msize)) {
                    sf_result = 0;
                    // loop not strongly fair
                    ltl_free(matching_transitions);
                    break;
                }
                if (msize>0) ltl_free(matching_transitions);
            }
            if (sf_result) {
                result = 1;
                ltl_free(sf_arguments);
                break;
            }
            ltl_free(sf_arguments);
        }
    } else {
        result = 1;
    }
    ltl_free(sf_assumptions);
    return result;
}

int check_loop_weak_and_strong_fairness(ltl_formula *f, ltlatom **loop_atoms, int loop_size, SP_term_ref loop_node_ids, SP_term_ref wf_assumption, SP_pred_ref trans_ids_predicate) {
    
    int i,msize;
    int wf_loop_result = 1;
    SP_integer* matching_transitions;
    int wfargs_size;
    SP_term_ref *wf_arguments = prolog_list_to_array(wf_assumption,&wfargs_size);
    
    for (i=0; i<wfargs_size; i++) {
        matching_transitions = get_matching_transitions(trans_ids_predicate,wf_arguments[i],loop_node_ids,&msize,f,0);
        
        // when we test a single path for weak fairness, we just need to exercise the nodes
        // and the transitions in the loop
        if (    msize >0
            && !check_if_loop_contains_transitions(f,loop_atoms,loop_size,matching_transitions,msize)
            && enabled_in_each_node1(f,loop_atoms,loop_size,matching_transitions,msize)) {
            
            wf_loop_result = 0;
            ltl_free(matching_transitions);
            break;
        }
        
        ltl_free(matching_transitions);
    }
    ltl_free(wf_arguments);
    
    if (!wf_loop_result) {
        return 0;
    } else {
        return check_loop_strong_fairness(f,loop_atoms,loop_size,loop_node_ids,trans_ids_predicate);
    }
}

int check_loop_fairness_all(ltl_formula *f, ltlatom **loop_atoms, int loop_size, SP_term_ref loop_node_ids) {

    int result;
    SP_pred_ref predicate_enabled = f->get_enabled_actions_predicate;
    SP_term_ref enabled_actions = get_enabled_actions(predicate_enabled,loop_node_ids,f,0);
    
    SP_pred_ref callback_predicate = f->get_transition_ids_predicate;

    if (f->strong_fairness_all_operations) {
        result = check_loop_strong_fairness1(f,loop_atoms,loop_size,loop_node_ids,enabled_actions,callback_predicate);
    } else if (f->weak_fairness_all_operations) {
        result = check_loop_weak_and_strong_fairness(f,loop_atoms,loop_size,loop_node_ids,enabled_actions,callback_predicate);
    } else {
        fprintf(stderr, "Checking fainress for all operations is not enabled.\n");
        exit(1);
    }
    
    return result;
}

ltlatom **save_loop(ltlarray *bf_items, unsigned int solution, int size) {
    ltlatom **saved;
    int i;
    saved = (ltlatom **) ltl_malloc(sizeof(ltlatom*) * size);
    loopitem *item;
    unsigned int index = solution;

    item = (loopitem*) get_array_element(bf_items, solution);

    for (i=0; i<size; i++) {
        item = (loopitem*) get_array_element(bf_items, index);
        saved[size-1-i] = item->atom;
        index = item->previous;
    }

    return saved;
}

int is_fair_loop(ltl_formula *f, ltlarray *bf_items, unsigned int solution, int size) {
    
    printf("call is_fair_loop\n");
  
    SP_term_ref loop_node_ids = create_loop_nodes_list(bf_items,solution,size);
    print_scc_state_ids(loop_node_ids);
    // In addition to guaranteeing that the SCC of the tarjan search is fair,
    // we also need to assure that the found loop is also fair in regard to the fairness conditions

    // the loop legth does not correspond to the number of elements in bf_items
    int loop_size = size +1;
    int result;

    ltlatom **loop_atoms = save_loop(bf_items,solution,loop_size);

    if (f->strong_fairness_all_operations || f->weak_fairness_all_operations) {
        result = check_loop_fairness_all(f,loop_atoms,loop_size,loop_node_ids);
    } else {
        int wf_size;
        SP_term_ref *wf_assumptions = prolog_list_to_array(f->weak_fairness_arguments,&wf_size);
        
        if (wf_size>0) {
            result = 0;
            int i;
            for (i=0; i<wf_size; i++) {
                if (check_loop_weak_and_strong_fairness(f,loop_atoms,loop_size,loop_node_ids,wf_assumptions[i],f->get_transition_ids_predicate)) {
                    result = 1;
                    break;
                }
            }
        } else {
            result = check_loop_strong_fairness(f,loop_atoms,loop_size,loop_node_ids,f->get_transition_ids_predicate);
        }
        ltl_free(wf_assumptions);
    }
    ltl_free(loop_atoms);
    return result;
}
