//
//  tarjan.h
//  
//
//  Created by Ivaylo Dobrikov on 12/10/13.
//
//

#ifndef _tarjan_h
#define _tarjan_h

#include "ltlc.h"

typedef struct {
    ltlatom *from;
    int iterator;
    int lowlink;
} tarjanstate;

typedef struct {
    ltl_formula *f;
    
    int current_number;
    
    ltlarray scc_stack;
    int scc_stacksize;
    
    ltlarray rec_stack;
    int rec_stacksize;
    
    bitfield *scccheck_bf_used;
    bitfield *scccheck_bf_goal;
    
    int checked_transitions;
} tarjandata;

void delete_tarjan_data(tarjandata *t);
void init_tarjan_data(tarjandata *t, ltl_formula *f);
ltlatom *peek_stack(tarjandata *t, int pos);
void push_to_stack(tarjandata *t, ltlatom *atom, int set_on_tarjan_stack);
void push_tarjan_stack(tarjandata *s, ltlatom *from, int i, int lowlink);
ltlatom *peek_tarjan_stack(tarjandata *s, int *i, int *lowlink, int pos);
int check_all_until_goals_present(tarjandata *t, int scc_size);
int is_nontrivial_scc(tarjandata *t, int scc_size);
SP_term_ref get_node_id(ltlnode *node);
SP_term_ref get_scc_as_list(tarjandata *t, int scc_size);
SP_integer* extract_transids_from_prolog_list(SP_term_ref list, int *rsize);
int check_if_scc_contains_transitions(tarjandata *t, int scc_size, SP_integer *matching_transitions, int msize);
void extract_unfair_atoms(tarjandata *t, int scc_size, SP_integer *matching_transitions, int msize, ltlatom ***runfair, int *rsize);
int enabled_in_each_atom(tarjandata *t, int scc_size, SP_integer *matching_transitions, int msize);
SP_term_ref create_scc_nodes_list(tarjandata *t, int scc_size);
ltlatom **save_stack(tarjandata *t, int size, ltlatom **entry, int sort_scc, int debug);
void sortscc(int atomnumber, ltlatom **scc);

IFLTLDEBUG( void ltldebug_store_scc(tarjandata *t, unsigned int size); );

#endif
