% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(myset, [myset_init/0,
  myset_reset/0, myset_add_node/1, myset_delete_node/1,
  myset_size/1,
  myset_empty/1, 

  test/0]).


/* the interface to a heap global variable stored and maintained in C */

foreign(myset_initialize,myset_initialize). % initializes the data structure
foreign(myset_reset,myset_reset). % resets the heap to the empty heap; frees memory
foreign(myset_add_node,myset_add_node(+integer)).  % ID
foreign(myset_delete_node,myset_delete_node(+integer)).  % ID
foreign(myset_node_exists,myset_node_exists(+integer,[-integer])).
foreign(myset_size,myset_size([-integer])).
foreign(myset_empty,myset_empty([-integer])). %return 1 if empty, 0 if non-empty


foreign_resource(myset,
  [
   myset_reset
  ,myset_initialize
  ,myset_add_node
  ,myset_delete_node
  ,myset_node_exists
  ,myset_size
  ,myset_empty
  
  ]).
  
  
  :- dynamic loaded/0.

myset_init :- loadfr, %print(initializing_myset),nl,
               myset_initialize.

loadfr :-
  (loaded -> true
   ; (assert(loaded), 
      %assert_dir, 
      %assert(user:library_directory(prob_home('.'))),
      %print(loading_foreign_resource(myset)),nl,
      load_foreign_resource(library(myset)))
   ).

assert_dir :- (user:library_directory('.') -> true ; assert(user:library_directory('.'))).

myset_isempty :- myset_empty(X),X=1.

test :-  assert(user:library_directory('.')),
         myset_init,
         myset_add_node(22),
         myset_add_node(33),
         myset_add_node(221),
         myset_size(Sz),
         print(size(Sz)),nl, flush_output(user),
         myset_delete_node(33),
         myset_delete_node(33),
         myset_size(Sz2),
         print(size(Sz2)),nl, flush_output(user),
         myset_node_exists(22,X22),
         myset_node_exists(33,X33),
         format('Exists 22:~w, 33:~w~n',[X22,X33]).
   
 bench :- bench(1000000).
 bench(X) :- assert(user:library_directory('.')), myset_init,
         statistics(runtime,_),
         add(X),
         print_time(myset_add),
         ass(X),
         print_time(assert),
         del(X),
         print_time(myset_del),
         ret(X),
         print_time(retract).
/*
myset_reset(); 
myset_add : 670 ms
assert : 4920 ms
myset_del : 490 ms
retract : 4050 ms
*/

add(X) :- X>0,!, myset_add_node(X), X1 is X-1, add(X1).
add(_).
del(X) :- X>0,!, myset_delete_node(X), X1 is X-1, del(X1).
del(_).
:- dynamic factset/1.
ass(X) :- X>0,!, assert(factset(X)), X1 is X-1, ass(X1).
ass(_).
ret(X) :- X>0,!, (retract(factset(X)) -> true ; print(fail(X)),nl), X1 is X-1, ret(X1).
ret(_).

 print_time(T) :- statistics(runtime,[_,T2]), print(T), print(' : '), print(T2), print(' ms'),nl.