#!/usr/bin/env bash

# Auto generate single AllTests file for CuTest.
# Searches through all *.c files in the current directory.
# Prints to stdout.
# Author: Asim Jalis
# Date: 01/08/2003

if test $# -eq 0 ; then FILES=*.c ; else FILES=$* ; fi

echo '

/* This is auto-generated code. Edit at your own peril. */
#include <stdio.h>
#include <stdlib.h>

#include "../include/zmq.h"
#include "../include/czmq.h"
#include "../master/mprint.h"

#include "CuTest.h"

int port = 5000;

'

cat $FILES | grep '^void Test' | 
    sed -e 's/(.*$//' \
        -e 's/$/(CuTest*);/' \
        -e 's/^/extern /'

echo \
'

zctx_t *ctx;
void *global_printer;

static void uberinit(void) {
    ctx = zctx_new();
    global_printer = zsocket_new(ctx, ZMQ_ROUTER);
    zsocket_bind(global_printer, "tcp://localhost:4999");
}
static void uberteardown(void) {
    zctx_destroy(&ctx);
}

void RunAllTests(void) 
{
    CuString *output = CuStringNew();
    CuSuite* suite = CuSuiteNew();

'
cat $FILES | grep '^void Test' | 
    sed -e 's/^void //' \
        -e 's/(.*$//' \
        -e 's/^/    SUITE_ADD_TEST(suite, /' \
        -e 's/$/);/'

echo \
'
    CuSuiteRun(suite);
    CuSuiteSummary(suite, output);
    CuSuiteDetails(suite, output);
    printf("%s\n", output->buffer);
    CuStringDelete(output);
    CuSuiteDelete(suite);
}

int main(void)
{
    uberinit();
    RunAllTests();
    uberteardown();
    return 0;
}
'
