#include <unistd.h>

#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../common/queue.h"
#include "../worker/wstats.h"


extern zctx_t *ctx;
static void *tmp_socket;
extern wQueue *local_queue;
extern int port;

static void init(void) {
    init_send_stats(ctx, "localhost", "5", port-4);
    tmp_socket = zsocket_new(ctx, ZMQ_PULL);
    zsocket_bind (tmp_socket, "tcp://*:%d", port);
    local_queue = init_queue(0);
}

static void teardown(void) {
    teardown_queue(&local_queue);
    teardown_send_stats(ctx);
    zsocket_destroy(ctx, tmp_socket);
}

void Test_full_result(CuTest *tc) {
    init();
    int i = 0;
    for (i = 0; i < 10; i++) {
        start_timer();
        unsigned int usecs = 1000;
        usleep(usecs); // working hard
        stop_timer();
        send_statistics("foobartest7777777777");

        int rc = zsocket_poll(tmp_socket, 5000);
        CuAssertTrue(tc, rc);
        zmsg_t *msg = zmsg_recv(tmp_socket);
        CuAssertIntEquals(tc, 5, zmsg_size(msg));
    
        char *str = zmsg_popstr(msg);
        CuAssertStrEquals(tc, "5 0", str);
        free(str);
    
        str = zmsg_popstr(msg);
        CuAssertStrEquals(tc, "foobartest7777777777", str);
        free(str);
    
        str = zmsg_popstr(msg);
        float f;
        CuAssertIntEquals(tc, 1, sscanf(str, "%f", &f));
        CuAssertTrue(tc, f > 1);
        free(str);
        
        str = zmsg_popstr(msg);
        int x;
        CuAssertIntEquals(tc, 1, sscanf(str, "%d", &x));
        CuAssertTrue(tc, x);
        free(str);

        str = zmsg_popstr(msg);
        CuAssertIntEquals(tc, 1, sscanf(str, "%d", &x));
        CuAssertIntEquals(tc, i, x);
        free(str);
        zmsg_destroy(&msg);
    }
    teardown();
}

void Test_small_package(CuTest *tc) {
    init();
    int i;
    for (i = 0; i < 10; i++) {
        send_statistics("foobartest7777777777");

        int rc = zsocket_poll(tmp_socket, 5000);
        CuAssertTrue(tc, rc);
        zmsg_t *msg = zmsg_recv(tmp_socket);
        CuAssertIntEquals(tc, 1, zmsg_size(msg));
    
        char *str = zmsg_popstr(msg);
        CuAssertStrEquals(tc, "5 0", str);
        free(str);
        zmsg_destroy(&msg);
    }
    teardown();
}

CuSuite *WStatsGetSuite() {
    CuSuite *suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, Test_full_result);
//    SUITE_ADD_TEST(suite, Test_small_package);
    return suite;
}
