#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <sicstus/sicstus.h>

#include "../../../include/zmq.h"
#include "../../../include/czmq.h"

#include "../../../common/hashtrie.h"
#include "../../../common/shm_common.h"
#include "../../../common/shm_cln.h"
#include "../../../common/ring.h"
#include "../../../common/ringcovery.h"
#include "../../../common/proxy-connector.h"

#include "wid.h"
#include "whash_send.h"
#include "wtok.h"
#include "wwork.h"
#include "wctrl.h"
#include "wstats.h"
#include "wdebug.h"
#include "worker.h"
#include "wbeat.h"

static SP_pred_ref init_predicate(const char *name, int arity, const char *module);
static void init(char *masteraddr, int port, zctx_t **ctx, int strategy, int lvl, int recover);
static void init_modules(char *masteraddr, int port, zctx_t *ctx, int lvl, int recover);
static void teardown_worker(zctx_t *ctx, int strategy);
static void load_model(char *model);
static int get_strategy(char *mode);

wQueue *local_queue = NULL;

char *worker_id;

// TODO: recover flag
int worker_main(char *masteraddr, int port, char *mode, int debug_level, int recover) {
    int strategy = get_strategy(mode);
    zctx_t *ctx;
    init(masteraddr, port, &ctx, strategy, debug_level, recover);
    teardown_worker(ctx, strategy);
    return 0;
}

static void init(char *masteraddr, int port, zctx_t **ctx, int strategy, int lvl, int recover) {
    *ctx = zctx_new();
    connect_to_proxy(*ctx);

#ifndef TMPFOLDER
#define TMPFOLDER "/var/tmp/"
#endif
    if (recover) {
        local_queue = recover_queues(TMPFOLDER, *ctx, masteraddr, port);
    } else {
        local_queue = init_queue(0);
    }
    init_shm(*ctx);

    init_modules(masteraddr, port, *ctx, lvl, recover);
}


static void init_modules(char *masteraddr, int port, zctx_t *ctx, int lvl, int recover) {
    init_id(ctx, masteraddr, port);
    struct Preferences *p = get_id(ctx, masteraddr, port);
    init_worker_printer(ctx, lvl, masteraddr, port, p->id);
    worker_id = malloc(strlen(p->id) + 1);
    strcpy(worker_id, p->id);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE, , ,":enter :printer");
    load_model(p->model);

    zmq_pollitem_t poller_work_recv;
    //zmq_pollitem_t poller_send_work;
    init_work_handler(ctx, &poller_work_recv,
                      /*&poller_send_work,*/ masteraddr, p->id, port);

    zmq_pollitem_t poller_recv_ctrl;
    SP_pred_ref stats_pred = init_predicate("print_stats",1,"worker");
    init_recv_control(ctx, &poller_recv_ctrl, masteraddr, port, stats_pred);

    init_send_hashes(ctx, masteraddr, port);
    init_send_stats(ctx, masteraddr, p->id, port);

    SP_pred_ref check_pred_ref = init_predicate("zmq_check",1,"worker");
    zmq_pollitem_t poller_tick;
    init_tok(ctx, &poller_tick, check_pred_ref);

    if (recover) {
        tick();
    }

    DEBUG_PRINT(DEBUG_VERBOSE,,,":msg :ready");
    //init_wbeat();
    signal_ready(p->id);

    teardown_id(ctx);
    free(p->model);
    free(p->id);
    free(p);

    zloop_t *reactor = zloop_new();

    zloop_poller(reactor, &poller_tick, h_tick, NULL);
    zloop_poller(reactor, &poller_work_recv, h_work_recv, NULL);
    zloop_poller(reactor, &poller_recv_ctrl, h_control, NULL);
    //zloop_timer(reactor, TIMEOUT_CONSTANT * 1000, 0, send_heartbeat, NULL);

    zloop_start(reactor);
    zloop_destroy(&reactor);
    free(worker_id);
}

static void teardown_worker(zctx_t *ctx, int strategy) {
    //teardown_wbeat();
    teardown_recv_control(ctx);
    teardown_send_hashes(ctx);
    teardown_work_handler(ctx);
    teardown_tok(ctx);
    stop_worker_printer(ctx);
    zctx_destroy(&ctx);
    teardown_hashtrie();
    teardown_queue(&local_queue);
}

static void load_model(char *model) {
    SP_pred_ref pred_load_model = init_predicate("load",1,"worker");
    SP_term_ref val = SP_new_term_ref(); 
    SP_put_address(val, model);
    int status = SP_query_cut_fail(pred_load_model,val);
    assert(status == SP_SUCCESS);   
}

static int get_strategy(char *mode) {
    if (strcmp(mode, "mixed") == 0) {
        return 50;
    } else if (strcmp(mode, "breadth_first") == 0) {
        return 0;
    } else {
        return 100;
    }
}

static SP_pred_ref init_predicate(const char *name, int arity, const char *module) {
    SP_pred_ref pred = SP_predicate(name,arity,module);
    if (pred == NULL) {
        fprintf(stderr, "Unable to find callback module %s:%s/%i for zmq worker, exiting.\n",
        module, name, arity);
        exit(1);
    }
    return pred;
}
