% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% bmachine provides access to the currently loaded B-machine and all its subsidiary machines

:- module(bmachine,[bmachine_is_precompiled/0,
                    b_set_initial_machine/0,
                    b_set_empty_machine/0,
                    b_set_machine/3,
                    b_set_typed_machine/1,
                    b_load_eventb_project/1, % used for loading .eventb files
                    b_set_eventb_project_flat/3, % set an Event-B project from terms

                    full_b_machine/1,

                    get_proven_invariant/2,
                    load_additional_information/1,
                    %% generate_specialized_invariants/0, %% no longer exists
                    get_operation_info/2,

                    b_load_machine_from_file/1, b_load_machine_probfile/1,
                    b_load_machine_from_list_of_facts/2,
                    b_load_machine_from_term/2,
                    b_machine_precompile/0,  other_spec_precompile/0,
                    b_machine_reset/0,

                    b_machine_name/1,

                    b_get_all_used_filenames/1, b_get_main_filename/1,
                    b_get_main_filenumber/1, b_filenumber/4,
                    b_get_animated_sections/1,

                    b_get_machine_set/1,
                    b_get_named_machine_set/2,
                    b_get_disjoint_constants_of_type/3, b_get_constant_represented_inside_global_set/2,
                    b_get_machine_constants/1, b_get_machine_all_constants/1,
                    b_machine_has_constants/0, b_machine_has_constants_or_properties/0,
                    b_is_constant/1, b_is_unused_constant/1,
                    b_machine_has_variables/0,
                    b_is_variable/1, b_is_variable/2,
                    b_get_machine_variables/1,
                    b_get_all_used_identifiers/1, b_get_all_used_identifiers_in_section/2,
                    source_code_for_identifier/6,

                    b_get_properties_from_machine/1,
                    b_get_invariant_from_machine/1,
                    b_get_linking_invariant_from_machine/1,
                    b_machine_has_assertions/0,
                    b_get_static_assertions_from_machine/1, b_machine_has_static_assertions/0,
                    b_get_unproven_static_assertions_from_machine/1,
                    b_get_dynamic_assertions_from_machine/1, b_machine_has_dynamic_assertions/0,
                    b_get_unproven_dynamic_assertions_from_machine/1,
                    b_get_assertions_from_main_machine/2, b_main_machine_has_no_assertions/0,
                    b_machine_has_unproven_assertions/0, get_assertions_from_machine/2,
                    b_machine_has_proven_invariants/0,
                    b_get_assertions/3,
                    b_get_assertion_count/3,
                    is_discharged_assertion/1,

                    b_get_initialisation_from_machine/2,
                    b_get_machine_operation/4, b_get_machine_operation/5,
                    b_get_promoted_machine_operations/1,
                    b_top_level_operation/1, b_is_operation_name/1,
                    b_get_machine_operation_parameter_types/2,
                    b_get_machine_operation_parameter_names/2,
                    b_machine_operation_names_in_reverse_order/1,

                    b_get_machine_setscope/2,
                    b_get_machine_goal/1,
                    b_set_machine_goal/1, b_set_machine_goal/2,
                    b_set_machine_searchscope/1, b_set_machine_searchscope/2,
                    b_set_parsed_typed_machine_goal/1,
                    b_reset_machine_goal_from_DEFINITIONS/0,
                    b_parse_machine_expression_from_codes/2,
                    b_parse_machine_expression_from_codes_with_prob_ids/2,
                    b_parse_machine_expression_from_codes/5,
                    b_parse_machine_expression_from_codes/6,
                    b_parse_machine_predicate/2, b_parse_machine_predicate/3,
                    b_parse_optional_machine_predicate/2,
                    b_parse_machine_predicate_from_codes/3,
                    b_parse_machine_predicate_from_codes_open/5,
                    b_get_machine_searchscope/1,
                    b_get_machine_animation_function/2,
                    b_get_machine_heuristic_function/1,
                    b_get_machine_custom_edges_function/1,
                    b_get_machine_custom_nodes_function/1,
                    b_machine_temp_predicate/1, set_temp_predicate/2, assert_temp_typed_predicate/1,
                    reset_temp_predicate/0,

                    get_animation_image/2,
                    b_get_definition/5,
                    b_get_definition_string_from_machine/2,
                    b_definition_prefixed/4,
                    b_get_typed_definition/3, b_get_typed_predicate_definition/3,
                    b_get_typed_expression_definition/3, b_get_true_expression_definition/1,

                    b_specialized_invariant_for_op/2, b_operation_preserves_full_invariant/1,
                    tcltk_get_specialized_invariant_for_op/2, tcltk_get_specialized_invariants_for_ops/1,
                    b_get_operation_normalized_read_write_info/3, % used to be b_normalized_rwsets_for_op/3,
                    reuse_operation_effect/2, candidate_operation_for_reuse/1,
                    b_operation_cannot_modify_state/1,

                    b_type_expression/5,b_type_expressions/5,
                    b_type_open_predicate/5,b_type_open_predicate_with_errors/4,
                    b_type_open_exists_predicate/3,
                    type_with_errors/4,

                    b_show_machine_representation/3,  b_write_machine_representation_to_file/2,
                    b_write_eventb_machine_to_classicalb_to_file/1,
                    b_show_eventb_as_classicalb/2,

                    b_get_machine_refinement_hierarchy/1,b_get_refined_machine/1,
                    b_machine_statistics/2,

                    % More Proof Information:
                    discharged_guard_strengthening/4,

                    % Flow information
                    wp_untyped/3, clear_wp/0,
                    nonchanging_guard/2,
                    typecheck_predicates/4
                   ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(bmachine_construction,[check_machine/4,type_in_machine_l/6,type_open_predicate_with_quantifier/6]).
:- use_module(bsyntaxtree).
:- use_module(bmachine_structure,[get_section/3,write_section/4,get_section_texprs/3]).
:- use_module(parsercall).
:- use_module(tools).
:- use_module(translate,[translate_machine/3,translate_eventb_to_classicalb/3]).
:- use_module(bmachine_eventb,[check_event_b_project/4,is_eventb_additional_info/1]).
:- use_module(kernel_freetypes,[reset_freetypes/0,add_freetype/2]).
:- use_module(b_machine_hierarchy).
:- use_module('kodkod/kodkod', [replace_by_kodkod/3]).

:- use_module(b_global_sets,[find_inequal_global_set_identifiers/4,
                             b_get_prob_deferred_set_elements/2]).

% for applying transformations on the machines
:- use_module(record_detection,[replace_sets_by_records/2]).
:- use_module(partition_detection,[detect_partitions/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides access to the various parts of the loaded B machine.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% bmachine: module for storing the currently loaded machine
%   Contains predicates for getting (already typed) information about
%   different parts of the machine
%  many predicates are precompiled by b_precompile_machine/0, see below for
%  details.

:- volatile bmachine_is_precompiled/0.
:- dynamic bmachine_is_precompiled/0.  /* TRUE if the precompiled predicates have been set */

assert_bmachine_is_precompiled :-
   (bmachine_is_precompiled -> true ; assert(bmachine_is_precompiled), debug_println(4,bmachine_is_precompiled)).

:- volatile machine/2.
:- dynamic machine/2.

get_machine(Name,M) :- (machine(Name,M) -> true; add_error(bmachine,'No B machine loaded'),fail).

full_b_machine(machine(Name,M)) :- get_machine(Name,M).

b_set_machine(Main,Machines,Errors) :-
    b_machine_reset,
    ~~mnf( check_machine(Main,Machines,M1,Errors1) ),!,
    apply_machine_transformations(M1,ResultMachine),
    ( no_real_perror_occurred(Errors1) ->
        %% get_memory_used(M1), %%
        assert_machine(ResultMachine)
       %%, get_memory_used(M2),print_memory_used(M2),  print_memory_used_difference(M1,M2),nl %%
    ; otherwise ->
        true),
    Errors=Errors1.

assert_machine(X) :- nonvar(X), X=machine(A,B),!,assert(machine(A,B)). %, print(machine_asserted),nl.
assert_machine(X) :- add_internal_error('Illegal format: ',assert_machine(X)),fail.

:- meta_predicate apply_transformation_step(*,2,*,*).
% apply some transformations on the machine that require global information
% (other local transformations can be found in ast_cleanup)
apply_machine_transformations -->
    apply_transformation_step('record definition detection', replace_sets_by_records),
    apply_transformation_step('partition detection', detect_partitions).

apply_transformation_step(Name,Pred,In,Out) :-
    ( call(Pred, In, Out) -> true
    ; otherwise ->
        ajoin(['applying ', Name, ' failed, skipping.'], Msg),
        add_error(bmachine,Msg),
        In=Out).

% try to translate (parts of) the properties to Kodkod
% if it succeeds, replace the translated predicates by
% calls to the Kodkod process
try_kodkod :-
    get_preference(try_kodkod_on_load,true),
    b_get_machine_constants(Constants),
    Constants = [_|_],!,
    b_get_properties_from_machine(OldPredicate),
    ( replace_by_kodkod(Constants,OldPredicate,NewPredicate) ->
        retract(machine(N,OldMachine)),
        retractall( b_get_properties_from_machine(_) ),
        write_section(properties,NewPredicate,machine(N,OldMachine),NewMachine),
        assert( NewMachine ),
        assert( b_get_properties_from_machine(NewPredicate) )
    ; otherwise ->
        true).
try_kodkod.

% set a machine term as current machine, without calling type checking
% (used for Z translations)
b_set_typed_machine(M1) :-
    b_machine_reset,
    apply_machine_transformations(M1,ResultMachine),
    assert_machine(ResultMachine).

b_set_eventb_project_flat(Models,Contextes,Proofs) :-
    b_machine_reset,
    analyse_eventb_hierarchy(Models,Contextes),
    ( check_event_b_project(Models,Contextes,Proofs,M1) ->
        apply_machine_transformations(M1,ResultMachine),
        assert_machine(ResultMachine) % has the form machine/2
    ; otherwise ->
        add_internal_error('Loading Event-B Model failed; Please submit bug report.',b_set_eventb_project_flat), %,project(Models,Contextes)),
        fail).


:- volatile package/1, stored_preference/2.
:- dynamic package/1, stored_preference/2.
clear_package :- retractall(package(_)),retractall(stored_preference(_,_)).

b_load_eventb_project(Filename) :-
  clear_package,
  consult_without_redefine_warning(Filename),
  (package(load_event_b_project(Machines,Contexts,Proofs,Errors))
    -> true
    ; package(load_event_b_project(Machines,Contexts,Errors))
      -> Proofs=[], print('% Deprecated EventB Project without Proof Info.'),nl
      ; add_error_and_fail(b_load_eventb_project,'This is not an EventB Project file:',Filename)
  ),
  (Errors=[] -> true ; add_error(b_load_eventb_project,'Errors in Event-B Package: ',Errors)),
  read_stored_prefs,
   b_set_eventb_project_flat(Machines,Contexts,Proofs),
   clear_wp,
   (load_additional_information(Proofs) -> true ; add_error(b_load_eventb_project,'Failed to load proof information',Proofs)).
%   generate_specialized_invariants.

read_stored_prefs :- stored_preference(Pref,Val), print(stored_preference(Pref,Val)),nl,
      preferences:set_preference(Pref,Val),fail.
read_stored_prefs.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Collect Proof Information from Event B

:- volatile discharged/3.
:- volatile discharged_theorem/3, wd_po/3, wp_untyped/3.
:- volatile nonchanging_guard/2, discharged_guard_strengthening/4.
:- dynamic discharged/3.
% discharged(Model,Event,Name) : invariant Name in model Model is preserved by Event (no matter which Event at which refinement level)
:- dynamic discharged_theorem/3.
:- dynamic wd_po/3.
:- dynamic wp_untyped/3.
:- dynamic nonchanging_guard/2.
:- dynamic discharged_guard_strengthening/4.

:- volatile b_specialized_invariant_for_op/2.
:- dynamic b_specialized_invariant_for_op/2.
:- volatile get_proven_invariant/2.
:- dynamic get_proven_invariant/2.



clear_wp :-
    retractall( wp_untyped(_,_,_) ).
    %retractall( wp_typed(_,_,_) ). only in flow.pl

load_additional_information([]).
load_additional_information([H|T]) :-
      load_additional_information_fact(H),!,
      load_additional_information(T).
load_additional_information([H|T]) :-
      add_internal_error('Unknown additional info: ',load_additional_information_fact(H)),
      load_additional_information(T).

% old style proof info
load_additional_information_fact(discharged(Machine,Theorem)) :- % a discharged theorem
      debug_println(9,discharged_theorem(Machine,old_style_proof_info,Theorem)),
      assert(discharged_theorem(Machine,invariant,Theorem)).
load_additional_information_fact(discharged(Machine,Event,Invariant)) :-
      debug_println(9,discharged(Machine,Event,Invariant)),
      assert(discharged(Machine,Event,Invariant)).
load_additional_information_fact(Term) :- % theories are handled in bmachine_eventb.pl already
    is_eventb_additional_info(Term).

% new style proof info po(Machine,Info,SourceList,Discharged)

load_additional_information_fact(po(Machine,'Invariant  preservation',Source,true)) :- % invariant preservation
      add_discharged_po_infos(Source,Machine).

load_additional_information_fact(po(Machine,'Invariant  establishment',Source,true)) :- % invariant establishment
      add_discharged_po_infos(Source,Machine).

load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % invariant or theorem wd
      ( Text = 'Well-definedness of Invariant'
       ; Text = 'Well-definedness of Theorem'),
      member(invariant(Invariant),Source),
      debug_println(9,wd_po(Machine,Invariant,Proven)),
      assert(wd_po(Machine,Invariant,Proven)).
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % axiom or axiom theorem wd
      ( Text = 'Well-definedness of Axiom'
       ; Text = 'Well-definedness of Theorem'),
      member(axiom(Axiom),Source),
      debug_println(9,wd_axiom(Machine,Axiom,Proven)).   % TO DO: assert and use
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % proven theorem as guard
      Text = 'Well-definedness of Theorem',
      member(guard(Grd),Source),
      member(event(Evt),Source),
      debug_println(9,wd_axiom(Machine,Evt/Grd,Proven)).   % TO DO: assert and use

load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % wd of action
      Text = 'Well-definedness of action',
      member(action(ACT),Source), % probably the guard in the last event with discharged info ??
      debug_println(9,wd_action(Machine,ACT,Proven)).   % TO DO: assert and use
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % wd of guard
      Text = 'Well-definedness of Guard',
      member(guard(GRD),Source), % probably the guard in the last event with discharged info ??
      debug_println(9,wd_guard(Machine,GRD,Proven)).  % TO DO: assert and use

load_additional_information_fact(po(Machine,'Theorem',[invariant(Theorem)],true)) :- % a theorem
      debug_println(9,discharged_theorem(Machine,invariant,Theorem)),
      assert(discharged_theorem(Machine,invariant,Theorem)).
load_additional_information_fact(po(Machine,'Theorem',[axiom(Theorem)],true)) :- % a theorem
      debug_println(9,discharged_theorem(Machine,axiom,Theorem)),
      assert(discharged_theorem(Machine,axiom,Theorem)).
load_additional_information_fact(po(Machine,'Theorem',[guard(GuardTheorem),event(Event)],true)) :- % a theorem in a guard
      debug_println(9,discharged_theorem(Machine,guard(Event),GuardTheorem)),
      assert(discharged_theorem(Machine,guard(Event),GuardTheorem)).

% stuff we don't yet care about
load_additional_information_fact(po(Machine,Text,Infos,Proven)) :-
      ( Text = 'Guard strengthening (split)'
      ; Text = 'Guard strengthening (merge)'
      ; Text = 'Guard strengthening' ),
      !,
      ((Proven=true, Infos = [event(AbsEvent),guard(Guard),event(ConcreteEvent)])
        -> assert(discharged_guard_strengthening(Machine,AbsEvent,Guard,ConcreteEvent)),
           print(discharged_guard_strengthening(Machine,AbsEvent,Guard,ConcreteEvent)),nl
        ;  true
      ).
load_additional_information_fact(po(_,'Variant of event',_,_)).
load_additional_information_fact(po(_,'Natural number variant of event',_,_)).
load_additional_information_fact(po(_,'Action simulation',_,_)).
load_additional_information_fact(po(_,'Feasibility of witness',_,_)).
load_additional_information_fact(po(_,'Feasibility of action',_,_)).
load_additional_information_fact(po(_,'Well-definedness of witness',_,_)).
load_additional_information_fact(po(_,'Well-definedness of variant',_,_)).
load_additional_information_fact(po(_,'Finiteness of variant',_,_)).
load_additional_information_fact(po(_,'Equality of common variables',_,_)).

load_additional_information_fact(po(_Machine,_Kind,_Source,false)). % ignore everything that is not proven

load_additional_information_fact(po(_Machine,_Kind,_Source,reviewed)). % ignore everything that is only reviewed

% ignore exporter version for now
% we are compatible to both versions 2 and 3 (added abstract constants)
load_additional_information_fact(exporter_version(V)) :- V=2 ; V=3.

% ignored because the pragmas are attached to the machine in load_event_b_project/4
load_additional_information_fact(pragma(_Type,_Source,_AttachedTo,_Content)).

load_additional_information_fact(wp(S,D,P)) :-
      assert( wp_untyped(S,D,P) ).

load_additional_information_fact(nonchanging_guard(E,P)) :-
      typecheck_predicates(P,TypedPred,E,_),
      assert(nonchanging_guard(E,TypedPred)).

load_additional_information_fact(invariant(_I)). %ignore

typecheck_predicates([],[],_,_).
typecheck_predicates([H|T],[TH|TT],S,D) :-
   (S='INITIALISATION' -> Scope1 = []; Scope1=[operation(S)]),
   ((D='INITIALISATION';var(D)) -> Scope2 = Scope1; Scope2=[operation(D)|Scope1]),
   Scope = [variables|Scope2],
   b_type_expression(H,Scope, _, TH, _),
   typecheck_predicates(T,TT,S,D).


add_discharged_po_infos(Source,Machine) :-
  % Source = [event(A1),...,event(CurMachineEvent),invariant(Inv)]
  % note that CurMachineEvent could be equal to A1; A1 can be split up into multiple events (one of which could have the same name as A1; see Ticket PROB-292)
      get_bottom_event(Source,Event),
      member(invariant(Invariant),Source),
      debug_println(9,discharged(Machine,Event,Invariant)),
      assert(discharged(Machine,Event,Invariant)),
      % TO DO: should we propagate info up to abstract events which are not split up or where all refinements preserve a certain invariant ??
      fail.
add_discharged_po_infos(_Source,_Machine).

% get the last event entry in proof info source list
get_bottom_event(Source,Event) :-
 (bottom_event(Source,'$NONE'(none),Event) -> true
   ; add_error(bmachine,'Could not determine event for proof info: ',Source),fail).
bottom_event([],Ev,Ev) :- Ev \= '$NONE'(none).
bottom_event([event(E)|T],_,Res) :- !,bottom_event(T,E,Res).
bottom_event([_|T],E,Res) :- bottom_event(T,E,Res).

:- use_module(debug).

%b_specialized_invariant_for_op_calc(Event,Invariant) :-
% b_get_invariant_from_machine(Invariant).

tcltk_get_specialized_invariants_for_ops(list(Res)) :-
   findall([Name,' INVARIANT ',TI, '\n', '  NON-TRIVIAL PROVEN:', PI, 'END;'],
            ((b_top_level_operation(Name) ; Name='$initialise_machine'),
              tcltk_get_specialized_invariant_for_op(Name,TI),
              tcltk_get_proven_invariant_for_op(Name,PI)), Res).
tcltk_get_specialized_invariant_for_op(Event,TI) :-
   (b_specialized_invariant_for_op(Event,Invariant) ->
    translate:translate_bexpression(Invariant,TI)
    ; TI = 'not_simplified').
tcltk_get_proven_invariant_for_op(Event,TI) :-
   (get_proven_invariant(Event,Invariant) ->
    translate:translate_bexpression(Invariant,TI)
    ; TI = 'none').

% use proof information to compute subset of invariant that needs to be checked for operations
:- public b_specialized_invariant_for_op_calc/2.
b_specialized_invariant_for_op_calc(Event,Invariant) :-
    preferences:get_preference(use_po,true),
    b_get_invariant_from_machine(I), % Note: we can have interference with CSE here, as lazy_let_pred at top-level can merge several conjuncts (but CSE is now disabled for the invariant if PROOF_INFO = TRUE)
    conjunction_to_list(I,IL),
    IL \= [], % otherwise: Invariant already trivial; no sense in keeping track of information
    get_op_or_init_and_syntax_filter(Event,ProofInfoEventName,IL,RIL),
    exclude(is_proven_invariant(ProofInfoEventName),RIL,RL), % Warning !! The invariants are not all from Model; the discharged information could be for a proven invariant with the same label from another model, see examples/EventBPrologPackages/ProofDirected/TestR_label_reuse.eventb; this should now properly be dealt with by filter_proven_invariants
    \+ (IL=RL),  /* something filtered out */
    debug_println(9,simplified(Event)),
    conjunct_predicates(RL,Invariant). %, print(simplified),nl,translate:print_bexpr(Invariant),nl.


get_op_or_init_and_syntax_filter('$initialise_machine','INITIALISATION',IL,IL). % INIT always sets all variables
get_op_or_init_and_syntax_filter(Event,Event,IL,RIL) :-
    b_get_machine_operation(Event,_,_,_),
    get_operation_info(Event,SubstitutionInfo),
    filter_syntactically_relevant_invariants(IL,SubstitutionInfo,RIL). % filter out all invariants not modified by Event

% check if an operation preserves the full invariant; according to proof information
b_operation_preserves_full_invariant(Event) :-
    b_specialized_invariant_for_op(Event,Invariant),
    is_truth(Invariant).

:- public get_proven_invariant_calc/2.
% computes proven, relevant invariants (not trivial ones which are unaffected by Event) for main model
get_proven_invariant_calc(Event, Invariant) :-
    preferences:get_preference(use_po,true),
    b_machine_name(Model),
    b_get_invariant_from_machine(I),
    conjunction_to_list(I,IL),
    get_operation_info(Event,SubstitutionInfo),
    ~~mnf(filter_syntactically_relevant_invariants(IL,SubstitutionInfo,RIL)),
    b_get_machine_operation(Event,_,_,_),
    ~~mnf(select_proven_invariants(RIL,Model,Event,RL)),
    conjunct_predicates(RL,Invariant).

% filter_syntactically_relevant_invariants(+InvariantsList, +Substitution, -RelevantItemsFromInvariantsList)
filter_syntactically_relevant_invariants(Invariants,SubstitutionInfo,Result) :-
    include(is_syntactically_relevant_invariant(SubstitutionInfo),Invariants,Result).
is_syntactically_relevant_invariant(SubstitutionInfo,Inv) :-
  find_identifier_uses(Inv,[],IDs),
  list_to_ord_set(IDs,I),
  ( ground(SubstitutionInfo), memberchk(modifies(ModID),SubstitutionInfo) ->
      list_to_ord_set(ModID,M),
      ord_intersect(I,M)
  ; otherwise -> true % no info; assume all variables modified
  ).


% filter_proven_invariants(+InvariantsList, +ModelName, +EventName, -RelevantItemsFromInvariantsList)
% this gets called during model checking time
is_proven_invariant(Event,E) :-
    % print('Filter Check: '), print(Event), print(' : '),translate:print_bexpr(E),nl,
   get_texpr_pos(E,POS),
   POS = rodinpos(Model,Name,_), % only accept new rodinpos information with three parameters !
   !,    % the old one with two parameters could lead to confusion as labels can be reused
   % TO DO: check that if E is not always well-defined that there are no unproven WD proof obligations !
  discharged_invariant(Model,Name,Event),
  check_wd_discharged_inv_or_thm(E,Model,Name,Event). %% print('DISCHARGED'(Model,Event,Name)),nl.

check_wd_discharged_inv_or_thm(E,_,_,_) :- always_well_defined(E),!.
check_wd_discharged_inv_or_thm(_,Model,Name,_Event) :- wd_po(Model,Name,ProofStatus),!,
    ProofStatus=true.
check_wd_discharged_inv_or_thm(_,Model,Name,Event) :-
    format('% Assuming invariant/theorem ~w is well-defined for event ~w:~w.~n',[Name,Model,Event]).

% select_proven_invariants(+InvariantsList, +ModelName, +EventName, -RelevantItemsFromInvariantsList)
select_proven_invariants(Invariants,Model,Event,Result) :-
   include(is_proven_invariant_for_model(Model,Event),Invariants,Result).

% Note: different from is_proven_invariant; TO DO: merge (here we assume main model)
is_proven_invariant_for_model(MainModel,Event,E) :- %% print(chck(MainModel,Event)),nl,
   get_rodin_name(E,Name), % if we do not have a rodin position we consider this as not proved
   (get_rodin_model_name(E,Model) -> true
     ; format('Could not get Rodin model name for invariant: ~w~n',[Name]),
       Model=MainModel,
       fail  % we have an old style .eventb file; safer to assume not proven as there can be clashes of labels
    ),
   discharged_invariant(Model,Name,Event).

is_discharged_assertion(Theorem) :-
   if( get_texpr_pos(Theorem,rodinpos(Model,Name,_)),
        (discharged_theorem(Model,_,Name), % does not matter if invariant or axiom
         check_wd_discharged_inv_or_thm(Theorem,Model,Name,assertion) ) % check that well-defined
      , (specfile:animation_minor_mode(eventb),
         print('Could not get position information for theorem: '),
         translate:print_bexpr(Theorem),nl,
         get_texpr_pos(Theorem,POS),print(get_texpr_pos(POS)),nl,
         fail)
      ).
% derived predicate
discharged_invariant(Model,InvariantName,Event) :-
  % InvariantName from model Model is preserved by Event and any of its refinements
  discharged(Model,Event,InvariantName).
discharged_invariant(Model,InvariantName,Event) :-
  % InvariantName from model Model is preserved by Event and any of its refinements
  also_discharged(Model,Event,InvariantName).

b_machine_has_proven_invariants :-
  (discharged(_,_,_) -> true).

% complete the discharged/3 information: propagating information from abstract events
% down to refined top-level events
:- public complete_discharged_info_calc/0.
complete_discharged_info_calc :- debug_println(4,'completing discharged info: '),
    statistics(runtime,[Start,_]),
    (complete_discharged_info2 -> true ; true),
    statistics(runtime,[Stop,_]), T is Stop-Start, debug_print(4,T), debug_println(4,' ms').
:- volatile discharged2/2, also_discharged/3.
:- dynamic discharged2/2.
:- dynamic also_discharged/3.
discharged_info_exists :- discharged2(_,_),!.
complete_discharged_info2 :- retractall(also_discharged(_,_,_)),
    retractall(discharged2(_,_)),
    discharged(Model,Event,_), % find out which Model and Events are used in discharged info
   \+ discharged2(Model,Event), %print(discharged(Model,Event)),nl,
   assert(discharged2(Model,Event)),fail.
complete_discharged_info2 :- discharged_info_exists,
   b_get_animated_sections(Sections),
   discharged2(Model,Event),   % Event in Model was proven to preserve an invariant in Model
   b_get_abstract_events(Event,RModel,AbstractEvents), % search for a refinement of Model where the same event exists
   RModel \= Model,
   % i.e., invariant Name was proven in an abstract Model for Event
   \+ is_abstract_event_in_list(Event,Model,AbstractEvents), % this event is *NOT* a refinement of the proven one
   debug_println(4,removing_discharged(Model,Event)),
   retractall(discharged(Model,Event,_Name)),
   member(model(Model),Sections),  % only generate error message if the Model is also animated using b_get_animated_sections, see example  EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb
   %  (otherwise the refinement may well exist but has been filtered out by restricting the animation levels)
   add_error(event_refinement,'An event does not refine an abstract event of the same name: ',Event:Model:RModel),
   % This will could cause errors in the proof information usage !
   %print(AbstractEvents),nl,
   fail.
complete_discharged_info2 :- discharged_info_exists,
   complete_discharged_info3(_Model,_Event,_Name).
complete_discharged_info3(InvNameModel,AbsEvent,Name) :-
   b_get_abstract_events(EventInRefinement,_RefModel,AbstrEvents),
   is_abstract_event_in_list(AbsEvent,InvNameModel,AbstrEvents), % Event in InvNameModel is an abstract event of EventInRefinement
   %print(check(InvNameModel,Event, EventInRefinement,RefModel)),nl,
   discharged(InvNameModel,AbsEvent,Name),  % print(chck2(Model,Event,Name)),nl,
   %InvNameModel:Name has been proven for AbsEvent *at all relevant* levels
   \+ discharged(InvNameModel,EventInRefinement,Name),
   \+ also_discharged(InvNameModel,EventInRefinement,Name),
   % this invariant has not yet been marked as proven for EventInRefinement
   debug_println(4,also_discharged_in_refinement(InvNameModel:Name,EventInRefinement,refined_from(AbsEvent))),
   assert(also_discharged(InvNameModel,EventInRefinement,Name)),
   fail.

%abstract_event_of_calc(TopLevelEvent,RModel,Event,Model) :-
%      b_get_abstract_events(TopLevelEvent,RModel,AbstrEvents),
%      is_abstract_event_in_list(Event,Model,AbstrEvents) , print(abs(TopLevelEvent,RModel,Event,Model)),nl.

is_abstract_event_in_list(Event,Model,AbstractEvents) :-
   member(b(rlevent(AEvent,AModel,_S,_P,_G,_Thm,_Act,_VW,_PW,_Unm,AbstractEvents2),_,_),AbstractEvents),
   (Event=AEvent, Model=AModel;
    is_abstract_event_in_list(Event,Model,AbstractEvents2)).

% get events that are refined by an event
b_get_abstract_events(Name,Section,AbsEvents) :-
    b_get_machine_operation(Name,_R,_P,Body,_OType),
    get_texpr_expr(Body,RLEVENT),
    RLEVENT = rlevent(_Name,Section,_Stat,_Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,AbsEvents).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% test if we have a query or skip operation which cannot modify the state
% (preconditions, assertions could still fail though !)
% TO DO: precompile
b_operation_cannot_modify_state(Event) :-
   b_get_operation_normalized_read_write_info(Event,_ReadVariables,[]).

% Reuse information

:- volatile b_get_operation_normalized_read_write_info/3.
:- dynamic b_get_operation_normalized_read_write_info/3.
:- volatile reuse_operation_effect/2, candidate_operation_for_reuse/1.
:- dynamic reuse_operation_effect/2, candidate_operation_for_reuse/1.

:- public b_get_operation_normalized_read_write_info_calc/3.
/* compute Read and Written Variables by an operation */
b_get_operation_normalized_read_write_info_calc(Event,ReadVariables,WrittenVariables) :-
   get_operation_info(Event,SubstitutionInfo),
   (memberchk(modifies(WrittenVariables),SubstitutionInfo)
    -> true ; print(no_modifies_info(Event)),nl,fail),
   (memberchk(reads(ReadVariables),SubstitutionInfo)
    -> true ; debug_println(9,no_reads_info(Event)),fail). % happens currently in Event-B !
 %% ,  print(read_write_set(Event,ReadVariables,WrittenVariables)),nl.

:- use_module(specfile, [animation_minor_mode/1]).
:- public reuse_operation_effect_calc/2.
% True if after Executing Operation G:
%       the effect of the operation H (which writes variables in WriteH) is unaffected
reuse_operation_effect_calc(G,H) :- get_preference(try_operation_reuse,true),
   (specfile:animation_minor_mode(eventb) -> print('% OPERATION_REUSE not available for Event-B models'),nl,fail
    % this doesn't work ; see eg. prob_examples/examples/EventBPrologPackages/Advance_WP2/v6_Sep2014/ex_mch.eventb -animate 300 which generates deadlock
    ; true),
   b_get_machine_operation(G,_,_,_),
   b_get_operation_normalized_read_write_info(G,_ReadG,WriteG), %print(analyse_reuse(G,_ReadG,WriteG)),nl,
   WriteG \= [], /* otherwise we have a skip operation; we will not recompute state anyway */
   b_get_machine_operation(H,_Results,_Parameters,_Body,OpType), % TO DO: check if _Body complicated enough
   b_get_operation_normalized_read_write_info(H,ReadH,WriteH),
   \+ ord_intersect(WriteG,ReadH),
   (OpType = classic
    -> \+ ord_intersect(WriteG,WriteH)  % because for the moment we do not yet detect whether a potentially written variable by H is really written !
   % TO DO: either store in reuse_operation the actual variables written, then we can remove the above line
   % or compute the variables that are potentially written PotWriteH and intersect with those
    ; true
    ),
   (debug:debug_mode(off) -> true
    ; format('~nAfter: ~w reuse: ~w~n',[G,H]))
   %,format('~nAfter:  ~w (writes ~w)~n reuse: ~w (reads ~w, writes ~w)~n',[G,WriteG,H,ReadH,WriteH]) %%
   .
reuse_operation_effect_calc(_,_) :- get_preference(try_operation_reuse,true),
   findall(G,b_get_machine_operation(G,_,_,_),LG), length(LG,NrOps), TotPairs is NrOps*NrOps,
   findall((G,H),reuse_operation_effect(G,H),Pairs), length(Pairs,NrReuse),
   format('OPERATION_REUSE Statistics: ~w reuse pairs (for ~w operations, i.e. ~w pairs)~n',[NrReuse,NrOps,TotPairs]),
   fail.

:- public candidate_operation_for_reuse_calc/1.
% True if for that operation we may be able sometimes to reuse previously calculated effects
candidate_operation_for_reuse_calc(Op) :-  reuse_operation_effect(_,Op). % no need to avoid multiple solutions; auto_precompile will remove them

:- public print_reuse/0.
print_reuse :-
  findall((G,H),reuse_operation_effect(G,H),L),
  print_reuse_info(L).

print_reuse_info([]).
print_reuse_info([(G,H)|T]) :- print('When applying '),print(H),print(' after '),
   print(G),print(' recalculate only '),
   b_get_operation_normalized_read_write_info(H,_ReadH,WriteH),
   print(WriteH),nl, print_reuse_info(T).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    (% load in dynamic mode: so that we can retract later
     my_consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    (OK=true -> true ;
      add_error_and_fail(bmachine,'Consulting of File failed: ',File)).

my_consult(File) :- on_exception(error(existence_error(_,_),_),
                      load_files(File,[compilation_mode(assert_all)]),
                    add_error_fail(my_consult,'File does not exist: ',File)).

:- volatile  b_get_all_used_filenames/1, b_get_main_filename/1, b_get_main_filenumber/1.
:- dynamic b_get_all_used_filenames/1. % can be used, e.g., for second argument of pos info field
:- dynamic b_get_main_filename/1. % can be used in above list; there is also specfile:currently_opened_file
:- dynamic b_get_main_filenumber/1. % the number of the main_file in the b_get_all_used_filenames list (useful for pos information)

b_load_machine_from_file(Filename) :-
    (debug_mode(on)
     -> get_parser_version(Version),
        format('Java Parser Version: ~w~n',[Version]) % used to be called unconditionally to be sure to catch error if java or BParser.jar not available; but now we do not want the Java startup time unless really necessary
     ; true),
    b_load_reset,
    debug_stats(loading(Filename)),
    catch(load_b_machine_as_term(Filename,Machine),
          parse_errors(Errors),
          (add_all_perrors(Errors,[],parse_error),fail)),
    b_load_machine_from_term(Filename,Machine).


% a way to load .prob files directly
b_load_machine_probfile(Filename) :- b_load_reset,
    load_b_machine_probfile_as_term(Filename,Machine),
    b_load_machine_from_term(Filename,Machine).

% a way to load form a list of Facts which would be found in the .prob file
b_load_machine_from_list_of_facts(Filename,ListOfFacts) :- % Filename only used below for find_main_file_name and later to find out about position infos if they are in the main file or not
    load_b_machine_list_of_facts_as_term(ListOfFacts,Machine),
    b_load_machine_from_term(Filename,Machine).



b_load_reset :- retractall(b_get_all_used_filenames(_)), retractall(b_get_main_filename(_)),
    retractall(b_get_main_filenumber(_)).

b_load_machine_from_term(Filename,complete_machine(Main,Machines,Filenames)) :-
    b_load_reset, % call in case we call this predicate directly from outside
    !,
    assert(b_get_main_filename(Filename)),assert(b_get_all_used_filenames(Filenames)),
    store_filenumbers(Filenames),
    (find_main_file_name(Nr,Filenames,Filename) -> assert(b_get_main_filenumber(Nr))
     ; print('*** Warning: main filename does not appear in filenames list: '),nl,
       print(Filename),nl, print(Filenames),nl),
    %print(find_main_file_name(Nr,Filenames,Filename)),nl,
    debug_stats(analyse_hierarchy(Main,Filename)),
    (analyse_hierarchy(Main,Machines)
      -> true ; add_error(b_load_machine,'Analyse Hierarchy failed'),fail),
    debug_stats(type_check(Main,Filename)),
    ~~mnf( b_set_machine(Main,Machines,Errors) ), % Typechecking
    debug_stats(checked_machine(Main,Filename)),
    add_all_perrors(Errors,Filenames,type_error),
    no_real_perror_occurred(Errors).
b_load_machine_from_term(F,X) :- add_internal_error('Illegal machine: ',b_load_machine_from_term(F,X)),
    fail.

find_main_file_name(Nr,_Filenames,File) :- absolute_file_name(File,AF),
   tools:get_modulename_filename(AF,TAF), tools:get_filename_extension(AF,ExtF),
   (ExtF = prob -> member(EXT,[mch,ref,imp,tla]) /* any machine extension is ok */
     ; EXT =ExtF),
   b_filenumber(TAF,EXT,Nr,_).
   %find_main_file_name_aux(Nr,Filenames,TAF,EXT).
%find_main_file_name_aux(1,[H|_],TAF,ExtF) :- absolute_file_name(H,AF),tools:get_modulename_filename(AF,TAF),
%   tools:get_filename_extension(AF,ExtF).
%find_main_file_name_aux(Nr,[_|T],TAF,ExtF) :- find_main_file_name_aux(N1,T,TAF,ExtF), Nr is N1+1.


% remember as facts all B filenames, their extension and their number
:- volatile b_filenumber/4.
:- dynamic b_filenumber/4.
store_filenumbers(Filenames) :- retractall(b_filenumber(_,_,_,_)),
  store_filenumbers(Filenames,1).
store_filenumbers([],_).
store_filenumbers([H|T],Nr) :-
   absolute_file_name(H,AF),tools:get_modulename_filename(AF,ModuleName),
   tools:get_filename_extension(AF,ExtF),
   assert(b_filenumber(ModuleName,ExtF,Nr,AF)),
   debug_println(9,b_filenumber(ModuleName,ExtF,Nr,AF)),
   N1 is Nr+1,
   store_filenumbers(T,N1).


:- use_module(btypechecker,[reset_typechecker/0]).
:- use_module(b_global_sets,[b_clear_global_set_type_definitions/0]).
:- use_module(bmachine_construction,[reset_bmachine_construction/0]).
b_machine_reset :- %print(b_machine_reset),nl,
    retractall(machine(_,_)),
    retractall(discharged_theorem(_,_,_)),
    retractall(discharged_guard_strengthening(_,_,_,_)),
    retractall(discharged(_,_,_)),
    retractall(wd_po(_,_,_)),
    retractall(also_discharged(_,_,_)),
    btypechecker:reset_typechecker,
    b_clear_global_set_type_definitions,
    reset_bmachine_construction,
    startup_precompiled. % change so that error messages generated if we call them

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,b_machine_reset,
                           'Reset bmachine facts.').

% :- use_module(library(random)).
:- use_module(preferences,[get_preference/2]).




% precompiled
:- volatile b_machine_name/1.
:- dynamic b_machine_name/1.
:- public b_machine_name_calc/1.
b_machine_name_calc(Name) :-
    get_machine(Name,_).

b_get_animated_sections(Sections) :-
    get_section_from_current_machine(meta,MetaInfos),
    ( member(animated_sections(Sections), MetaInfos) -> true
    ; otherwise -> add_error(bmachine,'No animated sections available'), fail).

:- volatile b_get_named_machine_set/3.
:- dynamic b_get_named_machine_set/3.
:- public b_get_named_machine_set_calc/3.
b_get_named_machine_set(Set,Members) :- b_get_named_machine_set(Set,_,Members).
% precompiled: all enumerated sets with their elements (all in atomic form)
b_get_named_machine_set_calc(Set,TExpr,Members) :-
    get_section_from_current_machine(enumerated_sets,Sets),
    get_section_from_current_machine(enumerated_elements,Elems),
    get_texpr_id(TExpr,Set),
    get_texpr_type(TExpr,set(Type)),
    get_texpr_type(TElem,Type),
    member(TExpr,Sets),
    findall(E, (member(TElem,Elems), get_texpr_id(TElem,E)), Members).

:- volatile b_get_machine_set/2.
:- dynamic b_get_machine_set/2.
:- public b_get_machine_set_calc/2.
b_get_machine_set(Set) :- b_get_machine_set(Set,_).
% precompiled: all deferred and enumerated sets (in atomic form)
b_get_machine_set_calc(Set,TExpr) :-
    get_section_from_current_machine(deferred_sets,Sets),
    get_texpr_id(TExpr,Set),
    member(TExpr,Sets).
b_get_machine_set_calc(Set,TExpr) :-
    b_get_named_machine_set(Set,TExpr,_).

:- volatile b_get_machine_constants/1, b_get_machine_all_constants/1, b_get_machine_variables/1, b_is_constant/1, b_is_variable/2.
:- dynamic b_get_machine_constants/1, b_get_machine_all_constants/1, b_get_machine_variables/1, b_is_constant/1, b_is_variable/2.
b_machine_has_constants :- b_get_machine_constants([_|_]).
:- public b_get_machine_all_constants_calc/1.
% precompiled: all constants (abstract and concrete) in typed form
b_get_machine_all_constants_calc(Csts) :-
    findall(C,get_multi_sections([abstract_constants,concrete_constants],C),Csts).
:- public b_get_machine_constants_calc/1.
% precompiled: all constants (abstract and concrete) in typed form; except for constant_represented inside global set
b_get_machine_constants_calc(Csts) :-
    findall(C,
      (get_multi_sections([abstract_constants,concrete_constants],C),
       get_texpr_id(C,CID),
       \+ b_get_constant_represented_inside_global_set(CID,_)), Csts).

b_machine_has_variables :- b_get_machine_variables([_|_]).
:- public b_get_machine_variables_calc/1.
% precompiled: all variables (abstract and concrete) in typed form
b_get_machine_variables_calc(Vars) :-
    findall(V,get_multi_sections([abstract_variables,concrete_variables],V),Vars).
:- public b_is_constant_calc/1.
% precompiled: all constants (abstract and concrete) in atomic form
b_is_constant_calc(Cst) :-
    b_get_machine_constants(Constants),
    get_texpr_id(Term,Cst),
    member(Term,Constants).
b_is_variable(V) :- b_is_variable(V,_).
:- public b_is_variable_calc/2.
% precompiled: all variables (abstract and concrete) in atomic form
b_is_variable_calc(Cst,Type) :-
    b_get_machine_variables(Vars),
    get_texpr_id(Term,Cst),
    get_texpr_type(Term,Type),
    member(Term,Vars).
% look up elements in several sections of the machine
get_multi_sections(Sections,Elem) :-
    member(Section,Sections),
    get_section_from_current_machine(Section,List),
    member(Elem,List).

:- volatile b_get_invariant_from_machine/1,
    b_get_linking_invariant_from_machine/1,
    b_get_properties_from_machine/1,
    b_get_static_assertions_from_machine/1,
    b_get_unproven_static_assertions_from_machine/1,
    b_get_dynamic_assertions_from_machine/1,
    b_get_unproven_dynamic_assertions_from_machine/1,
    b_get_assertions_from_main_machine/2,
    b_get_initialisation_from_machine/2,
    b_get_machine_operation/5,b_top_level_operation/1,
    b_get_promoted_machine_operations/1.
:- dynamic b_get_invariant_from_machine/1,
    b_get_linking_invariant_from_machine/1,
    b_get_properties_from_machine/1,
    b_get_static_assertions_from_machine/1,
    b_get_unproven_static_assertions_from_machine/1,
    b_get_dynamic_assertions_from_machine/1,
    b_get_unproven_dynamic_assertions_from_machine/1,
    b_get_assertions_from_main_machine/2,
    b_get_initialisation_from_machine/2,
    b_get_machine_operation/5,b_top_level_operation/1,
    b_get_promoted_machine_operations/1.

:- public b_get_invariant_from_machine_calc/1.
% precompiled: typed invariant (a predicate)
b_get_invariant_from_machine_calc(Inv) :-
    get_section_from_current_machine(invariant,Inv).

:- public b_get_linking_invariant_from_machine_calc/1.
% precompiled: typed linking invariant (a predicate)
b_get_linking_invariant_from_machine_calc(Inv) :-
    get_section_from_current_machine(linking_invariant,Inv).

:- public b_get_properties_from_machine_calc/1.
% precompiled: typed properties (a predicate)
b_get_properties_from_machine_calc(SortedProp) :-
    get_section_from_current_machine(properties,Prop),
    (sort_properties(Prop,SortedProp) -> true %(Prop=SortedProp->true ; print(sorted(SortedProp)))
     ; (add_error(sort_properties,'sort_properties failed: ',Prop),
        print(sort_properties_failed),nl,SortedProp=Prop)).

sort_properties(Prop,SortedProp) :-  % print(sort_properties),nl,statistics(runtime, [R1,_]),
   ~~mnf(weigh_properties(Prop,1,_,[],SProp)),
    % statistics(runtime, [R2,_]), RDiff is R2-R1,print(done_weigh(RDiff,SProp)),nl,
    sort(SProp,SSProp),
  %  statistics(runtime, [R3,_]), RDiff2 is R3-R2,print(done_sort(RDiff2)),nl,
   ~~mnf(remove_weights(SSProp,PropList)),
   ~~mnf(conjunct_predicates(PropList,SortedProp)).
  % translate:print_bexpr(SortedProp).

remove_weights([],[]).
remove_weights([prop(_W,_Nr,E)|T],[E|RT]) :- remove_weights(T,RT).

weigh_properties(Pred,Nr,ONr,InL,OutL) :-
   is_a_conjunct(Pred,LHS,RHS), !,
   weigh_properties(LHS,Nr,N1,InL,L2),
   weigh_properties(RHS,N1,ONr,L2,OutL).
weigh_properties(b(Pred,S,I),Nr,N1,T,[prop(Weight,Nr,b(Pred,S,I))|T]) :-
   N1 is Nr+1, weigh_properties2(Pred,Weight).
   % translate:print_bexpr(b(Pred,S,I)), print(' WEIGHT: '), print(Weight),nl

weigh_properties2(truth,W) :- !, W=1.
weigh_properties2(falsity,W) :- !, W=1.
weigh_properties2(equal(L,R),W) :- weigh_term(L,W1), !,
  ( weigh_term(R,W2) -> W is W1*W2
  ; W is 3200).
%  ; W is W1*20). %causes issue with ParityFunction test
%weigh_properties2(equal(_,R),W) :- weigh_term(R,W2),  !,  W is W2*20.
weigh_properties2(SOp,W) :- comparison(SOp,L,R),!,
  ( (weigh_term(L,W1), weigh_term(R,W2))
    -> W is 7+W1+W2 % try and give things like x>10 priority over x>y (to overcome CLP(FD) limitations)
    ; W=10).
weigh_properties2(partition(LHS,RHS),W) :- weigh_term(LHS,W1), l_weigh_term(RHS,W2), !, W is W1*W2.
weigh_properties2(not_equal(L,R),W) :- ((id(L),empty(R)) ; (id(R),empty(L))),!, W=1.
weigh_properties2(not_equal(L,R),W) :- weigh_term(L,W1), weigh_term(R,W2),!,
   W is 20+W1*W2.
weigh_properties2(_Rest,W) :- W=32000.

comparison(less(L,R),L,R).
comparison(less_equal(L,R),L,R).
comparison(greater(L,R),L,R).
comparison(greater_equal(L,R),L,R).


id(b(identifier(_),_,_)).
empty(b(empty_set,_,_)).
empty(b(empty_sequence,_,_)).

l_weigh_term([],0).
l_weigh_term([H|T],R) :- l_weigh_term(T,TR), weigh_term(H,TH), R is TH+TR.

weigh_term(b(T,_,_),R) :- (weigh_term2(T,R) -> true ; fail). %R=20).

weigh_term2(value(_),1).
weigh_term2(boolean_true,1).
weigh_term2(boolean_false,1).
weigh_term2(empty_set,1).
weigh_term2(empty_sequence,1).
weigh_term2(integer(_),1).
weigh_term2(min_int,1).
weigh_term2(max_int,1).
weigh_term2(string(_),1).
weigh_term2(bool_set,2).
weigh_term2(identifier(_),2).
weigh_term2(set_extension(_),3).
weigh_term2(integer_set(_),3).
weigh_term2(interval(L,R),W) :- weigh_term(L,WL), weigh_term(R,WR), W is WL*WR.



b_machine_has_constants_or_properties :-
  (b_machine_has_constants -> true
   ; (b_get_properties_from_machine(Prop)), \+ is_truth(Prop)).

b_get_machine_operation_parameter_types(OpName,ParTypes) :-
   b_get_machine_operation_typed_parameters(OpName,Parameters),
   maplist(get_texpr_type,Parameters,ParTypes).

b_get_machine_operation_typed_parameters(Operation,Res) :-
   if(b_get_machine_operation_typed_parameters_aux(Operation,Paras),Res=Paras,
      (add_internal_error('Cannot get parameters: ',Operation),fail)).
% also recognises SETUP_CONSTANTS & INITIALISE_MACHINE
b_get_machine_operation_typed_parameters_aux('$setup_constants',Constants) :-
   b_machine_has_constants_or_properties,
   b_get_machine_constants(Constants).
b_get_machine_operation_typed_parameters_aux('$initialise_machine',Variables) :-
   b_get_machine_variables(Variables).
b_get_machine_operation_typed_parameters_aux(OpName,Parameters) :-
   b_get_machine_operation(OpName,_R,Parameters,_Body,_OType).

b_get_machine_operation_parameter_names('@INITIALISATION',ParaNames) :- % for Tcl/Tk
   b_get_machine_operation_parameter_names('$initialise_machine',ParaNames).
b_get_machine_operation_parameter_names('@PROPERTIES',ParaNames) :- % for Tcl/Tk
   b_get_machine_operation_parameter_names('$setup_constants',ParaNames).
b_get_machine_operation_parameter_names(OpName,ParaNames) :-
   b_get_machine_operation_typed_parameters(OpName,Parameters),
   get_texpr_ids(Parameters,ParaNames).

b_get_machine_operation(Name,Results,Parameters,Body) :-
    b_get_machine_operation(Name,Results,Parameters,Body,_OType).

:- public b_get_assertions_from_main_machine_calc/2.
b_get_assertions_from_main_machine_calc(static,FProp) :-
    ((b_get_static_assertions_from_machine(Prop),
      exclude(is_not_main_assertion,Prop,FProp)) -> true
      ; add_error(bmachine,'b_get_assertions_from_main_machine_calc static fails'), FProp=[]).
b_get_assertions_from_main_machine_calc(dynamic,FProp) :-
    ((b_get_dynamic_assertions_from_machine(Prop),
      exclude(is_not_main_assertion,Prop,FProp)) -> true
      ; add_error(bmachine,'b_get_assertions_from_main_machine_calc dynamic fails'), FProp=[]).
is_not_main_assertion(Assertion) :-
    get_texpr_pos(Assertion,Pos),
    error_manager:position_is_not_in_main_file(Pos).

% get nr of assertions and nr of main assertions for static/dynamic
b_get_assertion_count(StaticOrDynamic,NrAll,NrMain) :-
   b_get_assertions(all,StaticOrDynamic,A), length(A,NrAll),
   b_get_assertions_from_main_machine(StaticOrDynamic,MA),length(MA,NrMain).

b_get_assertions(main,Mode,Ass) :- !,
   b_get_assertions_from_main_machine(Mode,Ass).
b_get_assertions(specific(Label),Mode,LabelAss) :- !,
   b_get_assertions(all,Mode,Ass),
   %maplist(get_texpr_label,Ass,Labels), print(labels(Labels)),nl,
   include(has_label(Label),Ass,LabelAss),
   (LabelAss=[] -> tools:ajoin(['No ',Mode,' assertion has label: '],Msg),
                   add_warning(b_get_assertions,Msg,Label) ; true).
b_get_assertions(_,static,Ass) :- b_get_static_assertions_from_machine(Ass).
b_get_assertions(_,dynamic,Ass) :- b_get_dynamic_assertions_from_machine(Ass).
b_main_machine_has_no_assertions :-
  b_get_assertions_from_main_machine(static,[]),
  b_get_assertions_from_main_machine(dynamic,[]).


%has_label(Label,Expr) :- get_texpr_label(Expr,Label1), format('Labels ~w =? ~w~n',[Label1,Label]), fail.
has_label(Label,Expr) :- get_texpr_label(Expr,Label).

:- public b_get_static_assertions_from_machine_calc/1.
% precompiled: typed assertions (a list of predicates)
b_get_static_assertions_from_machine_calc(Prop) :-
    get_section_from_current_machine(assertions,Prop1),
    exclude(predicate_uses_variables,Prop1,Prop),!.
b_get_static_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_static_assertions_from_machine_calc fails'), Prop=[].

:- dynamic b_machine_has_static_assertions/0.
:- public b_machine_has_static_assertions_calc/0.
b_machine_has_static_assertions_calc :- 
    b_get_static_assertions_from_machine([_|_]).

:- public b_get_unproven_static_assertions_from_machine_calc/1.
b_get_unproven_static_assertions_from_machine_calc(FAssertions) :-
    b_get_static_assertions_from_machine(Assertions),
    filter_out_proven_assertions(Assertions,FAssertions),!.
b_get_unproven_static_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_unproven_static_assertions_from_machine_calc fails'), Prop=[].

filter_out_proven_assertions([],[]).
filter_out_proven_assertions([A1|T],R) :-
     (is_discharged_assertion(A1)
      -> R = TR   %, print('Proven: '), translate:print_bexpr(A1),nl
      ; R = [A1|TR]),
      filter_out_proven_assertions(T,TR).

:- public b_get_dynamic_assertions_from_machine_calc/1.
b_get_dynamic_assertions_from_machine_calc(Prop) :-
    get_section_from_current_machine(assertions,Prop1),
    include(predicate_uses_variables,Prop1,Prop),!.
b_get_dynamic_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_dynamic_assertions_from_machine_calc fails'), Prop=[].


:- dynamic b_machine_has_dynamic_assertions/0.
:- public b_machine_has_dynamic_assertions_calc/0.
b_machine_has_dynamic_assertions_calc :- 
    b_get_dynamic_assertions_from_machine([_|_]).

b_machine_has_assertions :- (b_machine_has_static_assertions -> true ; b_machine_has_dynamic_assertions).

b_machine_has_unproven_assertions :-
    b_get_unproven_dynamic_assertions_from_machine([_|_]) -> true
    ; b_get_unproven_static_assertions_from_machine([_|_]).


get_assertions_from_machine(Type,Assertions) :-
    ( Type=(dynamic) ->
        (preferences:get_preference(use_po,true)
         -> b_get_unproven_dynamic_assertions_from_machine(Assertions)
         ;  b_get_dynamic_assertions_from_machine(Assertions)
        )
    ; Type=(static)  ->
        (preferences:get_preference(use_po,true)
          -> b_get_unproven_static_assertions_from_machine(Assertions)
          ;  b_get_static_assertions_from_machine(Assertions)
        )
    ),
    Assertions = [_|_].

:- public b_get_unproven_dynamic_assertions_from_machine_calc/1.
b_get_unproven_dynamic_assertions_from_machine_calc(FAssertions) :-
    b_get_dynamic_assertions_from_machine(Assertions),
    filter_out_proven_assertions(Assertions,FAssertions),!.
b_get_unproven_dynamic_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_unproven_dynamic_assertions_from_machine_calc fails'), Prop=[].

predicate_uses_variables(Pred) :-
    b_get_machine_variables(Vars),
    get_texpr_ids(Vars,Ids),
    list_to_ord_set(Ids,OrdVars),
    find_identifier_uses(Pred,[],UsedVars),
    list_to_ord_set(UsedVars,OrdUsedVars),
    ord_intersect(OrdVars,OrdUsedVars).

:- public b_get_initialisation_from_machine_calc/2.
% precompiled: typed initialisation (a substitution)
b_get_initialisation_from_machine_calc(Init,OType) :-
    get_section_from_current_machine(initialisation,Init),
    ( get_texpr_expr(Init,rlevent(_Name,_Section,_Status,Params,_Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents)) ->
        get_operation_type(Init,'INITIALISATION',Params,OType)
    ; otherwise ->
        OType = classic).
% precompiled: operations, Name is in atomic form, Results and Parameters are lists
%   of typed identifiers, Body is a typed substitution

:- volatile b_machine_operation_names_in_reverse_order/1.
:- dynamic b_machine_operation_names_in_reverse_order/1.
% a list of operation names in reverse order; this way we can randomize the
% generation of the successor states
precompile_operation_names_in_reverse_order :-
   retractall(b_machine_operation_names_in_reverse_order(_)),
   b_get_machine_operation(Name,_Results,_Parameters,_Body,_OType),
   asserta(b_machine_operation_names_in_reverse_order(Name)),
   fail.
precompile_operation_names_in_reverse_order.

:- public b_get_machine_operation_calc/5.
b_get_machine_operation_calc(Name,Results,Parameters,OpBody,OType) :-
    is_ground(Name,NameGround),
    get_section_from_current_machine(operation_bodies,Ops),
    get_texpr_expr(Op,operation(TId,Results,Parameters,Body)),
    get_texpr_id(TId,op(Name)),
    member(Op,Ops),
    % remove choicepoints when name was ground
    (NameGround==ground -> !; true),
    get_operation_type(Op,Name,Parameters,OType),
    ((get_preference(use_common_subexpression_also_for_substitutions,true),
      get_preference(use_common_subexpression_elimination,true),
      \+ specfile:animation_minor_mode(eventb) % CSE_SUBST not yet available for Event-B
      )
       -> print(applying_cse_to(Name)),nl,
          b_expression_sharing:cse_optimize_substitution(Body,OpBody)
       ; OpBody = Body).
get_operation_type([],_Name,_Parameters,classic).
get_operation_type(Op,Name,Parameters,OType) :-
    get_texpr_info(Op,Info),
    ( member(eventb(_),Info) ->
        OType=eventb_operation(SortedChangeSet,Ps,OpPattern),
        member(modifies(ChangeSet),Info),!,
        sort(ChangeSet,SortedChangeSet), % maybe not necessary ?
        same_length(Parameters,Ps),
        OpPattern =.. [Name|Ps]
    ; otherwise ->
        OType=classic).

:- volatile get_operation_info/2.
:- dynamic get_operation_info/2.
:- public get_operation_info_calc/2.
% precompiled: operation infos
get_operation_info_calc(OpName,Info) :-
    get_section_from_current_machine(operation_bodies,Ops),
    create_texpr(operation(TId,_Results,_Parameters,_Body),_,Info,TOp),
    get_texpr_id(TId,op(OpName)),
    member(TOp,Ops).

:- public b_top_level_operation_calc/1.
% precompiled: operation names (in atomic form) of top-level operations (promoted operations)
b_top_level_operation_calc(Name) :-
    is_ground(Name,NameGround),
    get_section_from_current_machine(promoted,Promoted),
    get_texpr_id(TId,op(Name)),
    member(TId,Promoted),
    (NameGround==ground -> !; true).
:- public b_get_promoted_machine_operations_calc/1.
b_get_promoted_machine_operations_calc(Ops) :-
    get_section_from_current_machine(promoted,Ops).

% this succeeds not just for top-level operations
b_is_operation_name(Name) :-
   b_get_machine_operation(Name,_,_,_,_).

% the names of definitions which have a given Prefix,
% Type is expression or predicate, Name is an atomic name
b_definition_prefixed(Type,Prefix,Suffix,Name) :-
    safe_atom_chars(Prefix,PrefixChars,b_definition_prefixed1),
    append(PrefixChars,SuffixChars,FullChars),
    (ground(Suffix) -> safe_atom_chars(Suffix,SuffixChars,b_definition_prefixed2),safe_atom_chars(Name,FullChars,b_definition_prefixed3); true),
    b_get_definition(Name,Type,_,_,_),
    safe_atom_chars(Name,FullChars,b_definition_prefixed4),
    safe_atom_chars(Suffix,SuffixChars,b_definition_prefixed5).

b_get_definition(Name,DefType,Args,RawExpr,Deps) :-
    get_section_from_current_machine(definitions,Defs),
    member(definition_decl(Name,DefType,_Pos,Args,RawExpr,Deps),Defs).
b_get_typed_predicate_definition(Name,Scope,TExpr) :-
    b_get_definition(Name,predicate,[],RawExpr,_Deps),
    type_with_errors(RawExpr,Scope,TExpr).
b_get_typed_expression_definition(Name,Scope,TExpr) :-
    b_get_definition(Name,expression,[],RawExpr,_Deps),
    type_with_errors(RawExpr,Scope,TExpr).
b_get_true_expression_definition(Name) :-
    b_get_typed_expression_definition(Name,[variables],PF),
    PF = b(boolean_true,boolean,_).
b_get_typed_definition(Name,Scope,TExpr) :-
    b_get_definition(Name,_DefType,[],RawExpr,_Deps),
    type_with_errors(RawExpr,Scope,TExpr).
type_with_errors(RawExpr,Scope,TExpr) :-
    type_with_errors(RawExpr,Scope,_,TExpr).
type_with_errors(RawExpr,Scope,Type,TExpr) :- %print(scope(Scope)),nl,
    b_type_expression(RawExpr,Scope,Type,TExpr,Errors),!,
    add_all_perrors(Errors,[],type_expression_error),
    no_real_perror_occurred(Errors).
b_type_open_predicate_with_errors(OptionalQuantifier,RawPred,Scope,TPred) :-
    b_type_open_predicate(OptionalQuantifier,RawPred,Scope,TPred,Errors),
    add_all_perrors(Errors,[],type_expression_error),
    no_real_perror_occurred(Errors).
b_type_open_predicate(OptionalQuantifier,RawPred,Scope,TPred,Errors) :-
    full_b_machine(Machine),
    replace_prob_set_elements_in_scope(Scope,Scope1),
    type_open_predicate_with_quantifier(OptionalQuantifier,RawPred,Scope1,Machine,TPred,Errors),!.
b_type_open_exists_predicate(Raw,Typed,Errors) :-
    b_type_open_predicate(open(exists),Raw,[prob_ids(visible),variables],Typed,Errors).

b_type_expression(RawExpr,Scope,Type,TExpr,Errors) :-
    b_type_expressions([RawExpr],Scope,[Type],[TExpr],Errors).
b_type_expressions(RawExprs,Scope,Types,TExprs,Errors) :-
    full_b_machine(Machine),
    replace_prob_set_elements_in_scope(Scope,Scope1),
    % we need this, for example for tcltk_check_state_sequence_from_file, when deferred set elements written by TLC are in the file
    ~~mnf( type_in_machine_l(RawExprs,Scope1,Machine,Types,TExprs,Errors) ),!.

replace_prob_set_elements_in_scope(Scope,NewScope) :-
    ( select(prob_ids(AllOrVisible),Scope,identifier(Ids),NewScope) ->   % replaces prob by deferred set element identifiers
     % for example if we have SETS PID then we would have PID1,... in Ids unless PID1 is known to be another identifier
        b_get_prob_deferred_set_elements(DefIDs,AllOrVisible),
        exclude(known_identifier,DefIDs,Ids) % exclude those ids that would clash
    ; otherwise ->
        Scope = NewScope).

known_identifier(X) :- get_texpr_id(X,ID),
     (bmachine:b_is_constant(ID) ;
      bmachine:b_is_variable(ID,_)).


:- use_module(b_global_sets,[b_global_deferred_set/1,all_elements_of_type/2]).
:- use_module(input_syntax_tree,[get_raw_position_info/2]).
% try and get source code for an identifier
source_code_for_identifier(ID,constant,Type,OriginStr,OriginTerm,Source) :- b_is_constant(ID),!,
      get_constant_type_and_origin(ID,Type,OriginStr,OriginTerm),
      findall(Def,get_constant_definition(ID,Def),AllDefs),
      (AllDefs=[_|_] -> conjunct_predicates(AllDefs,Source)
        ; % try and find other type of predicates defining ID
          findall(Pred,get_constant_predicate(ID,Pred),AllPreds),
          conjunct_predicates(AllPreds,Source)).
source_code_for_identifier(ID,variable,Type,OriginStr,OriginTerm,Source) :- b_is_variable(ID),!,
      get_variable_type_and_origin(ID,Type,OriginStr,OriginTerm),
      Source = b(truth,pred,[]). % TO DO: improve
source_code_for_identifier(ID,Kind,Type,OriginStr,OriginTerm,Source) :- b_get_machine_set(ID,TID),!,
      (b_global_deferred_set(ID) -> Kind = deferred_set ; Kind = enumerated_set),
      Type = set(global(ID)),
      all_elements_of_type(ID,All),
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID,
      Source = b(equal(TID,b(value(All),Type,[])),pred,[]).
source_code_for_identifier(ID,Kind,Type,OriginStr,OriginTerm,Source) :-
      global_set_element(ID,GID),!,
      (b_global_deferred_set(GID)
       -> Kind = deferred_set_element ; Kind = enumerated_set_element),
      b_get_machine_set(GID,TGID),
      Type = global(ID),
      all_elements_of_type(GID,All),
      translate:translate_span_with_filename(TGID,OriginStr), OriginTerm=TGID,
      Source = b(equal(TGID,b(value(All),set(Type),[])),pred,[]).
source_code_for_identifier(ID,definition,Type,OriginStr,OriginTerm,Source) :-
      b_get_definition(ID,DefType,Args,RawExpr,_Deps),!,
      Type = DefType,
      % Definitions are only present in RAW form
      get_raw_position_info(RawExpr,PosInfo),
      translate:translate_span_with_filename(PosInfo,OriginStr), % the position info does not always seem to be valid (when coming from DEFINITION FILE) !??! TO DO fix
      OriginTerm=PosInfo,
      source_code_for_definition(DefType,ID,Args,RawExpr,Source).
source_code_for_identifier(ID,Kind,subst,OriginStr,OriginTerm,Source) :-
      b_get_machine_operation(ID,Res,TParas,Body,OType),
      (OType=classic -> Kind = operation ; Kind=event),
      translate:translate_span_with_filename(Body,OriginStr),Body=OriginTerm,
      get_texpr_info(Body,Info),
      Source=b(operation(b(identifier(ID),subst,[]),Res,TParas,Body),subst,Info).
source_code_for_identifier(ID,file,Type,OriginStr,OriginTerm,b(string(Source),string,[])) :-
      b_filenumber(ID,Kind,_,Filename),
      extension_kind(Kind,Type),
      ajoin([Type,' ',ID],Source),
      OriginTerm =src_position_with_filename(1,0,0,Filename),
      OriginStr = Filename.
% TO DO: look for invariant and assertion labels as ids ?

extension_kind(mch,'MACHINE').
extension_kind(ref,'REFINEMENT').
extension_kind(def,'DEFINITIONS FILE').
extension_kind(imp,'IMPLEMENTATION').

source_code_for_definition(predicate,ID,_,RawExpr,Source) :-
     b_type_open_exists_predicate(RawExpr,Typed,Errors),
     (no_real_perror_occurred(Errors) -> true ; debug_println(9,errs(ID,Errors))),
     Source=Typed.
source_code_for_definition(expression,ID,Args,RawExpr,Source) :-
     % TO DO: add parameters to avoid type errors; but we get Typed anyway
     b_type_expression(RawExpr,[prob_ids(visible),variables],Type,Typed,Errors),
     (no_real_perror_occurred(Errors) -> true ; debug_println(9,errs(ID,Args,Type,Errors))),
     Source=Typed.
% TO DO: treate substitutions
source_code_for_definition(DefType,ID,_Args,_RawExpr,Source) :-
     Source=b(identifier(ID),DefType,[]).

:- use_module(b_global_sets,[lookup_global_constant/2,prob_deferred_set_element/4]).
global_set_element(ID,GlobalSetID) :- lookup_global_constant(ID,fd(_,GlobalSetID)).
global_set_element(ID,GlobalSetID) :-
       prob_deferred_set_element(GlobalSetID,_Elem,ID,all).


get_constant_type_and_origin(ID,Type,OriginStr,OriginTerm) :-
      b_get_machine_constants(Constants),
      TID = b(identifier(ID),Type,_Infos),
      member(TID,Constants),
      % we could also use member(origin(Origin),Infos)
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID.
get_variable_type_and_origin(ID,Type,OriginStr,OriginTerm) :-
      b_get_machine_variables(Vars),
      TID = b(identifier(ID),Type,_Infos),
      member(TID,Vars),
      % we could also use member(origin(Origin),Infos)
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID.

get_constant_definition(ID,FullDef) :-
      b_get_properties_from_machine(Prop),
      EqDef = b(equal(LHS,RHS),pred,_),
      member_in_conjunction_cse(FullDef,EqDef,Prop),
      (get_texpr_id(LHS,ID) ; get_texpr_id(RHS,ID)).

get_constant_predicate(ID,FullPred) :-
      b_get_properties_from_machine(Prop),
      member_in_conjunction_cse(FullPred,Pred,Prop),
      involves_id(Pred,ID).
involves_id(ID,b(E,_,_)) :- involves_id_aux(E,ID).
involves_id_aux(member(LHS,_),ID) :- get_texpr_id(LHS,ID).
involves_id_aux(subset(LHS,_),ID) :- get_texpr_id(LHS,ID).
involves_id_aux(strict_subset(LHS,_),ID) :- get_texpr_id(LHS,ID).

:- volatile b_get_machine_setscope/2.
:- dynamic b_get_machine_setscope/2.
:- public b_get_machine_setscope_calc/2.
% precompiled: definitions of the form "scope_SET == Expr",
%   Set is the name of the Set, TExpr the typed expression
b_get_machine_setscope_calc(Set,TExpr) :-
    b_definition_prefixed(expression, scope_, Set, Name),
    b_get_typed_definition(Name,[],TExpr).

:- volatile b_get_machine_goal/1.
:- dynamic b_get_machine_goal/1.
:- public b_get_machine_goal_calc/1.
% precompiled: definition of GOAL, a typed predicate
b_get_machine_goal_calc(TPred) :- get_goal(TPred).

b_reset_machine_goal_from_DEFINITIONS :-
    get_goal(TPred),
    b_set_parsed_typed_machine_goal(TPred).
get_goal(Goal) :-
    get_proz_settings(Settings),memberchk(goal(Goal),Settings),!.
get_goal(Goal) :-
    get_texpr_type(Goal,pred),
    b_get_typed_predicate_definition('GOAL',[variables],Goal).

:- volatile b_get_machine_searchscope/1.
:- dynamic b_get_machine_searchscope/1.
:- public b_get_machine_searchscope_calc/1.
% precompiled: definition of SCOPE, a typed predicate
b_get_machine_searchscope_calc(TPred) :-
    get_texpr_type(TPred,pred),
    b_get_typed_definition('SCOPE',[variables],TPred).

:- volatile b_get_definition_string_from_machine/2.
:- dynamic b_get_definition_string_from_machine/2.
:- public b_get_definition_string_from_machine_calc/2.
% precompiled: string definitions, Name == "String"
b_get_definition_string_from_machine_calc(Name,String) :-
    b_get_definition(Name,expression,[],string(Pos,Str),_Deps),
    get_texpr_expr(TExpr,string(String)),
    type_with_errors(string(Pos,Str),[],TExpr).

get_animation_image(Nr,S) :-
    get_proz_animation_function(_,_,Images),!,
    nth1(Nr,Images,S).
get_animation_image(Nr,S) :-
    b_get_definition_string_from_machine(Def_Name,S),
    /* ANIMATION_IMG */
    atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,73,77,71|TailCodes]),
    number_codes(Nr,TailCodes).


get_proz_animation_function(Function,Default,Images) :-
    get_proz_settings(Settings),
    memberchk(animation_function(Function,Default,Images),Settings).
get_proz_settings(Settings) :-
    machine(_,Sections),memberchk(meta/Meta,Sections),
    memberchk(proz_settings(Settings),Meta).


:- volatile b_get_machine_animation_function/2.
:- dynamic b_get_machine_animation_function/2.
:- public b_get_machine_animation_function_calc/2.
:- use_module(tools,[safe_number_codes/2]).
% precompiled: definition of ANIMATION_FUNCTION, a typed expression
b_get_machine_animation_function_calc(AFun,Nr) :-
    get_proz_animation_function(TFun,Default,_),!,
    ( Nr = -1, AFun = Default
    ; Nr = 0, TFun \= none, AFun=TFun
    ).
b_get_machine_animation_function_calc(TFun,Nr) :-
    get_texpr_type(TFun,set(couple(_,_))),
    b_get_definition(Def_Name,expression,[],_RawExpr,_Deps),
    atom_codes(Def_Name,AC),
    append("ANIMATION_FUNCTION",TailCodes,AC),
    (TailCodes = [] -> Nr=0
     ; TailCodes = "_DEFAULT" -> Nr = -1
     ; safe_number_codes(Nr,TailCodes)),
    b_get_typed_expression_definition(Def_Name,[variables],TFun). % only type it afterwards; avoid throwing type errors for other definitions

:- volatile b_get_machine_heuristic_function/1.
:- dynamic b_get_machine_heuristic_function/1.
:- public b_get_machine_heuristic_function_calc/1.
b_get_machine_heuristic_function_calc(TFun) :- %print(extracting_heuristic_fun(TFun)),nl,
    get_texpr_type(TFun,integer),
    b_get_typed_expression_definition('HEURISTIC_FUNCTION',[variables],TFun).

% Custom nodes and edges for a custom graph representation of a state
:- volatile b_get_machine_custom_nodes_function/1, b_get_machine_custom_edges_function/1.
:- dynamic b_get_machine_custom_nodes_function/1, b_get_machine_custom_edges_function/1.
:- public b_get_machine_custom_nodes_function_calc/1.
b_get_machine_custom_nodes_function_calc(TFun) :-
    get_texpr_type(TFun,set(_)),
    b_get_typed_expression_definition('CUSTOM_GRAPH_NODES',[variables],TFun).
:- public b_get_machine_custom_edges_function_calc/1.
b_get_machine_custom_edges_function_calc(TFun) :- %print(extracting_heuristic_fun(TFun)),nl,
    get_texpr_type(TFun,set(couple(_,_))),
    b_get_typed_expression_definition('CUSTOM_GRAPH_EDGES',[variables],TFun).

% ------------------------

:- volatile b_get_constant_represented_inside_global_set/2.
:- dynamic b_get_constant_represented_inside_global_set/2.

:- public b_get_constant_represented_inside_global_set_calc/2.
b_get_constant_represented_inside_global_set_calc(X,GlobalSetName) :-
  b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants,_),
  member(X,DisjointConstants). /* X will be integrated into the global_set */


:- volatile b_get_disjoint_constants_of_type/3.
:- dynamic b_get_disjoint_constants_of_type/3.
:- public b_get_disjoint_constants_of_type_calc/3.
b_get_disjoint_constants_of_type_calc(GlobalSetName, DisjointConstants, AllConstantsIDs) :-
    b_get_machine_all_constants(Constants), /* get all the constants */
    b_get_properties_from_machine(Properties),
    b_get_machine_set(GlobalSetName),
    \+ b_get_named_machine_set(GlobalSetName,_), % not an enumerated set
    find_constants_of_type(Constants,GlobalSetName,AllConstantsIDs,_GSConstants),
    find_inequal_global_set_identifiers(AllConstantsIDs,global(GlobalSetName),Properties,DisjointConstants),
    DisjointConstants \= [],
    debug_println(9,disjoint_constants(GlobalSetName,DisjointConstants)).
   % TO DO: remove from Properties irrelevant conjuncts; sort AllConstantsIDs by having first those with max disequalities

find_constants_of_type([],_GS,[],[]) :- !.
find_constants_of_type([b(identifier(ID),global(GS),INFO)|T], GS, [ID|IT],
                       [b(identifier(ID),global(GS),INFO)|TT]) :- !,
          find_constants_of_type(T,GS,IT,TT).
find_constants_of_type([b(_,_,_)|T], GS, IT, TT) :- !,
          find_constants_of_type(T,GS,IT,TT).
find_constants_of_type(X,GS,_,_) :- print(find_constants_of_type_error(X,GS)),nl,fail.




% ------------------------
:- volatile b_get_all_used_identifiers_in_section/2.
:- dynamic b_get_all_used_identifiers_in_section/2.
:- public b_get_all_used_identifiers_in_section_calc/2.

b_get_all_used_identifiers_in_section_calc(SECTION,Identifiers) :-
    full_b_machine(BMachine),
    member(SECTION,[constraints,properties,invariant,assertions,
                initialisation,operation_bodies]),
    findall(I, find_used_identifier_in_machine_section(BMachine,SECTION,I), Ids),
    %print(section_ids(SECTION,Ids)),nl,
    sort(Ids,Identifiers).

:- volatile b_get_all_used_identifiers/1.
:- dynamic b_get_all_used_identifiers/1.
:- public b_get_all_used_identifiers_calc/1.
:- use_module(library(ordsets),[ord_union/2]).
% Precompiled: b_get_all_used_identifiers_calc/1 returns a list of all
% identifiers used in machine sections that contain expressions and predicates
b_get_all_used_identifiers_calc(Identifiers) :-
    findall(I, b_get_all_used_identifiers_in_section(_,I), ListOfList),
    ord_union(ListOfList,Identifiers). % ,print(all(Identifiers)),nl.

find_used_identifier_in_machine_section(Machine,Sec,Identifier) :-
    % treat it like a list of expressions
    get_section_texprs(Sec,Machine,TExprs),
    % and find all identifiers in each expression
    member(TExpr,TExprs),
    find_identifier_uses(TExpr,[],Ids),
    member(Identifier,Ids).

:- volatile b_is_unused_constant/1.
:- dynamic b_is_unused_constant/1.
:- public b_is_unused_constant_calc/1.
b_is_unused_constant_calc(UnusedConstant) :-
      findall(C,b_is_constant(C),CL), %b_get_machine_constants(Constants), gives typed ids
      CL\=[] ,sort(CL,Constants),
      b_get_all_used_identifiers_in_section(invariant,I1),
      b_get_all_used_identifiers_in_section(assertions,I2),
      b_get_all_used_identifiers_in_section(initialisation,I3),
      b_get_all_used_identifiers_in_section(operation_bodies,I4),
      ord_union([I1,I2,I3,I4],UsedIds),
      %print(used(UsedIds)),nl,
      ord_subtract(Constants,UsedIds,UnusedConstants),
      UnusedConstants\=[],
      (b_get_machine_variables([]) -> true
       ; print_message_with_max_depth(unused_constants(UnusedConstants),500)),
      member(UnusedConstant,UnusedConstants).


% precompiled: Skel is the operation skeleton, SpecializedInv a typed predicate
% (Skel,SpecializedInv) :-
%  b_get_machine_operation(Name,_Res,Params,_Body),
%  atom_concat('SIMPLIFIED_INV_OP_',Name,ASNC),
%  b_get_typed_definition(ASNC,[variables,identifiers(Params)],SpecializedInv),
%  length(Params,Len),
%  functor(Skel,Name,Len),
%  print(b_specialized_invariant_for_op(Skel)).


get_section_from_current_machine(Section,Content) :-
    full_b_machine(BMachine),
    ~~mnf( get_section(Section,BMachine,Content) ).

is_ground(X,R) :- ground(X),!,R=ground.
is_ground(_,nonground).

b_set_empty_machine :- debug_println(9,setting_empty_machine),
    empty_machine(Main,Machines),
    b_set_machine(Main,Machines,[]).

b_get_machine_refinement_hierarchy(Hierarchy) :-
    \+ machine(_Name,_M),!,Hierarchy=[]. %No B Machine loaded
b_get_machine_refinement_hierarchy(Hierarchy) :-
    get_section_from_current_machine(meta,Infos),
    member(hierarchy/Hierarchy,Infos),!.

b_get_refined_machine(M) :-
    get_section_from_current_machine(meta,Infos),
    member(refined_machine/M,Infos), !.


% TODO(DP,6.8.2008)
:- volatile b_machine_temp_predicate/1.
:- dynamic b_machine_temp_predicate/1.

set_temp_predicate(CustomPredicate) :-
  b_parse_machine_predicate(CustomPredicate,TypedPred),
  assert_temp_typed_predicate(TypedPred).

% set a temporary predicate as additional guard
set_temp_predicate(CustomPredicate,'$initialise_machine') :- !,set_temp_predicate(CustomPredicate).
set_temp_predicate(CustomPredicate,'$setup_constants') :- !,set_temp_predicate(CustomPredicate).
set_temp_predicate(CustomPredicate,OpName) :-
  b_parse_machine_operation_pre_predicate(CustomPredicate,TypedPred, OpName),
  assert_temp_typed_predicate(TypedPred).

assert_temp_typed_predicate(TypedPred) :-
  reset_temp_predicate,
  assert(b_machine_temp_predicate(TypedPred)).


reset_temp_predicate :- retractall(b_machine_temp_predicate(_)).

% -------------------

:- use_module(specfile,[unset_animation_minor_modes/1,reset_animation_minor_modes/1]).
b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes) :-
    full_b_machine(BMachine),
    (UnsetMinorModes==true -> unset_animation_minor_modes(L) ; L=[]),
    translate_machine(BMachine,Rep,AdditionalInfo),
    (UnsetMinorModes==true -> reset_animation_minor_modes(L) ; true).

:- use_module(tools,[get_tail_filename/2,split_filename/3,safe_atom_chars/3]).
b_write_machine_representation_to_file(Level,PPFILE) :-
    %get_machine(Name,M),
    get_tail_filename(PPFILE,Tail),
    split_filename(Tail,MachName,_Ext),
    b_write_machine_representation_to_file3(MachName,Level,PPFILE).

:- use_module(translate,[set_print_type_infos/1]).
b_write_machine_representation_to_file3(MachName,Level,PPFILE) :-
    debug_println(20,'% Pretty-Printing Internal Representation to File: '),
    debug_println(20,PPFILE),
    translate:set_print_type_infos(Level), % needed or all
    get_machine(_Name,M),
    (translate_machine_to_classicalb(MachName,M,Rep)
      -> translate:set_print_type_infos(none)
       ; add_internal_error('Translating machine failed: ',MachName),set_print_type_infos(none),
        fail),
    tools:write_to_file(PPFILE,Rep).

:- use_module(specfile).
translate_machine_to_classicalb(MachName,M,Rep) :-
   (animation_minor_mode(X)
    -> remove_animation_minor_mode,
       call_cleanup(translate_machine(machine(MachName,M),Rep,false), set_animation_minor_mode(X))
     ; translate_machine(machine(MachName,M),Rep,false)).

b_write_eventb_machine_to_classicalb_to_file(PPFILE) :-
  get_tail_filename(PPFILE,Tail),
  split_filename(Tail,MachName,_Ext),
  b_show_eventb_as_classicalb1(MachName,Rep,_AdditionalInfo),
  tools:write_to_file(PPFILE,Rep).

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).
b_show_eventb_as_classicalb1(MachName,Rep,AdditionalInfo) :-
  full_b_machine(machine(_Name,MachineBody)),
  % limit abstract level to 0
  b_limit_abstract_level_to_zero(MachineBody,MachineBodyLevel0),
  temporary_set_preference(translate_print_typing_infos,true,CHNG),
  (translate_eventb_to_classicalb(machine(MachName,MachineBodyLevel0),AdditionalInfo,Rep)
    -> preferences:reset_temporary_preference(translate_print_typing_infos,CHNG)
    ;  add_internal_error('Translating Event-B machine to Classical B failed: ',MachName),
       reset_temporary_preference(translate_print_typing_infos,CHNG),
       fail).

b_limit_abstract_level_to_zero(MachineBody,MachineBodyLevel0) :-
  maplist(b_set_to_level_zero,MachineBody,MachineBodyLevel0).

% TODO: set to level zero also the other events of the Event-B machine
b_set_to_level_zero(initialisation/InitEvent,initialisation/InitEventLevelZero) :-
  get_texpr_expr(InitEvent,rlevent(Id,Sec,St,Par,Grd,Thms,Act,VW,PW,Ums,_Abs)),
  get_texpr_info(InitEvent,Info),
  create_texpr(rlevent(Id,Sec,St,Par,Grd,Thms,Act,VW,PW,Ums,[]),subst,Info,InitEventLevelZero).
b_set_to_level_zero(X,X).

b_show_eventb_as_classicalb(Rep,AdditionalInfo) :-
    full_b_machine(BMachine),
    translate_eventb_to_classicalb(BMachine,AdditionalInfo,Rep).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% precompilation: predicates are precompiled by calling the ..._calc predicate

:- use_module(state_packing).
% maybe this predicate should be directly called by b_load_machine
b_machine_precompile :- debug_println(9,'Precompiling B Machine'),
    get_machine(_,_),
    auto_precompile,
    precompile_operation_names_in_reverse_order,
    % TODO(DP, 14.8.2008): remove reference to b_global_sets
    ~~mnf( b_global_sets:b_check_and_precompile_global_set_type_definitions ),
    % TODO(DP, 12.9.2008): remove reference to kernel_freetypes
    reset_freetypes,
    get_section_from_current_machine(freetypes,Freetypes),
    add_all_freetypes(Freetypes),
    state_packing:precompile_state_packing,
    do_machine_consistency_check,
    try_kodkod.
add_all_freetypes([]).
add_all_freetypes([freetype(Name,Cases)|Rest]) :-
    add_freetype(Name,Cases),
    add_all_freetypes(Rest).


other_spec_precompile :- retractall(bmachine_is_precompiled),
  /* call if you do not animate a B specification */
    precompiled_predicate(Pred/Arity),
    functor(Pattern,Pred,Arity),
    retractall(Pattern),
    (precompiled_predicate_no_error_when_not_precompiled(Pred/Arity)
      -> assert( (Pattern :- fail ))
      ;  assert( (Pattern :- print_message('No B machine available for '),print_message(Pattern),fail) )
    ),
    fail.
other_spec_precompile :-
    retractall(b_get_machine_constants(_)),
    assert( b_get_machine_constants([]) ),
    retractall(b_get_machine_variables(_)),
    assert( b_get_machine_variables([]) ),
    retractall(b_get_static_assertions_from_machine(_)),
    assert( b_get_static_assertions_from_machine([]) ),
    retractall(b_machine_has_static_assertions),
    retractall(b_get_unproven_static_assertions_from_machine(_)),
    assert( b_get_unproven_static_assertions_from_machine([]) ),
    retractall(b_get_dynamic_assertions_from_machine(_)),
    assert( b_get_dynamic_assertions_from_machine([]) ),
    retractall(b_machine_has_dynamic_assertions),
    retractall(b_get_unproven_dynamic_assertions_from_machine(_)),
    assert( b_get_unproven_dynamic_assertions_from_machine([]) ),
    retractall( b_get_machine_searchscope(_) ),
    retractall( b_get_machine_goal(_) ),
    assert_bmachine_is_precompiled, debug_println(4,other_spec_precompile).


precompiled_predicate(b_machine_name/1).
precompiled_predicate(b_get_named_machine_set/3).
precompiled_predicate(b_get_properties_from_machine/1).
precompiled_predicate(b_get_machine_all_constants/1).
precompiled_predicate(b_get_machine_set/2).
precompiled_predicate(b_get_disjoint_constants_of_type/3).
precompiled_predicate(b_get_constant_represented_inside_global_set/2).
precompiled_predicate(b_get_machine_constants/1).
precompiled_predicate(b_get_machine_variables/1).
precompiled_predicate(b_get_invariant_from_machine/1).
precompiled_predicate(b_get_linking_invariant_from_machine/1).
precompiled_predicate(b_is_constant/1).
precompiled_predicate(b_is_variable/2).
precompiled_predicate(b_get_static_assertions_from_machine/1).
precompiled_predicate(b_machine_has_static_assertions/0).
precompiled_predicate(b_get_unproven_static_assertions_from_machine/1).
precompiled_predicate(b_get_dynamic_assertions_from_machine/1).
precompiled_predicate(b_machine_has_dynamic_assertions/0).
precompiled_predicate(b_get_unproven_dynamic_assertions_from_machine/1).
precompiled_predicate(b_get_assertions_from_main_machine/2).
precompiled_predicate(b_get_initialisation_from_machine/2).
precompiled_predicate(b_get_machine_operation/5).
precompiled_predicate(b_top_level_operation/1).
precompiled_predicate(b_get_promoted_machine_operations/1).
precompiled_predicate(get_operation_info/2).
precompiled_predicate(b_get_machine_goal/1).
precompiled_predicate(b_get_machine_setscope/2).
precompiled_predicate(b_get_machine_searchscope/1).
precompiled_predicate(b_get_definition_string_from_machine/2).
precompiled_predicate(b_get_machine_animation_function/2).
precompiled_predicate(b_get_machine_heuristic_function/1).
precompiled_predicate(b_get_machine_custom_nodes_function/1).
precompiled_predicate(b_get_machine_custom_edges_function/1).
precompiled_predicate(complete_discharged_info/0). % just call it before calling b_specialized_invariant_for_op_calc !
precompiled_predicate(get_proven_invariant/2).
precompiled_predicate(b_specialized_invariant_for_op/2).
precompiled_predicate(b_get_operation_normalized_read_write_info/3).
precompiled_predicate(reuse_operation_effect/2).
precompiled_predicate(candidate_operation_for_reuse/1).
precompiled_predicate(b_get_all_used_identifiers_in_section/2).
precompiled_predicate(b_get_all_used_identifiers/1).
precompiled_predicate(b_get_all_used_identifiers/1).
precompiled_predicate(b_is_unused_constant/1).

% these functions sometimes get called also in csp mode; simply fail silently for those cases in other_spec_precompile
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_animation_function/2).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_heuristic_function/1).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_custom_nodes_function/1).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_custom_edges_function/1).

% on startup, all precompiled predicates yield error message
startup_precompiled :- retractall(bmachine_is_precompiled),
    precompiled_predicate(Pred/Arity),
    functor(Pattern,Pred,Arity),
    retractall(Pattern),
    assert( (Pattern :- add_error(bmachine,'B machine not precompiled for ',Pattern),fail) ),
    fail.
startup_precompiled :- debug_println(4,startup_precompiled).

:- startup_precompiled.

% for each predicate in "precompiled_predicate", auto_precompile computes
% all solutions by calling the predicate with an appended _calc
auto_precompile :-
    % failure-driven loop
    precompiled_predicate(Pred/Arity), %print(precompile(Pred/Arity)),nl,
    auto_precompile2(Pred,Arity,Pattern),
    (Pattern -> true /* this pattern already asserted */
             ; assert(Pattern)), fail.
auto_precompile :- assert_bmachine_is_precompiled, debug_println(4,auto_precompile).
% if the precompiled predicates is e.g. foo/2 then
% Pattern=foo(A,B) and Call=foo_calc(A,B)
auto_precompile2(Pred,Arity,Pattern) :-
    atom_concat(Pred,'_calc',CallFunctor),
    functor(Pattern,Pred,Arity),
    functor(Call,CallFunctor,Arity),
    unify_args(Arity,Call,Pattern),
    % delete all existing solutions
    retractall(Pattern), %print(retractall(Pattern)),nl,
    % the failure-driven loop for one predicate
    call(Call).
unify_args(0,_,_) :- !.
unify_args(N,A,B) :-
    arg(N,A,Arg), arg(N,B,Arg),
    N1 is N-1, unify_args(N1,A,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set machine goal

b_set_machine_goal(Goal) :- b_set_machine_goal(Goal,false).
b_set_machine_goal(Goal,WithDeferredSets) :-
    (WithDeferredSets=with_deferred -> Scope = [prob_ids(visible),variables] ; Scope = [variables]),
    b_parse_machine_predicate(Goal,Scope,TypedPred),
    b_set_parsed_typed_machine_goal(TypedPred).
b_set_parsed_typed_machine_goal(TypedPred) :-
    retractall(b_get_machine_goal(_)),
    assert(b_get_machine_goal(TypedPred)).


b_set_machine_searchscope(Goal) :- b_set_machine_searchscope(Goal,with_deferred).
b_set_machine_searchscope(Goal,WithDeferredSets) :-
    (WithDeferredSets=with_deferred -> Scope = [prob_ids(visible),variables] ; Scope = [variables]),
    b_parse_machine_predicate(Goal,Scope,TypedPred),
    b_set_parsed_typed_machine_searchscope(TypedPred).
b_set_parsed_typed_machine_searchscope(TypedPred) :-
    retractall(b_get_machine_searchscope(_)),
    assert(b_get_machine_searchscope(TypedPred)).

% also allow empty string
b_parse_optional_machine_predicate(TargetString,Target) :-
    is_empty_string(TargetString),!,create_texpr(truth,pred,[],Target).
b_parse_optional_machine_predicate(TargetString,Target) :-
    b_parse_machine_predicate(TargetString,Target).

is_empty_string(Atom) :-
    atom_codes(Atom,Codes),
    is_empty_string2(Codes).
is_empty_string2([]).
is_empty_string2([32|Codes]) :- is_empty_string2(Codes).

% parsing a string as predicate with deferred set elements

b_parse_machine_predicate(Pred,TypedPred) :- b_parse_machine_predicate(Pred,[variables],TypedPred).
b_parse_machine_predicate(Pred,Scope,TypedPred) :-
    atom_codes(Pred,Codes),
    b_parse_machine_predicate_from_codes(Codes,Scope,TypedPred).
b_parse_machine_predicate_from_codes(Codes,Scope,TypedPred) :-
    b_parse_machine_predicate_from_codes2(Codes,[],Scope,TypedPred,closed).
b_parse_machine_predicate_from_codes_open(OptionalQuantifier,Codes,Defs,Scope,TypedPred) :-
    b_parse_machine_predicate_from_codes2(Codes,Defs,Scope,TypedPred,open(OptionalQuantifier)).

b_parse_machine_predicate_from_codes2(Codes,Defs,Scope,TypedPred,Mode) :-
    debug_println(9,parse_exprs_preds(Mode)),
    catch( ((Defs=[] -> true ; print('% Ignoring DEFINITIONS'),nl),
               parse_predicate(Codes,Parsed)), % ,print(Parsed),nl),
           Exception,
           ( Exception = parse_errors(Errors) -> %print(errors(Errors)),nl,
               add_all_perrors(Errors,[],parse_machine_predicate_error),fail
           ; otherwise ->
               add_error(bmachine,'Exception while parsing predicate: ',Exception),
               fail)),
    debug_println(9,type_with_errors(Parsed)),
    ( Mode == closed ->
        type_with_errors(Parsed,Scope,pred,TypedPred)
    ; Mode = open(OptionalQuantifier) ->
        b_type_open_predicate_with_errors(OptionalQuantifier,Parsed,Scope,TypedPred)).

b_parse_machine_operation_pre_predicate(Pred,TypedPred,OpName) :-
    atom_codes(Pred,Codes),
    catch( parse_predicate(Codes,Parsed),
           Exception,
           ( Exception = parse_errors(Errors) ->
               add_all_perrors(Errors,[],parse_pre_predicate_error),fail
           ; otherwise ->
               add_error(bmachine,'Exception while parsing predicate: ',Exception),
               fail)),
    ( b_get_machine_operation(OpName,_,V,_) -> true
    ; add_error_fail(b_parse_machine_operation_pre_predicate,'Unknown operation:', OpName)),
    Scope = [identifier(V),variables],
    print(typing(Parsed)),nl,
    type_with_errors(Parsed,Scope,pred,TypedPred).

b_parse_machine_expression_from_codes_with_prob_ids(Codes,TypedExpr) :-
    b_parse_machine_expression_from_codes(Codes,[prob_ids(visible),variables],TypedExpr,_Type,true,Error),!,
    (Error=none -> true
     ; add_error(b_parse_machine_expression_from_codes_with_prob_ids,'Errors occured during parsing: ', Error),fail).
b_parse_machine_expression_from_codes(Codes,TypedExpr) :-
    b_parse_machine_expression_from_codes(Codes,[variables],TypedExpr,_Type,true,Error),!,
    (Error=none -> true
     ; add_error(b_parse_machine_expression_from_codes,'Errors occured during parsing: ', Error),fail).


%b_parse_machine_expression_from_codes(Codes,TypedExpr) :-
%    b_parse_machine_expression_from_codes(Codes,TypedExpr,_Type,false,Error), Error=none.

b_parse_machine_expression_from_codes(Codes,Typed,Type,GenerateParseErrors,Error) :-
    b_parse_machine_expression_from_codes(Codes,[variables],
                                          Typed,Type,GenerateParseErrors,Error).
b_parse_machine_expression_from_codes(Codes,Scope,
                                      Typed,Type,GenerateParseErrors,Error) :-
    catch( parse_expression(Codes,Parsed),
           Exception,
           ( GenerateParseErrors=true,
               ( Exception = parse_errors(Errors) ->
                   add_all_perrors(Errors,[],parse_machine_expression_error),
                   Error=parse_error
               ; otherwise ->
                   add_error(bmachine,'Exception while parsing expression: ',Exception),
                   Error=exception))),
    %print(parsed(Parsed,Error)),nl,
    ( nonvar(Error) -> true /* either exception or parse error occured */
      ; type_with_errors(Parsed,Scope,Type,Typed) -> Error=none
     ; otherwise -> Error=type_error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% consistency check to spot some errors
do_machine_consistency_check :-
    specfile:animation_mode(b),            % TODO (dp,27.09.2012): this is just an ugly hack
    minor_mode_to_check,
    %debug:time(bmachine:all_operations_have_valid_read_modifies_infos).
    !,
    all_operations_have_valid_read_modifies_infos.
do_machine_consistency_check.

% In future, all formalisms should provide the relevent data
minor_mode_to_check :- specfile:animation_minor_mode(eventb),!.
minor_mode_to_check :- specfile:animation_minor_mode(_),!,fail.
minor_mode_to_check. % classical B

all_operations_have_valid_read_modifies_infos :-
    %findall(O,b_get_machine_operation(O,_,_,_),OpNames),
    %maplist(operation_has_valid_read_modifies_infos,OpNames).
    b_get_machine_operation(OpName,_,_,_),
    operation_has_valid_read_modifies_infos(OpName),fail.
all_operations_have_valid_read_modifies_infos.

operation_has_valid_read_modifies_infos(OpName) :- % print(check(OpName)),nl,
    ( operation_has_valid_read_modifies_infos2(OpName) -> true
    ; otherwise ->
        add_failed_call_error(operation_has_valid_read_modifies_infos2(OpName))).
operation_has_valid_read_modifies_infos2(OpName) :-
    b_get_operation_normalized_read_write_info(OpName,Reads,Writes),
    maplist(check_read(OpName),Reads),
    %print(operation(OpName,reads(Reads),writes(Writes))),nl,
    maplist(check_write(OpName),Writes).


check_read(OpName,Id) :-
    ( (check_is_var(Id);check_is_constant(Id)) -> true
    ; otherwise ->
        add_internal_error(bmachine,'Unknown identifier in operation "reads" information ',
                           OpName/Id,unknown)).
check_write(OpName,Id) :-
    ( check_is_var(Id) -> true
    ; otherwise ->
        add_internal_error(bmachine,'Unknown variable in operation "modifies" information ',
                           OpName/Id,unknown)).

check_is_var(Id) :-
    b_is_variable(Id,_).
    %get_texpr_id(E,Id),
    %b_get_machine_variables(Vars),
    %memberchk(E,Vars).
check_is_constant(Id) :- b_is_constant(Id).
check_is_constant(Id) :- b_get_constant_represented_inside_global_set(Id,_).
   % b_get_machine_all_constants(Constants),
   % get_texpr_id(TId,Id),
   % memberchk(TId,Constants).
check_is_constant(Id) :- b_get_machine_set(Id).
check_is_constant(Id) :- b_get_named_machine_set(Id,_).
check_is_constant(Id) :-
    b_get_named_machine_set(_,Elems),member(Id,Elems).


% provide access to number of constants, ....:
b_machine_statistics(constants,Nr) :- b_get_machine_constants(L), length(L,Nr).
b_machine_statistics(variables,Nr) :- b_get_machine_variables(L), length(L,Nr).
b_machine_statistics(properties,Nr) :-
   b_get_properties_from_machine(C), conjunction_to_list(C,L),length(L,Nr).
b_machine_statistics(invariants,Nr) :-
   b_get_invariant_from_machine(C), conjunction_to_list(C,L),length(L,Nr).
b_machine_statistics(static_assertions,Nr) :-
   b_get_static_assertions_from_machine(L), length(L,Nr).
b_machine_statistics(dynamic_assertions,Nr) :-
   b_get_dynamic_assertions_from_machine(L), length(L,Nr).
b_machine_statistics(operations,Nr) :- findall(N,b_get_machine_operation(N,_,_,_),L),length(L,Nr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% default and initial machines
b_set_initial_machine :-
    initial_machine(Main,Machines),
    b_set_machine(Main,Machines,[]).

initial_machine(phonebook,[M]) :-
    M = abstract_machine(none,machine(none),machine_header(none,phonebook,[]),Body),
    Body = [sets(none, [deferred_set(none,'Name'),
                        deferred_set(none,'Code')]),
            definitions(none,[ScopeName,ScopeCode,TraceTest0,TraceTest1]),
            variables(none, [identifier(none, db)]),
            invariant(none, member(none,
                                   identifier(none, db),
                                   partial_function(none,
                                                    identifier(none, 'Name'),
                                                    identifier(none, 'Code')))),
            initialisation(none,
                           assign(none, [identifier(none, db)], [empty_set(none)])),
            operations(none, [Add,Lookup,Update,Reset])],

    % definitions
    ScopeName = expression_definition(none,
                                      scope_Name,
                                      [],
                                      interval(none, integer(none, 1), integer(none, 3))),
    ScopeCode = expression_definition(none,
                                      scope_Code,
                                      [],
                                      interval(none, integer(none, 1), integer(none, 2))),
    TraceTest0 = expression_definition(none, trace_Test0, [], empty_sequence(none)),
    TraceTest1 = expression_definition(none, trace_Test1, [], sequence_extension(none, TraceSeq)),
    TraceSeq = [identifier(none,db),identifier(none,'Name1'),identifier(none,'Code7'),
                identifier(none,update),identifier(none,'Name1'),identifier(none,'Code8')],

    % add operation
    Add = operation(none,
                    identifier(none, add),
                    [],
                    [identifier(none,cc),identifier(none,nn)],
                    precondition(none, AddPre, AddAssign)),
    AddPre = conjunct(none,
                      conjunct(none,
                               member(none, identifier(none,nn), identifier(none,'Name')),
                               member(none, identifier(none,cc), identifier(none,'Code'))),
                      not_member(none,
                                 identifier(none,nn),
                                 domain(none, identifier(none,db)))),
    AddAssign = assign(none,
                       [identifier(none,db)],
                       [union(none,
                              identifier(none,db),
                              set_extension(none,
                                            [couple(none,[identifier(none,nn),
                                                          identifier(none,cc)])]))]),

    % lookup operation
    Lookup = operation(none,
                       identifier(none, lookup),
                       [identifier(none, cc)],
                       [identifier(none, nn)],
                       precondition(none, LookupPre, LookupAssign)),
    LookupPre = member(none, identifier(none,nn), domain(none,identifier(none,db))),
    LookupAssign = assign(none,
                          [identifier(none, cc)],
                          [function(none, identifier(none, db), identifier(none, nn))]),

    % update operation
    Update = operation(none,
                       identifier(none, update),
                       [],
                       [identifier(none,nn),identifier(none,cc)],
                       precondition(none, UpdatePre, UpdateAssign)),
    UpdatePre = conjunct(none,
                         conjunct(none,
                                  member(none, identifier(none,nn), identifier(none,'Name')),
                                  member(none, identifier(none,cc), identifier(none,'Code'))),
                         member(none,
                                identifier(none,nn),
                                domain(none, identifier(none,db)))),
    UpdateAssign = assign(none,
                          [function(none, identifier(none,db), identifier(none,nn))],
                          [identifier(none, cc)]),

    % reset operation
    Reset = operation(none,
                      identifier(none,reset),
                      [],
                      [],
                      precondition(none, ResetPre, ResetAssign)),
    ResetPre = member(none,
                      identifier(none, db),
                      total_function(none, identifier(none,'Name'), identifier(none,'Code'))),
    ResetAssign = assign(none,[identifier(none,db)],[empty_set(none)]).

empty_machine(empty_machine,[M]) :-
    M = abstract_machine(none,machine(none),machine_header(none,empty_machine,[]),Body),
    Body = [].
