% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(haskell_csp_analyzer,[is_csp_process/1, is_possible_csp_process/1, agent_functor/2, is_csp_constructor/1,
				cspPrintCompiled/2, channel_type_list/2, csp_full_type_constructor/3,csp_full_type_constant/2,
				%% csp_sub_type_constructor/3, csp_sub_type_constant/2, % reported as unnecessary exports by infolog
				name_type/2, agent_compiled/3, csp_constant/2, analyze/0,
				undefined_process_construct/1, definite_cspm_process_construct/4, csp_constructor/3,
				is_a_channel_name/1,agent_parameters/3, is_list_skeleton/1,
				lift/4, is_defined_agent/1,
				definite_cspm_process_expression/1, 
        get_internal_csp_representation/1,
				%% input_agent_compiled/2, % reported as an unnecessary export by infolog
        %% blocking_append/3, % reported as an unnecessary export by infolog
        mynumbervars/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,csp).
:- module_info(description,'Pre-compiler and analyzer for CSP.').

/******** SICSTUS libraries ********/
:- use_module(library(lists)).
:- use_module(library(codesio)).
/******** ----------------- ********/

/************** PROB modules **************/
:- use_module(probsrc(self_check)).
:- use_module(probsrc(debug),[debug_println/2, debug_mode/1]).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(tools), [string_concatenate/3,remove_variables/3]).
:- use_module(probsrc(gensym),[gensym/2]).
:- use_module(probsrc(tools),[list_difference/3,exact_member/2]).
%--------- CSP modules:
:- use_module(probcspsrc(haskell_csp),[add_symbol_error/4, is_a_datatype/2, 
		channel/2, bindval/3, agent/3, agent_curry/3, dataTypeDef/2, subTypeDef/2, nameType/2,
		symbol/4, cspTransparent/1, cspPrint/1, valid_constant/1,
		extract_span_info/2,
		evaluate_type_list/2, evaluate_argument/2]).
:- use_module(probcspsrc(csp_sets),[extract_variables_from_generator_list/2]).
:- use_module(probcspsrc(csp_tuples),[pattern_match_function_argument/3, is_constructor/3]).
/************** ------------ **************/

:- dynamic is_csp_constructor/1.
:- dynamic csp_full_type_constructor/3.
:- dynamic csp_sub_type_constructor/3.
:- dynamic csp_full_type_constant/2.
:- dynamic csp_sub_type_constant/2.

undefined_process_construct(X) :- \+ definite_cspm_process_construct(X,_,_,_),
   \+ potential_cspm_process_construct(X,_,_,_).
/*defined_process_constructs(R) :- findall(F/N,(haskell_csp:clause(cspm_trans(X,_,_),_),nonvar(X), functor(X,F,N)),L),sort(L,R).
  haskell_csp:defined_process_constructs(L), write_term(L,[quoted(true)]),nl. */

%:- dynamic cspm_trans/3.
/* findall(F/N,(haskell_csp:clause(cspm_trans(X,_,_),_),nonvar(X), functor(X,F,N)),L),sort(L,R),
   member(F2/N2,R), functor(X2,F2,N2) ,write_term(cspm_process_construct(X2),[quoted(true)]), print('.'), nl, fail  */
 
% definite_cspm_process_construct(Construct, Span(s), OtherSubConstructs, DefiniteCSPSubConstructs)
definite_cspm_process_construct(&(Grd,B),          [],[Grd],[B]).
%%definite_cspm_process_construct('&'(_G,B),[_G],[B]).
definite_cspm_process_construct('/\\'(A,B,Span),  [Span],[],[A,B]).
definite_cspm_process_construct(';'(A,B,Span),    [Span],[],[A,B]).
definite_cspm_process_construct('[>'(A,B,Span),   [Span],[],[A,B]).
definite_cspm_process_construct('[]'(A,B,Span),   [Span],[],[A,B]).
definite_cspm_process_construct('\\'(A,Set,Span), [Span],[Set],[A]).
definite_cspm_process_construct(ehide(A,Set,Span),[Span],[Set],[A]).
definite_cspm_process_construct(exception(Set,A,B,Span), [Span],[Set],[A,B]).
definite_cspm_process_construct(eexception(Set,A,B,Span),[Span],[Set],[A,B]).
definite_cspm_process_construct(omega,            [],[],[]).
definite_cspm_process_construct(skip(Span),       [Span],[],[]).
definite_cspm_process_construct(stop(Span),       [Span],[],[]).
definite_cspm_process_construct('CHAOS'(Span,B),  [Span],[B],[]).
definite_cspm_process_construct('|||'(A,B,Span),  [Span],[],[A,B]).
definite_cspm_process_construct('|~|'(A,B,Span),  [Span],[],[A,B]).
definite_cspm_process_construct(val_of(A,Span),   [Span],[],[A]).
definite_cspm_process_construct(eprocRenaming(RenList,B,Span),[Span],[RenList],[B]).
definite_cspm_process_construct(prefix(Span,A,B,C,Span2),  [Span,Span2],[A,B],[C]).
definite_cspm_process_construct(procRenaming(RenList,B,Span),[Span],[RenList],[B]).
definite_cspm_process_construct(repSequence(Gen,A,Span),      [Span],[Gen],[A]).
definite_cspm_process_construct(elinkParallel(A,B,C,Span),  [Span],[A],[B,C]).
definite_cspm_process_construct(repChoice(Gen,A,Span),        [Span],[Gen],[A]).
definite_cspm_process_construct(repInterleave(Gen,A,Span),    [Span],[Gen],[A]).
definite_cspm_process_construct(repInternalChoice(Gen,A,Span),[Span],[Gen],[A]).
definite_cspm_process_construct(lParallel(LinkList,B,C,Span),[Span],[LinkList],[B,C]).
definite_cspm_process_construct(sharing(Set,B,C,Span),       [Span],[Set],[B,C]).
definite_cspm_process_construct(esharing(Set,B,C,Span),      [Span],[Set],[B,C]).
definite_cspm_process_construct(aParallel(E,A,C,B,Span),     [Span],[E,C],[A,B]).
definite_cspm_process_construct(eaParallel(C,A,D,B,Span),    [Span],[C,D],[A,B]).
definite_cspm_process_construct(procRepAParallel(Generators,pair(SYNC,A),Span),[Span],[Generators,SYNC],[A]).
definite_cspm_process_construct(procRepLinkParallel(Generators,SYNC,A,Span),   [Span],[Generators,SYNC],[A]).
definite_cspm_process_construct(procRepSharing(SYNC,Generators,A,Span),       [Span],[SYNC,Generators],[A]).
definite_cspm_process_construct(procRenamingComp(A,Gen,Ren),[],[Gen,Ren],[A]).
definite_cspm_process_construct(builtin_call(X), Span,Gen,P) :- definite_cspm_process_construct(X,Span,Gen,P).
definite_cspm_process_construct(P,[],[],[]) :- is_csp_process(P).
% potential CSPM constructs:

% potential_cspm_process_construct(Construct, Span(s), OtherSubConstructs, PossibleCSPSubConstructsIfConstructItselfIsCSP)
potential_cspm_process_construct(head(L),[],[],[L]).
potential_cspm_process_construct(ifte(G,B,C,S1,S2,S3),[S1,S2,S3],[G],[B,C]). % version with source-spans
potential_cspm_process_construct(agent_call(Span,A,Par),      [Span],[A],[Par]).
potential_cspm_process_construct(agent_call_curry(A,Par),     [],    [A],[Par]).




:- dynamic is_function/1.
:- dynamic is_csp_process/1.
:- dynamic is_possible_csp_process/1.
:- dynamic change_happened/0.
:- dynamic cspPrintCompiled/2.

definite_cspm_process_expression(X) :- var(X),!,fail.
definite_cspm_process_expression(ifte(_Tst,T,E,_,_,_)) :- !, 
   (definite_cspm_process_expression(T) ; definite_cspm_process_expression(E)). /* assume no typing error */
% let(_,E) and val_of(X,_) will be precompiled to agents, therefore the two clauses below will never matched
%definite_cspm_process_expression(let(_,E)) :- !,
%    definite_cspm_process_expression(E). 
%definite_cspm_process_expression(val_of(X,_)) :- !,is_csp_process(X).
definite_cspm_process_expression(agent_call(_Span,F,Par)) :- 
  atomic(F), !,   /* Note: with currying we can have expressions here !   TO DO: ADAPT */
  X=..[F|Par],is_csp_process(X).
definite_cspm_process_expression(agent_call(_Span,lambda(_Args,Body),_Par)) :- !, 
  %%print(check(Body)),nl,
    definite_cspm_process_expression(Body).
definite_cspm_process_expression(agent_call_curry(F,[Par1|_])) :- !, 
  atomic(F),  /* Note: with currying we can have expressions here !   TO DO: ADAPT */
  X=..[F|Par1],is_csp_process(X).
definite_cspm_process_expression(head(X)) :- !, /* shouldn't X be a list of CSP constructs ??? */
   definite_cspm_process_expression(X).
definite_cspm_process_expression(lambda(_Head,Body)) :- \+var(Body), !, definite_cspm_process_construct(Body,_,_,_).
definite_cspm_process_expression(X) :- definite_cspm_process_construct(X,_,_,_).

possible_cspm_process_expression(X) :- var(X),!.
possible_cspm_process_expression(ifte(_Tst,T,E,_,_,_)) :- !, 
   (possible_cspm_process_expression(T) ; possible_cspm_process_expression(E)). /* assume no typing error */
% the same as above for definite_cspm_process_expression let(_,E) and val_of(X,_) (see comment above)
%possible_cspm_process_expression(let(_,E)) :- !,
%    possible_cspm_process_expression(E). 
%possible_cspm_process_expression(val_of(X,_)) :- !, is_possible_csp_process(X).
possible_cspm_process_expression(agent_call(_Span,F,Par)) :- atomic(F),!,   
    X=..[F|Par], is_possible_csp_process(X).
possible_cspm_process_expression(agent_call(_,_,_)). /* TO DO: refine this */
possible_cspm_process_expression(agent_call_curry(_,_)). /* TO DO: refine this */
possible_cspm_process_expression(head(_)).

check_definition_body(X,_,_) :- var(X),!,true.
check_definition_body('$VAR'(X),_,_) :- !, cur_numbervars_index(Idx), X =< Idx.
% let expressions will be first precomiled into agent calls (case for check_definition_body/3 will never occur)
%check_definition_body(let(_,E),Def,Span) :- !,
%    check_definition_body(E,Def,Span).
check_definition_body(lambda(_,E),Def,Span) :- !,
    check_definition_body(E,Def,Span).
%check_definition_body(head(L),Def,Span) :- !,  %% covered below  by potential_cspm_process_construct
% val_of(_,X) expressions will be first precomiled into agent calls (case for check_definition_body/3 will never occur)
%check_definition_body(val_of(X,CallSpan),_Def,_Span) :- !,
%  (is_defined_value(X) -> true ; add_internal_error(haskell_csp_analyzer,'Undefined Value: ',X,CallSpan)).
check_definition_body(agent_call(CallSpan,F,Par),_Def,_Span) :- atomic(F),!, X=..[F|Par],
  ((is_defined_agent(X); is_builtin_agent(X)) -> true ; add_error(haskell_csp_analyzer,'Undefined Agent Call: ',X,CallSpan)).
check_definition_body(agent_call(CallSpan,X,Par),_Def,_Span) :- 
   nonvar(X), X=lambda(Var,Body),!,
   length(Par,PL), length(Var,VL),
   (PL=VL -> true
   ; add_error(haskell_csp_analyzer,'Lambda parameters and arguments do not match: ',(Par,Var,Body),CallSpan)).
check_definition_body(agent_call(CallSpan,F,Par),Def,_Span) :- !, 
   (var(F) -> true ; 
      (F=agent_call(Span2,F2,Par2)
        -> check_definition_body(agent_call(Span2,F2,Par2),Def,CallSpan)
       ;   add_error(haskell_csp_analyzer,'Complicated LHS not yet supported in agent_call: ',
                       agent_call(F,Par),CallSpan)
      )
   ).
check_definition_body(X,Def,Span) :-
   definite_cspm_process_construct(X,Spans,OtherSubExprs,List),!,
   l_check_spans(Spans,Span), l_check_non_csp_args(OtherSubExprs,Span),
   l_check_definite_csp_args(List,Def,Span).
check_definition_body(X,Def,Span) :- 
   potential_cspm_process_construct(X,Spans,OtherSubExprs,List),
   l_check_spans(Spans,Span), l_check_non_csp_args(OtherSubExprs,Span),
   l_check_possible_csp_args(List,Def,Span).

check_definite_cspm_body_construct(X,D,S) :-
    check_definition_body(X,D,S),!.
check_definite_cspm_body_construct(X,_Def,Span) :-
   add_internal_error(haskell_csp_analyzer,'Unknown CSP construct: ',X,Span).

l_check_spans([],_Span).
l_check_spans([H|T],OuterSpan) :- 
   (extract_span_info(H,_Info) -> true
             ; add_internal_error('Illegal AST; unknown Span info: ',H:OuterSpan)),
   l_check_spans(T,OuterSpan).

l_check_non_csp_args([],_Span).
l_check_non_csp_args([H|T],OuterSpan) :- 
   (\+ definite_cspm_process_expression(H) -> true
             ; (definite_cspm_process_construct(H,[Span|_],_,_) -> true ; Span=OuterSpan),
                add_error(haskell_csp_analyzer,'Illegal AST; CSP process in illegal position: ',H,Span)),
   l_check_non_csp_args(T,OuterSpan).
l_check_definite_csp_args([],_,_Span).
l_check_definite_csp_args([H|T],Def,Span) :- 
    check_definite_cspm_body_construct(H,Def,Span),l_check_definite_csp_args(T,Def,Span).
l_check_possible_csp_args([],_,_Span).
l_check_possible_csp_args([H|T],Def,Span) :- 
    check_definition_body(H,Def,Span),l_check_possible_csp_args(T,Def,Span).

is_defined_agent(EX) :- var(EX),!,fail.
is_defined_agent(EX) :- functor(EX,F,N), functor(Skel,F,N), 
    agent_compiled_without_constraints(Skel,_,_). % avoid executing Constraint
%is_defined_value(X) :- bindval(X,_,_).
 
is_builtin_agent(Term) :- var(Term),!,fail.
is_builtin_agent(Term) :-
  member(Term,[ union(_,_), inter(_,_), diff(_,_), 'Union'(_), 'Inter'(_), member(_,_), card(_), empty(_),
                set(_), head(_), tail(_), concat(_), null(_), elem(_,_), length(_), 'Set'(_), 'Seq'(_), seq(_)]),!.


agent_compiled_without_constraints(Head,Body,Span) :-
   clause(agent_compiled(Head,Body,Span),_Constr).
agent_compiled_with_constraints(Head,Body,Span,Constr) :-
   clause(agent_compiled(Head,Body,Span),Constr).
 
/* compute which values/agents are CSP processes and which ones are just ordinary functions/values */
analyze :-  reset_csp_analyzer,
            precompile_datatypes(init),% initialiazing constructors before starting to check the process bodies
            precompile_constants,
	          compile_nested_let_expressions,
            opt_csp_listing,
            print('% Finding Definite CSP Processes:'),
            analyze_recursively,
            nl,print('% Finding Possible CSP Processes: '),
            analyze_recursively_possible,
            check,
            precompile_nametypes, /* must be put before datatypes */
            precompile_datatypes(eval),
            generate_channel_type_lists,
            precompile_cspPrint.

reset_csp_analyzer :-
   retractall(is_csp_process(_)),
   retractall(is_function(_)),
   retractall(is_possible_csp_process(_)),
   retractall(multiple_let_equations_for(_,_)),
   retractall(cspPrintCompiled(_,_)),
   retractall(csp_full_type_constructor(_,_,_)),
   retractall(csp_sub_type_constructor(_,_,_)),
   retractall(csp_full_type_constant(_,_)),
   retractall(csp_sub_type_constant(_,_)),
   retractall(is_csp_constructor(_)),
   retractall(channel_type_list(_,_)),
   retractall(name_type(_,_)),
   retractall(agent_functor(_,_)),
   retractall(agent_compiled(_,_,_)), 
   retractall(agent_parameters(_,_,_)),
   retractall(input_agent_compiled(_,_)).			

precompile_cspPrint :- cspPrint(Expr),   
    haskell_csp_analyzer:compile_body(Expr,'print',[],[],CompiledExpr),
    assert(cspPrintCompiled(Expr,CompiledExpr)),
    fail.
precompile_cspPrint.

analyze_recursively :- 
            retractall(change_happened),
            agent_compiled_without_constraints(X,Body,_Span),
            functor(X,F,N),functor(X2,F,N),
            \+ is_csp_process(X2),
            (is_function(X2) -> true ; assert(is_function(X2))),
            definite_cspm_process_expression(Body),
            assert(is_csp_process(X2)),
            assert(change_happened),
            print(' '),print_agent(X2),
            fail.
analyze_recursively :-   print(';'),
       (change_happened -> analyze_recursively ; nl).
            
analyze_recursively_possible :- 
            retractall(change_happened),
            is_function(X2), 
            \+ is_csp_process(X2), \+ is_possible_csp_process(X2),
            findall(Body,agent_compiled_without_constraints(X2,Body,_Span),AllBodies),
            maplist(possible_cspm_process_expression,AllBodies),
            assert(is_possible_csp_process(X2)),
            assert(change_happened),
            print(' '),print_agent(X2),
            fail.
analyze_recursively_possible :-   print(';'),
       (change_happened -> analyze_recursively_possible ; nl).
            

check :- print('% Checking Definitions:'),
       findall(V,bindval(V,_,_),LBV),
       check_for_multiples(LBV),
       findall(W,(agent(W,_,_);agent_curry(W,_,_)),LAG),
       check_for_contig(LAG),
           % Note: we cannot detect if a function definition is interrupted by a bindval or channel def
       nl,
       debug_println(9,'% Checking Definition bodies:'),
       agent_compiled_without_constraints(X,Body,Span),
       (debug_mode(on) -> print(' '),print_agent(X) ; true),
       check_definition_body(Body,X,Span),
       fail.
 check :- (debug_mode(on) -> nl ; true).

% ----------------------------------------------



csp_constant(C,T) :- csp_full_type_constant(C,T).
csp_constant(C,T) :- csp_sub_type_constant(C,T).

csp_constructor(C,T,A) :- csp_full_type_constructor(C,T,A).
csp_constructor(C,T,A) :- csp_sub_type_constructor(C,T,A).

csp_full_type_constant(apples,'FRUIT'). csp_full_type_constant(oranges,'FRUIT'). csp_full_type_constant(pears,'FRUIT'). % complementary to default CSP-M spec

precompile_datatypes(E) :-
  (E = eval -> print('% Precompiling datatype constructor types: '); print('% Initialising datatype constructor types: ')),
  retractall(csp_full_type_constructor(_,_,_)),
  dataTypeDef(Type,ConstructorList), print(Type), print(' '),
  member(X,ConstructorList),
  (X = constructor(_)
    -> fail /* already asserted */
    ;  (X = 'constructorC'(C,dotTupleType(CT)),(E = eval -> evaluate_type_list(CT,CTypes) ; true)
         -> assert(csp_full_type_constructor(C,Type,CTypes))
         ;  (X = 'constructorC'(C,typeTuple([dotTuple(CT)])),(E = eval -> evaluate_type_list(CT,CTypes) ; true) % can appear for example for sq.(x.y)
               -> assert(csp_full_type_constructor(C,Type,CTypes))
               ;  (X = 'constructorC'(C,typeTuple(CT)),(E = eval -> evaluate_type_list(CT,CTypes) ; true)
                    -> assert(csp_full_type_constructor(C,Type,[typeTuple(CTypes)]))
                    ;  add_internal_error('Unknown constructor in precompile_datatypes: ',(Type:X))
                  )
            )
       )
  ),
  fail.
precompile_datatypes(E) :- nl,
  retractall(csp_sub_type_constructor(_,_,_)),
  subTypeDef(Type,ConstructorList), print(Type), print(' '),
  member(X,ConstructorList),
  (X = constructor(_)
    -> fail /* already asserted */
    ;  (X = 'constructorC'(C,dotTupleType(CT)),(E = eval -> evaluate_type_list(CT,CTypes) ; true) % eval only when all declarations are pre-compiled
         -> assert(csp_sub_type_constructor(C,Type,CTypes))
         ;  (X = 'constructorC'(C,typeTuple([dotTuple(CT)])),(E = eval -> evaluate_type_list(CT,CTypes) ; true) % can appear for example for sq.(x.y)
               -> assert(csp_sub_type_constructor(C,Type,CTypes))
               ;  (X = 'constructorC'(C,typeTuple(CT)),(E = eval -> evaluate_type_list(CT,CTypes) ; true)
                     -> assert(csp_sub_type_constructor(C,Type,[typeTuple(CTypes)]))
                      ;  add_error(precompile_datatypes,'Unknown constructor: ',(Type:X))
                  )
            )
       )
  ),
  fail.
precompile_datatypes(_E) :- nl.

/* can be executed before definitions are parsed; as no need to call evaluate_list */
precompile_constants :- print('% Precompiling datatype constants: '),
  retractall(csp_full_type_constant(_,_)),
  retractall(csp_sub_type_constant(_,_)),
  retractall(is_csp_constructor(_)),
  dataTypeDef(Type,ConstructorList), print(Type), print(' '),
  member(constructor(C),ConstructorList),
  assert(csp_full_type_constant(C,Type)),
  fail.
precompile_constants :-
  is_a_datatype(_Type,ConstructorList),
  member('constructorC'(C,_),ConstructorList),
  assert(is_csp_constructor(C)),fail.  
precompile_constants :-
  subTypeDef(SubType,ConstructorList), print(SubType), print(' '),
  member(constructor(C),ConstructorList),
  (csp_full_type_constant(C,_SuperType) -> true
   ; add_error(precompile_constants,'Subtype introduces new constant: ',(SubType:C))
  ),
  assert(csp_sub_type_constant(C,SubType)),
  fail.
precompile_constants :- nl.
 
% ----------------------------------------------
 
:- dynamic channel_type_list/2, name_type/2.

channel_type_list(left,[dataType('FRUIT')]).
channel_type_list(right,[dataType('FRUIT')]).
channel_type_list(mid,[dataType('FRUIT')]).
channel_type_list(ack,[]).
channel_type_list(a,[intType]).  /* for test cases */
channel_type_list(b,[intType]).  /* for test cases */
channel_type_list(c,[intType]).  /* for test cases */
channel_type_list(gen1,[intType,intType]).  /* for test cases */
channel_type_list(gen2,[intType,intType]).  /* for test cases */

/* test cases for the is_not_infinite_type/1 predicate in haskell_csp module  (see lines 374 - 398) */

channel_type_list(left,[dataType('FRUIT')]).
channel_type_list(mid,[dataType('FRUIT')]).
channel_type_list(right,[dataType('FRUIT')]).
channel_type_list(receive,[dataType('SubMsg')]).
channel_type_list(contact,[dataType('SubSubMsg'),intType,boolType]).
channel_type_list(infinite,[dataType('Inf')]).

is_a_channel_name(Ch) :- channel(Ch,_). % use channel/2 rather than channel_type_list/2, as channel/2 available before precompilation

 /* TO DO: check if there can be a nesting of associative dotTuples and non-associative Tuples */
generate_channel_type_lists :- print('% Analyzing channels:'),
     retractall(channel_type_list(_,_)),
     channel(X,CType), print(' '),print(X),
     (CType = type(Type) -> true ; add_error_fail(haskell_csp_analyzer,'Channel type in wrong format: ',channel(X,Type))),
     (Type = dotTupleType(List)
       -> ((evaluate_type_list(List,EList),flatten_dotTupleTypes(EList,FlatEList))
           -> assert(channel_type_list(X,FlatEList)) %% EList ????? <----------
               %, nl, print(dotTupleType(List,EList,FlatEList)),nl
           ; add_symbol_error(haskell_csp_analyzer,'Could not evaluate channel type list: ',channel(X,Type),X)
           )
       ;  Type = dotUnitType ->  assert(channel_type_list(X,[]))
       ;  Type = typeTuple(List)
               -> (evaluate_type_list(List,EList)
                   -> assert(channel_type_list(X,[typeTuple(EList)]))
                   ; add_symbol_error(haskell_csp_analyzer,'Could not evaluate non-associative tuple:',channel(X,Type),X)
                  )
      ;  otherwise -> add_symbol_error(haskell_csp_analyzer,'Unexpected channel type: ',channel(X,Type),X)
     ),
     fail.
generate_channel_type_lists :- nl.

flatten_dotTupleTypes([],[]).
flatten_dotTupleTypes([dotTupleType(L)|T],R) :- !,
   flatten_dotTupleTypes(L,RL),
   append(RL,RR,R),
   flatten_dotTupleTypes(T,RR).
flatten_dotTupleTypes([H|T],[H|R]) :- flatten_dotTupleTypes(T,R).


precompile_nametypes :- print('% Analyzing nametypes:'),
     retractall(name_type(_,_)),
     nameType(X,CType), print(' '),print(X),
     (CType = type(Type) -> true ; add_error_fail(haskell_csp_analyzer,'Nametype type in wrong format: ',nameType(X,Type))),
     (Type = dotTupleType(List)
       -> (evaluate_type_list(List,EList)
           -> (EList = [TheType] -> true /* we have a single type; not really a dotTuple */
                ; TheType = dotTupleType(EList)), %% was typeTuple ??? <----------
               assert(name_type(X,TheType))
            ;  add_symbol_error(haskell_csp_analyzer,'Could not evaluate nametype list:',nameType(X,Type),X)
	    )
       ;  Type = dotUnitType ->  assert(name_type(X,dotUnitType))
       ;  Type = typeTuple(List)
               -> (evaluate_type_list(List,EList)
                   -> assert(name_type(X,typeTuple(EList)))
                   ; add_symbol_error(haskell_csp_analyzer,'Could not evaluate non-associative tuple:',nameType(X,Type),X)
                  )
      ;  otherwise -> add_symbol_error(haskell_csp_analyzer,'Unexpected nametype: ',nameType(X,Type),X)
     ),
     fail.
precompile_nametypes :- nl.


check_for_multiples([]).
check_for_multiples([H|T]) :- 
  (member(H,T) -> add_symbol_error(haskell_csp,'Multiple Definitions of: ',H,H) ; true),
  check_for_multiples(T).
  
check_for_contig([]).
check_for_contig([H,H2|T]) :- functor(H,F,N), functor(H2,F,N), !, check_for_contig(T).
check_for_contig([H|T]) :- 
  (member(H,T) -> add_symbol_error(haskell_csp,'Definition not contiguous: ',H,H) ; true),
  check_for_contig(T).

print_agent(T) :- nonvar(T),functor(T,F,N), print(F/N).



/* ------------------------ LET COMPILATION ----------------------------- */

:- dynamic agent_compiled/3.
:- dynamic agent_parameters/3.
%agent_compiled(H,B) :- agent_compiled(H,B_SRCSPAN).


get_agent(Head,Body,SRCSPAN) :-  bindval(F,FBody,SRCSPAN),
 %  (FBody = lambda(Args,Body)
 %    -> Head =.. [F|Args]   /* translate m = \ x,y @ BODY into  m(x,y) = BODY */
 %    ;  
       Head=F, Body=FBody
 % )
   .
get_agent(Head,Body,SRCSPAN) :- agent(Head,Body,SRCSPAN).
get_agent(Head,Body,SRCSPAN) :- agent_curry(CHead,CBody,SRCSPAN),
                                translate_agent_curry(CHead,CBody,Head,Body,SRCSPAN).

    
translate_agent_curry(CHead,CBody,Head,Body,SPAN) :- 
  CHead =.. [Function,Args1|Rest],
  Head =.. [Function|Args1], translate_into_lambda(Rest,CBody,Body,SPAN).
  %,nl,print(agent_curry(Head,Body)),nl,nl,nl.
    
translate_into_lambda([],Body,Body,_SPAN).
translate_into_lambda([Args1|Rest],Body,Res,SPAN) :-
    l_compile_head_para(Args1,NewParas,Constraint,lambda(Args1),SPAN),
    (Constraint = true
      -> LBody=RB
       ; S=no_loc_info_available,
         LBody = ifte(prolog_constraint(Constraint),RB,error(illegal_curried_arg),S,S,S)
    ),
    Res  = lambda(NewParas,LBody),
    translate_into_lambda(Rest,Body,RB,SPAN).


:- dynamic agent_functor/2.
init_agent_functor :- retractall(agent_functor(_,_)),
   get_agent(X,_B,_SRCSPAN), nonvar(X), 
   functor(X,F,Arity),
   assert(agent_functor(F,Arity)),
   fail.
init_agent_functor.

%agent_functor(F,Arity) :- agent_parameters(F,Arity,_).

:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

% below we compile let expression into separate agent definitions
% so that we get the treatment of local variables from Prolog
% the same is applied to prefixes with input variables (which are also local variables)
compile_nested_let_expressions :- debug_println(9,'% Compiling nested let expressions:'),
   retractall(agent_compiled(_,_,_)), 
   retractall(agent_parameters(_,_,_)),
   init_agent_functor, reset_cur_numbervars_index,
   get_agent(X,Body,SRCSPAN), 
   (var(X) ->
     add_internal_error(haskell_csp,'Error in compiled .pl File: Uninstantiated LHS in bindval or agent',(X,Body),SRCSPAN),fail
     ; true),
   compile_head(X,NX,Constraints,SRCSPAN),
   debug_println(9,compiled_head(X,NX,Constraints)),
   
   NX=..[F|_N_XARGS], 
   (debug_mode(on) -> print(' '),print_agent(X),nl ; true),
   term_variables(X,InnerVariables), XARGS=InnerVariables, % these are the variables before extracting pattern matches into Constraints
   %debug_println(9,compiling_body(Body,F,_N_XARGS,XARGS)),
   (compile_body(Body,F,XARGS,[],Res)
     -> true
     ;  add_internal_error(haskell_csp,'Unable to compile body of agent: ',agent(X,Body),SRCSPAN), fail
   ),
   assert_agent_compiled_head(NX,Res,SRCSPAN,Constraints),
   % (Constraints = true -> assert_agent_compiled(NX,Res,SRCSPAN) ; assert_agent_compiled(NX,Res,SRCSPAN,Constraints)),
   %portray_clause( (agent_compiled(NX,Res) :- Constraints)),
   fail.
compile_nested_let_expressions :- (debug_mode(on) -> nl ; true).   


:- dynamic multiple_let_equations_for/2.

multiple_definitions(Head) :- functor(Head,F,_),
   multiple_let_equations_for(F,_),!.
multiple_definitions(Head) :- functor(Head,F,N), functor(CH,F,N), functor(CH2,F,N),
   get_agent(CH,Body,SRCSPAN), get_agent(CH2,Body2,SRCSPAN2),
  (CH,Body,SRCSPAN) \= (CH2,Body2,SRCSPAN2).
 
compile_head(Head,NewHead,Constraints,SPAN) :-
      Head =.. [Func|Args],
      l_compile_head_para(Args,NewArgs,Constraints,Head,SPAN),
      NewHead =.. [Func|NewArgs].

l_compile_head_para([],[],true,_,_).
l_compile_head_para([H|T],[NH|NT],Constraints,Head,SPAN) :-
      compile_head_para(H,NH,CH,Head,SPAN),
      l_compile_head_para(T,NT,CT,Head,SPAN),
      (CH=true -> Constraints=CT ; (CT=true -> Constraints=CH ; Constraints = ','(CH,CT))).


compile_head_para(Var,NewPara,Constraint,_Head,_Span) :- var(Var),!, NewPara=Var, Constraint=true.
compile_head_para(emptySet,NewPara,Constraint,_Head,_Span) :- !, Constraint=true, NewPara = setValue([]).
compile_head_para(listPat(List),NewPara,RestConstr,Head,Span) :- !, NewPara=list(NList),
      l_compile_head_para(List,NList,RestConstr,Head,Span).
compile_head_para(singleSetPat(List),NewPara,RestConstr,Head,Span) :- !, NewPara=setValue(NList),
      l_compile_head_para(List,NList,RestConstr,Head,Span).
compile_head_para(dotpat(List),NewPara,Constraint,Head,Span) :- !,
  % TO DO: check if we have a record constructor and can construct a record pattern: we would need no constraint
      Constraint = (pattern_match_function_argument(tuple(WNList),NewPara,Head),RestConstr),
      l_compile_head_para(List,NList,RestConstr,Head,Span),
      wrap_free_vars(NList,WNList). % wrap free vars with in(.)
compile_head_para(tuplePat(List),NewPara,Constraint,Head,Span) :- !, NewPara = na_tuple(NList),
      l_compile_head_para(List,NList,Constraint,Head,Span).
compile_head_para(alsoPattern([Pat1]),NewPara,Constraint1,Head,Span) :- !,
      compile_head_para(Pat1,NewPara,Constraint1,Head,Span).
compile_head_para(alsoPattern([Pat1|Tail]),NewPara,Constr,Head,Span) :- !,
      compile_head_para(Pat1,NewPara,Constraint1,Head,Span),
      compile_head_para(alsoPattern(Tail),NewPara2,Constraint2,Head,Span),
      Constr = (NewPara=NewPara2,Constraint1,Constraint2).
compile_head_para(appendPattern([Pat1]),NewPara,Constraint1,Head,Span) :- !,
      compile_head_para(Pat1,NewPara,Constraint1,Head,Span).
compile_head_para(appendPattern([H|T]),NewPara,Constraints,Head,Span) :-
      nonvar(H), H=listPat(HList),is_list_skeleton(HList),!,
      compile_head_para(H,NH,ConstraintH,Head,_), NH=list(NHList),
      NewPara = list(NewList),
      compile_head_para(appendPattern(T),NewT,ConstraintsT,Head,Span),
      Constraints = ','(ConstraintH,ConstraintsT),
      NewT = list(TList), append(NHList,TList,NewList).
compile_head_para(appendPattern(Pat),NewPara,Constraint,Head,Span) :-
      is_list_skeleton(Pat),
      append(T,[H],Pat),
      nonvar(H), H=listPat(HList),is_list_skeleton(HList),!,
      compile_head_para(H,NH,ConstraintH,Head,_), NH=list(NHList),
      compile_head_para(appendPattern(T),NewT,ConstraintsT,Head,Span),
      NewPara = list(NewList), NewT = list(TList),
      Constraint=','(ConstraintH,','(append(TList,NHList,NewList),ConstraintsT)).
% to do:  appendPattern with more parameters
compile_head_para(appendPattern(Pat),_,_,_Head,Span) :- 
      %haskell_csp:get_symbol_span(Head,HeadSpan),
      add_error(compile_head_para,'Unsupported Append Pattern: ',Pat,Span),fail.
compile_head_para(X,_,_,Head,Span) :-
      atomic(X),
      channel(X,_), %print(Head),nl,
      (multiple_definitions(Head)
       -> add_message(compile_head_para,'Warning: Channel name used as Function/Process argument: ',X)
       ;  % haskell_csp:get_symbol_span(Head,HeadSpan),
          add_error(compile_head_para,'Warning: Channel name used as Function/Process argument: ',X, Span)
      ),
      fail.
compile_head_para(X,NewPara,Constraint,_Head,_Span) :-
      atomic(X), evaluate_argument(X,NewPara), !,
  %%    print(compiled_head_para(X,NewPara)),nl,
      Constraint=true.
compile_head_para(P,P,true,_Head,_Span).


wrap_free_vars([],[]).
wrap_free_vars([H|T],[WH|WT]) :- 
   (var(H) -> WH=in(H) ; WH=H),
   wrap_free_vars(T,WT).

is_list_skeleton(X) :- var(X),!,fail.
is_list_skeleton([]).
is_list_skeleton([_|T]) :- is_list_skeleton(T).

/* compile_body(ExpressionToCompile, Top-Level-FunName, ArgumentsToAddToNestedLetsOrPrefixes,LocalLets, CompiledExpr) */
compile_body(Body,F,XARGS,NDEFS,Res) :- compile_body(Body,F,XARGS,NDEFS,false,Res).
   /* false = we are not in a dangerous context where prefixes with input variables need to be lifted out */
   
%compile_body(X,F,XA,D,DC,R) :- print(compile_body(X,F,XA,D,DC,R)),nl,fail.
compile_body(X,_F,_XARGS,_DEFS,_DC,Res) :- var(X),!, Res=X.
   %add_error(haskell_csp_analyzer,'Variable in compile_body:'(X,F)),Res=X.
compile_body(lambda(Args,Body),F,XARGS,DEFS,DC,Res) :- !, %nl,print(lambda1(Args,Body)),nl,
   l_compile_head_para(Args,CArgs,Constraints,lambda(Args,Body),unknown_span),
   (Constraints=true -> true
     ; add_error(compile_body,'ProB does not (yet) support complicated patterns inside lambda: ',lambda(Args,Body))),
   mynumbervars(CArgs), % note: important that compilation of body does detect $VAR
   % otherwise will fail to add necessary parameters
   Res = lambda(CArgs,ResBody),
   append(Args,XARGS,NewXARGS),
   compile_body(Body,F,NewXARGS,DEFS,DC,ResBody).
compile_body(val_of(X,Span),_F,_XARGS,DEFS,_DC,Res) :- !,   /* we do not need to compile X itself !? */
  % haskell_csp:get_symbol_span(X,Span),  % does this work for local lets ???
   (is_local_let2(X,0,[],DEFS,NX,NPar,_)
     -> Res = agent_call(Span,NX,NPar) 
     ; Res=agent_call(Span,X,[])).
compile_body(agent_call(_Span,TFun,[Arg]),F,XARGS,DEFS,DC,Res) :- 
    cspTransparent([TFun]), !,
    compile_body(Arg,F,XARGS,DEFS,DC,Res).
compile_body(agent_call_curry(X,Paras),F,XARGS,DEFS,DC,Res) :- !,
  /* translate agent_call_curry into nested agent_calls : */
    (Paras=[Paras1|RestParas]
      ->    compile_body(agent_call_curry(agent_call(no_loc_info_available,X,Paras1),RestParas),F,XARGS,DEFS,DC,Res)
      ;     compile_body(X,F,XARGS,DEFS,DC,Res)
    ).
compile_body(agent_call(_Span,X,Para),F,XARGS,DEFS,DC,Res) :- atomic(X),
    Term=..[X|Para], length(Para,N), length(NPara,N), Fun=..[X|NPara],
    (\+ agent(Fun,_Body,_Src), is_builtin_agent(Term), \+is_local_let(Fun,DEFS,_NewFun,_FXArgs,_)),!,
    compile_body(builtin_call(Term),F,XARGS,DEFS,DC,Res).
compile_body(agent_call(Span,X,Para),F,XARGS,DEFS,_DC,Res) :- atomic(X),
    CX=X, %%%%compile_body(X,F,XARGS,DEFS,true,CX),
    l_compile_body(Para,F,XARGS,DEFS,true,CPara),
    (is_local_let2(CX,_Arity,CPara,DEFS,NX,NPara,_) -> Res = agent_call(Span,NX,NPara) ; Res=agent_call(Span,CX,CPara)).
compile_body(prefix(Span,Values,ChannelExpr,CSP,Span2),F,XARGS,DEFS,DC,Res) :-  !,
    compile_body(ChannelExpr,F,XARGS,DEFS,DC,NewChExpr),
    l_compile_channel_value(Values,F,XARGS,DEFS,NewVals,NewXARGS,Span),
    %print(comp_channel(Values,NewVals)),nl,
    compile_body(CSP,F,NewXARGS,DEFS,DC,NewCSP),
    ResBody = prefix(Span,NewVals,NewChExpr,NewCSP,Span2),
    (NewXARGS \= XARGS %,DC = true   always lift out so that states are ground and that we need no variant/instance checks !
       ->  /* need to lift prefix */
            gensym('->',GSF),  debug_println(9,generating_agent_for_prefix(F,GSF, ResBody)),
            string_concatenate(F,GSF,PrefixName),
            Call =.. [PrefixName|XARGS],
            Res = agent_call(Span,PrefixName,XARGS),
          %  portray_clause( prefix_agent(Call,ResBody) ),
            assert_agent(Call, ResBody,Span),
            functor(Call,CF,CN), functor(CallSkel,CF,CN),
            assert( is_csp_process(CallSkel) )
        ; %print(not_lifting_prefix(Values,F,XARGS,DC, NewXARGS),nl,
           Res = ResBody
    ).
compile_body(X,F,XARGS,DEFS,DC,Res) :-  
    replicated_or_comprehension(X,Op,Vars,OutsideOfScope,InScope,NewDC,ISCSP,Span),!, 
   % print(repl(X,Op,Vars)),nl,print(inscope(InScope)),nl, print(outofscope(OutsideOfScope)),nl,
    (var(Vars)
      -> add_error(haskell_csp_analyzer,'Replicated variables var: ',X),fail ; true),
    %%print(compiling_set(OutsideOfScope,XARGS)),nl,
    % This part does not need the extra variables Vars as parameters
    compile_body(OutsideOfScope,F,XARGS,DEFS,DC,NewOutsideOfScope),
    % print(done_outside(NewOutsideOfScope)),nl,
    append(Vars,XARGS,NewXARGS), % add the variables of the comprehension as additional parameters to nested calls
    %%print(compiling_csp(InScope,NewXARGS)),nl,
    compile_body(InScope,F,NewXARGS,DEFS,NewDC,NewInScope),
   % print(done_csp(InScope,NewInScope)),nl,
    replicated_or_comprehension(ResBody,Op,Vars,NewOutsideOfScope,NewInScope,_,ISCSP,Span),
    (true %DC = true    always lift out so that states are ground and that we need no variant/instance checks !
       ->   gensym('@',GSF),  debug_println(9,generating_agent_for_repOp(F,GSF, ResBody)),
            string_concatenate(F,GSF,PrefixName),
            Call =.. [PrefixName|XARGS],
            Res = agent_call(Span,PrefixName,XARGS), % use Span ?? or no_loc_info_available
           % portray_clause( replicated(Call,ResBody) ),
            assert_agent(Call, ResBody, unknown_span),
            functor(Call,CF,CN), functor(CallSkel,CF,CN),
            (ISCSP=true -> assert( is_csp_process(CallSkel) ) ; true)
        ; Res = ResBody
    ).
compile_body(let(LIST,Body),F,XARGS,DEFS,DC,Res) :- !, %print(let(LIST)),nl,
     expand_let_definitions(LIST,ExpLIST), %print(extracting(ExpLIST)),nl,
     extract_new_funs_from_lets(ExpLIST,F,XARGS,[],LetDefs),
     append(LetDefs,DEFS,NDEFS),
     debug_println(9,new_funs(NDEFS)),
     compile_lets(ExpLIST,F,XARGS,NDEFS,DC),
     debug_println(9,nested_let_compile(Body,NDEFS)),
     compile_body(Body,F,XARGS,NDEFS,DC,Res). % check builtin calls in let declarations list
compile_body(builtin_call(X),F,XARGS,DEFS,DC,Res) :- !, 
   translate_builtin_call(X,TX),
  %(translate_builtin_call(X,TX) -> true ; TX=X),
  compile_body(TX,F,XARGS,DEFS,DC,Res).
compile_body([H|T],F,XARGS,DEFS,DC,Res) :- !, l_compile_body([H|T],F,XARGS,DEFS,DC,Res).
compile_body(FunName,_F,_XARGS,DEFS,_DC,Res) :- atomic(FunName),
   %(FunName=getPrio -> print(compile(FunName,DEFS)),nl ; true),
   is_local_let2(FunName,OriginalArity,[],DEFS,NewF,_,LLXARGS) ,!,   %% append(LLXARGS,_,_XARGS) ??
   %print(replacing(FunName,NewF,LLXARGS)),nl,
   % We have a local function whose function name is used: replace by a lambda closure which hides
   %   the additional parameters required to call the lifted version NewF
   length(OrigArgs,OriginalArity),
   mynumbervars(OrigArgs),
   (LLXARGS=[]
    -> Res = NewF
    ;  (append(OrigArgs,LLXARGS,FullArgs),
        haskell_csp:get_symbol_span(FunName,Span),
        Res = lambda(OrigArgs,agent_call(Span,NewF,FullArgs))
      %  ,print(generated(Res)),nl
       )
   ).  %%  ,print(res(Res)),nl.
compile_body(NV,F,XARGS,DEFS,DC,Res) :- nonvar(NV), NV=..[FN|Args], !,  /* TO DO: IMPROVE ! */
   l_compile_body(Args,F,XARGS,DEFS,DC,ResArgs),
   Res =.. [FN|ResArgs]    %%,print(compiled(NV,Res)),nl
   .
compile_body(X,_F,_X,_D,_DC,X).

l_compile_body([],_F,_XARGS,_DEFS,_DC,[]).
l_compile_body([H|T],F,XARGS,DEFS,DC,[RH|RT]) :-
   %print(compiling(H)),nl,
   compile_body(H,F,XARGS,DEFS,DC,RH), %print(finished(H)), nl,
   add_new_parameters(H,XARGS,NewXARGS), %print(newargs(NewXARGS,H)),nl,
   l_compile_body(T,F,NewXARGS,DEFS,DC,RT).

% check whether we have to add parameters of the body to calls to the right
add_new_parameters(X,Vars,NewVars) :- var(X),!,Vars=NewVars.
add_new_parameters(comprehensionGenerator(VAR,_SET),Vars,NewVars) :- !,
 (var(VAR) -> NewVars=[VAR|Vars]
  ; %nl,print(comprehensionGenerator(VAR,_SET)),nl,
    NewVars=Vars)
 . %, check_var(VAR,_SET).
add_new_parameters(_,V,V).

:- public check_var/2.
:- block check_var(-,?).
check_var(V,SET) :- nl, print(not_var(V)),nl,print(SET),nl.

% give some of CSP's operators are more distinctive name
translate_builtin_call(set(X),R) :- !, R=seq_to_set(X).
translate_builtin_call(seq(X),R) :- !, R=set_to_seq(X).
translate_builtin_call(X,X).

assert_agent(Head,Body,Span) :-
    compile_head(Head,NewHead,Constraints,Span),
    assert_agent_compiled_head(NewHead,Body,Span,Constraints).
   
assert_agent_compiled_head(Head,OrigBody,Span,Constr) :-
   (preferences:get_preference(cspm_strip_source_location,true)
      -> clear_span(OrigBody,Body)  %,print(clear_span(Res,CRes)),nl
      ;  Body=OrigBody),
     lift(Head,LHead,(Body,Constr),(LBody,LConstr)),
     assert((agent_compiled(LHead,LBody,Span) :- LConstr)),
     LHead =.. [Functor|Args],
     length(Args,Arity),
     compute_lazy_strict_types(Args,TL),
     (retract(agent_parameters(Functor,Arity,OldTL))
       -> merge_lazy_strict_types(TL,OldTL,NewTL)  ;  NewTL = TL ),
  %%     print(agent_parameters(Functor,Arity,NewTL)),nl,
     assert(agent_parameters(Functor,Arity,NewTL)),
     (agent_functor(Functor,Arity) -> true ; assert(agent_functor(Functor,Arity))).

compute_lazy_strict_types([],[]).
compute_lazy_strict_types([H|T],[HT|TT]) :-
    (var(H) -> HT=lazy ; HT=strict),
    compute_lazy_strict_types(T,TT).
merge_lazy_strict_types([],[],[]).
merge_lazy_strict_types([H1|T1],[H2|T2],[HH|TT]) :-
   (H1=strict -> HH=strict ; HH=H2),
   merge_lazy_strict_types(T1,T2,TT).

:- assert_must_succeed( (X=p('$VAR'(1),R,f('$VAR'(1),'$VAR'(2))),haskell_csp_analyzer:lift(X,R,true,true),
                         R=X) ).

/* lift all the variables which have been numbervar'd in the head to real variables */
lift(Head,LiftedHead,Body,LiftedBody) :-
  lift2(Head,LiftedHead,[],VL,true),
  lift2(Body,LiftedBody,VL,_,false).  /* do not lift $VARs inside Body which are not in head */

lift2(X,R,VL,AVL,_INS) :- (var(X); atomic(X)),!,R=X, VL=AVL.
lift2('$VAR'(X),R,VarList,RVL,INS) :- !,lookup_insert(VarList,X,R,RVL,INS).
lift2(lambda(Args,Body),ResTerm,VarList,RVL,false) :- !,
   ResTerm = lambda(Args,LiftedBody), RVL = VarList,
   lift2(Args,_LiftedArgs,[],LambdaVARs,true),
   list_difference(VarList,LambdaVARs,InnerVarList), % do not lift local $VARS (which hide outer $VAR)
   lift2(Body,LiftedBody,InnerVarList,_,false).
  % ,print(lifted_lambda(Args,Body,ResTerm)),nl.
lift2(Term,ResTerm,VarList,RVL,INS) :-
   Term =.. [Functor|Args],
   l_lift2(Args,RArgs,VarList,RVL,INS),
   ResTerm =.. [Functor|RArgs].

% TO DO: check if we need to treat lambda's here: what if a lambda by accident has as
% its local variable a $VAR which is currently being lifted !!


l_lift2([],[],V,V,_INS).
l_lift2([H|T],[LH|LT],Vin,Vout,INS) :-
   lift2(H,LH,Vin,V2,INS),
   l_lift2(T,LT,V2,Vout,INS).

lookup_insert([],X,R,[X/R],INSERT) :-
        (INSERT=true -> true %Vout = [X/R]
        ; %Vout=[], 
          R='$VAR'(X)).
lookup_insert([K/V|T],X,R,[K/V|RT],INS) :-
    (K==X
       -> V=R,T=RT
       ;  lookup_insert(T,X,R,RT,INS)
    ).

%translate_builtin_call(set(X),set_to_seq(X)).

replicatedOperator(repChoice(Generators,Body,Span),repChoice,Vars,Generators,Body,false,Span) :-
  /* false -> Body is not in a dangerous context (even if repChoice itself is in 
     a dangerous context) */
    extract_variables_from_generator_list(Generators,Vars).  /* should the generators be put together with Body ?? */
replicatedOperator(repInternalChoice(Generators,Body,Span),repInternalChoice,Vars,Generators,Body,false,Span) :-
    extract_variables_from_generator_list(Generators,Vars).
replicatedOperator(repInterleave(Generators,Body,Span),repInterleave,Vars,Generators,Body,true,Span) :-
    extract_variables_from_generator_list(Generators,Vars).
replicatedOperator(repSequence(Generators,Body,Span),repSequence,Vars,Generators,Body,true,Span)  :-
    extract_variables_from_generator_list(Generators,Vars).
replicatedOperator(procRepAParallel(Generators,pair(Sync,Body),Span),procRepAParallel,Vars,Generators,[Sync,Body],true,Span) :-
    %print(repA(Generators)),nl,
    extract_variables_from_generator_list(Generators,Vars).
replicatedOperator(procRepLinkParallel(Sync,Generators,Body,Span),procRepLinkParallel,Vars,[Generators,Sync],Body,true,Span) :-
    extract_variables_from_generator_list(Generators,Vars).
replicatedOperator(procRepSharing(Sync,Generators,Body,Span),procRepSharing,Vars,[Generators,Sync],Body,true,Span) :-
    extract_variables_from_generator_list(Generators,Vars).
 /* last arg == true ->means-> dangerous as several branches active simultaneously -> local variables need
    to be extracted*/
    
% replicated_or_comprehension(EXPR, OperatorAsAtom, Variables,  
%         Term(s)OutsideOfReplicatedVariablesScope (possibly defining new ones),
%                                 TermInScopeofVariables, Replicated_true/false, Span)
replicated_or_comprehension(X,Op,Vars,Set,CSP,NewDC,true,Span) :-
   replicatedOperator(X,Op,Vars,Set,CSP,NewDC,Span).
replicated_or_comprehension(X,Op,Vars,OutsideOfScope,InScope,NewDC,false,no_loc_info_available) :-
   comprehension_aux(X,Op,Vars,OutsideOfScope,InScope), NewDC=true.

comprehension_aux(listExp(rangeEnum(Tuples),Generators),listComp, Vars , Generators, InScope) :-
    InScope = Tuples,
    extract_variables_from_generator_list(Generators,Vars).
comprehension_aux(setExp(rangeEnum(Tuples),Generators), setComp, Vars , Generators, InScope) :-
    InScope = Tuples,
    extract_variables_from_generator_list(Generators,Vars).
% TO DO: other setExp patterns !
comprehension_aux(procRenamingComp(A,Generators,RenameList),procRenamingComp,Vars,OutsideOfScope,InScope) :-
    OutsideOfScope = [A|Generators],
    InScope = RenameList,
    extract_variables_from_generator_list(Generators,Vars).

l_compile_channel_value([],_F,XA,_DEFS,Res,XA,_Span) :- !, Res = [].
l_compile_channel_value([H|T],F,XA,DEFS,Res,NewXA,Span) :- !,
   compile_channel_value(H,F,XA,DEFS,ResH,XA1,Span), append(ResH,ResT,Res),
   l_compile_channel_value(T,F,XA1,DEFS,ResT,NewXA,Span).
l_compile_channel_value(E,F,XA,_,R,NXA,Span) :-
   add_internal_error(haskell_csp_analyzer,'Unkown channel value list: ',(E,F),Span), R=E,NXA=XA.


compile_channel_value(VAR,F,XA,_DEFS,Res,NXA,Span) :- var(VAR),!,
  add_internal_error(compile_channel_value,'Variable as channel value: ',F,Span),NXA=XA, Res=[VAR].
compile_channel_value(out(Expr),F,XA,DEFS,Res,NXA,_Span) :- !,
  Res=[out(CE)],NXA=XA,compile_body(Expr,F,XA,DEFS,CE).
compile_channel_value(dot(Expr),F,XA,DEFS,Res,NXA,_Span) :- !,
  Res=[out(CE)],NXA=XA,compile_body(Expr,F,XA,DEFS,CE).
compile_channel_value(in(Expr),F,XA,DEFS,Res,NXA,Span) :- ground(Expr), !,
  add_message(haskell_csp_analyzer,'Warning: Channel input contains no variable: ',Expr:Span),
  Res=[out(CE)],NXA=XA,compile_body(Expr,F,XA,DEFS,CE).
compile_channel_value(in(Expr),F,XA,DEFS,Res,NXA,Span) :- 
   compile_top_level_in_channel_value(Expr,F,XA,DEFS,Res,NXA,Span,warn),!.
compile_channel_value(inGuard(Var,Guard),F,XA,DEFS,Res,NXA,Span) :- Res = [inGuard(VarNew,NewGuard)],
    compile_in_channel_value(Var,XA,VarRes,NXA,Span,warn),
    %print(compile_inGuard(Var,Guard,Res)),nl,
    VarRes=[in(VarNew)],!,
    compile_body(Guard,F,XA,DEFS,NewGuard). % XA instead of NXA: the input value cannot be used in the set expression
compile_channel_value(E,F,XA,_,Res,NXA,Span) :-
  add_internal_error(haskell_csp_analyzer,'Illegal channel value or guard: ',E:F,Span), Res=[E],NXA=XA.

compile_top_level_in_channel_value(Expr,F,XA,DEFS,Res,NXA,Span,_) :- % print(compile(Expr,XA,Res,NXA)),nl,
  nonvar(Expr),Expr = dotpat(List),
  List=[H|T], \+ is_csp_constructor(H),!, % we have a ?(x.y) --> treate like ?x?y
  compile_top_level_in_channel_value(H,F,XA,DEFS,R1,NXA1,Span,nowarn),append(R1,R2,Res),
  compile_top_level_in_channel_value(dotpat(T),F,NXA1,DEFS,R2,NXA,Span,nowarn).
% here we need a case for treating channel fields like a?P.x as a.P?x or a?x.P.y as a?x.P?y where P is a constructor
compile_top_level_in_channel_value(Expr,F,XA,DEFS,Res,NXA,Span,_) :- 
  nonvar(Expr),Expr = dotpat(List),
  List = [H|T], nonvar(H),is_constructor(H,C,_Args),!, (length(T,1) -> T=[H1|_], LL = [out(C),in(H1)]; LL = [out(C),in(dotpat(T))]),
  l_compile_channel_value(LL,F,XA,DEFS,Res,NXA,Span).
compile_top_level_in_channel_value(Expr,_F,XA,_DEFS,Res,NXA,_Span,_) :- % print(compile(Expr,XA,Res,NXA)),nl,
  nonvar(Expr),Expr = dotpat(List),List=[],!, % we have a ?(x.y) --> treate like ?x.?y
  Res = [],NXA=XA.
compile_top_level_in_channel_value(Expr,_F,XtraArgs,_DEFS,Res,NewXtraArgs,Span,WarnIfNoVariables) :-
  compile_in_channel_value(Expr,XtraArgs,Res,NewXtraArgs,Span,WarnIfNoVariables).

compile_in_channel_value(Expr,XA,[ResTerm],NXA,Span,Warn) :- !,
   add_extra_input_variable(Expr,XA,NXA,Res,Span),
   (XA==NXA -> 
     %add_error(haskell_csp_analyzer,'Channel input contains no variable: ',Expr,Span)
     (Warn=nowarn -> true
       ;add_message(haskell_csp_analyzer,'Warning: Channel input contains no variable: ',Expr:Span)),
       ResTerm = out(Res)
     ; ResTerm=in(Res)).

% add_extra_input_variable(InputVariableTerm,  ExtraLocalVariablesSoFar,
%     NewExtraLocalVariablesAfter, ResTransformedInputVariableTerm, Span)
    
% Note: expression wrapped in in(.) can contain variables; they can thus not
% yet be fully evaluated (mainly due to the CSP dot constructor)

:- dynamic input_agent_compiled/2.

:- block blocking_append(?,?,-).
blocking_append(L1,L2,L3) :- append(L1,L2,L3).

add_extra_input_variable(Var,XA,NXA,Res,Span) :- %print(add(Var,XA,NXA)),nl,
   var(Var),!,
   (exact_member(Var,XA) %% case will be catched from the CSP-M parser (e.g. c?x?x => Error message: Redefinition of x)
          -> (add_internal_error(haskell_csp_analyzer,'Internal Error: Channel input variable already used: ',Var,Span),NXA=XA)
          ;  NXA=[Var|XA]
   ),Res=Var.
add_extra_input_variable(listPat(List),XA,NXA,Res,Span) :- !, Res = list(RList),
    add_extra_input_variables(List,XA,NXA,RList,Span).
add_extra_input_variable(singleSetPat(List),XA,NXA,Res,Span) :- !, Res = setValue(RList),
    add_extra_input_variables(List,XA,NXA,RList,Span).
add_extra_input_variable(tuplePat(List),XA,NXA,Res,Span) :- !, Res = na_tuple(RList),
    add_extra_input_variables(List,XA,NXA,RList,Span).
add_extra_input_variable(dotpat(List),XA,NXA,Res,Span) :- !, Res = dotTuple(RList),
   /* dotTuple or dotpat because it still needs to be evaluated */
    /* TODO: CHECK if we don't have to detect records here ?? */
    add_extra_input_tuple_variables(List,XA,NXA,RList,Span).
add_extra_input_variable(alsoPattern([X,Y]),XA,NXA,Res,Span) :- !, Res = alsoPat(RX,RY),
    add_extra_input_variable(X,XA,NXX,RX,Span),
    add_extra_input_variable(Y,XA,NXY,RY,Span), append(NXX,NXY,NXA). % was NXA=NXX.
add_extra_input_variable(appendPattern(List),_XA,_NXA,Res,Span) :- !, 
    compile_append_pat(appendPattern(List),LRes,Constr),
    get_append_agent_result('AppPatFun',LRes,Constr,Res,Span).
add_extra_input_variable(Expr,XA,NXA,Res,_Span) :- 
    (Expr = int(_); valid_constant(Expr)),!, XA=NXA, Res=Expr.
add_extra_input_variable(Expr,XA,NXA,Res,Span) :-
    add_error(haskell_csp_analyzer,'Channel input not recognized: ',Expr,Span),
    NXA=XA, Res=Expr.
% TODO: check if other patterns can appear here,...

get_append_agent_result(FunName,VarExpr,Constr,Res,Span) :- 
    gensym(FunName,NewFun),
    %functor(Head,NewFun,1),arg(1,Head,VarExpr),
    Head =.. [NewFun|[VarExpr]],
    Res = appendPat(VarExpr,Head),
    assert_agent_compiled_head(Head,VarExpr,Span,Constr). %,print(assert_agent_compiled_head(Head,VarExpr,Span,Constr)),nl.

compile_append_pat(appendPattern([Pat1]),NewPara,Constraint1) :- !,
      compile_head_para(Pat1,NewPara,Constraint1,_Head,_Span).
compile_append_pat(appendPattern([H|T]),NewPara,ConstraintsT) :-
      nonvar(H), H=listPat(HList),is_list_skeleton(HList),!,
      NewPara = list(NewList), 
      compile_append_pat(appendPattern(T),NewT,ConstraintsT),
      NewT = list(TList), append(HList,TList,NewList).
compile_append_pat(appendPattern(Pat),NewPara,Constraint) :-
      is_list_skeleton(Pat),
      append(T,[H],Pat),
      nonvar(H), H=listPat(HList),is_list_skeleton(HList),!,
      compile_append_pat(appendPattern(T),NewT,_ConstraintsT),
      NewPara = list(NewList), NewT = list(TList),
      Constraint=blocking_append(TList,HList,NewList).
compile_append_pat(Pat,_NewPara,_Constraint) :- 
      add_internal_error('Internal Error: Expected append pattern instead of ', Pat),fail.

add_extra_input_tuple_variable(Expr,XA,NXA,Res,_Span) :- % allow record constructors
    %print(addi(Expr)),nl,
    nonvar(Expr), is_csp_constructor(Expr),!, XA=NXA, Res=Expr. % TO DO: should we check the args ?!
add_extra_input_tuple_variable(Expr,XA,NXA,Res,Span) :- add_extra_input_variable(Expr,XA,NXA,Res,Span).

add_extra_input_tuple_variables([],XA,XA,[],_).
add_extra_input_tuple_variables([V|T],XA,NXA,[CV|CT],Span) :-
	add_extra_input_tuple_variable(V,XA,XA1,CV,Span),
	add_extra_input_tuple_variables(T,XA1,NXA,CT,Span).

add_extra_input_variables([],XA,XA,[],_).
add_extra_input_variables([V|T],XA,NXA,[CV|CT],Span) :-
	add_extra_input_variable(V,XA,XA1,CV,Span),
	add_extra_input_variables(T,XA1,NXA,CT,Span).

compile_lets([],_F,_XA,_D,_DC).
compile_lets([L1|T],F,XA,D,DC) :-
   compile_let(L1,F,XA,D,DC), compile_lets(T,F,XA,D,DC).


compile_let(agent_curry(Fun,Body,SRCSPAN),F,XARGS,DEFS,DC) :- !,
     translate_agent_curry(Fun,Body,TFun,TBody,SRCSPAN),
     compile_let(agent(TFun,TBody,SRCSPAN),F,XARGS,DEFS,DC).
compile_let(agent(Fun,Body,SRCSPAN),F,XARGS,DEFS,DC) :- 
   /* TO DO: treat multiple agent facts for same fun ;; DONE ?? */
    ( is_local_let(Fun,DEFS,NewFun,_FXArgs,_)
      ->   term_variables(Fun,ArgumentsToFun),
           append(XARGS,ArgumentsToFun,NewXARGS),
           debug_println(9,compile_within_let(Fun,NewXARGS)),
           compile_body(Body,F,NewXARGS,DEFS,DC,Res),  /* one could pass NewF instead of F */
       %    portray_clause( let_agent(NewFun,Res) ),
           assert_agent(NewFun,Res,SRCSPAN)
	   %print(assert_agent(NewFun,Res,SRCSPAN)),nl
      ; add_internal_error(haskell_csp,'Could not find let function in defs: ',(Fun,DEFS),SRCSPAN)
    ). 

:- meta_predicate findall_keepvars(-,0,-).
expand_let_definitions(Lets,ExpLets) :- %print(expanding(Lets)),nl,
  findall_keepvars(EA, member_expand(Lets,EA), ExpLets). % important to use findall_keepvars instead of findall to avoid free variable (parameters of outer function) renaming apart
  
member_expand(Lets,EA) :- member(A,Lets),expand(A,EA).

/* a version of findall that does not rename apart unbound variables */
findall_keepvars(V,Goal,SolList) :- term_variables(Goal,GVars), term_variables(V,VVars),
    remove_variables(GVars,VVars,Vars),
    findall( sol(V,Vars), Goal, ISol),
   % print(findall( sol(V,Vars), Goal, ISol)),nl,
    l_unify(ISol,Vars,SolList).
l_unify([],_,[]).
l_unify([sol(V,SVars)|T],Vars,[V|TT]) :- 
  (SVars=Vars -> l_unify(T,Vars,TT)
  ; add_internal_error('Variable Binding Clash in l_unify: ',l_unify([sol(V,SVars)|T],Vars,[V|TT])),
    fail).

expand(bindval(Fun,Body,Span),R) :- !,expand(agent(Fun,Body,Span),R).
expand(agent(Fun,Body,Span),R) :- !, R = agent(EFun,EBody,Span),
    if(expand_agent(Fun,Body,outer,Span,EFun,EBody),
        true, %(print(expanded_agent(R)),nl),
        (add_internal_error(expand_agent,'Expand agent failed: ',agent(Fun,Body),Span),fail)).
expand(X,X).

% expand definitions such as (p1,p2,p3) = c into three definitions of agent
%expand_agent(F,B,_,_,_FF,_BB) :- print(ea(F,B,_FF,_BB)),nl,fail.
expand_agent(Var,_Body,_,_,_RV,_RB) :- var(Var),!,fail. /* we have a variable; we assume this combes from a Wildcard _  underscore !!!  example let (_,a) = Body ....  */
/* the wildcard value is thrown away: we do not need to compute this projection */ %RV=Var,RB=Body.
expand_agent('tuplePat'(Pat),Body, _,Span,Fun, ExpandedBody) :- !,
      nth1(Nr,Pat,Head), 
      expand_agent(Head,na_tuple_projection(int(Nr),Body,Span),inner,Span,Fun,ExpandedBody).
      %,print(expanded(Fun,ExpandedBody)),nl.
expand_agent(dotpat(Pat),Body, _,Span,Fun, ExpandedBody) :- !,
      nth1(Nr,Pat,Head), 
      expand_agent(Head,tuple_projection(int(Nr),Body,Span),inner,Span,Fun,ExpandedBody).
expand_agent(singleSetPat([Pat]),Body, _,Span,Fun, ExpandedBody) :- !,
      expand_agent(Pat,singleSetPatValue(Body,Span),inner,Span,Fun,ExpandedBody).
expand_agent(listPat(Pat),Body, _,Span,Fun, ExpandedBody) :- !,
      nth1(Nr,Pat,Head), length(Pat,Len),
      expand_agent(Head,list_projection(int(Nr),int(Len),Body,Span),inner,Span,Fun,ExpandedBody).
% TO DO: add more patterns
% TO DO: generate error if pattern contains constructs not yet translated
expand_agent(Fun,Body,inner,Span,FunR,BodyR) :- !,
  ((atomic(Fun), \+ number(Fun)) -> FunR=Fun, BodyR=Body
     ; add_error(expand_agent,'Cannot treat expression in LHS of let (use fresh variable): ',Fun,Span),fail).
expand_agent(Fun,Body,_,_,Fun,Body).  %% check if Fun is atomic ??


 
extract_new_funs_from_lets([],_F,_XA,D,D).
extract_new_funs_from_lets([L1|T],F,XA,D,ND) :-
   extract_new_funs_from_let(L1,F,XA,D,ND1),
   extract_new_funs_from_lets(T,F,XA,ND1,ND).
   
extract_new_funs_from_let(bindval(Fun,Body,SRCSPAN),F,XARGS,DEFS, NewDEFS) :- !,
   extract_new_funs_from_let(agent(Fun,Body,SRCSPAN),F,XARGS,DEFS, NewDEFS).
extract_new_funs_from_let(agent_curry(Fun,Body,SRCSPAN),F,XARGS,DEFS, NewDEFS) :- !,
   translate_agent_curry(Fun,Body,TFun,TBody,SRCSPAN),
   extract_new_funs_from_let(agent(TFun,TBody,SRCSPAN),F,XARGS,DEFS, NewDEFS).
extract_new_funs_from_let(agent(Fun,_Body,_SRCSPAN),F,XARGS,DEFS, NewDEFS) :- 
     %Fun=.. [FunF|_FunArgs],
     functor(Fun,FunF,Arity),
     (builtin_functor(FunF,Arity)
       -> add_error(extract_new_funs_from_let,'Not a user-definable function:',FunF/Arity,_SRSCPAN),
          NewF = FunF
        ;  (is_local_let(Fun,DEFS,_,_,NewF)
		   ->   %print(multiple_equations_for(F,Fun,SRCSPAN,NewF)),nl,
				assert(multiple_let_equations_for(NewF,Fun))
		   ;  gensym(FunF,FunFGS),
			  string_concatenate('*',FunFGS,P2),
			  string_concatenate(F,P2,NewF)   /* This is not really required: but makes it more readable?? */
		  )
      ),
	  debug_println(9,rename_let(FunF,NewF,XARGS)),
	  NewDEFS = [  local_let(FunF,Arity, NewF,XARGS) | DEFS].
% commented out: is_possible_csp_process is computed by analyze_recursively_possible ????
%	  functor(NewFun,CF,CN), functor(CallSkel,CF,CN).
%	  (is_possible_csp_process(CallSkel) -> true
%	   ; assert( is_possible_csp_process(CallSkel) ) ).




builtin_functor('tuplePat',1).
builtin_functor(dotpat,1).
builtin_functor(Nr,0) :- number(Nr).
% TO DO: add more cases


   
   /* lookup if agent is a locally defined one and if so return the new call */
is_local_let(X,DEFS,Res,XARGS,NewF) :-  % print(check_is_local_let(X,DEFS,Res)),nl,
   functor(X,F,Arity),
   member(local_let(F,Arity,NewF,XARGS),DEFS),
   X =.. [_|Args],
   append(Args,XARGS,NA),
   Res =.. [NewF|NA].%, print(res(Res)),nl.
   
% same as above but with functor and args treated seperately
is_local_let2(Fun,OldArity,Args,DEFS,NewF,NA,XARGS) :-  % print(check_is_local_let(X,DEFS,Res)),nl,
   member(local_let(Fun,OldArity,NewF,XARGS),DEFS),
   append(Args,XARGS,NA). %, print(res(Res)),nl. 
   
   
:- dynamic cur_numbervars_index/1.
cur_numbervars_index(0).

reset_cur_numbervars_index :- retractall(cur_numbervars_index(_)),
   assert(cur_numbervars_index(0)).
mynumbervars(Term) :-
   cur_numbervars_index(Idx),
   numbervars(Term,Idx,NewIdx),
   retract(cur_numbervars_index(Idx)),
   assert(cur_numbervars_index(NewIdx)).

% ------------------------------------------


csp_span(no_loc_info_available).
csp_span(src_span(_,_,_,_,_,_)).
csp_span(src_position(_,_,_,_)).
csp_span(span_info(_,_)).
csp_span(multi_span(_,_,_,_,_,_,_)).
csp_span(src_span_operator(_,_)).

% visited_expression(ID,E), haskell_csp:clear_span(E,CE), visited_expression(ID2,E2), ID2@>ID, haskell_csp:clear_span(E2,CE).
clear_span(X,R) :- var(X),!,R=X.
clear_span('|~|'(X,Y,SrcSpan),R) :- R='|~|'(CX,CY,SrcSpan), 
   % keep span for internal choice; it is currently being used for tau-priority partial order reduction
   clear_span(X,CX), clear_span(Y,CY).
%clear_span(prefix(Span,NewVals,NewChExpr,X,_),R) :- !,
%   R=prefix(Span,NewVals,NewChExpr,CX,no_loc_info_available), 
%   clear_span(X,CX).
clear_span(X,R) :- csp_span(X),!,R=no_loc_info_available.
clear_span(X,R) :- atomic(X),!,R=X.
clear_span(X,R) :- X=..[F|A], maplist(clear_span,A,CA), R=..[F|CA].

% ------------------------------------------
   
opt_csp_listing :- debug_mode(on) -> csp_listing ; true.

csp_listing :- 
   nl,
   agent_compiled_without_constraints(Head,Body,_Span),
   portray_clause((Head :- Body)),
   functor(Head,F,A),
   (agent_parameters(F,A,List)
     ->  (List=[] -> true ; (print(' /* '), print(List), print('*/')))
     ;   print('  **** NO agent_parameter fact ! ****')
   ),nl,
   fail.
csp_listing.

get_internal_csp_representation(CodesClauses) :-
   findall(Clause, csp_clause(Clause), CL), append(CL,CodesClauses).

csp_clause(Clause) :- channel_type_list(C,T),
  format_to_codes('~q.~N',[channel_type_list(C,T)],Clause).
csp_clause(Clause) :- name_type(C,T),
  format_to_codes('~q.~N',[name_type(C,T)],Clause).
csp_clause(Clause) :- 
   haskell_csp_analyzer:agent_compiled_with_constraints(Head,Body,_Span,Constr),
   functor(Head,F,A),
   (agent_parameters(F,A,List) ->  true ;   List = no_agent_parameter_fact),
   numbervars((Head,Body,Constr),0,_),
   %portray_clause((Head :- Body)),
   ((List=[],Constr=true) -> format_to_codes('~q :-~N     ~q.~N',[Head,Body],Clause)
    ; List=[] -> format_to_codes('~q :- ~q, /* Constraints */~N     ~q.~N',[Head,Constr,Body],Clause)
    ; Constr\=true -> format_to_codes('~q :- /* ~p */~N     ~q, /* Constraints */~N     ~q.~N',[Head,List,Constr,Body],Clause)
    ; format_to_codes('~q :- /* ~p */~N     ~q.~N',[Head,List,Body],Clause)).
