% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(disprover_test_runner, [test_repl/0, run_disprover_on_all_pos/1, load_po_file/1,
       set_disprover_timeout/1, reset_disprover_timeout/0,
       set_disprover_options/1,
       print_disprover_stats/0]).

:- use_module(pathes,[set_search_pathes/0]). % called first to set_compile_time_search_pathes
%:- set_search_pathes.

portray_message(informational, _).

:- use_module(module_information).
:- module_info(group,cbc).
:- module_info(description,'Rodin Prover / Disprover').


:- use_module(eclipse_interface).
:- use_module(prob2_interface).
:- use_module(disprover).
:- use_module(error_manager).
:- use_module(tools).
:- use_module(solver_interface).
:- use_module(debug).
:- use_module(preferences).

:- use_module(library(system)).
:- use_module(library(lists)).



test_repl :- set_search_pathes,
   init_eclipse_preferences,
   disprover_load_at_repl_startup,
   disprover_help,
   disprover_repl_loop.


disprover_repl_loop :- print('DISPROVER ==> '),safe_read(T), 
  (quit_command(T) -> true; 
   test_eval(T) -> disprover_repl_loop
   ; print('Command failed !!'),nl, disprover_repl_loop).

safe_read(T) :- on_exception(E,read(T),
 ( format('Exception while reading user input: ~w ~n',[E]), T=help)).

disprover_help :- print('Available commands: run, run(PO), timeout(T), stats, list, q, x, FILE, v, vv, write(S,FILE)'),nl.

:- meta_predicate wall(0).
wall(Call) :-
    statistics(walltime,[Start,_]),
    call(Call),
    statistics(walltime,[Stop,_]), WT is Stop-Start,
    format('Walltime: ~w ms~n',[WT]).

quit_command(q).
quit_command(quit).
quit_command(halt).

test_eval(run) :- !,wall(run).
test_eval(run(PO)) :- !,disprover_timeout(TO), wall(run_on_po(PO,TO)).
test_eval(timeout(T)) :- !, set_disprover_timeout(T).
test_eval(time_out(T)) :- !, set_disprover_timeout(T).
test_eval(stats) :- !,print_disprover_stats.
test_eval(x) :- !,halt.
test_eval(list) :- !,list(_).
test_eval(lup) :- !,list(unknown).
test_eval(lf) :- !,list(false).
test_eval(help) :- !,disprover_help.
test_eval(h) :- !,disprover_help.
test_eval(v) :- !,tcltk_turn_debugging_on(19).
test_eval(vv) :- !,tcltk_turn_debugging_on(4).
test_eval(load(FILE)) :- !,load_po_file(FILE).
test_eval(wp) :- !, write_pos(true,_).
test_eval(write(Status)) :- !, write_pos(Status,_).
test_eval(wc) :- !, write_pos(true,contradiction_in_hypotheses).
test_eval(wp(FILE)) :- !, tell(FILE), write_pos(true,_), told.
test_eval(write(Status,FILE)) :- !, tell(FILE), write_pos(Status,_), told.
test_eval(cse) :- !,enable_cse.
test_eval(core) :- !,add_opts([disprover_option(unsat_core),unsat_core_algorithm/linear]).
test_eval(corem) :- !,add_opts([disprover_option(unsat_core),disprover_option(unsat_core_as_machine),unsat_core_algorithm/linear]).
test_eval(coredc) :- !,add_opts([disprover_option(unsat_core)]).
test_eval(observe) :- !,add_opts([disprover_option(observe_state)]).
test_eval(pob) :- !,add_opts([disprover_option(show_po_as_machine)]).
test_eval(kodkod) :- !,enable_kodkod.
test_eval(abort) :- !, print('TRY FIND ABORT true'),nl,set_preference(find_abort_values,true).
test_eval(nocounter) :- print('ENUMERATE_INFINITE_TYPES=FALSE'),nl,!, set_preference(allow_enumeration_of_infinite_types,false).
test_eval(trace_info) :- !, print('TRACE_INFO true'),nl,set_preference(provide_trace_information,true).
test_eval(trace) :- !, print('Tracing upon error'),nl,set_preference(trace_upon_error,true).
test_eval(debug) :- !,
  on_exception(error(existence_error(_,_),_),debug, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')).
test_eval(eval(Atom)) :- !,
   eval_strings:eval_string(Atom,Result),
   format('Evaluation result: ~w~n',[Result]).
test_eval(dir) :- 
    loaded_po_file(File),!,
    directory_listing_for_file(File).
test_eval(PO) :- disprover_po(PO,_,_,_,_,_),!,
   test_eval(run(PO)).
test_eval(Filename) :- format('Trying to consult file: ~w~n',[Filename]),
    load_po_file(Filename), print_disprover_stats,!.
test_eval(Cmd) :- format('Could not executed command: ~w~n',[Cmd]).


:- use_module(library(file_systems)).
% list the directory in which a file is in and allow use to load a po file in that directory
directory_listing_for_file(File) :- 
   get_parent_directory(File,Dir),
   findall(FF,(file_member_of_directory(Dir,F,FF),get_filename_extension(F,'pl')),List),
   sort(List,SList),
   prdir(SList),
   write('==Nr==>'), read(Nr),
   ((number(Nr),nth1(Nr,SList,NewFile)) -> load_po_file(NewFile) ; print('No file loaded'),nl).
prdir(List) :-   nth1(N,List,File), format(' ~w -> ~w~n',[N,File]), fail.
prdir(_).

print_disprover_stats :- 
         disprover_timeout(TO),
         project_name(N), format('Project ~w~n',[N]),
         generated(_,D),
         machine_name(M), format('Machine ~w (exported ~w)~n',[M,D]),
         print_nr_pos('',_,_),
         print_nr_pos('proven (within Rodin)',true,_),
         format('ProB Disprover Analysis with timeout ~w ms~n',[TO]),
         print_nr_pos('proven     (by ProB)',_,true),
         (disprover_result(_,_,contradiction_in_hypotheses)
           -> print_nr_pos('those with contradiction in hypotheses',_,true,contradiction_in_hypotheses) ; true),
         print_nr_pos('disproven  (by ProB)',_,false),
         print_nr_pos('unproven   (by ProB)',_,unknown),
         (raw_result_exists(failure) -> print_nr_pos('failure    (of ProB)',_,failure) ; true).

raw_result_exists(RawResult) :- disprover_po_with_info(_,_,_,RawResult).
print_nr_pos(Msg,RodinStatus,ProBStatus) :- print_nr_pos(Msg,RodinStatus,ProBStatus,_).
print_nr_pos(Msg,RodinStatus,ProBStatus,RawResult) :-
         findall(Name,disprover_po_with_info(Name,RodinStatus,ProBStatus,RawResult),L),
         length(L,Len),
         ((L = [First|_],last(L,Last))
           -> format('Number of ~w POs: ~w (~w -- ~w)~n',[Msg,Len,First,Last])
           ;  format('Number of ~w POs: ~w~n',[Msg,Len])
          ).

disprover_po_with_info(Name,RodinStatus,ProBStatus,RawResult) :-
   disprover_po(Name,_Context,_Goal,_All,_Sel,RodinStatus),
   (disprover_result(Name,ProBStatus,RawResult) -> true ; ProBStatus='not_run', RawResult='not_run').

:- dynamic disprover_timeout/1.
disprover_timeout(300).
set_disprover_timeout(N) :- retractall(disprover_timeout(_)), assert(disprover_timeout(N)).
reset_disprover_timeout :- retractall(disprover_timeout(_)), assert(disprover_timeout(300)).

run :- run_disprover_on_all_pos(_).

:- use_module(tools_lists,[count_occurences/2]).

run_disprover_on_all_pos(_) :- retractall(disprover_result(_,_,_)), disprover_timeout(TO),run_on_po(_,TO), fail.
run_disprover_on_all_pos(Summary) :- findall(R,disprover_result(_,R,_),L),
       count_occurences(L,Occs),
       format('~nDisprover Run Summary: ~w~n',[Occs]),
       add_info(true,Occs,O1), add_info(false,O1,O2), add_info(unknown,O2,O3),
       add_info(failure,O3,Summary).

add_info(Info,List,Result) :- member(Info-_,List) -> List=Result ; Result = [Info-0|List].


run_on_po(PO_Name,TimeOut) :-
    disprover_po(PO_Name,Context,Goal,AllHypotheses,SelectedHypotheses,Status),
    format('Testing ~w (~w)~n',[PO_Name,Status]),
    load_event_b_project([],[Context],[],_),
    start_animation,
    debug_println(5,loaded_context),
    (my_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeOut,OutResult)
     -> translate_result_for_po(OutResult,ProverResult),
        format('Disprover Result ~w (PO is ~w)~n',[OutResult,ProverResult]),
        functor(OutResult,RawOutFunctor,_),
        assert(disprover_result(PO_Name,ProverResult,RawOutFunctor))
     ;  add_error(disprover,'Disprover failed: ',(PO_Name:Status)),
        assert(disprover_result(PO_Name,'failure',unknown))
     ).

translate_result_for_po(solution(_),R) :- !, R=false.
translate_result_for_po(contradiction_found,R) :- !,R=true.
translate_result_for_po(contradiction_in_hypotheses,R) :- !,R=true.
translate_result_for_po(_,unknown).

:- dynamic extra_opts/1.
extra_opts([]).
add_opts(Extra) :-
  retract(extra_opts(Old)),
  append(Extra,Old,New), sort(New,SNew),
  assert(extra_opts(SNew)).
enable_cse :- XO = [use_common_subexpression_elimination/true,
                     use_common_subexpression_also_for_predicates/true],
        add_opts(XO).
enable_kodkod :- XO = [try_kodkod_on_load/true, kodkod_for_components/true], add_opts(XO).
set_disprover_options(L) :-
   retractall(extra_opts(_)),
   assert(extra_opts(L)).

my_disprove(Goal,AllHypotheses,SelectedHypotheses,Timeout,OutResult) :- 
    Opts = [use_smt_mode/true,use_clpfd_solver/true,use_chr_solver/true],
    % interesting extra opts : disprover_option(unsat_core),unsat_core_algorithm/linear
    extra_opts(X), append(X,Opts,AllOpts),
    % Note: disprover_mode is always set to true
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,Timeout,AllOpts,OutResult).

:- use_module(translate).
:- use_module(bmachine,[type_with_errors/4]).

list(DPRes) :- disprover_po(Name,Context,Goal,_All,_Sel,Status),
   get_disprover_result(Name,DPRes,RawResult),
   format('~w (~w : ~w [~w]) : ',[Name,Status,DPRes,RawResult]),
   load_event_b_project([],[Context],[],_),
   start_animation,
   type_check_in_machine_context([Goal],[TypedGoal]),
   translate:print_bexpr(TypedGoal),nl,fail.
list(_).

:- dynamic project_name/1.
:- dynamic machine_name/1.
:- dynamic generated/2.
:- dynamic disprover_po/6.
:- dynamic disprover_result/3.
project_name(unknown).
machine_name(unknown).
generated(unknown,unknown).
disprover_po(unknown,unknown,unknown,unknown,unknown,unknown).
disprover_result(unknown,unknown,unknown).

reset_po_facts :- retractall(loaded_po_file(_)),
   retractall(project_name(_)),
   retractall(machine_name(_)), retractall(generated(_,_)),
   retractall(disprover_po(_,_,_,_,_,_)), retractall(disprover_result(_,_,_)).

% write POs which have a certain status; can be used to generate .pl files with only proven or not-proven POs
write_pos(Status,RawResult) :- 
  format('% Proof Obligations with status = ~w~n',[Status]),
  write_facts(project_name(_)), write_facts(machine_name(_)),
  write_facts(generated(_,_)),
  disprover_result(Name,Status,RawResult),
  write_facts(disprover_po(Name,_Context,_Goal,_All,_Sel,_RodinStatus)),
  fail.
write_pos(_,_) :- nl.

write_facts(Fact) :- Fact, 
   write_term(Fact,[quoted(true),numbervars(true)]), write('.'),nl,fail.
write_facts(_).

get_disprover_result(PO_Name,Result,RawResult) :-
   (disprover_result(PO_Name,R,Raw) -> Result=R,RawResult=Raw
    ; Result = 'not_yet_run', RawResult='not_yet_run').

:- dynamic loaded_po_file/1.
load_po_file(File) :-  reset_po_facts,
    % consult_without_redefine_warning
    prolog_flag(redefine_warnings, Old, off),
    (% load in dynamic mode: so that we can retract later
     my_po_consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    (OK=true -> assert(loaded_po_file(File)) ;
      add_error_and_fail(bmachine,'Consulting of File failed: ',File)).

my_po_consult(File) :- on_exception(error(existence_error(_,_),_),
                    load_files(File,[compilation_mode(assert_all)]), 
                    add_error_fail(my_consult,'File does not exist: ',File)).


disprover_load_at_repl_startup :- prolog_flag(argv,ArgV),ArgV = [FILE|T],
  (T=[] -> load_po_file(FILE)
   ; format('*** Please provide only one command-line argument: ~w~n',[ArgV]),fail).
disprover_load_at_repl_startup  :-
  print('Loading default file'),nl,
  load_po_file('../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl').
disprover_load_at_repl_startup.
  
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M2_mch.pl').
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M3_mch.pl').
% :- load_po_file('../../prob_examples/examples/ProofObligations/ABZ14_Mery/M4_mch.pl').
% :- load_po_file('../../prob_examples/public_examples/EventBPrologPackages/Disprover/ProofTests/DefSetCst_ctx.pl').
% '/Users/leuschel/git_root/prob_examples/public_examples/SMT/QF_IDL/queens_bench/toroidal_bench/toroidal_queen_2_QF_IDL_ctx.pl'.

%:- test_repl.

