% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_dif,[frozen_dif/2, syntactic_equality_test/3, post_dif/2]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides a dif which works with reified equality.').

% check if a frozen dif co-routine is pending and stating that the variables are different
% cf testcase 1175
%frozen_dif(X,Y) :- preferences:preference(use_smt_mode,false),!,fail.
frozen_dif(X,Y) :-
   % print(check_frozen(X,Y)),nl, translate:l_print_frozen_info([X,Y]),nl,
   var(X),!,frozen(X,Goal),
   frozen_dif_aux(Goal,5,X,Y).
frozen_dif(X,Y) :- var(Y), frozen(Y,Goal), % do we need this in clpfd mode ?
   frozen_dif_aux(Goal,2,Y,X).
frozen_dif_aux(prolog:dif(A,B),_,X,Y) :-
   (A==X,B==Y ; A==Y,B==X),
   !. % , print(frozen_dif(X,Y)),nl.
% Nr: put an upper bound on checks; avoid slow-down when frozen goal very large; see e.g. alloc_large in test 1220
% TO DO: use attribute and get more efficient lookup
frozen_dif_aux((A,B),Nr,X,Y) :- Nr>1, N1 is Nr-1,
  (frozen_dif_aux(A,N1,X,Y) -> true ; frozen_dif_aux(B,N1,X,Y)).


% a reified version for syntactic equality ==

% trigger Res to pred_true when two variables become identical (as CLPFD does not do this)
% it is supposed to be used in conjunction with ((X#=Y) #<=> R01)
%syntactic_equality_test(X,Y,Res) :- Res==pred_true, !, print(useless_syntactic_test(X,Y,Res)),nl.
syntactic_equality_test(X,Y,Res) :-
  % TO DO: attach attribute/co-routine so that dif will trigger Res to be pred_false ?
  % e.g., (X=Y <=> r=TRUE) & X:1..100 & Y /= X does not deterministically set r to FALSE (but does if Y/=X moved to the left)
  %reify(X,Y,Res), % a way to attach information for frozen_eqeq check
   %when(?=(X,Y),(X==Y -> Res=pred_true ; true)). % CLPFD does not detect variable equality
   when((?=(X,Y) ; nonvar(Res)),syntactic_eq_prop(Res,X,Y)). % CLPFD does not detect variable equality

syntactic_eq_prop(Res,X,Y) :- var(Res), !, (X==Y -> Res=pred_true
   ; true). % syntactically not equal, but they could still be of identical value ?!
syntactic_eq_prop(pred_true,X,Y) :- ((var(X),var(Y)) -> X=Y ; true). % CLPFD reification does not unify ! required for test 1525
syntactic_eq_prop(pred_false,X,Y) :- dif(X,Y). % or frozen_dif ??

:- block reify(-,-,-).
reify(_,_,_).



post_dif(X,Y) :- dif(X,Y),
  (frozen_eqeq(X,Y,Res) -> print(frz(X,Y,Res)),nl, Res=pred_false ; true).

%kernel_dif:syntactic_equality_test(X,Y,R), kernel_dif:post_dif(X,Y), -> R==pred_false

frozen_eqeq(X,Y,R) :-
   % print(check_frozen(X,Y)),nl, translate:l_print_frozen_info([X,Y]),nl,
   var(X),!,frozen(X,Goal),
   frozen_eqeq_aux(Goal,5,X,Y,R).
frozen_eqeq(X,Y,R) :- var(Y), frozen(Y,Goal), % do we need this in clpfd mode ?
   frozen_eqeq_aux(Goal,2,Y,X,R).
frozen_eqeq_aux(kernel_dif:reify(A,B,R),_,X,Y,R) :-
   (A==X,B==Y ; A==Y,B==X),
   !. % , print(frozen_dif(X,Y)),nl.
% Nr: put an upper bound on checks; avoid slow-down when frozen goal very large; see e.g. alloc_large in test 1220
% TO DO: use attribute and get more efficient lookup
frozen_eqeq_aux((A,B),Nr,X,Y,R) :- Nr>1, N1 is Nr-1,
  (frozen_eqeq_aux(A,N1,X,Y,R) -> true ; frozen_eqeq_aux(B,N1,X,Y,R)).
