% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(prob2_interface,
   [initialise_specification/0,
    get_version/7,

    load_classical_b_from_list_of_facts/2,
    update_preferences_from_spec/0,

    load_event_b_project/4,

    load_cspm_spec_from_cspm_file/1,

    reset_animator/0, clear_animator/0, start_animation/0,

    serialize/2,
    deserialize/2,

    get_state/2,
    get_b_state/2,

    compute_operations_for_state/2,
    prob2_execute_custom_operations/6,

    get_op_from_id/4,

    is_initialised_state/1,
    state_property/3,
    op_timeout_occurred/2,

    get_state_errors/2,

    evaluate_formula/3,
    evaluate_registered_formulas/3,
    register_prob2_formula/2,
    evaluate_formula_typecheck/5,

    get_states_for_predicate/3,

    filter_states_for_predicate/3,

    insert_formula_for_expansion/2,
    expand_formula/3,

    do_modelchecking/4,
    set_goal_for_model_checking/1,
    compute_coverage/5,
    prob2_deadlock_freedom_check/2,
    prob2_invariant_check/2,

    prob2_do_ltl_modelcheck/4,

    activate_plugin/1,
    get_plugin_output/2,

    find_trace_to_node/2,
    find_trace_from_node_to_node/3,
    find_state_for_predicate/2,

    cbc_type_check_predicate/2,
    cbc_disprove/5, cbc_disprove/6,
    cbc_solve/3,
    cbc_generate_test_cases/3,

    check_csp_assertions/3,

    list_eclipse_preferences/1,
    list_current_eclipse_preferences/1,
    get_eclipse_preference/2,
    set_eclipse_preference/2,

    get_signature_merge_state_space/2,
    get_transition_diagram/2,

    write_dotty_transition_diagram/2,
    write_dotty_signature_merge/2,
    write_dot_for_state_viz/2,
    write_dotty_state_space/1,

    get_error_messages/2,

    es_copy_from_statespace/2,
    es_eval/4,

    generate_trace_until_condition_fulfilled/4,
    get_unsat_core_with_fixed_conjuncts/3,
    get_minimum_unsat_core_with_fixed_conjuncts/3,

    prob2_construct_trace/6,

    % These predicates should probably not be exported after the end of the migration
    convert_cli_arg/2,


    get_pretty_print/1
   ]).

:- use_module(state_space,              [time_out_for_node/1, time_out_for_node/3,
                                         time_out_for_invariant/1, max_reached_for_node/1,
                                         max_reached_or_timeout_for_node/1,
                                         visited_expression/2, transition/4, visited_expression_id/1,
                                         set_context_state/1, clear_context_state/0,
                                         state_error/3, current_state_id/1,
                                         get_state_space_stats/3, state_space_initialise/0,
                                         history/1, any_transition/3,set_current_options/1]).
% new profiler
%:- use_module('../extensions/profiler/profiler.pl').

% old profiler
:- use_module(runtime_profiler,         [reset_runtime_profiler/0]).

:- use_module(symmetry_marker,          [precompile_marker_typing_info/0]).
:- use_module(value_persistance,        [initialise_operation_caching/0]).
:- use_module(eventhandling,            [announce_event/1]).
:- use_module(self_check,               [turn_off_run_time_type_checks/0]).
:- use_module(flow,                     [reset_flow/0]).
:- use_module(refinement_checker,       [reset_refinement_checker/0]).
:- use_module(reduce_graph_state_space, [reduce_graph_reset/0]).
:- use_module(version,                  [version/4, revision/1, lastchangeddate/1]).
:- use_module(pref_definitions,         [b_get_preferences_from_machine/0]).
:- use_module(bsyntaxtree,              [conjunct_predicates/2, get_texpr_ids/2,
                                         get_texpr_expr/2,
                                         find_identifier_uses/3]).
:- use_module(evalstores,               [evalstore_bindings/2, evalstore_create_store_by_state/2, evalstore_eval_formula/4]).
:- use_module(model_checker,            [is_deadlocked/1]).
:- use_module(state_space_exploration_modes,[set_depth_breadth_first_mode/1]).
:- use_module(bmachine,                 [b_type_expression/5, b_get_machine_operation/4,
                                         b_type_open_predicate/5, b_top_level_operation/1,
                                         b_get_machine_operation/5, b_load_machine_from_list_of_facts/2,
                                         b_set_eventb_project_flat/3, load_additional_information/1,
                                         b_machine_precompile/0, other_spec_precompile/0,
                                         b_set_parsed_typed_machine_goal/1,
                                         b_type_open_predicate_with_errors/4]).
:- use_module(probltlsrc(ltl),          [ltl_model_check_with_ce/4,preprocess_formula/2]).
:- use_module(probltlsrc(ltl_tools),    [typecheck_temporal_formula/3,temporal_parser/3]).
:- use_module(probltlsrc(ltl_verification), [evaluate_ltl_formula/6]).
:- use_module(b_interpreter,            [b_compute_expression/4]).
:- use_module(b_global_sets,            [add_prob_deferred_set_elements_to_store/3]).
:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2,
                                                     ignore_user_interrupt_det/1]).
:- use_module(library(lists),           [maplist/3, maplist/4, reverse/2]).
:- use_module(library(codesio),         [read_from_codes/2, write_term_to_codes/4]).
:- use_module(specfile,                 [expand_const_and_vars_to_full_store/2, animation_mode/1,
                                         b_or_z_mode/0, set_currently_opened_file/1,
                                         set_animation_mode/1, set_animation_minor_mode/1]).
:- use_module(translate,                [translate_bvalue/2, suppress_rodin_positions/0,
                                         translate_event/2, translate_bstate/2, translate_state_error/2,
                                         translate_bexpression/2, translate_bvalue_with_limit/3,
                                         translate_event_error/2, explain_event_trace/2, explain_state_error/3]).
:- use_module(error_manager,            [add_error_and_fail/3, get_all_errors_and_reset/1,
                                         add_error/3, add_error/2, add_error_and_fail/2,
                                         add_message/2, reset_errors/0, get_error/2,
                                         catch_enumeration_warning_exceptions/2,
                                         real_error_occurred/0]).
:- use_module(tools,                    [string_concatenate/3, split_atom/3,
                                         safe_atom_codes/2,ajoin/2]).
:- use_module(kernel_tools,             [map_over_bvalue/3]). % not used ??
:- use_module(bvisual2,                 [bv_insert_formula/3, bv_expand_formula/3,
                                         bv_get_values/3]).
:- use_module(preferences,              [eclipse_preference/2, get_preference/2,
                                         preference_description/2, preference_val_type/2,
                                         preference_default_value/2, preference_category/2,
                                         advanced_eclipse_preference/2, set_preference/2,
                                         obsolete_eclipse_preference/1]).
:- use_module(disprover,                [disprove/5, disprove_with_opts/6]).
:- use_module(solver_interface,         [solve_predicate/4, type_check_in_machine_context/2,
                                         call_with_smt_mode_enabled/1]).
:- use_module(plugins,                  [start_plugin_for_current_mode/1,plugin_get_output/2]).
:- use_module(probcspsrc(haskell_csp),  [parse_and_load_cspm_file/1,load_cspm_pl_file/1]).
:- use_module(state_space_reduction,    [reset_ignored_events/0, set_ignored_events/1,
                                         compute_signature_merge/0, compute_transition_diagram/1,
                                         get_reduced_node/4, reduced_trans/5,
                                         generate_node_color/5, generate_node_labels/3,
                                         generate_transition_label/3, generate_transition_color_and_style/6,
                                         write_transition_diagram_for_expr_to_dotfile/2,
                                         write_signature_merge_to_dotfile/2]).
:- use_module(graph_canon,              [print_cstate_graph/2]).
:- use_module(debug, [debug_println/2]).
:- use_module(unsat_cores, [unsat_core_with_fixed_conjuncts/3,
                            minimum_unsat_core_with_fixed_conjuncts/3]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/**
Access information about the current version of the ProB core.

#### called by:
* ProB 2.0: GetVersionCommand
*/
get_version(Major,Minor,Service,Qualifier,SvnRevision,LastChangedDate,PrologInfo) :-
   version(Major,Minor,Service,Qualifier),
   revision(SvnRevision),  lastchangeddate(LastChangedDate), prolog_flag(version,PrologInfo).

/* ------------------------- */
/* B Loading Interface       */
/* ------------------------- */

/**
Loads a classical b model.

#### called by:
* ProB 2.0: LoadBProjectCommand
*/
:- mode load_classical_b_from_list_of_facts(+MainFilename,+ListOfFacts).
load_classical_b_from_list_of_facts(MainFilename,ListOfFacts) :-
   clear_animator,
   b_load_machine_from_list_of_facts(MainFilename,ListOfFacts),
   update_preferences_from_spec.

/**
update preferences from SET_PREF Definitions in B machines:
should be called after loading a model and before start_animation
*/
update_preferences_from_spec :- (b_or_z_mode -> b_get_preferences_from_machine ; true).

/* ------------------------- */
/* Event-B Loading Interface */
/* ------------------------- */

/**
Loads an Event-B model.

#### called by:
* ProB Plugin: de.prob.eventb.translator.internal.EventBTranslator, DisproverLoadCommand
* ProB 2.0: de.prob.model.eventb.translate.EventBModelTranslator (LoadEventBProjectCommand)
*/
load_event_b_project(Machines,Contexts,Proofs,Errors) :-
    clear_animator,
    set_animation_mode(b), set_animation_minor_mode(eventb),
    b_set_eventb_project_flat(Machines,Contexts,Proofs),
    load_additional_information(Proofs),
    Errors = [].

/* ------------------------- */
/* CSP Loading Interface     */
/* ------------------------- */

/**
Takes a path to a CSPM specification and loads the file using the CSPM parser.

#### called by:
* ProB 2.0: LoadCSPCommand
*/
:- mode load_cspm_spec_from_cspm_file(+CSPMFile).
load_cspm_spec_from_cspm_file(CSPMFile) :-
   clear_animator,
   set_animation_mode(cspm),
   parse_and_load_cspm_file(CSPMFile),
   set_currently_opened_file(CSPMFile).


/* ------------------------- */
/* Animation                 */
/* ------------------------- */

reset_animator :- announce_event(reset_specification). % just reset; keeping same spec
clear_animator :- reset_animator,announce_event(clear_specification).

start_animation :- initialise_animator,!.
start_animation :- add_error(start_animation,'Start Animation Failed').

initialise_animator :-

   reset_errors,
   %NOW done by events: reset_flow,    state_space_initialise, reduce_graph_reset, reset_runtime_profiler,
   %  reset_refinement_checker reset_model_checker,
%   reset_dynamics, % new profiler
   initialise_specification.


initialise_specification :-  /* call once before starting animation or model checking */
   announce_event(start_initialising_specification),
   ( b_or_z_mode -> b_machine_precompile
   ; otherwise ->
       preferences:get_preference(symmetry_mode,X),
       ( X=off -> true
       ; otherwise ->
           add_message(initialise_specification,'Symmetry can only be used for B & Z specifications'),
           preferences:set_preference(symmetry_mode,off) /* Turn symmetry off */
       ),
       tools:print_bt_message(other_spec_precompile),
       ~~mnf(other_spec_precompile)
   ),
      %TODO(DP,11.8.2008)
   user:initialise_operation_not_yet_covered,
   % now done via event_handline: reset_external_functions,
   preferences:get_preference(symmetry_mode,SymMode),
   symmetry_marker:precompile_marker_typing_info,
   ( SymMode=nauty ->
       (graph_canon:initialise_nauty -> true ; preferences:set_preference(symmetry_mode,off))
   ; otherwise -> graph_canon:clear_nauty),
   announce_event(specification_initialised),
   ( initialise_operation_caching -> true
   ; otherwise -> add_error_and_fail(b_load_machine,'Initialising operation caching failed')).

/* ------------------------- */
/* Handling States           */
/* ------------------------- */

/**
Produces a prolog serialization of a given state id

#### called by:
* ProB Plugin: SerializeStateCommand
* ProB 2.0: SerializeStateCommand
*/
:- mode serialize(+Id,-SerializedState).
serialize(Id,SerializedState) :-
    visited_expression(Id,State),
    write_term_to_codes(State,Codes,".",[quoted(true)]),
    atom_codes(SerializedState,Codes).

/**
Deserialize a state and add it if necessary

#### called by:
* ProB Plugin: DeserializeStateCommand
* ProB 2.0: DeserializeStateCommand
*/
:- mode deserialize(-NewId,+SerializedState).
deserialize(NewId,SerializedState) :-
    atom_codes(SerializedState,Codes),
    read_from_codes(Codes,State),
    user:get_id_of_node_and_add_if_required(State,NewId,_Res,root).

get_state(es(ID),Store) :-           % use a store from the evalstore module instead of
    !,evalstore_bindings(ID,Store).  % a state from the state space
get_state(ID,BState) :-
    visited_expression(ID,State),
    get_state2(State,BState).
get_state2(csp_and_b(CSPState,BState),[bind('CSP_CONTROLLER',CSPState)|ExpandedBState]) :- !,
    get_state2(BState,ExpandedBState).
get_state2(concrete_constants(BState),BState) :- !.
get_state2(root,BState) :- !, BState=[].
get_state2(State,Res) :- expand_const_and_vars_to_full_store(State,BState),
    (BState=[_|_] -> Res=BState ; Res=[] /* we have [] or a CSP process or similar */).


/**
Get the predicate representation of a BState

#### called by:
* ProB 2.0: GetBStateCommand
*/
get_b_state(StateId, BState) :-
    get_state(StateId, S), translate_bstate(S, BState).

/**
Compute the enabled operations (without the backtrack options) for a given state id.
Transitions is a list of operation tuples with the form op(TransitionId,SrcId,DestId)

#### called by:
* ProB 2.0: GetEnabledOperationsCommand
*/
:- mode compute_operations_for_state(+StateID,-Transitions).
compute_operations_for_state(StateID, Transitions) :-
   visited_expression_id(StateID),
   user:tcltk_compute_options(StateID,OpsSTAndIDs),
   (create_simple_op_terms(OpsSTAndIDs,StateID,Transitions) -> true
     ; add_error_and_fail(prob2_interface, 'Creating op terms failed', OpsSTAndIDs)).

create_simple_op_terms([],_,[]).
create_simple_op_terms([(Id,Op,Dst)|T],StateID,[op(Id,Name,StateID,Dst)|FT]) :-
    extract_op_name(Op,Name),
    create_simple_op_terms(T,StateID,FT).

set_current_state(ID) :- /* jumps to the given node; can be backtracked */
   visited_expression_id(ID),
   user:tcltk_goto_state(jump,ID).

/**
Calculates an operation given a predicate from the user.

#### called by:
* ProB 2.0: GetOperationByPredicateCommand
*/
:- mode prob2_execute_custom_operations(+CurID, +OpName, +ParsedPredicate, +MaxNrOfSolutions, -TOperations, -Errors).
prob2_execute_custom_operations(_CurID, _OpName, _Pred, MaxNrOfSolutions, TOps, Errs) :-
    MaxNrOfSolutions < 1,
    !,
    TOps = [],Errs = ['max nr of solutions too small'].
prob2_execute_custom_operations(_CurID, OpName, _Pred, _Max, TOps, Errs) :-
    \+ valid_op_name(OpName),
    !,
    TOps = [], ajoin(['Unknown Operation ', OpName], Msg), Errs = [Msg].
prob2_execute_custom_operations(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut) :-
    set_current_state(CurID),
    prob2_execute_custom_operations_aux(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut).

% special case for TRUE=TRUE predicate to avoid parser overhead
prob2_execute_custom_operations_aux(CurID, OpName, equal(_,boolean_true(_),boolean_true(_)), MaxNrOfSolutions, TOperations, ErrorsOut) :-
    findall(TO, prob2_execute_custom_operation(CurID,OpName,TO,MaxNrOfSolutions), TOperations),
    clear_context_state,
    (TOperations = []
     -> ajoin(['Could not execute Operation ', OpName, ' with additional predicate'], Msg),
        ErrorsOut = [Msg]
     ;  ErrorsOut = []).
prob2_execute_custom_operations_aux(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut) :-
    ( (OpName = '$initialise_machine'; OpName =  '$setup_constants' )  -> Scope = [prob_ids(visible),variables];
      otherwise ->    b_top_level_operation(OpName),
                      b_get_machine_operation(OpName,_,Parameters,_),
                      Scope = [identifier(Parameters),prob_ids(visible),variables] % prob scope allows one to use identifiers for deferred set elements
    ),
    b_type_expression(ParsedPredicate, Scope, _, TypedPred, Errors),
    (Errors = [] ->
        retractall(bmachine:b_machine_temp_predicate(_)),
        assert(bmachine:b_machine_temp_predicate(TypedPred)),
        findall(TO, prob2_execute_custom_operation(CurID,OpName,TO,MaxNrOfSolutions), TOperations),
        retractall(bmachine:b_machine_temp_predicate(_)),
        clear_context_state,
        (TOperations = []
         -> ajoin(['Could not execute Operation ', OpName, ' with additional predicate'], Msg),
            ErrorsOut = [Msg]
         ;  ErrorsOut = [])
    ;   TOperations = [],
        ErrorsOut = [typechecker_errors(Errors)]
    ).

valid_op_name('$initialise_machine').
valid_op_name('$setup_constants').
valid_op_name(OpName) :- b_top_level_operation(OpName).
%valid_op_name(OpName) :- b_get_promoted_machine_operations(OpName).

/**

Calculates a trace given a list of operation names and a list of guards.
In case of errors, a partial trace is generated that jumps over the errornous operation / event.
In addition, a list of integers should be given as argument that make possible to execute some
operations in the oprations list multiple times or as long as they are disabled (in this case one should give -1).
E.g. by calling prob2_construct_trace(0,[e1,e2,e3],[TRUE,TRUE,TRUE],[2,1,-1],OpsOut,ErrOut) the predicate calculates
a trace starting at the state with the ID 0 where e1 is executed sequentially 2 times, after that e2 once,
and finally e3 until it becomes disabled.

#### called by:
* ProB 2.0: ConstructTraceCommand
*/

prob2_construct_trace(CurId,Names,Preds,Nrs,OpsOut,ErrOut) :-
    construct_trace(CurId,Names,Preds,Nrs,[],OpsOut,[],ErrOut).

construct_trace(_CurId,[],[],[],OpsOut,OpsOut,ErrOut,ErrOut).
construct_trace(CurId,[Name|Names],[Pred|Preds],[NrName|NrNames],CurOpsOut,OpsOut,CurErrOut,ErrOut) :-
    execute_nr_of_custom_operations(CurId,Name,Pred,NrName,NewId,ListOfOps,ListOfErrs),
    append(CurOpsOut,ListOfOps,NewOpsOut),
    append(CurErrOut,ListOfErrs,NewErrOut),
    construct_trace(NewId,Names,Preds,NrNames,NewOpsOut,OpsOut,NewErrOut,ErrOut).

execute_nr_of_custom_operations(CurId,_Name,_Pred,0,CurId,[],[]).
execute_nr_of_custom_operations(CurId,Name,Pred,Nr,NewId,ListOfOps,ListOfErrs) :-
    ( Nr = -1 ->
        execute_until_disabled(CurId,Name,Pred,NewId,ListOfOps,ListOfErrs)
    ; otherwise -> % Nr is greater than 0
        prob2_execute_custom_operations(CurId,Name,Pred,1,ListContainingSingleOp,SingleErr),
        ( ListContainingSingleOp = [] ->
            NextId=CurId
        ;   ListContainingSingleOp = [op(_Trans,_Name,_From,To)],
            NextId = To
        ),
        append(ListContainingSingleOp,Ops,ListOfOps),
        append(SingleErr,Errs,ListOfErrs),
        NewNr is Nr - 1,
        execute_nr_of_custom_operations(NextId,Name,Pred,NewNr,NewId,Ops,Errs)
    ).


execute_until_disabled(CurId,OpName,Pred,NewId,ListOfOps,ListOfErrs) :-
    execute_until_disabled(CurId,OpName,Pred,NewId,[],ListOfOps,ListOfErrs).

execute_until_disabled(CurId,OpName,Pred,NewId,CurListOfOps,ListOfOps,ListOfErrs) :-
    prob2_execute_custom_operations(CurId,OpName,Pred,1,ListContainingSingleOp,SingleErr),
    ( ListContainingSingleOp = [] -> % reached a state where OpName is not executed
        NewId=CurId,
        ListOfOps=CurListOfOps,
        ajoin(['Could not execute Operation ', OpName, ' with additional predicate'], Msg),
        (member(Msg,SingleErr) -> % Error message generated because OpName is disabled at some state
            ListOfErrs = []
        ; otherwise -> % unexpected error occurred, need to be reported
            ListOfErrs = SingleErr
        )
    ;   ListContainingSingleOp = [op(Trans,Name,From,To)],
        ( member(op(_T,_N,To,_To),CurListOfOps) -> % in case of a loop we terminate, otherwise we will go forever
            ListOfOps = CurListOfOps,
            NewId=From
        ; otherwise ->
            NextId = To,
            append(CurListOfOps,[op(Trans,Name,From,To)],NewCurListOfOps),
            execute_until_disabled(NextId,OpName,Pred,NewId,NewCurListOfOps,ListOfOps,ListOfErrs)
        )
    ).

% TO DO: check if we need to recompute the operation effect: if the ParsedPredicate is TRUE
%  (TRUE = equal(none,integer(none,1),integer(none,1)) ) & MaxNrOfSolutions <=
%        what has already been used previously, we can simply reuse transition from the state space

prob2_execute_custom_operation(CurID,OpName,Transition,Max) :-
    set_context_state(CurID),
    visited_expression(CurID,InState),
    specfile:compute_operation_effect_max(InState,OpName,Operation,NewState,_TransPathInfo,Max),
    % prob_socketserver:writeln_log(sol(OpName,NewState)), %%
    user:add_trans_id(CurID,Operation,NewState,[],NewID,[],TransId),
    % prob_socketserver:writeln_log(ids(OpName,Operation,NewID,TransId)), %%
    Transition = op(TransId,OpName,CurID,NewID).

/**
Extracts information about the parameters and return values for the
specified transition.

#### called by:
* ProB 2.0: GetOpFromId
*/
:- mode get_op_from_id(+Id,+Truncate,-Params,-RetVals).
get_op_from_id(Id,Truncate,Params,RetVals) :-
    transition(_Src,Op,Id,_Dest),
    create_op_tuple(Op,Truncate,Params,RetVals).

/**
Creates an operation tuple from transition id, source id, an op term, and a destination id.
If creation is unsuccessful, an error is added and the predicate fails.
See comment for extract_op_tuple for a description of OpTuple
*/
:- mode create_op_tuple(+OpTerm, +Truncate, -Params, -RetVals).
create_op_tuple(OpTerm, Truncate, Params, RetVals) :-
    (extract_op_tuple(OpTerm, Truncate, Params, RetVals) -> true;
        add_error_and_fail(prob2_interface, 'Could not create OpTuple ', extract_op_tuple(OpTerm, Params, RetVals))).

/**
Extracts the parameters and return values of the operations
*/
:- mode extract_op_tuple(+OpTerm, +Truncate, -Params, -RetVals).
extract_op_tuple(OpTerm, Truncate, Params, RetVals) :-
    (animation_mode(cspm) ->
        extract_csp_op(OpTerm, Params, RetVals)
    ;   extract_b_op(OpTerm, Truncate, Params, RetVals)
    ).

/**
Extracts information for a CSP operation.
*/
:- mode extract_csp_op(+OpTerm, -Params, -RetVals).
extract_csp_op(OpTerm, Params, []) :-
    %OpTuple = op(TransId, Src, Dest, Name, params(Source, Pretty), return([],[]), TargetState), /* TODO: CSP operations do not have return types, is this correct??? */
    %extract_csp_name_and_args(OpTerm, _Name, _Source),
    translate_event(OpTerm, PPEvent),
    split_atom(PPEvent,['.','!'],[_Op|Params]).

:- mode extract_csp_name_and_args(+OpTerm, -Name, -Args).
extract_csp_name_and_args(io(Args,ChName,_SPAN),ChName,Args).
extract_csp_name_and_args(start_cspm(Name), Name, []).
extract_csp_name_and_args(start_cspm_MAIN,'start_cspm_MAIN',[]).
extract_csp_name_and_args(tick(_),tick,[]).
extract_csp_name_and_args(tau(_),tau,[]).
extract_csp_name_and_args(_OP,'?',[]).

/**
Extracts information for a B operation.
*/
:- mode extract_b_op(+OpTerm, +Truncate, -Params, -RetVals).
extract_b_op(OpTerm, Truncate, Params, RetVals) :-
    extract_b_op_infos(OpTerm, _Name, PSource, RSource),
    maplist(pretty_print_formula(Truncate), PSource, Params),
    maplist(pretty_print_formula(Truncate), RSource, RetVals).

:- mode extract_b_op_infos(+Term, -Name, -Arguments, -RetVals).
extract_b_op_infos(Term, Name, Arguments, RetVals) :-
    (Term = '-->'(OpTerm,RetVals) -> true;
        OpTerm = Term, RetVals = []),
    OpTerm =.. [Name|Arguments].

extract_op_name(OpTerm,Name) :-
    (animation_mode(cspm) ->
        extract_csp_name_and_args(OpTerm, Name, _Args)
    ;   extract_b_op_infos(OpTerm, Name, _Params, _RetVals)
    ).

pretty_print_formula(truncate,Formula,Result) :-
    translate_bvalue(Formula, Result).
pretty_print_formula(expand,Formula,Result) :-
    translate_bvalue_with_limit(Formula,-1,Result).

/* ------------------------- */
/* Boolean Properties        */
/* ------------------------- */

is_initialised_state(ID) :-
    visited_expression(ID,State), ID \= root,
    specfile:state_corresponds_to_initialised_b_machine(State).


invariantKO(StateID) :-  is_initialised_state(StateID),
    get_state(StateID,State),
    user:check_invariantKO(StateID,State),
    state_space:invariant_violated(StateID).

/**
Takes a state id and finds all of the operations for which a timeout occurred

#### called by:
* ProB Plugin: GetTimeoutedOperationsCommand
* ProB 2.0: GetOperationsWithTimeout
*/
op_timeout_occurred(StateID,OpNameList) :-
    findall(OpName,time_out_for_node(StateID,OpName,_Type),OpNameList).

inv_timeout_occurred(StateID) :- time_out_for_invariant(StateID).
timeout_occurred(StateID) :-  time_out_for_node(StateID).
max_operations_reached(StateID) :- max_reached_for_node(StateID).

/**
Finds the status for a given property
Properties can be: invariantKO, timeout_occurred, max_operations_reached, initialised
Statuses are expected to be boolean values: either true or false

#### called by:
* ProB Plugin: CheckBooleanPropertyCommand
* ProB 2.0: CheckBooleanPropertyCommand
*/
:- mode state_property(+Property,+StateId,-Status).
state_property(Property,StateId,Status) :-
    ( state_property2(Property,StateId) -> Status = true
    ; otherwise -> Status = false).
state_property2(invariantKO,StateId) :- invariantKO(StateId).
state_property2(timeout_occurred,StateId) :- inv_timeout_occurred(StateId) ; timeout_occurred(StateId).
state_property2(max_operations_reached,StateId) :- max_operations_reached(StateId).
state_property2(valid_state,StateId) :- visited_expression(StateId, _State).
state_property2(initialised,StateId) :-
    (specfile:csp_mode ->
        true % in CSP the content of the state is irrelevant for the semantic of the state space
      ; is_initialised_state(StateId)
    ).


/**
Takes a id for a given state in the state space and produces a list of all
state based errors.

#### called by:
* ProB Plugin: GetStateBasedErrorsCommand
* ProB 2.0:    GetStateBasedErrorsCommand
*/
:- mode get_state_errors(+StateId,-Errors).
get_state_errors(StateId,Errors) :-
  findall(E, (state_error(StateId,_,E),E \== invariant_violated), Errs),
  convert_errors(Errs,Errors).
convert_errors([],[]).
convert_errors([Error|InRest],[error(Event,Short,Long)|OutRest]) :-
  ( Error = eventerror(Event,EError,Trace) ->
      translate_event_error(EError,Short),
      explain_event_trace(Trace,LongStr),
      safe_atom_codes(Long,LongStr)
  ; otherwise ->
      Event = '*unknown*',
      translate_state_error(Error,Short),
      explain_state_error(Error,_Span,LongStr),
      safe_atom_codes(Long,LongStr)),
  convert_errors(InRest,OutRest).

/* ------------------------- */
/* Evaluate Formulas         */
/* ------------------------- */

% term in string (see String argument) should be terminated by full-stop!!!
evaluate_csp_expression_string(PlClause,R) :-
    read_from_codes(PlClause,PlTerm),
    evaluate_csp_expression_aux(PlTerm,R).

evaluate_csp_expression_aux(PlTerm,R) :-
    haskell_csp:evaluate_expression(PlTerm,Res),
    translate:translate_cspm_expression(Res, PrettyPrintedResult),
    ajoin(PrettyPrintedResult,R).
evaluate_csp_expression_aux(Other,_R) :-
    add_error_and_fail(prob2_interface, 'Unexpected CSP Expression: ', Other).

/**
Evaluate a formula given a state id and the raw prolog representation of the formula

#### Params:
* StateId - state for which the formula should be evaluated
* Element - Triple in the form eval(Formula,Type,Code)
    * Formula - Prolog representation of non-typechecked formula
    * Type - csp,'#PREDICATE',or '#EXPRESSION'
    * Code - String representation of Formula
    * Truncate - if the result should be truncated or expanded
* Result - Triple result(Value,Solutions,Code)
    * Value - String rep of the value calculated by ProB or tuple: v(SRes,Term)
        * SRes - String rep of the value calculated by ProB
        * Term - Prolog representation of the value
    * Solutions - List of solutions as a triple: bind(Name,Solution,PPSol)
        * Name - String free variable from formula
        * Solution - Prolog representation of the solution bound to the name
        * PPSol - String representation of the solution bound to the name
    * Code - String representation of evaluated formula

#### called by:
ProB 2.0: EvaluateFormulaCommand
*/
:- mode evaluate_formula(+StateId,+Element,-R).
evaluate_formula(_StateId, eval(Raw,csp,Code,_Truncate), R) :- !,
    ( evaluate_csp_expression_string(Raw,Res) ->
        R = result(Res,[]);
        R = [Code,"Unexpected CSP Expression"]).

evaluate_formula(StateId, eval(Raw,PredOrExpr,Code,Truncate), R) :-
    evaluate_formula_typecheck(PredOrExpr,Raw,Typed,Type,Errors),
    ( Errors=[] ->
        evaluate_typechecked_b_formula(StateId,Typed,Type,Truncate,R)
    ; R = [Code|Errors]).

/**
Type checks a given formula.

#### Params:
* PredOrExpr - '#PREDICATE' or '#EXPRESSION'
* RawFormula - Prolog representation of non-typechecked formula
* TypedFormula - Prolog representation of typechecked formula
* Type - type of formula
* Errors - any errors that have occured during typechecking

#### called by:
ProB 2.0: EvaluateFormulaCommand
*/
:- mode evaluate_formula(+PredOrExpr,+RawFormula,-TypedFormula,-Type,-Errors).
evaluate_formula_typecheck('#PREDICATE',Raw,Typed,Type,Errors) :-
    !,b_type_open_predicate(open(exists),Raw,[prob_ids(visible),variables],Typed,Errors),
    Type=pred.
evaluate_formula_typecheck('#EXPRESSION',Raw,Typed,Type,Errors) :-
    !,b_type_expression(Raw,[prob_ids(visible),variables],Type,Typed,Errors).
evaluate_formula_typecheck(Else,_Raw,_Typed,_Type,_Errors) :-
    add_error_and_fail(prob2_interface,'expected #PREDICATE or #EXPRESSION', Else),fail.

evaluate_typechecked_b_formula(StateId,Typed,Type,Truncate,Result) :-
    get_state(StateId, State1),
    add_prob_deferred_set_elements_to_store(State1, State, visible),
    evaluate_formula_eval(Type,State,Typed,Res,Solution,EnumWarning),
    get_error_messages(_WarningsOnly,Errors),
    extract_result(EnumWarning,Errors,Res,Solution,Truncate,Result).

extract_result(true,_,_,_,_,enum_warning).
extract_result(_EnumW,[],v(_SRes,Term),[],Truncate,Result) :-
    pretty_print_formula(Truncate,Term,R),
    Result = result(R,[]).
extract_result(_EnumW,[],SRes,Solutions,Truncate,Result) :-
    prettyprint_solutions(Truncate,Solutions,PPSol),
    Result = result(SRes,PPSol).
extract_result(_EnumW,Errors,v(SRes,_),_Sol,_Truncate,Result) :-
    Result = errors(SRes, Errors).
extract_result(_EnumW,Errors,SRes,_Sol,_Truncate,Result) :-
    Result = errors(SRes, Errors).

evaluate_formula_eval(pred,State,Typed,Res,LocalState,EnumWarning) :-
    !,catch_enumeration_warning_exceptions(
         eval_strings:eval_predicate(State,Typed,Res,EnumWarning,LocalState),
         (Res = time_out,EnumWarning = true)).

evaluate_formula_eval(_Type,State,Typed,Res,[],EnumWarning) :-
    catch_enumeration_warning_exceptions(
        eval_strings:eval_expression(State,Typed,SRes,EnumWarning,Term),
        (SRes = time_out, EnumWarning = true)),
    Res = v(SRes,Term).

:- dynamic prob2_formula/4.

/**
Type checks a specified formula and asserts so that it can be found later
and evaluated later.

#### called by:
* ProB 2.0: RegisterFormulaCommand
*/
:- mode register_prob2_formula(+FormulaUUID, +Formula).
register_prob2_formula(FormulaUUID, _) :- prob2_formula(FormulaUUID,_,_,_),!.
register_prob2_formula(FormulaUUID, eval(Raw,PredOrExpr,Truncate)) :-
    evaluate_formula_typecheck(PredOrExpr,Raw,Typed,Type,Errors),
    ( Errors=[] ->
        register_prob2_formula2(FormulaUUID, Typed, Type, Truncate)
    ; otherwise ->
        add_error_and_fail(register_prob2_formula, 'Type-Errors: ', Errors)
    ).

register_prob2_formula2(FUUID, Typed, Type, Truncate) :-
    assert(prob2_formula(FUUID, Typed, Type, Truncate)).

/**
Takes a state id and a list of formula ids (that were registered already)
via register_prob2_formula predicate. Then it evaluates all of the formulas
for the given state id.

#### called by:
* ProB 2.0: EvaluateRegisteredFormulasCommand
*/
:- mode evaluate_registered_formulas(+StateId, +FormIds, -Results).
evaluate_registered_formulas(StateId,FormIds,Results) :-
    maplist(evaluate_registered_formula(StateId), FormIds, Results).

evaluate_registered_formula(StateId,FormulaUUID,Result) :-
    prob2_formula(FormulaUUID,Typed,Type,Truncate),
    evaluate_typechecked_b_formula(StateId,Typed,Type,Truncate,Result).


/**
Takes a predicate and finds a list of all state ids for which the
predicate holds. The states that are not intitialized (i.e. root) are
included in the list. The list that is returned is therefore the union
of the uninitialised states and the states for which the predicate holds.

#### called by:
* ProB 2.0: GetStatesFromPredicate
*/
:- mode get_states_for_predicate(+Raw,-States,-Errors).
get_states_for_predicate(Raw,States,Errors) :-
      evaluate_formula_typecheck('#PREDICATE',Raw,Typed,_,Errors),
      findall(StateId,get_state_for_predicate(StateId,Typed),States).

get_state_for_predicate(StateId,Typed) :-
      state_space:visited_expression(StateId,StatePacked),
      ( is_initialised_state(StateId) ->
          (expand_const_and_vars_to_full_store(StatePacked,State),
          eval_strings:eval_predicate(State,Typed,'TRUE',_,_))
          ;
          true
      ).

/**
This cycles through all of the solutions and extracts the string
representation of the solutions.
*/
prettyprint_solutions(_Truncate,[],[]).
prettyprint_solutions(Truncate,[bind(Name,Res)|T],[PP|R]) :-
    pretty_print_formula(Truncate,Res,PPRes),
    PP = solution(Name,PPRes),
    prettyprint_solutions(Truncate,T,R).

/**
<<<<<<< HEAD
Allows Strings to be parsed and evaluated. Used for EvalB

#### called by:
* ProB 2.0: RemoteEvaluateCommand
*/
evalb_evaluate_formula(Codes, result(R,Solutions,Quantor,Vars,Warn)):-
    eval_strings:eval_codes(Codes,exists,R,Warn,LocalState,TypeInfo),
    extract_typeinfo(TypeInfo,Quantor,Vars),
    prettyprint_solutions(truncate,LocalState,Solutions).

/**
Allows Strings to be parsed and evaluated. Used for EvalB

#### called by:
* ProB 2.0: RemoteEvaluateCommand
*/
evalb_evaluate_tautology(Codes, result(R,Solutions,Quantor,Vars,Warn)):-
    eval_strings:eval_codes(Codes,forall,R,Warn,LocalState,TypeInfo),
    extract_typeinfo(TypeInfo,Quantor,Vars),
    prettyprint_solutions(truncate,LocalState,Solutions).

extract_typeinfo(expression(_),expression,[]).
extract_typeinfo(predicate(forall(BL)),forall,L) :- translate_list(BL,L).
extract_typeinfo(predicate(exists(BL)),exists,L) :- translate_list(BL,L).
extract_typeinfo(predicate(no_outer_quantifier),predicate,[]).

translate_list([],[]).
translate_list([H|T],[HT|TT]) :- translate_bexpression(H,HT), translate_list(T,TT).

/**
Takes a list of state ids and a predicate and finds all of the states
for which the predicate is true

#### called by:
* ProB 2.0: FilterStatesForPredicateCommand
*/
:- mode filter_states_for_predicate(+Raw,+States,-Filtered).
filter_states_for_predicate(Raw,States,Filtered) :-
      evaluate_formula_typecheck('#PREDICATE',Raw,Typed,_,Errors),
      (Errors=[] ->
        filter_states_for_typed_predicate(Typed,States,Filtered)
        ;
        Filtered = errors(Errors)
      ).

filter_states_for_typed_predicate(_,[],[]).
filter_states_for_typed_predicate(Typed,[Id|States],Filtered) :-
      get_state(Id,State1),
      add_prob_deferred_set_elements_to_store(State1,State,visible),
      ( is_initialised_state(Id) ->
          (eval_strings:eval_predicate(State,Typed,'TRUE',_,_) ->
              Filtered = [Id|Rest] ;
              Filtered = Rest
          ) ;
          Filtered = [Id|Rest]
      ),
      filter_states_for_typed_predicate(Typed,States,Rest).

/* ------------------------- */
/* Formula Expansion         */
/* ------------------------- */

/**
Inserts a formula into bvisual2 module
The formula is inserted as a child of level "user" in bvisual2

#### called by:
* ProB 2.0: InsertFormulaForVisualizationCommand
*/
:- mode insert_formula_for_expansion(+AST,-Id).
insert_formula_for_expansion(AST,Id) :-
    suppress_rodin_positions,
    b_type_expression(AST,[variables],_,Typed,Errors),
    ( Errors == [] ->
        bv_insert_formula(Typed,user,Id)
    ; otherwise ->
        add_error_and_fail(eclipse_interface,'Could not type-check AST','')).

/**
Uses the bvisual2 module to recursively expand and evaluate a specified formula.
This formula must first be inserted into the bvisual2 module using the predicate
insert_formula_for_expansion. The formula is then identified by the Id specified
at the time of insertion.

#### called by:
* ProB 2.0: ExpandFormulaCommand
*/
:- mode expand_formula(+Id,+StateId,-Tree).
expand_formula(Id,StateId,Tree) :-
    bv_expand_formula(Id,Label,Children),
    bv_get_values([Id],StateId,[Value]),
    bv_get_values(Children,StateId,ChildrenValues),
    find_all_subformulas(StateId,Children,ChildrenValues,AsTrees),
    Tree = formula(Label,Value,Id,AsTrees).

find_all_subformulas(_,[],_,[]).
find_all_subformulas(StateId,[Id|RestI],[Value1|RestV],[formula(Label,Value1,Id,Trees)|RestT]) :-
    bv_expand_formula(Id,Label,Children),
    bv_get_values(Children,StateId,CValues),
    find_all_subformulas(StateId,Children,CValues,Trees),
    find_all_subformulas(StateId,RestI,RestV,RestT).


/* ------------------------- */
/* Model Checking            */
/* ------------------------- */

/**
#### do_modelchecking(+Time,+Options,-Result,stats(-NrNodes,-NrTrans,-NrProcessed))
* +Time : Timeout specified by the user in ms
* +Options : List of options specified by the user. Used for predicate do_modelchecking(Time,Options,Result)
* -NrNodes : total number of nodes in state space. Calculated with get_state_space_stats
* -NrTrans : total number of nodes in state space. Calculated with get_state_space_stats
* -NrProcessed : total number of calculated nodes in state space. Calculated with get_state_space_stats
 When Time Milliseconds have elapsed the modelchecker should stop after its next step

#### called by:
* ProB Plugin: ModelCheckingCommand
* ProB 2.0: ModelCheckingStepCommand
*/
:- mode do_modelchecking(+Time,+Options,-Result,-Stats).
do_modelchecking(Time, Options, Result, stats(NrNodes,NrTrans,NrProcessed)) :-
    statistics(walltime, [CurTime,_]), /* get current time in ms */
    LimitTime is CurTime+Time,
    option_set(find_deadlocks, Options, Deadlock),
    option_set(find_invariant_violations, Options, Invariant),
    option_set(find_assertion_violations, Options, Assertions),
    option_set(inspect_existing_nodes, Options, InspectExistingNodes),
    option_set(stop_at_full_coverage, Options, StopAtFullCoverage),
    (option_set(breadth_first_search, Options, 1)
      -> set_depth_breadth_first_mode(breadth_first)
      ;  set_depth_breadth_first_mode(mixed)),
    option_set(find_goal, Options, Goal),
    option_set(partial_guard_evaluation, Options, PartialGuardEvaluation),
    (option_set(partial_order_reduction, Options, 1)
      -> WithPOR = ample_sets
      ;  WithPOR = off),
    % no partial guards evaluation
    MaxNumberOfStatesToCheck=100000,
    user:do_model_check(MaxNumberOfStatesToCheck, _, LimitTime, Res, Deadlock, Invariant, Goal,Assertions,
                        StopAtFullCoverage, WithPOR, PartialGuardEvaluation, InspectExistingNodes),
    build_modelcheck_return(Res, Result),
    get_state_space_stats(NrNodes, NrTrans, NrProcessed).

/**
Sets the goal for model checking. The Option search_for_goal needs to be set when starting
model checking in order for this to have an effect on the model checking.

#### called by:
* ProB 2.0: SetBGoalCommand
*/
:- mode set_goal_for_model_checking(+Goal).
set_goal_for_model_checking(Goal) :-
    evaluate_formula_typecheck('#PREDICATE',Goal,TypedGoal,_Type,Errors),
    (Errors = [] ->
        b_set_parsed_typed_machine_goal(TypedGoal);
        add_error_and_fail(set_goal_for_model_checking,typeerror,Errors)
      ).

/**
Takes an atom and unifies Result with 1 if the atom is in the List.
Otherwise, Result is unified with 0.
*/
:- mode option_set(+Element, +List, -Result).
option_set(Element, List, Result) :-
    ( member(Element,List)
        -> Result = 1
        ;  Result = 0).

:- mode build_modelcheck_return(+MCRes, -JavaResult).
build_modelcheck_return(MCRes, JavaResult) :-
    build_modelcheck_return2(MCRes, JavaResult) -> true
    ; nl, print('### Unknown Model Check Error Result:'), print(MCRes), nl, /* TODO: Why do we print something here? Can we get rid of this print? (Q? from Joy)*/
        current_state_id(State), JavaResult=general_error(State, MCRes).

build_modelcheck_return2(no, not_yet_finished(100000)).
build_modelcheck_return2([timeout,N], not_yet_finished(N1)) :- N1 is 100000 - N.
build_modelcheck_return2(deadlock, deadlock(State)):- current_state_id(State).
build_modelcheck_return2(invariant_violation, invariant_violation(State)):- current_state_id(State).
build_modelcheck_return2(assertion_violation, assertion_violation(State)):- current_state_id(State).
build_modelcheck_return2(state_error, state_error(State)):- current_state_id(State).
build_modelcheck_return2(state_error(_), state_error(State)):- current_state_id(State).
build_modelcheck_return2(goal_found, goal_found(State)) :- current_state_id(State).
build_modelcheck_return2(well_definedness_error, well_definedness_error(State)) :- current_state_id(State).
build_modelcheck_return2(general_error_occurred, general_error(State)):- current_state_id(State).
build_modelcheck_return2(full_coverage, full_coverage).
build_modelcheck_return2(all, ok_not_all_nodes_considered) :- max_reached_or_timeout_for_node(_),!.
build_modelcheck_return2(all, ok).

/**
Computes the coverage statistics of the current state space at any given time.
The information of interest includes the total number of nodes and transitions, as well as
a list of statistics about the nodes and operations and a list of the operations that have been uncovered sofar.

#### called by:
* ProB Plugin: ComputeCoverageCommand
* ProB 2.0: ComputeCoverageCommand
*/
:- use_module(coverage_statistics, [compute_the_coverage/5, operation_hit/2,query_node_hit/2, uncovered_operation/1]).
:- mode compute_coverage(-TotalNodeNr,-TotalTransSum,-NodeStat,-OpStat,-Uncovered).
compute_coverage(TotalNodeNr,TotalTransSum,NodeStat,OpStat,Uncovered)  :-
    compute_the_coverage(_,TotalNodeNr,TotalTransSum,false,false),
    findall(S2,(coverage_statistics:operation_hit(OpS,Nr),string_concatenate(':',Nr,S1),string_concatenate(OpS,S1,S2)),OpStat),
    findall(S2,(coverage_statistics:query_node_hit(Prop,Nr),string_concatenate(':',Nr,S1),string_concatenate(Prop,S1,S2)),NodeStat),
    findall(OpName, coverage_statistics:uncovered_operation(OpName),Uncovered).

/**
Performs deadlock freedom checking with constraint Predicate and calulates the Result.

#### called by:
* ProB 2.0: ConstraintBasedDeadlockCheckCommand
*/
:- mode prob2_deadlock_freedom_check(+Predicate,-Result).
prob2_deadlock_freedom_check(Predicate,Result) :-
    b_type_expression(Predicate,[variables],pred,TPredicate,Errors),
    ( Errors == [] ->
        prob2_deadlock_freedom_check1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
prob2_deadlock_freedom_check1(Predicate,Result) :-
    % always do a deadlock check with SMT mode enabled
    call_with_smt_mode_enabled(prob2_deadlock_freedom_check2(Predicate,Result)).
prob2_deadlock_freedom_check2(Predicate,Result) :-
    user_interruptable_call_det(user:catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,Predicate,0)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = deadlock(Transition,StateId),
        add_artificial_transition(root,deadlock_check,State,StateId,Transition)).
prob2_deadlock_freedom_check2(_Predicate,no_deadlock_found).

/**
Performs invariant checking for either for all operations or a list of operations.

#### called by:
* ProB 2.0: ConstraintBasedInvariantCheckCommand
*/
:- mode prob2_invariant_check(+Ops,-Result).
prob2_invariant_check(all,Result) :-
    findall(OpName,b_get_machine_operation(OpName,_,_,_),Ops),
    prob2_invariant_check2(Ops,Result).
prob2_invariant_check(ops(Ops),Result) :-
    prob2_invariant_check2(Ops,Result).
prob2_invariant_check2(Ops,Result) :-
    call_with_smt_mode_enabled(prob2_invariant_check3(Ops,Result,[])).
prob2_invariant_check3([]) --> !.
prob2_invariant_check3([Op|Rest]) -->
    prob2_invariant_check_for_single_op(Op),
    prob2_invariant_check3(Rest).
prob2_invariant_check_for_single_op(OpName,In,Out) :-
    ( clpfd_interface:catch_clpfd_overflow_call1(b_state_model_check:state_model_check(OpName,State1,Operation,State2)) ->
        In = [counterexample(OpName,Trans1,Trans2)|Out],
        atom_concat( invariant_check_ , OpName, RootTrans),
        add_artificial_transition(root,    RootTrans,State1,StateId1,Trans1),
        add_artificial_transition(StateId1,Operation,State2,_StateId2,Trans2)
    ; otherwise ->
        In = Out).

/**
creates a helper transition that is artificially added to the state space (e.g. during deadlock checking)
This transition is added to the state space.
A triple TransitionTuple in the form op(TransId,OpName,SrcId,DstId) for this transition is generated.
*/
:- mode add_artificial_transition(+SrcId,+Operation,+DstState,+DstId,-TransitionTuple).
add_artificial_transition(SrcId,Operation,DstState,DstId,TransitionTuple) :-
    user:tcltk_add_new_transition_transid(SrcId,Operation,DstId,DstState,[],TransId),
    extract_op_name(Operation,OpName),
    TransitionTuple = op(TransId,OpName,SrcId,DstId).


/**
Performs an LTL model checking step.

#### called by:
* ProB 2.0: LtlCheckingCommand
*/
:- mode prob2_do_ltl_modelcheck(+Formula,+Max,-Result,-Errors).
prob2_do_ltl_modelcheck(Formula,Max,Result,Errors) :-
   typecheck_temporal_formula(Formula,TypeCheckedFormula,Status),
   ( Status=ok ->
       ltl_model_check_with_ce(TypeCheckedFormula,Max,init,Result1),
       prob2_ltl_adapt_operations(Result1,Result),
       get_error_messages(_WarningsOnly,Errors)
   ; otherwise ->
       Result=typeerror,
       get_error_messages(_WarningsOnly,Errors)).

prob2_ltl_adapt_operations(counterexample(CE1,LoopEntry,PathToCE1), counterexample(CE,LoopEntry,PathToCE)) :- !,
    create_simple_op_terms(PathToCE1,root,PathToCE),
    prob2_ltl_adapt_ce(CE1,CE).
prob2_ltl_adapt_operations(Result,Result).

prob2_ltl_adapt_ce([],[]).
prob2_ltl_adapt_ce([atom(StateId,_,OpTuple)|Irest],[Transition|Orest]) :-
    prob2_ltl_adapt_ce2(OpTuple,StateId,Transition),
    prob2_ltl_adapt_ce(Irest,Orest).
prob2_ltl_adapt_ce2(none,_StateId,none).
prob2_ltl_adapt_ce2((TransId,Action,DestId),StateId,op(TransId,Name,StateId,DestId)) :-
    extract_op_name(Action,Name).

/* ------------------------- */
/* Plugin Handling           */
/* ------------------------- */

/**
Activates a plugin with the given ID

#### called by:
* ProB Plugin: ActivateUnitPluginCommand
* ProB 2.0: ActivateUnitPluginCommand
*/
:- mode activate_plugin(+ID).
activate_plugin(ID) :-
    start_plugin_for_current_mode(ID).

/**
Gets the output for a given plugin

#### called by:
* ProB Plugin: GetPluginResultCommand
* ProB 2.0: GetPluginResultCommand
*/
:- mode get_plugin_output(+OutputID,-Output).
get_plugin_output(OutputID,Output) :-
    plugin_get_output(OutputID,Output).


/* ------------------------- */
/* Find Traces               */
/* ------------------------- */

/**
Finds a trace from the root state to the specified state in the current state space.

#### Parameters:
* StateId
* Trace - List of op tuples op(OpId,SrcId,DestId) corresponding to the trace calculated by ProB or atom no_trace_found if the call was unsuccessful

#### called by:
* ProB 2.0: GetShortestTraceCommand
*/
:- mode find_trace_to_node(+StateId,-Trace).
find_trace_to_node(StateId, Trace) :-
    find_trace_from_node_to_node(root, StateId, Trace).

/**
Finds a trace from one state to a goal state in the current state space.

#### Parameters:
* FromId - Id of source node
* ToId - Id of destination node
* Trace - List of op tuples op(OpId,SrcId,DestId) corresponding to the trace calculated by ProB or atom no_trace_found if the call was unsuccessful

#### called by:
* ProB 2.0: GetShortestTraceCommand
*/
:- mode find_trace_from_node_to_node(+FromId,+ToId,-Trace).
find_trace_from_node_to_node(FromId, ToId, Trace) :-
    (user:find_shortest_trace_to_node(FromId, ToId, T, _TraceIDs) ->
        trace_to_op_triple(T,Trace);
        Trace = no_trace_found
    ).

/**
Translates a list of transition ids to a list of op tuples op(TransId,OpName,SrcId,DestId)
*/
:- mode trace_to_op_triple(+ListIds,-ListTuples).
trace_to_op_triple([],[]).
trace_to_op_triple([Op|T], [op(Op,Name,Src,Dest)|OpT]) :-
    transition(Src,Action,Op,Dest),
    extract_op_name(Action,Name),
    trace_to_op_triple(T,OpT).

/**
Takes a given predicate and finds a state in the state space that satisfies the predicate.
A helper transition is then added to go to the goal state.

#### called by:
* ProB 2.0: FindValidStateCommand
*/
find_state_for_predicate(Predicate,Result) :-
    evaluate_formula_typecheck('#PREDICATE',Predicate,TPredicate,_Type,Errors),
    ( Errors == [] ->
        find_state_for_predicate1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
find_state_for_predicate1(Predicate,Result) :-
    user_interruptable_call_det(user:catch_clpfd_overflow_call1(b_set_up_valid_state_with_pred(State,Predicate)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = state_found(Transition,StateId),
        add_artificial_transition(root,find_valid_state,State,StateId,Transition)).
find_state_for_predicate1(_Predicate,no_valid_state_found).

/* ------------------------------- */
/* CBC Disproving & solving        */
/* ------------------------------- */

% Takes an untyped Predicate AST and returns the type checked AST
% be sure to getErrors in case it fails
cbc_type_check_predicate(ParsedPredicate,TypedPredicate) :-
  type_check_in_machine_context([ParsedPredicate],[TypedPredicate]).
  % an alternative would be:
  % TO DO: check if the scope is correct: [prob_ids(visible),variables]; maybe we should remove variables
  %b_type_open_predicate_with_errors(exists,ParsedPredicate,[prob_ids(visible),variables],TypedPredicate).

/**
#### called by:
* ProB Plugin (de.prob.eventb.disprover.core): DisproverCommand
*/
:- mode cbc_disprove(+Goal,+AllHypotheses,+SelectedHypotheses,+TimeoutFactor,-OutResult).
cbc_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,OutResult) :-
    disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,OutResult),
    % remove the warning regarding double check, in rodin it is shown in the proof tree anyway
    (get_error(warning(disprover_inconsistent_hypotheses),_) ; true).

:- mode cbc_disprove(+Goal,+AllHypotheses,+SelectedHypotheses,+TimeoutFactor,-OutResult).
cbc_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,Options,OutResult) :-
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,Options,OutResult),
    % remove the warning regarding double check, in rodin it is shown in the proof tree anyway
    (get_error(warning(disprover_inconsistent_hypotheses),_) ; true).

/**
Takes a list of Predicates and produces a result from the CBC solver
#### called by:
* ProB 2.0: CbcSolveCommand
*/
:- mode cbc_solve(+Predicates,+Identifiers,-Result).
cbc_solve(Predicates,Identifiers,Result) :-
    cbc_solve(Predicates,_State,Identifiers,Result).
cbc_solve(Predicates,State,Identifiers,Result) :-
    cbc_solve(Predicates,State,1,Identifiers,Result).
cbc_solve(Predicates,State,TimeoutFactor,Identifiers,Result) :-
    type_check_in_machine_context(Predicates,TPredicates),
    cbc_solve_typed(TPredicates,State,TimeoutFactor,Identifiers,Result).

cbc_solve_typed(TPredicates,State,TimeoutFactor,Identifiers,Result) :-
    conjunct_predicates(TPredicates,Predicate),
    find_identifier_uses(Predicate,[],Identifiers),
    solve_predicate(Predicate,State,TimeoutFactor,Result).

:- use_module(smt_solvers_interface(smt_solvers_interface),[smt_solve_predicate/2]).
smt_solve_typed(TPredicates,_TimeoutFactor,Result) :-
    conjunct_predicates(TPredicates,Predicate),
    smt_solve_predicate(Predicate,Result).

% Constraint-Based Test-case generation
% TO DO: return values
:- use_module(sap,[cbc_gen_test_cases/5]).
cbc_generate_test_cases(TargetPred,MaxDepth,OutputFile) :-
   % I am not sure whether we should call b_parse_machine_predicate(TargetPred,...) or not
    Events = all,
    cbc_gen_test_cases(Events,TargetPred,MaxDepth,OutputFile,_Uncovered).
/* ------------------------- */
/* Check CSP Assertion       */
/* ------------------------- */

/**
Takes a list of assertions and produce a list of results and result traces.

TODO: We should modify the result traces so that they are useful for ProB 2.0. (or at least document what the result traces mean)

#### called by:
* ProB 2.0: CSPAssertionsCommand
*/
:- mode check_csp_assertions(+Assertions,-Results,-ResultTraces).
check_csp_assertions(Assertions,Results,ResultTraces) :-
    maplist(check_csp_assertion,Assertions,Results,ResultTraces).

check_csp_assertion(AssClause,Res,ResTrace1) :-
    read_from_codes(AssClause,Assertion),
    user:checkAssertion(Assertion,_PP,_Negated,Res,ResTrace),
    (ResTrace = no_counter_example -> ResTrace1 = []; ResTrace1=ResTrace).

/* ------------------------- */
/* Preferences Interface     */
/* ------------------------- */

/**
Returns a list of all the preferences with their current values

#### called by:
* ProB 2.0: GetCurrentPreferencesCommand
*/
:- mode list_current_eclipse_preferences(-L).
list_current_eclipse_preferences(L) :-
    findall(preference(A,B),find_eclipse_preference(A,B),L).

find_eclipse_preference(A,B) :-
   list_eclipse_preferences(X),
   member(preference(A,_,_,_,_),X),
   get_eclipse_preference(A,B).

/**
Returns the current value of a specified preference.

#### called by:
* ProB 2.0: GetPreferenceCommand
*/
get_eclipse_preference(PrefS,PrefVal) :-
    eclipse_preference(PrefS,PS),
    get_preference(PS,PrefVal).

/**
Returns a list of all normal eclipse preferences as well as their information
(i.e. type, description, category, and default value)

#### called by:
* ProB Plugin: GetPreferencesCommand
* ProB 2.0: GetDefaultPreferencesCommand
*/
:- mode list_eclipse_preferences(-L).
list_eclipse_preferences(L) :-
    findall(preference(A,B,C,D,E),is_normal_eclipse_preference(A,B,C,D,E),L).

is_normal_eclipse_preference(PrefString,Type,Description,Category,DefaultValue) :-
    eclipse_preference(PrefString,PS),
    \+ advanced_eclipse_preference(PrefString,PS),
    preference_val_type(PS,Type),
    preference_description(PS,Description),
    preference_category(PS,Category),
    preference_default_value(PS,DefaultValue).

/**
Sets a preference

#### called by:
* ProB Plugin: SetPreferenceCommand
* ProB 2.0: SetPreferenceCommand
*/
:- mode set_eclipse_preference(+PrefS,+PrefVal).
set_eclipse_preference(PrefS,PrefVal) :-
    convert_cli_arg(PrefVal,Value),
    convert_pref_value(Value,CValue),
    (eclipse_preference(PrefS,P) -> true
      ; preferences:obsolete_eclipse_preference(PrefS)
        -> add_error_and_fail(eclipse_preference,'Obsolete Eclipse Preference: ',PrefS)
        ; add_error_and_fail(eclipse_preference,'Unknown Eclipse Preference: ',PrefS)
    ),
    (set_preference(P,CValue) -> true ; add_error(eclipse_preference,'Could not set preference: ',Value)).

convert_cli_arg(PrefVal,Value) :-
    (atomic(PrefVal) ->
        (name(PrefVal,ASCIIPREF), name(Value,ASCIIPREF))  /* convert '12' to 12 */
      ;  Value=PrefVal
    ).

convert_pref_value('TRUE',V) :- !, V=true.
convert_pref_value('FALSE',V) :- !, V=false.
convert_pref_value(X,X).

/* ------------------------- */
/* Apply Graph reduction     */
/* ------------------------- */

/**
Takes a list of ignored events, and applies the signature merge algorithm
from module `state_space_reduction` to the current state space.

#### called by:
* ProB 2.0: ApplySignatureMergeCommand
*/
:- mode get_signature_merge_state_space(+IgnoredEvents,-Space).
get_signature_merge_state_space(IgnoredEvents,Space) :-
    reset_ignored_events,
    set_ignored_events(IgnoredEvents),
    compute_signature_merge,
    findall(node(NodeId,Count,Color,Labels),extract_node_info(NodeId,Count,Color,simple_list,Labels),Nodes),
    findall(trans(TransId,Src,Dest,Label,Style,Color),extract_trans_info(true,TransId,Src,Dest,Label,Style,Color),Trans),
    Space = [Nodes,Trans].

/**
Takes a list of ignored events, and calculates a transition diagram
using module `state_space_reduction` for the current state space.

#### called by:
* ProB 2.0: CalculateTransitionDiagramCommand
*/
:- mode get_transition_diagram(+ParsedExpr,-Space).
get_transition_diagram(ParsedExpr,Space) :-
    evaluate_formula_typecheck('#EXPRESSION',ParsedExpr,TypedExpr,_Type,Errors),
    (Errors = [] -> true; add_error_and_fail(get_transition_diagram,typeerror,Errors)),
    compute_transition_diagram(TypedExpr),
    findall(node(NodeId,Count,Color,Labels),extract_node_info(NodeId,Count,Color,gen_label,Labels),Nodes),
    findall(trans(TransId,Src,Dest,Label,Style,Color),extract_trans_info(false,TransId,Src,Dest,Label,Style,Color),Trans),
    Space = [Nodes,Trans].

write_dotty_transition_diagram(Expression,Filename) :-
    write_transition_diagram_for_expr_to_dotfile(Expression,Filename).

write_dotty_signature_merge(IgnoredEvents,Filename) :-
    write_signature_merge_to_dotfile(IgnoredEvents,Filename).

:- use_module(visualize_graph,[tcltk_print_states_for_dot/1]).
write_dotty_state_space(Filename) :-
    tcltk_print_states_for_dot(Filename).
/**
Writes a dot representation of the given state to the specified file.

*/
:- mode write_dot_for_state_viz(+StateId, +Filename).
write_dot_for_state_viz(StateId, Filename) :-
    get_state(StateId, State),
    print_cstate_graph(State, Filename).

/**
Generates information about the nodes found during state space reduction

#### Generated Information:
* Count - number of concrete states combined in the abstract state
* Color - the color used to represent this state type in a visualization
* Labels - determine the labels that should appear on a node in a visualization
*/
:- mode extract_node_info(+NodeId,-Count,-Color,+LabelGenerator,-Labels).
extract_node_info(NodeId,Count,Color,LabelGenerator,Labels) :-
    get_reduced_node(AbsState,Count,Witness,NodeId),
    generate_node_color(NodeId,Witness,AbsState,Count,Color),
    generate_node_labels(LabelGenerator,AbsState,Labels).

/**
Generates information about the transitions created during state space reduction

#### Generated Information:
* Src - ID corresponding to the abstract state that is the source of the abstract transition with TransId
* Dest - ID corresponding to the abstract state that is the destination of the abstract transition with TransId
* Label - the label that should appear on a transition in a visualization
* Style - the style that should be applied to a transition of this type in a visualization (e.g. dashed)
* Color - the color used to represent a transition of this type in a visualization
*/
:- mode extract_trans_info(+ShowSelfLoops,+TransId,-Src,-Dest,-Label,-Style,-Color).
extract_trans_info(ShowSelfLoops,TransId,Src,Dest,Label,Style,Color) :-
    reduced_trans(Src,AbsAction,Count,Dest,TransId),
    generate_transition_label(AbsAction,Count,Label),
    generate_transition_color_and_style(ShowSelfLoops,Src,AbsAction,Dest,Color,Style).


/* ------------------ */
/* Get Errors         */
/* ------------------ */

/**
Returns a list of strings containing all error messages (of the error_manager).
Clears all the errors. Returns the empty list if no errors occurred since
the last call to get_error_messages.

#### called by:
* ProB 2.0: GetErrorsCommand
*/
:- mode get_error_messages(-WarningsOnly,-ListOfErrMsgs).
get_error_messages(WarningsOnly,ListOfErrMsgs) :-
    (real_error_occurred -> WarningsOnly = false ; WarningsOnly = true),
    ignore_user_interrupt_det((get_all_errors_and_reset(ListOfErrMsgs) ->  true  ; ListOfErrMsgs=[])).



/* ------------------------- */
/* Evalstore                 */
/* ------------------------- */

/**
TODO: Please document what this does.

#### called by:
* ProB 2.0: EvalstoreCreateByStateCommand
*/
es_copy_from_statespace(StateId,EvalstoreId) :-
    evalstore_create_store_by_state(StateId,EvalstoreId).

/**
TODO: Please document what this does.

#### called by:
* ProB 2.0: EvalstoreEvalCommand
*/
:- mode es_eval(+EvalstoreId,+UntypedFormula,+Timeout,-Result).
es_eval(EvalstoreId,UntypedFormula,Timeout,Result) :-
    evalstore_eval_formula(EvalstoreId,UntypedFormula,Timeout,Result1), !,
    (Result1 = ok(V,TypedIds,StoreId) ->
        Result = ok(V,PP,Ids,StoreId),
        get_texpr_ids(TypedIds,Ids),
        es_eval_translate(V,PP)
    ; otherwise                       -> Result = Result1).
es_eval_translate(pred(V),V).
es_eval_translate(value(V),PP) :-
    translate_bvalue(V,PP).

% ####################

/**

Animates randomly number of steps until a certain (LTL?) condition is fulfilled.

#### Called by:
* ProB 2.0: ExecuteUntilCommand

#### Arguments
* CurState - the ID of the current state
* Condition - a condition that should be satisfied at some point (this come as parsed term from the ProB2)

#### Generated Information:
* Trace - a list of triples representing a trace in the state space of the model being analysed
* Result - the result found: ltl_found, typeerror, max_nr_of_steps_reached, deadlock
*/
:- mode generate_trace_until_condition_fulfilled(+CurState,+Condition,-Trace,-Result).
generate_trace_until_condition_fulfilled(CurState,Condition,Trace,Result) :-
   typecheck_temporal_formula(Condition,TypeCheckedCondition,Status),
   ( Status=ok -> find_trace1(CurState,TypeCheckedCondition,no_loop,100000,Trace, Result) %FIXME: turn maximal numver of steps into an argument?
   ; otherwise -> get_error_messages(_WOnly,_Errors), % Reset errors. We don't want to throw a ProBError on the Java side because the result is capsuled in Result.
                  Result=typeerror, Trace=[]).

find_trace1(StateId,Ltl,Type,MaxSteps,OpTripleResultTrace, Result) :-
    (MaxSteps =< 0
       -> add_error(find_trace,'Number of maximum animation steps should be a positiv integer. The number of steps which was given is ',MaxSteps)
        ; true),
    set_current_state(StateId), !, % can be backtracked
    % negate -> counterexample is the trace we are looking for
    preprocess_formula(Ltl,Ltl2),
    find_trace_aux(StateId,not(Ltl2),Type,0,MaxSteps,ResultTrace,ResultTrace,Names,Result),
    gen_op_triples(ResultTrace,Names,OpTripleResultTrace).
    %(gen_op_triples(ResultTrace,OpTripleResultTrace) -> print(ok(OpTripleResultTrace,result(Result))),nl
    %  ; add_error_and_fail(prob2_interface,'trace not correctly generated',ResultTrace)).

find_trace_aux(CurID,Condition,Type,_N,_Max,StateTransitionHistory,[],[],ltl_found) :-
    debug_println(9,checking(CurID,StateTransitionHistory)),
    evaluate_ltl_formula(Condition,StateTransitionHistory,Type,ltl:check_ap,ltl:callback_tp,Result),
    Result = false,!.
find_trace_aux(CurID,Condition,Type,N,Max,StateTransitionHistory,STTail,Names,RESULT) :-
    user:tcltk_compute_options(CurID,ActionsAndIDs),
    debug_println(9,opts(ActionsAndIDs)),
    ( N=Max
       -> debug_println(9,'Maximum number of animation steps reached.'),
          Names = [],
          RESULT = maximum_nr_of_steps_reached
    ; pick_action(random,ActionsAndIDs,ActionId,Name,DstID) % pick first one; we could do random
       -> debug_println(9,performing_action(ActionId,from_to(CurID,DstID),opts(ActionsAndIDs))),
          STTail = [strans(CurID,ActionId)|STTail2],
          Names = [Name|Names2],
          N1 is N+1,
          find_trace_aux(DstID,Condition,Type,N1,Max,StateTransitionHistory,STTail2,Names2,RESULT)
    ; otherwise
       -> STTail=[], Names = [], RESULT = deadlock
     ).

gen_op_triples([],[],[]).
gen_op_triples([_X],[_Y],[]).
gen_op_triples([strans(CurID,ActionId),strans(DstID,ActId)|T],[Name|NameT],[op(ActionId,Name,CurID,DstID)|Rest]) :-
   gen_op_triples([strans(DstID,ActId)|T],NameT,Rest).

:- use_module(library(random)).
:- use_module(library(lists)).
pick_action(first,[(ActionId,Term,DstID)|_], ActionId, Name, DstID) :- extract_op_name(Term,Name).
pick_action(random,Options,ActionId, Name, DstID) :-
    length(Options,Len),
    L1 is Len+1,
    random(1,L1,RanChoice),
    debug_println(9,random(RanChoice,Len)),
    nth1(RanChoice,Options,(ActionId,ActionAsTerm,DstID)),
    extract_op_name(ActionAsTerm,Name).

% history_to_stransitions([],[]).
% history_to_stransitions([_],[]).
% history_to_stransitions([S1,S2|T],[strans(S1,Trans)|TransT]) :-
%     history_to_stransitions([S2|T],TransT),
%     any_transition(S1,Trans,S2).
%
% find_trace_aux(CurState,Ltl,Type,ResultTrace) :-
%     history(History),
%     print(history(History)),nl,
%     reverse([CurState|History],H1),
%     history_to_stransitions(H1,StateTransitions),
%     (is_deadlocked(CurState) ->
%         add_error(trace_animation,'DEADLOCK state occurred: ', CurState),
%         get_result_trace(StateTransitions,ResultTrace)
%     ; otherwise ->
%         evaluate_ltl_formula(Ltl,StateTransitions,Type,ltl:check_ap,ltl:callback_tp,Result),
%         Result = false,!,
%         get_result_trace(StateTransitions,ResultTrace)
%     ).
% find_trace_aux(_CurState,Ltl,Type,ResultTrace) :-
%     user:tcltk_random_perform,
%     user:current_state_id(CurID),
%     print(current_state(CurID)),nl,
%     user:tcltk_compute_options(CurID,TransSpecs),
%     user:extract_actions(TransSpecs,ActionsAndIDs,_Actions),
%     set_current_options(ActionsAndIDs),
%     find_trace_aux(CurID,Ltl,Type,ResultTrace).
%
% get_result_trace(StateTransitions,ResultTrace) :-
%     maplist(transid,StateTransitions,THistory),
%     trace_to_op_triple(THistory,ResultTrace).
%
% transid(strans(_StateId,TransId),TransId).

% alternative version which does not use Tcl/Tk history: ...
% use_module(probsrc(prob2_interface)),prob2_interface:find_trace2(root,ap(bpred(greater(none,card(none,identifier(none,active)),integer(none,0)))),Trace).
% use_module(probsrc(prob2_interface)),prob2_interface:find_trace2(root,ap(enabled(btrans(event(new)))),Trace).
% use_module(probsrc(prob2_interface)),prob2_interface:find_trace2(root,ap(enabled(btrans(event(del)))),Trace).

% find_trace2(StateId,Condition,OpTripleResultTrace) :-
%     set_current_state(StateId), !, % can be backtracked
%     % negate -> counterexample is the trace we are looking for
%     print(start_state(StateId)),nl,
%     find_trace_aux(StateId,Condition,ResultTrace,ResultTrace,RESULT),
%      % we do not initialise trace with [strans(CurID,...)] -> we execute at least one step !
%     print(result_trace(ResultTrace,RESULT)),nl,
%     % check RESULT = ltl_found to avoid also finding deadlocks
%     (gen_op_triples(ResultTrace,OpTripleResultTrace) -> print(ok(OpTripleResultTrace)),nl
%       ; add_internal_error('Call failed: ',gen_op_triples(ResultTrace,OpTripleResultTrace)),fail).
%
% % find_trace_aux(+CurrentStateID, History1,History2 (DifferenceList), -CheckingResult)
% find_trace_aux(CurID,Condition,StateTransitionHistory,[],ltl_found) :-
%     print(checking(CurID,StateTransitionHistory)),nl,
%     evaluate_ltl_formula(Condition,StateTransitionHistory,_,ltl:check_ap,ltl:callback_tp,Result),
%     print(result(Result)),nl,
%     Result = true, !.
% find_trace_aux(CurID,Condition,StateTransitionHistory,STTail,RESULT) :-
%     user:tcltk_compute_options(CurID,ActionsAndIDs), print(opts(ActionsAndIDs)),nl,
%     (pick_action(random,ActionsAndIDs,ActionId,DstID) % pick first one; we could do random
%       -> print(performing_action(ActionId,from_to(CurID,DstID),opts(ActionsAndIDs))),nl,
%          STTail = [strans(DstID,ActionId)|STTail2],
%          find_trace_aux(DstID,Condition,StateTransitionHistory,STTail2,RESULT)
%       ;  STTail=[], RESULT = deadlock,
%          format('Deadlock reached for state ~w (after ~w)~n',[CurID,StateTransitionHistory])
%      ).

get_unsat_core_with_fixed_conjuncts(Pred,FixedPreds,CoreOut) :-
    typecheck_pred_for_unsat_core(Pred,TypedPred),
    maplist(typecheck_pred_for_unsat_core,FixedPreds,TypedFixedPreds),
    conjunct_predicates(TypedFixedPreds,Conj),
    unsat_core_with_fixed_conjuncts(TypedPred,Conj,Core),
    translate_bexpression(Core,CoreOut).

get_minimum_unsat_core_with_fixed_conjuncts(Pred,FixedPreds,CoreOut) :-
    typecheck_pred_for_unsat_core(Pred,TypedPred),
    maplist(typecheck_pred_for_unsat_core,FixedPreds,TypedFixedPreds),
    conjunct_predicates(TypedFixedPreds,Conj),
    minimum_unsat_core_with_fixed_conjuncts(TypedPred,Conj,Core),
    translate_bexpression(Core,CoreOut).

typecheck_pred_for_unsat_core(PIn,POut) :-
    evaluate_formula_typecheck('#PREDICATE',PIn,POutT,_,[]),
    get_texpr_expr(POutT,exists(_,POut)).

/**
Access information about the current version of the ProB core.

#### called by:
* ProB 2.0: GetInternalRepresentationPrettyPrintCommand
*/
:- use_module(specfile,[get_internal_representation/1]).
get_pretty_print(PP) :-
    get_internal_representation(PPC),
    atom_codes(PP,PPC).
