% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% probsocket_proxy.pl: tool to call probcli via socket interface

:- use_module(library(sockets)).
:- use_module(library(system),[sleep/1]).

:- dynamic client_stream/1.
:- dynamic server_port/1, server_stream/1.
:- dynamic revision/1.

start :-
    revision(Revision),
    socket_server_open(Port, Socket),
    format('Starting proxy ProB socket server.~n',[]),
    format('Port: ~w~n', [Port]),
    format('probcli revision: ~w~n',[Revision]),
    format('-- starting command loop --~n', []),
    flush_output,
    socket_server_accept(Socket, _Client, Stream, [type(text)]),
    assert( client_stream(Stream) ),
    read_socket,
    socket_server_close(Socket).

read_term_from_socket(Term,Vars) :-
    client_stream(Stream),
    catch(read_term(Stream, Term, [variable_names(Vars)]), E, syntax_error(E)).
syntax_error(E) :-
    format('Got syntax error, get your Java code right!~n~w~n', [E]),
    stop.

codes_to_client(Codes) :-
    client_stream(Stream),
    put_codes(Codes, Stream),
    flush_output(Stream).
put_codes([],_).
put_codes([C|Rest],Stream) :-
    put_code(Stream,C),
    put_codes(Rest,Stream).

term_to_server(Term,Vars) :-
    ( server_stream(Stream) ->
        write_varterm(Stream,Term,Vars),
        write(Stream,'.\n'),
        flush_output(Stream)
    ; otherwise ->
        connect_to_server,
        term_to_server(Term,Vars)).

write_varterm(Stream,Term,Vars) :-
    ( var(Term) ->
        get_variable_name(Vars,Term,Varname),
        write(Stream,Varname)
    ; ground(Term) ->
        write_term(Stream,Term,[quoted(true)])
    ; otherwise ->
        Term =.. [Functor|Args],
        write_term(Stream,Functor,[quoted(true)]),
        write(Stream,'('),
        write_varterm_l(Args,Stream,Vars),
        write(Stream,')')).
write_varterm_l([Arg|Rest],Stream,Vars) :-
    write_varterm(Stream,Arg,Vars),
    write_varterm_l2(Rest,Stream,Vars).
write_varterm_l2([],_Stream,_Vars).
write_varterm_l2([Arg|Rest],Stream,Vars) :-
    write(Stream,','),
    write_varterm_l([Arg|Rest],Stream,Vars).
get_variable_name([Name=Var|_],V,N) :-
    Var==V,!,N=Name.
get_variable_name([_|Rest],V,N) :-
    get_variable_name(Rest,V,N).


connect_to_server :-
    retractall(server_stream(_)),
    server_port(Port),
    catch( ( socket_client_open(localhost:Port,Stream,[type(text)]),
               assert( server_stream(Stream) )),
           error(system_error, system_error(E)),
           retry_connect_to_server(E)).
retry_connect_to_server(E) :-
    format('Error when trying to connect to server - retry in a second (~w)~n',[E]),
    sleep(1),
    connect_to_server.
    
read_socket :-
    read_term_from_socket(Term,Vars),
    process_term(Term,Vars),
    read_socket.

process_term(end_of_file,_) :- !, stop.
process_term(halt,_) :- !, stop.
process_term(Term,Vars) :-
    term_to_server(Term,Vars),
    codes_from_server(Codes),
    codes_to_client(Codes).

codes_from_server(Codes) :-
    server_stream(Stream),
    read_from_server2(Stream,Codes).
read_from_server2(Stream,Codes) :-
    get_code(Stream, Code),
    ( Code == -1 ->
        Codes = [],
        retractall(server_stream(_))
    ; Code == 1 ->
        Codes = [1]
    ; otherwise ->
        Codes = [Code|Rest],
        read_from_server2(Stream,Rest)).

stop :-
    format('Stopping proxy server, bye!~n',[]),
    halt.



runtime_entry(start) :- start.

save(Port,Revision) :-
    assert( server_port(Port) ),
    assert( revision(Revision) ),
    save_program('probproxy.sav').
