% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(runtime_profiler,[reset_runtime_profiler/0,
                            profile_failure_driven_loop/1, profile_failure_driven_loop/2,
                            profile_single_call/2, profile_single_call/3,
                            tcltk_get_profile_info/1,
                            print_runtime_profile/0]).

% TO DO: optionally load more precise microsecond timer

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
 
:- if(\+ environ(prob_profile, true)).

reset_runtime_profiler.
:- meta_predicate profile_single_call(-,-,0).
profile_single_call(_Label,_ID,C) :- call(C).
:- meta_predicate profile_single_call(-,0).
profile_single_call(_Label,C) :- call(C).
profile_failure_driven_loop(_).
profile_failure_driven_loop(_,_).
print_runtime_profile :-
    print('No profiling information available'),nl,
    print('Recompile ProB with -Dprob_profile=true'),nl.
tcltk_get_profile_info(list([I1,I2])) :-
    I1='No profiling information available',
    I2='Recompile ProB with -Dprob_profile=true'.

:- else.

:- volatile total_runtime/5.
:- dynamic total_runtime/5.

reset_runtime_profiler :-
  retractall(total_runtime(_,_,_,_,_)).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_runtime_profiler,
                           'Reset runtime profiler.').

:- meta_predicate profile_single_call(-,-,0).
profile_single_call(Label,ID,C) :-
   start_profile(Label,S),
   if(call(C),stop_profile(Label,ID,S),(stop_profile(Label,ID,S),fail)).

:- meta_predicate profile_single_call(-,0).
profile_single_call(Label,C) :- profile_single_call(Label,unknown,C).


% register runtime for a failure driven loop:
profile_failure_driven_loop(ActionName) :- profile_failure_driven_loop(ActionName,unknown).
profile_failure_driven_loop(ActionName,ID) :-
   start_profile(ActionName,Info),
   register_time(ActionName,ID,Info).
register_time(_,_,_).
register_time(AN,ID,Info) :-
   stop_profile(AN,ID,Info),
   fail.

start_profile(_Label,[RT1,WT1]) :-
   statistics(runtime,[RT1,_Delta]),
   statistics(walltime,[WT1,_]).
stop_profile(Label,ID,[RT1,WT1]) :- 
   statistics(runtime,[RT2,_]),
   statistics(walltime,[WT2,_]),
   DRT is RT2-RT1, DWT is WT2-WT1,
   %format('Runtime for ~w : ~w ms (~w wall)~n',[AN,DRT,DWT]),
   add_runtime(Label,ID,DRT,DWT).

add_runtime(_,_,0,0) :- !.
add_runtime(PP,ID,RTime,WTime) :- (var(PP) -> PP2='unknown' ; PP2 = PP),
   (retract(total_runtime(PP2,OldR,OldW,Max,MaxID)) -> true ; OldR=0,OldW=0,Max=0,MaxID=unknown),
   NewR is OldR+RTime, NewW is OldW+WTime,
   (WTime > Max -> NewMax = WTime, NewMaxID=ID ; NewMax = Max, NewMaxID=MaxID),
   assert(total_runtime(PP2,NewR,NewW,NewMax,NewMaxID)).


:- use_module(library(lists)).
print_runtime_profile :-
   findall(profile(RT,WT,MaxWT,AN,MaxID),total_runtime(AN,RT,WT,MaxWT,MaxID),List),
   sort(List,SL),
   reverse(SL,RSL),
   maplist(print_runtime_profile_info,RSL).

print_runtime_profile_info(profile(RT,WT,MaxWT,AN,unknown)) :-
   format(' ~w : ~w ms (~w ms total & ~w ms max. walltime)~n',[AN,RT,WT,MaxWT]).
print_runtime_profile_info(profile(RT,WT,MaxWT,AN,MaxID)) :-
   format(' ~w : ~w ms (~w ms total & ~w ms max. walltime for ~w)~n',[AN,RT,WT,MaxWT,MaxID]).

tcltk_get_profile_info(list([list(['Operation Name','Total Runtime (ms)','Total Walltime (ms)','Max. Walltime (ms)','Max. Witness ID'])|L])) :-
   findall(profile(RT,WT,MaxWT,AN,MaxWitnessID),total_runtime(AN,RT,WT,MaxWT,MaxWitnessID),List),
   sort(List,SL),
   reverse(SL,RSL),
   print(RSL),nl,
   maplist(get_profile_list,RSL,L).
get_profile_list(profile(RT,WT,MaxWT,AN,MaxID),list([AN,RT,WT,MaxWT,MaxID])).


:- endif.
