% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

/* 
 ensure_loaded('images/state_viewer_images').
*/

/* ------------------------------------------------------------------ */

:- use_module(bmachine,[b_machine_name/1,
                        b_get_machine_animation_function/2,
                        b_get_definition_string_from_machine/2,
                        get_animation_image/2]).

/* this file is not a module (yet):
:- use_module(module_information).
:- module_info(group,gui).
:- module_info(description,'This module can provide a Tk graphical representation of a B state.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').
*/

:- use_module(specfile,[state_corresponds_to_fully_setup_b_machine/2]).

/* First argument: image list of pairs: ImageName, ImageFilePath
   Second argument: list of triplets {type framename label}
   */

:- use_module(translate,[    print_bexpr/1 ]).


eval_definition_funs([],_,[]).
eval_definition_funs([anim_fun(_Nr,AnimFunction1)|T],BState,Result) :-
    eval_definition_fun(AnimFunction1, BState, FunctionRes1),
    eval_definition_funs(T,BState,FunctionRes0),
    override_anim_fun(FunctionRes1,FunctionRes0,Result).

:- use_module(xtl_interface,[xtl_animation_function_result/2]).
:- public tcltk_get_image_list/2.
tcltk_get_image_list(list(ImageList),list(FramesAndItems)) :-
       current_expression(_CurID,State),
       specfile:state_corresponds_to_fully_setup_b_machine(State,BState),
       findall(anim_fun(Nr,AF),b_get_machine_animation_function(AF,Nr),AnimFuns),
       AnimFuns \= [],
       sort(AnimFuns,SAnimFuns),
       get_animation_images(ImageList),
       %print(animation_images(ImageList)),
      % print('Animation Function: '),
      % print_bexpr(SAnimFuns),
       eval_definition_funs(SAnimFuns, BState, FunctionRes),
       % print('full_fun'(FunctionRes)),nl,
       display_animation_function_result(FunctionRes,FramesAndItems).
tcltk_get_image_list(list([]),list(FramesAndItems)) :-
       animation_mode(xtl),
       current_expression(_CurID,State),
       xtl_interface:xtl_animation_function_result(State,FunctionRes),
       clear_animation_images,
       display_animation_function_result(FunctionRes,FramesAndItems).
       

display_animation_function_result(FunctionRes,FramesAndItems) :-
       (FunctionRes = int(ImageNr)
         ->  /* simple style animation: just a single image */
             convert_nr_into_tk_label(ImageNr,"IMG",Label),
             FramesAndItems = [frame, row1, top, img11, row1, Label]
          ;  ( get_min_max_row(FunctionRes,Min,Max)
               ->  debug_println(9,'Matrix function'),
                   debug_println(9,min_max(Min,Max)),
                   setup_frames(Min,Max,Frames), debug_println(9,frames(Frames)),
                   sort(FunctionRes,SortedFR),
                   get_frame_items(SortedFR,Items),
                   % print(items(Items)),nl,
                   pad_text_frame_items(Items,PItems),
                   append(Frames,PItems,FramesAndItems)
               ;   FunctionRes=[] -> print('**** Empty ANIMATION_FUNCTION result'),nl,
                   fail
               ;   print('**** Unknown Format for ANIMATION_FUNCTION result'),nl,
                   print('**** must be of type INT or (INT*INT)+->INT'),nl,
                   % print(FunctionRes),nl, %%
                   fail
             )
       ).

:- use_module(custom_explicit_sets,[try_expand_custom_set/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :-
       b_interpreter:b_compute_expression(AnimFunction,[],BState,FunctionResCl),
     %%  nl, print('FunctionResult'(FunctionResCl)),nl,
       ~~mnf( custom_explicit_sets:try_expand_custom_set(FunctionResCl, ExpCl) ),
       (translate_to_ints(ExpCl, FunctionRes)
          -> true /* already translate to ints so that override will work properly */
          ; check_anim_function(ExpCl),fail
       )  % ,print('fres'(FunctionRes)),nl
       .

check_anim_function([]) :- !.
check_anim_function(int(_)) :- !. % just a single picture
check_anim_function([H|_]) :- !, 
  (H = ((Row,Col),_Img) ->
         (((get_int(Row,_),get_int(Col,_)) -> true
          ;  add_error(state_viewer_images,'Elements of ANIMATION_FUNCTION should be convertible to (INT*INT)*INT: ',H))
         )
   ; H = (Row,(Col,_Img)) ->
         (((get_int(Row,_),get_int(Col,_)) -> true
          ;  add_error(state_viewer_images,'Elements of ANIMATION_FUNCTION should be convertible to INT*(INT*INT): ',H))
        )
     ;  add_error(state_viewer_images,
            'Elements of the ANIMATION_FUNCTION should be of type (INT*INT)*INT: ',H)
   ).
check_anim_function((A,B)) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be a set of pairs, not a pair: ',(A,B)).
check_anim_function(X) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be of type (INT*INT) +-> INT (or similar): ',X).

:- use_module(library(avl)).

translate_to_ints(In,Out) :- empty_avl(E),translate_to_ints(In,E,Out).
translate_to_ints([],_,[]) :-!.
translate_to_ints(X,_,int(IX)) :- get_int(X,IX),!.
translate_to_ints([H|T], AVL, Res) :- 
   translate_element(H,AVL, NewAVL, Res, NewRes),
   translate_to_ints(T,NewAVL,NewRes).
  
translate_element(((Row,Col),Img),AVL, NewAVL, Res, NewRes) :- !,
   get_int(Row,RI), get_int(Col,CI), (Img=II),
   add_image(RI,CI,II, AVL, NewAVL, Res, NewRes).
translate_element((Row,(Col,Img)),AVL, NewAVL, Res, NewRes) :- !,
   translate_element(((Row,Col),Img), AVL, NewAVL, Res, NewRes).
translate_element((Row,Set),AVL,NewAVL,Res,NewRes) :- % Set can be relation or sequence
   % expand Set and associate elements with Row
   try_expand_custom_set(Set, ESet),
   findall((Row,El),member(El,ESet),All),
   l_translate(All,AVL,NewAVL,Res,NewRes).

l_translate([],AVL,AVL,Res,Res).
l_translate([H|T],AVL,NewAVL,Res,NewRes) :- translate_element(H,AVL,NewAVL1,Res,NewRes1),
   l_translate(T,NewAVL1,NewAVL,NewRes1,NewRes).

add_image(RI,CI,II, AVL, NewAVL, Res, NewRes) :-
   (avl_fetch((RI,CI),AVL,ExistingValue)
    -> add_error(state_viewer_images,
          'Animation function defines two images for (Row,Col):[img1,img2] = ',((RI,CI):[ExistingValue,II])),
       NewAVL = AVL, Res=NewRes
     ; avl_store((RI,CI),AVL,II,NewAVL), Res=[((RI,CI),II)|NewRes]
   ).
 
/* a version of override which does not worry about correct typing */
% override_anim_fun(Fun1, Fun2, Res)  :: Fun2 overrides Fun1
override_anim_fun([],Fun2,Res) :- !, Res=Fun2.
override_anim_fun(int(_),Fun2,Res) :- !, Res=Fun2.
override_anim_fun([((RowT,Col),Img)|T],Fun2,Res) :- !,
   (member(((RowT,Col),_),Fun2)
     -> Res = ResT % we use version from Fun2
      ; Res = [((RowT,Col),Img)|ResT]
   ), override_anim_fun(T,Fun2,ResT).
override_anim_fun(X,F,R) :-
   add_error(state_viewer_images,'Unknown Format for ANIMATION_FUNCTION_DEFAULT', X),
   R=F.

get_min_max_row([((Row,_Col),_Img)|T],Min,Max) :-
   number(Row), % this is already in a format for Tcl/Tk; not int(.)
   get_min_max_row2(T,Row,Row,Min,Max).
get_min_max_row2([],Min,Max,Min,Max).
get_min_max_row2([((Row,_Col),_Img)|T],Min,Max,MinR,MaxR) :-
    (Row<Min -> Min1=Row ; Min1=Min),
    (Row>Max -> Max1=Row ; Max1=Max),
    get_min_max_row2(T,Min1,Max1,MinR,MaxR).

get_int(Nr,R) :- number(Nr),!,R=Nr.
get_int(int(R),R).
get_int(fd(R,_Name),R).
get_int(pred_true /* bool_true */,1).
get_int(pred_false /* bool_false */,0).

% add extra whitespace to make all text fields the same length
% TO DO: maybe only pad within a column; not overall
pad_text_frame_items(FrameItems,NewFrameItems) :- b_or_z_mode,
   (b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_LEFT') ;
    b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_RIGHT') ),
   longest_text(FrameItems,0,Len),
   Len>0,
   !,
   %print(longest(Len)),nl,
   pad_text_frame_items_to(FrameItems,Len,NewFrameItems).
pad_text_frame_items(F,F).


pad_text_frame_items_to([],_,[]).
pad_text_frame_items_to([ItemLabel,Frame,Text|TR],Len,[ItemLabel,Frame,NewText|NTR]) :-
   (is_text_type(ItemLabel)
      -> atom_codes(Text,Codes),
         length(Codes,TLen),
         AdditionalSpaces is Len-TLen, gen_spaces(AdditionalSpaces,AS),
         (b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_RIGHT')
           -> append(AS,Codes,NewCodes)
           ;  append(Codes,AS,NewCodes) % 34 -> ' 39 -> "
         ),
         atom_codes(NewText,NewCodes)
       ; NewText=Text),
   pad_text_frame_items_to(TR,Len,NTR).

gen_spaces(Nr,R) :- Nr<1,R=[]. % 34 -> '
gen_spaces(Nr,[32|T]) :- N1 is Nr-1, gen_spaces(N1,T).


% find longest text label; so that we can pad the other ones with spaces
longest_text([],R,R).
longest_text([ItemLabel,_Frame,Text|TR],Acc,R) :-
   is_text_type(ItemLabel),!,
   atom_length(Text,Len),
   (Len>Acc -> NAcc = Len ; NAcc=Acc),
   longest_text(TR,NAcc,R).
longest_text([_,_,_|TR],Acc,R) :- longest_text(TR,Acc,R).

is_text_type(ItemLabel) :- atom_codes(ItemLabel,Codes), append("text",_,Codes).

get_frame_items([],[]).
get_frame_items([((Row,Col),Img)|T],[ItemLbl,FrameLbl,ImageLbl|TR]) :-
    convert_nr_into_tk_label(Row,"frame",FrameLbl),
    ((get_int(Img,ImgNr),images(ImgNr,_,ImageLbl))
      -> convert_2nr_into_tk_label(Row,Col,"item",ItemLbl)
      ;  convert_2nr_into_tk_label(Row,Col,"text",ItemLbl),
         ((get_int(Img,ImgNr),get_animation_string(ImgNr,ImageLbl)) -> true
            ; translate:translate_bvalue(Img,ImageLbl)
         )
    ),
    get_frame_items(T,TR).

setup_frames(Min,Max,[]) :- Min>Max,!.
setup_frames(Min,Max,[frame, FrameLbl, top |T]) :-
    convert_nr_into_tk_label(Min,"frame",FrameLbl), M1 is Min+1,
    setup_frames(M1,Max,T).

:- dynamic images/3.

get_animation_images(Res) :- clear_animation_images,
    findall( img(Nr,S,Lbl), get_animation_image_and_label(Nr,S,Lbl), List),
    flatten_img_list(List,Res).
clear_animation_images :- retractall(images(_,_,_)).

flatten_img_list([],[]).
flatten_img_list([img(Nr,S,Lbl)|T],[Lbl,S|FT]) :- assert(images(Nr,S,Lbl)), flatten_img_list(T,FT).

get_animation_string(Nr,S) :- b_or_z_mode,
         number_codes(Nr,TailCodes),
         /* ANIMATION_STR */
         atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,83,84,82|TailCodes]),
         b_get_definition_string_from_machine(Def_Name,S).
         
get_animation_image_and_label(Nr,S,Lbl) :-
    get_animation_image(Nr,S),
    convert_nr_into_tk_label(Nr,"IMG",Lbl).

convert_nr_into_tk_label(Nr,Prefix,Label) :- number_codes(Nr,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).
        
convert_2nr_into_tk_label(Nr1,Nr2,Prefix,Label) :-
        number_codes(Nr1,Codes1), number_codes(Nr2,Codes2),
        append("_",Codes2,C2), append(Codes1,C2,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).
 