% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(succeed_max,[reset_max_reached/0,
                       max_reached/0, max_reached/1,
                       succeed_max_call/2, succeed_max_call_id/3,
                       succeed_max_initialise/0]).

:- use_module(tools).

:- use_module(module_information).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides a meta call to retrieve a specified maximum number of solutions.').

:- use_module(extension('counter/counter')).

:- dynamic max_reached/1.

max_reached :- max_reached(_),!.

assert_max_reached(Id) :- 
   (max_reached(Id) -> true ; assert(max_reached(Id))).
reset_max_reached :- retractall(max_reached(_)).

:- meta_predicate succeed_max_call(0,-).
:- meta_predicate succeed_max_call_id(-,0,-).

succeed_max_call(Module:Call,MaxNrOfSols) :-
    succeed_max_call_id('$unknown',Module:Call,MaxNrOfSols).


succeed_max_initialise :- counter_init,new_counter(inits_found),new_counter(constants_found),new_counter(ops_found).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,succeed_max_initialise,
                           'Initialise sols_found counter.').


succeed_max_call_id(Id,Call,MaxNrOfSols) :- 
   get_counter_name(Id,Counter),
   reset_counter(Counter),
   succeed_max_call_id_aux(Id,Counter,Call,MaxNrOfSols).

succeed_max_call_id_aux(Id,_,_Call,MaxNrOfSols) :- 
   MaxNrOfSols==0,!,
   assert_max_reached(Id),fail.
succeed_max_call_id_aux(Id,Counter,Call,MaxNrOfSols) :- %print_message(max_call(Call,MaxnrOfSols)), 
    call(Call),
    inc_counter(Counter,X1),
    %  print_message(sol(X1)), %
    ((X1>=MaxNrOfSols) 
      -> (!, %print_message(max_sol_found(Call,X1,MaxNrOfSols)),
          assert_max_reached(Id)
         )
      ;  true).

% use different counters so that we can have at least one of each succeed_max_call pending without interaction
get_counter_name(V,Name) :- var(V),!, Name=ops_found.
get_counter_name('$initialise_machine',Name) :- !, Name=inits_found.
get_counter_name('$setup_constants',Name) :- !,Name=constants_found.
get_counter_name(_,ops_found).

