:- module(library_setup,[library_setup/4,is_seq/1]).

:- use_module(library(lists),[append/2]).

library_setup(Type,LibExpansion,Example,(Operators,Arithmetic)) :- 
    count_var_types(Example,Int,Bool,(Set,SetTypes),(Seq,SeqTypes)) , 
    library_add_operator(Type,LibExpansion,intOp,Int,[],TLib1) , 
    library_add_operator(Type,LibExpansion,boolOp,Bool,TLib1,TLib2) , 
    library_add_operator(Type,LibExpansion,setOp-SetTypes,Set,TLib2,TLib3) , 
    library_add_operator(Type,LibExpansion,seqOp-SeqTypes,Seq,TLib3,TOperators) , 
    library_add_conn(LibExpansion,Example,Or,And) , 
    append([Or,And,TOperators],Operators) , 
    library_add_operator(Type,LibExpansion,arithmetic,Int,[],Arithmetic).

count_var_types(Example,LInt,LBool,(LSet,SetTypes),(LSeq,SeqTypes)) :- 
    findall(Val,member(b(Val,integer,_),Example),Int) , 
    findall(Val,member(b(Val,boolean,_),Example),Bool) , 
    findall(Type,(member(b(Node,Type,_),Example) , is_seq(b(Node,Type,_))),SeqTypes) , 
    findall(Type,(member(b(Node,Type,_),Example) , Type = set(_) , \+is_seq(b(Node,Type,_))),SetTypes) , 
    length(Int,LInt) , length(Bool,LBool) , length(SetTypes,LSet) , length(SeqTypes,LSeq).

is_seq(b(_,seq(_),_)).
is_seq(Seq) :- 
    Seq = b(_,set(couple(integer,_)),_) , 
    % if the domain is enumerated from 1 to n we suppose it is a sequence
    b_interpreter:b_test_boolean_expression_wf(b(equal(b(domain(Seq),set(integer),[]),
        b(interval(b(integer(1),integer,[]),b(size(Seq),integer,[])),set(integer),[])),pred,[]),[],[]).

% add conjunct/disjunct to library
library_add_conn(1,Example,Or,And) :- 
    length(Example,VarAmount) , 
    C is floor((VarAmount-1)/2) , 
    library_add_conn_aux(C,Or,And).
library_add_conn(_,Example,Or,And) :- 
    length(Example,VarAmount) , 
    C is floor(VarAmount/2) , 
    library_add_conn_aux(C,Or,And).
library_add_conn_aux(C,Or,And) :- 
    length(And,C) , length(Or,C) , 
    findall(A,(member(A,And), A = conjunct),And) , 
    findall(O,(member(O,Or), O = disjunct),Or).

% limited arithmetic library for guards and invariants
library_add_operator(Type,1,OpType,1,Acc,Acc) :- 
    (OpType = arithmetic ; OpType = intOp) , 
    (Type = guard ; Type = invariant) , !.
library_add_operator(Type,_,arithmetic,1,Acc,Acc) :- 
    (Type = guard ; Type = invariant) , !.
library_add_operator(_,_,_,0,Acc,Acc).
library_add_operator(Type,LibExpansion,OpType,VarAmount,Acc,Library) :- 
    NVarAmount is VarAmount - 1 , 
    (OpType = TOp-[SetType|_] 
    % get the type from the list of set types
    ->  Op = TOp-SetType
    ;   Op = OpType) , 
    library_add_operator_aux(Type,LibExpansion,Op,Acc,Operators) , 
    library_add_operator(Type,LibExpansion,OpType,NVarAmount,Operators,Library).

% we define several library steps for each synthesis type to improve the performance
% start with small library and expand successively

% integer
library_add_operator_aux(action,_,intOp,Acc,[equal-integer,constant-integer|Acc]).
library_add_operator_aux(_,_,intOp,Acc,[greater,equal-integer,constant-integer|Acc]).

% boolean
library_add_operator_aux(_,_,boolOp,Acc,[equal-boolean,constant-boolean,negation|Acc]).

% arithmetic
library_add_operator_aux(action,_,arithmetic,Acc,[add,minus,multiplication,div|Acc]).
library_add_operator_aux(_,1,arithmetic,Acc,[add,minus|Acc]).
library_add_operator_aux(_,_,arithmetic,Acc,[add,minus,multiplication|Acc]). 

% set
% use min/max only if we have set(integer), general_union/intersection only if we have set(set(_))
library_add_operator_aux(action,1,setOp-SetType,Acc,[constant-SetType,union-SetType,general_union-SetType,general_intersection-SetType|Acc]) :- 
    SetType = set(set(_)).
library_add_operator_aux(action,1,setOp-SetType,Acc,[constant-SetType,union-SetType|Acc]) :- 
    SetType \= set(set(_)).
library_add_operator_aux(action,2,setOp-SetType,Acc,[constant-SetType,union-SetType,subset-SetType,subset_strict-SetType|Acc]).
library_add_operator_aux(action,3,setOp-SetType,Acc,[constant-SetType,union-SetType,set_subtraction-SetType|Acc]).
library_add_operator_aux(action,_,setOp-SetType,Acc,[constant-SetType,union-SetType,subset-SetType,subset_strict-SetType,set_subtraction-SetType,
                                                     general_union-SetType,general_intersection-SetType|Acc]) :- 
    SetType = set(set(_)).
library_add_operator_aux(action,_,setOp-SetType,Acc,[constant-SetType,union-SetType|Acc]) :- 
    SetType \= set(set(_)).
library_add_operator_aux(_,1,setOp-SetType,Acc,[card-SetType,constant-SetType,constant-integer,greater|Acc]). 
library_add_operator_aux(_,2,setOp-SetType,Acc,[card-SetType,constant-SetType,max-SetType,min-SetType,equal-SetType,constant-integer,greater|Acc]) :- 
    SetType = set(integer).
library_add_operator_aux(_,2,setOp-SetType,Acc,[card-SetType,constant-SetType,equal-SetType,constant-integer,greater|Acc]) :- 
    SetType \= set(integer).
library_add_operator_aux(_,3,setOp-SetType,Acc,[constant-SetType,equal-SetType,subset_strict-SetType,subset-SetType|Acc]). 
library_add_operator_aux(_,4,setOp-SetType,Acc,[constant-SetType,equal-SetType,subset-SetType,subset_strict-SetType,general_union-SetType,general_intersection-SetType|Acc]) :- 
    SetType = set(set(_)).
library_add_operator_aux(_,_,setOp-SetType,Acc,[constant-SetType,equal-SetType,subset-SetType,subset_strict-SetType|Acc]) :- 
    SetType \= set(set(_)).

% sequence
library_add_operator_aux(Type,LibExpansion,seqOp-SetType,Acc,Library) :- 
    (SetType = seq(InnerType) ; SetType = set(couple(integer,InnerType))) , 
    library_add_operator_aux_seq(Type,LibExpansion,SetType,TLibrary) , 
    % use constant of type seq(_)
    append([[constant-seq(InnerType)],TLibrary,Acc],Library).

library_add_operator_aux_seq(action,1,SetType,[overwrite-SetType]).
library_add_operator_aux_seq(action,2,SetType,[concat-SetType]).
library_add_operator_aux_seq(action,3,SetType,[overwrite-SetType,constant-SetType]).
library_add_operator_aux_seq(action,4,SetType,[concat-SetType,constant-SetType]).
library_add_operator_aux_seq(action,_,SetType,[concat-SetType,overwrite-SetType,constant-SetType]).

library_add_operator_aux_seq(_,1,SetType,[size-SetType,constant-integer,greater]).
library_add_operator_aux_seq(_,2,SetType,[size-SetType,card-SetType,equal-integer,constant-integer,greater]).
library_add_operator_aux_seq(_,3,SetType,[size-SetType,card-SetType,range-SetType,domain-SetType,equal-SetType,constant-SetType]).
library_add_operator_aux_seq(_,4,SetType,[range-SetType,domain-SetType,image-SetType,reverse-SetType,equal-SetType,constant-SetType]).
library_add_operator_aux_seq(_,5,SetType,[range_restriction-SetType,range_subtraction-SetType,domain_restriction-SetType,domain_subtraction-SetType,
                                          equal-SetType,constant-SetType]).
library_add_operator_aux_seq(_,_,SetType,[restrict_front-SetType,restrict_tail-SetType,constant-SetType,range_restriction-SetType,range_subtraction-SetType,
                                          domain_restriction-SetType,domain_subtraction-SetType,range-SetType,domain-SetType,image-SetType,equal-SetType,
                                          first-SetType,last-SetType,union-SetType,equal-SetType]).