% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(table_tools, [expand_and_translate_to_table/3, print_table/2]).

:- use_module(translate).
:- use_module(tools_strings,[ajoin/2]).
:- use_module(library(lists)).
:- use_module(self_check).



% expand a SetValue into a Header List and a Table List

:- assert_must_succeed(( expand_and_translate_to_table([rec([field(a,int(1)),field(b,int(2))])],H,T),
                          H == ['Nr',a,b], T==[list([1,'1','2'])] )).
:- assert_must_succeed(( expand_and_translate_to_table([(int(1),int(2)),(int(33),int(44))],H,T),
                          H == ['Nr',prj1,prj2], T==[list([1,'1','2']),list([2,'33','44'])] )).
expand_and_translate_to_table(Value,Header,Table) :-  
    expand(Value, ExpVal),
    translate_to_table(ExpVal,1,Table,Header). %, print(table(Res)),nl.

:- use_module(custom_explicit_sets,[is_custom_explicit_set/1]).
expand((A,B),[(A,B)]) :- !. % wrap into a singleton set
expand(rec(Fields),[rec(Fields)]) :- !. % wrap into a singleton set
expand(Value,ExpVal) :- is_custom_explicit_set(Value),!,
      custom_explicit_sets:try_expand_custom_set(Value, ExpVal).
expand([H|T],V) :- !, V=[H|T].
expand([],V) :- !, V=[].
expand(Val,[Val]).


translate_to_table([],_RowNr,[],['Nr','Elements']) :- !.
translate_to_table([A|T],RowNr,[list([RowNr|LA])|TT],['Nr'|Header]) :- !,
    get_table_entry(A,LA,Header),
    R1 is RowNr+1,
    translate_to_table(T,R1,TT,_).
translate_to_table(A,RowNr,[list([RowNr|AT])],['Nr','Value']) :- !,
    translate:translate:translate_bvalue(A,AT).

get_table_entry(rec(Fields),Vals,FieldNames) :- !,
    maplist(get_record_field_val_name,Fields,Vals,FieldNames).
get_table_entry(A,[AT],['Elements']) :- A \= (_,_),!,
    translate:translate:translate_bvalue(A,AT).
get_table_entry(A,AT,HA) :- %A=(A1,A2), (A1 = (_,_) ; A2=(_,_)),
    !,
    get_table_entry2(A,[],AT,HA).
%get_table_entry((A,B),[AT,BT],['prj1','prj2']) :- !,
%    translate:translate:translate_bvalue(A,AT),
%    translate:translate:translate_bvalue(B,BT).

get_record_field_val_name(field(Name,Value),ValS,Name) :-
    translate:translate:translate_bvalue(Value,ValS).

get_table_entry2((A,B),Path,ABT,HAB) :- !,
    get_table_entry2(A,[1|Path],AT,HA),
    get_table_entry2(B,[2|Path],BT,HB),
    append(AT,BT,ABT), append(HA,HB,HAB).
%get_table_entry2(rec([field(Name,Val)|Rest]),Path,ABT,HAB) :- !,
get_table_entry2(A,Path,[AT],[Header]) :- !,
    reverse(Path,RP), ajoin(['prj'|RP],Header), % to do: do all this more efficiently
    translate:translate:translate_bvalue(A,AT).
    

print_table(Header,Table) :-
   maplist(print_entry,Header),
   nl,
   maplist(print_row,Table).


print_row(list(Row)) :- maplist(print_entry,Row),nl.
% TO DO: determine column length dynamically
print_entry(E) :- print(E), my_atom_length(E,Len), WS is 10-Len, print_ws(WS).
print_ws(X) :- X>1, !, print(' '), X1 is X-1, print_ws(X1).
print_ws(_) :- print(' ').

my_atom_length(AtomOrNumber,Len) :- atom(AtomOrNumber),!, atom_length(AtomOrNumber,Len).
my_atom_length(A,Len) :- name(A,List), length(List,Len).