% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% tcltk_interface.pl

% do we still need this ?
prob_use_module(X) :- load_files(X,[if(changed),load_type(source),compilation_mode(compile)]).
prob_use_module(X,I) :-
  load_files(X,[if(changed),load_type(source),compilation_mode(compile),imports(I)]).

/* load all necessary modules */
:- use_module(library(random)).
:- use_module(library(lists)).
:- use_module(library(system)).

:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(process)).

:- prob_use_module(typechecker).
:- prob_use_module(self_check).
:- prob_use_module(kernel_waitflags).
                           
:- prob_use_module(specfile).
:- prob_use_module(translate).
:- prob_use_module(tools).
:- prob_use_module(b_enumerate).
:- prob_use_module(b_state_model_check).
:- prob_use_module(bmachine).


:- prob_use_module(debug).
:- prob_use_module(preferences).
:- prob_use_module(version).
:- prob_use_module(b_interpreter).
:- prob_use_module(predicate_debugger).
:- prob_use_module(b_trace_checking).
:- prob_use_module(gensym).
:- use_module(delay).
:- prob_use_module(bsyntaxtree).

:- prob_use_module(kernel_objects).
:- prob_use_module(bsets_clp).

:- prob_use_module(state_space).
:- prob_use_module(state_space_exploration_modes).


:- prob_use_module(b_compiler).
:- prob_use_module(error_manager).
:- prob_use_module(model_checker).

:- prob_use_module(visualize_graph).
:- prob_use_module(state_space_reduction).
:- prob_use_module(state_custom_dot_graph).
:- prob_use_module(fdr_csp_generator).
:- prob_use_module(coverage_statistics).
:- prob_use_module(succeed_max).

%:- prob_use_module(testcase_generator).

:- use_module(prozsrc(proz),[open_proz_file/2]).

:- prob_use_module(reduce_graph_state_space).
:- prob_use_module(flow).

:- prob_use_module(probltlsrc(ltl)).
:- prob_use_module(probltlsrc(ctl)).
:- prob_use_module(kodkodsrc(kodkod)).
:- prob_use_module(graph_canon).
:- prob_use_module('promela/promela_ncprinter').

:- prob_use_module(sap).
:- prob_use_module(plugins,[notify_new_transition_to_plugin/5]).

:- prob_use_module(system_call,[generate_temporary_cspm_file/2, get_writable_compiled_filename/3]).

:- prob_use_module(bvisual2).
:- prob_use_module(b_show_history).
:- prob_use_module(enabling_analysis).
:- prob_use_module(mcdc_coverage).
:- prob_use_module(prologTasks).

% Modules from por folder
:- prob_use_module(probporsrc(dot_graphs_static_analysis)).
:- prob_use_module(probporsrc(static_analysis)).
:- prob_use_module(probporsrc(ample_sets)).
:- prob_use_module(probpgesrc(pge_algo)).

:- use_module(probcspsrc(haskell_csp),[parse_and_load_cspm_file/1, cspm_file_add_remove_unicode/3,
                                       get_csp_assertions_as_string/2,get_csp_processes/1,
                                       parse_and_load_cspm_file_into_specific_pl_file/2, load_cspm_pl_file/1,
					evaluate_csp_expression/2,evaluate_csp_expression/3,
					parse_single_csp_expression_file/3, parse_single_csp_declaration/3]).
%:- use_module(probcspsrc(slicer_csp),[slice_from_program_point/14]).

:- use_module(value_persistance, [save_constants/1,add_new_transitions_to_cache/1]).

%:- use_module(ast_inspector).
:- use_module(tcltk_tree_inspector).

:- prob_use_module(symbolic_model_checker(cbc_ba)).

/* main program 
 * ============
 */

 % use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(tcltk_interface,group,infrastructure).
:- module_info(tcltk_interface,description,'Interface between the Tcl/Tk GUI, the CLI and the internal modules.').


   

/* -------------------------------------------------------------------- */

%tcltk_find_untyped_vars(Vs) :- find_untyped_vars(Vs), Vs \== [].

%tcltk_find_untyped_consts(Vs) :- find_untyped_consts(Vs), Vs \== [].

tcltk_find_max_reached_node :- tcltk_find_max_reached_node(_).
tcltk_find_max_reached_node(ID) :- max_reached_or_timeout_for_node(ID).

tcltk_current_node_has_no_real_transition :- 
  \+ tcltk_current_node_has_real_transition.
tcltk_current_node_has_real_transition :-
  current_state_id(CurID),
  transition(CurID,Trans,_),
  (CurID \= root ; Trans \= '$partial_setup_constants'),!.
tcltk_current_node_has_partial_transition :-
  current_state_id(CurID),
  transition(CurID,'$partial_setup_constants',_),!.

/* --------------------------------------------------------------------- */
tcltk_goto_state(ActionAsTerm,StateID) :-
  (var(StateID) -> print_message(var_state(tcltk_goto_state(ActionAsTerm,StateID)))  ;true),
   (visited_expression_id(StateID)
    -> true
    ;  print(state_does_not_exist(tcltk_goto_state(ActionAsTerm,StateID))),nl,
       fail
    ),
   (retract(current_state_id(CurID))->true;print(no_current_state_id),nl),
   assert(current_state_id(StateID)),
   (transition(CurID,ActionAsTerm,OpID,StateID)
    -> /* ok: proper transition */
       add_to_op_trace_ids(OpID),
       add_id_to_history(CurID)
     ; (CurID=StateID
        -> true
        ;  debug_print(9,jumping(CurID,ActionAsTerm,StateID)),debug_nl(9),
           add_to_op_trace_ids(jump(StateID)),
           add_id_to_history(CurID)
       )
    ).

add_id_to_history(CurID) :-
   (retract(history(History))->true),
   assert(history([CurID|History])), retractall(forward_history(_)).

initialise_operation_not_yet_covered :- retractall(operation_not_yet_covered(_)),
   b_or_z_mode,
   b_top_level_operation(Name),
  % b_get_machine_operation(Name,_,Par,_),  length(Par,Arity), functor(Op,Name,Arity), 
   % Note: no '-->' added 
   assert(operation_not_yet_covered(Name)), 
   debug_println(operation_not_yet_covered(Name)),
   fail.
   /* Missing: treat operations with return values */
initialise_operation_not_yet_covered.


tcltk_add_new_transition(FromID,Action,ToID,ToTempl,TransInfo) :-
    tcltk_add_new_transition_transid(FromID,Action,ToID,ToTempl,TransInfo,_TransId).

tcltk_add_new_transition_transid(FromID,Action,ToID,ToTempl,TransInfo,TransId) :-
   get_id_of_node_and_add_if_required(ToTempl,ToID,Exists,FromID),
   %% print_message(tcltk_add_new_transition(FromID,Action,ToID,Exists)), %%
   ((Exists=exists,transition(FromID,Action,TransId,ToID)) 
     -> true
     ;  (preferences:preference(forget_state_space,true)
          -> (any_transition(FromID,_,_) -> true
               ; store_transition(FromID,'$NO_DEADLOCK',FromID,TransId)) /* just add dummy transition */
          ;  store_transition(FromID,Action,ToID,TransId),
			 (store_transition_infos(TransInfo,TransId) -> true
					; add_internal_error('storing transition info failed', store_transition_infos(TransInfo,TransId)))
		),
	    add_cover(Action,OpName),
	    animation_mode(Mode),
        ((FromID \= ToID, b_or_z_mode(Mode), not_all_transitions_added(ToID))
            -> add_additional_destination_infos(FromID,OpName,ToID)
             ; true
         ),
        notify_new_transition_to_plugin(Mode,FromID,TransId,ToID,Exists)
   ).

add_cover('-->'(Action1,_),OpName) :- !,add_cover_aux(Action1,OpName).
add_cover(Action1,OpName) :- add_cover_aux(Action1,OpName).
add_cover_aux(Action1,OpName) :-  functor(Action1,OpName,_),
         (retract(operation_not_yet_covered(OpName))
           -> (preferences:get_preference(provide_trace_information,true)
                -> print(cover(OpName)),nl ; true),
              (operation_not_yet_covered(_) -> true ; nl,print('ALL OPERATIONS COVERED'),nl,nl)
           ; true
         ).


add_additional_destination_infos(FromID,ActionName,ToID) :- 
          preference(try_operation_reuse,true), 
          ToID \= FromID,
          animation_mode(b),
          preference(symmetry_mode,MODE), (MODE=off ; MODE=flood),
          % TO DO: can we enable this trick in symmetry mode ?
          bmachine:reuse_operation_effect(ActionName,OtherAction),
          \+ reuse_operation(ToID,OtherAction,_,_),
          assert(reuse_operation(ToID,OtherAction,FromID,ActionName)),
          % print(reuse_operation(ToID,OtherAction,FromID,ActionName)),nl, %%
          fail.
add_additional_destination_infos(FromID,ActionName,ToID) :- %print(add_dest(FromID,ActionName,ToID)),nl,
          ((b_specialized_invariant_for_op(ActionName,_),
            \+ invariant_not_yet_checked(FromID),
            \+ invariant_violated(FromID) % if Invariant does not hold in previous state, we have no idea about invariants in the new state (proof assumes invariant holds before)
            )
            -> 
              (retract(specialized_inv(ToID,OldAVL)) -> true ; empty_avl(OldAVL)),
               avl_store(ActionName,OldAVL,nothing,NewAVL),
               % TO DO: provide a special value if an operation preserves the entire invariant ??
               %print(specialized_inv(ToID,NewAVL)),nl,
               assert(specialized_inv(ToID,NewAVL))
             ; true
           ).
/*
Store names of incoming transitions in an AVL tree
Used to calculate invariant in presence of proof information
*/



  

:- use_module(state_space_exploration_modes,[set_depth_breadth_first_mode/1,dbf_modes/3]).


tcltk_set_dbf_mode(Nr) :- dbf_modes(Nr,Mode,_),!,
   set_depth_breadth_first_mode(Mode).
tcltk_set_dbf_mode(Nr) :- add_error(tcltk_set_dbf_mode,'Unknown mode: ',Nr).

tcltk_dbf_modes(list(Names)) :-  findall(N,dbf_modes(_,_,N),Names).


:- dynamic prev_randomise_enumeration_order/1.

tcltk_mark_current_node_to_be_recomputed_wo_timeout :-
  current_state_id(CurID), print(recomputing(CurID)),nl,
  tcltk_mark_node_for_recomputed(CurID),
   (use_no_timeout(CurID) -> true ; assert(use_no_timeout(CurID))).
tcltk_mark_current_node_to_be_recomputed_with_random_enum :-
  current_state_id(CurID), print(recomputing(CurID)),nl,
  tcltk_mark_node_for_recomputed(CurID),
  % TO DO: delete all destinations only reached by CurID
  retractall(transition(CurID,_,_,_)),
  temporary_set_preference(randomise_enumeration_order,true,PREV),
  assert(prev_randomise_enumeration_order(PREV)).
tcltk_finish_current_node_to_be_recomputed_with_random_enum :-
  retract(prev_randomise_enumeration_order(PREV)),
  reset_temporary_preference(randomise_enumeration_order,PREV).

tcltk_mark_node_for_recomputed(ID) :-
  (not_all_transitions_added(ID) -> true
    ; add_id_at_front(ID)),
  retractall(max_reached_for_node(ID)),
  retractall(time_out_for_node(ID,_,_)),
  (transition(ID,'$partial_setup_constants',ID2)
     -> delete_node(ID2) ; true),
  retractall(state_space:transition(ID,'$partial_setup_constants',_,_)).

transitions_computed_for_node(ID) :-
  \+(not_all_transitions_added(ID)),  \+ is_not_interesting(ID).


  
  

  



equivalent_states(S1,true,S2,true) :- \+(\+(S1=S2)).  /* IMPROVE, may even
 allow hash functions */
    
tcltk_set_initial_machine :-
  tcltk_clear_machine,
  b_set_initial_machine.

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').

% old profiler
:- use_module(runtime_profiler,[reset_runtime_profiler/0]).

:- use_module(eval_strings,[reset_eval_strings/0]).
:- use_module(eventhandling, [announce_event/1]).
tcltk_clear_machine :- 
   announce_event(reset_specification),
   announce_event(clear_specification),
   retractall(cached_state_as_strings(_,_,_)),
   set_animation_mode(b),
   bmachine:b_set_empty_machine,
   reset_errors,
   reset_eval_strings,
%   reset_dynamics, % new profiler
   % Now done via event handling:  b_interpreter:reset_b_interpreter,reset_model_checker,
    % bmachine:b_machine_reset, 
    % haskell_csp:clear_cspm_spec, state_space_initialise,  reduce_graph_state_space:reduce_graph_reset, reset_runtime_profiler reset_refinement_checker 
    % b_machine_hierarchy:reset_hierarchy, reset_all_temporary_preferences,
    % retractall(refinement_checker: generated_predeterministic_refinement(_,_,_)),
    % retractall(refinement_checker: determinism_check_div_found(_)),
    % retractall(counterexample_node(_)), retractall(counterexample_op(_))
    % clear_dynamic_predicates_for_POR,
    % clear_dynamic_predicates_for_PGE,
   garbage_collect_atoms. % reclaim atoms if possible

:- use_module(prob2_interface, [start_animation/0, update_preferences_from_spec/0]).
tcltk_initialise :-
                prob2_interface:update_preferences_from_spec,
                prob2_interface:start_animation,
                tcltk_try_reload.

tcltk_reset :- reset_errors,
               history(H), op_trace_ids(T), 
               current_state_id(CurID),
               (forward_history(FH) -> true ; FH=[]),
               state_space_reset,
               gen_forward_history(T,[CurID|H],FH,FwdHist),
               retractall(forward_history(_)),
               assert(forward_history(FwdHist)).

gen_forward_history([],H,R,R) :- (H=[root] -> true ; print(unexpected_remaining_hist(H)),nl).
gen_forward_history([A|TT],[ID|TH],Acc,Res) :-
  gen_forward_history(TT,TH,[forward(ID,A)|Acc],Res).

      
      
/* ---- getting available options ----- */


tcltk_get_options_dest_info(list(DestinationInfos)) :- debug_println(9,start_tcltk_get_options_dest_info),
  % get a list of all options which are skip events
  current_state_id(CurID),
  compute_all_transitions_if_necessary(CurID),
  findall(Res, (transition(CurID,ActionAsTerm,ID),
               option_dest_info(ID,ActionAsTerm,CurID,Res)),DestinationInfos).

option_dest_info(DestID,_,CurID,Res) :- DestID=CurID,!,Res=skip.
option_dest_info(DestID,_,_,Res) :- not_all_transitions_added(DestID),!,Res=open.
option_dest_info(DestID,_,_,Res) :- invariant_violated(DestID),!,Res=invariant_violated.
option_dest_info(DestID,_,_,Res) :- is_deadlocked(DestID),!,Res=deadlock.
option_dest_info(_DestID,_,_,unknown).

tcltk_get_options(list(Options)) :-
  current_state_id(CurID), debug_println(9,start_tcltk_get_options(CurID)),
  on_exception(user_interrupt_signal,
     tcltk_compute_options(CurID,TransSpecs),
     (assert_time_out_for_node(CurID,'unspecified_operation',user_interrupt_signal),
      tcltk_get_computed_options(CurID,TransSpecs))
     ),
  extract_actions(TransSpecs,ActionsAndIDs,Actions),
  set_current_options(ActionsAndIDs),
  %print(Actions),nl,
  debug_println(9,finished_tcltk_get_options(CurID)),
  Actions = Options.

:- use_module(tools_timeout,[time_out_call/2]).
:- use_module(translate,[translate_event_with_target_id/3]).
extract_actions([],[],[]).
extract_actions([(ActionId,Term,Dst)|T],[(ActionId,Str,Term,Dst)|Srest],[Str|ET]) :-
	time_out_call(translate_event_with_target_id(Term,Dst,Str),
				  Str='**TIMEOUT**'),
	extract_actions(T,Srest,ET).


compute_state_information_if_necessary(CurID) :-
  compute_all_transitions_if_necessary(CurID,true).

tcltk_compute_options(CurID,ActionsAndIDs) :-
  compute_state_information_if_necessary(CurID),
  %print(found_transitions(CurID)),nl,
  tcltk_get_computed_options(CurID,ActionsAndIDs).

tcltk_get_computed_options(CurID,ActionsAndIDs) :-
  findall((ActId,ActionAsTerm,DestID),
	      transition(CurID,ActionAsTerm,ActId,DestID),
	                   % TO DO: sort or group, especially if we randomize transition/3
	      ActionsAndIDs).
  

tcltk_get_status(INVVIOLATED,MAXREACHED,TIMEOUT) :- debug_println(9,start_tcltk_get_status),
  current_state_id(CurID),
  (time_out_for_invariant(CurID) -> INVVIOLATED = 2
    ; invariant_violated(CurID)  -> INVVIOLATED = 1
    ; not_invariant_checked(CurID) -> INVVIOLATED = 3
    ; otherwise -> INVVIOLATED = 0
   ),
  (max_reached_for_node(CurID) -> MAXREACHED = 1 ; MAXREACHED = 0),
  (time_out_for_node(CurID,_,time_out) -> TIMEOUT = 1 
   ; time_out_for_node(CurID,_,user_interrupt_signal) -> TIMEOUT = 3 % CTRL-C
   ; time_out_for_node(CurID,_,virtual_time_out(kodkod_timeout)) -> TIMEOUT = 1 % KODKOD; regular time-out
   ; time_out_for_node(CurID,_,_) -> TIMEOUT = 2 % virtual_time_out
   ; TIMEOUT = 0),
  debug_println(9,got_status(INVVIOLATED,MAXREACHED,TIMEOUT)).


%%fd_copy_term(X,Y,true) :- copy_term(X,Y).

    

/* --------------------------------------------------------------------- */

tcltk_get_history(list(StringHistory)) :- debug_println(9,start_tcltk_get_trace),
   get_action_trace(History),
   convert_trace_to_list_of_strings(History,StringHistory).

convert_trace_to_list_of_strings([],[]).
convert_trace_to_list_of_strings([action(AS,_)|T],[AS|CT]) :- !, 
      %truncate_atom(AS,1000,TruncatedAS), % especially SETUP_CONSTANTS, INITIALISATION can be very large; leading to performance issues
      % get_action_trace now truncates
      convert_trace_to_list_of_strings(T,CT).
convert_trace_to_list_of_strings([H|T],[H|CT]) :- convert_trace_to_list_of_strings(T,CT).


/* --------------------------------------------------------------------- */

:- dynamic cached_state_as_strings/3.

tcltk_get_state(list(List)) :- debug_println(9,start_tcltk_get_state),
  current_expression(CurID,CurState),
  compute_state_information_if_necessary(CurID),
  (time_out_for_invariant(CurID) -> List = [invariant_time_out|StateAsStrings]
    ; invariant_violated(CurID) -> List = [invariant_violated|StateAsStrings]
    ; invariant_not_yet_checked(CurID) -> List = [invariant_not_checked|StateAsStrings]
    ; CurID=root ->  List = StateAsStrings
    ; CurState = concrete_constants(_) -> List = StateAsStrings
    ; specfile:b_or_z_mode -> List=[invariant_ok|StateAsStrings] % only in b_or_z_mode do we add not_invariant_checked facts and check invariant while model checking/animating
    ; List=StateAsStrings),
  (cached_state_as_strings(CurID,CurState,StateAsStrings)
   -> true
   ;  retractall(cached_state_as_strings(_,_,_)),
      time_out_call(tcltk_get_state(StateAsStrings,CurState),
                     StateAsStrings = ['***TIMEOUT-PROPERTY***']),
      assert(cached_state_as_strings(CurID,CurState,StateAsStrings))
  ),
  debug_println(9,finished_tcltk_get_state).
  

tcltk_get_state(StateAsStrings,State) :-
  %print(get_state),nl,
  findall(Prop,property(State,Prop),StateProps),
  %print(translating(StateProps)),nl,
  ~~mnf(tcltk_get_state,translate:translate_properties_with_limit(StateProps,StateAsStrings)).
  %print(translated(StateAsStrings)),nl.

:- dynamic current_state_errors/1.

tcltk_get_state_errors(list(ErrorsAsStrings)) :-
  current_state_id(CurID),
  findall(Error,( state_error(CurID,_,Error), Error \== invariant_violated ), Errors),
  ~~mnf(tcltk_get_state,translate:translate_state_errors(Errors,ErrorsAsStrings)),
  retractall(current_state_errors(_)),
  assert(current_state_errors(Errors)).

:- dynamic last_detailed_error/2.
tcltk_get_detailed_state_error(Num,ErrorMsg,Srow,Scol,Erow,Ecol) :-
  retractall(last_detailed_error(_,_)),
  current_state_errors(Errors),
  nth0(Num,Errors,Error),
  !,  %print(explaining(Error)),nl,
  explain_state_error(Error,Span,ErrorMsg), %print(span(Span)),nl,
  assert(last_detailed_error(Span,ErrorMsg)), % save info, so that we can debug if Span contains a span_predicate
  %(generate_dot_from_last_span_predicate('~/Desktop/err.dot') -> true ; true),
  (extract_line_col_for_main_file(Span,Srow,Scol,Erow,Ecol)
    -> true
    ;  Srow= -1, Scol= -1, Erow= -2, Ecol= -2).

can_generate_dot_from_last_state_error :- last_detailed_error(span_predicate(_,_,_),_).
generate_dot_from_last_span_predicate(FileName) :- 
  last_detailed_error(span_predicate(P,LS,State),_ErrorMsg), % also works with expressions
  write_dot_file_for_pred_expr_and_state(P, LS, State,FileName).
  
tcltk_current_state_invariant_violated :-
  current_state_id(CurID),
  invariant_violated(CurID).
  
  

/* --------------------------------------------------------------------- */
 
:- use_module(tools_printing, [print_error/1]).
tcltk_save_history_as_trace_file(File) :-
   print_message('saving history to: '), print_message(File),
   (get_action_term_trace(TrA),reverse(TrA,Tr)
   ;
   find_trace_to_current_node(Tr)),!,
   tcltk_save_history_as_trace_file(Tr,File).
tcltk_save_history_as_trace_file(_File) :-
   print_error('### Cannot produce trace to current state !'),
   fail.
   
tcltk_save_history_as_trace_file(Tr,File) :-
   tell(File),
   (specfile:b_or_z_mode
    -> b_machine_name(OurName)
     ; specfile:currently_opened_file(OurName)
   ),
   print_quoted(machine(OurName)), print('.'),nl,
   member(O,Tr),
   print_quoted(O), print('.'),nl,
   fail.
tcltk_save_history_as_trace_file(_,_) :-
   nl, told, 
   print_message(done).   

      
/* --------------------------------------------------------------------- */

:- use_module(translate,    [get_bexpression_column_template/4]).
:- meta_predicate call_pred_on_expanded_state(3,-,-,-).
:- meta_predicate map_over_history(3,-).

evaluate_bexpr_over_state(Stream,Expr,ValueTemplate,Columns,ResultColStrings,BState,OpName) :- 
    format(Stream,'~w,',[OpName]),
    b_compute_expression_with_prob_ids(Expr,BState,Value),
    copy_term((ValueTemplate,Columns),(Value,ResultCols)),
    my_print_csv_value(ResultCols,Stream),
    maplist(translate:translate_bvalue,ResultCols,ResultColStrings),
    nl(Stream).

% evaluate an expression over the history and save the result to a CSV file
evaluate_expression_over_history_to_csv_file(Expr,File) :-
   atom_codes(Expr,VorECodes),
   bmachine:b_parse_machine_expression_from_codes_with_prob_ids(VorECodes,TypedExpr),
   get_bexpression_column_template(TypedExpr,ValueTemplate,ColHeaders,Columns),
   format('Writing CSV file: ~w~n',[File]),
   open(File,write,Stream),
   format(Stream,'OPERATION,',[]),
   my_print_csv_atoms(ColHeaders,Stream),
   call_cleanup(map_over_history(evaluate_bexpr_over_state(Stream,TypedExpr,ValueTemplate,Columns),_),
                close(Stream)),
   format('Finished writing CSV file: ~w~n',[File]).
   
tcltk_evaluate_expression_over_history(Expr,list([list(['OPERATION','State Id'|ColHeaders])|ColumnLists])) :-
   atom_codes(Expr,VorECodes),
   bmachine:b_parse_machine_expression_from_codes_with_prob_ids(VorECodes,TypedExpr),
   get_bexpression_column_template(TypedExpr,ValueTemplate,ColHeaders,Columns),
   map_over_history(evaluate_bexpr_over_state(user_output,TypedExpr,ValueTemplate,Columns),
                    ColumnLists).

map_over_history(Pred,ResColumnLists) :-
   history(H), current_state_id(CurID),
   reverse([CurID|H],RH),
   state_space:trace(T), reverse(T,RT),
   maplist(call_pred_on_expanded_state(Pred),RH,[action(root,root)|RT],ColumnLists),
   prune_list(ColumnLists,ResColumnLists).
% remove unknown elements at front for Tk
prune_list([unknown|T],Res) :- !, prune_list(T,Res).
prune_list(R,R).

call_pred_on_expanded_state(_Pred,root,_,ColList) :- !, ColList=unknown.
call_pred_on_expanded_state(Pred,StateId,action(_,ActionTerm),ResColList) :- print(StateId),nl,
   specfile:get_operation_name(ActionTerm,OpName),
   format('Processing state ~w (reached via ~w)~n',[StateId,OpName]),
   visited_expression(StateId,State),
   (state_corresponds_to_initialised_b_machine(State,BState)
     -> call(Pred,ColList,BState,OpName), ResColList = list([OpName,StateId|ColList])
      ; ResColList = unknown ).

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
b_compute_expression_with_prob_ids(Expr,BState,Value) :-
    add_prob_deferred_set_elements_to_store(BState,BState1,visible),
    b_interpreter:b_compute_expression(Expr,[],BState1,Value).


my_print_csv_atoms([A,B|T],Stream) :- !, format(Stream,'"~w",',[A]),
    my_print_csv_atoms([B|T],Stream).
my_print_csv_atoms([A],Stream) :-  format(Stream,'"~w"~n',[A]).

my_print_csv_value([A,B|T],Stream) :- !,
    my_print_csv_value([A],Stream),
    write(Stream,','),
    my_print_csv_value([B|T],Stream).
my_print_csv_value([V],Stream) :-   
    (no_quotes_necessary(V) -> true ; write(Stream,'"')), 
    translate:print_bvalue_stream(Stream,V),
    (no_quotes_necessary(V) -> true ; write(Stream,'"')).

no_quotes_necessary(int(_)).
no_quotes_necessary(pred_false).
no_quotes_necessary(pred_true).

    
/* --------------------------------------------------------------------- */

tcltk_execute_until(FormulaAsAtom) :-
    (ltl_tools: temporal_parser(FormulaAsAtom,ltl,Ltl) ->
        true
    ; otherwise ->
        add_error(ltl,'LTL Parser failed: ',FormulaAsAtom),fail),
    ( Ltl=syntax_error ->
        add_error(ltl,'Syntax error while parsing: ',FormulaAsAtom),fail
    ; otherwise ->
        preprocess_formula(Ltl,Ltl2)
    ),
    current_state_id(CurID),
    prob2_interface: find_trace1(CurID,Ltl2,no_loop,1000,Trace,_Result),
    %TODO: check if result is deadlock. Will not fail.
    update_history_from_trace(Trace).

update_history_from_trace([]). % TODO: Can we get an empty trace from find_trace predicate?
update_history_from_trace([op(OpID,_Name,CurID,NextID)]) :-
    add_id_to_history(CurID),
    add_to_op_trace_ids(OpID),
    add_id_to_history(NextID),
    set_id_as_current_state(NextID).
update_history_from_trace([op(OpID,_Name,CurID,NextID)|T]) :-
    add_id_to_history(CurID),
    add_to_op_trace_ids(OpID),
    add_id_to_history(NextID),
    update_history_from_trace(T).

set_id_as_current_state(CurID) :-
    retract(current_state_id(_ID)),
     assert(current_state_id(CurID)).

tcltk_random_perform :- tcltk_random_perform(_).
tcltk_random_perform(ActionId) :- 
  current_options(Options),
  length(Options,Len),
  (Len>0
  ->  L1 is Len+1,
      random(1,L1,RanChoice),
      tcltk_perform_nr_option(RanChoice,Options,ActionId)
  ;   true
  ).
   
 
tcltk_get_line_col(Nr,list(StartLine),list(StartCol),list(EndLine),list(EndCol)) :-  /* just get the location in the source code */
   current_options(Options),
   nth1(Nr,Options,(_Id,_Action,ActionAsTerm,_NewID)),
   %print(get_source_text(ActionAsTerm)),nl,
   time_out_call(get_source_text_positions(ActionAsTerm,StartLine,StartCol,EndLine,EndCol),
     add_error_and_fail(tcltk_get_line_col,'Timeout determining source code position: ',ActionAsTerm)),
   StartLine \= []. % otherwise fail
%   print(startlines(StartLine)),nl,print(startcols(StartCol)),nl,
%   print(endlines(EndLine)),nl, print(endcol(EndCol)),nl.
   
tcltk_is_sync_event(Nr) :-
   current_options(Options),
   nth1(Nr,Options,(_Id,_Action,ActionAsTerm,_NewID)),
   extract_span_from_event(ActionAsTerm,SPAN,_,_), 
   extract_span_info(SPAN,Info),
  % print(extracted_span_info_for_sync(Info,SPAN)),nl,
   member(sharing,Info).
   
tcltk_perform_nr(Nr) :-
   current_options(Options),
   tcltk_perform_nr_option(Nr,Options,_).
  

tcltk_perform_nr_option(Nr,Options,ActionId) :-
   (nth1(Nr,Options,(ActionId,_Action,ActionAsTerm,NewID))
     ->  (forward_history([forward(NewID,ActionId)|_FwdHist])
          -> tcltk_forward
          ;  tcltk_perform_action_term(ActionAsTerm,NewID),
             retractall(forward_history(_))
         )
	 ;   tcltk_backtrack /* interpret as backtrack */
   ).
   

tcltk_can_backtrack :- history([_|_]).
tcltk_backtrack :- \+ tcltk_can_backtrack,!,
               print_message('Cannot backtrack'),fail.
tcltk_backtrack :-
               remove_from_op_trace_ids(LastTraceH),
               retract(history(History)),
               retract(current_state_id(CurID)),!,
	           History= [LastID|EHist],
               assert(history(EHist)),
               %visited_expression(LastID,LastExpr,LastBody),
               assert(current_state_id(LastID)),
               (retract(forward_history(FwdHist)) -> true ; FwdHist=[]),
               assert(forward_history([forward(CurID,LastTraceH)|FwdHist])).
tcltk_backtrack(N) :- N =< 0, !.
tcltk_backtrack(N) :- tcltk_backtrack, N1 is N-1, tcltk_backtrack(N1).

:- dynamic saved_forward_history/1.
:- volatile saved_forward_history/1.

% save certain information to restore when reload is completed:
tcltk_prepare_for_reload :- retractall(saved_forward_history(_)),
   get_action_term_trace(TrA),reverse(TrA,Tr),
   %print(prepare_reloading(Tr)),nl,
   maplist(gen_unknown_forward,Tr,SavedForwardHistory),
   % TO DO: add current forward history as well
   assert(saved_forward_history(SavedForwardHistory)).
   
gen_unknown_forward(Step,forward('$UNKNOWN',Step)).

% try reloading certain information from previous load of same model
tcltk_try_reload :- 
   retract(saved_forward_history(SavedForwardHistory)),!,
    % TO DO: also retractall in other places
   %print(reloading(SavedForwardHistory)),nl,
   retractall(forward_history(_)),
   assert(forward_history(SavedForwardHistory)).
tcltk_try_reload.

% go forward while possible
tcltk_fast_forward :- tcltk_can_forward,!, tcltk_forward,
    current_state_id(CurID),
    compute_state_information_if_necessary(CurID),
    tcltk_fast_forward.
tcltk_fast_forward.

tcltk_can_forward :- forward_history([_|_]).
tcltk_forward :- \+ tcltk_can_forward,!,
               print_message('Cannot go forward'),fail.
tcltk_forward :-
               retract(forward_history([forward(FwdID,LastTraceH)|FwdHist])),
               % TO DO: check if FwdID is '$UNKNOWN' -> then try to find successor which can be reached via LastTraceH
               retract(history(EHist)),
               retract(current_state_id(CurID)),!,
               go_forward(FwdID,LastTraceH,FwdHist,CurID,EHist).
go_forward('$UNKNOWN',ActionAsTerm,FwdHist,CurID,EHist) :-
               transition(CurID,ActionAsTerm,TransitionID,FwdID),!,
               %print(found(CurID,ActionAsTerm,TransitionID,FwdID)),nl,
               go_forward(FwdID,TransitionID,FwdHist,CurID,EHist).
% TO DO: try and execute by predicat in case max_reached is true !
go_forward('$UNKNOWN',ActionAsTerm,FwdHist,CurID,EHist) :-
               is_initialisation_op(ActionAsTerm),
               % maybe PROPERTIES/INITIALISATION has changed; try and replay with other values
               transition(CurID,ActionAsTerm2,TransitionID,FwdID),
               is_initialisation_op(ActionAsTerm2),
               !,
               %print(found(CurID,ActionAsTerm,TransitionID,FwdID)),nl,
               translate_event(ActionAsTerm,Action),
               format('Could not replay ~w.~nUsing other possible transition instead.~n',[Action]),
               go_forward(FwdID,TransitionID,FwdHist,CurID,EHist).
go_forward('$UNKNOWN',ActionAsTerm,_FwdHist,CurID,EHist) :- !,
               assert(forward_history([])),
               assert(history(EHist)),
               assert(current_state_id(CurID)),
               translate_event(ActionAsTerm,Action),
               add_error(tcltk_forward,'Cannot replay step after reload: ',Action).
go_forward(FwdID,LastTraceH,FwdHist,CurID,EHist) :- 
               assert(forward_history(FwdHist)),
               assert(history([CurID|EHist])),
               add_to_op_trace_ids(LastTraceH),
               assert(current_state_id(FwdID)),
               re_execute_transition_if_necessary(CurID,LastTraceH,FwdID).
               %visited_expression(FwdID,FwdExpr,FwdBody).

is_initialisation_op(Op) :- 
  (functor(Op,'$setup_constants',_) ;functor(Op,'$initialise_machine',_)).

% re-execute operation with side-effects if necessary and if preference set
% TO DO: tie into tcltk_perform so that not only the forward button but also double click
%        in operations pane works
re_execute_transition_if_necessary(FromID,TransitionID,ToID) :-
    b_or_z_mode,
    get_preference(re_execute_operations_with_side_effects,true),
    bmachine_construction:external_procedure_used(_), % TO DO: check if operation itself has side-effects
    !,
    (re_execute_transition(FromID,TransitionID,ToID)
      -> true
      ; add_error(tcltk_forward,'Could not re-execute operation: ',TransitionID)).
re_execute_transition_if_necessary(_,_,_).

re_execute_transition(FromID,TransitionID,ToID) :-
    transition(FromID,ActionTerm,TransitionID,ToID),
    %%print(trans(ActionTerm)),nl,
    specfile:get_operation_name(ActionTerm,OpName),
    debug_println(9,re_executing_operation(OpName,FromID,TransitionID,ToID)),
    visited_expression(FromID,InState),
    specfile:b_trans(InState,OpName,ActionTerm,_NewState,_TransInfo),
    debug_println(9,re_executed(OpName)).    

% a version of tcltk_perform_action_term which can be given only the String (e.g., for trace checking)
tcltk_perform_action_string(ActionString,ActionAsTerm,NewID) :-
    tcltk_goto_state(ActionAsTerm,NewID),
    translate_event_with_target_id(ActionAsTerm,NewID,ActionString).
%    translate_event_with_limit(ActionAsTerm,5000,ActionString). % should use same limit as below; otherwise trace checking will fail; ideally we should check ActionString is prefix
    %translate_event(ActionAsTerm,ActionString).

% perform the Action with term ActionAsTerm leading to new state NewID
tcltk_perform_action_term(ActionAsTerm,NewID) :-  %print(tcltk_perform_action_term(Action,ActionAsTerm,NewID)),nl,
   (var(ActionAsTerm)
     -> add_internal_error('Illegal call: ',tcltk_perform_action_term(ActionAsTerm,NewID)) ; true),
	tcltk_goto_state(ActionAsTerm,NewID).

	/* allows one to execute an action in term form, from the current state: */	  
%tcltk_perform_action(T,Cur) :- print_message(perf(T,Cur)),fail.

tcltk_perform_action(ActionAsTerm,NewID) :-
  current_state_id(CurID),
  transition(CurID,ActionAsTerm,NewID),
  tcltk_goto_state(ActionAsTerm,NewID).
 
 tcltk_no_constants_or_no_inititalisation_found :-
    (current_expression(root,_) ; current_expression(_,concrete_constants(_))),
    current_options([]).

/*  ------------ */
/*  tcltk_open/1 */
/*  ------------ */
  

%prolog_check_load_errors :-
%  (tcltk_find_untyped_consts(ErrRes)
%   -> add_error(prolog_open,'These constants were not given a proper type:', ErrRes)
%   ;  true),
%  (tcltk_find_untyped_vars(ErrRes)
%   -> add_error(prolog_open,'These variables were not given a proper type:', ErrRes)
%   ;  true).
   
tcltk_open_b_file_for_minor_mode(File,Minor) :- tcltk_open_b_file(File),
  set_animation_minor_mode(Minor).

tcltk_open_b_file(File) :- print(tcltk_open_b_file(File)),nl,
  tcltk_clear_machine,
  set_animation_mode(b),
  print_message(loading(File)),
  (bmachine:b_load_machine_from_file(File)
    -> print_message(loaded),
       set_currently_opened_file(File)
    ;  set_failed_to_open_file(File),
       fail
  ).

:- use_module(xtl_interface,[open_xtl_file/1]).
tcltk_open_xtl_file(File) :- 
  tcltk_clear_machine,
  set_animation_mode(xtl),
  xtl_interface:open_xtl_file(File), set_currently_opened_file(File).

  
tcltk_open_cspm_file(File) :- 
  tcltk_clear_machine,
  set_animation_mode(cspm),
  xtl_interface:open_cspm_file(File), set_currently_opened_file(File).
  
:- prob_use_module(xtl_interface).
tcltk_open_promela_file(File) :- 
  tcltk_clear_machine,
  set_animation_mode(promela), 
  xtl_interface:open_promela_file(File), set_currently_opened_file(File).

% disabled because Java parser seems broken  
%tcltk_open_smv_file(File) :- 
%  tcltk_clear_machine,
%  set_animation_mode(smv), 
%  xtl_interface:open_smv_file(File), set_currently_opened_file(File).

:- dynamic currently_opened_cspm_file_in_csp_and_b_mode/1.
:- use_module(specfile,[type_check_csp_and_b/0]).
tcltk_add_csp_file(File) :- 
  unset_animation_minor_modes(L), % we could be in z or tla or eventb mode
  set_animation_mode(csp_and_b), 
  reset_animation_minor_modes(L),
  retractall(currently_opened_cspm_file_in_csp_and_b_mode(_)),
  xtl_interface:open_cspm_file(File),
  assert(currently_opened_cspm_file_in_csp_and_b_mode(File)),
  notify_change_of_animation_mode,
  specfile:type_check_csp_and_b.

notify_change_of_animation_mode :-
   announce_event(change_of_animation_mode).
   % clear_applicable_flag. % no done via event handling
  

  
tcltk_open_z_file(File) :- 
  open_proz_file(File,BMachine),
  tcltk_clear_machine,
  set_animation_mode(b), set_animation_minor_mode(z),
  b_set_typed_machine(BMachine),
  set_currently_opened_file(File).

:- use_module(eclipse_interface,[load_eventb_file/1]).
tcltk_load_packaged_eventb_file(File) :-
  tcltk_clear_machine,
  load_eventb_file(File).
   
    
/* --------------------------------------------------------------------- */

tcltk_exists_an_open_node :-
  not_all_transitions_added(_) ; not_interesting(_).
tcltk_goto_an_open_node :-
  (not_all_transitions_added(ID) ; not_interesting(ID)),!,
  tcltk_goto_state(find_open_node,ID).
tcltk_goto_max_reached_node :-
  (max_reached_for_node(ID) ; time_out_for_node(ID)),!,
  tcltk_goto_state(find_max_reached_node,ID).

tcltk_goto_an_invariant_violation :-
  invariant_violated(ID),!,
  tcltk_goto_state(find_inv_violation,ID).  

tcltk_goto_node_with_id(ID) :-
  visited_expression_id(ID),
  tcltk_execute_trace_to_node(ID).
/* --------------------------------------------------------------------- */

% TO DO: clean up and move into model_checker module

tcltk_state_space_only_has_root_node :- transition(root,_,ID),
    \+ not_all_transitions_added(ID),!,fail.
tcltk_state_space_only_has_root_node.

tcltk_search_among_existing_nodes(TclTkRes,FindDeadlocks,FindInvViolations,
                                                      FindGoal,FindAssViolations) :-         
   search_among_existing_nodes(Res,FindDeadlocks,FindInvViolations,
                                                      FindGoal,FindAssViolations),
   translate_error_for_tclk(Res,TclTkRes).

% if tcltk_state_space_only_has_root_node is true: then we should call search_among_existing_nodes before doing open_search to ensure that we uncover state_errors in the root node
search_among_existing_nodes(invariant_violation,_FindDeadlocks,FindInvViolations,
                                                      _FindGoal,_FindAssViolations) :-
   ((get_state_space_stats(_,_,PrN), PrN>100)
     -> print_message('Search Among Processed Nodes'), print_message(number_of_nodes(PrN)) ; true),
   FindInvViolations=1,
   (invariant_violated(ResID) -> true
     ; find_invariant_violation_among_not_checked_nodes(ResID)),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(DEADLOCK,FindDeadlocks,_,_,_) :-
   FindDeadlocks=1,
   visited_expression_id(ResID),
   transitions_computed_for_node(ResID),
   is_deadlocked(ResID),
   (time_out_for_node(ResID)
    -> DEADLOCK = possible_deadlock_timeout ; DEADLOCK = deadlock),
   print_message(deadlock(ResID)),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(goal_found,_,_,FindGoal,_) :-
   FindGoal=1,
   %visited_expression(ResID,_CurState),
   %\+(not_all_transitions_added(ResID)),  % if commented out: will also look at open nodes
   test_goal_in_node(ResID),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(assertion_violation,_,_,_,FindAssViolations) :-
   FindAssViolations=1,
   b_machine_has_assertions,
   visited_expression(ResID,CurState),
   transitions_computed_for_node(ResID),
   state_violates_assertions(ResID,CurState),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(ERR,_,_,_,_) :-
   state_space:state_error(ResID,_,Error), Error \== invariant_violated,
   (Error = abort_error(ERR,_,_,_) -> true ; ERR = state_error),
   tcltk_execute_trace_to_node(ResID).
search_among_existing_nodes(all,_,_,_,_).

test_goal_in_node(ResID) :- b_get_machine_goal(Goal), 
  eclipse_interface:test_boolean_expression_in_node(ResID,Goal).

compute_all_transitions_if_necessary :-
  current_state_id(CurID),
  compute_all_transitions_if_necessary(CurID,false).
compute_all_transitions_if_necessary(CurID) :-
  compute_all_transitions_if_necessary(CurID,false).
  
compute_all_transitions_if_necessary(CurID,CheckInvariant) :-
  (retract_open_node(CurID) % will assert not_invariant_checked
    -> set_context_state(CurID),
       % first check invariant, relevant if we want to use specialized_inv proof information
       visited_expression(CurID,State),
       % TO DO: this does not expand concrete_constants: the expansion may be done twice below:
       % once for compute_invariant (corresponds_to_b...) and once for compute_all_transitions (prepare_state_for_specfile_trans)
       prepare_state_for_specfile_trans(State,PreparedState),
       (CheckInvariant==true
        -> compute_invariant_if_necessary_and_requested(CurID,PreparedState) ; true),
       compute_all_transitions(CurID,PreparedState),
       clear_context_state
   ; true).
%compute_all_transitions_if_necessary(CurID,CurState) :-
%  ((not_all_transitions_added(CurID);not_interesting(CurID))
%    -> compute_all_transitions(CurID,CurState) ; true).


compute_invariant_if_necessary_and_requested(ID,State) :-
  enter_new_error_scope(ErrScID), % here we setup a new error scope; see comment for force_check_invariantKO below
  compute_invariant_if_necessary(ID,State),
  exit_error_scope(ErrScID,_ErrOcc). % should we do something with error occured result ?

compute_invariant_if_necessary(ID,State) :-
    animation_mode(MODE),compute_invariant_if_necessary(ID,State,MODE).
compute_invariant_if_necessary(ID,State,MODE) :- 
   ( check_invariantKO(MODE,ID,State) -> true ; true).

check_invariantKO(ID,State) :-
    animation_mode(MODE),check_invariantKO(MODE,ID,State).
check_invariantKO(b,ID,State) :- !, check_invariantKO2(ID,State).
check_invariantKO(csp_and_b,ID,State) :- !, check_invariantKO2(ID,State).
check_invariantKO(MODE,ID,State) :-  % print(check_invariant(MODE,ID,State)),nl,
    MODE \= cspm, MODE \= xtl, MODE \= promela,
    %MODE \= smv, 
    plugins:plugin_check_invariantKO(MODE,ID,State).
    
check_invariantKO2(ID,_) :- invariant_violated(ID),!.
check_invariantKO2(ID,CurState) :- 
    set_invariant_checked(ID), % only succeeds if we haven't checked invariant for ID yet
    % print(check_inv(ID)),nl,
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    ( preferences:preference(do_invariant_checking,false) -> assert(not_invariant_checked(ID))
     ; force_check_invariantKO(ID,CurBState)).

:- meta_predicate catch_clpfd_overflow_call_for_state(-,0,0).

force_check_invariantKO(ID,CurBState) :-
  % Note: this checks the invariant without setting up a new error scope
  % (Reason: in the model checker it would be relatively expensive to set up an error scope for each state
  %  the model checker stops anyway when the first error is found and then exits its error scope)
   catch_clpfd_overflow_call_for_state(ID,
        b_state_violates_invariant(ID,CurBState),true) ->
    %(b_state_violates_invariant(ID,CurBState) ->
        set_invariant_violated(ID).
        % TO DO ?: call state_satisfies_negation_of_invariant if double_evaluation_when_analysing is true ?

% try and find an invariant violation among those nodes that have not yet been checked
find_invariant_violation_among_not_checked_nodes(ID) :-
   retract(not_invariant_checked(ID)), %print(checking_invariant(ID)),nl,
   visited_expression(ID,State),
   state_corresponds_to_initialised_b_machine(State,CurBState),
   force_check_invariantKO(ID,CurBState).


% computing ample sets (ltl.pl)
compute_ample_set(CurID,Trans,POROptions) :-
  catch_clpfd_overflow_call_for_state(CurID,compute_ample_set2(CurID,POROptions),clear_context_state),
  findall(op(Id,ActionAsTerm,To),
    transition(CurID,ActionAsTerm,Id,To),
    Trans).

% consider to rename, compute_ample_set2/5 is defined in ample_sets.pl
compute_ample_set2(CurID,POROptions) :-
  (retract_open_node(CurID)
    -> set_context_state(CurID),
       compute_ample_set3(CurID,POROptions),
       clear_context_state
   ; true).
   
compute_ample_set3(CurID,POROptions) :-
   visited_expression(CurID,CurState),
   CheckStutterCondition=1, % the stutter condition should be checked for LTL model checking
   %ample_sets: enable_graph(EnableGraph), % the enable graph should be already computed (is this always guaranteed?)
   compute_ample_set2(CurID,CurState,CheckStutterCondition,POROptions). % stutter action

% computing ample sets (model_checker.pl)
compute_ample_set(CurID,CurState,FindInvViolations,POROptions) :-
  catch_clpfd_overflow_call_for_state(CurID,compute_ample_set2(CurID,CurState,FindInvViolations,POROptions),clear_context_state).

% preferences:preference(use_clpfd_solver,false)
compute_all_transitions(CurID,CurState) :-
  catch_clpfd_overflow_call_for_state(CurID,compute_all_transitions2(CurID,CurState),clear_context_state).


%catch_clpfd_overflow_call_for_state(_CurID, Call,_ExtraCleanUpCode) :-
%    preferences:preference(use_clpfd_solver,false),!, % Overflow should not occur
%    call(Call).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
catch_clpfd_overflow_call_for_state(CurID, Call,ExtraCleanUpCode) :-
   catch_clpfd_overflow_call2(Call,
             ( store_state_error(CurID,'CLPFD_integer_overflow',_),
               call(ExtraCleanUpCode)) ).


:- use_module(clpfd_interface,[catch_clpfd_overflow_call1/1]).

:- meta_predicate catch_interrupt_assertion_call(0).
:- use_module(extension('user_signal/user_signal'), [get_user_signal_ref/1]).
catch_interrupt_assertion_call(Call) :- 
   on_exception(error(forced_interrupt_error('User has interrupted the current execution'),_),
                Call,
                (process_interrupted_error_message,fail)).

process_interrupted_error_message :- 
    add_error(probcli,'User has interrupted current execution.').

compute_all_transitions2(CurID,CurState) :- /* compute all outgoing transitions for a given node */
   % set_context_state(CurID), % error_context is set for each operation below anyway.
   reset_max_reached,
   prepare_state_for_specfile_trans(CurState,PreparedCurState), % will perform unpacking of constants just once, for example
   ((time_out_preference_disabled ; use_no_timeout(CurID))
     -> add_transitions_fail_loop(CurID,PreparedCurState)
     ;   preferences:preference(time_out,CurTO),
         add_transitions__with_timeout_fail_loop(CurID,PreparedCurState,CurTO),
	     add_partial_transitions(CurID,PreparedCurState)
   ),
   (max_reached -> assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
   (CurID==root -> save_constants(MaxReached) ; add_new_transitions_to_cache(CurID))
   . %,clear_context_state.  
   
add_transitions_fail_loop(CurID,CurState) :-
    specfile_possible_trans_name(CurState,ActionName),
   %print(adding_transition_no_timeout(CurID,ActionName)),nl,
   set_error_context(operation(ActionName,CurID)),
   add_transition(CurID,CurState,ActionName,_NewId), % TO DO: transmit animation_mode value?
   fail.
add_transitions_fail_loop(_,_) :- clear_error_context.

:- use_module(library(timeout),[time_out/3]).
my_timeout_test(1000000) :- print('.'), !, my_timeout_test(0).
my_timeout_test(X) :- X1 is X+1, my_timeout_test(X1).
my_timeout_test :- time_out(my_timeout_test(0),500,TO), nl,
	    print(time_out_result(TO)),nl, TO==time_out.
:- assert_must_succeed(user:my_timeout_test).

add_transitions__with_timeout_fail_loop(CurID,CurState,CurTO) :-  
   enter_new_error_scope(Level),
   call_cleanup(add_transitions__with_timeout_fail_loop2(Level,CurID,CurState,CurTO),
                exit_error_scope(Level,_)).

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(add_transions__with_timeout_fail_loop2/4).

% old profiler
:- use_module(runtime_profiler,[profile_failure_driven_loop/2]).

add_transitions__with_timeout_fail_loop2(Level,CurID,CurState,CurTO) :-    % statistics(runtime,_),
   % TO DO: prepare computing the state: expanding const and vars, unpacking, ...
    specfile_possible_trans_name(CurState,ActionName),
    profile_failure_driven_loop(ActionName,CurID),
    % Note: we enumerate possible transition names so that we can catch the time-out per name !
       %  statistics(runtime,[T1,DeltaT]), print(delta_time(ActionName,DeltaT)),nl, %%
         % first enumerate possible Operations/... to obtain timeout per operation/...
   % TO DO: divide CurTO by number of transitions ???
       %% statistics(runtime,[T1,_]), print(try(ActionName,T1)),nl,%%
      set_error_context(operation(ActionName,CurID)),
      time_out_with_enum_warning_for_findall_in_current_error_scope(Level,add_transition(CurID,CurState,ActionName,_NewId),CurTO,TimeOutRes),
       %% statistics(runtime,[T2,_]), TDiff is T2-T1, print(added_transition(CurID,ActionName,TDiff,TimeOutRes)),nl, %%
   %%  add_transition(CurID,CurState,ActionName) ,
	  ((TimeOutRes==time_out ; nonvar(TimeOutRes),TimeOutRes=virtual_time_out(_), \+ max_reached(ActionName)) % virtual time outs not relevant if we have already stopped earlier anyway; TO DO: does not seem to work for INITIALISATION
	   ->  (TimeOutRes=virtual_time_out(_) -> true
	           ; print_message('TIME OUT: '),print_message(CurID), 
	             print_message(CurTO),print_message(ActionName)
	        ),
		   (var(ActionName)
		     -> assert_time_out_for_node(CurID,'unspecified_operation',TimeOutRes)
		      ; assert_time_out_for_node(CurID,ActionName,TimeOutRes))  % TO DO: assert info per ActionSkeleton ?
		; true
	  ),fail.
add_transitions__with_timeout_fail_loop2(_,_,_,_) :- clear_error_context.
    % statistics(runtime,[_,DeltaT]), print(delta_time(end,DeltaT)),nl.


add_transition(CurID,CurState,ActionName,PreviousSuccessorID) :- 
    %(var(ActionName) -> add_error_fail(add_transition,'ActionName is a variable: ',CurID) ; true),
    nonvar(ActionName), % can happen in CSP, CSP||B, XTL, ... modes  but also in B mode
    candidate_operation_for_reuse(ActionName), % can only be true when try_operation_reuse preference is true
    retract(reuse_operation(CurID,ActionName,PreviousID,_FromOp)),!,
    % we don't have to compute the operation again; simply reuse pre-computed effect
    b_get_operation_normalized_read_write_info(ActionName,_Read,WrittenVariables),
    build_up_b_real_operation(ActionName,ActionTemplate), % missing add any parameters
   %  print(reusing_operation(ActionName,ActionTemplate, from(PreviousID), for(CurID),via(FromOp))),nl,
    transition(PreviousID,ActionTemplate,TransID,PreviousSuccessorID),
    findall(Info,transition_info(TransID,Info),TransInfos),
     visited_expression(PreviousSuccessorID,PreviousExpression,true),
     ~~mnf(patch_state(CurState,WrittenVariables,PreviousExpression,NewExpression)),
%     (specfile_trans(CurState,ActionName,ActionTemplate,NewExpression,_,_)
%      -> true ; add_error(add_transition,'Mismatch in operation reuse',ActionName),
%                print(CurID),nl, print(ActionName), print(' via '),  print(via(PrevOp)),nl),
    ~~mnf(add_trans_id(CurID,ActionTemplate,NewExpression,[],_NewID,TransInfos,_)).  
:- if(debug:global_debug_flag).
add_transition(CurID,CurState,ActionName,NewId) :-
   %debug:watch(30,specfile:specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)),
   %debug:watch(30,user:add_trans_id(CurID,Act,NewExpression,Residue,_NewId,TransInfo,_)).
   debug:timer_call(ActionName,specfile:specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)),
   debug:timer_call(user:add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_)).
:- else.
add_transition(CurID,CurState,ActionName,NewId) :-
  /* ActionName maybe pre-instantiated */
   specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue),
 %% print_bt_message(specfile_trans(CurState,ActionName,Act,NewExpression,TransInfo,Residue)), %%
   add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_).
:- endif.

patch_state(const_and_vars(ConstID,Vars),WrittenVariables,PrevState,
            const_and_vars(ConstID,PatchedVars)) :- !,
   patch_state(Vars,WrittenVariables,PrevState,PatchedVars).
patch_state(expanded_const_and_vars(ConstID,Vars,_FullState),WrittenVariables,PrevState,
            const_and_vars(ConstID,PatchedVars)) :- !,
   patch_state(Vars,WrittenVariables,PrevState,PatchedVars).
   %patch_state(FullState,WrittenVariables,PrevState,PatchedFullState). % do we need to patch both ?
patch_state([],_Write,_Prev,[]) :- !.
patch_state([bind(Variable,Value)|T],WrittenVariables,PrevState,[bind(Variable,NewValue)|PT]) :- !,
    (ord_member(Variable,WrittenVariables)
     ->  lookup_variable(PrevState,Variable,NewValue)
     ; NewValue=Value
    ), patch_state(T,WrittenVariables,PrevState,PT).
patch_state(S,WV,PS,Res) :- add_internal_error('Illegal state: ',patch_state(S,WV,PS,Res)), Res=S.
  
lookup_variable(const_and_vars(_ConstID,Vars),Var,Val) :- !,
    lookup_value_for_existing_id(Var,Vars,Val).
lookup_variable(expanded_const_and_vars(_ConstID,Vars,_FullState),Var,Val) :- !,
    lookup_value_for_existing_id(Var,Vars,Val).
lookup_variable(Store,Var,Val) :- lookup_value_for_existing_id(Var,Store,Val). 
   

add_trans_id(CurID,Act,NewExpression,Residue,NewID,TransInfo,TransId) :-  
  (Residue=[] ->  true ; add_error(add_trans_id,'Residue: ',CurID:Act:Residue)),
  tcltk_add_new_transition_transid(CurID,Act,NewID,NewExpression,TransInfo,TransId).
    

add_partial_transitions(root,CurState) :-
   /* for the moment the only state where partial transitions are computed */
   \+ any_transition(root,_,_), /* adding normal transitions was not successful */
   partial_trans(CurState,Act,NewExpression,_Residue),
   add_trans_id(root,Act,NewExpression,_Residue,_NewID,[],_),fail.
add_partial_transitions(_CurID,_).

/* --------------------------------------------------------------------- */

/* A bit like AO* algorithm: expands open nodes (i.e., states whose transitions
 have not yet been computed */
 
%:- use_module(linda_slave).
:- use_module(model_checker).

tcltk_model_check(Nr,TclTkRes,FindDeadlocks,FindInvViolations,FindGoal,FindAssViolations,StopFullCoverage,PartialOrderReduction,PartialGuardsEvaluation,TIMELIMIT,TotalTime) :-
   statistics(walltime,[CurTime,_]), /* get current time in ms */
   LimitTime is CurTime+TIMELIMIT,
   %debug_println(10,start_tcltk_model_check(Nr,TIMELIMIT)),
   InspectExistingNodes = 0,
   (do_model_check(Nr,_A,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction,PartialGuardsEvaluation,InspectExistingNodes)
        -> translate_error_for_tclk(Res,TclTkRes)
        ; add_internal_error('Call failed: ',do_model_check(Nr,_A,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction,PartialGuardsEvaluation,InspectExistingNodes)),
          TclTkRes=error),
   statistics(walltime,[CurTime2,_]),
   TotalTime is CurTime2-CurTime,
   debug_println(10,tcltk_model_check(Nr,TotalTime,Res)).

translate_error_for_tclk([timeout,Nr],Res) :- !, Res=[timeout,Nr].
translate_error_for_tclk(state_error(S),String) :- !, translate_state_err_for_tcltk(S,String).
translate_error_for_tclk(S,R) :- atomic(S),!,R=S.
translate_error_for_tclk(S,R) :- nonvar(S), add_error(translate_error_for_tclk,'Unknown Model Check result: ',S), R=unknown.

translate_state_err_for_tcltk(abort_error(TYPE,_,_,_),ERR) :- !, ERR = TYPE.
translate_state_err_for_tcltk(eventerror(Event,Error,_),ERR) :- !,
    functor(Error,F,_),
    ajoin(['event_error:',Event,':',F],ERR).
translate_state_err_for_tcltk(Error,ERR) :-
   functor(Error, StateError, _), atom_concat('state_error:',StateError,ERR).
   


do_model_check(_Nr,NodesAnalysed,_LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,_StopFullCoverage,_PartialOrderReduction,_PartialGuardsEvaluation,ForceInspectExistingNodes) :-
    (tcltk_state_space_only_has_root_node -> true % Then also look at the root node
      ; ForceInspectExistingNodes == 1),
    search_among_existing_nodes(Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations), Res \=all, NodesAnalysed=0.
do_model_check(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
               FindAssViolations,StopFullCoverage,PartialOrderReduction,PartialGuardsEvaluation,_) :- 
    update_ass(FindAssViolations,FindAssViolations2),
    open_search(LimitNr,NodesAnalysed,LimitTime,Res,FindDeadlocks,FindInvViolations,FindGoal,
                FindAssViolations2,StopFullCoverage,PartialOrderReduction,PartialGuardsEvaluation).

update_ass(1,FindAssViol) :- \+ b_machine_has_assertions,!, % no need to check assertions
    FindAssViol=0.
update_ass(X,X).

/* --------------------------------------------------- */

/* constraint-based checking of refinements */
:- use_module(cbc_refinement_checks).
tcltk_cbc_refinement_check(list(List),ErrorOccured) :-
    cbc_refinement_check(List,ErrorOccured).

/* constraint-based/model search for invariant violations */

:- dynamic cbc_inv_error_found/2.
tcltk_constraint_based_check(L) :- tcltk_constraint_based_check(L,_).
tcltk_constraint_based_check(list(List),ErrorsOrTimeoutWereFound) :- 
   retractall(cbc_inv_error_found(_,_)),
   findall(Res, (tcltk_constraint_based_check_with_timeout(OpName,ErrorDesc),
                   (ErrorDesc=ok -> true ; assert(cbc_inv_error_found(OpName,ErrorDesc))),
                   string_concatenate(' == ',ErrorDesc,Res1),
                   string_concatenate(OpName,Res1,Res)), List),
   %%print(list(List)),nl, %%
   (cbc_inv_error_found(_,_) -> 
      ((cbc_inv_error_found(_,X),X\=time_out)
        -> ErrorsOrTimeoutWereFound=true ; ErrorsOrTimeoutWereFound=time_out
      )
      ; ErrorsOrTimeoutWereFound = false).

% can also be called with uninstantiated OpName
tcltk_constraint_based_check_with_timeout(OpName,ErrorDesc) :-
   preferences:get_computed_preference(debug_time_out,DebugTimeOut),
   b_get_machine_operation(OpName,_,_,_),
   (time_out_with_enum_warning_one_solution(tcltk_constraint_based_check_op(OpName,ErrorDesc),DebugTimeOut,TimeOutRes) 
    -> (is_time_out_result(TimeOutRes) 
		   ->  print_message('TIME OUT: '), print_message(OpName), print_message(DebugTimeOut),
			   ErrorDesc = timeout
			; true
	)).
tcltk_constraint_based_check_op(OpName,ErrorDesc) :-
  catch_clpfd_overflow_call1(state_model_check(OpName,State,Operation,NewState)),!,
  %fd_copy_term(NewState,ResTempl,ResBody), print(ResTempl),print(ResBody),nl,
  %fd_copy_term(State,PriorTempl,PriorBody),
  ~~mnf(tcltk_add_new_transition_transid(root,'$JUMP'(constraint_based_check),PriorID,State,[],TransId1)),
  ~~mnf(translate_event(Operation,Action)),
  ~~mnf(tcltk_add_new_transition_transid(PriorID,Operation,NewID,NewState,[],TransId2)),
  retractall(current_state_id(_)),
  assert(current_state_id(NewID)),
  reset_op_trace_ids,
  add_to_op_trace_ids(TransId1), add_to_op_trace_ids(TransId2),
  retractall(history(_)),
  retractall(forward_history(_)),
  assert(history([PriorID,root])), print(State),nl,
  ajoin(['invariant_violated_after(',Action,')'],ErrorDesc).
tcltk_constraint_based_check_op(_Depth,'ok').


% try to find constants, parameters that make a given trace feasible
tcltk_cbc_find_trace(TraceToBeFound,Res) :-
   tcltk_cbc_find_trace(TraceToBeFound,'',first_solution,Res).
tcltk_cbc_find_trace(TraceToBeFound,TargetPredString,FindAll,Res) :-
    b_parse_optional_machine_predicate(TargetPredString,TargetPred),
    print(looking_for_solution_for_trace(TraceToBeFound)),nl,
    split_atom(TraceToBeFound,[',',';',' '],TraceAsList),
    (maplist(check_valid_operation,TraceAsList,AdaptedTraceAsList)
     -> start_ms_timer(Timer),
        tcltk_cbc_find_trace_aux(Timer,AdaptedTraceAsList,TargetPred,FindAll,Res)
      ; Res = error).
tcltk_cbc_find_trace_aux(Timer,TraceAsList,TargetPred,FindAll,Res) :-
    preferences:get_computed_preference(debug_time_out,DebugTimeOut),
    reset_nr_cbc_sols,
    time_out_with_enum_warning_one_solution(sap:create_testcase_path_nondet(init,TraceAsList,TargetPred,Transitions),DebugTimeOut,TimeOutRes),
    stop_ms_timer(Timer),
    (is_time_out_result(TimeOutRes)
      -> !,
         print_message('*** TIME OUT: '), print_message(DebugTimeOut), Res = time_out
      ; FindAll = findall -> inc_nr_cbc_sols,
          print('% *** FOUND PATH: '), maplist(print_trans,Transitions),nl,
          fail % backtrack to find further solutions
      ;  !,Res = ok,
          print('% *** FOUND PATH: '), maplist(print_trans,Transitions),nl,
          maplist(extract_trans_id,Transitions,TransIds),
          set_trace_by_transition_ids(TransIds)
    ).
tcltk_cbc_find_trace_aux(Timer,_TraceToBeFound,_TargetPred,findall,R) :- !, stop_ms_timer(Timer),
    R=nr_cbc_sols(Res),
    nr_cbc_sols(Res),
    print_message('*** ALL SOLUTIONS FOUND ').
tcltk_cbc_find_trace_aux(Timer,_TraceToBeFound,_TargetPred,_FindAll,no_solution_found) :-
    stop_ms_timer(Timer),
    print_message('*** NO SOLUTION FOUND'),nl.

:- dynamic nr_cbc_sols/1.
nr_cbc_sols(0).
reset_nr_cbc_sols :- retractall(nr_cbc_sols(_)), assert(nr_cbc_sols(0)).
inc_nr_cbc_sols :- 
   retract(nr_cbc_sols(N)), N1 is N+1, assert(nr_cbc_sols(N1)).
extract_trans_id( (TransId,_,_,_), TransId).
print_trans( (_,Event,_,_)) :- translate_event(Event,S), print(S), print(' ').
check_valid_operation(Op,Res) :-
   (get_possible_language_specific_top_level_event(Op,InternalName) -> Res=InternalName
    ; add_error(cbc_find_sequence,'Unknown operation/event: ',Op),fail).

   

% ------------


tcltk_constraint_find_valid_state_with_goal_pred :-
   ( b_get_machine_goal(Goal) -> true
   ; otherwise                -> create_texpr(truth,pred,[],Goal)),
   catch_clpfd_overflow_call1(b_set_up_valid_state_with_pred(State,Goal)), % TO DO: does not yet separate constants and variables !
   tcltk_add_cbc_state(State,find_valid_state).
tcltk_constraint_find_valid_state :-
   catch_clpfd_overflow_call1(b_set_up_valid_state(State)), % TO DO: does not yet separate constants and variables !
   tcltk_add_cbc_state(State,find_valid_state).
 
tcltk_constraint_find_deadlock_state(Res) :-
   check_we_are_in_b_mode(cbc_deadlock_freedom_check),
   create_texpr(truth,pred,[],True),
   catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,True,0)), % TO DO: does not yet separate constants and variables !
   (State == time_out -> Res = time_out
    ; tcltk_add_cbc_state(State,cbc_deadlock), Res=deadlock).
tcltk_constraint_find_deadlock_state_with_goal(Filter,Res) :-
   check_we_are_in_b_mode(cbc_deadlock_freedom_check),
   ( b_get_machine_goal(Goal) -> true
   ; otherwise                -> create_texpr(truth,pred,[],Goal)),
   catch_clpfd_overflow_call1(cbc_deadlock_freedom_check(State,Goal,Filter)), % TO DO: does not yet separate constants and variables !
   (State == time_out -> Res = time_out
    ; tcltk_add_cbc_state(State,cbc_deadlock), Res=deadlock).

check_we_are_in_b_mode(_) :- b_or_z_mode,!.
check_we_are_in_b_mode(Cmd) :- add_error(check_we_are_in_b_mode,'This command can only be performed in B,Z or TLA+ mode: ',Cmd),fail.
 
tcltk_constraint_find_deadlock_state_with_goal(true,_Filter,Res) :- 
  !,tcltk_constraint_find_deadlock_state(Res).
tcltk_constraint_find_deadlock_state_with_goal(GoalPred,Filter,Res) :-
  bmachine:b_set_machine_goal(GoalPred),
  tcltk_constraint_find_deadlock_state_with_goal(Filter,Res).

tcltk_add_cbc_state(State,OpName) :- 
  ~~mnf(tcltk_add_new_transition_transid(root,'$JUMP'(OpName),NewID,State,[],TransId)),
  retractall(current_state_id(_)),
  assert(current_state_id(NewID)),
  reset_op_trace_ids,
  add_to_op_trace_ids(TransId), 
  retractall(history(_)),
  retractall(forward_history(_)),
  assert(history([root])). %, print(State),nl.

tcltk_constraint_find_static_assertion_violation(TclResult) :-
   cbc_constraint_find_static_assertion_violation(Result,[]),
   functor(Result,TclResult,_).
cbc_constraint_find_static_assertion_violation(Result,Options) :-
    catch_clpfd_overflow_call1(cbc_static_assertions_check(State,Options)),
    ( State = counterexample_found(CE) ->
        Result = counterexample_found,
        tcltk_add_cbc_state(CE,static_assertion_check)
    ; otherwise ->
        Result = State).
        
/* --------------------------------------------------- */
  

:- use_module(error_manager,[extract_all_line_col/5]).
:- use_module(probcspsrc(haskell_csp),[extract_span_from_event/4,extract_span_info/2]).

get_source_text_positions(Event,Line,Col,EndLine,EndCol) :- 
      extract_span_from_event(Event,SPAN,_,_),  
      extract_all_line_col(SPAN,Line,Col,EndLine,EndCol).


/* -------------------------------------------------------------------- */

get_current_option_eventtrace(OptionNum,Trace) :-
    % the option number starts with 0 (coming from a TK listbox)
    current_options(ActionsAndIDs),
    nth0(OptionNum,ActionsAndIDs,(Id,_,_,_)),
    transition_info(Id,eventtrace(Trace)),!.

tcltk_has_eventtrace(OptionNum) :-
    get_current_option_eventtrace(OptionNum,_Trace).

tcltk_show_eventtrace(OptionNum,Desc) :-
    get_current_option_eventtrace(OptionNum,Trace),
    explain_event_trace(Trace,Desc).

    
    
/* --------------------------------------------------- */

   
tcltk_find_shortest_trace_to_current_node(list(StringList)) :-
   find_trace_to_current_node(OpL),
   translate_events(OpL,StringList).
   
tcltk_execute_trace_to_current_node :-
   current_state_id(CurID),
   tcltk_execute_trace_to_node(CurID).
   
tcltk_execute_trace_to_node(ToID) :- /* can be useful for TestCase Generation */
  set_target_id(ToID),
  tcltk_execute_trace_from_to_found_predicate(root,_,_).
 
tcltk_execute_trace_from_to_found_predicate(From,OpL,IDList) :-
     print(finding_trace_from_to(From)),nl, flush_output(user),
   find_shortest_path_from_to_found_predicate(From,OpL,IDList,ToID), !,
       /* mal: replaced by shortest 5/12/2006 */
   (current_state_id(From) -> true ; state_space_reset),
     debug_println(9,executing_trace(From,OpL,ToID)), flush_output(user),
   execute_id_trace_from_current(ToID,OpL,IDList).
 %  print(executing_trace_to(ID,OpL,IDList)),nl,
   %compute_history(OpL,IDList,History), print(new_history(History)),nl,


find_trace_to_current_node(Trace) :-
   current_state_id(CurID),
   find_shortest_path_from_to(root,CurID,L,_),
   extract_term_trace_from_transition_ids(L,Trace).
   
:- use_module(state_space_dijkstra).
 
find_shortest_trace_to_node(FromID,ToID,ActionTermList,IDList) :-
   find_shortest_path_from_to(FromID,ToID,ActionTermList,IDList).

    
 /* --------------------------------------------------- */ 

    

tcltk_goto_a_non_deterministic_node :- /* same transition leads to different resulting state */
    transition(ID,X,ID2),
    transition(ID,X,ID3), dif_ID(ID2,ID3),
    print_message(found_non_det_op(X)),
    %tcltk_goto_state(find_non_det_node,ID).
    tcltk_execute_trace_to_node(ID).

dif_ID(ID1,ID2) :- ((number(ID1),number(ID2)) -> ID2>ID1 ; ID1 \= ID2).

tcltk_goto_a_branching_node :- /* two different events/operations enabled */
    transition(ID,X,ID2), get_operation_name(X,NX),
    transition(ID,Y,ID3), dif_ID(ID2,ID3), get_operation_name(Y,NY), NX\=NY,
    print_message(found_branching_op(ID,NX,NY)),
    %tcltk_goto_state(find_branching_node,ID),
    tcltk_execute_trace_to_node(ID).

tcltk_goto_a_non_resetable_node :-
    find_resetable_nodes,
    visited_expression_id(ID),
    \+(resetable_node(ID)),
    \+(not_all_transitions_added(ID)), !,
    %tcltk_goto_state(find_non_resetable_node,ID).
    tcltk_execute_trace_to_node(ID).
    
operation_name_covered(OpName) :- operation_name_covered(OpName,_).
operation_name_covered(OpName,Template) :- b_get_machine_operation(OpName,Results,Par,_),
   \+ operation_not_yet_covered(OpName),
   length(Par,Arity),
   functor(Op,OpName,Arity),
   (Results\=[] -> Template = '-->'(Op,_) ; Template = Op).

tcltk_goto_event_list_property_violation(Property,EventSelection,list(OperationList)) :-
    % first translate Tk EventSelection Number list into Events
    sap:get_selected_events(EventSelection,_AllEvents,OperationList),
    include(operation_name_covered,OperationList,_,EnabledOpList), % only keep operations that can be enabled anyway
    maplist(operation_name_covered,EnabledOpList,TemplateList), % replace by template
    print(find_property_violation_for_template(Property,TemplateList)),nl,
    visited_expression(ID,S), ID \= root, S\= concrete_constants(_),
    \+(not_all_transitions_added(ID)),
    \+( check_event_list_property(Property,ID,TemplateList) ),
    print(found_event_list_property_violation(Property,ID)),nl,
    tcltk_execute_trace_to_node(ID).

check_event_list_property(relative_deadlock_freedom,ID,TemplateList) :-
    relative_deadlock_freedom(ID,TemplateList).
check_event_list_property(valid_controller,ID,TemplateList) :-
    is_valid_controller_state(ID,TemplateList).
check_event_list_property(det_controller,ID,TemplateList) :-
    is_det_controller_state(ID,TemplateList).

relative_deadlock_freedom(ID,TemplateList) :- member(Template,TemplateList),
    transition(ID,Template,_).
is_valid_controller_state(ID,TemplateList) :- % check that exactly one event from TemplateList is enabled
    append(_,[LiveEvent|RestTemplate],TemplateList),
    transition(ID,LiveEvent,_),!, % we found one enabled event
    \+ relative_deadlock_freedom(ID,RestTemplate). % all other events from list must be disabled
is_det_controller_state(ID,TemplateList) :- % check that *at most* one event from TemplateList is enabled
    append(_,[LiveEvent|RestTemplate],TemplateList),
    transition(ID,LiveEvent,_),!, % we found one enabled event from list
    \+ relative_deadlock_freedom(ID,RestTemplate). % all other events from list must be disabled
is_det_controller_state(_,_) :- true. % no event enabled is also ok

tcltk_goto_state_enabling_operation(TclOpName,FromCurrent) :-
   predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
   operation_name_covered(OpName,Template),
   set_found_predicate(ID, transition(ID,Template,_)),
   print(template(Template,FromCurrent)),nl,
   print_message('Computing trace to node'),
    (FromCurrent=1
      -> current_state_id(FromID) % inefficient: better to do one dijkstra from FromID !
      ; FromID = root
    ),
    tcltk_execute_trace_from_to_found_predicate(FromID,_,_).
%    tcltk_goto_state(find_enabled(OpName),ID).

% quickly compute which operations have already been covered and which ones not
tcltk_operations_covered_info(list([list(['Operation','Covered'])|OpCov]),Precise) :-
    findall(list([OpName,Cov]),
             (specfile:get_possible_event(OpName),
              (operation_name_not_yet_covered(OpName)
               -> not_yet_covered_info(Precise,OpName,Cov)
                ; Cov = 'yes')),OpCov).

not_yet_covered_info(precise,OpName,Cov) :-
    enabling_analysis:feasible_operation(OpName,Res),!, Res=Cov.
not_yet_covered_info(precise,_,R) :- !, R='unknown'.
not_yet_covered_info(_,_,'uncovered').
   
:- dynamic resetable_node/1.  

resetable_node(a).

:- dynamic find_flag/2.
find_resetable_nodes :-
    retractall(find_flag(_,_)),
    retractall(resetable_node(_)),
    any_transition(root,_,ID), /* add nodes reachable from root */
    assert(find_flag(ID,0)),
    fail.
find_resetable_nodes :-
    visited_expression(CID,concrete_constants(_C)),
    any_transition(CID,_,ID), /* add nodes reachable after a setup_constants */
    assert(find_flag(ID,0)),
    fail.
find_resetable_nodes :- find_resetable_nodes_loop,
    print_message('Not resetable:'),
    visited_expression_id(ID),
    \+(resetable_node(ID)),
    print_message(ID),
    fail.
find_resetable_nodes :-
    print_message('Done').
   
    
find_resetable_nodes_loop :-
    retract(find_flag(ID,N)), !, find_resetable_nodes_loop(ID,N).
find_resetable_nodes_loop.    

find_resetable_nodes_loop(ID,N) :- N1 is N+1,
    assert(resetable_node(ID)),
    any_transition(NewID,_,ID),
    \+(resetable_node(NewID)),
    \+(find_flag(NewID,_)),
    assert(find_flag(NewID,N1)),
    fail.
find_resetable_nodes_loop(_,_) :- find_resetable_nodes_loop.



tcltk_show_expression_as_dot(_VorE,_File) :-
    \+ current_state_corresponds_to_fully_setup_b_machine,!,
    add_error(tcltk_show_expression_as_dot,'Please initialise machine first: '), % TO DO:
    fail.
tcltk_show_expression_as_dot(VorE,File) :-
    atom_codes(VorE,VorECodes),
    print('Type checking'),nl,
    bmachine:b_parse_machine_expression_from_codes_with_prob_ids(VorECodes,TypedExpr),
    current_expression(_CurID,State),
    % state_corresponds_to_initialised_b_machine(State,BState),
    state_corresponds_to_set_up_constants(State,BState), % TO DO: either check if no variables referenced or catch errors below
    decompose_expression_for_dot(TypedExpr,ExprList,[]), % decompose pairs so that user can provide multiple values/graphs
    compute_for_dot(ExprList,BState,GraphState), % compute value for each sub-component
    format('Writing value to DOT file: ~w~n',[File]),
    graph_canon:print_cstate_graph(GraphState,File),
    format('Finished writing value to DOT file: ~w~n',[File]).
tcltk_show_expression_as_dot(_VorECodes,File) :-
    add_error(tcltk_show_expression_as_dot,'Computing Value Dot Representation Failed for: ',File),
    fail.

compute_for_dot([],_,[]).
compute_for_dot([String,TypedExpr|Tail],BState,[bind(Str,Value)|TS]) :-
    % allow user to specify labels for graphs ("lbl1", e1, "lbl2", e2, ...)
    get_texpr_expr(String,string(Str)),!,
    translate:translate_bexpression_with_limit(TypedExpr,100,EStr),
    format('Computing value for expression: ~w~n',[EStr]),
    b_compute_expression_with_prob_ids(TypedExpr,BState,Value),
    compute_for_dot(Tail,BState,TS).
compute_for_dot([TypedExpr|Tail],BState,[bind(Str,Value)|TS]) :-
    translate:translate_bexpression_with_limit(TypedExpr,100,Str),
    format('Computing value for expression: ~w~n',[Str]),
    b_compute_expression_with_prob_ids(TypedExpr,BState,Value),
    compute_for_dot(Tail,BState,TS).
  


decompose_expression_for_dot(b(couple(A,B),_,_)) --> !,
     decompose_expression_for_dot(A), decompose_expression_for_dot(B).
decompose_expression_for_dot(X) --> [X].
    
/* ------------------------------------------------------------------ */

:- ensure_loaded(state_viewer_images).
 
/* ------------------------------------------------------------------ */



print_property_partitions :- print('PARTITIONS OF PROPERTIES'),nl,
   b_interpreter:b_get_properties_from_machine(Properties),
   bsyntaxtree:predicate_components(Properties,Comp),
   length(Comp,Len), print(Len), print(' partitions found'),nl,
   member(component(P,Vars),Comp), nl,print(Vars),nl,translate:print_bexpr(P),nl,fail.
print_property_partitions :- nl, print(' ============== '),nl.


tcltk_add_user_executed_operation(TclOpName,CustomPredicate) :- 
    predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
    %print(tcltk_add_user_executed_operation(OpName,CustomPredicate)),nl,
    bmachine:set_temp_predicate(CustomPredicate,OpName),
    %print(exec(OpName)),nl,
    call_cleanup(b_trans_one_solution(OpName,_),
                  bmachine:reset_temp_predicate).

% just as above but with typed predicate
tcltk_add_user_executed_operation_typed(OpName,TypedCustomPredicate,NewID) :- 
    bmachine:assert_temp_typed_predicate(TypedCustomPredicate),
    call_cleanup(b_trans_one_solution(OpName,NewID),
                  bmachine:reset_temp_predicate).
    
% find one solution from current state using OpName and respecting temp_predicate
b_trans_one_solution(OpName,NewID) :-
     current_expression(CurID,InState),
     set_context_state(CurID),
     temporary_set_preference(maxNrOfEnablingsPerOperation,1,Max),
     temporary_set_preference(maxNrOfInitialisations,1,MaxI),
     start_ms_timer(Timer),
     call_cleanup(specfile:b_trans(InState,OpName,Operation,NewState,TransInfo),
                  (stop_ms_timer(Timer),
                   reset_temporary_preference(maxNrOfEnablingsPerOperation,Max),
                   reset_temporary_preference(maxNrOfInitialisations,MaxI),
                   clear_context_state)
       ),
       %print(found_solution(NewState)),nl,
       % now add a transition to the state space:
       Residue=[],
       add_trans_id(CurID,Operation,NewState,Residue,NewID,TransInfo,_),
       tcltk_goto_state(Operation,NewID).

% --------------------------------------------------------------------------------------


:- use_module(bmachine).
:- use_module(specfile).
:- use_module(tools).
:- use_module(library(lists)).
:- use_module(store).
:- use_module(self_check).
:- use_module(b_global_sets).
:- use_module(translate).
:- use_module(predicate_debugger).


/* --------------------------------------------------------- */
/* --------------------------------------------------------- */

        
/* ----------------------------------*/
/*       b_analyse_conjunction       */
/* ----------------------------------*/

/* provide list of possible options for analysis */
tcltk_analyse_option(invariant,'INVARIANT') :- b_or_z_mode.
tcltk_analyse_option(properties,S) :- b_or_z_mode,
    get_specification_description(properties,S).
tcltk_analyse_option(assertions,S) :- (b_or_z_mode ; csp_mode),
    get_specification_description(assertions,S).
tcltk_analyse_option(deadlock,'DEADLOCKED') :- b_or_z_mode.

:- use_module(predicate_evaluator).
/* perform the various analysis */
tcltk_analyse(Mode,Type,ListWithResults,Total,False,Unknown) :-
   (Mode == unicode
    -> set_unicode_mode,
       call_cleanup(tcltk_analyse2(Type,ListWithResults,Summary),unset_unicode_mode)
     ; tcltk_analyse2(Type,ListWithResults,Summary)),
   (member(total/Total,Summary) -> true ; Total=0),
   (member(false/False,Summary) -> true ; False=0),
   (member(unknown/Unknown,Summary) -> true ; Unknown=0).
tcltk_analyse2(invariant,L,Summary) :- !,tcltk_analyse_invariant(L,Summary).
tcltk_analyse2(properties,L,Summary) :- !,tcltk_analyse_properties(L,Summary).
tcltk_analyse2(assertions,L,Summary) :- !,tcltk_analyse_assertions(L,Summary).
tcltk_analyse2(deadlock,L,Summary) :- !,tcltk_analyse_deadlock(L,Summary).
tcltk_analyse2(X,L,Summary) :- add_error(tcltk_analyse, 'Unknown type of formula: ', X),
 L=[], Summary=[].

   
/* --------------------------------- */

tcltk_show_current_state(TransBState) :-
   current_b_expression(B),
   mnf(translate_bstate(B,TransBState)).
        
tcltk_show_typing_of_variables_and_constants(List) :-
      b_get_machine_variables(TypedVarList),
      convert_typed_varlist(TypedVarList,VarList,Types),
      generate_typing_list(VarList,Types,VList),
      
      b_get_machine_all_constants(TypedCstList),
      convert_typed_varlist(TypedCstList,CVarList,CTypes),
      find_non_det_constants,
      generate_typing_list(CVarList,CTypes,CList),
      ( CList = []
        -> List = [' VARIABLES;'|VList]
        ;  append([' CONSTANTS;'|CList],['VARIABLES;'|VList],List)
      ),!.
tcltk_show_typing_of_variables_and_constants('TYPE ERROR in MACHINE').

%%% Used to get the constant values for the currently loaded B machine in form of a B predicate.
%%% In case that the constants in the model has not been set up then all possible constants'
%%% evaluations will be provided in form of a predicate in Disjunctive Normal Form.
%%% Predicate used for setting up constants when model checking with TLC. (TLA)
tcltk_get_constants_predicate(P,N) :-
    current_state_id(CurID),
    (CurID==root
      ->  get_constants_as_disjoint_predicate(N,P)
      ;   prob2_interface:get_state(CurID,EState),
          filter_constants_bindings(EState,X), % this will only get current constants values
          translate: translate_bstate(X,P)
      ).

get_constants_as_disjoint_predicate(N,Res) :-
   findall(R,get_set_up_constants_solution(R),L),
   get_a_part_of_the_found_solutions(L,N,Part),
   add_or_between_constant_predicates(Part,Res).

get_a_part_of_the_found_solutions(L,N,Part) :-
   length(L,Length),
   (N>=Length
       -> Part=L
       ;  sublist(L,Part,_B,N,_A)).


get_set_up_constants_solution(R) :-
   %\+ max_reached_or_timeout_for_node(root),!, % then lookup constants rather than computing them
   transition(root,_,Dst),
   visited_expression(Dst,concrete_constants(X)),
   translate:translate_bstate(X,R).
%get_set_up_constants_solution(R) :-
%    b_interpreter: b_set_up_concrete_constants(X),
%    translate:translate_bstate(X,R).

add_or_between_constant_predicates([],'1=1').
add_or_between_constant_predicates(L,Res) :-
   tools: ajoin_with_sep(L,' or ',Res).

filter_constants_bindings(EState,R) :-
   b_get_machine_all_constants(C),
   convert_typed_varlist(C,CVarList,_CTypes),
   filter_constants_bindings1(EState,CVarList,R).

filter_constants_bindings1([],_,[]).
filter_constants_bindings1([bind(Ident,Val)|T],CVarList,Res) :-
   (member(Ident,CVarList) -> Res = [bind(Ident,Val)|Rest]; Res = Rest),
   filter_constants_bindings1(T,CVarList,Rest).

convert_typed_varlist([],[],[]).
convert_typed_varlist([b(identifier(ID),Type,_)|T],[ID|IT],[Type|TT]) :-
  %print(ID), print(' : '), print(Type),nl,
  % db : set(couple(global(Name),global(Code)))   a : set(integer)  boolean
  convert_typed_varlist(T,IT,TT).
%--------------------------------------


:- use_module(symmetry_marker,[type_allows_for_precise_marker/1]).
:- use_module(bmachine,[b_is_unused_constant/1]).
:- use_module(library(codesio),[format_to_codes/3]).
generate_typing_list([],[],[]).
generate_typing_list([Var|VT],[BasicType|TT],[Translation|Rest]) :-
   pretty_type(BasicType,PrettyType),
   (kernel_objects:max_cardinality(BasicType,Card) -> true ; Card='??'),
   (b_get_constant_represented_inside_global_set(Var,BasicType)
     -> Imp='(treated as enumerated) '
    ; (preference(symmetry_mode,hash),
         \+type_allows_for_precise_marker(BasicType))
       -> Imp = '(IMPRECISE for HASH) '
       ;  Imp = ''
   ),
   (non_det_constant(Var)
     -> MulSol ='[multiple solutions in SETUP_CONSTANTS]'
      ; MulSol=''),
   (b_is_unused_constant(Var) -> Unused=' (unused)' ; Unused=''),
%   mnf(translate:translate_bexpression(b(member(identifier(Var),BasicType),pred,[]),Translation)),
   format_to_codes('~w:~w [card=~w]~w~w~w;;',[Var,PrettyType,Card,MulSol,Imp,Unused],Codes),
   atom_codes(Translation,Codes),
   generate_typing_list(VT,TT,Rest).
  
:- use_module(symmetry_marker,[hash_model_checking_imprecise/2]).
:- use_module(translate,[pretty_type/2]).
tcltk_hash_model_checking_imprecise :-
   hash_model_checking_imprecise(ID,BasicType),!,
   print(' * Note: Hashing may be imprecise for: '),
   print(ID), print(' : '),translate:pretty_type(BasicType,S),print(S), print(' !'),nl.


/* find constants which are assigned multiple values */
:- dynamic non_det_constant/1.
find_non_det_constants :-
   retractall(non_det_constant(_)),
   findall(Csts,visited_expression(_,concrete_constants(Csts),_),CL),
   ((CL=[Valuation1|O],O=[_|_]) -> find_non_det_constants1(Valuation1,O)
    ;  true /* 0 or 1 valuation only */
    ).

find_non_det_constants1([],_).
find_non_det_constants1([bind(C,Value)|T],Other) :- check_cst(Other, C, Value, PeeledOther),
  find_non_det_constants1(T,PeeledOther).

check_cst([],_,_,[]).
check_cst([ [bind(C,OtherVal)|T1] | TT], C, Value, [ T1 | PTT]) :-
   (Value=OtherVal -> check_cst(TT,C,Value,PTT)
    ; assert(non_det_constant(C)), peel2(TT,PTT)).
peel2([],[]).
peel2([ [_|T1] | TT], [ T1 | PTT]) :- peel2(TT,PTT).


/* --------------------------------- */

	
   /* for tcltk: */
:- use_module(pref_definitions,[b_get_definition_string_list_from_machine/2]).
/* --------------------------------- */

%%
% Visualize Invariant...
%
% Siehe bvisual.pl

:- use_module(bvisual).

generate_dot_from_invariant(FileName):-
    bmachine:b_get_invariant_from_machine(BExpr),
    write_dot_file_for_pred(BExpr,FileName).


generate_dot_from_operation(TclOpName,FileName) :-  
    predicate_debugger:adapt_tcl_operation_name(TclOpName,OpName),
    bvisual:get_operation_pre(OpName, BExpr, BecomesSuchVars),
    print_bexpr(BExpr),nl,
    generate_tree_from_boolean_expression(BExpr, Tree,BecomesSuchVars),
    write_dot_graph_to_file(Tree,FileName).
   	

generate_dot_from_properties(FileName) :- 
    (current_state_corresponds_to_fully_setup_b_machine -> CreateExists=false; CreateExists=true),
    bvisual:get_properties(BExpr,CreateExists), 
    write_dot_file_for_pred(BExpr,FileName).
    
:- use_module(bsyntaxtree,[conjunct_predicates/2]).

generate_dot_from_assertions(FileName) :- 
    ((current_state_corresponds_to_setup_constants_b_machine,
     bmachine:b_get_dynamic_assertions_from_machine(BExpr), BExpr \=[])
      -> true
       ; bmachine:b_get_static_assertions_from_machine(BExpr)
    ),
    conjunct_predicates(BExpr,BE),
    write_dot_file_for_pred(BE,FileName).
    
generate_dot_from_goal(FileName) :-  
    b_get_machine_goal(BExpr),
    print('% CHECKING: '),translate_bexpression(BExpr,PT), print(PT),nl,
    write_dot_file_for_pred(BExpr,FileName).
 
generate_dot_from_deadlock_po(FileName) :-  
    b_state_model_check:get_unsorted_all_guards_false_pred(BExpr),
    print('% CHECKING: '),translate_bexpression(BExpr,PT), print(PT),nl,
    write_dot_file_for_pred(BExpr,FileName).


write_dot_file_for_pred(BExpr,FileName) :- 
    generate_tree_from_boolean_expression(BExpr, Tree,[]),
    print('% Writing to: '), print(FileName),nl,
    mnf(write_dot_graph_to_file(Tree,FileName)).
write_dot_file_for_pred_expr_and_state(BExpr, LocalState, State,FileName) :- 
    mnf(get_tree_from_expr(Tree, BExpr, LocalState, State)),
    print('% Writing to: '), print(FileName),nl,
    mnf(write_dot_graph_to_file(Tree,FileName)).
 
    
    
generate_tree_from_boolean_expression(BExpr,Tree,BSVars):-
    ( 
        current_state_corresponds_to_setup_constants_b_machine,
        current_b_expression(CurState0), 
        % print(cur(CurState0)),nl,
   ~~mnf(insert_before_substitution_variables(BSVars,[],CurState0,CurState0,CurState))
   
        ;
        CurState = []
    ),!,
    mnf(get_tree_from_expr(Tree, BExpr, [], CurState)).

tcltk_bv_get_tops(Tops) :-
    bv_get_top_level(Tops).    
tcltk_bv_get_structure(Id,Label,list(Subs)) :-
    bv_expand_formula(Id,Label,Subs).
tcltk_bv_is_leaf(Id,Label) :-
    bv_expand_formula(Id,Label,[]).
tcltk_bv_get_value(Id,RText,Tag) :-
    current_state_id(StateId),
    bv_get_values([Id],StateId,[Value]),
    retrieve_tag_text_for_id(Id,Value,Tag,Text) ,
    (Text = '[]' -> RText = '[ ]' ; RText=Text). % for some reason the empty sequence translation [] gets transformed into '' by the tcltk library
retrieve_tag_text_for_id(Id,Value,FullTag,Text) :-
    retrieve_tag_text(Value,Tag,Text),
    (bv_is_explanation_node(Id) -> convert_explain_tag(Tag,FullTag) ; FullTag=Tag).
retrieve_tag_text(i,       inac,   '').
retrieve_tag_text(e(Text), error,  Text).
retrieve_tag_text(v(Text), value,  Text).
retrieve_tag_text(p(true), ptrue,  true).
retrieve_tag_text(p(false),pfalse, false).

tcltk_bv_get_unlimited_value(Id,Text,Tag) :-
    current_state_id(StateId),
    bv_get_value_unlimited(Id,StateId,Value),
    retrieve_tag_text_for_id(Id,Value,Tag,Text1),
    atom_codes(Text1,Text).   % TODO: In the end, the result was converted from codes to atom and back

convert_explain_tag(ptrue,explaintrue).
convert_explain_tag(pfalse,explainfalse).
convert_explain_tag(X,X).

% -----------------------------------------

% STILL TO DO: refactor eval from here and probcli into separate module

:- dynamic eval_elapsed_time/1.
:- meta_predicate tcltk_time_call(0).
tcltk_time_call(Call) :-
  cputime(T1),Call,cputime(T2),ElapsedTime is T2-T1,
  retractall(eval_elapsed_time(_)),assert(eval_elapsed_time(ElapsedTime)).

:- use_module(eval_strings).  

tcltk_eval(String,StringResult,EnumWarning,Solution) :-
  get_computed_preference(debug_time_out,DTO),
  %print(debug_time_out(DTO)),nl,
  time_out_with_enum_warning_one_solution(eval_string(String,StringResult,EnumWarning,LocalState),DTO,TimeOutRes),
  (LocalState=[] -> Solution='' ; translate_bstate(LocalState,Solution)),
  (is_time_out_result(TimeOutRes) ->
    StringResult = '**** TIME-OUT ****', print(StringResult), print(' ('),print(DTO), print('ms)'),nl,
    EnumWarning = no
    ; print_last_info).


tcltk_eval_string_in_cur_state(String,TclValue) :-
    eval_string_in_cur_state(String,Value),
    convert_value_to_tcl(Value,TclValue).
convert_value_to_tcl(int(X),R) :- !, R=X.
convert_value_to_tcl(pred_true,R) :- !, R=1.
convert_value_to_tcl(pred_false,R) :- !, R=0.
convert_value_to_tcl(string(X),R) :- !, R=X.
% TO DO: add sets,...
convert_value_to_tcl(R,R) :- print(cannot_convert_to_tcl(R,R)),nl.

eval_string_in_cur_state(String,Value) :-
    atom_codes(String,Codes),
    %format('Parsing : "~w"~n',[String]),
    bmachine:b_parse_machine_expression_from_codes_with_prob_ids(Codes,Typed),
    current_state_id(CurID), prob2_interface:get_state(CurID,EState),
    %print(evaluating_in_state(CurID)),nl,
    b_compute_expression_with_prob_ids(Typed,EState,Value).

% code to translate an expression into a list of lists that can be displayed as a table in Tk
% TO DO: make more robust and add more fancy translations
:- use_module(table_tools).
tcltk_eval_as_table(String,Res) :-
    eval_string_in_cur_state(String,Value),
    Res = list([list(Header)|Table]),
    expand_and_translate_to_table(Value,Header,Table).

:- use_module(user_interrupts,[interruptable_call/1, interruptable_call/2]).
interruptable_tcltk_eval(String,StringResult,EnumWarning,Solution) :-
  if(interruptable_call(tcltk_eval(String,StringResult,EnumWarning,Solution)),true,
     (StringResult = 'Error or User-Interrupt', EnumWarning=unknown, Solution='')).

% -----------------------------------------
% B Simplifier
:- use_module(probporsrc(b_simplifier),[tcltk_simplify_b_predicate/2]).
tcltk_simplify_predicate(String,StringResult) :-
   tcltk_simplify_b_predicate(String,StringResult).


% -----------------------------------------
% CBC Check

get_cbc_data_base_id_checks(IDs) :-
	findall(ID,b_state_model_check: cbc_check(ID,_Text,_Call,_Res,_Ok),IDs).
get_cbc_data_base_text_checks(Strings) :-
	findall(Text1,(b_state_model_check: cbc_check(_ID,Text,_Call,_Res,_Ok),tools: string_concatenate(Text,';',Text1)),Strings).


% -----------------------------------------

% CSP Refinement Search

:- dynamic tcltk_cspm_mode/0.

set_tcltk_cspm_mode :- assert(tcltk_cspm_mode).
unset_tcltk_cspm_mode :- retractall(tcltk_cspm_mode).

:- meta_predicate add_csp_process_id1(-,-,1).

get_new_csp_process_id(E,File,ID) :- 
	( parse_single_csp_expression_file(E,File,Proc) -> 
		haskell_csp: check_compiled_term(Proc),	add_csp_process_id1(Proc,ID,animatable_process)
	; add_error(parse_single_csp_expression_file,'Invalid process expression: ',E),fail
	).

%parse_and_analyze_process_expression(ProcExp,File,Proc) :- 
%	( parse_single_csp_expression_file(ProcExp,File,Proc) -> true
%	; add_error(parse_single_csp_expression_file,'Invalid process expression: ',ProcExp),fail
%	).

:- use_module(refinement_checker).
get_csp_process_id(Process,ID) :- get_start_transition_term(Process,Trans),
   transition(root,Trans,ID).

:- use_module(probcspsrc(haskell_csp),[animatable_process_without_arguments/1,animatable_process/1,animatable_process_cli/1]).
add_csp_process_id(Process,NewID) :-
    add_csp_process_id1(Process,NewID,animatable_process_cli).

add_csp_process_id1(Process,NewID,Pred) :-
    (call(Pred,Process)
     -> get_start_transition_term(Process,Trans),
        (transition(root,Trans,_,DestID) -> NewID=DestID % process already set up
         ;  xtl_interface:get_start_expr(Process,NewState),
            add_trans_id(root,Trans,NewState,[],NewID,[],_)
        )
      ; add_error_fail(add_csp_process_id,'Not an animatable CSP Process: ',Process)).

get_start_transition_term('MAIN',R) :- !, R= start_cspm_MAIN.
get_start_transition_term(agent_call(_Src,'MAIN',[]),R) :- !, R= start_cspm_MAIN.
get_start_transition_term(P,start_cspm(P)).

add_csp_general_process_id(val_of(Process,_),ID) :- !,
   ( animatable_process_without_arguments(Process) % in case the called processes is one without arguments, then we don't have to parse the expression
	-> add_csp_process_id(Process,ID)
  % code below not covered; loaded_main_file only defined for probcli entry point !
   %; animation_mode(cspm),
   %  loaded_main_file(CSPFile),
   %  parse_and_analyze_process_expression(Process,CSPFile,Proc),
	% (transition(root,start_cspm(Proc),_TransID,ID) -> true
	% ; otherwise -> add_trans_id(root,start_cspm(Proc),Proc,[],ID,[],_)
	% )
   ).
add_csp_general_process_id(GenProcExpr,ID) :-
   % stripping out the source-loc info
   haskell_csp_analyzer: clear_span(GenProcExpr,GenProcExpr1),
   get_start_transition_term(GenProcExpr1,Trans),
   (transition(root,Trans,_TransID,ID)
    -> true
    ;  add_trans_id(root,Trans,GenProcExpr1,[],ID,[],_)
   ).

tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style) :-
   format('Checking ~w ~w ~w~n',[AbsProcess,Style,ImplProcess]),
   tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style,10000000).
csp_ref_style('[T=',trace). % Trace Refinement
csp_ref_style('T',trace). % Trace Refinement
csp_ref_style('[SF=',singleton_failures). % Singleton Failures Refinement
csp_ref_style('[F=',failures). % Failures Refinement
csp_ref_style('[R=',refusals). % Refusals Refinement
csp_ref_style('[RD=',refusals_div). % Refusals-Divergence Refinement
csp_ref_style('F',failures). % Failures Refinement
csp_ref_style('[FD=',X) :- csp_ref_style('FailureDivergence',X).
csp_ref_style('FD',X) :- csp_ref_style('FailureDivergence',X).
csp_ref_style('Trace',trace).
csp_ref_style('Failure',failures).
csp_ref_style('FailureDivergence',failure_divergences).
csp_ref_style('RefusalTesting',refusals).
csp_ref_style('RefusalTestingDiv',refusals_div).
csp_ref_style('RevivalTesting',revival) :- add_error(unsupported_refinement_operator,'[V= refinement not yet supported.').
csp_ref_style('RevivalTestingDiv',revival_div) :- add_error(unsupported_refinement_operator,'[VD= refinement not yet supported.').

translate_csp_ref_style(Style,RefStyle) :- (csp_ref_style(Style,RefStyle) -> true ;
     add_error_fail(csp,'Unknown refinement style: ',Style)).

% tcltk_csp_in_situ_refinement('P','Q',R,0,1000). tcltk_csp_in_situ_refinement('Q','P',R,1000).
tcltk_csp_in_situ_refinement(ImplProcess,AbsProcess,ResTrace,Style,MaxNrOfNewNodes) :-
    add_csp_process_id(ImplProcess,ImplNodeID),
    add_csp_process_id(AbsProcess,SpecNodeID),
    translate_csp_ref_style(Style,RefStyle),
    refinement_checker:in_situ_ref_search(ImplNodeID,SpecNodeID,ResTrace,RefStyle,MaxNrOfNewNodes).

:- use_module(probcspsrc(haskell_csp),[get_csp_assertions/1,translate_csp_assertion/2]).
tcltk_check_csp_assertions([],Summary) :- get_csp_assertions(Assertions),
   treat_csp_ass(Assertions,_List,0,0,0,Summary). % TO DO: display _List

treat_csp_ass([],[],T,F,U,[total/Tot, true/T, false/F, unknown/U]) :- Tot is T+F+U.
treat_csp_ass([Assertion|T],['='(PP,Ok)|TPP],True,False,U,Summary) :- !,
   (checkAssertion(Assertion,PP,_,Ok,_) 
    -> U1=U, (Ok=true -> T1 is True+1, F1=False ; T1=True, F1 is False+1) 
    ;  U1 is U+1, T1=True, F1=False,Ok=unknown),
   treat_csp_ass(T,TPP,T1,F1,U1,Summary).
treat_csp_ass([H|T],TP,T1,F1,U,Summary) :- add_error(treat_csp_ass,'Unknown CSP assertion: ',H),
   U1 is U+1, treat_csp_ass(T,TP,T1,F1,U1,Summary).

% assertModelCheckExt(False,val_of(P3,src_span(25,8,25,10,562,2)),DeadlockFree,F)

run_assertion_check(assertModelCheckExt(Negated,Spec,Type,ModelStyle),Negated,ResTrace) :-
  haskell_csp_analyzer: compile_body(Spec,'assertModelCheck',[],[],CSpec),
  debug_println(9,assertModelCheckExt(Negated,CSpec,Type,ModelStyle)),
  add_csp_general_process_id(CSpec,SpecNodeID),
  translate_csp_ref_style(ModelStyle,RefStyle),
  refinement_checker:in_situ_model_check(SpecNodeID,ResTrace,Type,RefStyle,10000000),
  print(result_trace),nl,print(ResTrace),nl.
run_assertion_check(assertModelCheck(Negated,Spec,Style),Negated,ResTrace) :-
  run_assertion_check(assertModelCheckExt(Negated,Spec,Style,'FailureDivergence'),Negated,ResTrace).
run_assertion_check(assertRef(Negated,Spec,Style,Impl,Span),Negated,ResTrace) :-
  haskell_csp_analyzer: compile_body(Spec,'assertRef',[],[],CSpec),
  haskell_csp_analyzer: compile_body(Impl,'assertRef',[],[],CImpl),
  debug_println(9,assertRef(Negated,CSpec,Style,CImpl,Span)),
  add_csp_general_process_id(CSpec,SpecNodeID),
  add_csp_general_process_id(CImpl,ImplNodeID),
  translate_csp_ref_style(Style,RefStyle),
  refinement_checker:in_situ_ref_search(ImplNodeID,SpecNodeID,ResTrace,RefStyle,10000000).
run_assertion_check(assertLtl(Negated,Spec,FormulaAsAtom,_Span),Negated,ResTrace) :-
  add_csp_process_id_normalized(Spec,'assertLtl',CSpec,CSpecNodeID),
  %% print(add_csp_general_process_id(Spec,CSpec,SpecNodeID)),nl,
  tcltk_reset,
  ltl:ltl_model_check_with_ce1(CSpec,FormulaAsAtom,-1,specific_node(CSpecNodeID),Result),
  translate_ltl_result(Result,ResTrace),
  debug_println(9,ltl_result(Negated,CSpec,CSpecNodeID,FormulaAsAtom,Result)).
run_assertion_check(assertCtl(Negated,Spec,FormulaAsAtom,_Span),Negated,ResTrace) :-
  add_csp_process_id_normalized(Spec,'assertCtl',CSpec,CSpecNodeID),
  tcltk_reset,
  ctl: ctl_model_check_with_ce(CSpec,FormulaAsAtom,-1,specific_node(CSpecNodeID),Result,Trace),
  debug_println(9,ctl_result(Negated,CSpec,CSpecNodeID,FormulaAsAtom,Result)),
  translate_ctl_result(Result,Trace,ResTrace).

checkAssertion(Assertion,PP,Negated,Ok,ResTrace) :-
  print('CHECKING '),
  translate_csp_assertion(Assertion,PP),
  print(PP),nl,
  debug_println(9,checkAssertion(Assertion)),
  reset_refinement_checker,
  (run_assertion_check(Assertion,Negated,ResTrace)
    -> debug_println(9,result(ResTrace)),
      (ResTrace=no_counter_example
         -> (Negated = 'False' -> Ok=true
               ; %add_error(check_assertRef,'Assertion FAILS (expected refinement not to hold): ',PP),
                 Ok=false)
          ; (ResTrace=[interrupted|_]
                -> Ok=interrupted
               ;  (Negated = 'False' -> %add_error(check_assertRef,'Assertion FAILS: ',PP),
                                  Ok=false ; Ok=true)
            )
       ),
       (Ok=true -> print('OK'),nl ; (Ok=interrupted -> print('*** Execution interrupted by user ***'),nl
                                                 ;  print('*** FALSE ***')),nl
                                 )
    ;  add_error_fail(checkAssertion,'Internal Error, Assertion Checking Failed for: ',PP)
  ).

add_csp_process_id_normalized(Spec,Functor,CSpec,CSpecNodeID) :-
  % stripping out the src_span info
  haskell_csp_analyzer: clear_span(Spec,SpecNoSpan),
  haskell_csp_analyzer: compile_body(SpecNoSpan,Functor,[],[],CSpec),
  add_csp_general_process_id(CSpec,CSpecNodeID).

translate_ltl_result(ok,Res) :- !,
  Res = no_counter_example.
translate_ltl_result(no,Res) :- !,
  Res = no.
translate_ltl_result(ce(Atom),Res) :-
  !, % we have a counter-example for the LTL formula
  Res=Atom.
translate_ltl_result(interrupted,Res) :-
  !,
  Res=[interrupted].
translate_ltl_result(Status,_Res) :-
  add_error_fail(ltl_assertions_result, 'Internal Error: unknown LTL assertion result ', Status).

translate_ctl_result(true,_Trace,ResTrace) :- !,
  ResTrace = no_counter_example.
translate_ctl_result(false,ce(Atom),ResTrace) :- !,
  ResTrace = Atom.
translate_ctl_result(interrupted,_TRace,ResTrace) :-
  !,
  ResTrace=[interrupted].
translate_ctl_result(Status,_Trace,_ResTrace) :-
  add_error_fail(ctl_assertions_result, 'Internal Error: unknown CTL assertion result ', Status).

tcltk_check_csp_assertion(Assertion,CSPFile,Negated,Result) :-
   tcltk_check_csp_assertion(Assertion,CSPFile,Negated,_PlTerm,Result).

tcltk_check_csp_assertion(Assertion,CSPFile,Negated,PlTerm,Result) :-
	parse_single_csp_declaration(Assertion,CSPFile,PlTerm),
	checkAssertion(PlTerm,_PP,Negated,_Ok,Result).

tcltk_play_ltl_ctl_counterexample_trace(Assertion,CSPFile,Type,FairnessChecked) :-
  announce_event(play_counterexample),
  ( Type == ltl -> Functor = assertLtl
  ; Type == ctl -> Functor = assertCtl
  ; otherwise -> add_error_fail(ltl_ctl_counterexample, 'Internal Error: Model checker type unknown: ',Type)),
  % constructing the prolog declaration term
  PlDeclTerm =.. [Functor,_Negated,Spec,FormulaAsAtom,_Span],
  parse_single_csp_declaration(Assertion,CSPFile,PlDeclTerm),
  haskell_csp_analyzer: clear_span(Spec,SpecNoSpan),
  haskell_csp_analyzer: compile_body(SpecNoSpan,'assertLtl',[],[],CSpec),
  % parsing and preprocessing the LTL/CTL formula
  parse_and_preprocess_formula(FormulaAsAtom,Type,PF),
  (is_fairness_implication(PF) -> FairnessChecked='yes'; FairnessChecked='no'),
  % need to reset history and current state before loading the counter-example 
  tcltk_reset,
  tcltk_play_counterexample(Type,CSpec,PF).

tcltk_play_counterexample(ltl,Spec,Formula) :- !,
  ltl: tcltk_play_ltl_counterexample(Spec,Formula).
tcltk_play_counterexample(ctl,Spec,Formula) :- !,
  ctl: tcltk_play_ctl_counterexample(Spec,Formula).
tcltk_play_counterexample(Type,Spec,Formula) :-
  add_error_fail(ltl_ctl_counterexample, 'Internal Error: Unknown domain type: ', (Type,Spec,Formula)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% Parse and get LTL formulas from LTL File %%%%
get_ltl_formulas_from_file(Filename,Text) :-
    parse_ltlfile(Filename, Formulas),
    extract_formulas(Formulas,Fs,_,_),
    translate_ltl_formulas(Fs,Text).

translate_ltl_formulas(Formulas,Text) :-
    maplist(translate_ltl_formula,Formulas,Strings),
    haskell_csp: concatenate_string_list(Strings,Text).

translate_ltl_formula(Formula,Text) :-
    pp_ltl_formula(Formula,T),
    string_concatenate(T,'$',Text).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tcltk_explore_csp_process(ProcString,SpecNodeID) :-
  haskell_csp: parse_single_csp_expression(ltl,ProcString,Res),
  haskell_csp_analyzer: clear_span(Res,ResNoSpan),
  haskell_csp_analyzer: compile_body(ResNoSpan,'assertLtl',[],[],CRes),
  add_csp_general_process_id(CRes,SpecNodeID),
  refinement_checker:dfs_check(SpecNodeID,_ResTrace,false,false).

tcltk_visualize_csp_process(SpecNodeID,F) :-
  visualize_graph: tcltk_print_csp_process_states_for_dot(F,SpecNodeID).

:- dynamic nr_csp_process_state/1.
get_csp_process_stats(NrNodes) :-
  assert(nr_csp_process_state(0)),
  count_nodes,
  nr_csp_process_state(NrNodes),
  retractall(nr_csp_process_state(_)).

count_nodes :-
  refinement_checker: dvisited(_NodeID),
  nr_csp_process_state(NrNodes),
  NewNrNodes is NrNodes +1,
  retractall(nr_csp_process_state(_)),
  assert(nr_csp_process_state(NewNrNodes)),
  fail.
count_nodes.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tcltk_machine_has_assertions :-
    ( b_get_static_assertions_from_machine([_|_]); b_get_dynamic_assertions_from_machine([_|_]) ), !.
    
tcltk_unique_top_level_operation(Name) :- b_top_level_operation(Name),
  \+ (b_top_level_operation(N2), N2 \= Name).

tcltk_top_level_operations(list(Names)) :-
    findall(Name,b_top_level_operation(Name),Names).

tcltk_get_vacuous_guards(list(R)) :-
   eclipse_interface:get_vacuous_guards(L),
   (L=[] -> R = ['No vacuous guards']
     ; R=L).
   
tcltk_get_vacuous_invariants(list(L)) :-
   eclipse_interface:get_vacuous_invariants(LP),
   (LP=[] -> L = ['No vacuous invariants']
    ; maplist(translate:translate_bexpression,LP,L)).
