% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(tools_commands,
    [ edit_file/1,edit_file/2, diff_files_with_editor/2 ]).
    

:- use_module(preferences).
:- use_module(debug).
:- use_module(tools).
:- use_module(system_call,[system_call/4]).

edit_file(F) :- edit_file(F,none).
edit_file(FILE,LINE) :-
     get_preference(path_to_text_editor_launch,EDITOR),
     edit_file_with_editor(EDITOR,FILE,LINE).

edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,EdName),
     editor_understands_plus_line(EdName),!, % bbedit/vim can handle +LINE arg
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     tools:string_concatenate('+',LINE,LinePosArg),
     system_call:system_call(EDITOR,[LinePosArg,FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,LINE) :- number(LINE),
     tools:get_tail_filename(EDITOR,atom),!, % atom can handle FILE:LINE
     format('Opening file "~w" at LINE ~w using EDITOR "~w"~n',[FILE,LINE,EDITOR]),
     ajoin([FILE,':',LINE],FILE_LINE),
     system_call:system_call(EDITOR,[FILE_LINE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).
edit_file_with_editor(EDITOR,FILE,_) :-
     format('Opening file "~w" using EDITOR "~w"~n',[FILE,EDITOR]),
     system_call:system_call(EDITOR,[FILE],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).


editor_understands_plus_line(bbedit).
editor_understands_plus_line(vim).



diff_files_with_editor(FILE1,FILE2) :- 
     DIFFER = '/usr/local/bin/bbdiff', % TO DO: provide preference
     system_call:system_call(DIFFER,[FILE1,FILE2],ErrorTextAsCodeList,ExitCode),
     debug_println(9,exit(ExitCode,ErrorTextAsCodeList)).