% (c) 2004-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bsets_clp,
   [empty_sequence/1,
    is_sequence/2,  is_sequence_wf/3, not_is_sequence/2, not_is_sequence_wf/3,
    not_is_non_empty_sequence_wf/3,
    injective_sequence_wf/3,
    not_injective_sequence/3,
    not_non_empty_injective_sequence/3,
    injective_non_empty_sequence/3,
    finite_non_empty_sequence/3,
    permutation_sequence_wf/3,
    not_permutation_sequence/3,
    not_surjective_seq/3,
    size_of_sequence/3,
    prepend_sequence/4, append_sequence/4, prefix_sequence_wf/4,
    suffix_sequence/4, concat_sequence/4,
    disjoint_union_wf/4,
    concatentation_of_sequences/3,
    tail_sequence/4, first_sequence/4, front_sequence/4, last_sequence/4,
    rev_sequence/3,


   % maplet/3,
   % relation/1,
    relation_over/3, relation_over_wf/4,
    not_relation_over/4,
    domain_wf/3,

    range_wf/3,
    identity_relation_over_wf/3, in_identity/3, not_in_identity/3,
    invert_relation_wf/3,
    tuple_of/3,
    rel_composition/3, rel_composition_wf/4,
    direct_product_wf/4,
    parallel_product/3, in_parallel_product_wf/4, not_in_parallel_product_wf/4,
    rel_iterate/4,


    not_partial_function/4,
    partial_function/3, partial_function_wf/4,

    total_function/3, total_function_wf/4,

%    enumerate_total_bijection/3,
    total_bijection/3, total_bijection_wf/4,

    not_total_function/4,
    not_total_bijection/4,


    range_restriction_wf/4, range_subtraction_wf/4,
    domain_restriction_wf/4, domain_subtraction_wf/4,
    override_relation/4,
    image_wf/4, image_for_closure1_wf/4,
    element_of_closure1_wf/3, not_element_of_closure1_wf/3,
    in_domain_wf/3, not_in_domain_wf/3,
    apply_to/4, apply_to/5,
    override/5,

    %sum_over_range/2, mul_over_range/2,

    disjoint_union_generalized_wf/3,

    partial_surjection/3, not_partial_surjection_wf/4,
    total_relation_wf/4,
    not_total_relation_wf/4,

    surjection_relation_wf/4, total_surjection_relation_wf/4,
    not_surjection_relation_wf/4, not_total_surjection_relation_wf/4,

    total_surjection/3, total_surjection_wf/4,
    not_total_surjection_wf/4,

    partial_injection/3, partial_injection_wf/4,
    not_partial_injection/4,

    total_injection/3, total_injection_wf/4,
    not_total_injection/4,

    partial_bijection/3, partial_bijection_wf/4,
    not_partial_bijection/4,

    relational_trans_closure/2, relational_reflexive_closure/2
   ]).


%:- print(loading_bsets_clp),nl.
%:- use_module(library(clpfd)).
%portray_message(informational, _).
:- use_module(library(terms)).
:- use_module(self_check).

:- use_module(debug).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides more advanced operations for the basic datatypes of ProB (mainly for relations, functions, sequences).').

:- use_module(tools_printing,[print_term_summary/1]).

:- use_module(delay).

%:- use_module(labeling).

:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(kernel_objects).
:- use_module(kernel_records).
:- use_module(kernel_tools).

:- use_module(kernel_waitflags).
:- use_module(kernel_equality,[equality_objects_wf/4]).

%:- use_module(b_enumerate,[b_enumerate_values/1]).
:- use_module(custom_explicit_sets).
:- use_module(avl_tools,[avl_fetch_pair/3]).
:- use_module(bool_pred,[negate/2]).

:- use_module(bsyntaxtree, [conjunct_predicates/2,
                            mark_bexpr_as_symbolic/2,
                            create_texpr/4
                            ]).

/* --------- */
/* SEQUENCES */
/* ------- - */

:- assert_must_succeed((bsets_clp:empty_sequence([]))).
:- assert_must_fail((bsets_clp:empty_sequence([int(1)]))).
empty_sequence(X) :- empty_set(X). % TO DO: add WF

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_empty_sequence([(int(2),int(33)),(int(1),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_empty_sequence([(int(1),int(33))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_empty_sequence([]))).

not_empty_sequence(X) :- var(X),!,
   X = [(int(1),_)|_].
not_empty_sequence(X) :- is_custom_explicit_set_nonvar(X),!,
   is_non_empty_explicit_set(X).
not_empty_sequence([(int(_),_)|_]). % clousure, avl_set dealt with clause above

not_empty_sequence_wf(X,_WF) :- nonvar(X),!, not_empty_sequence(X).
not_empty_sequence_wf(X,WF) :- (preferences:preference(use_smt_mode,true) -> not_empty_sequence(X)
   ; get_enumeration_starting_wait_flag(not_empty_sequence_wf,WF,LWF),  not_empty_sequence_lwf(X,LWF)).

:- block not_empty_sequence_lwf(-,-).
not_empty_sequence_lwf(S,_) :- nonvar(S),!,not_empty_sequence(S).
not_empty_sequence_lwf([(int(1),_)|_],_).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:is_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(bsets_clp:is_sequence(closure(['_zzzz_unit_tests'],[couple(integer,integer)],b(member(b(identifier('_zzzz_unit_tests'),couple(integer,integer),[generated]),b(value([(int(1),int(22))]),set(couple(integer,integer)),[])),pred,[])),[int(22)])).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:is_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(1),int(33)),(int(0),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:is_sequence([(int(3),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed((is_sequence(R,global_set('Name')),R = [])).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((is_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name')) )).
:- assert_must_succeed((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_fail((is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(3))] )).
:- assert_must_fail((is_sequence(R,global_set('Name')),
                        R = [(int(0),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((is_sequence(X,global_set('Name')),
                        kernel_objects:enumerate_basic_type(X,seq(global('Name'))),
                        X = [(int(1),fd(1,'Name'))])).

is_sequence(X,Type) :- init_wait_flags(WF),
  is_sequence_wf(X,Type,WF),
  ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([int(1),int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_domain([],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:is_sequence_domain([int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:is_sequence_domain([int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_is_sequence_domain([int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_is_sequence_domain([int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_is_sequence_domain([int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_is_sequence_domain([],WF),WF)).

% check if a set is the domain of a sequence, i.e., an interval 1..n with n>=0
:- use_module(custom_explicit_sets,[construct_interval_closure/3]).
:- block is_sequence_domain(-,?).
is_sequence_domain(Domain,WF) :-
   kernel_objects:finite_cardinality_as_int(Domain,int(Max),WF),
   %print(size(Max)),nl,
   construct_interval_closure(1,Max,Interval), equal_object_wf(Domain,Interval,is_sequence_domain,WF).
:- block not_is_sequence_domain(-,?).
not_is_sequence_domain(Domain,WF) :-
   kernel_objects:finite_cardinality_as_int(Domain,int(Max),WF),
   %print(size(Max)),nl,
   construct_interval_closure(1,Max,Interval), not_equal_object_wf(Domain,Interval,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_wf([(int(1),pred_true)],
[pred_true,pred_false],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:is_sequence_wf([(int(1),pred_true),(int(2),pred_false),(int(3),pred_true)],
[pred_true,pred_false],WF),WF)).
:- assert_must_succeed((bsets_clp:is_sequence_wf([(int(X),R)],[pred_true],_WF),X==1, R==pred_true)).
:- assert_must_succeed((bsets_clp:is_sequence_wf([(int(X),R),(int(Y),R)],[pred_true],_WF),X=2,
  (preferences:preference(use_clpfd_solver,true) -> Y==1 ; Y=1), R==pred_true)).

is_sequence_wf(Seq,Range,WF) :- is_sequence_wf_ex(Seq,Range,WF,_).
% is_sequence_wf_ex also returns expansion; if it was done
:- block is_sequence_wf_ex(-,?,?,?).
is_sequence_wf_ex(FF,Range,WF,FF) :-  % print(is_sequence_wf_ex(FF,Range,WF,FF)),nl,
    nonvar(FF), FF = closure(_,_,_),
    custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(checking_sequence(FFDomain)),nl, %%
    is_sequence_domain(FFDomain,WF).
is_sequence_wf_ex(Seq,Range,WF,Res) :-
   expand_and_convert_to_avl_set_warn(Seq,AER,is_sequence_wf_ex,'ARG : seq(?)'),!,
   is_avl_sequence(AER),
   is_avl_relation_over_range(AER,Range,WF),
   custom_explicit_sets:construct_avl_set(AER,Res).
is_sequence_wf_ex(X,Type,WF,EX) :- %writeq(is_sequence(X,Type)),nl,
     % try_ensure_seq_numbering(X,1),
     expand_custom_set_to_list(X,EX,_,is_sequence_wf_ex),
     is_sequence2(EX,[],Type,0,_MinSize,WF).

% will make this much faster x:seq(STRING) & card(x)=400 & 401:dom(x)  (40 ms rather than > 2 secs)
% but this does not work -eval_file /Users/leuschel/git_root/prob_examples/examples/Setlog/prob-ttf/plavis-TransData_SP_21_simplified.prob
:- block try_ensure_seq_numbering(-,?).
try_ensure_seq_numbering([H|T],NextNr) :- var(H),!, print(nr(NextNr)),nl,
    H = (int(NextNr),_), N1 is NextNr+1,
    try_ensure_seq_numbering(T,N1).
try_ensure_seq_numbering(_,_).

:- block is_sequence2(-,?,?,?,?,?).
is_sequence2([],IndexesSoFar,_Type,Size,MinSize,_WF) :- MinSize = Size,
   contiguous_set_of_indexes(IndexesSoFar,Size).
   /* not very good to do the checking at the end; can we move part of the checking earlier ? */
is_sequence2([(int(Idx),X)|Tail],IndexesSoFar,Type,Size,MinSize,WF) :-
    less_than_direct(0,Idx), %is_index_greater_zero(Idx),
    not_element_of_wf(int(Idx),IndexesSoFar,WF),
    check_element_of_wf(X,Type,WF), S1 is Size+1,
    clpfd_interface:clpfd_leq(Idx,MinSize,_Posted),
    (var(Tail)
      -> clpfd_interface:clpfd_domain(MinSize,Low,_Up), %print(is_seq_index(MinSize,Low,_Up,Tail,S1,IndexesSoFar)),nl, % TO DO: ensure that final size at least Low
         ((number(Low),Low>S1) -> Tail = [_|_]  % TO DO: proper reification; what if MinSize gets constrained later
          ; expand_seq_if_necessary(Idx,S1,Tail)) % the sequence must be longer; force it
      ; true
    ), %print(rec_call(Tail)),nl,
    is_sequence2(Tail,[int(Idx)|IndexesSoFar],Type,S1,MinSize,WF).

:- block expand_seq_if_necessary(-,?,-).
expand_seq_if_necessary(MinSize,S1,Tail) :- % TO DO: proper reification on MinSize above
      number(MinSize), MinSize>S1, (var(Tail) ; Tail==[]),
      %print(expand(MinSize,S1,Tail)),nl,
      !,
      Tail = [_|_].
expand_seq_if_necessary(_,_,_).

:- block contiguous_set_of_indexes(-,?).
contiguous_set_of_indexes([],_).
contiguous_set_of_indexes([H|T],Size) :- contiguous_set_of_indexes1(T,H,Size).

:- block contiguous_set_of_indexes1(-,?,?).
contiguous_set_of_indexes1([],int(1),_).
contiguous_set_of_indexes1([int(H2)|T],int(H1),Size) :- less_than_equal_direct(H1,Size),
   less_than_equal_direct(H2,Size), less_than_equal_indexes(T,[H1,H2],Size).
/*
   Indexes = [H1,H2|T],
   when(ground(Indexes),(sort(Indexes,Sorted),contiguous_set_of_indexes2(Sorted))).
contiguous_set_of_indexes2([]).
contiguous_set_of_indexes2([int(1)|T]) :- contiguous_set_of_indexes3(T,1).
contiguous_set_of_indexes3([],_).
contiguous_set_of_indexes3([int(N)|T],N1) :- N is N1+1, contiguous_set_of_indexes3(T,N).
*/

less_than_equal_indexes([],All,_) :- clpfd_interface:clpfd_alldifferent(All).
less_than_equal_indexes([int(H)|T],All,Size) :- less_than_equal_direct(H,Size),less_than_equal_indexes(T,[H|All],Size).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(4),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(3),int(8))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(2),int(6)),(int(3),int(7)),(int(4),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_is_sequence_wf([(int(1),int(6)),(int(0),int(7)),(int(2),int(7))],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_sequence([(int(3),int(33)),(int(1),int(22))],[int(22),int(33),int(44)]))).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),R = [])).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((not_is_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_fail((not_is_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name')) )).
:- assert_must_fail((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_succeed((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_succeed((not_is_sequence(R,[int(1),int(2)]),
                        R = [(int(2),int(2)),(int(1),int(3))] )).


not_is_sequence(X,Type) :- init_wait_flags(WF),
  not_is_sequence_wf(X,Type,WF),
  ground_wait_flags(WF).

:- block not_is_sequence_wf(-,?,?).
not_is_sequence_wf(FF,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
     % print_term_summary(not_checking_sequence(FFDomain)),nl, %
    not_is_sequence_domain(FFDomain,WF).
not_is_sequence_wf(Seq,Range,WF) :-
   expand_and_convert_to_avl_set_warn(Seq,AER,not_is_sequence_wf,'ARG /: seq(?)'),
   !, %print_term_summary(not_is_sequence_wf(Seq,Range,WF)),
   (is_avl_sequence(AER) -> is_not_avl_relation_over_range(AER,Range,WF)
      ; true).
not_is_sequence_wf(X,Type,WF) :- expand_custom_set_to_list(X,EX,_Done,not_is_sequence_wf),
    not_is_sequence2(EX,[],Type,WF).

:- block not_is_sequence2(-,?,?,?).
not_is_sequence2([],IndexesSoFar,_,_WF) :- not_contiguous_set_of_indexes(IndexesSoFar).
not_is_sequence2([(int(Idx),X)|Tail],IndexesSoFar,Type,WF) :-
    membership_test_wf(IndexesSoFar,int(Idx),MemRes,WF),
    not_is_sequence3(MemRes,Idx,X,Tail,IndexesSoFar,Type,WF).

:- block not_is_sequence3(-,?,?,?,?,?,?).
not_is_sequence3(pred_true,_Idx,_X,_Tail,_IndexesSoFar,_Type,_WF).
not_is_sequence3(pred_false,Idx,_X,_Tail,_IndexesSoFar,_Type,_WF) :- nonvar(Idx),Idx<1,!.
not_is_sequence3(pred_false,Idx,X,Tail,IndexesSoFar,Type,WF) :-
      membership_test_wf(Type,X,MemRes,WF),
      not_is_sequence4(MemRes,Idx,Tail,IndexesSoFar,Type,WF).

:- block not_is_sequence4(-,?,?,?,?,?).
not_is_sequence4(pred_false,_Idx,_Tail,_IndexesSoFar,_Type,_WF).
not_is_sequence4(pred_true,Idx,Tail,IndexesSoFar,Type,WF) :-
     not_is_sequence2(Tail,[int(Idx)|IndexesSoFar],Type,WF).

not_contiguous_set_of_indexes(Indexes) :-
   when(ground(Indexes),(sort(Indexes,Sorted),not_contiguous_set_of_indexes2(Sorted,1))).
not_contiguous_set_of_indexes2([int(N)|T],N1) :-
       when(?=(N,N1),
          ((N \= N1) ;  (N=N1, N2 is N1+1, not_contiguous_set_of_indexes2(T,N2)))).





:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:not_is_non_empty_sequence([(int(1),int(22))],[int(22)]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_is_non_empty_sequence([(int(1),int(2))],[int(22)]))).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),R = [])).
:- assert_must_fail((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(2),fd(1,'Name')),(int(4),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_is_non_empty_sequence(R,global_set('Name')),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_is_non_empty_sequence(R,[int(1),int(2)]),
                        R = [(int(1),int(2)),(int(2),int(3))] )).

not_is_non_empty_sequence_wf(S,T,_) :- not_is_non_empty_sequence(S,T).
:- block not_is_non_empty_sequence(-,?).
not_is_non_empty_sequence(X,Type) :-
      empty_sequence(X) ; not_is_sequence(X,Type).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective_sequence_wf([(int(2),int(22)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_sequence_wf([],global_set('Name'),WF),WF)).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
            kernel_waitflags:ground_det_wait_flag(WF), R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
            ground_det_wait_flag(WF),  R = [(int(1),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
           ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,global_set('Name'),WF),
             ground_det_wait_flag(WF), R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_sequence_wf([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name'),WF),WF) ).
:- assert_must_succeed((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_fail((bsets_clp:injective_sequence_wf(R,[int(1),int(2)],WF),
                       ground_det_wait_flag(WF), R = [(int(2),int(2)),(int(1),int(3))] )).



:- block injective_sequence_wf(-,-,?).
injective_sequence_wf(Seq,Type,WF) :- /* corresponds to iseq */
   get_enumeration_starting_wait_flag(injective_sequence_wf,WF,LWF),  % before we start to enumerate infinite types: ensure we enumerate bounded length of Seq !
   injective_sequence_wf_aux(Seq,Type,WF,LWF).

:- block injective_sequence_wf_aux(-,?,?,-).
injective_sequence_wf_aux(Seq,Type,WF,_) :- /* corresponds to iseq */
   expand_and_convert_to_avl_set_warn(Seq,AER,injective_sequence_wf_aux,'ARG : iseq(?)'),!,
   %  print_term_summary(inj_avl_seq(AER,Type)),
   is_avl_sequence(AER),
   is_injective_avl_relation(AER,_ExactRange), % Should we check _ExactRange <: Type ??
   is_avl_relation_over_range(AER,Type,WF).
injective_sequence_wf_aux(Seq,Type,WF,LWF) :-
   expand_custom_set_to_list(Seq,ESeq,_,injective_sequence_wf),
   is_sequence_wf(ESeq,Type,WF),
   cardinality_as_int_for_wf(Type,MaxCard), % print(iseq_wf(Seq,MaxCard,LWF)),nl,
   injective_sequence2(ESeq,0,[],Type,WF,MaxCard,LWF).

:- block injective_sequence2(-,?,?,?,?,?,-),injective_sequence2(-,?,?,?,?,-,?).
injective_sequence2([],_,_,_Type,_WF,_MaxCard,_LWF).
injective_sequence2([(int(_),X)|Tail],CardSoFar,SoFar,Type,WF,MaxCard,LWF) :-
    (number(MaxCard) -> CardSoFar< MaxCard ; true),
    check_element_of_wf(X,Type,WF),
    not_element_of_wf(X,SoFar,WF),
    add_new_element_wf(X,SoFar,SoFar2,WF),
    C1 is CardSoFar+1,
    (C1 == MaxCard -> Tail=[] ; true),
    injective_sequence2(Tail,C1,SoFar2,Type,WF,MaxCard,LWF).



:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_injective_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_injective_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_injective_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_injective_sequence([(int(2),int(22)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),_WF),R = [])).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(1),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_fail((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_injective_sequence([(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))],
                       global_set('Name'),WF),
                        ground_det_wait_flag(WF) )).
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(3),int(1))] )).
:- assert_must_succeed((bsets_clp:not_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),
                        R = [(int(2),int(2)),(int(1),int(3))] )).
:- block not_injective_sequence(-,?,?), not_injective_sequence(?,-,?).
not_injective_sequence(Seq,Type,WF) :- nonvar(Seq),
    expand_and_convert_to_avl_set_warn(Seq,AER,not_injective_sequence,'ARG /: iseq(?)'),!,
    %print_term_summary(not_inj_seq(Seq,Type,WF)),
    (\+ is_avl_sequence(AER) -> true
     ; is_injective_avl_relation(AER,ExactRange) -> not_subset_of_wf(ExactRange,Type,WF)
     ; true).
not_injective_sequence(Seq,Type,WF) :- /* corresponds to Iseq */
    %get_middle_wait_flag(not_injective_sequence,WF,LWF),
    kernel_tools:ground_value_check(Seq,SV),
    not_injective_sequence1(Seq,Type,WF,SV).
:- block not_injective_sequence1(?,?,?,-).
not_injective_sequence1(Seq,Type,WF,_) :-
   expand_custom_set_to_list(Seq,ESeq,_,not_injective_sequence1),
   (not_is_sequence_wf(ESeq,Type,WF)
    ;  /* CHOICE POINT !! */
    (is_sequence_wf(ESeq,Type,WF),not_injective_sequence2(ESeq,[],Type,WF))).
:- block not_injective_sequence2(-,?,?,?).
not_injective_sequence2([(int(_),X)|Tail],SoFar,Type,WF) :-
    membership_test_wf(SoFar,X,MemRes,WF),
    not_injective_sequence3(MemRes,X,Tail,SoFar,Type,WF).

:- block not_injective_sequence3(-,?,?,?,?,?).
not_injective_sequence3(pred_true,_X,_Tail,_SoFar,_Type,_WF).
not_injective_sequence3(pred_false,X,Tail,SoFar,Type,WF) :-
         add_new_element_wf(X,SoFar,SoFar2,WF),
         not_injective_sequence2(Tail,SoFar2,Type,WF).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_non_empty_injective_sequence([(int(2),int(33)),(int(1),int(33))],[int(44),int(33),int(22)],WF),WF)).
:- assert_must_succeed((bsets_clp:not_non_empty_injective_sequence(R,global_set('Name'),WF),
  ground_det_wait_flag(WF), R = [])).
:- assert_must_fail((bsets_clp:not_non_empty_injective_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_non_empty_injective_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF), R = [(int(2),int(2)),(int(1),int(3))] )).

:- block not_non_empty_injective_sequence(-,?,?).
not_non_empty_injective_sequence(X,Type,WF) :-
      empty_sequence(X) ; not_injective_sequence(X,Type,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_non_empty_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective_non_empty_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_non_empty_sequence([(int(2),int(33)),(int(1),int(23))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:injective_non_empty_sequence([(int(2),int(44)),(int(1),int(44))],[int(22),int(33),int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:injective_non_empty_sequence(R,global_set('Name'),WF),
                       ground_det_wait_flag(WF),R = [])).
:- assert_must_succeed((bsets_clp:injective_non_empty_sequence(R,global_set('Name'),WF),
             ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- block injective_non_empty_sequence(-,-,?). /* corresponds to iseq1 */
injective_non_empty_sequence(A,Type,WF) :- nonvar(A),A=avl_set(AS), !,
   injective_sequence_wf(avl_set(AS),Type,WF),is_non_empty_explicit_set_wf(avl_set(AS),WF).
injective_non_empty_sequence(Seq,Type,WF) :- % print_term_summary(iseq1(Seq,Type,WF)),nl,
   ((nonvar(Seq),Seq=closure(_,_,_)) -> try_expand_custom_set(Seq,ESeq) ; ESeq=Seq),
   injective_sequence_wf(ESeq,Type,WF),not_empty_sequence_wf(ESeq,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:finite_non_empty_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:finite_non_empty_sequence([(int(1),int(33)),(int(2),int(33))],[int(22),int(33)],WF),WF)).
:- assert_must_fail((bsets_clp:finite_non_empty_sequence(R,global_set('Name'),WF),ground_det_wait_flag(WF),ground_det_wait_flag(WF),R = [])).
:- assert_must_succeed((bsets_clp:finite_non_empty_sequence(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- block finite_non_empty_sequence(-,?,?).
finite_non_empty_sequence(Seq,Type,WF) :- /* corresponds to Seq1 */
   is_sequence_wf_ex(Seq,Type,WF,ESeq),(var(ESeq) -> not_empty_sequence_wf(Seq,WF) ; not_empty_sequence_wf(ESeq,WF)).




:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:permutation_sequence_wf([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:permutation_sequence_wf([(int(2),int(33)),(int(1),int(22))],[int(22),int(33)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:permutation_sequence_wf([(int(2),int(33)),(int(1),int(23))],[int(23),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:permutation_sequence_wf([(int(2),int(44)),(int(1),int(44))],[int(44)],WF),WF)).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(2))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(2)),(int(2),int(1))] )).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:permutation_sequence_wf(R,[pred_true /* bool_true */,pred_false /* bool_false */],WF), kernel_waitflags:ground_wait_flags(WF), nonvar(R),
                        R = [(int(1),pred_false /* bool_false */),(int(2),pred_true /* bool_true */)] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(1))] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),ground_det_wait_flag(WF),R = [])).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
                 ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
   ground_det_wait_flag(WF),
  kernel_objects:equal_object(R,[(int(1),fd(1,'Name')),(int(3),fd(2,'Name')),(int(2),fd(3,'Name'))]) )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(3))] )).
:- assert_must_fail((bsets_clp:permutation_sequence_wf(R,global_set('Name'),WF),
                        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- assert_must_succeed((bsets_clp:permutation_sequence_wf(R,[int(4),int(3),int(2),int(1)],WF),
                       ground_det_wait_flag(WF), R=[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4))])).

:- block permutation_sequence_wf(-,-,?).
permutation_sequence_wf(SeqFF,Type,WF) :- nonvar(SeqFF),
    custom_explicit_sets:dom_range_for_specific_closure(SeqFF,FFDomain,FFRange,function(bijection)),!,
    equal_object_wf(FFRange,Type,permutation_sequence_wf_1,WF),
    is_sequence_domain(FFDomain,WF).
permutation_sequence_wf(Seq,Type,WF) :-
   expand_and_convert_to_avl_set_warn(Seq,AER,permutation_sequence_wf,'ARG : perm(?)'),!,
  %% print_term_summary(checking_avl_perm_seq(AER,Type)),nl, %%
   is_avl_sequence(AER),
   %% print(is_sequence),nl, %%
   is_injective_avl_relation(AER,Range),
%%   print_term_summary(is_injective(Range)), %%
   kernel_objects:equal_object_wf(Range,Type,permutation_sequence_wf_2,WF).
permutation_sequence_wf(Seq,Type,WF) :-
   try_expand_custom_set(Seq,ESeq),
   cardinality_peano_wf(Type,Card,WF),
   when(ground(Card), (setup_sequence(Card,SkelSeq,1), %print(setup(SkelSeq,ESeq)),nl,
                        CardGround=true,
                       kernel_objects:equal_object_wf(SkelSeq,ESeq,permutation_sequence_wf_3,WF))),
   %injective_sequence_wf(ESeq,Type,WF,LWF),
   surjective_iseq_0(SkelSeq,ESeq,Type,WF,Card,CardGround).
  % quick_all_different_range(ESeq,[],Type,WF).

:- block surjective_iseq_0(-,-,?,?,?,-).
surjective_iseq_0(SkelSeq,_ESeq,Type,WF,_Card,Ground) :-   % print(perm(SkelSeq,Type,_Card,Ground)),nl,
    nonvar(Ground),
    nonvar(SkelSeq),
    preference(use_clpfd_solver,true), % try and use an optimized version calling global_cardinality in CLPFD module
    get_global_cardinality_list(Type,YType,GCL,_),
    % this dramatically reduces runtime for NQueens40_perm; maybe we should do this only when necessary, i.e., when surjective_iseq blocks on PreviousRemoveDone
    % check why it slows down SortByPermutation_v2
    !,
    global_cardinality_range(SkelSeq,[],YType,GCL,WF).
surjective_iseq_0(_,ESeq,Type,WF,Card,_) :-
    %quick_propagate_range(ESeq,Type,WF), % ensure that we propagate type information to all elements; p:perm(5..20) & p(10)=21 will fail straightaway (surjective_iseq will block);
      % but this slows down EulerWay.mch ; maybe because it sets up enumerators ? TO DO: investigate
    surjective_iseq(ESeq,Type,WF,Card).

:- use_module(library(lists),[maplist/2]).
%:- use_module(clpfd_interface,[clpfd_alldifferent/1]).
% collect range and then call CLPFD global_cardinality using GCL (Global Cardinality List Ki-Vi)
:- block global_cardinality_range(-,?,?,?,?).
global_cardinality_range([],Acc,_Type,GCL,WF) :-
   %print(global_cardinality(Acc,_Type,GCL)),nl,
   clpfd:global_cardinality(Acc,GCL,[consistency(value)]),
   add_fd_variables_for_labeling(Acc,WF). % this is needed for efficiency for NQueens40_perm !!
global_cardinality_range([(_,Y)|T],Acc,Type,GCL,WF) :-
   get_simple_fd_value(Type,Y,FDYVAL),
   global_cardinality_range(T,[FDYVAL|Acc],Type,GCL,WF).


:- use_module(b_global_sets,[all_elements_of_type/2,b_integer_set/1]).
% try and convert a B set into a list suitable for calling clpfd:global_cardinality
% get_global_cardinality_list(avl_set(A) % TO DO: extend to integer_lists
get_global_cardinality_list(global_set(G),Type,GCL,list) :- !,
   all_elements_of_type(G,Values),
   (b_integer_set(G) -> Type=integer ; Type=global(G)),
   findall(X-1,(get_simple_fd_value(Type,VV,X),member(VV,Values)),GCL).
get_global_cardinality_list(avl_set(A),Type,GCL,list) :- !,
   A = node(TopValue,true,_,_,_),
   get_simple_fd_value(Type,TopValue,_), % we have CLPFD values
   avl:avl_domain(A,Values),
   findall(X-1,(get_simple_fd_value(Type,VV,X),member(VV,Values)),GCL).
get_global_cardinality_list(Set,integer,GCL,interval(L1,U1)) :- nonvar(Set),
   is_interval_closure_or_integerset(Set,L1,U1), number(L1),number(U1),
   global_cardinality_list_interval(L1,U1,GCL).

global_cardinality_list_interval(From,To,[]) :- From>To, !.
global_cardinality_list_interval(From,To,[From-1|T]) :-
   F1 is From+1, global_cardinality_list_interval(F1,To,T).

%try_get_simple_fd_value(Type,V,Val) :- nonvar(V),get_simple_fd_value(Type,V,Val).
get_simple_fd_value(integer,int(X),X).
get_simple_fd_value(global(T),fd(X,T),X).
% try_get_simple_fd_value(pred_false,0). try_get_simple_fd_value(pred_true,1). ??
% TO DO: maybe also treat pairs ? but we need complete values; see module clpfd_lists !

setup_sequence(0,[],_).
setup_sequence(s(C),[(int(Nr),_)|T], Nr ) :-
   N1 is Nr+1,
   setup_sequence(C,T,N1).

:- block surjective_iseq(?,?,?,-),surjective_iseq(?,-,?,?), surjective_iseq(-,?,?,?).
surjective_iseq(avl_set(S),Type,WF,Done) :- expand_custom_set(avl_set(S),ES),
  surjective_iseq(ES,Type,WF,Done).
surjective_iseq(closure(P,T,B),Type,WF,Done) :- expand_custom_set(closure(P,T,B),ES),
  surjective_iseq(ES,Type,WF,Done).
% no case for global_set: cannot be a relation
surjective_iseq([],T,WF,_) :- empty_set_wf(T,WF).
surjective_iseq([(int(_Nr),El)|Tail],Type,WF,_PreviousRemoveDone) :-
  % print(surjective_iseq([(int(_Nr),El)|Tail],Type,WF)),nl,
   remove_element_wf(El,Type,NType,WF,Done),
  % when(nonvar(Done), (print(removed(_Nr,El)),nl)),
   surjective_iseq(Tail,NType,WF,Done).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_permutation_sequence([(int(1),int(22))],[int(22)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(33)),(int(1),int(22))],[int(22),int(33)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(33)),(int(1),int(23))],[int(23),int(33),int(44)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_permutation_sequence([(int(2),int(44)),(int(1),int(44))],[int(44)],WF),WF)).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(2),int(2)),(int(1),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
                        ground_det_wait_flag(WF),R = [(int(1),int(2)),(int(2),int(1))] )).
:- assert_must_fail((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
   ground_det_wait_flag(WF), R = [(int(1),fd(1,'Name')),(int(3),fd(2,'Name')),(int(2),fd(3,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,[int(1)],WF),
      ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(1))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),_WF),R = [])).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
        ground_det_wait_flag(WF),R = [(int(1),fd(1,'Name')),(int(2),fd(2,'Name'))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,[int(1),int(2)],WF),
        ground_det_wait_flag(WF),R = [(int(1),int(1)),(int(2),int(3))] )).
:- assert_must_succeed((bsets_clp:not_permutation_sequence(R,global_set('Name'),WF),
        ground_det_wait_flag(WF),R = [(int(2),fd(1,'Name')),(int(1),fd(1,'Name'))] )).
:- block not_permutation_sequence(-,?,?).
not_permutation_sequence(SeqFF,Type,WF) :- nonvar(SeqFF),
    custom_explicit_sets:dom_range_for_specific_closure(SeqFF,FFDomain,FFRange,function(bijection)),!,
    equality_objects_wf(FFRange,Type,Result,WF),
    when(nonvar(Result),(Result=pred_false -> true ; not_is_sequence_domain(FFDomain,WF))).
not_permutation_sequence(Seq,Type,WF) :-
    kernel_tools:ground_value_check(Seq,SV),
    when((nonvar(SV),ground(Type)),not_permutation_sequence1(Seq,Type,WF)).
%:- block not_permutation_sequence1(?,?,?,-).
% TO DO: ? provide a version for avl trees ?
not_permutation_sequence1(Seq,Type,WF) :-
   expand_custom_set_to_list(Seq,ESeq,Done,not_permutation_sequence1),
   not_permutation_sequence2(ESeq,Type,WF,Done).

% here we could have a choice point in WF0
:- block not_permutation_sequence2(?,?,?,-).
not_permutation_sequence2(Seq,Type,WF,_) :- not_injective_sequence(Seq,Type,WF).
not_permutation_sequence2(Seq,Type,WF,_) :-
    injective_sequence_wf(Seq,Type,WF), not_surjective_seq(Seq,Type,WF).

:- block not_surjective_seq(-,?,?).
not_surjective_seq([],T,WF) :- not_empty_set_wf(T,WF).
not_surjective_seq([(int(_),El)|Tail],Type,WF) :-
   delete_element_wf(El,Type,NType,WF),
   not_surjective_seq(Tail,NType,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(1),int(22))],int(1),_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22))],int(2),_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22))],int(3),_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:size_of_sequence([(int(2),int(22)),(int(1),int(22)),(int(3),int(33))],int(3),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        R = int(2))).
:- assert_must_succeed((preferences:preference(use_clpfd_solver,false) -> true
                        ; bsets_clp:size_of_sequence(X,R,_WF), R=int(RI),
                          clpfd_interface:clpfd_geq2(RI,2,_), nonvar(X), X = [(I1,_),(I2,_)|T],
                          I1==int(1), I2==int(2), T=[], RI==2 )).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,R,_WF),X = [(int(1),_),(int(2),_)],R = int(2))).
:- assert_must_succeed((bsets_clp:size_of_sequence(X,_R,_WF),X =[(int(1),_),(int(2),_)] )).
:- assert_must_succeed_any((bsets_clp:size_of_sequence(X,int(2),_WF),nonvar(X),X=[_|Y],nonvar(Y),Y=[_|Z],Z==[])).
:- assert_must_succeed((bsets_clp:size_of_sequence([],int(0),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([],int(0),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([(int(1),int(4))],int(1),_WF))).
:- assert_must_succeed((bsets_clp:size_of_sequence([],_,_WF))).
:- assert_must_fail((bsets_clp:size_of_sequence(X,int(1),_WF),
                        X = [(int(1),_),(int(2),_)|_])).
:- block size_of_sequence(-,-,?).
size_of_sequence(Seq,int(Res),WF) :- size_of_sequence1(Seq,Res,WF),
   set_up_sequence_skel(Seq,Res).

% setup sequence skeleton if we have some CLPFD bounds information about the size
% currently still quite limited: only sets up if sequence is a variable; + does the setup only once
:- use_module(clpfd_interface,[clpfd_domain/3]).
set_up_sequence_skel(Seq,Res) :- %print(size(Seq,Res)),nl,
   var(Seq), % to do: also deal with cases when Seq partially instantiated
   var(Res),
   preferences:preference(use_clpfd_solver,true),
   !,
   clpfd_interface:clpfd_geq2(Res,0,_), % assert that size must not be negative
   clpfd_interface:post_constraint( (Res#>0) #<=> Trigger), % generate reified trigger for when we can instantiate Seq
   set_up_sequence_skel_aux(Seq,Res,Trigger).
set_up_sequence_skel(_,_). % TO DO: check if Size interval shrinks
:- block set_up_sequence_skel_aux(-,?,-).
set_up_sequence_skel_aux(Seq,_Res,_Trigger) :-
   nonvar(Seq),!.  % to do: also deal with cases when Seq partially instantiated
set_up_sequence_skel_aux(Seq,Res,_Trigger) :-
   clpfd_domain(Res,FDLow,FDUp), % FDLow could also be 0
   %print(setup_sequence(Seq,Res,FDLow)),nl, % <=>
   gen_sequence_skeleton(1,FDLow,FDUp,S), %print(S),nl,
   Seq=S.
gen_sequence_skeleton(N,M,FDUp,S) :- N>M,!,(FDUp==M -> S=[] ; true).
gen_sequence_skeleton(N,Max,FDUp,[(int(N),_)|T]) :-
    N1 is N+1,
    gen_sequence_skeleton(N1,Max,FDUp,T).

:- block size_of_sequence1(-,-,?).
size_of_sequence1(Seq,ResInt,WF) :-
  nonvar(Seq),is_custom_explicit_set_nonvar(Seq),
  size_of_custom_explicit_set(Seq,Size,WF),!,
  equal_object_wf(Size,int(ResInt),size_of_sequence1,WF).
  /* TO DO: CHECK BELOW: would it not  be better to use cardinality ?? */
/*
size_of_sequence1(Seq,Size,WF) :- !,finite_cardinality_as_int(Seq,int(Size),WF), check_indexes(Seq,Size).

   construct_interval_closure(1,Size,Domain),
   total_function_wf(FF,Domain,Range,_WF)
% we could also call total_function 1..Size --> _RangeType; would setup domain ?
:- block check_indexes(-,?).
check_indexes([],_) :- !.
check_indexes([(int(X),_)|T],Size) :- !,
   less_than_equal_direct(X,Size), check_indexes(T,Size).
check_indexes(_,_).
*/
size_of_sequence1(Seq,Size,_WF) :- Size==0,!, empty_sequence(Seq).
size_of_sequence1(Seq,Size,_WF) :-
   expand_custom_set_to_list(Seq,ESeq,_,size_of_sequence1),  % print_term_summary(size1(Seq,ESeq,Size)),
   ((var(ESeq),nonvar(Size)) -> size_of_var_seq(ESeqR,0,Size),
                                ESeqR=ESeq % unify after to do propagation in one go, without triggering coroutines inbetween
                ; size_of_seq2(ESeq,0,Size),
                  ((var(Size),var(ESeq)) -> less_than_equal_direct(0,Size) % propagate that Size is positive
                        ; true)
    ).
           /* small danger of expanding closure while still var !*/
:- block size_of_seq2(-,?,-).
size_of_seq2([],Size,Size).
size_of_seq2([I|Tail],SizeSoFar,Res) :-
    S2 is SizeSoFar + 1,
    check_index(I,Res), % don't instantiate I yet; allow other kernel_predicates to freely instantiate it
    less_than_equal_direct(S2,Res),
    %(ground(Res) -> safe_less_than_equal(size_of_seq2,S2,Res) ; true),
    size_of_seq2(Tail,S2,Res).
size_of_var_seq([],Size,Size).
size_of_var_seq([(int(S2),_)|Tail],SizeSoFar,Res) :-
    S2 is SizeSoFar + 1,safe_less_than_equal(size_of_var_seq,S2,Res),
    (var(Tail) -> size_of_var_seq(Tail,S2,Res) ; size_of_seq2(Tail,S2,Res)).


:- block check_index(-,?).
check_index((I,_),Res) :- check_index1(I,Res).
:- block check_index1(-,?).
check_index1(int(Idx),Res) :- less_than_equal_direct(Idx,Res).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(1),int(22))],[(int(2),int(22)),(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[],[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(2),int(44)),(int(1),int(22))],[(int(1),int(33)),(int(3),int(44)),(int(2),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:prepend_sequence(int(33),[(int(1),int(22))],[(int(1),int(22)),(int(2),int(33))],WF),WF)).
:- assert_must_succeed((bsets_clp:prepend_sequence(int(7),[],[(int(1),int(7))],_WF))).
:- assert_must_succeed((bsets_clp:prepend_sequence(int(7),X,R,_WF),
                        X = [(int(2),int(4)),(int(1),int(3))],
                        kernel_objects:equal_object(R,[(int(1),int(7)),(int(2),int(3)),(int(3),int(4))]))).
:- block prepend_sequence(?,-,-,?).
prepend_sequence(El,Seq,Res,_WF) :- Seq==[],!,
   equal_object_optimized([(int(1),El)],Res,prepend_sequence).
prepend_sequence(El,Seq,Res,WF) :- nonvar(Seq),is_custom_explicit_set(Seq,prepend_sequence),
   prepend_custom_explicit_set(Seq,El,ERes),!, %print(prepend(Seq,El)),nl,
   equal_object_wf(Res,ERes,prepend_sequence,WF).
prepend_sequence(El,Seq,Res,WF) :- %print_message(prepend_sequence(El,Seq,Res)),
   equal_cons_wf(Res,(int(1),El),ShiftSeq,WF),
   shift_seq_indexes(Seq,1,ShiftSeq,WF).
%   print_message(prepend_result(El,Res)).

:- block shift_seq_indexes(-,-,?,?),shift_seq_indexes(-,?,-,?).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :-
   Offset == 0,!,  equal_sequence(Seq,ShiftedSeq,WF).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :- nonvar(Seq),!,
   expand_custom_set_to_list(Seq,ESeq,_,shift_seq_indexes),
   shift_seq_indexes2(ESeq,Offset,ShiftedSeq,WF,Done),
   (Done == done
      -> true
      ;  % also propagate in the other way: TO DO: make a more efficient fine-grained two-ways propagation; maybe using CHR
         NegOffset is -Offset,
         expand_custom_set_to_list(ShiftedSeq,ESeq1,_,shift_seq_indexes),
         shift_seq_indexes2(ESeq1,NegOffset,ESeq,WF,_)).
shift_seq_indexes(Seq,Offset,ShiftedSeq,WF) :- NegOffset is -Offset,
   % compute in the other direction; TO DO: make a more efficient fine-grained two-ways propagation; maybe using CHR
   expand_custom_set_to_list(ShiftedSeq,ESeq,_,shift_seq_indexes),
   shift_seq_indexes2(ESeq,NegOffset,Seq,WF,Done),
   (Done == done
      -> true
      ;  % also propagate in the original way:
         expand_custom_set_to_list(Seq,ESeq1,_,shift_seq_indexes),
         shift_seq_indexes2(ESeq1,Offset,ESeq,WF,_)).

:- block shift_seq_indexes2(-,?,?,?,?).
shift_seq_indexes2([],_,R,WF,Done) :- !, Done = done, empty_set_wf(R,WF).
shift_seq_indexes2([Pair|Tail],Offset,Res,WF,Done) :- !,
   Pair = (int(N),El),
   equal_cons_wf(Res,(int(NewN),El),ShiftTail,WF),
   int_plus(int(N),int(Offset),int(NewN)),
   shift_seq_indexes2(Tail,Offset,ShiftTail,WF,Done).
shift_seq_indexes2(Seq,Offset,Res,WF,Done) :-
    add_internal_error('Unexpected set argument: ',shift_seq_indexes2(Seq,Offset,Res,WF,Done)), fail.

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([(int(1),int(22))],int(33),[(int(2),int(33)),(int(1),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([],int(33),[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:append_sequence([(int(2),int(44)),(int(1),int(22))],int(33),[(int(1),int(22)),(int(3),int(33)),(int(2),int(44))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:append_sequence([(int(1),int(22))],int(33),[(int(1),int(33)),(int(2),int(22))],WF),WF)).
:- assert_must_succeed((bsets_clp:append_sequence([],int(7),[(int(1),int(7))],_WF))).
:- assert_must_succeed((bsets_clp:append_sequence(X,int(7),R,_WF),
                        X = [(int(2),int(4)),(int(1),int(3))],
                        kernel_objects:equal_object(R,[(int(1),int(3)),(int(2),int(4)),(int(3),int(7))]))).

:- block append_sequence(-,?,-,?).
append_sequence(Seq,El,Res,_WF) :- Seq==[],!,
   equal_object_optimized([(int(1),El)],Res,append_sequence).
append_sequence(Seq,El,Res,WF) :- %print_term_summary(append_sequence(Seq,El,Res)),
   nonvar(Seq),is_custom_explicit_set_nonvar(Seq),
   append_custom_explicit_set(Seq,El,ERes),!,
   equal_sequence(Res,ERes,WF).
append_sequence(Seq,El,Res,WF) :- equal_cons_wf(Res,(int(NewSize),El),ResT,WF),
     append_sequence2(Seq,ResT,NewSize,WF).

:- block append_sequence2(-,?,?,?).
append_sequence2(Seq,ResT,NewSize,WF) :-
   try_expand_custom_set(Seq,ESeq), equal_sequence(ESeq,ResT,WF),
   size_of_sequence(ESeq,Size,WF),
   int_plus(Size,int(1),int(NewSize)).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:prefix_sequence([(int(1),int(22))],int(1),[(int(1),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(bsets_clp:prefix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(2),[(int(1),int(11)),(int(2),int(22))]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:prefix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(3),[(int(1),int(11)),(int(2),int(22))]))).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(1),X),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(0),[]),X = [(int(1),int(1))])).
:- assert_must_abort_wf((bsets_clp:prefix_sequence_wf(X,int(-1),_R,WF),X = [(int(1),int(1))]),WF).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(2),Y),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(1)),(int(2),int(3))]) )).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(1),Y),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(1))]) )).
:- assert_must_succeed((bsets_clp:prefix_sequence(X,int(3),Y),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,X) )).

prefix_sequence(Seq,N,R) :- init_wait_flags(WF),
    prefix_sequence_wf(Seq,N,R,WF),
    ground_wait_flags(WF).

% Prefix of a sequence (s /|\ n)
prefix_sequence_wf(Seq,int(Num),Res,WF) :-
   prefix_sequence1(Seq,Num,Res,WF),
   propagate_size(Res,Num,WF).

:- block propagate_size(-,-,?).
propagate_size(Res,Num,WF) :- %print(prop(Res,Num,WF)),nl,
   var(Res),!, %print(setup_sequence(Num)),nl,
   (Num<0 -> preferences:preference(disprover_mode,false) % don't do anything; we may want to generate WD error
    ; size_of_sequence(Res,int(Num),WF)).
propagate_size(_,Num,_) :- number(Num), !. % no need to propagate
propagate_size(_,_Num,_) :- preferences:preference(find_abort_values,true),
    !. % do not propagate as we could prevent detection of WD errors below
propagate_size([],Num,_WF) :- !,
   Num=0. % Note: this could prevent a wd-error being detected
propagate_size(avl_set(A),Num,WF) :- var(Num),
   % with partially instantated sets we get slowdowns (SimpleCSGGrammar2_SlowCLPFD)
   % TO DO: treat list skeletons
   !,
   size_of_sequence(avl_set(A),int(Num),WF). % Note: this could prevent a wd-error being detected
propagate_size(_,_,_). % should we also propagate the other way around ?

:- block prefix_sequence1(-,?,?,?), prefix_sequence1(?,-,?,?).
prefix_sequence1(_Seq,Num,Res,WF) :- Num==0,!, empty_set_wf(Res,WF).
prefix_sequence1(_Seq,Num,_Res,WF) :- Num<0,!,  % according to version 1.8.8 of Atelier-B manual Num must be in 0..size(_Seq)
   add_wd_error('negative index in prefix_sequence (/|\\)! ', Num,WF).
prefix_sequence1(Seq,Num,Res,WF) :-
   is_custom_explicit_set(Seq,prefix),
   prefix_of_custom_explicit_set(Seq,Num,ERes,WF),!, % TO DO: check Num <= size(Seq)
   equal_object_wf(Res,ERes,prefix_sequence1,WF).
prefix_sequence1(Seq,Num,Res,WF) :-
  expand_custom_set_to_list(Seq,ESeq,_,prefix_sequence1),
  unify_same_index_elements(Res,ESeq),
  unify_same_index_elements(Seq,Res),
  prefix_seq(ESeq,Num,0,Res,WF). %,print(prefix_seq_exit(ESeq,Num,0,Res)),nl
:- block  prefix_seq(-,?,?,?,?).
prefix_seq([],Max,Sze,Res,WF) :-
   (less_than_direct(Sze,Max)
         -> add_wd_error('index larger than size of sequence in prefix_sequence (/|\\)! ', (Max,Sze),WF)
          ; true),
    empty_set_wf(Res,WF).
          %(less_than(int(_Sze),int(_Max))
         %  ->  (print_message('Index bigger than sequence size in prefix_sequence (/|\\) !'),
         %       print_message(Max))
         %     /* in the AtelierB book this is allowed, in Wordsworth + AMN on web it is not ?? */
         %  ;  true).
prefix_seq([(int(N),El)|Tail],Max,SizeSoFar,Res,WF) :-
    prefix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF).
:- block prefix_seq2(-,?,?,?,?,?,?).
prefix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF) :- % SizeSoFar is always ground
    (less_than_equal_direct(N,Max), equal_cons_wf(Res,(int(N),El),PTail,WF)
      ;
     less_than_direct(Max,N), equal_object_wf(Res,PTail,prefix_seq2,WF)
    ),
    ( SizeSoFar<N -> NewSizeSoFar=N ; NewSizeSoFar = SizeSoFar ),
    prefix_seq(Tail,Max,NewSizeSoFar,PTail,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:suffix_sequence([(int(1),int(22))],int(0),[(int(1),int(22))],WF),ground_det_wait_flag(WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(bsets_clp:suffix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(1),[(int(1),int(22)),(int(2),int(33))],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(bsets_clp:suffix_sequence([(int(2),int(22)),(int(3),int(33)),(int(1),int(11))],int(2),[(int(1),int(22)),(int(2),int(33))],_WF))).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(0),X,_WF),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(1),[],_WF),X = [(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(2),Y,_WF),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(4))]) )).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(1),Y,_WF),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(3)),(int(2),int(4))]) )).
:- assert_must_succeed((bsets_clp:suffix_sequence(X,int(2),Y,_WF),
             X = [(int(2),int(3)),(int(3),int(4)),(int(1),int(1))],
             kernel_objects:equal_object(Y,[(int(1),int(4))]) )).
:- assert_must_abort_wf(bsets_clp:suffix_sequence([(int(1),int(11)),(int(2),int(22))],int(-1),_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:suffix_sequence([(int(1),int(11)),(int(2),int(22))],int(3),_R,WF),WF).

% kernel_waitflags:assert_must_abort2_wf(bsets_clp:suffix_sequence([int(11),int(22)],int(-1),_R,WF),WF)

% suffix of a sequence (s \|/ n); also called restrict at tail (Atelier B) or Drop
:- block suffix_sequence(-,?,?,?).
suffix_sequence(Seq,int(Num),Res,WF) :- % print(suffix_sequence(Seq)),nl,
   suffix_sequence1(Seq,Num,Res,WF).
:- block suffix_sequence1(?,-,?,?).
suffix_sequence1(Seq,Num,Res,WF) :- Num==0, !, equal_object_wf(Res,Seq,suffix_sequence1_1,WF).
suffix_sequence1(_Seq,Num,_Res,WF) :- Num<0, !, add_wd_error('negative index in suffix_sequence (\\|/)! ', Num,WF).
suffix_sequence1(Seq,Num,Res,WF) :- is_custom_explicit_set(Seq,suffix),
   suffix_of_custom_explicit_set(Seq,Num,ERes,WF),!,
   equal_object_wf(Res,ERes,suffix_sequence1_2,WF).
suffix_sequence1(Seq,Num,Res,WF) :-
  expand_custom_set_to_list(Seq,ESeq,_,suffix_sequence), suffix_seq(ESeq,Num,0,Res,WF).
:- block suffix_seq(-,?,?,?,?).
suffix_seq([],Max,Sze,Res,WF) :-
   (less_than_direct(Sze,Max)
          -> add_wd_error('index larger than size of sequence in suffix_sequence (\\|/)! ', '>'(Max,Sze),WF)
          ; true), empty_set_wf(Res,WF).
suffix_seq([(int(N),El)|Tail],Max,SizeSoFar,Res,WF) :-
    suffix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF).
:- block suffix_seq2(-,?,?,?,?,?,?).
suffix_seq2(N,El,Tail,Max,SizeSoFar,Res,WF) :-
    (less_than_equal_direct(N,Max), equal_object_wf(Res,PTail,suffix_seq2,WF)
      ;
     less_than_direct(Max,N),int_minus(int(N),int(Max),int(NN)),
     equal_cons_wf(Res,(int(NN),El),PTail,WF)
    ),
    (N>SizeSoFar -> (NewSizeSoFar=N)
                  ; (NewSizeSoFar = SizeSoFar)),
    suffix_seq(Tail,Max,NewSizeSoFar,PTail,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:concat_sequence([],[(int(1),int(33))],[(int(1),int(33))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:concat_sequence([(int(1),int(22)),(int(2),int(33))],[(int(1),int(33)),(int(2),int(44))],[(int(2),int(33)),(int(3),int(33)),(int(1),int(22)),(int(4),int(44))],WF),WF)). % not wfdet because of pending label_el_nr from clpfd_lists
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:concat_sequence([(int(1),int(22))],[(int(1),int(33))],[(int(2),int(33)),(int(1),int(22))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:concat_sequence([(int(1),int(22))],[(int(1),int(33))],[(int(2),int(22)),(int(1),int(33))],WF),WF)).
:- assert_must_succeed((bsets_clp:concat_sequence([],X,Y,_WF),
                       X = [(int(1),int(1))], Y==X)).
:- assert_must_succeed((bsets_clp:concat_sequence(X,[],Y,_WF), X = [(int(1),int(1))], Y==X)).
:- assert_must_succeed((bsets_clp:concat_sequence([(int(1),int(1))],[],Y,_WF), Y==[(int(1),int(1))])).
:- assert_must_succeed((bsets_clp:concat_sequence(X,X,Y,_WF),
                X = [(int(1),int(1))], kernel_objects:equal_object(Y,[(int(1),int(1)),(int(2),int(1))]))).
:- assert_must_succeed((bsets_clp:concat_sequence(X,X,Y,_WF),
      X = [(int(2),int(88)),(int(1),int(77))],
      kernel_objects:equal_object(Y,[(int(1),int(77)),(int(2),int(88)),(int(3),int(77)),(int(4),int(88))]))).

:- block /* concat_sequence(-,-,?,?), */
         concat_sequence(?,-,-,?), concat_sequence(-,?,-,?).
concat_sequence(S1,S2,Res,WF) :- Res==[],!, empty_set_wf(S1,WF), empty_set_wf(S2,WF).
concat_sequence(S1,S2,Res,WF) :-
   (var(S1),var(S2) -> get_wait_flag(2,concat,WF,LWF)  % we have at least two solutions; TODO maybe use cardinality as wait_flag?
    ; LWF=1),
    concat_sequence2(LWF,S1,S2,Res,WF).

:- block concat_sequence2(-,?,-,?,?), concat_sequence2(-,-,?,?,?).
concat_sequence2(_,S1,S2,Res,WF) :- S1==[],!,equal_sequence(S2,Res,WF).
concat_sequence2(_,S1,S2,Res,WF) :- S2==[],!,equal_sequence(S1,Res,WF).
concat_sequence2(LWF,S1,S2,Res,WF) :-
   try_expand_and_convert_to_avl_with_check(S1,AS1,concat1),
   try_expand_and_convert_to_avl_with_check(S2,AS2,concat2),
   concat_sequence3(LWF,AS1,AS2,Res,WF).

concat_sequence3(_,S1,S2,Res,WF) :- nonvar(S1),is_custom_explicit_set(S1,concat_sequence),
   concat_custom_explicit_set(S1,S2,ERes),!,
   %print_term_summary(concat(S1,S2,ERes)),nl,
   equal_sequence(Res,ERes,WF).
concat_sequence3(_LWF,S1,S2,Res,WF) :-
  % print_term_summary(ordinary_concat_sequence(_LWF,S1,S2,Res)),
  %try_expand_custom_set(S1,ES1),
  size_of_sequence(S1,int(Size1),WF),
  (number(Size1) -> true
    ; size_of_sequence(S2,Size2,WF),
      size_of_sequence(Res,SizeRes,WF),
      int_minus(SizeRes,Size2,int(Size1)),
      in_nat_range_wf(int(Size1),int(0),SizeRes,WF)
      % is this required: ?? ,in_nat_range_wf(Size2,int(0),SizeRes,WF)
  ),
   %print(call_concat_sequence_aux(Size1,S1,S2,Res,WF)),nl,
   concat_sequence_aux(Size1,S1,S2,Res,WF).


:- assert_must_succeed( (bsets_clp:equal_sequence([(int(1),A)|T1],[(int(1),int(22))|T2],_WF),
A==int(22),T2=[],T1==[] )) .
:- assert_must_succeed( (bsets_clp:equal_sequence([(int(1),A)|T],avl_set(node((int(2),string(a)),true,0,node((int(1),string(c)),true,0,empty,empty),node((int(3),string(b)),true,0,empty,empty))),_WF),
A==string(c), T = [(int(2),B)|T2],  B==string(a), T2=[(int(3),C)], C==string(b)) ).
% equal_object optimized for sequences; can infer that pairs with same index have same value
% TO DO: complete and make more efficient
%equal_sequence(X,Y,_WF) :- nonvar(X),nonvar(Y),
%   is_custom_explicit_set(X,eval_sequence), is_custom_explicit_set(Y,eval_sequence),!,
%   equal_explicit_sets(X,Y).
equal_sequence(X,Y,WF) :- nonvar(X),nonvar(Y),
   get_seq_head(X,XI,XEl,XT), get_seq_head(Y,YI,YEl,YT), XI==YI,!,
   % THIS CURRENTLY ONLY CHECKS FRONTMOST indexes
   %print(eq_seq(XI,XEl,YI,YEl,XT,YT)),nl,
   equal_object_wf(XEl,YEl,equal_sequence_1,WF), %print(done(XEl,YEl)),nl,
   equal_sequence(XT,YT,WF).
equal_sequence(X,Y,WF) :-
   %print(equal_sequence(X,Y)),nl,
   /* (is_custom_explicit_set(Y) -> monitor_equal_sequence_againts_custom_set(X,Y,WF)
    ; is_custom_explicit_set(X) -> monitor_equal_sequence_againts_custom_set(Y,X,WF)
    ; true),  does not seem to buy anything; equal_object already powerful enough */
   equal_object_wf(X,Y,equal_sequence_2,WF).

% enforces the constraint that there is only one possible elemenent per index:
%:- block monitor_equal_sequence_againts_custom_set(-,?,?).
%monitor_equal_sequence_againts_custom_set([],_,_) :- !.
%monitor_equal_sequence_againts_custom_set([El|T],CS,WF) :- !,
%   element_of_custom_set_wf(El,CS,WF),
%   monitor_equal_sequence_againts_custom_set(T,CS,WF).
%monitor_equal_sequence_againts_custom_set(_,_,_).

get_seq_head([(Idx,El)|Tail],Idx,El,Tail).
%get_seq_head(avl_set(AVL),Idx,El,Tail) :-  does not seem to buy anything; equal_object already powerful enough
%   custom_explicit_sets:avl_min_pair(AVL,Idx,El),
%   custom_explicit_sets:direct_remove_element_from_avl(AVL,(Idx,El),Tail). % TO DO: only compute if indexes ==


:- block concat_sequence_aux(-,?,?,?,?).
concat_sequence_aux(Size1,_S1,_S2,Res,WF) :- nonvar(Res),Res=avl_set(_),
   size_of_custom_explicit_set(Res,int(RSize),WF), number(RSize),
   Size1 > RSize,!, % S1 is longer than Res; no solution (prevent WD error below)
   fail.
concat_sequence_aux(Size1,S1,S2,Res,WF) :- nonvar(Res),Res=avl_set(_),
   % split Result into prefix and suffix
   prefix_of_custom_explicit_set(Res,Size1,Prefix,WF), % we could call versions which do not check WD
   suffix_of_custom_explicit_set(Res,Size1,Postfix,WF),
   %print(split_avl_set(Size1,S1,S2,Res,Prefix,Postfix,WF)),nl,
   !,
   equal_sequence(S1,Prefix,WF), equal_sequence(S2,Postfix,WF).
concat_sequence_aux(Size1,S1,S2,Res,WF) :-
   % print(concat_size1(S1,Size1)),nl,
  shift_seq_indexes(S2,Size1,NewS2,WF),  %%      print(shifted(S2,Size1,NewS2)),nl,
  % We can do something stronger than disjoint union: we know that the indexes are disjoint!
  % Hence: if (int(X),Y) : Res & (int(X),Z) : S1   =>   Y=Z
  % Hence: if (int(X),Y) : Res & (int(X),Z) : S2   =>   Y=Z
  unify_same_index_elements(S1,Res),
  unify_same_index_elements(Res,S1),
  unify_same_index_elements(NewS2,Res),
  unify_same_index_elements(Res,NewS2),
  disjoint_union_wf(S1,NewS2,Res,WF). % , print(disj_res(S1,NewS2,Res)),nl.

% Check if (int(X),Y) pairs in Seq2 have a matching (int(X),Z) in Seq1 and then unify(Y,Z)
:- block unify_same_index_elements(-,?).
unify_same_index_elements(avl_set(A),Seq) :- !, % print(check_seq(Seq,A)),nl,
   unify_same_index_elements_aux(Seq,A).
unify_same_index_elements(_,_). % TO DO: maybe also support other partially instantiated lists

:- block unify_same_index_elements_aux(-,?).
unify_same_index_elements_aux([],_) :- !.
unify_same_index_elements_aux([(int(Idx),El)|T],A) :- !,
   try_find_index_element(Idx,El,A),
   unify_same_index_elements_aux(T,A).
unify_same_index_elements_aux(_,_).

:- block try_find_index_element(-,?,?).
try_find_index_element(Idx,El,AVL) :-
    avl_tools:avl_fetch_pair(int(Idx),AVL,AvlEl),
    % print(fetch(Idx,AvlEl)),nl,
    !,
    % We have found an entry with the same index: El and AvlEl must be identical:
    equal_object(El,AvlEl).
try_find_index_element(_Idx,_El,_AVL). % :- print(not_found(_Idx,_AVL)),nl.

% TO DO: add waitflags + use within partition_wf
% computes union of two sets which are guaranteed to be disjoint: means that if two of three sets known the other one can be determined

:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([int(3)],[int(2),int(1)],[int(1),int(3),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([],[int(2),int(1)],[int(1),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf([commutative],bsets_clp:disjoint_union_wf([int(1),int(2)],[],[int(2),int(1)],WF),WF)).
:- assert_must_succeed((bsets_clp:disjoint_union_wf([int(1)],[int(2)],Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:disjoint_union_wf(A,B,[int(1)],_WF),B=[H],H==int(1),A==[])).

disjoint_union_wf(Set1,Set2,Res,WF) :-
  (var(Res)
   -> disjoint_union_wf0(Set1,Set2,DRes,DRes,WF), equal_object_optimized(Res,DRes) % try and convert result to AVL
    ; disjoint_union_wf0(Set1,Set2,Res,Res,WF)).

% disjoint_union_wf0(Set1,Set2,UnionOfSet1Set2, SuperSet, WF)
:- block disjoint_union_wf0(-,-,-,?,?).
disjoint_union_wf0(Set1,Set2,Res,_,WF) :- Set1==[],!,equal_object_wf(Set2,Res,disjoint_union_wf0_1,WF).
disjoint_union_wf0(Set1,Set2,Res,_,WF) :- Set2==[],!,equal_object_wf(Set1,Res,disjoint_union_wf0_2,WF).
disjoint_union_wf0(Set1,Set2,Res,_,WF) :- Res==[],!,empty_set_wf(Set1,WF), empty_set_wf(Set2,WF).
disjoint_union_wf0(Set1,Set2,Res,FullRes,WF) :-
    ((nonvar(Set1);nonvar(Set2)) -> true ; get_cardinality_powset_wait_flag(Res,disjoint_union_wf0,WF,_Card,CWF)),
    disjoint_union0(Set1,Set2,Res,FullRes,WF,CWF).

:- block disjoint_union0(-,-,?,?,?,-), disjoint_union0(-,?,-,-,?,?), disjoint_union0(?,-,-,-,?,?).
disjoint_union0(Set1,Set2,Res,_,WF,_) :- Set1==[],!,equal_object_wf(Set2,Res,disjoint_union0_1,WF).
disjoint_union0(Set1,Set2,Res,_,WF,_) :- Set2==[],!,equal_object_wf(Set1,Res,disjoint_union0_2,WF).
disjoint_union0(S1,S2,Res,_F,WF,_CWF) :-  % print(disjoint_union0(S1,S2,Res,_F)),nl,
     ground_value(Res),
    ( ground_value(S1) -> !, %print(diff(Res,S1,S2)),nl,
                       check_subset_of_wf(S1,Res,WF), % TO DO: check if we can merge the check_subset and difference set in one predicate
                       difference_set_wf(Res,S1,S2,WF)
     ; ground_value(S2) -> !, %print(diff(Res,S2,S1)),nl,
                       check_subset_of_wf(S2,Res,WF),
                       difference_set_wf(Res,S2,S1,WF)
     ; var(S1),var(S2) -> !, %print(disj_enum(Res,CWF)),nl, % CWF nonvar
             % see test 1408; allows to generate subsets of Res and avoid enumeration warnings
                       check_subset_of_wf(S1,Res,WF),
                       %check_subset_of(S1,Res), % without waitflag: will generate ground version
                       difference_set_wf(Res,S1,S2,WF)
    ).
disjoint_union0(Set1,Set2,Res,_,WF,_) :- nonvar(Set1),
    is_custom_explicit_set_nonvar(Set1),
    union_of_explicit_set(Set1,Set2,Union), !,
    equal_object_wf(Union,Res,disjoint_union0_3,WF).
disjoint_union0(Set1,Set2,Res,Full,WF,_) :- expand_custom_set_to_list(Set1,ESet1,_,disjoint_union0_1),
    expand_custom_set_to_list(Set2,ESet2,_,disjoint_union0_2),
    %print(disjoint_union0(ESet1,ESet2,Res,WF)),nl,
    disj_union1(ESet1,ESet2,Res,Full,WF).

:- block disj_union1(-,-,?,?,?).
disj_union1(X,Y,Res,FullRes,WF) :-
   var(X) -> disj_union2(Y,X,Res,FullRes,WF) ; disj_union2(X,Y,Res,FullRes,WF).

disj_union2([],Y,Res,_,_WF) :- equal_object_optimized(Y,Res,disj_union2).
disj_union2([X|TX],Y,Res,FullRes,WF) :-
   remove_element_wf(X,Res,TR,WF), % was: equal_cons_wf(Res,X,TR,WF)  but error was that it could force X to be a certain value
   kernel_tools:ground_value_check(X,XV),
   (nonvar(XV) -> equal_cons_wf(Res,X,TR,WF)
    ; check_element_of_wf(X,FullRes,WF), % ensure that we set up proper constraints for X; e.g., for x \/ y = 1..10 & x /\ y = {}
      when(nonvar(XV), equal_cons_wf(Res,X,TR,WF))
    ), % ensure that we instantiate Res if TR known; otherwise we may get pending co-routines, e.g. test 506, SyracuseGrammar
   disj_union3(TX,Y,TR,FullRes,WF).

:- block disj_union3(-,-,-,?,?).
disj_union3(X,Y,Res,_,WF) :- Res==[],!,empty_set_wf(X,WF),empty_set_wf(Y,WF).
disj_union3(X,Y,Res,FullRes,WF) :- disj_union1(X,Y,Res,FullRes,WF).


:- block disjoint_union_generalized_wf(-,?,?).
%disjoint_union_generalized_wf([Set1|T],Res,_WF) :- T==[],!, % just one set; probably not covered at the moment (ast_cleanup simplifies partition with single set
%        equal_object(Set1,Res).
disjoint_union_generalized_wf(SetsOfSets,Res,WF) :-
        expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,disjoint_union_generalized_wf),
        disjoint_union_generalized2(ESetsOfSets,[],Res,WF). %, print(res(Res)),nl.
:- block disjoint_union_generalized2(-,?,?,?).
disjoint_union_generalized2([],S,Res,WF) :-  equal_object_optimized_wf(S,Res,disjoint_union_generalized2,WF).
disjoint_union_generalized2([H|T],UnionSoFar,Res,WF) :-
     disjoint_union_wf0(H,UnionSoFar,UnionSoFar2,Res,WF),
     %%  print_message(called_disjoint_union(H,UnionSoFar,UnionSoFar2)), %%
     disjoint_union_generalized2(T,UnionSoFar2,Res,WF).

% TO DO: if there are more than two sets: it may be interesting to set up constraint that
%        each set is a subset of the full set;
%  (would avoid enumeration warning in, e.g.,  x \/ y \/ z = 1..10 & x /\ y = {} & x /\ z = {} & y /\ z = {} & card(x)=card(y)+2 )

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:concatentation_of_sequences([(int(1),[]),(int(3),[(int(1),int(22)),(int(2),int(33))]),(int(2),[(int(1),int(11))])],
   [(int(1),int(11)),(int(2),int(22)),(int(3),int(33))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:concatentation_of_sequences([(int(1),[]),(int(2),[(int(1),int(33))])],[(int(1),int(33))],_WF))).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:concatentation_of_sequences([(int(1),[]),(int(2),[(int(1),int(55))])],Res,WF),
                        kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(Res,[(int(1),int(55))]) )).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:concatentation_of_sequences([(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])],Res,WF),
                        kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(Res,[(int(1),int(22)),(int(2),int(55))]) )).
:- block concatentation_of_sequences(-,?,?).
concatentation_of_sequences(SeqOfSeq,Res,WF) :- %print_message(conc(SeqOfSeq)),
    try_expand_and_convert_to_avl_with_check(SeqOfSeq,ES,conc),
    concs2(ES,Res,WF).

concs2(SeqOfSeq,Res,WF) :- is_custom_explicit_set(SeqOfSeq,conc),
   conc_custom_explicit_set(SeqOfSeq,CRes),!,
   equal_object_wf(CRes,Res,concs2,WF).
concs2(SeqOfSeq,Res,WF) :- empty_set_wf(SeqOfSeq,WF),empty_set_wf(Res,WF).
concs2(SeqOfSeq,Res,WF) :- not_empty_set_wf(SeqOfSeq,WF),
    front_sequence(SeqOfSeq,Front,WF),
    concatentation_of_sequences(Front,FrontRes,WF),
    last_sequence(SeqOfSeq,Last,WF),
    concat_sequence(FrontRes,Last,Res,WF).

:- assert_must_abort_wf(bsets_clp:tail_sequence([],_R,unknown,WF),WF).
:- assert_must_abort_wf(bsets_clp:tail_sequence([],[],unknown,WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check(
 bsets_clp:tail_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(5))],unknown,_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:tail_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],
[(int(1),int(5)),(int(2),int(6))],unknown,_WF)) ).
:- assert_must_succeed((bsets_clp:tail_sequence(X,R,unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5))],
                        kernel_objects:equal_object(R,[(int(1),int(5))]) )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(1),int(5))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5))] )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(1),int(5)),(int(2),int(7))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5)),(int(3),int(7))] )).
:- assert_must_succeed((bsets_clp:tail_sequence(X,[(int(2),int(7)),(int(1),int(5))],unknown,_),
                        X = [(int(1),int(6)),(int(2),int(5)),(int(3),int(7))] )).
:- block tail_sequence(-,?,?,?).
tail_sequence(S1,Res,Span,WF) :- is_custom_explicit_set(S1,tail_sequence),
  tail_sequence_custom_explicit_set(S1,TRes,Span,WF),!,
  equal_object_wf(TRes,Res,tail_sequence,WF).
tail_sequence(S1,Res,Span,WF) :- expand_custom_set_to_list(S1,ES1,_,tail_sequence),
  tail2(ES1,Res,Span,WF).

tail2([],_,Span,WF) :-
    add_wd_error_span('tail applied to empty sequence!',[],Span,WF).
tail2([H|T],Res,_Span,WF) :- domain_subtraction_wf([int(1)],[H|T],IntRes,WF),
   shift_seq_indexes(IntRes,-1,Res,WF).


:- assert_must_abort_wf(bsets_clp:first_sequence([],_R,unknown,WF),WF).
:- assert_must_abort_wf(bsets_clp:first_sequence([],int(1),unknown,WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:first_sequence([(int(1),int(4)),(int(2),int(5))],int(4),unknown,_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:first_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],int(4),unknown,_WF)) ).
:- assert_must_succeed((bsets_clp:first_sequence(X,R,unknown,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        R = int(2))).

:- block first_sequence(-,?,?,?).
%first_sequence(Seq,Res,Span,_) :- % print(first_seq(Seq,Res,Span)),nl,fail.
%  is_custom_explicit_set(Seq,first_sequence),
%  first_sequence_explicit_set(Seq,First), !,
%  equal_object(First,Res).
first_sequence([],_,Span,WF) :- !,add_wd_error_span('first applied to empty sequence!',[],Span,WF).
first_sequence(Seq,Res,Span,WF) :- apply_to(Seq,int(1),Res,Span,WF).



:- assert_must_abort_wf(bsets_clp:front_sequence([],_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:front_sequence([],[],WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:front_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(4))],_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:front_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],[(int(1),int(4)),(int(2),int(5))],_WF)) ).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:front_sequence(X,R,WF),
                        X = [(int(1),int(2)),(int(2),int(55))],kernel_waitflags:ground_wait_flags(WF),
                        kernel_objects:equal_object(R,[(int(1),int(2))]))).
:- assert_must_succeed((kernel_waitflags:init_wait_flags(WF),bsets_clp:front_sequence(X,R,WF),
                        X = [(int(3),int(33))|R], kernel_waitflags:ground_wait_flags(WF),
                        print(R),nl,
                        kernel_objects:equal_object(R,[(int(1),int(2)),(int(2),int(55))]) )).

front_sequence(Seq,Res,WF) :- front_sequence(Seq,Res,unknown,WF).
:- block front_sequence(-,?,?,?).
front_sequence(S1,Res,_Span,WF) :-
  is_custom_explicit_set(S1,front_sequence),
  front_sequence_custom_explicit_set(S1,FRes),!,
  equal_object_wf(FRes,Res,front_sequence,WF).
front_sequence(Seq,Res,Span,WF) :-  expand_custom_set_to_list(Seq,ESeq,_,front_sequence),
  front2(ESeq,Res,Span,WF).
front2([],_,Span,WF) :- add_wd_error_span('front applied to empty sequence!',[],Span,WF).
front2([H|T],Res,_Span,WF) :- size_of_sequence([H|T],int(Size),WF),
  (number(Size) -> true ; size_of_sequence(Res,SizeRes,WF), int_plus(SizeRes,int(1),int(Size))),
  when(ground(Size), domain_subtraction_wf([int(Size)],[H|T],Res,WF)).


:- assert_must_abort_wf(bsets_clp:last_sequence([],_R,WF),WF).
:- assert_must_abort_wf(bsets_clp:last_sequence([],int(1),WF),WF).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:last_sequence([(int(1),int(4)),(int(2),int(5))],int(5),_WF)) ).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:last_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],int(6),_WF)) ).
:- assert_must_succeed((bsets_clp:last_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(55))],R = int(55))).
:- assert_must_succeed((bsets_clp:last_sequence(X,R,_WF), X = [(int(1),int(55))], R = int(55))).
:- assert_must_succeed((bsets_clp:last_sequence([(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])],R,_WF), R == [(int(1),int(55))])).

last_sequence(Seq,Res,WF) :- last_sequence(Seq,Res,unknown,WF).
:- block last_sequence(-,?,?,?).
last_sequence(Seq,Res,_Span,WF) :- %print(last_sequence(Seq,Res,_)),nl,
  is_custom_explicit_set(Seq,last_sequence),
  last_sequence_explicit_set(Seq,Last), !, %print(last(Last)),nl,
  equal_object_wf(Last,Res,last_sequence,WF).
last_sequence([],_,Span,WF) :- !,add_wd_error_span('last applied to empty sequence!',[],Span,WF).
%last_sequence(Seq,Res,Span,WF) :- !,  /* I am not sure whether this brings any advantage */
%   expand_custom_set_to_list(Seq,ESeq),
%   last_seq(ESeq,0,abort(undefined_last),0,Res,Span,WF), print(last(ESeq,Res)),nl.
last_sequence(Seq,Res,Span,WF) :-
    size_of_sequence(Seq,int(Size),WF),
    last_sequence_aux(Size,Seq,Res,Span,WF).
:- block last_sequence_aux(-,?,?,?,?).
last_sequence_aux(Size,Seq,Res,Span,WF) :-
    apply_to(Seq,int(Size),Res,Span,WF).
/*
:- block last_seq(-, ?,?,?,?, ?,?).
last_seq([],LastIdx,LastVal,SizeSoFar,LastResult, Span,WF) :-
   (LastIdx=SizeSoFar -> LastResult = LastVal ;
     add_wd_error_span('last applied to invalid sequence (biggest index different from size): ',LastIdx/SizeSoFar,Span,WF),
     LastResult=LastVal).
last_seq([(ICurIdx,CurVal)|T],LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
  last_seq2(ICurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF).
:- block last_seq2(-,?,?, ?,?,?,?, ?,?).
last_seq2(int(CurIdx),CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
   last_seq3(CurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF).
:- block last_seq3(-,?,?, ?,?,?,?, ?,?).
last_seq3(CurIdx,CurVal,T,LastIdx,LastVal,SizeSoFar,LastResult,Span,WF) :-
   S1 is SizeSoFar+1,
   (CurIdx>LastIdx -> last_seq(T,CurIdx,CurVal,S1,LastResult,Span,WF)
     ; last_seq(T,LastIdx,LastVal,S1,LastResult,Span,WF)).
*/


:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4)),(int(2),int(5))],[(int(1),int(5)),(int(2),int(4))],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4))],[(int(1),int(4))],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([],[],WF),WF )).
:- assert_must_succeed(exhaustive_kernel_check_wfdet( bsets_clp:rev_sequence([(int(1),int(4)),(int(3),int(6)),(int(2),int(5))],[(int(1),int(6)),(int(3),int(4)),(int(2),int(5))],WF),WF )).
:- assert_must_succeed((bsets_clp:rev_sequence([],[],_WF))).
:- assert_must_succeed((bsets_clp:rev_sequence(X,R,_WF),
                        X = [(int(1),int(2)),(int(2),int(1))],
                        kernel_objects:equal_object(R,[(int(2),int(2)),(int(1),int(1))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(X,R,_WF),
                        X = [(int(1),int(23)),(int(2),int(1)),(int(3),int(55))],
              kernel_objects:equal_object(R,[(int(3),int(23)),(int(2),int(1)),(int(1),int(55))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(R,X,_WF),
                        X = [(int(1),int(23)),(int(2),int(1)),(int(3),int(55))],
              kernel_objects:equal_object(R,[(int(3),int(23)),(int(2),int(1)),(int(1),int(55))]) )).
:- assert_must_succeed((bsets_clp:rev_sequence(X,_R,_WF),
                        X = [(int(2),int(1)),(int(1),int(23)),(int(3),int(55))] )).
:- assert_must_succeed((bsets_clp:rev_sequence(_R,X,_WF),
                        X = [(int(3),int(55)),(int(1),int(23)),(int(2),int(1))] )).

/* reverse of sequence */
:- block rev_sequence(-,-,?).
rev_sequence(S1,Res,WF) :-
  (nonvar(S1) -> rev_sequence2(S1,Res,WF)
               ; rev_sequence2(Res,S1,WF)).

rev_sequence2(S1,Res,WF) :- reverse_custom_explicit_set(S1,RS1),!,
  equal_object_wf(Res,RS1,WF).
rev_sequence2(S1,Res,WF) :-
  expand_custom_set_to_list(S1,ES1,_,rev_sequence2),
 % print(rev(ES1)),nl,
  size_of_sequence(ES1,int(Size1),WF),
  rev_sequence3(ES1,Size1,Res,WF).

:- block rev_sequence3(?,-,?,?).
rev_sequence3(E,S,R,WF) :- rev_sequence4(E,S,R,WF).

:- block rev_sequence4(-,?,?,?).
rev_sequence4([],_,Res,WF) :- empty_set_wf(Res,WF).
rev_sequence4([(int(N),El)|Tail],Size1,Res,WF) :-
   equal_cons_wf(Res,(NewN,El),RTail,WF),
   % compute NewN = Size - (N-1)
   int_minus(int(N),int(1),N1),
   int_minus(int(Size1),N1,NewN),
   %print(rev(N,Size1,NewN)),nl,
   (ground(NewN) -> true ; in_nat_range(NewN,int(0),int(Size1))),
   rev_sequence4(Tail,Size1,RTail,WF).


/* --------- */
/* RELATIONS */
/* --------- */

%maplet(X,Y,(X,Y)).

% relation([]).
% relation([(_X,_Y)|T]) :- relation(T).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:relation_over_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:relation_over([],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(int(1),int(2))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([([int(1)],[int(2)])],[[int(1)],[],[int(2)]],[[int(2)]]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(pred_true /* bool_true */,pred_false /* bool_false */)],[pred_false /* bool_false */,pred_true /* bool_true */],[pred_false /* bool_false */]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((pred_true /* bool_true */,int(2)),fd(1,'Name'))],[(pred_false /* bool_false */,int(1)),(pred_true /* bool_true */,int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((rec([field(a,fd(1,'Name'))]),int(2)),fd(1,'Name'))],[(rec([field(a,fd(1,'Name'))]),int(1)),(rec([field(a,fd(1,'Name'))]),int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((rec([field(a,fd(2,'Name')),field(b,fd(1,'Name'))]),int(2)),fd(1,'Name'))],[(rec([field(a,fd(1,'Name')),field(b,fd(1,'Name'))]),int(1)),(rec([field(a,fd(1,'Name')),field(b,fd(2,'Name'))]),int(2)),(rec([field(a,fd(2,'Name')),field(b,fd(1,'Name'))]),int(2))],[fd(2,'Name'),fd(1,'Name')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([((pred_true /* bool_true */,int(2)),string('STRING1'))],[(pred_false /* bool_false */,int(1)),(pred_true /* bool_true */,int(2))],[string('STRING2'),string('STRING1')]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( /* multiple solutions !!*/ bsets_clp:relation_over([(int(1),int(2)),(int(2),int(2))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_succeed_check( bsets_clp:relation_over([(int(1),int(2)),(int(1),int(3))],[int(1),int(2)],[int(3),int(2)]) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:relation_over([(int(1),int(2)),(int(2),int(1))],[int(1),int(2)],[int(2)]) )).
:- assert_must_fail(( bsets_clp:relation_over([(int(1),int(1))],[int(1),int(2)],[int(2)]) )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(3)]),
                         X==[(int(1),int(3))] )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(3)]),
                         X==[(int(1),int(3)),(int(2),int(3))] )).
:- assert_must_succeed(( bsets_clp:relation_over(X,[int(1),int(2)],[int(4),int(5)]),
                         X==[(int(2),int(4)),(int(2),int(5))] )).

relation_over(R,Dom,Ran) :- init_wait_flags(WF),
  relation_over_wf(R,Dom,Ran,WF),
  ground_wait_flags(WF).

:- block relation_over_wf(-,-,-,?).
relation_over_wf(R,Dom,Ran,WF) :-
  kernel_equality:get_cardinality_relation_over_wait_flag(Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels),
  relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels).

:- block relation_over1(-,?,?,?,-,?,?).
relation_over1(FF,Domain,Range,WF,_WFR,_MaxCard,_MaxNrOfRels) :-
    nonvar(FF),
    custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,_),!,
    check_subset_of_wf(FFDomain,Domain,WF).
relation_over1(FF,Domain,Range,WF,_WFR,_MaxCard,_MaxNrOfRels) :- nonvar(FF),
    % print_term_summary(relation_over1(FF,Domain,Range,WF)), %
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,_),!,
    check_subset_of_wf(FFDomain,Domain,WF),
    check_subset_of_wf(FFRange,Range,WF).
relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels) :- var(R),!,
    expand_custom_set_to_list(R,ER,_,relation_over1),
    relation_over2(ER,[],Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels,none).
relation_over1(R,Domain,Range,WF,_WFR,_MaxCard,_) :-
   expand_and_convert_to_avl_set_warn(R,AER,relation_over1,'ARG : ? <-> ?'),!,
   is_avl_relation_over_domain(AER,Domain,WF), %print(dom_ok),nl,
   is_avl_relation_over_range(AER,Range,WF).
relation_over1(R,Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels) :-
    expand_custom_set_to_list(R,ER,_,relation_over1),
    relation_over2(ER,[],Dom,Ran,WF,WFR,MaxCard,MaxNrOfRels,none).

expand_and_convert_to_avl_set_warn(R,AS,Origin,Operator) :-
    observe_enumeration_warnings(expand_and_convert_to_avl_set(R,AS),
       add_message(Origin,'Enumeration warning occured while expanding argument ARG for predicate ',Operator)).
%expand_and_convert_to_avl_set(R,AS,_,Operator,Values) :-
%    observe_enumeration_warnings(expand_and_convert_to_avl_set(R,AS),
%       display_warning_message(Operator,Values)).
%display_warning_message(Operator,Values) :-
%       format(user_error,'Enumeration Warning for Operator ~w~n',[Operator]),
%       maplist(translate:print_bvalue,Values),nl.

:- block relation_over2(-,?,?,?,?,-,?,?,?).
relation_over2([],_,_,_,_WF,_WFR,_MaxCard,_MaxNrOfRels,_LastPair).
relation_over2(REL,SoFar,Domain,Range,WF,WFR,MaxCard,MaxNrOfRels,LastPair) :-
   (var(REL) -> NewLastPair=(X,Y) ; NewLastPair=none), %remember whether we freely chose X,Y
   REL = [(X,Y)|T],
    % print(rel_over2(X,Y,T,sofar(SoFar),WFR,MaxCard,LastPair)),nl,
      (number(MaxCard)
       -> MaxCard>0,C1 is MaxCard-1  ,(C1=0 -> T=[] ; true)
        ; C1=MaxCard),
      % TO DO: try to enumerate elements in order
      ordered_pair(LastPair,X,Y,not_equal),
      check_element_of_wf(X,Domain,WF),
      check_element_of_wf(Y,Range,WF),
      not_element_of_wf((X,Y),SoFar,WF),
      %print_bt_message(done_enum_relation_over2(X,Y,T,sofar(SoFar),MaxCard)),nl,
      update_waitflag(MaxNrOfRels,WFR,NewWFR,WF),
      relation_over2(T,[(X,Y)|SoFar],Domain,Range,WF,NewWFR,C1,MaxNrOfRels,NewLastPair).

% check that new pair is greater than previous pair, if that pair was freely chosen
ordered_pair(none,_,_,_).
ordered_pair((LastX,LastY),NewX,NewY,Eq) :- ordered_value(LastX,NewX,EqualX),
  check_second_component(EqualX,LastY,NewY,Eq).

:- block check_second_component(-,?,?,?).
check_second_component(equal,X,Y,EqRes) :- ordered_value(X,Y,EqRes).
check_second_component(not_equal,_X,_Y,not_equal). % no need to check 2nd component

:- block ordered_value(-,?,?), ordered_value(?,-,?).
ordered_value(pred_true /* bool_true */,B,Eq) :- !, (B=pred_true /* bool_true */ -> Eq=equal ; Eq=not_equal).
ordered_value(pred_false /* bool_false */,B,Eq) :- !, B=pred_false /* bool_false */, Eq=equal.
ordered_value(int(X),int(Y),Eq) :- !,
   kernel_objects:less_than_equal_direct(X,Y), equal_atomic_term(X,Y,Eq).
ordered_value(fd(NrX,T),fd(NrY,T),Eq) :- !,
   kernel_objects:less_than_equal_direct(NrX,NrY), equal_atomic_term(NrX,NrY,Eq).
ordered_value((X1,X2),(Y1,Y2),Eq) :- !, ordered_pair((X1,X2),Y1,Y2,Eq).
ordered_value(string(X),string(Y),Eq) :- !, less_equal_atomic_term(X,Y,Eq).
ordered_value(rec(FX),rec(FY),Eq) :- !,
   ordered_fields(FX,FY,Eq).
ordered_value([],Y,Eq) :- !, (Y==[] -> Eq=equal ; Eq=not_equal).
ordered_value(avl_set(A),Y,Eq) :- !,
  (Y==[] -> fail
    ; Y=avl_set(B) -> (A @< B -> Eq=not_equal ; A@>B -> fail ; Eq=equal)
    ; print(assuming_strictly_ordered(avl_set(A),Y)),nl,
      Eq=not_equal). % TO DO: treat sets better
ordered_value([H|T],Y,Eq) :- !,
  (Y==[] -> fail ; (Y==[H|T] -> Eq=equal
                    ; print(assuming_strictly_ordered([H|T],Y)),nl,
                      Eq=not_equal)).
ordered_value(A,B,not_equal) :- print(assuming_strictly_ordered(A,B)),nl.

:- block ordered_fields(-,?,?).
ordered_fields([],RHS,Eq) :- !,RHS=[], Eq=equal.
ordered_fields([field(Name,ValX)|TX],RHS,Eq) :- !,RHS=[field(Name,ValY)|TY],
  ordered_value(ValX,ValY,Equal1), check_next_field(Equal1,TX,TY,Eq).
ordered_fields(FX,FY,Eq) :- add_internal_error('Unknown fields: ',ordered_fields(FX,FY,Eq)), Eq=not_equal.

:- block check_next_field(-,?,?,?).
check_next_field(equal,TX,TY,EqRes) :- ordered_fields(TX,TY,EqRes).
check_next_field(not_equal,_X,_Y,not_equal). % no need to check next field

:- block less_equal_atomic_term(-,?,?), less_equal_atomic_term(?,-,?).
less_equal_atomic_term(A,B,Res) :- (A==B -> Res=equal ; A @<B, Res=not_equal).

:- block equal_atomic_term(-,?,?), equal_atomic_term(?,-,?).
equal_atomic_term(A,B,Res) :- (A==B -> Res=equal ; Res=not_equal).


:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:not_relation_over([(int(1),int(2)),(int(2),int(1))],[int(1),int(2)],[int(2)],_WF) )).
:- assert_must_succeed(exhaustive_kernel_check( bsets_clp:not_relation_over([(int(1),int(2))],[],[int(2)],_WF) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:not_relation_over([(int(1),pred_true)],[int(1)],[pred_true],_WF) )).
:- assert_must_succeed(exhaustive_kernel_fail_check( bsets_clp:not_relation_over([],[int(1)],[pred_true],_WF) )).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(2))],[int(3)],[int(1),int(2)],_) ).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(2))],[int(1)],[int(3)],_) ).
:- assert_must_succeed( bsets_clp:not_relation_over([(int(1),int(3)),(int(1),int(2))],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([(int(1),int(3))],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([],[int(1)],[int(3)],_) ).
:- assert_must_fail( bsets_clp:not_relation_over([],[],[],_) ).
:- block not_relation_over(-,?,?,?).

not_relation_over(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,_),!,
    % print_term_summary(not_relation_over(FF,FFDomain,Range)),
    not_subset_of_wf(FFDomain,Domain,WF).
not_relation_over(FF,Domain,Range,WF) :- nonvar(FF), % print_term_summary(not_relation_over(FF,Domain,Range,WF)), %
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,_),!,
    %print_term_summary(check_special_not_pf(FF,FFDomain,FFRange)),nl,
    not_both_subset_of(FFDomain,FFRange,Domain,Range,WF).
/* could be slightly more efficient: but not clear if warrants additional complexity in code:
not_relation_over(FF,Domain,Range,WF) :- nonvar(FF),
   check_element_can_be_decided(Domain), % ensures that check_element_of_wf will not block below
   check_element_can_be_decided(Range), % ensures that check_element_of_wf will not block below
   expand_and_convert_to_avl_set(FF,AER),!,
   (is_avl_relation_over_domain(AER,Domain,WF)
     -> \+ is_avl_relation_over_range(AER,Range,WF)
     ;  true).
check_element_can_be_decided(X) :- var(X),!,fail.
check_element_can_be_decided(avl_set(_)).
check_element_can_be_decided([]).
check_element_can_be_decided(closure(P,T,B)) :-
   custom_explicit_sets:is_interval_closure_or_integerset(closure(P,T,B),Low,Up),
   ground(Low), ground(Up).
*/
not_relation_over(R,Dom,Ran,WF) :-
   expand_custom_set_to_list(R,ER,_,not_relation_over),
 %%  print(not_rel(ER,Dom,Ran)),nl,
   not_relation_over2(ER,Dom,Ran,WF).


%not_relation_over2(R,_,_) :- when(nonvar(R), (R\=[], R\=[_|_])) . % TYPE ERROR !
:- block not_relation_over2(-,?,?,?).
not_relation_over2([(X,Y)|T],Domain,Range,WF) :- %print(not_rel2(X,Y,T)),nl,
      membership_test_wf(Domain,X,MemRes,WF),
      not_relation_over3(MemRes,Y,T,Domain,Range,WF).

:- block not_relation_over3(-,?,?,?,?,?).
not_relation_over3(pred_false,_Y,_T,_Domain,_Range,_WF).
not_relation_over3(pred_true,Y,T,Domain,Range,WF) :-
      membership_test_wf(Range,Y,MemRes,WF),
      not_relation_over4(MemRes,T,Domain,Range,WF).

:- block not_relation_over4(-,?,?,?,?).
not_relation_over4(pred_false,_T,_Domain,_Range,_WF).
not_relation_over4(pred_true,T,Domain,Range,WF) :-
      not_relation_over2(T,Domain,Range,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf([],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf([(int(1),int(3))],[int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf(
    [(int(0),int(55)),(int(2),int(3)),(int(1),int(3))],[int(1),int(2),int(0)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:domain_wf(
    [(int(99),int(55)),(int(2),int(3)),(int(99),int(4))],[int(2),int(99)],WF),WF)).
:- assert_must_succeed((bsets_clp:domain_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:domain_wf([(int(1),int(2))],Res,_WF),
        kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:domain_wf([(int(1),int(2)),(int(1),int(1))],Res,_WF),
            kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:domain_wf([(int(2),int(2)),(int(1),int(2))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(3),int(2)]),
        kernel_objects:equal_object(X,[(int(2),int(2)),(int(1),int(1)),(int(3),int(2))]))).
:- assert_must_succeed((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]),
        kernel_objects:equal_object(X,[(int(2),int(2)),(int(1),int(1)),(int(1),int(2))]))).
:- assert_must_fail((bsets_clp:domain_wf(X,Res,_WF),kernel_objects:equal_object(Res,[int(1),int(2)]),
        kernel_objects:equal_object(X,[(int(2),int(2)),(int(1),int(1)),(int(3),int(2))]))).

:- block domain_wf(-,-,?).
domain_wf(Rel,Res,WF) :- Res == [],!,
   % print(empty_domain_res(Rel,Res)),nl,
   empty_set_wf(Rel,WF).
domain_wf(Rel,Res,WF) :- var(Rel),!, % hence Res must me nonvar
   (is_custom_explicit_set(Res,domain_wf)
     -> expand_custom_set_to_list(Res,Res2,_,propagate_result_to_input2) % avoid expanding twice
      ;  Res2 = Res),
   propagate_result_to_input(Res2,Rel,domain,WF),
   domain_wf1(Rel,Res2,WF).
domain_wf(Rel,Res,WF) :- domain_wf1(Rel,Res,WF).


propagate_result_to_input(Result,Rel,DomOrRange,WF) :-
   propagate_empty_set(Result,Rel), % this will trigger before LWF ground
   (preferences:preference(use_smt_mode,true)
    -> propagate_result_to_input1(Result,Rel,1,DomOrRange) % hopefully full CHR implementation will avoid the need for this hack
  %  ; kernel_objects:is_marked_to_be_computed(Rel) -> true % get_last_wait_flag(propagate_result_to_input,WF,LWF)
    ; %print(propagating_to_input(Rel)),nl,
      get_wait_flag(2000,propagate_result_to_input,WF,LWF), % TO DO: determine right value for Priority ?
      propagate_result_to_input1(Result,Rel,LWF,DomOrRange) % this slows down test 289 if not guarded, 1088 if guarded
    ).

:- block propagate_result_to_input1(-,?,?,?), propagate_result_to_input1(?,-,-,?).
                                           % Note: if arg 2 (Rel) is known we will not propagate
propagate_result_to_input1([],Rel,_,_) :- !, empty_set(Rel).
propagate_result_to_input1(Result,Input,LWF,DomOrRange) :-    
   (kernel_objects:is_marked_to_be_computed(Input) -> true %print(not_prop(Input,DomOrRange)),nl,nl
    ; %print(prop(Result,Input,LWF)),nl,
      propagate_result_to_input2(Result,Input,LWF,DomOrRange)).

%:- block propagate_result_to_input2(-,?).
:- block propagate_result_to_input2(-,?,?,?), propagate_result_to_input2(?,-,-,?).
% maybe do in CHR in future:  x:dom(R) => #z.(x,z) : R
% TO DO: make stronger; also support avl_set ...
propagate_result_to_input2([],_Rel,_,_) :- !. % nothing can be said; we could have repeated entries for previous domain elements
propagate_result_to_input2([D|T],Rel,LWF,DomOrRange) :- %print(propagate_result_to_input2([D|T],Rel,LWF,DomOrRange)),nl,
   !,
   (Rel == [] -> fail % we would need more relation elements to generate the domain/range
    ; nonvar(Rel) -> true % no propagation
    ; (DomOrRange=domain -> Rel = [(D,_)|RT] ; Rel = [(_,D)|RT]),
      propagate_result_to_input2(T,RT,LWF,DomOrRange)
    ).
propagate_result_to_input2(CS,Rel,LWF,DomOrRange) :- var(Rel), is_custom_explicit_set(CS),!,
   expand_custom_set_to_list(CS,Res,_,propagate_result_to_input2),
   propagate_result_to_input2(Res,Rel,LWF,DomOrRange).
propagate_result_to_input2(_1,_2,_LWF,_DomOrRange). %  :- print(unknown_prop(_1,_2,_LWF)),nl.

:- block domain_wf1(-,?,?).
domain_wf1(Rel,Res,WF) :- is_custom_explicit_set(Rel,domain_wf),
 % print_term_summary(try_dom_explicit(Rel,Res)),
  domain_of_explicit_set(Rel,Dom), !,
  % print_term_summary(dom_res(Rel,Dom)),%%
  equal_object_wf(Dom,Res,domain_wf1,WF).
domain_wf1(Rel,Res,WF) :-   %% print_term_summary(domain_normal(Rel,Res)), %%
   expand_custom_set_to_list(Rel,Relation,_,domain_wf),
   newdomain1(Relation,[],Res,WF),
   quick_propagate_domain(Relation,Res,WF).


%:- block newdomain1(-,?,-,?). % why was this commented out ?
:- block newdomain1(-,?,?,?).
/* newdomain1(Rel,SoFar,Res,WF) :- var(Rel), !, % print(prop(Res)),nl,
    domain_propagate_result(Res,Rel,SoFar,WF). */
newdomain1(Dom,SoFar,Res,WF) :- newdomain2(Dom,SoFar,Res,WF).

/* no longer reachable because of new block declaration for newdomain1 above
% propagate information from domain to relation
domain_propagate_result([],Rel,SoFar,WF) :- !, check_domain_subset(Rel,SoFar,WF).
domain_propagate_result([H|TR],Rel,SoFar,WF) :-  !,
   % propagate known domain element into Relation Rel
    Rel = [(H,_)|TD], % equal_cons_wf(Rel,(H,_),TD,WF) ?
    not_element_of_wf(H,SoFar,WF), add_element_wf(H,SoFar,SoFar2,WF),
    newdomain1(TD,SoFar2,TR,WF).
domain_propagate_result(Res,Rel,SoFar,WF) :- % not a list
    expand_custom_set_to_list(Res,ResultList,_,domain_propagate_result),
    newdomain1(Rel,SoFar,ResultList,WF).
:- block check_domain_subset(-,?,?).
check_domain_subset([],_,_).
check_domain_subset([(X,_Y)|T],SoFar,WF) :- % print(check_dom(X,_Y,SoFar)),nl,
  check_element_of_wf(X,SoFar,WF),
  check_domain_subset(T,SoFar,WF).
  */
:- block quick_propagate_domain(-,?,?).
quick_propagate_domain([],_,_WF).
quick_propagate_domain([(X,_)|T],FullRes,WF) :-
   quick_propagation_element_information(FullRes,X,WF,FullRes1), % should we use a stronger check ?
   quick_propagate_domain(T,FullRes1,WF).

%:- block newdomain2(-,?,?,?).
newdomain2([],_SoFar,Res,WF) :- empty_set_wf(Res,WF).
newdomain2([(X,Y)|T],SoFar,Res,WF) :-
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(X,SoFar,WF)
    ;  membership_test_wf(SoFar,X,MemRes,WF),
       card_greater_equal_check([(X,Y)|T],Res,MemRes) % check that card of Relation is greater or equal to Result; if equal set MemRes to pred_false
   ),
   newdomain3(MemRes,X,T,SoFar,Res,WF).

:- block newdomain3(-,?,?,?,?,?).
newdomain3(pred_true,_,T,SoFar,Res,WF) :- newdomain1(T,SoFar,Res,WF).
newdomain3(pred_false,X,T,SoFar,Res,WF) :-
    kernel_objects:mark_as_non_free(X), % X is linked to a particular Y -> it is not free
    add_element_wf(X,SoFar,SoFar2,WF),
    equal_cons_wf(Res,X,Res2,WF), %print(newdomain3_added(X,SoFar,SoFar2,Res,Res2)),nl,
    newdomain1(T,SoFar2,Res2,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(2),[(int(2),int(7))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(2),[(int(1),int(6)),(int(2),int(7))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_domain_wf(int(22),[(int(1),int(6)),(int(22),int(7)),(int(33),int(7))],WF),WF)).
:- assert_must_succeed((bsets_clp:in_domain_wf(int(1),[(int(1),int(2))],_))).
:- assert_must_succeed((bsets_clp:in_domain_wf(int(3),[(int(1),int(2)),(int(3),int(4))],_))).
:- assert_must_fail((bsets_clp:in_domain_wf(int(3),[],_))).
:- assert_must_fail((bsets_clp:in_domain_wf(int(3),[(int(1),int(2))],_))).
 /* a more efficient version than using element_of and computing domain */

% just like not_empty_set_wf but instantiates with (El,_) as first element
in_domain_wf(El,S,WF) :- var(S),!,
  (preferences:preference(use_smt_mode,true) -> S=[(El,_)|_]
   ; get_enumeration_starting_wait_flag(not_empty_domain_wf,WF,LWF),  in_domain_lwf(El,S,LWF,WF)).
in_domain_wf(El,Rel,WF) :- in_domain_wf_lazy(El,Rel,WF).
:- block in_domain_lwf(-,?,-,?).
in_domain_lwf(El,Rel,_,WF) :-
    (var(Rel) -> Rel = [(El,_)|_]
               ; not_empty_set_unless_closure_wf(Rel,WF),
                 in_domain_wf_lazy(El,Rel,WF)).

not_empty_set_unless_closure_wf(closure(_,_,_),_) :- !. % do not check this; in_domain_wf or other call will find a solution anyway; no need to set up closure constraints twice
not_empty_set_unless_closure_wf(Rel,WF) :- not_empty_set_wf(Rel,WF).

% does not instantiate to [(El,_)|_]
:- block in_domain_wf_lazy(?,-,?).
%in_domain_wf_lazy(DomainElement,ES,WF) :- print(in_domain_wf_lazy(DomainElement,ES,WF)),nl,fail.
in_domain_wf_lazy(_DomainElement,[],_WF) :- !,fail.
in_domain_wf_lazy(DomainElement,avl_set(A),_WF) :-
     kernel_tools:ground_value(DomainElement), !, %print(enter_in_dom_avl(DomainElement)),nl,
     check_in_domain_of_avlset(DomainElement,A).
 % TO DO: check for infinite closures
in_domain_wf_lazy(DomainElement,ES,WF) :-
   is_custom_explicit_set(ES,in_domain_wf_lazy), %print(enter_in_dom(ES)),nl,
    domain_of_explicit_set(ES,Dom),!,
   % print(in_domain_of_explicit_set(DomainElement,Dom,ES)),nl,
    check_element_of_wf(DomainElement,Dom,WF).
in_domain_wf_lazy(El,Rel,WF) :- %print(in_dom(El)),nl,
   expand_custom_set_to_list(Rel,Relation,Done,in_domain_wf_lazy),
   get_binary_choice_wait_flag(in_domain_wf_lazy(El),WF,LWF),
   % print(in_domain2(El,Relation,WF,LWF,Done)),nl,
   % if Done == true -> we can use maybe clpfd_inlist or clpfd:element or quick_propagate
   quick_propagation_domain_element_list(Done,Relation,El,WF),
   in_domain2(El,Relation,WF,LWF).

% a custom implementation of quick_propagation_element_information for checking domain elements and lists only
:- use_module(clpfd_lists,[try_in_fd_value_list_check/4]).
:- block quick_propagation_domain_element_list(-,?,?,?).
quick_propagation_domain_element_list(_,_,El,_) :- ground(El),!.
quick_propagation_domain_element_list(_,RelList,El,WF) :-
    try_in_fd_value_list_check(RelList,(El,_),couple_left(_),WF). % use couple_left to ignore range values


:- block in_domain2(?,-,?,?).
in_domain2(El,[(X,_Y)|T],WF,LWF) :-
   kernel_objects:equality_objects_lwf(El,X,EqRes,LWF),
   in_domain3(EqRes,El,T,WF,LWF).

:- block in_domain3(-,?,?,?,?).
in_domain3(pred_true,_El,_T,_WF,_LWF).
in_domain3(pred_false,El,T,WF,LWF) :-
   get_new_subsidiary_wait_flag(LWF,in_domain2(El,T),WF,NewLWF),
   in_domain2(El,T,WF,NewLWF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_domain_wf(int(3),[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_domain_wf(int(3),[(int(2),int(7))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_domain_wf(int(3),[(int(2),int(7)),(int(4),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_in_domain_wf(int(4),[(int(2),int(7)),(int(4),int(3))],WF),WF)).
:- assert_must_fail((bsets_clp:not_in_domain_wf(int(1),[(int(1),int(2))],_))).
:- assert_must_fail((bsets_clp:not_in_domain_wf(int(3),[(int(1),int(2)),(int(3),int(4))],_))).
:- assert_must_succeed((bsets_clp:not_in_domain_wf(int(3),[],_))).
:- assert_must_succeed((bsets_clp:not_in_domain_wf(int(3),[(int(1),int(2))],_))).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_domain_wf(int(3),[(int(1),int(2)),(int(2),int(3))],WF),WF)).
 /* a more efficient version than using not_element_of and computing domain */


:- block not_in_domain_wf(?,-,?).
not_in_domain_wf(DomainElement,ES,WF) :- is_custom_explicit_set(ES,not_in_domain),
    domain_of_explicit_set(ES,Dom),!,
    %print_term_summary(not_in_domain2(DomainElement,Dom,ES)),
    not_element_of_wf(DomainElement,Dom,WF).
not_in_domain_wf(El,Rel,WF) :- %%print(not_in_dom(El,Rel)),nl,
   expand_custom_set_to_list(Rel,Relation,_,not_in_domain),
   not_in_domain2(Relation,El,WF).
:- block not_in_domain2(-,?,?).
not_in_domain2([],_,_WF).
not_in_domain2([(X,_Y)|T],E,WF) :- not_equal_object_wf(E,X,WF), not_in_domain2(T,E,WF).




:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf([],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf([(int(1),int(3))],[int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:range_wf(
    [(int(0),int(55)),(int(2),int(3)),(int(1),int(3))],[int(3),int(55)],WF),WF)).
:- assert_must_succeed((bsets_clp:range_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2))],Res,_WF),
           kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(1)),(int(2),int(1))],Res,_WF),
           kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2)),(int(1),int(1))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf([(int(1),int(2)),(int(1),int(1)),(int(2),int(3))],Res,_WF),
                        kernel_objects:equal_object(Res,[int(1),int(3),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,_WF),
                        X = [(int(1),int(2)),(int(1),int(1)),(int(2),int(3))],
                        kernel_objects:equal_object(Res,[int(1),int(3),int(2)]))).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [(int(1),int(2)),(int(1),int(1)),(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [(int(2),int(1)),(int(1),int(2)),(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:range_wf(X,Res,WF), bsets_clp:domain_wf(X,Res2,WF), kernel_objects:equal_object(Res,Res2),
                        X = [])).
:- assert_must_succeed((bsets_clp:range_wf([([],[]),([int(0)],[int(0)]),
 ([int(0),int(1)],[int(0),int(1)]),([int(0),int(2)],[int(0),int(2)]),
 ([int(0),int(3)],[int(0),int(3)]),([int(0),int(4)],[int(0),int(4)]),([int(1)],[int(1)]),
 ([int(1),int(2)],[int(1),int(2)]),([int(1),int(3)],[int(1),int(3)]),
 ([int(1),int(4)],[int(1),int(4)]),([int(2)],[int(2)]),([int(2),int(3)],[int(2),int(3)]),
 ([int(2),int(4)],[int(2),int(4)]),([int(3)],[int(3)]),([int(3),int(4)],
                                   [int(3),int(4)]),([int(4)],[int(4)])],_Res,_WF))).
:- assert_must_succeed((bsets_clp:range_wf([([],[]),([int(0)],[int(0)]),
 ([int(0),int(1)],[int(0),int(1)]),
 ([int(0),int(3)],[int(0),int(3)]),([int(0),int(4)],[int(0),int(4)]),([int(1)],[int(1)]),
 ([int(1),int(2)],[int(1),int(2)])],_Res,_WF))).

% TO DO: also use propagate_result_to_input
:- block range_wf(-,-,?).
range_wf(Rel,Res,WF) :- Res ==[],!, empty_set_wf(Rel,WF).
range_wf(Rel,Res,WF) :- Rel ==[],!, empty_set_wf(Res,WF).
range_wf(Rel,Res,WF) :- range_wf1(Rel,Res,WF),
  propagate_result_to_input(Res,Rel,range,WF).

:- block range_wf1(-,?,?).
range_wf1(Rel,Res,WF) :-  % print_term_summary(range_wf1(Rel,Res)),
  is_custom_explicit_set(Rel,range_wf1),
  range_of_explicit_set(Rel,Range), !,
   % print_term_summary(range_of_explicit_set(Rel,Range)),
  equal_object_wf(Range,Res,range_wf1,WF).
range_wf1(Rel,Res,WF) :-  % print_term_summary(range_normal(Rel,Res)),
 % TO DO : propagate information that card of Res <= card of Rel; similar thing for domain
   expand_custom_set_to_list(Rel,Relation,_,range_wf1), %print(exp(Relation)),nl,
   % print(newrange2(Relation,[],Res)),nl,
   newrange2(Relation,[],Res,WF),
   quick_propagate_range(Relation,Res,WF).


:- block quick_propagate_range(-,?,?).
quick_propagate_range([],_,_WF).
quick_propagate_range([(_,Y)|T],FullRes,WF) :-
   quick_propagation_element_information(FullRes,Y,WF,FullRes1), % should we use a stronger check ?
   quick_propagate_range(T,FullRes1,WF).

:- block newrange2(-,?,?,?).
newrange2([],_SoFar,Res,WF) :-  %print(range2_base(Acc,Res)),nl,
   empty_set_wf(Res,WF). % ,print(eq),nl.
newrange2([(X,Y)|T],SoFar,Res,WF) :-  %print(range2(Y,T,SoFar,Res)),nl,
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(Y,SoFar,WF)
    ;  membership_test_wf(SoFar,Y,MemRes,WF),
       card_greater_equal_check([(X,Y)|T],Res,MemRes) % check that card of Relation is greater or equal to Result; if equal set MemRes to pred_false
   ),
   newrange3(MemRes,Y,T,SoFar,Res,WF).

% card_greater_equal_check(Set1,Set2,EqFlag) : check that cardinality of Set1 is greater or equal to that of Set2; set EqFlag to pred_false if they are equal
% checking is stopped if EqFlag becomes nonvar
% tested by testcase 1061
:- block card_greater_equal_check(-,?,-), card_greater_equal_check(?,-,-).
card_greater_equal_check(_,_,Flag) :- nonvar(Flag),!. % no longer required; even though we could prune failure !? done later in newrange2/newdomain2 ??!!
card_greater_equal_check([],Set2,Flag) :- !,empty_set(Set2),
   Flag=pred_false. % Flag set indicates that both sets have same size
card_greater_equal_check(_,[],_) :- !.
card_greater_equal_check([_|T],[_|R],Flag) :- !, card_greater_equal_check(T,R,Flag).
% To do: deal with AVL args as Result + also use efficient_card_for_set for closures
%card_greater_equal_check([_|T],Set,Flag) :- efficient_card_for_set(B,CardB,CodeB),!,
% f: 1..7 -->> 1..n & n>=7 & n<10 still does not work well
% TO DO: can we merge code with check_card_greater_equal
card_greater_equal_check(_,_,_).


:- block newrange3(-,?,?,?,?,?).
newrange3(pred_true,_Y,T,SoFar,Res,WF) :- newrange2(T,SoFar,Res,WF).
newrange3(pred_false,Y,T,SoFar,Res,WF) :-
    kernel_objects:mark_as_non_free(Y), % Y is linked to a particular X -> it is not free
    add_element_wf(Y,SoFar,SoFar2,WF),
    equal_cons_wf(Res,Y,Res2,WF),
    % print_bt_message(added(Y,T,SoFar,sofar2(SoFar2),Res,res2(Res2))),
    newrange2(T,SoFar2,Res2,WF).


:- assert_must_succeed((bsets_clp:identity_relation_over_wf([],Res,_WF),Res=[])).
:- assert_must_succeed((bsets_clp:identity_relation_over_wf([int(1),int(2)],Res,_WF),
    Res=[(int(1),int(1)),(int(2),int(2))])).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:identity_relation_over_wf([int(2),int(4)],[(int(4),int(4)),(int(2),int(2))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:identity_relation_over_wf([int(1),int(2),int(4)],[(int(4),int(4)),(int(2),int(2)),(int(1),int(1))],WF),WF)).
:- assert_must_fail((bsets_clp:identity_relation_over_wf([int(1)|_],_,_WF),fail)). /* check: no loop */

:- block identity_relation_over_wf(-,?,?).
identity_relation_over_wf(Set1,IDRel,WF) :-
    expand_custom_set_to_list(Set1,ESet1,_,identity_relation_over_wf),
    identity_relation_over2(ESet1,IDRel,WF).

:- block identity_relation_over2(-,?,?).
identity_relation_over2([],Res,WF) :- empty_set_wf(Res,WF).
identity_relation_over2([X|T1],Res,WF) :- equal_cons_wf(Res,(X,X),T2,WF), % equal_object([(X,X)|T2],Res),
    identity_relation_over2(T1,T2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:in_identity((int(1),int(1)),[int(1),int(2)],WF),WF)).
:- assert_must_fail((bsets_clp:in_identity((int(1),int(2)),[int(1),int(2)],_WF))).
:- assert_must_fail((bsets_clp:in_identity((int(3),int(3)),[int(1),int(2)],_WF))).
:- assert_must_fail((bsets_clp:in_identity((int(1),int(2)),[],_WF))).
in_identity((X,Y),Domain,WF) :- %print(in_identity((X,Y),Domain)),nl,
    equal_object_wf(X,Y,in_identity,WF), check_element_of_wf(X,Domain,WF).

:- assert_must_fail((bsets_clp:not_in_identity((int(1),int(1)),[int(1),int(2)],_WF))).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_identity((int(1),int(2)),[int(1),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_in_identity((int(3),int(3)),[int(1),int(2)],WF),WF)).
:- assert_must_succeed((bsets_clp:not_in_identity((int(1),int(2)),[],_WF))).
not_in_identity((X,Y),Domain,WF) :- %print(not_in_identity((X,Y),Domain)),nl,
    equality_objects_wf(X,Y,Eq,WF),
    not_in_id2(Eq,X,Domain,WF).

:- block not_in_id2(-,?,?,?).
not_in_id2(pred_true,X,Domain,WF) :- not_element_of_wf(X,Domain,WF).
not_in_id2(pred_false,_,_,_).


:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], [(int(6),int(5)),(int(2),int(1)),(int(4),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([(int(1),int(2))], [(int(2),int(1))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:invert_relation_wf([], [],WF),WF)).
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,X,_),X = [])).
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,X,_),X = [(int(2),int(2))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf(X,[(int(1),int(2)),(int(7),int(6))],_WF),
                     X = [(int(2),int(1)),(int(6),int(7))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf([(int(1),int(2)),(int(7),int(6))],X,_WF),
                     X = [(int(2),int(1)),(int(6),int(7))])).
:- assert_must_succeed((bsets_clp:invert_relation_wf([(int(1),int(2)),(int(7),int(6))],
                           [(int(6),int(7)),(int(2),int(1))],_WF))).
:- assert_must_succeed((bsets_clp:invert_relation_wf(closure([a,b],[string,boolean],b(truth,pred,[])),
					     closure([b,a],[boolean,string],b(truth,pred,[])),_WF))).

:- block invert_relation_wf(-,-,?).
invert_relation_wf(R,IR,WF) :- % print_message(invert_relation(R,IR)),
  %  (nonvar(R) -> invert_relation2(R,IR) ; invert_relation2(IR,R)).
     invert_relation2(R,IR,WF). % , print_term_summary(invert_relation(R,IR)).
/* Optimization for some types of closures: Instead of expanding the closures, we just
   swap the parameters. This does not work with closures wich have only one parameter
   wich is a pair */
invert_relation2(CS,R,WF) :- nonvar(CS),is_custom_explicit_set_nonvar(CS),!,
     invert_explicit_set(CS,ICS), equal_object_wf(R,ICS,invert_relation2_1,WF).
invert_relation2(R,CS,WF) :- nonvar(CS),is_custom_explicit_set_nonvar(CS),!,
     invert_explicit_set(CS,ICS), equal_object_wf(R,ICS,invert_relation2_2,WF).
%invert_relation2(closure([P1,P2],[T1,T2],Clo),closure([P2,P1],[T2,T1],Clo)) :- !.
invert_relation2(R,IR,WF) :- %try_expand_custom_set(R,ER),
  %  (nonvar(R) -> invert_relation3(R,IR)
  %              ; invert_relation3(IR,R),(ground(IR)-> true ; invert_relation3(R,IR))).
  % print(inverting(R,IR,WF)),nl,
    invert_relation3(R,IR,WF,1), invert_relation3(IR,R,WF,1).
  %  ,print(inverted(R,IR)),nl. % propagates both ways; but multiple solutions ?!

:- block invert_relation3(-,?,?,?).
invert_relation3(closure(P,T,B),Res,WF,_) :- invert_explicit_set(closure(P,T,B),ICS),
       equal_object_wf(Res,ICS,invert_relation3_1,WF).
invert_relation3(avl_set(S),Res,WF,_) :- invert_explicit_set(avl_set(S),ICS),
       equal_object_wf(Res,ICS,invert_relation3_2,WF).
invert_relation3([],Res,WF,_) :- empty_set_wf(Res,WF).
invert_relation3([(X,Y)|T],Res,WF,Depth) :- %prints(invert([(X,Y)|T],Res,Depth)),
       D1 is Depth+1, get_wait_flag(D1,invert_relation3,WF,LWF),
       equal_cons_lwf(Res,(Y,X),IT,LWF),
       invert_relation3(T,IT,WF,D1).




tuple_of(X,Y,R) :- check_element_of((X,Y),R).
%tuple_of_wf(X,Y,R,WF) :- check_element_of_wf((X,Y),R,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], [(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],
[(int(1),int(1)),(int(5),int(7)),(int(3),int(33))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([], [(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([(int(6),int(7)),(int(2),int(1)),(int(22),int(22)),(int(4),int(33))],[],[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_composition([],[],[]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],
                                             [(int(1),int(11))],X),X = [])).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],[],X),X = [])).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(6))],
                                                  [(int(2),int(11))],X),
                     kernel_objects:equal_object(X,[(int(1),int(11))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(2))],[(int(2),int(11))],X),
                    ground(X), bsets_clp:equal_object(X,[(int(1),int(11)),(int(7),int(11))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(7),int(5))],
                       [(int(2),int(11)),(int(2),int(4))],X),
                     kernel_objects:equal_object(X,[(int(1),int(11)),(int(1),int(4))]))).
:- assert_must_succeed((bsets_clp:rel_composition([(int(1),int(2)),(int(1),int(5))],
                       [(int(2),int(11)),(int(5),int(11))],X),
                    kernel_objects:equal_object(X,[(int(1),int(11))]))).

rel_composition(Rel1,Rel2,Comp) :-
      init_wait_flags(WF), % TO DO: remove and add WF arg
      rel_composition_wf(Rel1,Rel2,Comp,WF),
      ground_wait_flags(WF).

:- block rel_composition_wf(-,-,?,?).
rel_composition_wf(Rel1,Rel2,Comp,WF) :-  % print(rel_compose(Rel1,Rel2,Comp)),nl,
   (Rel1==[] ; Rel2==[]),
   !,
   empty_set_wf(Comp,WF).
rel_composition_wf(Rel1,Rel2,Comp,WF) :- rel_composition1(Rel1,Rel2,Comp,WF).

:- block rel_composition1(-,?,?,?),rel_composition1(?,-,?,?).
rel_composition1(Rel1,Rel2,Comp,WF) :-  % print(rel(Rel1,Rel2,Comp)),nl, trace,
   (Rel1==[] ; Rel2==[]),!, empty_set_wf(Comp,WF).
rel_composition1(Rel1,Rel2,Comp,WF) :- keep_symbolic(Rel1),
    (Rel2 = avl_set(_) -> SYMBOLIC=false ; SYMBOLIC=symbolic),
    symbolic_composition(Rel1,Rel2,SYMBOLIC,Rel3),
    !,
     % format('SYMBOLIC COMPOSITION (~w): ',[SYMBOLIC]), translate:print_bvalue(Rel3),nl,
    equal_object_wf(Comp,Rel3,rel_composition1_0,WF).
rel_composition1(Rel1,Rel2,Comp,WF) :- rel_composition_for_explicit_set(Rel1,Rel2,Res),!,
  % print_term_summary(rel_composition_for_explicit_set(Rel1,Rel2,Res)),
  equal_object_wf(Res,Comp,rel_composition1_1,WF).
rel_composition1(Rel1,Rel2,Comp,WF) :- Rel2=closure(_,_,_),   % nl,print(rel(Rel1,Rel2,Comp)),nl, trace,
  keep_symbolic(Rel2),
  (dom_for_specific_closure(Rel2,Domain,function(_)) % TO DO: also deal with relations
    -> !,
       expand_custom_set_to_list(Rel1,Relation1,_,rel_composition1),
       rel_compose_with_inf_fun(Relation1,Domain,Rel2,Comp,WF)
    ;  symbolic_composition(Rel1,Rel2,false,Rel3),
       !,
       % print('SYMBOLIC COMPOSITION2: '),translate:print_bvalue(Rel3),nl,
       expand_custom_set(Rel3,CRes), equal_object_optimized(CRes,Comp,rel_composition1_3) % do we need to expand ?
  ).
rel_composition1(Rel1,Rel2,Comp,WF) :-
  expand_custom_set_to_list(Rel1,Relation1,_,rel_composition1_2),
  expand_custom_set_to_list(Rel2,Relation2,_,rel_composition1_3),
  rel_compose2(Relation1,Relation2,Comp,WF).

:- use_module(btypechecker, [unify_types_strict/2]).
symbolic_composition(Rel1,Rel2,SYMBOLIC,Rel3) :-
       get_relation_types(Rel1,TX,TY),
       get_relation_types(Rel2,TY2,TZ),
       (unify_types_strict(TY,TY2) -> true
         ; add_internal_error('Could not unify range and domain types: ',unify_types_strict(TY,TY2)),fail),
       rel_comp_closure(Rel1,Rel2,TX,TY,TZ,SYMBOLIC,Rel3).
% generate a closure for {xx,zz | #(yy).(xx|->yy : Rel1 & yy|->zz : Rel2)}
rel_comp_closure(Rel1,Rel2,TX,TY,TZ,SYMBOLIC,closure(Args,Types,CBody)) :-
    Args = [xx,zz], Types = [TX,TZ],
    couple_member_pred(xx,TX,yy,TY,Rel1, Pred1),
    couple_member_pred(yy,TY,zz,TZ,Rel2, Pred2),
    conjunct_predicates([Pred1,Pred2],P12),
    b_interpreter_components:create_unsimplified_exists([b(identifier(yy),TY,[])],P12,Body),
    (SYMBOLIC==symbolic -> mark_bexpr_as_symbolic(Body,CBody) ; CBody=Body).

% generate predicate for X|->Y : Rel
couple_member_pred(X,TX,Y,TY,Rel, Pred) :-
   Pred = b(member(b(couple(b(identifier(X),TX,[]),
                            b(identifier(Y),TY,[])),couple(TX,TY),[]),
                   b(value(Rel),set(couple(TX,TY)),[])),pred,[]).

/* ------------------
% first version to compute the relational composition symbolically:
% we can now compute: i = %x.(x:INTEGER|x+1) & f = (i;i;i;i) & res = f(2000)
%  or even {x,y|x<y} = r & r2 = (r;r) &  20|->y : r2
% TO DO: treat closure([X],[couple(T1,T2)],P)  like closure([X1,X2],[T1,T2],P')
%        treat closure([X1,X2,X3],[T1,T2,T3],P) like closure([X12,X3],[couple(T1,T2),T3],P')
% THIS IS VERY SIMILAR TO symbolic_composition(C1,C2,symbolic,C3) if we have two closures; but seems a bit faster
% TO DO: maybe throw this code away nonetheless
symbolic_composition(Closure1,Closure2,closure(PR,TR,BRS)) :- % print(sym(Closure1,Closure2)),nl,trace,
% ERROR: what if identifier Y=V !!  <---------------- LEADS TO ERROR
    closure_mapping_to(Closure1, OtherParams,OtherTypes, Y,TY, B1),
    binary_closure(Closure2,  V,TV, W,TW, B2),
    TV=TY, % check that types match
    append(OtherParams,[W],PR),
    append(OtherTypes,[TW],TR),
    gensym:gensym('__FRESH__',FRESHID), % maybe we can avoid generating fresh ids if not necessary
    % if Rel2 is a function (which it currently is !) we can maybe just apply it if no restriction on domain ?
    bsyntaxtree:rename_bt(B1,[rename(Y,FRESHID)],RB1),
    bsyntaxtree:rename_bt(B2,[rename(V,FRESHID)],RB2),
    bsyntaxtree:conjunct_predicates([RB1,RB2],CB),
    FRESH = b(identifier(FRESHID),TY,[]),
    b_interpreter_components:create_unsimplified_exists([FRESH],CB,BR),
    bsyntaxtree:mark_bexpr_as_symbolic(BR,BRS) ,print('RESULT: '),translate:print_bvalue(closure(PR,TR,BRS)).

closure_mapping_to(closure([X],[TX],P),  [X1],[T1], X2,T2, NewP) :- !,
    % try and translate this into a binary closure; will arise e.g. in symbolic closures for closure1 operator
    binary_closure(closure([X],[TX],P),   X1,T1,  X2,T2, NewP).
closure_mapping_to(closure(Args,Types,Pred),  OtherParams,OtherTypes, RangeVar,RangeType, Pred) :-
    append(OtherTypes,[RangeType],Types), OtherTypes \= [],
    append(OtherParams,[RangeVar],Args).

binary_closure(closure([X,Y],[TX,TY],P),  X,TX, Y,TY, P) :- !.
binary_closure(closure([X],[couple(T1,T2)],P),   X1,T1,  X2,T2, NewP) :-
   gensym:gensym('__FRESH__',X1), gensym:gensym('__FRESH__',X2),
   BX1 = b(identifier(X1),T1,[]), BX2 = b(identifier(X2),T2,[]),
   X12_Pair = b(couple(BX1,BX2),couple(T1,T2),[]),
   bsyntaxtree:replace_id_by_expr(P,X,X12_Pair,NewP),
   print('BINARY: '),translate:print_bvalue(closure([X1,X2],[T1,T2],NewP)),nl.
-------------------- */

:- block rel_compose2(-,?,?,?).
rel_compose2([],_,Out,WF) :- empty_set_wf(Out,WF).
rel_compose2([(X,Y)|T],Rel2,Out,WF) :-
   rel_extract(Rel2,X,Y,OutXY,[],WF),
  % rel_extract(Rel2,X,Y,Out,OutRem),
   rel_compose2(T,Rel2,OutRem,WF),
   union_wf(OutRem,OutXY,Out,WF). % used to call union wihout wf; makes test 1394 fail

:- block rel_extract(-,?,?,?,?,?).
rel_extract([],_,_,Rem,Rem,_WF).  % should we use equal_object here ?????
rel_extract([(Y1,Z)|T],X,Y,Res,Rem,WF) :-
   rel_extract(T,X,Y,CT,Rem,WF),
   equality_objects_wf(Y1,Y,EqRes,WF),
   rel_extract2(EqRes,Z,X,CT,Res).

:- block rel_extract2(-,?,?,?,?).
rel_extract2(pred_true,  Z, X,CT,Res) :- add_element((X,Z),CT,Res).
rel_extract2(pred_false,_Z,_X,CT,Res) :- Res = CT.



:- block rel_compose_with_inf_fun(-,?,?,?,?).
rel_compose_with_inf_fun([],_Dom,_Rel2,Comp,WF) :- empty_set_wf(Comp,WF).
rel_compose_with_inf_fun([(X,Y)|T],Dom,Fun,CompRes,WF) :-
   membership_test_wf(Dom,Y,MemRes,WF),
   rel_compose_with_inf_fun_aux(MemRes,X,Y,T,Dom,Fun,CompRes,WF).

:- block rel_compose_with_inf_fun_aux(-,?,?,?, ?,?,?,?).
rel_compose_with_inf_fun_aux(pred_true,X,Y,T,Dom,Fun,CompRes,WF) :-
   % print(rel_compose_with_inf_fun_aux(X,Y)),nl,
   apply_to(Fun,Y,FY,WF), % TO DO: generalize to image so that we can apply it also to infinite relations ?
   add_element_wf((X,FY),CT,CompRes,WF),
   rel_compose_with_inf_fun(T,Dom,Fun,CT,WF).
rel_compose_with_inf_fun_aux(pred_false,_X,_Y,T,Dom,Fun,Comp,WF) :-
    % print(rel_compose_not_in_domain(_X,_Y)),nl,
    rel_compose_with_inf_fun(T,Dom,Fun,Comp,WF).

:- assert_must_abort_wf(bsets_clp:rel_iterate([],int(-1),_R,WF),WF).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([], int(2),[],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], int(1),[(int(1),int(2)),(int(3),int(4)),(int(5),int(6))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(3),int(4)),(int(5),int(6))], int(0),[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4)),(int(5),int(5)),(int(6),int(6))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:rel_iterate([(int(1),int(2)),(int(2),int(4)),(int(4),int(6))], int(2),[(int(1),int(4)),(int(2),int(6))],_WF))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(1),X,_WF), R=[],
                        bsets_clp:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(1),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
                        bsets_clp:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(2),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
         bsets_clp:equal_object(X,[(int(1),int(3))]))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(3),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3))],
         bsets_clp:equal_object(X,[]))).
:- assert_must_succeed((bsets_clp:rel_iterate(R,int(3),X,_WF),
                        R=[(int(1),int(2)),(int(2),int(3)),(int(1),int(1))],
         bsets_clp:equal_object(X,[(int(1),int(1)),(int(1),int(2)),(int(1),int(3))]))).

rel_iterate(Rel,int(Nr),Res,WF) :- rel_iterate2(Nr,Rel,Res,WF).

:- block rel_iterate2(-,?,?,?).
rel_iterate2(X,Rel,Res,WF) :-
    ( X=1 -> equal_object_wf(Res,Rel,rel_iterate2,WF)
    ; X>1 -> X1 is X-1,
        rel_iterate2(X1,Rel,R1,WF),
        rel_composition(Rel,R1,Res) % TO DO: call WF version
    ; X=0 -> rel_iterate0(Rel,Res)
    ; otherwise -> add_wd_error('negative index in iterate',X,WF)
    ).

:- block rel_iterate0(-,?).
rel_iterate0(Rel,Res) :-
    expand_custom_set_to_list(Rel,Relation,_,rel_iterate0),
   % nl,print(start_rel_iterate0(Relation)),nl,
   rel_iterate0(Relation,[],Res).

:- block rel_iterate0(-,?,?).
rel_iterate0([],D,Res) :-  %print(result_rel_iterate0(D,Res)),nl,nl,
    equal_object_optimized(D,Res,rel_iterate0).
rel_iterate0([(X,Y)|T],ResSoFar,Res) :-
 % TO DO: add (X,X) and (Y,Y) earlier to Res  to improve constraint propagation
   add_element((X,X),ResSoFar,Res2),
   add_element((Y,Y),Res2,Res3),
   rel_iterate0(T,Res3,Res).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([],[(int(1),int(11))],[],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
 [(int(1),int(11))],[(int(1),(int(2),int(11)))],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
 [(int(2),int(11))],[],_WF))).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
                                             [(int(2),int(11))],X,_WF),
                                             X = [])).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(7),int(6))],
                                             [(int(1),int(11))],X,_WF),
                                             kernel_objects:equal_object(X,[(int(1),(int(2),int(11)))]))).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(1),int(6))],
                                             [(int(1),int(11))],X,_WF),
          kernel_objects:equal_object(X,[(int(1),(int(2),int(11))),(int(1),(int(6),int(11)))]))).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(2),int(6))],
                                             [(int(1),int(11)),(int(1),int(12))],X,_WF),
          kernel_objects:equal_object(X,[(int(1),(int(2),int(11))),(int(1),(int(2),int(12)))]))).
:- assert_must_succeed((bsets_clp:direct_product_wf([(int(1),int(2)),(int(2),int(6))],
                                             [(int(1),int(11)),(int(1),int(12))],
                          [(int(1),(int(2),int(11))),(int(1),(int(2),int(12)))],_WF))).
:- assert_must_succeed((bsets_clp:direct_product_wf(avl_set(node((fd(1,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name')),true,0,empty,empty))),
avl_set(node((fd(1,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name')),true,0,empty,empty))),
avl_set(node((fd(1,'Name'),fd(2,'Name'),fd(2,'Name')),true,1,empty,node((fd(2,'Name'),fd(3,'Name'),fd(3,'Name')),true,0,empty,empty)))
,_WF))).

:- block direct_product_wf(-,?,?,?),direct_product_wf(?,-,?,?).
direct_product_wf(Rel1,Rel2,Prod,WF) :-
   try_expand_and_convert_to_avl_with_check(Rel1,E1,direct_product), % to do: try_expand_and_convert_to_avl_unless_very_large(Rel1,E1),
   try_expand_and_convert_to_avl_with_check(Rel2,E2,direct_product),
   direct_product_wf1(E1,E2,Prod,WF).

direct_product_wf1(Rel1,Rel2,Prod,WF) :-  % print_term_summary(direct_product(Rel1,Rel2,Prod)), %
  direct_product_explicit_set(Rel1,Rel2,Res),!,
 % print_term_summary(direct_product_result(Res)),
  equal_object_wf(Prod,Res,direct_product_wf1,WF).
direct_product_wf1(Rel1,Rel2,Prod,WF) :-  %print_term_summary(direct_product_list(Rel1,Rel2,Prod)),
  expand_custom_set_to_list(Rel1,Relation1,_,direct_product_wf1_1),
  expand_custom_set_to_list(Rel2,Relation2,_,direct_product_wf1_2),
  direct_product2(Relation1,Relation2,Prod,WF),
  direct_product_backwards(Relation1,Relation2,Prod,WF).

%direct_product2(R1,R2,Res) :- print_message(direct_product2(R1,R2,Res)),fail.
:- block direct_product2(-,?,?,?).
direct_product2([],_,Out,WF) :- equal_object_wf(Out,[],direct_product2,WF).
direct_product2([(X,Y)|T],Rel2,Out,WF) :-
   direct_product_tuple(Rel2,X,Y,Out,OutRem,WF),
   direct_product2(T,Rel2,OutRem,WF).

:- block direct_product_tuple(-,?,?,?,?,?).
direct_product_tuple([],_,_,Res,Rem,WF) :- equal_object_optimized_wf(Res,Rem,direct_product_tuple,WF).
direct_product_tuple([(X2,Z)|T],X,Y,Res,Rem,WF) :-
   direct_product_tuple(T,X,Y,CT,Rem,WF),
   equality_objects_wf(X2,X,EqRes,WF),
   direct_product_tuple3(EqRes,X,Y,Z,CT,Res,WF).

:- block direct_product_tuple3(-,?,?,?,?,?,?).
direct_product_tuple3(pred_true,X,Y,Z,CT,Res,WF) :-
    equal_cons_wf(Res,(X,(Y,Z)),CT,WF).  /* no need for add_element as output uniquely determines X,Y,Z !?*/
direct_product_tuple3(pred_false,_X,_Y,_Z,CT,Res,WF) :-  equal_object_optimized_wf(Res,CT,direct_product_tuple3,WF).

:- block direct_product_backwards(?,?,-,?).
% Propagate information backwards from result to arguments
direct_product_backwards(R1,R2,Prod,WF) :-
   ((kernel_tools:ground_value(R1);kernel_tools:ground_value(R2)) -> true
    ; expand_custom_set_to_list(Prod,ProdList,_,direct_product_backwards),
      direct_product_propagate_back(ProdList,R1,R2,WF)
   ).

:- block direct_product_propagate_back(-,?,?,?).
direct_product_propagate_back([],_,_,_WF).
direct_product_propagate_back([(X,(Y,Z))|T],R1,R2,WF) :- %% print(prop_back(X,Y,Z,R1,R2)),nl, %
   check_element_of_wf((X,Y),R1,WF), check_element_of_wf((X,Z),R2,WF),
   direct_product_propagate_back(T,R1,R2,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:parallel_product([],[(int(3),int(4))],[]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:parallel_product([(int(1),int(2))],
                        [(int(3),int(4))],[((int(1),int(3)),(int(2),int(4)))]))).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))],
                                             [(int(3),int(4))],X), ground(X),
                                             equal_object(X,[((int(1),int(3)),(int(2),int(4)))]))).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))],
                                             [(int(3),int(4))],[((int(1),int(3)),(int(2),int(4)))]))).
:- assert_must_succeed((bsets_clp:parallel_product([(int(1),int(2))], [],X),X == [])).
:- assert_must_succeed((bsets_clp:parallel_product([], [(int(3),int(4))],X),X == [])).

:- block parallel_product(-,?,?),parallel_product(?,-,?).
% NOTE: we now have in_parallel_product; as such parallel products are kept symbolic
%parallel_product(Rel1,Rel2,Prod) :- (keep_symbolic(Rel1) -> true ; keep_symbolic(Rel2)),
%    print_term_summary(parallel_product(Rel1,Rel2,Prod)),nl,
%%   % TO DO: generate closure
%   %{xy,mn|#(x,y,m,n).(xy=(x,y) & mn=(m,n) & (x,m):S & (y,n):R)}
%    fail.
parallel_product(Rel1,Rel2,Prod) :-
   expand_custom_set_to_list(Rel1,Relation1,_,parallel_product_1),
   expand_custom_set_to_list(Rel2,Relation2,_,parallel_product_2),
   parallel_product2(Relation1,Relation2,ProdRes),
   equal_object_optimized(ProdRes,Prod,parallel_product).

:- use_module(kernel_equality,[conjoin_test/4]).
 %(Rel1||Rel2) = {(x,y),(m,n)| (x,m):Rel1 & (y,n):Rel2}

 % TO DO: use this in b_interpreter_check:
in_parallel_product_test(((X,Y),(M,N)),Rel1,Rel2,Result,WF) :-
   conjoin_test(MemRes1,MemRes2,Result,WF),
   % print_term_summary(in_parallel_product_test(Rel1,Rel2,Result,MemRes1,MemRes2)),nl,
   membership_test_wf(Rel1,(X,M),MemRes1,WF),
   membership_test_wf(Rel2,(Y,N),MemRes2,WF).

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:in_parallel_product_wf(((int(1),int(2)),(int(11),int(22))),[(int(1),int(11))],[(int(2),int(22))],_WF))).

in_parallel_product_wf(El,Rel1,Rel2,WF) :-
   in_parallel_product_test(El,Rel1,Rel2,pred_true,WF).


:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:not_in_parallel_product_wf(((int(1),int(11)),(int(2),int(22))),[(int(1),int(11))],[(int(2),int(22))],_WF))).

not_in_parallel_product_wf(El,Rel1,Rel2,WF) :-
   in_parallel_product_test(El,Rel1,Rel2,pred_false,WF).


:- block parallel_product2(-,?,?).
parallel_product2([],_,Out) :- empty_set(Out).
parallel_product2([(X,Y)|T],Rel2,Out) :-
   parallel_product_tuple(Rel2,X,Y,Out,Tail),
   parallel_product2(T,Rel2,Tail).

:- block parallel_product_tuple(-,?,?,?,?).
parallel_product_tuple([],_,_,Tail1,Tail2) :- equal_object(Tail1,Tail2).
parallel_product_tuple([(X2,Y2)|T],X,Y,Rel2,Tail) :-
    equal_object(Rel2,[((X,X2),(Y,Y2))|RT]),
    parallel_product_tuple(T,X,Y,RT,Tail).


% -------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1)],[int(7),int(6)],WF),WF)). %% with wf_det leads to residue custom_explicit_sets:b_not_test_closure_enum
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(1),int(7))],[int(1)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_partial_function([(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_partial_function([],[int(1)],[int(7)],_WF))).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1)],[[int(7),int(6)]],_WF),
                     X = [(int(1),[int(6),int(7)])])).
:- assert_must_fail((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(2))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(3),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(5))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(1),int(7)),(int(2),int(6)),(int(2),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(5),int(75))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(-1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_function(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('Name'),global_set('Code'),_WF),
                     X = [(fd(1,'Name'),fd(1,'Code'))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(88),fd(2,'Code'))])).
:- assert_must_fail((bsets_clp:not_partial_function(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(2),fd(2,'Code'))])).
:- assert_must_succeed((bsets_clp:not_partial_function([(fd(1,'Code'),int(1)),(fd(1,'Code'),int(2))],
                       global_set('Code'),global_set('NAT1'),_WF) )).

:- block not_partial_function(-,?,?,?).
not_partial_function([],_Domain,_Range,_WF) :- !,fail.
not_partial_function(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    %print_term_summary(not_partial_fun_dom(FF,FFDomain,Range)),
    not_subset_of_wf(FFDomain,Domain,WF).
not_partial_function(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    %print_term_summary(check_special_not_pf(FF,FFDomain,FFRange)),nl,
    not_both_subset_of(FFDomain,FFRange,Domain,Range,WF).
not_partial_function(R,Domain,Range,WF) :-
   expand_and_convert_to_avl_set(R,AER),!,
   % print_term_summary(check_is_not_avl_partial_function(avl_set(AER),Domain,Range)),
   (is_avl_partial_function(AER)
     ->  % print(is_pf),nl,
        is_not_avl_relation_over_domain_range(AER,Domain,Range,WF)
        %, print(not_in_domain_range),nl
     ;  true %,print(not_pf),nl
   ).
not_partial_function(R,Domain,Range,WF) :-
   expand_custom_set_to_list(R,ER,_,not_partial_function),
   not_pf(ER,[],Domain,Range,WF).

:- block not_pf(-,?,?,?,?).
not_pf([],_,_,_,_) :- fail.
not_pf([(X,Y)|T],SoFar,Dom,Ran,WF) :-
    membership_test_wf_with_force(SoFar,X,MemRes,WF),
    not_pf2(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_pf2(-,?,?,?,?,?,?,?).
not_pf2(pred_true,_X,_Y,_T,_SoFar,_Dom,_Ran,_WF). /* then not a function */
not_pf2(pred_false,X,Y,T,SoFar,Dom,Ran,WF) :-
    membership_test_wf_with_force(Dom,X,MemRes,WF),
    not_pf2a(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_pf2a(-,?,?,?,?,?,?,?).
not_pf2a(pred_false,_X,_Y,_T,_SoFar,_Dom,_Ran,_WF). /* function, but domain wrong */
not_pf2a(pred_true,X,Y,T,SoFar,Dom,Ran,WF) :-
    remove_element_wf_if_not_infinite_or_closure(X,Dom,Dom2,WF,_LWF,Done),  %% provide _LWF ??
    not_pf2b(Done,X,Y,T,SoFar,Dom2,Ran,WF).

:- block not_pf2b(-, ?,?,?, ?,?,?, ?).
not_pf2b(_Done, X,Y,T, SoFar,Dom2,Ran, WF) :- %print_term_summary(not_pf2b(X,Y,T,SoFar,Dom2,Ran,WF)),nl,
    add_element_wf(X,SoFar,SoFar2,WF),
    (T==[] -> not_element_of_wf(Y,Ran,WF)
     ; membership_test_wf_with_force(Ran,Y,MemRes,WF),
       prop_empty_pred_false(T,MemRes), % if T=[] -> Y must not be in Ran
       not_pf3(MemRes,T,SoFar2,Dom2,Ran,WF)).

:- block prop_empty_pred_false(-,?).
prop_empty_pred_false([],R) :- !, R=pred_false.
prop_empty_pred_false(_,_).

:- block not_pf3(-,?,?,?,?,?).
not_pf3(pred_false,_T,_SoFar,_Dom2,_Ran,_WF).   /* illegal range */
not_pf3(pred_true,T,SoFar,Dom2,Ran,WF)  :-
    not_pf(T,SoFar,Dom2,Ran,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(1),int(1)),(int(2),int(1))],global_set('NATURAL'),global_set('NATURAL'),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_function_wf([(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_function_wf([(int(2),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:partial_function([],[int(1)],[int(7)]))).
:- assert_must_succeed((bsets_clp:partial_function(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:partial_function(X,[int(1),int(2)],[int(7)]),
                     equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((findall(X,bsets_clp:partial_function(X,[int(1),int(2)],[int(7)]),L),
                          length(L,4) )).
:- assert_must_succeed((bsets_clp:partial_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)],_WF),
                     X = [([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,[int(1)],[[int(7),int(6)]],_WF),
                     X = [(int(1),[int(6),int(7)])])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(5),int(75))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(-1),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,global_set('NATURAL1'),global_set('NATURAL1'),_WF),
                     X = [(int(1),int(7)),(int(0),int(7))])).
:- assert_must_fail((bsets_clp:partial_function_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('Name'),global_set('Code'),_WF),
                     X = [(fd(1,'Name'),fd(1,'Code'))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(88),fd(2,'Code'))])).
:- assert_must_succeed((bsets_clp:partial_function_wf(X,global_set('NATURAL'),global_set('Code'),_WF),
                     X = [(int(1),fd(1,'Code')),(int(0),fd(1,'Code')),(int(2),fd(2,'Code'))])).

partial_function(R,Domain,Range) :- init_wait_flags(WF),
  partial_function_wf(R,Domain,Range,WF),
 % print(grounding_pf_wait_flags(WF)),nl,
  ground_wait_flags(WF).

:- use_module(kernel_equality,[get_cardinality_powset_wait_flag/5, get_cardinality_partial_function_wait_flag/7]).
:- block partial_function_wf(-,-,?,?).
partial_function_wf(R,_Domain,_Range,_WF) :- R==[], !.
partial_function_wf(R,Domain,Range,WF) :- (Domain==[] ; Range==[]), !, empty_set_wf(R,WF).
partial_function_wf(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(partial_fun_dom(FF,FFDomain,Range)),
    check_subset_of_wf(FFDomain,Domain,WF).
partial_function_wf(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    % same as for total_function_wf check
    %print(check_special(FF,FFDomain,FFRange)),nl,
    check_subset_of_wf(FFDomain,Domain,WF),
    check_subset_of_wf(FFRange,Range,WF). %, print(special_partial_fun(FF)),nl.
partial_function_wf(R,Domain,Range,WF) :- %print(pf(R)),nl,
   expand_and_convert_to_avl_set_warn(R,AER,partial_function_wf,'ARG : ? +-> ?'),!,
   % print_term_summary(check_is_avl_partial_function(avl_set(AER),Domain,Range)),
   is_avl_partial_function(AER),
   is_avl_relation_over_domain(AER,Domain,WF), %print(dom_ok),nl,
   is_avl_relation_over_range(AER,Range,WF).
partial_function_wf(R,Domain,Range,WF) :-  %% print(pf(R,Domain,Range,WF)),nl, %%
   %we used to call:
   get_cardinality_powset_wait_flag(Domain,partial_function_wf,WF,Card,CWF),
   % probably we should compute real cardinality of set of partial functions over Domain +-> Range ?
   % the powset waitflag uses 2^Card as priority; is the number of partial functions when Range contains just a single element
   % slows down test 1088: TO DO investigate
   % get_cardinality_partial_function_wait_flag(Domain,Range,partial_function_wf,WF,Card,_,CWF),
   %% Maybe we should only enumerate partial functions for domain variables ; e.g., not  f <+ {x |-> y} : T +-> S
    %%  print_bt_message(pf_dom_card(Card)),nl, %%%
   % probably we should use a special version when R is var
   propagate_empty_set(Domain,R),
   propagate_empty_set(Range,R),
    (var(R) -> pf_var_r(R,var,Domain,Range,Card,WF,CWF) ; pf_var_r(R,nonvar,Domain,Range,Card,WF,CWF)).

% if first argument is empty, second argument must also be empty
:- block propagate_empty_set(-,?).
propagate_empty_set([],A) :- !,
  % (A==[] -> true ; print(prop_empty(A)),nl), %%
  kernel_objects:empty_set(A). % TO DO: add WF
propagate_empty_set(_,_).

:- block pf_var_r(-,?,?,?,?,?,-).
pf_var_r(R,var,Domain,Range,_Card,WF,_CWF) :- % if R was var: see if it is now an AVL set; otherwise we have already checked it
   expand_and_convert_to_avl_set_warn(R,AER,pf_var_r,'ARG : ? +-> ?'),!,
   % print_term_summary(check_is_avl_partial_function(avl_set(AER),Domain,Range)),
   is_avl_partial_function(AER),
   is_avl_relation_over_domain(AER,Domain,WF), %print(dom_ok),nl,
   is_avl_relation_over_range(AER,Range,WF).
pf_var_r(R,_,Domain,Range,Card,WF,CWF) :-
    expand_custom_set_to_list(R,ER,_,partial_function_wf),
    % print(expanded(R,ER,CWF)),nl,
    %get_last_wait_flag(partial_fun(Domain),WF,LWF),
    pf_w(ER,[],Domain,Range,Card,_Large,WF,CWF).

pf_w(T,SoFar,Dom,Ran,Card,Large,WF,LWF) :-
    (Card==0 -> T=[]
     ; pf(T,SoFar,Dom,Ran,Card,Large,WF,LWF)).

:- block pf(-,?,?,?,?,?,?,-).
pf([],_,_,_,_,_WF,_,_LWF).
pf(LIST,SoFar,Dom,Ran,Card,Large,WF,LWF) :-
    (var(LIST) -> ListWasVar = true ; ListWasVar = false), % is ListWasVar = true we are doing the enumeration driven by LWF being ground
    LIST = [(X,Y)|T],
    dec_card(Card,NC),/* Card ensures we do not build too big lists */
    Dom \== [],
    remove_domain_element(ListWasVar,X,Y,Dom,Dom2,Large,WF,LWF,Done),
        %% prints(rem_dom(X,Dom,Dom2,Card,Done)),nl, %%
    check_element_of_wf(Y,Ran,WF),
    pf1(Done, X,Y,T,SoFar,Dom2,Ran,NC,Large,WF,LWF).
pf(AVL,SoFar,Dom,Ran,Card,Large,WF,LWF) :- nonvar(AVL),AVL=avl_set(_A),
   add_internal_error('AVL arg: ',pf(AVL,SoFar,Dom,Ran,Card,Large,WF,LWF)),fail.

:- block dec_card(-,?).
dec_card(inf,NewC) :- !, NewC=inf.
dec_card(C,NewC) :- C>0, NewC is C-1.

:- block pf1(-, ?,?,?,?,?,?,?,?,?,?).
pf1(_Done, X,_Y,T,SoFar,Dom2,Ran,Card,Large,WF,LWF) :-
    not_element_of_wf(X,SoFar,WF), /* check that it is a function */
    %% check_element_of_wf(Y,Ran,WF), % this check is now done above in pf
     % prints(check(Y,Ran)), %%
    add_new_element_wf(X,SoFar,SoFar2,WF), %print(added(SoFar2)),nl,
    pf_w(T,SoFar2,Dom2,Ran,Card,Large,WF,LWF).

remove_domain_element(ListWasVar,X,Y,Dom,Dom2,Large,WF,LWF,Done) :- compute_large(Dom,Large),
   ((ListWasVar==true,var(X),var(Y),Large==false,
     kernel_tools:ground_value(Dom))
      -> %% (X, Y are free and we drive the enumeration: we can influence which element is taken from Dom
           % print_term_summary(removing_minimal_element(X,Y,Dom)),
          remove_a_minimal_element(X,Dom,Dom2,WF,Done)  %%%%%%%%%% added Jul 15 2008
          %,print_term_summary(done(X,Dom2,Done)) )
      ;  remove_element_wf_if_not_infinite_or_closure(X,Dom,Dom2,WF,LWF,Done)
    ).
compute_large(Dom,Large) :- % check if the domain is large; ensure that we compute this only once
 (nonvar(Large) -> true
   ; var(Dom) -> true
   ; dont_expand_this_explicit_set(Dom) -> Large=large
   ; Large=false).

:- assert_must_succeed(( bsets_clp:remove_a_minimal_element(X,[int(1)],R,_WF,Done),
                        X==int(1), Done==true, R=[] )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(2), Done==true, R=[int(3)] )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(1), R=[int(2),int(3)], Done==true )).
:- assert_must_succeed(( init_wait_flags(WF), bsets_clp:remove_a_minimal_element(X,[int(1),int(2),int(3)],R,WF,Done), ground_wait_flags(WF),
                        X==int(3), R=[], Done==true )).
:- assert_must_succeed((  init_wait_flags(WF), CL=closure(['_zzzz_binary'],[integer],b(member( b(identifier('_zzzz_binary'),integer,[]),
 b(interval(b(value(int(1)),integer,[]),b(value(int(10)),integer,[])),set(integer),[])),pred,[])),
               bsets_clp:remove_a_minimal_element(X,CL,R,WF,Done), ground_wait_flags(WF),
                        X=int(9), Done==true, kernel_objects:equal_object(R,[int(10)]) )).

/* usage: restrict number of possible choices if element to remove is free */
/* select one element; and disallow all elements appearing before it in the list */
remove_a_minimal_element(X,Set,Res,WF,Done) :-
  expand_custom_set_to_list(Set,ESet,EDone,remove_a_minimal_element),
  remove_a_minimal_element2(X,ESet,EDone,Res,WF,Done).

:- use_module(kernel_equality,[get_cardinality_wait_flag/4]).
:- block remove_a_minimal_element2(?,?,-,?,?,?).
remove_a_minimal_element2(X,ESet,EDone,Res,WF,Done) :- var(ESet),
 % should not happen as we wait for EDone
  add_internal_error('Illegal call: ',remove_a_minimal_element2(X,ESet,EDone,Res,WF,Done)),
  fail.
remove_a_minimal_element2(X,ESet,_EDone,Res,WF,Done) :-
  % print(removing_minimal_element2(X,ESet,Res,WF)),nl,%
  ESet \= [],
  (ESet = [El]
   -> X=El, empty_set_wf(Res,WF), Done=true % only one choice
    ; get_cardinality_wait_flag(ESet,remove_a_minimal_element2,WF,CWF),
      remove_a_minimal_element3(X,ESet,Res,WF,Done,CWF)
  ).

:- block remove_a_minimal_element3(?,?,?,?,?,-).
remove_a_minimal_element3(X,ESet,Res,WF,Done,_) :- var(Res), !,
    append(_,[X|TRes],ESet), % WHAT IF Res has been instantiated in the meantime ???
    equal_object_wf(Res,TRes,remove_a_minimal_element2_2,WF),Done=true.
remove_a_minimal_element3(X,ESet,Res,WF,Done,_) :- print(remove_min_nonvar_res(Res)),nl,
   equal_cons_wf(ESet,X,Res,WF), Done=true.





:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_function_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3))]],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(3))],int(7)),([(int(1),int(2))],int(7))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(2),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_succeed((bsets_clp:total_function(X,[int(1)],[[int(7),int(6)]]),
                     kernel_objects:equal_object(X,[(int(1),[int(6),int(7)])]))).
:- assert_must_succeed((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [[int(7),int(6)]]),
                    kernel_objects:equal_object(X,[([(int(1),int(2))],[int(6),int(7)]),
                          ([(int(2),int(3)),(int(1),int(3))],[int(6),int(7)])]))).
:- assert_must_succeed((bsets_clp:total_function(X,[ [(int(1),int(3)),(int(2),int(3))]],
                                                [int(6)]),
                     kernel_objects:equal_object(X,[ ([(int(2),int(3)),(int(1),int(3))], int(6)) ]))).
:- assert_must_succeed((bsets_clp:total_function(X,global_set('Name'),
                 [[],[fd(1,'Code'),fd(2,'Code')],[fd(1,'Code')],[fd(2,'Code')]]),
   kernel_objects:enumerate_basic_type(X,set(couple(global('Name'),set(global('Code'))))),
    kernel_objects:equal_object(X,[(fd(3,'Name'),[fd(2,'Code')]),(fd(1,'Name'),[fd(2,'Code')]),(fd(2,'Name'),[])]))).

%:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),bsets_clp:total_function_wf(TF,global_set('Code'),
% closure([zzzz],[set(set(couple(integer,boolean)))],
%  member(identifier(zzzz),
%  pow_subset(value(closure([zzzz],[set(couple(integer,boolean))],
%  member('ListExpression'(['Identifier'(zzzz)]),
%  'Seq'(value([pred_true /* bool_true */,pred_false /* bool_false */])))))))),WF),
%  % print(tf(TF)),nl,
%   kernel_objects:equal_object(TF,[ (fd(1,'Code'), [[],[(int(1),pred_true /* bool_true */)],[(int(1),pred_true /* bool_true */),(int(2),pred_true /* bool_true */)]]),
%   (fd(2,'Code'), [[],[(int(1),pred_true /* bool_true */)],[(int(1),pred_true /* bool_true */),(int(2),pred_true /* bool_true */)]]) ]),
  % print(gr(WF)),nl,
%   kernel_waitflags:ground_wait_flags(WF) )).

:- assert_must_succeed((bsets_clp:total_function([],[],[int(7)]))).

:- assert_must_fail((bsets_clp:total_function([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(5))]))).
:- assert_must_fail((bsets_clp:total_function(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7))]))).
:- assert_must_fail((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(2))],int(7)),
                          ([(int(1),int(3)),(int(1),int(3))],int(6))]))).
:- assert_must_fail((bsets_clp:total_function(X,[[(int(1),int(2))],[(int(1),int(3)),(int(2),int(3))]],
                                                [int(7),int(6)]),
                     kernel_objects:equal_object(X,[([(int(1),int(3)),(int(1),int(3))],int(6))]))).

total_function(R,Domain,Range) :- init_wait_flags(WF),
  total_function_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).


:- assert_must_succeed((bsets_clp:total_function_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
 nonvar(X),X=[(A,B),(C,D)],A==int(1),C==int(2),\+ ground(B),\+ ground(D),  B=int(7),D=int(7)     )).

:- block total_function_wf(-,-,-,?).
total_function_wf(FF,Domain,_Range,WF) :- FF == [],!,
   empty_set_wf(Domain,WF).
total_function_wf(FF,Domain,Range,WF) :-
   Range == [],!,
   empty_set_wf(FF,WF), empty_set_wf(Domain,WF).
total_function_wf(FF,Domain,Range,WF) :-
    % TO DO: if FF or Domain nonvar but \= [] -> check if other variable becomes []
    total_function_wf1(FF,Domain,Range,WF).

:- block total_function_wf1(?,-,?,?).
total_function_wf1(FF,Domain,_Range,WF) :-
    FF==[],!,
    empty_set_wf(Domain,WF).
total_function_wf1(FF,Domain,Range,WF) :-
    nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    % print_term_summary(total_fun_dom(FF,FFDomain,Range)),
    equal_object_wf(FFDomain,Domain,total_function_wf1_1,WF).
total_function_wf1(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    equal_object_wf(FFDomain,Domain,total_function_wf1_2,WF), check_subset_of_wf(FFRange,Range,WF)
    .%, print(special_total_fun(FF)),nl.
total_function_wf1(R,Domain,Range,WF) :- nonvar(R), R=avl_set(AEF), !,
   total_function_avl_set(AEF,Domain,Range,WF).
total_function_wf1(FF,Domain,Range,WF) :-
   % TODO: error handling in case infer_value_type fails
   % want to replace FF by closure: needs to be a variable!
   var(FF),
   % if the total function can not be build up explicitely (i.e. infinite domain)
   % TODO: can / should this be relaxed?
   custom_explicit_sets:get_card_for_specific_custom_set(Domain,Card),
   Card == inf, !,
   % build up a partial function instead (fulfilling all constraints)
   partial_function_wf(SFF,Domain,Range,WF),
   % next, build up a total function mapping everything to a default value
   % this function will be overriden by the partial function to fulfilling
   % given constraints
   % 1. identifiers for closure
   kernel_objects:infer_value_type(Domain,set(DomT)),
   kernel_objects:infer_value_type(Range,set(RanT)),
   create_texpr(identifier('__domid__'),DomT,[],TDomId),
   create_texpr(identifier('__ranid__'),RanT,[],TRanId),
   % 2. domain identifier might take all values of the domain
   create_texpr(member(TDomId,b(value(Domain),set(DomT),[])),pred,[],DomMember),
   % 3. pick a single value for the range identifier
   check_element_of_wf(RangeElement,Range,WF),
   create_texpr(equal(TRanId,b(value(RangeElement),RanT,[])),pred,[],RanMember),
   % 4. conjunct and form closure (should be treated symbolically)
   conjunct_predicates([RanMember,DomMember],Pred),
   Default = closure(['__domid__','__ranid__'],[DomT,RanT],Pred),
   % 5. override default values where needed
   override_relation(Default,SFF,FF,WF).
total_function_wf1(R,Domain,Range,WF) :-   %  print(total_function(R,Domain,Range,WF)),nl,
   try_expand_and_convert_to_avl_with_check(Domain,EDomain,total_function),  /* avoid multiple expansions */
   % print(EDomain),nl,
   % TO DO: maybe avoid converting intervals which are not fully instantiated ?
   % TODO: done by clause above? % TO DO ?: if Range singleton set {R} and Domain infinite: return %x.(x:Domain|R); if Range not empty choose one element
   try_expand_and_convert_to_avl_unless_large(R,ER),
   total_function_wf2(ER,EDomain,Range,WF).

:- block total_function_wf2(?,-,?,?).
total_function_wf2(R,Domain,Range,WF) :- nonvar(R), R=avl_set(AEF), !,
   total_function_avl_set(AEF,Domain,Range,WF).
total_function_wf2(R,Domain,Range,WF) :- %print_term_summary(total_function_wf2(R,Domain,Range,WF)),
    cardinality_peano_wf(Domain,Card,WF),
    % print(tf_card(R,Domain,Card)),nl, trace,
    ((nonvar(R);ground(Card))-> true
      ; get_last_wait_flag(total_fun(Domain),WF,WF1)),
    when((nonvar(R);ground(Card);(nonvar(Card),nonvar(WF1))), /* mal 12/5/04: changed , into ; 17/3/2008: added WF1 */
          /* reason for delaying nonvar(Card): Card grounded bit by bit by cardinality; avoid
             triggering too early and missing tf_var */
       total_function1(R,Card,Domain,Range,WF
     )).

total_function_avl_set(AEF,Domain,Range,WF) :-
    %print(total_function_avl(avl_set(AEF),Dom,Range)),nl,
   (Domain = avl_set(Dom) -> is_avl_total_function_over_domain(AEF,Dom)
     ; is_avl_partial_function(AEF),
       domain_of_explicit_set(avl_set(AEF),AEF_Domain),
       equal_object_wf(AEF_Domain,Domain,total_function_avl_set,WF)
    ),
   is_avl_relation_over_range(AEF,Range,WF).
%total_function1(FF,Card,Domain,Range,WF) :- nonvar(FF),
%   custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),
%   print(dom_range_for_specific_closure(FF,FFDomain,Domain,FFRange,Range)),nl,fail.
total_function1(FF,_,Domain,Range,WF) :-
    expand_and_convert_to_avl_set_warn(FF,AEF,total_function1,'ARG : ? --> ?'),!,
    total_function_avl_set(AEF,Domain,Range,WF).
total_function1(R,Card,Domain,Range,WF) :-
      try_expand_custom_set(R,ER),
      total_function2(ER,Card,Domain,Range,WF).

total_function2(ER,Card,Domain,Range,WF) :-
   var(ER),ground(Card),!, % print(tf_var(Card)),nl,
   tf_var(TotalFunction,[],Card,Domain,Range,WF),  %print(TotalFunction),nl,
   ER=TotalFunction.
total_function2(ER,Card,Domain,Range,WF) :-
   (ground(Card) -> get_wait_flag(0,tot_fun,WF,LWF) % we seem to know the domain exactly now; see e.g. test 1316
     ; get_wait_flag(2,total_function2,WF,LWF)), % ensure we don't start binding function as soon as Card is bound; important for test 1393; should we use another priority ?
   %print(tf(ER,Card,Domain,LWF)),nl,
   tf(ER,[],Card,Domain,Range,WF,LWF).

:- block tf(-,?,-,?,?,?,?),tf(-,?,?,?,?,?,-).
%   tf(X,SoFar,C,Dom,Ran,WF,LWF) :- print_message(tf(X,SoFar,C,Dom,Ran,WF,LWF)),nl,fail.
tf([],_,0,Dom,_,WF,_) :- empty_set_wf(Dom,WF).
tf(FUN,SoFar,s(Card),Dom,Ran,WF,LWF) :- var(FUN),nonvar(Dom), % try setting up skeleton for total fun
   remove_exact_first_element(X,Dom,Dom2),not_element_of_wf(X,SoFar,WF),var(FUN),!,
   FUN = [(X,Y)|T], tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF,LWF).
tf([(X,Y)|T],SoFar,s(Card),Dom,Ran,WF,LWF) :-
    not_element_of_wf(X,SoFar,WF), %print(not_el(X)),nl,
    remove_element_wf(X,Dom,Dom2,WF), %mal: 17/3/08 changed to _wf version
    tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF,LWF).
tf1(X,Y,T,SoFar,Card,Dom2,Ran,WF,LWF) :-
    check_element_of_wf(Y,Ran,WF), % print_message(chk(Y,Ran)),
    %when((nonvar(T);nonvar(Card)), /* mal 12/5/04: changed , into ; */
    add_new_element_wf(X,SoFar,SoFar2,WF),  %%% try_expand_and_convert_to_avl
    tf(T,SoFar2,Card,Dom2,Ran,WF,LWF).

:- block tf_var(-,?,-,?,?,?).
tf_var([],_,0,Dom,_,WF) :- empty_set_wf(Dom,WF). %, print_bt_message(finished_tf_var).
tf_var([(X,Y)|T],SoFar,s(Card),Dom,Ran,WF) :-  %print_bt_message(tf_var(X,Y,SoFar,Card)),
    /* supposes that X + Y are unbound */
    /* TO DO: rewrite like enumerate <-------------------------- */
    ((var(X),var(Y)) -> true ; (print_message(warning,'Nonvar in tf_var: '),
                                print_message(warning,((X,Y))))),
    remove_exact_first_element(X,Dom,Dom2),
    not_element_of_wf(X,SoFar,WF),
 %  print_message(check_element_of_wf(Y,Ran,WF)),
    check_element_of_wf(Y,Ran,WF),
   % print_message(checked(Y,Ran)),
    add_new_element_wf(X,SoFar,SoFar2,WF),
    tf_var(T,SoFar2,Card,Dom2,Ran,WF).


%:- assert_must_succeed((bsets_clp:enumerate_total_bijection([int(1)],[int(2),int(3)],R),
%                        kernel_objects:equal_object(R, []) )).
%:- assert_must_succeed((bsets_clp:enumerate_total_bijection([int(1)],[int(3)],R),
%                        kernel_objects:equal_object(R, [[(int(1),int(3))]]) )).
%enumerate_total_bijection(Dom,Range,AllTotalFunctions) :-
%   debug_print(9,enum_total_bijection(Dom,Range)),debug_nl(9),
%   OutputVars = TF, cardinality_peano_wf(Dom,Card,none),
%   delay_findall(TF,bsets_clp:enum_tbij(TF,Card,Dom,Range),AllTotalFunctions,OutputVars).
%enum_tbij([],0,Dom,Ran) :- empty_set(Dom),empty_set(Ran).
%enum_tbij([(X,Y)|T],s(Card),Dom,Ran) :-
%    /* supposes that X + Y are unbound */
%    ((var(X),var(Y)) -> true ; (print_message(warning,'Nonvar in enum_tf: '),
%                                print_message(warning,((X,Y))))),
%    remove_exact_first_element(X,Dom,Dom2),
%    remove_exact_element(Y,Ran,Ran2),
%    enum_tbij(T,Card,Dom2,Ran2).
/* not used for the moment
:- assert_must_succeed((kernel_objects:remove_exact_element([int(1),int(2)],X,[]),X = [[int(1),int(2)]])).
:- assert_must_fail((kernel_objects:remove_exact_element([int(1),int(2)],X,[]),X = [[int(2),int(1)]])).
:- block remove_exact_element(?,-,?).
remove_exact_element(X,Set,Res) :- expand_custom_set_to_list(Set,ESet,Done,remove_exact_element),
  remove_exact_element1(Done,X,ESet,Res).
:- block remove_exact_element1(-,?,?,?).
remove_exact_element1(_,X,S,R) :- remove_exact_element2(X,S,R).
:- block remove_exact_element2(?,-,?).
remove_exact_element2(H,[H|T],T).
remove_exact_element2(E,[H|T],[H|RT]) :- not_equal_object(E,H), remove_exact_element2(E,T,RT).
*/

:- assert_must_succeed((bsets_clp:total_bijection(X,[int(1)],[int(7)]),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     kernel_objects:equal_object(X,[(int(2),int(8)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1)],[int(7),int(3)]),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(3)]),
                     X = [(int(1),int(3)),(int(2),int(3))])).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_bijection(X,[int(1),int(2)],[int(7),int(8)]),
                     X = [(int(1),int(7)),(int(1),int(8))])).



total_bijection(R,Domain,Range) :- init_wait_flags(WF),
  total_bijection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

:- block total_bijection_wf(?,-,?,?).
total_bijection_wf(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
    equal_object_wf(FFDomain,Domain,total_bijection_wf_1,WF),
    equal_object_wf(FFRange,Range,total_bijection_wf_2,WF).
%(R,Domain,Range,WF) :- Domain==Range,!, print(eq_domain_range),nl, total_injection_wf(R,Domain,Range,WF).
total_bijection_wf(R,Domain,Range,WF) :-   % print_term_summary(total_bijection_wf(R,Domain,Range)),
    cardinality_peano_wf(Domain,Card,WF),
    %print(card_domain(Domain,Card)),nl,
    cardinality_peano_wf(Range,Card,WF), % Domain and Range must have same cardinality
    %print(card_range(Range,Card)),nl,
    %when(ground(Range),(print(ground_range(Range,Card,Domain)),nl)),
    %when(nonvar(Card), (print(nonvar_card(Card,Domain,Range)),nl)),
    %when(ground(Card), (print(ground_card(Card,Domain,Range)),nl)),
    total_injection_wf2(R,Domain,Range,WF). % TO DO: use cardinality_as_int_wf ? makes test 1194 fail

%Note: we used to call custom code: total_bijection_wf2(R,Domain,Card,Range,WF).
% total_injection_wf2 gives a considerable performance boost, e.g., for test 1222  ClearSy/alloc_large.mch or NQueens with >->>

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_function([(int(1),int(6)),(int(1),int(7))],[int(1)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_function(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_function([],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([],[global_set('NAT1')],[global_set('Name')],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([(int(7),int(7))],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_function(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).

:- block not_total_function(-,?,?,?), not_total_function(?,-,?,?).
not_total_function(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    not_equal_object_wf(FFDomain,Domain,WF).
not_total_function(FF,Domain,Range,WF) :- nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    %print_term_summary(check_special_not_tf(FF,FFDomain,FFRange)),nl,
    equality_objects_wf(FFDomain,Domain,Result,WF), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
    when(nonvar(Result),(Result=pred_false -> true ; not_subset_of_wf(FFRange,Range,WF))).
% this clause is unsound: it prevents finite partial functions as solutions to not_total_function over an infinite domain
%not_total_function(FF,Domain,Range,WF) :-
%    custom_explicit_sets:get_card_for_specific_custom_set(Domain,Card),
%    Card == inf, !, % cardinality is too large for avl expansion to be viable
%    not_partial_function(FF,Domain,Range,WF). % so search for something that is not even a partial function
not_total_function(R,Domain,Range,WF) :-
   % print_term_summary(not_total_function(R,Domain,Range,WF)),
     try_expand_and_convert_to_avl_with_check(R,ER,not_total_function_range),
     try_expand_and_convert_to_avl_with_check(Domain,EDomain,not_total_function_domain),  /* avoid multiple expansions */
     % print_term_summary(exp_domain(EDomain)),
     try_expand_and_convert_to_avl_unless_large(Range,ERange),
     % print_term_summary(exp_range(ERange)),
    not_total_function2(ER,EDomain,ERange,WF).

% repeat block, in case Domain or R is a closure
:- block not_total_function2(-,?,?,?), not_total_function2(?,-,?,?).
not_total_function2(R,Domain,Range,WF) :-    %print_term_summary(not_tot2(R,Domain,Range)),
   expand_and_convert_to_avl_set_warn(R,AER,not_total_function2,'ARG /: ? --> ?'),
   !,
   not_total_function_avl(AER,Domain,Range,WF).
not_total_function2(R,EDomain,ERange,WF) :-
    expand_custom_set_to_list(R,ER,_,not_total_function2),
    %print_term_summary(not_tf(R,ER)),nl,
    not_tf(ER,[],EDomain,ERange,WF).

not_total_function_avl(_AER,Domain,_Range,_WF) :- is_infinite_explicit_set(Domain),!,
   true. % a finite AVL set cannot be a total function over an infinite domain
not_total_function_avl(AER,Domain,Range,WF) :-  
   expand_and_convert_to_avl_set_warn(Domain,ADom,not_total_function2,'? /: ARG --> ?'),
   !,
    % print_term_summary(check_is_not_avl_total_function(avl_set(AER),Domain,Range)),
   (is_avl_total_function_over_domain(AER,ADom)
     ->  % print(is_tf),nl,
        is_not_avl_relation_over_range(AER,Range,WF)  % TO DO: check only Range !
     ;  true %,print(not_pf),nl
   ).
not_total_function_avl(AER,EDomain,ERange,WF) :- 
    expand_custom_set_to_list(avl_set(AER),ER,_,not_total_function_avl),
    not_tf(ER,[],EDomain,ERange,WF). 


:- use_module(kernel_equality,[membership_test_wf_with_force/4]).

:- block not_tf(-,?,?,?,?).
not_tf([],_,Domain,_,WF) :- not_empty_set_wf(Domain,WF).
not_tf([(X,Y)|T],SoFar,Dom,Ran,WF) :- membership_test_wf_with_force(SoFar,X,MemRes,WF),
   not_tf2(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_tf2(-,?,?,?, ?,?,?,?). %, not_tf2(?,?,?,?, -,?,?), not_tf2(?,?,?,?, ?,-,?).
not_tf2(pred_true,_X,_,_T,_SoFar,_Dom,_Ran,_WF).% :- check_element_of_lazy(X,SoFar,WF).
not_tf2(pred_false,X,Y,T,SoFar,Dom,Ran,WF) :-
    %not_element_of_wf(X,SoFar,WF),
    membership_test_wf_with_force(Dom,X,MemRes,WF),
    not_tf3(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_tf3(-, ?,?,?,?, ?,?,?).
not_tf3(pred_false,_X,_Y,_T,_SoFar,_Dom,_Ran,_WF).
not_tf3(pred_true,X,Y,T,SoFar,Dom,Ran,WF) :-
    remove_element_wf(X,Dom,Dom2,WF),
     membership_test_wf_with_force(Ran,Y,MemRes,WF),
     not_tf4(MemRes,X,Y,T,SoFar,Dom2,Ran,WF).

:- block not_tf4(-, ?,?,?,?, ?,?,?).
not_tf4(pred_false,_X,_Y,_T,_SoFar,_Dom2,_Ran,_WF).
not_tf4(pred_true,X,_Y,T,SoFar,Dom2,Ran,WF) :-
       %check_element_of_wf(Y,Ran,WF), %DO WE NEED THIS ????
       add_new_element_wf(X,SoFar,SoFar2,WF),
       not_tf(T,SoFar2,Dom2,Ran,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(8),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_bijection([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1)],[int(7)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_fail((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(1),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(3)],_WF),
                     X = [(int(1),int(3)),(int(2),int(3))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1)],[int(7),int(8)],_WF),
                     X = [(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7)],_WF),
                     X = [(int(2),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_bijection([],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection([(int(7),int(7))],[int(1)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)],_WF))).
:- assert_must_succeed((bsets_clp:not_total_bijection(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).

:- block not_total_bijection(-,?,?,?).
not_total_bijection(FF,Domain,Range,WF) :-
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
    %% print(not_total_bijection_wf_closure(FFDomain,FFRange)),nl, %%
    not_equal_object_wf((FFDomain,FFRange),(Domain,Range),WF).
not_total_bijection(R,Domain,Range,WF) :-
    try_expand_custom_set(R,ER),not_tot_bij(ER,[],Domain,Range,WF).

:- block not_tot_bij(-,?,?,?,?).
not_tot_bij([],_,Domain,Range,_WF) :- empty_not_tot_bij(Domain,Range).
not_tot_bij([(X,Y)|T],SoFar,Dom,Ran,WF) :- membership_test_wf(SoFar,X,MemRes,WF),
    not_tot_bij2(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- use_module(kernel_equality,[empty_set_test/2]).
:- block empty_not_tot_bij(-,?).
empty_not_tot_bij(Domain,Range) :-
      empty_set_test(Domain,EqRes), empty_not_tot_bij2(EqRes,Range).
:- block empty_not_tot_bij2(-,?).
empty_not_tot_bij2(pred_false,_).
empty_not_tot_bij2(pred_true,Range) :-  not_empty_set(Range).

:- block not_tot_bij2(-,?,?,?,?,?,?,?).
not_tot_bij2(pred_true,_X,_,_T,_SoFar,_Dom,_Ran,_WF).
not_tot_bij2(pred_false,X,Y,T,SoFar,Dom,Ran,WF) :-
    membership_test_wf(Dom,X,MemRes,WF),
    not_tot_bij3(MemRes,X,Y,T,SoFar,Dom,Ran,WF).

:- block not_tot_bij3(-,?,?,?,?,?,?,?).
not_tot_bij3(pred_false,_X,_,_T,_SoFar,_Dom,_Ran,_WF). %  X not a member of domain
not_tot_bij3(pred_true,X,Y,T,SoFar,Dom,Ran,WF) :-
    remove_element_wf(X,Dom,Dom2,WF),
    membership_test_wf(Ran,Y,MemRes,WF),
    not_tot_bij4(MemRes,X,Y,T,SoFar,Dom2,Ran,WF).

:- block not_tot_bij4(-,?,?,?,?,?,?,?).
not_tot_bij4(pred_false,_X,_,_T,_SoFar,_Dom2,_Ran,_WF). %  Y not a member of range
not_tot_bij4(pred_true,X,Y,T,SoFar,Dom2,Ran,WF) :-
        remove_element_wf(Y,Ran,Ran2,WF),
        add_element_wf(X,SoFar,SoFar2,WF),
        not_tot_bij(T,SoFar2,Dom2,Ran2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_restriction_wf([(int(1),int(2)),(int(2),int(3))],[int(3)],[(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_restriction_wf([(int(1),int(2)),(int(2),int(3))],[int(2),int(3)],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_restriction_wf([],[int(2),int(3)],[],WF),WF)).
:- assert_must_succeed((bsets_clp:range_restriction_wf([],[int(1)],[],_WF))).
:- assert_must_succeed((bsets_clp:range_restriction_wf([],[],[],_WF))).
:- assert_must_succeed((bsets_clp:range_restriction_wf([(int(1),int(2))],[int(1)],[],_WF))).
:- assert_must_succeed((bsets_clp:range_restriction_wf([(int(1),int(2))],[int(2)],[(int(1),int(2))],_WF))).
:- assert_must_succeed((bsets_clp:range_restriction_wf(X,[fd(3,'Name')],R,_WF),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).
:- assert_must_succeed((bsets_clp:range_restriction_wf(X,Y,R,_WF),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],Y=global_set('Name'),
                        kernel_objects:equal_object(X,R))).
:- assert_must_fail((bsets_clp:range_restriction_wf(X,[fd(3,'Name')],R,_WF),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(1,'Name'))],
                        kernel_objects:equal_object(X,R))).

:- block range_restriction_wf(-,?,?,?),range_restriction_wf(?,-,-,?).

range_restriction_wf(R,S,Res,WF) :-  /* R |> S */
    ok_to_try_restriction_explicit_set(S,R,Res),
    range_restriction_explicit_set_wf(R,S,SR,WF),!,
    equal_object_wf(SR,Res,range_restriction,WF).
range_restriction_wf(R,S,Res,WF) :-  /* R |> S */
    expand_custom_set_to_list(R,ER,_,range_restriction),
    relation_restriction_wf(ER,S,Res,pred_true,range,WF).

% heuristic: should we try restriction_explicit_set or
% is relation_restriction with its stronger constraint propagation better
ok_to_try_restriction_explicit_set(S,R,Res) :-
   nonvar(S),
   (var(Res) -> true
    ; S=avl_set(_),
      nonvar(R), R=avl_set(_)  % otherwise constraint propagation from normal relation_restriction better
   ).

:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_subtraction_wf([],[int(2)],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_subtraction_wf([(int(1),int(2)),(int(2),int(3))],[int(2)],[(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_subtraction_wf([(int(1),int(2)),(int(2),int(3))],[],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:range_subtraction_wf([(int(1),int(2)),(int(2),int(3))],[int(1)],[(int(1),int(2)),(int(2),int(3))],WF),WF)).

:- block range_subtraction_wf(-,?,?,?),range_subtraction_wf(?,-,-,?).
range_subtraction_wf(R,S,Res,WF) :-  /* R |>> S */
    S==[],!,
    equal_object_wf(R,Res,range_subtraction1,WF).
range_subtraction_wf(R,S,Res,WF) :-  /* R |>> S */
    % print(range_subtraction(R,S,Res)),nl,debug:trace_in_debug_mode,
    ok_to_try_restriction_explicit_set(S,R,Res),
    range_subtraction_explicit_set_wf(R,S,SR,WF),!,
    equal_object_wf(SR,Res,range_subtraction2,WF).
range_subtraction_wf(R,S,Res,WF) :-  /* R |>> S */
    expand_custom_set_to_list(R,ER,_,range_subtraction),
    relation_restriction_wf(ER,S,Res,pred_false,range,WF).


:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_restriction_wf([int(2),int(4)],[(int(1),int(4)),(int(2),int(3))],[(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_restriction_wf([int(1),int(2)],[(int(1),int(2)),(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_restriction_wf([int(2),int(3)],[],[],WF),WF)).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(1)],[],[],_WF))).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(1)],[],R,_WF), R==[])).
:- assert_must_fail((bsets_clp:domain_restriction_wf(_,[],R,_WF), R=[int(_)|_])).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(2)],[(int(1),int(2))],[],_WF))).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([],[(int(1),int(2))],[],_WF))).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(1)],[(int(1),int(2))],[(int(1),int(2))],_WF))).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(1)],[(int(1),int(2)),(int(2),_)],_,_WF))).
:- assert_must_succeed((bsets_clp:domain_restriction_wf([int(2),int(1)],X,R,_WF),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).


:- block domain_restriction_wf(?,-,?,?),domain_restriction_wf(-,?,-,?).
domain_restriction_wf(S,R,Res,WF) :-  /* S <| R */
    %print_term_summary(domain_restriction(S,R,Res)),
    ok_to_try_restriction_explicit_set(S,R,Res),
    domain_restriction_explicit_set_wf(S,R,SR,WF),!,
     %  print_term_summary(domain_restriction_explicit_set(S,R,SR)), %%
    equal_object_wf(SR,Res,domain_restriction,WF).
domain_restriction_wf(S,R,Res,WF) :-  /* S <| R */
    % print_term_summary(call_domain_restriction(S,R,Res)), translate:print_bvalue(S),nl,
    expand_custom_set_to_list(R,ER,_,domain_restriction),
    relation_restriction_wf(ER,S,Res,pred_true,domain,WF).

% a predicate to compute domain/range restriction/subtraction
:- block relation_restriction_wf(?,-,- ,?,?,?),
         relation_restriction_wf(-,?,? ,?,?,?).
relation_restriction_wf([],_S,Res,_AddWhen,_DomOrRange,WF) :-
    %print_term_summary(domain_restriction_finished([],_S,Res)),
    empty_set_wf(Res,WF).
relation_restriction_wf([(X,Y)|T],S,Res,AddWhen,DomOrRange,WF) :-
    (DomOrRange=domain
     -> membership_test_wf(S,X,MemRes,WF)   % TO DO: pass WF !
      ; membership_test_wf(S,Y,MemRes,WF)),
    (nonvar(MemRes)
     %MemRes==AddWhen  % MemRes already set; we will ensure that (X,Y) in Res below; this slows down Alstom Compilation Regle !
      % doing the membership_test on the result Res if MemRes\==AddWhen only makes sense if we cannot fully compute the restriction ?? i.e. if T is not a closed list ?
     -> true  %,(MemRes==AddWhen -> true ; print_term_summary(relation_restriction([(X,Y)|T],S,Res,AddWhen,DomOrRange)),nl)
     ;   (AddWhen=pred_true -> InResult=MemRes
           ; bool_pred:negate(InResult,MemRes)),
          membership_test_wf(Res,(X,Y),InResult,WF)
                    % TO DO: same for explicit version; gets called e.g. if S = 1..n  (1..n <| [1,2,3] = [1,2])
                    % can now solve e.g. {x|x <| [1,2,3] = [1,2] & card(x)=2} = {{1,2}}
                    % or  x <| s = [1,2,3] \/ {29|->29}  & x <: 1..100 & s = %i.(i:1..50|i)
     ),
    % print(restrict(X,Y,MemRes)),nl,
    relation_restriction_aux(MemRes,X,Y,T,S,Res,AddWhen,DomOrRange,WF).
:- block relation_restriction_aux(-,?,?,?,?,?, ?,?,?).
relation_restriction_aux(MemRes,X,Y,T,S,Res,AddWhen,DomOrRange,WF) :-
     MemRes==AddWhen,!, % (X,Y) should be added to result  %print(add(X,Y)),nl,
     equal_cons(Res,(X,Y),RT), % was : equal_object([(X,Y)|RT],Res),
     relation_restriction_wf(T,S,RT,AddWhen,DomOrRange,WF).
relation_restriction_aux(_MemRes,_X,_,T,S,RT,AddWhen,DomOrRange,WF) :-
     % the couple is filtered out   %print(filter(_X)),nl,
     relation_restriction_wf(T,S,RT,AddWhen,DomOrRange,WF).


:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_subtraction_wf([int(1),int(3)],[(int(1),int(4)),(int(2),int(3))],[(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_subtraction_wf([int(3),int(4)],[(int(1),int(2)),(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_subtraction_wf([int(1)],[],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:domain_subtraction_wf([],[(int(11),int(21))],[(int(11),int(21))],WF),WF)).
:- assert_must_succeed((bsets_clp:domain_subtraction_wf([int(1)],[(int(1),int(2))],[],_WF))).
:- assert_must_succeed((bsets_clp:domain_subtraction_wf([int(3)],[(int(1),int(2))],[(int(1),int(2))],_WF))).
:- assert_must_succeed((bsets_clp:domain_subtraction_wf([int(1)],[(int(1),int(2)),(int(2),int(X))],R,_WF),
                         R=[(int(2),int(YY))], YY==X)).
:- assert_must_succeed((bsets_clp:domain_subtraction_wf([int(5),int(3)],X,R,_WF),
                        X = [(int(1),fd(3,'Name')),(int(2),fd(3,'Name'))],
                        kernel_objects:equal_object(X,R))).
:- block domain_subtraction_wf(?,-,?,?),domain_subtraction_wf(-,?,-,?).
%domain_subtraction(S,R,Res) :-  /* S <<| R */
%    is_custom_explicit_set(R), nonvar(S),
%    fail.
domain_subtraction_wf(S,R,Res,WF) :- S==[],!,
    equal_object_wf(R,Res,domain_subtraction1,WF).
domain_subtraction_wf(S,R,Res,WF) :-  /* S <<| R */
    ok_to_try_restriction_explicit_set(S,R,Res),
    domain_subtraction_explicit_set_wf(S,R,SR,WF),!,
    equal_object_wf(SR,Res,domain_subtraction2,WF).
domain_subtraction_wf(S,R,Res,WF) :-  /* S <<| R */
    %print_message(dom_sub(S,R,Res)),
    expand_custom_set_to_list(R,ER,_,domain_subtraction),
    try_expand_and_convert_to_avl_with_check(S,AS,domain_subtraction),
   % print_term_summary(dom_sub2(ER,AS,Res)),
   % (ground(ER) -> domain_subtraction_acc(ER,AS,[],Res) ;
                 relation_restriction_wf(ER,AS,Res,pred_false,domain,WF)
  %               )
                 .

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override_relation([(int(1),int(2))],[(int(1),int(3))],[(int(1),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override_relation([(int(1),int(2))],[(int(2),int(3))],[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- assert_must_succeed((bsets_clp:override_relation([(int(1),int(2)),(int(2),int(4))],[(int(1),int(3))],X,_WF),
                        kernel_objects:equal_object(X,[(int(2),int(4)),(int(1),int(3))]))).
:- assert_must_succeed((bsets_clp:override_relation([(int(1),int(2)),(int(2),int(4))],[(int(3),int(6))],X,_WF),
                        kernel_objects:equal_object(X,[(int(2),int(4)),(int(1),int(2)),(int(3),int(6))]))).

:- block override_relation(-,-,?,?).
override_relation(R,S,Res,WF) :- R==[],!, equal_object_wf(S,Res,override_relation1,WF).
override_relation(R,S,Res,WF) :- S==[],!, equal_object_wf(R,Res,override_relation2,WF).
override_relation(R,S,Res,WF) :- Res==[],!, empty_set_wf(S,WF), empty_set_wf(R,WF).
override_relation(R,S,Res,WF) :-  /* R <+ S */
   % print_term_summary(override_relation(R,S,Res,_WF)),nl,
   override_custom_explicit_set(R,S,ORes),!,
   % print_term_summary(override_custom_explicit_set(R,S,ORes)),
   equal_object_wf(ORes,Res,override_relation3,WF).
override_relation(R,S,Res,WF) :-  /* R <+ S */
  % print('override_relation: '), translate:print_bvalue(S),nl,
  /* TO DO: first check if S is empty */
   % print(override__enter_domain(S,DS)),nl,
    domain_wf(S,DS,WF),
 %   when(ground(DS), (print(override_domain(DS)),nl)),
  % print_term_summary(override__exit__domain(S,DS)),nl,
 %   print_term_summary(override__enter_domain_subtraction(DS,R,DSR)),
    domain_subtraction_wf(DS,R,DSR,WF),
   % print(override__exit_domain_subtraction(DS,R,DSR)),nl,
  %  when(ground(DSR), (print(override_domain_subtraction(DSR)),nl)),
   % print(calling_union(DSR,S,Res)),nl,
    union_wf(DSR,S,Res,WF). % ,print(done_union(DSR,S,Res)),nl,trace.
  %  , print_term_summary(override__exit_union(DSR,S,Res)),nl.
 %  when(ground(Res), (print(override_UNION_RESULT(R,S,Res)),nl)).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([],int(1),int(3),[(int(1),int(3))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([(int(1),int(2)),(int(2),int(6))],int(1),int(3),[(int(1),int(3)),(int(2),int(6))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:override([(int(1),int(2)),(int(2),int(6))],int(2),int(3),[(int(1),int(2)),(int(2),int(3))],WF),WF)).
:- block override(-,?,?,?,?), override(?,-,?,?,?),
         override(?,?,-,?,?). % also wait on Y; try to generate avl if possible; can only be used in substitution anyway
/*  R <+ {X |-> Y} as used by substitution R(X) := Y */
override(R,X,Y,Res,WF) :- % print_term_summary(override(R,X,Y,Res)),
     override_pair_explicit_set(R,X,Y,ORes),!,
     %% print_term_summary(override_pair_explicit_set(R,X,Y,ORes)), %%
     equal_object_wf(ORes,Res,override1,WF).
override(R,X,Y,Res,WF) :-
    if(try_expand_custom_set_to_list(R,ER,_,override),
       ( %print_term_summary(override2(ER,X,Y)),
         override2(ER,X,Y,[(X,Y)],ORes,WF),
         equal_object_wf(ORes,Res,override2,WF)),
        ( %print_term_summary(exception(R)), % Virtual Timeout exception occured
          override_relation(R,[(X,Y)],Res,WF)
        )).

:- block override2(-,?,?,?,?,?).
override2([],_X,_Y,Remainder,Res,WF) :- equal_object_optimized_wf(Remainder,Res,override2,WF). %equal_object(Remainder,Res).
override2([(V,W)|T],X,Y,Remainder,Res,WF) :-
   equality_objects_wf(V,X,EqRes,WF),
   override2c(EqRes,V,W,T,X,Y,Remainder,Res,WF).

:- block override2c(-, ?,?,?, ?,?,?,?,?).
override2c(pred_true,_V,_W,T,X,Y,_Remainder,Res,WF) :-
    equal_cons_wf(Res,(X,Y),T2,WF), %equal_object([(X,Y)|T2],Res),
    override2(T,X,Y,[],T2,WF).  /* set remainder to [], we have already added (X,Y) */
override2c(pred_false,V,W,T,X,Y,Remainder,Res,WF) :-
    equal_cons_wf(Res,(V,W),T2,WF), %equal_object([(V,W)|T2],Res),
    override2(T,X,Y,Remainder,T2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(3),int(3))],[int(1),int(2)],[int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(1),int(3)),(int(4),int(4))],[int(1),int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2)),(int(1),int(3)),(int(4),int(4))],[int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(2)],_WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(2)],[],_WF)).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2))],[int(3)],[],_WF)).
:- assert_must_succeed((bsets_clp:image_wf([(int(1),int(2)),(int(1),int(3))],
                                  [int(X)],R,_WF), X=1, kernel_objects:equal_object(R,[int(2),int(3)]))).
:- assert_must_succeed((bsets_clp:image_wf([([int(1),int(2)],int(6)),
                                   ([int(1),int(2),int(3)],int(7)),
                                   ([int(2),int(1)],int(8))],
                                  [[int(X),int(1)]],R,_WF), X=2,
				  kernel_objects:equal_object(R,[int(6),int(8)]))).
:- assert_must_succeed(bsets_clp:image_wf([(int(1),int(2)),(int(2),int(2))],[int(1),int(2)],[int(2)],_WF)).
:- assert_must_fail(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[int(1)],_WF)).
:- assert_must_fail(bsets_clp:image_wf([(int(1),int(2))],[int(1)],[],_WF)).


:- block image_wf(-,?,?,?).
image_wf(Rel,_,Res,WF) :- Rel==[],!,empty_set_wf(Res,WF).
image_wf(Rel,S,Res,WF) :- image_for_id_closure(Rel,S,Img),!, % we don't require S to be known here
  % print(image_for_id(Rel,S,Img)),nl,
  equal_object_wf(Img,Res,image_wf_id_closure,WF). %, print_term_summary(image_for_id(Rel,S,Res)).
image_wf(Rel,S,Res,WF) :-   %print_term_summary(image_wf(Rel,S,Res)),
  image_wf0(Rel,S,Res,WF).

:- block image_wf0(?,-,?,?).
image_wf0(Rel,S,Res,WF) :-  /*  Res = Rel[S] */
     %  nl,print_term_summary(image_wf0(Rel,S,Res)),nl,
    % print(image(Rel,S,Res,WF)),nl,
    (S==[] -> empty_set_wf(Res,WF)
            ; image1(Rel,S,Res,WF)  ) . %,print(done_image(Rel,S,Res)),nl. %,watch(Rel,arg1_of_image,0).

keep_symbolic(R) :- var(R),!,fail.
keep_symbolic(closure(_,_,_)) :- preferences:get_preference(convert_comprehension_sets_into_closures,true),!.
keep_symbolic(R) :- dont_expand_this_explicit_set(R).

:- block image1(-,?,?,?).
image1(Rel,S,Res,WF) :-   %print_term_summary(image1(Rel,S,Res)),
  image_for_explicit_set(Rel,S,Img,WF),!,  % print(image_explicit(Img)),nl,
  equal_object_wf(Img,Res,image1_1,WF),
  quick_propagate_subset_range(Res,Rel,WF).
%image1(Rel,S,Res,WF) :- expand_custom_set_to_list(S,ES),!, image_of_set(ES,Rel,Res,WF).
image1(Rel,Set,Res,WF) :- %print(rel(Rel,Set)),nl,
   keep_symbolic(Rel), %print(don_expand_this),nl,
   ((preferences:get_preference(convert_comprehension_sets_into_closures,true), % in this case keep_symbolic is always true
     nonvar(Set),is_infinite_explicit_set(Set)) % in this case we have to expand Rel below; what if Rel also infinite ?? --> TO DO : symbolic treatment
     -> debug_println(9,infinite_for_image1(Set)),
        fail
     ; true),
   ( dom_for_specific_closure(Rel,Domain,function(_))
     % print(image_for_inf_fun(Rel,Set)),nl,
     -> !, expand_custom_set_to_list(Set,ESet,_,image1), % TO DO: what if keep_symbolic(Set)
        image_for_inf_fun(ESet,Domain,Rel,[],Res,WF)
     ;   get_relation_types(Rel,DomType,RangeType),!,
         %% print_term_summary(image_for_large_relation(DomType,RangeType,Set,Rel)),translate:print_bvalue(Set),nl,
         expand_custom_set_to_list(Set,ESet,_,image1_2),
         kernel_tools:ground_value_check(Rel,GRel),
         when(nonvar(GRel), image_for_large_relation(ESet,Rel,DomType,RangeType,[],Res,WF))
/*     ;  get_relation_types(Rel,DomType,RangeType),!,
        % Rel should not be expanded, compute closure by calculating {yy|#(xx).(xx:Set & xx|->yy:Rel)}
         print_term_summary(image_for_large_relation(DomType,RangeType,Set,Rel)),nl,
        image_closure(Set,Rel,DomType,RangeType,Closure ),
         translate:print_bvalue(Closure),nl,
        expand_custom_set(Closure,CRes),  print(expanded),nl,
        equal_object(CRes,Res,image1_2) */
   ).
image1(Rel,S,Res,WF) :-
    %((custom_explicit_sets:quick_custom_explicit_set_approximate_size(Rel,S), S>1000) -> print(image_for_large_relation(S)),nl ; true),
    expand_custom_set_to_list(Rel,Relation,_,image1_2), % bad if Rel is a big closure !
          % print(image2(Rel,Relation,S,Res)),nl,
 %        image_for_list_relation(Relation,S,Res).
 propagate_singleton_image(Relation,S,Res,WF),
         image_for_list_relation(Relation,S,[],Res,WF). %,print_term_summary(image2_res(Relation,S,Res)).

% propagate that f[{x}] = {r1,...,rk}  =>  x|->ri : f   (or {x}*{r1,...,rk} <: f); see test 1532
propagate_singleton_image(R,S,Res,_) :-
    (var(S) ; var(Res) ; nonvar(R), is_custom_explicit_set(R,psi)), !.
propagate_singleton_image(Relation,S,avl_set(Res),WF) :-
    custom_explicit_sets:singleton_set(S,El), % we have the image by a singleton set {El}
    expand_custom_set_to_list(avl_set(Res),LR,_,prop_singleton),
    !,
    %print_term_summary(prop(LR,El,Relation)),nl,
    l_check_element_of(LR, El, Relation, WF). % propagate x|->ri : f  (will force membership)
propagate_singleton_image(_,_,_,_).

l_check_element_of([],_,_,_).
l_check_element_of([H|T],El,Relation,WF) :-
   check_element_of_wf((El,H),Relation,WF),
   l_check_element_of(T,El,Relation,WF).

% quick_propagate_in_range(Set, Relation,WF) : propagate that Set <: ran(Relation)
:- block quick_propagate_subset_range(-,?,?).
quick_propagate_subset_range(avl_set(_),_,_) :- !.
quick_propagate_subset_range([],_,_) :- !.
quick_propagate_subset_range([H|T],Relation,WF) :- is_custom_explicit_set(Relation,range_wf1),
    range_of_explicit_set(Relation,Range), !,
    quick_propagation_element_information(Range,H,WF,NewRange),
    quick_propagate_subset_range2(T,NewRange,WF).
quick_propagate_subset_range(_,_,_).

:- block quick_propagate_subset_range2(-,?,?).
quick_propagate_subset_range2([H|T],NewRange,WF) :-  !,
    quick_propagation_element_information(NewRange,H,WF,NewRange1),
    quick_propagate_subset_range2(T,NewRange1,WF).
quick_propagate_subset_range2(_,_,_).

:- use_module(btypechecker, [unify_types_strict/2]).
get_relation_types(Value,Domain,Range) :-
    kernel_objects:infer_value_type(Value,VT),
    unify_types_strict(VT,set(couple(Domain,Range))). % deal also with seq types
%    VT=set(couple(Domain,Range)).

:- block image_for_large_relation(-,?,?,?,?,?,?), image_for_large_relation(?,?,?,?,-,?,?).
image_for_large_relation([],_,_,_,Acc,Res,WF) :- equal_object_wf(Acc,Res,WF).
image_for_large_relation([XX|T],Rel,DomType,RangeType,Acc,Res,WF) :-
    %print_term_summary(computing_image(XX,Acc)),nl,
    Body = b(member(b(couple(b(value(XX),DomType,[]),
                             b(identifier(yy),RangeType,[])),couple(DomType,RangeType),[]),
                    b(value(Rel),set(couple(DomType,RangeType)),[])),pred,[]),
    % TO DO: simplify above if we have Rel = closure(P,T,B); which we usually will
    custom_explicit_sets:expand_normal_closure_direct([yy],[RangeType],Body,HRes,_Done), % do not memoize this (many different values)
    union_wf(Acc,HRes,NewAcc,WF),
    (T == [] -> equal_object_wf(NewAcc,Res,WF)
     ; image_for_large_relation(T,Rel,DomType,RangeType,NewAcc,Res,WF)).

 /*  no longer used
% construct a closure for {yy|#(xx).(xx:Set & xx|->yy:Rel)}
image_closure(Set,Rel,DomType,RangeType,Closure ) :- custom_explicit_sets:singleton_set(Set,XX),!,
    % do not set up existential quantifier if Set is singleton set
    Closure = closure([yy],[RangeType],Body),
    Body = b(member(b(couple(b(value(XX),DomType,[]),
                             b(identifier(yy),RangeType,[])),couple(DomType,RangeType),[]),
                    b(value(Rel),set(couple(DomType,RangeType)),[])),pred,[]).
image_closure(Set,Rel,DomType,RangeType,Closure ) :-
    Closure = closure([yy],[RangeType],Body),
    couple_member_pred(xx,DomType,yy,RangeType,Rel, Predxxyy),
    Body =  b(exists([b(identifier(xx),DomType,[])],
      b(conjunct(
         b(member(b(identifier(xx),DomType,[]),b(value(Set),set(DomType),[])),pred,[]), % TO DO : force evaluation !
         Predxxyy),
      pred,[])),pred,[used_ids([yy])]).
      */

% very similar to rel_compose_with_inf_fun, indeed f[S] = ran((id(S);f))
   :- block image_for_inf_fun(-,?,?,?,?,?).
image_for_inf_fun([],_Dom,_Rel2,Acc,Comp,WF) :- equal_object_wf(Acc,Comp,WF).
image_for_inf_fun([X|T],Dom,Fun,Acc,CompRes,WF) :-
   membership_test_wf(Dom,X,MemRes,WF),
   image_for_inf_fun_aux(MemRes,X,T,Dom,Fun,Acc,CompRes,WF).

:- block image_for_inf_fun_aux(-,?,?, ?,?,?,?,?).
image_for_inf_fun_aux(pred_true,X,T,Dom,Fun,Acc,CompRes,WF) :-
    % print_term_summary(image_for_inf_fun_aux(pred_true,X,T,Dom,Fun,CompRes,WF)),
   apply_to(Fun,X,FX,WF), % TO DO: generalize to image so that we can apply it also to infinite relations ?
    %hashing:my_term_hash(FX,Hash), (Hash=353256514437402551 -> trace ; Hash=340015815059493514 -> trace ; true),
   add_element_wf(FX,Acc,NewAcc,WF), % will block until Acc Known !!
    % (Acc==NewAcc -> true ; translate:print_bvalue(NewAcc),nl), %, print(add(Hash,FX)),nl,nl),
   % TO DO USE: equal_cons_wf(CompRes,FX,CT,WF) + accumulator !,
   image_for_inf_fun(T,Dom,Fun,NewAcc,CompRes,WF).
image_for_inf_fun_aux(pred_false,_X,T,Dom,Fun,Acc,Comp,WF) :-
    % print(image_for_inf_fun_aux_not_in_domain(_X)),nl,
    image_for_inf_fun(T,Dom,Fun,Acc,Comp,WF).


/*
:- block image_of_set(-,?,?,?,?), image_of_set(?,?,-,?,?).
image_of_set([],Rel,ImageSoFar,Res,WF) :- equal_object(ImageSoFar,Res).
image_of_set([H|T],Rel,ImageSoFar,Res,WF) :-
   image_of_element(Rel,H,ImageSoFar,SF2,WF),
   image_of_set(T,Rel,SF2,Res,WF).

image_of_element([],_,Acc,Res,WF) :- equal_object(Acc,Res).
image_of_element([(A,B)|T],H,Acc,Res,WF) :- equality....
image_of_element(avl_set(),H,Acc,Res,WF) :- ....
image_of_element(closure(),....
*/

% Computing the image of a relation which is stored as a list: traverse the relation
:- block image_for_list_relation(-,?,?,?,?).
image_for_list_relation([],_,_,Res,WF) :- empty_set_wf(Res,WF).
image_for_list_relation([(X,Y)|T],S,ImageSoFar,Res,WF) :-  % prints(image(X,Y,T,ImageSoFar)),trace,
 %print_term_summary(image_for_list_relation(X,Y,T,S,ImageSoFar,Res,WF)),nl,
   ((T==[], definitely_not_empty(Res))
    -> MemRes=pred_true, % we need at least one more element for Res
       check_element_of_wf(X,S,WF)
    ;  (Res==[],ImageSoFar==[]) -> MemRes=pred_false, not_element_of_wf(X,S,WF) % Result empty: X cannot be in S
    ;  membership_test_wf(S,X,MemRes,WF)
   ),
   image4(MemRes,Y,T,S,ImageSoFar,Res,WF).

definitely_not_empty(Set) :- nonvar(Set), Set \== [], \+ functor(Set,closure,3). % Set \= closure(_,_,_).

:- block image4(-, ?,?,?, ?,?,?).
image4(pred_true, Y,T,S, ImageSoFar,Res,WF) :-
  (Res==[]
    -> MemRes=pred_true, check_element_of_wf(Y,ImageSoFar,WF)
    ;  membership_test_wf(ImageSoFar,Y,MemRes,WF)
  ),
  image5(MemRes,Y,T,S,ImageSoFar,Res,WF).
image4(pred_false, _Y,T,S, ImageSoFar,Res,WF) :-
  image_for_list_relation(T,S,ImageSoFar,Res,WF).

:- block image5(-, ?,?,? ,?,?,?).
image5(pred_true,_Y,T,S,ImageSoFar,Res,WF) :- /* we have already added Y to the image */
  image_for_list_relation(T,S,ImageSoFar,Res,WF).
image5(pred_false,Y,T,S,ImageSoFar,Res,WF) :- %print(add_to_image_sofar(Y,ImageSoFar)),nl,
  add_element_wf(Y,ImageSoFar,ImageSoFar2,WF), %print(adding_y_to_image_result(Y,Res)),nl,
  kernel_objects:mark_as_non_free(Y), % has been added to image, no longer freely choosable
  equal_cons_wf(Res,Y,Res2,WF), %print(added(Y,Res)),nl,
  image_for_list_relation(T,S,ImageSoFar2,Res2,WF).

% --------------------
% a new experimental version of image; unfortunately it does not yet seem to pay off :-(
% it is not yet called from above
% use_module(bsets_clp), bsets_clp:image_for_list_relation([(int(1),int(2))],[int(1)],Result)
% use_module(bsets_clp), bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result)
% bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result), T=[(int(2),A)], ground(Result), Result==[int(2)]
% bsets_clp:image_for_list_relation([(int(1),int(2))|T],[int(1),int(3)],Result), T=[(int(3),int(55))|T2], nonvar(Result),...
/*
image_for_list_relation(List,S,Result) :-
  image_for_list_relation_propagate(List,S,PQ,PQ,true,Count),
  terminate_propagator_queue(Count,PQ),
  add_image_to_result(PQ,[],Result).

:- block add_image_to_result(-,?,?).
add_image_to_result(empty,_,Result) :- empty_set(Result).
add_image_to_result(solution(YMEMRES,Y,PQ),ImageSoFar,Result) :- % print(sol(YMEMRES,Y,PQ)),nl, %%
   (YMEMRES=pred_true
      -> membership_test(ImageSoFar,Y,MemRes),add_image_to_result_mem(MemRes,Y,PQ,ImageSoFar,Result)
      ;  add_image_to_result(PQ,ImageSoFar,Result)
   ).
:- block add_image_to_result_mem(-,?,?,?,?).
add_image_to_result_mem(pred_false,Y,PQ,ImageSoFar,Result) :- % print(add_to_result(Y,ImageSoFar,Result)),nl,
   add_element(Y,ImageSoFar,ImageSoFar2),
   equal_cons(Result,Y,Res2), % was: equal_object([Y|Res2],Result), TO DO: use _wf version
   add_image_to_result(PQ,ImageSoFar2,Res2).
add_image_to_result_mem(pred_true,_Y,PQ,ImageSoFar,Result) :- % print(skip(_Y)),nl,
  add_image_to_result(PQ,ImageSoFar,Result).

:- block image_for_list_relation_propagate(-,?, ?,?,?,?).
image_for_list_relation_propagate([],_S, _PQ1,_PQ2,_Done,0).
image_for_list_relation_propagate([(X,Y)|T],S,PQ1,PQ2,PrevDone,s(Count)) :-
   membership_test(S,X,MemRes),
   (PrevDone==true -> PQ=PQ2 ; PQ=PQ1),
   add_to_propagator_queue(MemRes,Y,PQ,NewPQ,Done),
   image_for_list_relation_propagate(T,S,PQ,NewPQ,Done,Count).

:- block terminate_propagator_queue(-,?).
terminate_propagator_queue(0,PQ) :-
  (PQ=empty -> true ;
    var(PQ) -> fail ; add_error_fail(terminate_propagator_queue,'Illegal tail (should be var): ',PQ)).
terminate_propagator_queue(s(X),PQ) :- terminate_propagator_queue2(PQ,X).
:- block terminate_propagator_queue2(-,?).
terminate_propagator_queue2(PQ,X) :-
  (PQ=solution(_,_,PQT) -> terminate_propagator_queue(X,PQT)
    ; add_error(terminate_propagator_queue2,'Illegal tail (should be solution/3): ',PQ)).

:- block add_to_propagator_queue(-,?,?,?,?).
add_to_propagator_queue(Result,ID,PropagatorQueue,NewPQ,Done) :-
  add_solution_to_queue(PropagatorQueue,Result,ID,NewPQ,Done).

add_solution_to_queue(PropagatorQueue,Result,ID,NewPQ,Done) :- %% print(add_prop(Result,ID,PropagatorQueue)),nl, %%
  (var(PropagatorQueue) -> PropagatorQueue = solution(Result,ID,NewPQ), Done=true
   ; PropagatorQueue = solution(_,_,Tail) -> add_solution_to_queue(Tail,Result,ID,NewPQ,Done)
   ; add_error_fail(add_solution_to_queue,'Illegal PropagatorQueue: ', PropagatorQueue)
  ).
  */
% -------------------- END of experimental version for image


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[int(2)],[int(1),int(2)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:image_for_closure1_wf([(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],[int(3)],[int(3)],WF),WF)).
% first experimental version for computing closure1(Rel)[S]
:- block image_for_closure1_wf(-,?,?,?),image_for_closure1_wf(?,-,?,?).
image_for_closure1_wf(Rel,S,Res,WF) :- (Rel==[] ; S==[]),!,empty_set_wf(Res,WF).
image_for_closure1_wf(Rel,Set,Res,WF) :- try_expand_and_convert_to_avl_unless_large(Set,ESet),
   image_for_closure1_wf_aux(Rel,ESet,Res,WF).

:- use_module(library(avl),[avl_height/2]).
image_for_closure1_wf_aux(Rel,S,Res,WF) :- % print_term_summary(image_for_closure1_wf_aux(Rel,S,Res,WF)),nl,
   ((nonvar(S),S=avl_set(_))
     -> closure1_for_explicit_set_from(Rel,S,Closure1Rel),!, % if S is known: start from S (currently only deals with Rel=avl_set(_)
         % print_term_summary(closure1_from(Rel,S,Closure1Rel,Res)),nl,
        range_wf(Closure1Rel,Res,WF)
     ; Rel=avl_set(AR), avl_height(AR,AR_Height),
       ((set_smaller_than(S,4),AR_Height>4)
          -> !, % TO DO: we could do the same for small S if Rel is large
             % print_term_summary(clo1(Rel)),nl,
             when(ground(S), (expand_and_convert_to_avl_set(S,ES) ->
                              closure1_for_explicit_set_from(Rel,avl_set(ES),Closure1Rel),
                              range_wf(Closure1Rel,Res,WF)
                            ; image_for_closure1_iterate(Rel,S,[],Res,WF,first_iteration(S))
                             ))
          ;  % print_term_summary(clo1(Rel)),nl,
             % Don't do this if avl_height too large; then it is probably better to compute the image for S only
             AR_Height < 13, % how big should we make this magic constant; or should we time-out ? 2^14=16384
             closure1_for_explicit_set(Rel,Closure1Rel),!, % we can compute it effiently; don't use code below
             %print_term_summary(clo1_res(Closure1Rel,S,Res)),nl,
             image_wf(Closure1Rel,S,Res,WF)
        )
   ).
image_for_closure1_wf_aux(Rel,S,Res,WF) :-
    %when(nonvar(Res),(print(inst(Res,for(S))),nl,trace)),
    % print_term_summary(image_for_closure1_iterate(Rel,S,[],Res,WF)),debug:nl_time,
    image_for_closure1_iterate(Rel,S,[],Res,WF,first_iteration(S)). %  , print_term_summary(done(Res)),debug:nl_time.

% no need to treat avl_sets; already covered as special case above
set_smaller_than([],_).
set_smaller_than([_|T],N) :- N>1, nonvar(T), N1 is N-1, set_smaller_than(T,N1).

image_for_closure1_iterate(Rel,S,Acc,Res,WF,FIRST) :-
    % print_term_summary(image_wf0(Rel,S,Acc,Res,WF)),debug:nl_time,
    image_wf0(Rel,S,Res1,WF),
    % print_term_summary(done_image_wf0(Res1)),debug:nl_time,nl,
    ground_value_check(Res1,RV),
    image_for_closure1_check_fix(RV,Rel,Acc,Res1,Res,WF,FIRST).

:- block image_for_closure1_check_fix(-,?,?,?,?,?,?).
image_for_closure1_check_fix(_,Rel,Acc,Res1,Res,WF,FIRST) :-
    %try_expand_and_convert_to_avl_unless_large(Res1,ERes1),
    difference_set(Res1,Acc,New),
    try_expand_and_convert_to_avl(New,ENew), % we compute difference_set below; we most definitely will need an explicit finite representation
    % print_term_summary(new(ENew)),debug:nl_time,
    (not_empty_set_wf(ENew,WF),
     union(ENew,Acc,Acc1), % Note: we do not call union_wf - should we do this
    % upon first iteration remove also S from New -> New2 and pass New2 to image_for_closure1_iterate
    % TO DO: investigate whether this also makes sense for further iterations; always remove S
     (FIRST=first_iteration(S) -> difference_set(ENew,S,New2) ; New2=ENew),
     image_for_closure1_iterate(Rel,New2,Acc1,Res,WF,not_first)
    ;
    empty_set_wf(ENew,WF),equal_object_optimized_wf(Acc,Res,image_for_closure1_check_fix,WF)).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:element_of_closure1_wf((int(1),int(3)),[(int(1),int(2)),(int(2),int(1)),(int(2),int(3))],WF),WF)).

% the code below could be further optimized: merge with code above and stop iterating when B appears in New
element_of_closure1_wf((A,B),Relation,WF) :-
  in_domain_wf_lazy(A,Relation,WF), % for some reason calling in_domain_wf will substantially slowdown Loop.mch (tests 634, 637)
  % alternatively we could set Relation to not-empty and hope that image_for_closure1_wf propagates this (implicitly it is already not empty as we want to have B:Image)
  % or should we do something like quick_propagation_element_information ??
  % Question: should we also require in_range_wf(B,Relation,WF) ? in case B still variable ?
 % print(in_closure1(A,B,Relation)),nl,
  image_for_closure1_wf(Relation,[A],Image,WF), %print(computed_image(Image)),nl,
  check_element_of_wf(B,Image,WF).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_element_of_closure1_wf((int(1),int(3)),[(int(1),int(2)),(int(2),int(1)),(int(3),int(3))],WF),WF)).
not_element_of_closure1_wf((A,B),Relation,WF) :- %print(not_in_closure1(A,B,Relation)),nl,
  image_for_closure1_wf(Relation,[A],Image,WF), %print(computed_image(Image)),nl,
  not_element_of_wf(B,Image,WF).

% -----------------------------------

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(2),int(22))],int(2),int(22),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(1),int(22)),(int(3),int(33)),(int(4),int(44))],int(3),int(33),WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:apply_to([(int(1),[int(22)]),(int(3),[int(32),int(33)]),(int(4),[int(44)])],int(3),[int(32),int(33)],WF),WF)).
:- assert_must_succeed(bsets_clp:apply_to([(int(1),int(2))],int(1),int(2),_WF)).
:- assert_must_succeed((bsets_clp:apply_to(F,int(3),int(2),_WF),F=[(int(3),int(2)),(int(2),int(1))])).
:- assert_must_succeed((bsets_clp:apply_to(F,X,int(1),_WF),F=[(int(3),int(2)),(int(2),int(1))],X=int(2))).
:- assert_must_succeed((bsets_clp:apply_to(F,int(3),_,_WF),F=[(int(3),[int(2),int(3)]),(int(2),[])])).

:- assert_must_fail(bsets_clp:apply_to([(int(1),int(2)),(int(1),int(3))],int(1),int(3),_WF)).
  /* input not a function */
apply_to(R,X,Y,WF) :- apply_to2(R,X,Y,R,unknown,WF).

:- block apply_to(-,-,-,?,?).
apply_to(R,X,Y,Span,WF) :-   %prints(apply(R,X,Y,WF)), tools_printing:print_arg(R),nl,
  ((var(R),var(X),preferences:preference(find_abort_values,false)) -> force_in_domain_wf(X,R,WF) ; true),
  apply_to1(R,X,Y,R,Span,WF).

force_in_domain_wf(El,S,WF) :-
  (preferences:preference(use_smt_mode,true) -> S=[(El,_)|_]
   ; get_enumeration_starting_wait_flag(not_empty_domain_wf,WF,LWF),  in_domain_lwf(El,S,LWF,WF)).

:- use_module(preferences,[preference/2]).
:- block apply_to1(-,-,?,?,?,?).
apply_to1(R,X,Y,InitalRel,Span,WF) :-   % prints(apply2(R,X,Y,InitalRel,WF)),
  (var(R) -> apply_to_var(R,X,Y,InitalRel,Span,WF)
   ;  apply_to_nonvar(R,X,Y,InitalRel,Span,WF),
      propagate_range_membership(R,Y)
   ).
:- block apply_to2(-,-,?,?,?,?).
apply_to2(R,X,Y,InitalRel,Span,WF) :-   % prints(apply2(R,X,Y,InitalRel,WF)),
  (var(R) -> apply_to_var(R,X,Y,InitalRel,Span,WF)
   ;  apply_to_nonvar(R,X,Y,InitalRel,Span,WF)
   ).

:- use_module(clpfd_lists,[get_finite_fdset_information/2,combine_fdset_information/3,
                           assert_fdset_information/2,get_fdset_information/2]).
% tested in test 1478; initially slows down NQueens
%:- block propagate_range_membership(-,?). % not necessary
propagate_range_membership([(_,RanEl)|T],X) :- nonvar(RanEl),
     preferences:preference(use_clpfd_solver,true),
    % preferences:preference(use_smt_mode,true),
     preferences:preference(find_abort_values,false),
     get_finite_fdset_information(RanEl,Info),
     \+ ground(X),
     get_fdset_information(X,InfoX),
     %print(start(Info,X,InfoX)),nl,
     Info \= InfoX, % avoids NQueens slowdown; TO DO: check if more precise than InfoX; otherwise no use in collecting info
     !,
     propagate_range_membership(T,Info,X).
propagate_range_membership(_,_).
:- block propagate_range_membership(-,?,?).
propagate_range_membership([],Info,El) :- !,
   % note: the information for the first few elements might have become more precise; TO DO: wait until list known and then propagate ?+ keep on propagating ??
    %  print(mem_propagate(Info,El)),nl,
   assert_fdset_information(Info,El). %, tools_printing:print_arg(El),nl.
propagate_range_membership([(_,RanEl)|T],Acc,X) :-
    nonvar(RanEl), % otherwise we have no info: we may just as well stop
    get_finite_fdset_information(RanEl,RInfo),
    combine_fdset_information(Acc,RInfo,NewAcc),
    NewAcc \= no_fdset_info,
     !,
    propagate_range_membership(T,NewAcc,X).
propagate_range_membership(_,_,_).



apply_to_var(R,X,Y,InitalRel,Span,WF) :-
    when((ground(X);nonvar(R)),  % only instantiate R when X sufficiently instantiated (TO DO: maybe use some for of equality_objects with existing relation R set up so far ??)
        (var(R) -> apply_to_var2(R,X,Y,InitalRel,Span,WF) ; apply_to_nonvar(R,X,Y,InitalRel,Span,WF))).
apply_to_var2(R,X,Y,_InitalRel,_,WF) :-
   preferences:preference(find_abort_values,false),
   !,
   R=[(X,Y)|Tail],
   optional_functionality_check(Tail,X,WF).
apply_to_var2([(X,Y)|_],X,Y,_,_,_).
apply_to_var2(R,X,_Y,InitalRel,Span,WF) :- not_in_domain_wf(X,R,WF),
    kernel_tools:ground_value_check(R,RV),
    when((ground(X),nonvar(RV)),
            add_wd_error_span('function applied outside of domain (#1): ', '@fun'(X,InitalRel),Span,WF)).

optional_functionality_check(Tail,X,WF) :-
   preferences:preference(disprover_mode,true),!,
   not_in_domain_wf(X,Tail,WF). % we assert that R is a function ; when disproving we can assume well-definedness
   % Note: this can cut down the search space ; see e.g. test 1230 (but e.g. it will not find a problem with test 1169, RULE_r967_1)
optional_functionality_check(_,_X,_WF). % TO DO: maybe lazily check if we have other elements with X as first arg if find_abort_values is true

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(\+ environ(no_wd_checking,true)).
apply_to_nonvar([],X,_Y,InitalRel,Span,WF) :- preferences:preference(find_abort_values,true),
   %%print(apply_to_empty(X,_Y,InitalRel)),nl,
   when(ground(X),add_wd_error_span('function applied outside of domain (#2): ', '@fun'(X,InitalRel),Span,WF)).
:- endif.
apply_to_nonvar([(X2,Y2)|T],X,Y,InitalRel,Span,WF) :-
   equality_objects_wf(X2,X,EqRes,WF),
   (var(EqRes) -> equality_objects_wf(Y2,Y,EqResY,WF), %print(apply(X,X2,EqRes,Y2,Y,EqResY)),nl,
                  prop_apply_eqxy(EqResY,EqRes) % propagate: if Y/=Y2 => X/=X2
    ; EqResY=not_called),
   % this check above is important if both Y and Y2 are instantiated byt X,X2 not yet
   % example: aload_R07_cbc.mch (Savary) or cbc_sequence check for R08_ByteArray for aload_R07 event (test 1349)
   % however: slows down test 583 !
   apply_to4(EqRes,EqResY,Y2,T,X,Y,InitalRel,Span,WF).
%apply_to_nonvar(closure(Parameters,PT,Cond),X,Y,WF) :-  !,
%   check_element_of_wf((X,Y),closure(Parameters,PT,Cond),WF).

apply_to_nonvar(avl_set(A),X,Y,_InitalRel,Span,WF) :-
           %%debug:watch(10,custom_explicit_sets:apply_to_avl_set(A,X,Y,WF)). %%
           apply_to_avl_set(A,X,Y,Span,WF).
apply_to_nonvar(closure(P,T,B),X,Y,_InitalRel,Span,WF) :-
    %is_custom_explicit_set(Closure,apply),   % should also work for avl_set,...
    (is_recursive_closure(P,T,B)  % TO DO: maybe we should do the same for functions marked as symbolic/uni-directional/computed ? (although we have new rule for check_element_of_function_closure which makes this redundant ??)
     -> % print_term_summary(apply_recursive_closure(X,P,T,B)),
        %hit_profiler:add_profile_hit(rec_apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)),
        ground_value_check(X,XV), block_apply_closure_to_nonvar_groundx(XV,X,Y,P,T,B,Span,WF)
        %when(ground(X),apply_closure_to_nonvar_groundx(X,Y,P,T,B,Span,WF))
     ;  %hit_profiler:add_profile_hit(apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)),
         apply_closure_to_nonvar(X,Y,P,T,B,Span,WF)).

:- block block_apply_closure_to_nonvar_groundx(-,?,?,?,?,?,?,?).
block_apply_closure_to_nonvar_groundx(_,X,Y,P,T,B,Span,WF) :- apply_closure_to_nonvar_groundx(X,Y,P,T,B,Span,WF).

apply_closure_to_nonvar_groundx(X,Y,P,T,B,Span,WF) :-
    kernel_tools:ground_bexpr(B),
    !, % then if the element of function succeeds there is no need to check WD
    if(check_element_of_function_closure(X,Y,P,T,B,WF),
       true, % No need to check for well-definedness; no pending choice points
       apply_closure_to_nonvar_wd_check(X,P,T,B,Span,WF) % here we need to check; it could be that the result Y was instantiated
      ).
apply_closure_to_nonvar_groundx(X,Y,P,T,B,Span,WF) :-
   apply_closure_to_nonvar(X,Y,P,T,B,Span,WF).

% if we first check preferences:preference(find_abort_values,false) to avoid a choice
% point, we get a big slow-down on Alstom models; e.g., vesg_Mar12
% WARNING: This choice point can be set up in WF0 !
apply_closure_to_nonvar(X,Y,P,T,B,_,WF) :-
    %(preferences:preference(find_abort_values,true) -> true ; !),  % slow down ???!
    check_element_of_function_closure(X,Y,P,T,B,WF) .
apply_closure_to_nonvar(X,_,P,T,B,Span,WF) :- % removing this clause doubles runtime of COMPUTE_GRADIENT_CHANGE
    %hit_profiler:add_profile_hit(clause2(X,P,T,B,Span,WF),2),
    apply_closure_to_nonvar_wd_check(X,P,T,B,Span,WF).

apply_closure_to_nonvar_wd_check(X,P,T,B,Span,WF) :-
    preferences:preference(find_abort_values,true),
    not_in_domain_wf(X,closure(P,T,B),WF),
    when((ground(X),ground(closure(P,T,B))),
           add_wd_error_span('function applied outside of domain (#3): ', '@fun'(X,closure(P,T,B)),Span,WF)).


% propagate equality_objects between range and domain elements for function application:
:- block prop_apply_eqxy(-,-).
prop_apply_eqxy(Eqy,Eqx) :- var(Eqy),!, (Eqx = pred_true -> Eqy = pred_true ; true).
prop_apply_eqxy(pred_false,pred_false).
prop_apply_eqxy(pred_true,_).

:- block apply_to4(-,?,?,?,?,?,?,?,?).
apply_to4(pred_true,EqY,Y2,Tail,X,Y,_InitalRel,_,WF) :-
  (EqY==not_called -> equal_object_wf(Y2,Y,apply_to4,WF) ; EqY = pred_true),
  optional_functionality_check(Tail,X,WF).
apply_to4(pred_false,_EqY,_Y2,T,X,Y,InitalRel,Span,WF) :- apply_to2(T,X,Y,InitalRel,Span,WF).
  % (nonvar(T) -> apply_to2(T,X,Y,InitalRel,WF)
  %  ; when((ground(X);nonvar(T)), apply_to2(T,X,Y,InitalRel,WF))).


/********************************************/
/* surjection_relation(R,Domain,Range)      */
/*         R : Domain <->> Range            */
/********************************************/
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:surjection_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(6),int(7)],WF),WF)).

surjection_relation_wf(R,Domain,Range,WF) :-
  is_surjective(R,Range,WF),
  % TODO: is not optimal since ran(R)<:Range is already implied by is_surjective and
  % checked a second time by relation_over_wf/4
  relation_over_wf(R,Domain,Range,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(6),int(7)],WF),WF)).

not_surjection_relation_wf(R,Domain,Range,WF) :-
   expand_custom_set_to_list(R,ER,Done,not_surjection_relation_wf),
   not_tot_surj_rel(ER,Done,[],Domain,Range,Range,WF).

/*********************************************/
/* total_surjection_relation(R,Domain,Range) */
/*         R : Domain <<->> Range            */
/*********************************************/
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_surjection_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_surjection_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).


:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1)],[int(11),int(12)]),L),
    lists:maplist(sort,L,SL), sort(SL,SSL), % added May15th due to change in domain_wf (bsets_clp:propagate_result_to_input); TO DO: see if we can go back to just one solution
    length(SSL,1))).
%:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11),int(12)]),L), length(L,7))).
% the new domain predicate also instantiates from result; meaning that duplicate solutions are now generated
:- assert_must_succeed((findall(SR,(bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11),int(12)]),sort(R,SR)),L), sort(L,SL),length(SL,7))).
:- assert_must_succeed((findall(R,bsets_clp:total_surjection_relation(R,[int(1),int(2)],[int(11)]),L),
    length(L,1))).

total_surjection_relation(R,Domain,Range) :- init_wait_flags(WF),
    total_surjection_relation_wf(R,Domain,Range,WF), ground_wait_flags(WF).

total_surjection_relation_wf(R,Domain,Range,WF) :-
  relation_over_wf(R,Domain,Range,WF),
  check_relation_is_total(R,Domain,WF), % calls domain which now instantiates R if Domain known
  check_relation_is_surjective(R,Range,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_surjection_relation_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

not_total_surjection_relation_wf(R,Domain,Range,WF) :-
   expand_custom_set_to_list(R,ER,Done,not_total_surjection_relation_wf),
   not_tot_surj_rel(ER,Done,Domain,Domain,Range,Range,WF).


/********************************************/
/*  partial_surjection(R,DomType,RangeType) */
/*         R : DomType +->> RangeType       */
/********************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6),int(2)],WF),WF)).
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],global_set('Name')),
                     kernel_objects:equal_object(X,[(int(2),fd(1,'Name')),(int(1),fd(2,'Name')),(int(3),fd(3,'Name'))]))).
:- assert_must_succeed((bsets_clp:partial_surjection_wf(X,[int(1),int(2),int(3)],global_set('Name'),_WF),
                     kernel_objects:equal_object(X,[(int(2),fd(1,'Name')),(int(1),fd(2,'Name')),(int(3),fd(3,'Name'))]))).
:- assert_must_succeed((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed_multiple((bsets_clp:partial_surjection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]),
                    % print(X),nl,
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6)),(int(3),int(6))]))).  /* mult. */
:- assert_must_succeed((X=[(int(2),int(7)),(int(1),int(6)),(int(3),int(6))],
                        bsets_clp:partial_surjection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]))).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(3),int(8))])).
:- assert_must_succeed_multiple((bsets_clp:partial_surjection(_X,
                        [int(1),int(2),int(3),int(4),int(5),int(6),int(7)],[int(2),int(3),int(4)]) )).
:- assert_must_fail((bsets_clp:partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(2),int(6))])).

partial_surjection(R,Domain,Range) :- init_wait_flags(WF),
  partial_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

partial_surjection_wf(R,Domain,Range,WF) :-
  check_card_greater_equal(Domain,geq,Range,CardDom,CardRange),
  (surjection_has_to_be_total_injection(CardDom,CardRange)
     % LAW: card(setX) = card(setY) =>  ff: setX +->> setY  <=>  ff: setX >-> setY
     -> total_function_wf(R,Domain,Range,WF),
        injective(R,WF)
     ;  is_surjective(R,Range,WF),
        partial_function_wf(R,Domain,Range,WF)
   ).


% check_card_greater_equal(A,B) : quick check that card(A) >= card(B); also works with infinite cardinality
% TO DO: replace by a better constraint propagating predicate (also working for partially instantiated lists,...)
% compared with computing card and setting up < constraint: will only compute card if it can be done efficiently + deals with inf
% check_card_greater_equal(SetA,EQ,SetB) ;  EQ=eq or geq
:- block check_card_greater_equal(-,?,?,?,?).
check_card_greater_equal([],_,R,0,0) :- !, empty_set(R).
check_card_greater_equal(A,EQ,B,CA,CB) :- check_card_greater_equal2(A,EQ,B,CA,CB).

:- use_module(inf_arith,[block_inf_greater_equal/2]).
:- block check_card_greater_equal2(?,?,-,?,?).
check_card_greater_equal2(A,EQ,B,CardA,CardB) :-
   efficient_card_for_set(A,CardA,CodeA),
   efficient_card_for_set(B,CardB,CodeB),!,
   %print(check_geq(A,B,CardA,CodeA,CardB,CodeB)),nl,
   call(CodeA), call(CodeB),
   %% print(check_geq_val(CardA,CardB)),nl, %%
   (EQ=eq -> CardA=CardB ; inf_arith:block_inf_greater_equal(CardA,CardB)).
check_card_greater_equal2(_A,_,_B,'?','?').


:- block is_surjective(-,-,?).
is_surjective(R,Range,WF) :- % print(is_surjective(R,Range,WF)),nl, %
  (var(R) -> setup_surj_range(Range,R)
           ; range_wf(R,Range,WF)).

setup_surj_range(Range,R) :-
   setup_range(Range,Res,DONE),  %print(setup_range(Range,Res,DONE)),nl,
   equal_when_done(Res,R,DONE).
:- block equal_when_done(?,?,-).
equal_when_done(Res,R,_DONE) :- equal_object(Res,R).


:- block setup_range(-,?,?).
setup_range(global_set(G),Res,DONE) :-
    expand_custom_set(global_set(G),ES), setup_range(ES,Res,DONE).
setup_range(freetype(ID),Res,DONE) :-
    expand_custom_set(freetype(ID),ES), setup_range(ES,Res,DONE).
setup_range(avl_set(S),Res,DONE) :- expand_custom_set(avl_set(S),ES), setup_range(ES,Res,DONE).
setup_range(closure(P,T,B),Res,DONE) :-
    expand_custom_set(closure(P,T,B),ES), setup_range(ES,Res,DONE).
setup_range([],_,done).
setup_range([H|T],[(_,H)|ST],DONE) :- setup_range(T,ST,DONE).



:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],
                                                  [int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_surjection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],
                                                  [int(7),int(6),int(2)],WF),WF)).
:- assert_must_fail((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_fail((bsets_clp:not_partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_partial_surjection(X,[int(1),int(2),int(3)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(3),int(8))])).

/* /: Domain +->> Range */
not_partial_surjection(R,Domain,Range) :- init_wait_flags(WF),
  not_partial_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

:- block not_partial_surjection_wf(-,?,?,?), not_partial_surjection_wf(?,-,?,?),
   not_partial_surjection_wf(?,?,-,?).
not_partial_surjection_wf(R,DomType,RangeType,WF) :-
    % TO DO: create a new partial_function_test kernel_equality like function instead
    % The fact that R is not a partial surjection is difficult to exploit !?
    kernel_tools:ground_value_check(R,RV),
    not_partial_surjection2(R,DomType,RangeType,WF,RV).

:- block not_partial_surjection2(?,?,?,?,-).
not_partial_surjection2(R,DomType,RangeType,WF,_) :-
    not_partial_function(R,DomType,RangeType,WF).
not_partial_surjection2(R,DomType,RType,WF,_) :-
    partial_function_wf(R,DomType,RType,WF),
    invert_relation_wf(R,IR,WF),
    not_total_relation_wf(IR,RType,DomType,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7))]))).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6)),(int(1),int(8))]))).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[(int(2),int(7)),(int(3),int(6)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     kernel_objects:equal_object(X,[]))).


total_relation_wf(R,Domain,Range,WF) :- relation_over_wf(R,Domain,Range,WF),
   check_relation_is_total(R,Domain,WF).

% this predicates assume that the relation's range and domain have already been checked
check_relation_is_total(Relation,Domain,WF) :- domain_wf(Relation,Domain,WF).
check_relation_is_surjective(Relation,Range,WF) :-
    range_wf(Relation,Range,WF). % we could also call is_surjective (which does setup_surj_range) ?

:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_relation_wf([(int(1),int(6)),(int(2),int(7)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_relation_wf([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6))])).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(7))])).
:- assert_must_fail((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(2),int(6))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(1),int(6)),(int(1),int(8))])).
:- assert_must_succeed((bsets_clp:not_total_relation_wf(X,[int(1),int(2)],[int(7),int(6)],_WF),
                     X = [(int(2),int(7)),(int(3),int(6)),(int(1),int(7))])).

:- block not_total_relation_wf(-,?,?,?).
not_total_relation_wf(FF,Domain,Range,WF) :- nonvar(FF),custom_explicit_sets:is_definitely_maximal_set(Range),
    % we do not need the Range; this means we can match more closures (e.g., lambda)
    custom_explicit_sets:dom_for_specific_closure(FF,FFDomain,function(_)),!,
    not_equal_object_wf(FFDomain,Domain,WF).
not_total_relation_wf(FF,Domain,Range,WF) :-  nonvar(FF),
    custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(_)),!,
    equality_objects_wf(FFDomain,Domain,Result,WF), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
    when(nonvar(Result),(Result=pred_false -> true ; not_subset_of_wf(FFRange,Range,WF))).
not_total_relation_wf(R,Domain,Range,WF) :-
   expand_custom_set_to_list(R,ER,Done,not_total_relation_wf), %print(not_tot_rel(ER,Domain)),nl,trace,
   not_tot_surj_rel(ER,Done,Domain,Domain,[],Range,WF).

 % can be used to check not total, not surj, not total surj relation
:- block not_tot_surj_rel(-,?,?,?,?,?,?).
not_tot_surj_rel([],_,DelDomain,_,DelRange,_,WF) :-
    (not_empty_set_wf(DelDomain,WF) ; empty_set_wf(DelDomain,WF),not_empty_set_wf(DelRange,WF)).
not_tot_surj_rel([_|_],Done,_DelDom,Dom,_DelRan,_Ran,_WF) :- nonvar(Done),
     Done \= no_check_to_be_done,
     nonvar(Dom),is_infinite_explicit_set(Dom),
     !. % a finite expanded list can never be a total relation over an infinite domain
not_tot_surj_rel([(X,Y)|T],_Done,DelDom,Dom,DelRan,Ran,WF) :-
      membership_test_wf(Dom,X,MemRes,WF),
      not_tr2(MemRes,X,Y,T,DelDom,Dom,DelRan,Ran,WF).

:- block not_tr2(-,?,?,?,?,?,?,?,?).
not_tr2(pred_false,_X,_Y,_T,_DelDom,_Dom,_DelRan,_Ran,_WF).
not_tr2(pred_true,X,Y,T,DelDom,Dom,DelRan,Ran,WF) :-
       delete_element_wf(X,DelDom,DelDom2,WF),  % set DelDom initially to [] to avoid totality check
       membership_test_wf(Ran,Y,MemRes,WF),
       not_tr3(MemRes,Y,T,DelDom2,Dom,DelRan,Ran,WF).

:- block not_tr3(-,?,?,?,?,?,?,?).
not_tr3(pred_false,_Y,_T,_DelDom2,_Dom,_DelRan,_Ran,_WF).
not_tr3(pred_true,Y,T,DelDom2,Dom,DelRan,Ran,WF) :-
        delete_element_wf(Y,DelRan,DelRan2,WF),  % set DelRan initially to [] to avoid surjection check
        not_tot_surj_rel(T,no_check_to_be_done,DelDom2,Dom,DelRan2,Ran,WF).

/******************************************/
/*  total_surjection(R,DomType,RangeType) */
/*        R : DomType -->> RangeType      */
/******************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_succeed_check((bsets_clp:total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),kernel_waitflags:ground_det_wait_flag(WF)))). %% TO DO: get rid of multiple solutions
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_succeed((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_surjection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_surjection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).


total_surjection(R,Domain,Range) :- init_wait_flags(WF),
  total_surjection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

total_surjection_wf(R,DomType,RangeType,WF) :-
   check_card_greater_equal(DomType,geq,RangeType,CardDom,CardRange),
   total_function_wf(R,DomType,RangeType,WF),
  %  setup_surj_range(RangeType,R).
   %print(totsurj(R,DomType,RangeType,CardDom,CardRange)),nl,
   (surjection_has_to_be_total_injection(CardDom,CardRange)
   % LAW: card(setX) = card(setY) =>  ff: setX -->> setY  <=>  ff: setX >-> setY
    -> injective(R,WF) % if domain and range have same cardinality: injection ensures surjectivity, and is more efficient to check/propagate; example when using queens 1..n -->> 1..n for NQueens
    ; check_relation_is_surjective(R,RangeType,WF)).
   % invert_relation_wf(R,IR,WF), total_relation_wf(IR,RangeType,DomType,WF).

surjection_has_to_be_total_injection(CardDom,CardRange) :- number(CardDom), CardDom=CardRange.
% TO DO: determine the difference in size between Dom and Range and count how many times a range element can occur multiple times (would give better incremental checking)

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7)),(int(3),int(8))],[int(1),int(2),int(3)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:not_total_surjection_wf([(int(2),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- block not_total_surjection_wf(-,?,?,?), not_total_surjection_wf(?,-,?,?),
   not_total_surjection_wf(?,?,-,?).
not_total_surjection_wf(R,DomType,RangeType,WF) :-
    %get_middle_wait_flag(not_total_surjection,WF,LWF),
    % TO DO: create a new total_function_test kernel_equality like function instead; or write custom not_total_function code
    % The fact that R is not a total surjection is difficult to exploit !?
    ground_value_check(R,RV),
    not_total_surjection2(R,DomType,RangeType,WF,RV).
:- block not_total_surjection2(?,?,?,?,-).
not_total_surjection2(R,DomType,RangeType,WF,_) :-
    not_total_function(R,DomType,RangeType,WF).
not_total_surjection2(R,DomType,RangeType,WF,_) :-
    total_function_wf(R,DomType,RangeType,WF),
    not_partial_surjection_wf(R,DomType,RangeType,WF).

/*******************************************/
/*  partial_injection(R,DomType,RangeType) */
/*         R : DomType >+> RangeType       */
/*******************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(1),int(6)),(int(4),int(7)),(int(2),int(8))],[int(1),int(2),int(3),int(4)],[int(7),int(6),int(8),int(9)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_injection_wf([(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:partial_injection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_fail((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:partial_injection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:partial_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).


partial_injection(R,Domain,Range) :- init_wait_flags(WF),
  partial_injection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

:- block partial_injection_wf(-,-,?,?).
partial_injection_wf(FF,Domain,Range,WF) :- nonvar(FF),
   custom_explicit_sets:dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),!,
   check_subset_of_wf(FFDomain,Domain,WF),
   check_subset_of_wf(FFRange,Range,WF).
partial_injection_wf(R,DomType,RangeType,WF) :-
    try_expand_and_convert_to_avl_unless_large(R,ER),
    partial_function_wf(ER,DomType,RangeType,WF),
    injective(ER,WF).
 %   invert_relation_wf(R,IR,WF),
 %    print(pf(IR)),nl,
  %  partial_function_wf(IR,RangeType,DomType,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:injective([(int(1),int(6)),(int(4),int(7)),(int(2),int(8))],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:injective([(int(1),int(6)),(int(4),int(7)),(int(2),int(7))],WF),WF)).

:- block injective(-,?).
injective(FF,_WF) :- %tools_printing:print_term_summary(injective(FF)),nl,
   custom_explicit_sets:dom_range_for_specific_closure(FF,_FFDomain,_FFRange,function(bijection)),!.
injective(avl_set(AVL),_WF) :- !,
   is_injective_avl_relation(AVL,_Range). % seems slightly faster than code below
injective(Rel,WF) :- expand_custom_set_to_list(Rel,ERel,_,injective),
   injective(ERel,[],WF).

%:- use_module(library(lists),[maplist/3]).
% for FD-sets we could setup all_different constraint
:- block injective(-,?,?).
injective([],_SoFar,_). % :- print(all(SoFar)),nl, 
%  (maplist(get_fd_val,SoFar,FDL) -> clpfd:all_distinct(FDL) ; true). %clpfd_interface:clpfd_alldifferent(FDL) ; true).
%get_fd_val(int(H),H).
injective([(_From,To)|T],SoFar,WF) :-  %print(inj_check(_From,To,SoFar,T)),nl,
    not_element_of_wf(To,SoFar,WF), /* check that it is injective */
   %print(not_el(To)),nl,
    add_new_element_wf(To,SoFar,SoFar2,WF),     %SoFar2=[To|SoFar],  could also work and be faster ?
    %print(added(To,SoFar2)),nl,
   injective(T,SoFar2,WF).
% no case for global_set: it cannot be a relation; two cases below not required because of expand_custom_set_to_list
%injective(avl_set(S),SoFar,WF) :- expand_custom_set(avl_set(S),ES), injective(ES,SoFar,WF).
%injective(closure(P,T,B),SoFar,WF) :- expand_custom_set(closure(P,T,B),ES), injective(ES,SoFar,WF).



/* /: Dom >+> R */

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(1),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(2),int(8))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:not_partial_injection([(int(1),int(6)),(int(3),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- block not_partial_injection(-,?,?,?), not_partial_injection(?,-,?,?),
   not_partial_injection(?,?,-,?).
not_partial_injection(R,DomType,RangeType,WF) :-
    %get_middle_wait_flag(not_partial_injection,WF,LWF),
    % TO DO: create a new partial_function_test kernel_equality like function instead
    %ground_value_check(R,RV),
    kernel_tools:ground_value_check((R,DomType),RDV),
    not_partial_injection2(R,DomType,RangeType,WF,RDV).
% TO DO: should we check dom_range_for_specific_closure ?
:- block not_partial_injection2(?,?,?,?,-).
not_partial_injection2(R,DomType,RType,WF,_) :-
   not_partial_function(R,DomType,RType,WF).
not_partial_injection2(R,DomType,RType,WF,_) :-
    partial_function_wf(R,DomType,RType,WF),
    invert_relation_wf(R,IR,WF),
%    print(not_partial_function(IR,RType,DomType)),nl,
    not_partial_function(IR,RType,DomType,WF).

/*****************************************/
/*  total_injection(R,DomType,RangeType) */
/*       R : DomType >-> RangeType       */
/*****************************************/

:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:total_injection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:total_injection_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_total_injection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_total_injection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((bsets_clp:total_injection(X,[int(1)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(1),int(7))]))).
:- assert_must_succeed((bsets_clp:total_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(6))]))).
:- assert_must_fail((bsets_clp:total_injection(X,[int(1),int(2)],[int(7)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(1),int(7))]))).
:- assert_must_fail((bsets_clp:total_injection([],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection([(int(7),int(7))],[int(1)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection([(int(1),int(7)), (int(2),int(1))],
                                                 [int(1),int(2)],[int(7)]))).
:- assert_must_fail((bsets_clp:total_injection(X,[int(1),int(2)],[int(7),int(6)]),
                      kernel_objects:equal_object(X,[(int(2),int(7)),(int(2),int(6))]))).


total_injection(R,Domain,Range) :- init_wait_flags(WF),
  total_injection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

:- block total_injection_wf(?,-,?,?).
% Note: no need to check:  dom_range_for_specific_closure(FF,FFDomain,FFRange,function(bijection)),
total_injection_wf(R,DomType,RangeType,WF) :-   % print_message(total_inj(R)),
    check_card_greater_equal(RangeType,geq,DomType,_,_), % there must be more Range elements than domain elements; pigeonhole principle
    total_injection_wf2(R,DomType,RangeType,WF).
total_injection_wf2(R,DomType,RangeType,WF) :-
    try_expand_and_convert_to_avl_unless_large(R,ER),
    total_function_wf(ER,DomType,RangeType,WF),  %  print_message(total_f(ER)),
    injective(ER,WF).


:- block not_total_injection(-,?,?,?), not_total_injection(?,-,?,?),
   not_total_injection(?,?,-,?).
not_total_injection(R,DomType,RangeType,WF) :-
    %get_middle_wait_flag(not_total_injection,WF,LWF),
    % TO DO: create a new total_function_test kernel_equality like function instead
    kernel_tools:ground_value_check((R,DomType),RDV),
    not_total_injection2(R,DomType,RangeType,WF,RDV).

:- block not_total_injection2(?,?,?,?,-).
not_total_injection2(R,DomType,RangeType,WF,_) :- %print(not_tot_inj(R)),nl,
   not_total_function(R,DomType,RangeType,WF) . %, print(not_tot_fun(R)),nl.
not_total_injection2(R,DomType,RangeType,WF,_) :- % print(try2(R)),nl,
   total_function_wf(R,DomType,RangeType,WF), %print(tot_fun(R)),nl,
   invert_relation_wf(R,IR,WF), %print(inv(IR)),nl,
   not_partial_function(IR,RangeType,DomType,WF).

/***********************************/
/*  partial_bijection(R,DomType,RangeType) */
/*   R : DomType >+>> RangeType    */
/***********************************/


:- assert_must_succeed(exhaustive_kernel_check_wfdet(bsets_clp:partial_bijection_wf([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(bsets_clp:partial_bijection_wf([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed((partial_bijection(X,[int(1),int(2)],[int(7),int(6)]),
                     kernel_objects:equal_object(X,[(int(1),int(6)),(int(2),int(7))]))).
:- assert_must_succeed((partial_bijection(X,[int(1),int(2),int(3),int(4)],[int(7),int(6)]),
                     X = [(int(2),int(7)),(int(3),int(6))])).
:- assert_must_fail((partial_bijection(X,[int(1),int(2)],[int(7),int(6),int(5)]),
                     X = [(int(2),int(7)),(int(1),int(6))])).

partial_bijection(R,Domain,Range) :- init_wait_flags(WF),
  partial_bijection_wf(R,Domain,Range,WF),
  ground_wait_flags(WF).

partial_bijection_wf(R,DomType,RangeType,WF) :-
    partial_injection_wf(R,DomType,RangeType,WF),
    partial_surjection_wf(R,DomType,RangeType,WF).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(1),int(6)),(int(2),int(7))],[int(1),int(2)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_bijection([(int(1),int(6)),(int(2),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6)],WF),WF)).

:- assert_must_succeed(exhaustive_kernel_fail_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(3),int(6))],[int(1),int(2),int(3),int(4)],[int(7),int(6)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(bsets_clp:not_partial_bijection([(int(2),int(7)),(int(1),int(6))],[int(1),int(2)],[int(7),int(6),int(5)],WF),WF)).


:- block not_partial_bijection(-,?,?,?), not_partial_bijection(?,-,?,?),
   not_partial_bijection(?,?,-,?).
not_partial_bijection(R,DomType,RangeType,WF) :-
    %get_middle_wait_flag(not_partial_bijection,WF,LWF),
    kernel_tools:ground_value_check((R,DomType,RangeType),LWF),
    not_partial_bijection2(R,DomType,RangeType,WF,LWF).
:- block not_partial_bijection2(?,?,?,?,-).
not_partial_bijection2(R,DomType,RangeType,WF,_) :-
    not_partial_injection(R,DomType,RangeType,WF).
not_partial_bijection2(R,DomType,RangeType,WF,_) :-
    partial_injection_wf(R,DomType,RangeType,WF),
    not_partial_surjection_wf(R,DomType,RangeType,WF).


/* ------------------ */

/* The transitive & reflexive closure of a relation */

:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4))],X),
                        kernel_objects:equal_object(X,[(int(1),int(1)),(int(1),int(4)),(int(4),int(4))]))).
:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4)),(int(4),int(2))],X),
                        kernel_objects:equal_object(X,[(int(1),int(1)),(int(2),int(2)),(int(4),int(4)),(int(1),int(4)),(int(4),int(2)),
                                             (int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:relational_reflexive_closure([(int(1),int(4)),(int(4),int(2)),(int(2),int(3))],X),
kernel_objects:equal_object(X,[(int(1),int(1)),(int(2),int(2)),(int(3),int(3)),(int(4),int(4)),
                                      (int(1),int(4)),(int(4),int(2)),(int(2),int(3)),
                                             (int(4),int(3)),(int(1),int(2)),(int(1),int(3))]))).

:- block relational_reflexive_closure(-,?).
relational_reflexive_closure(Relation,Result) :-
    try_expand_and_convert_to_avl_with_check(Relation,ARelation,relational_reflexive_closure),
    relational_reflexive_closure2(ARelation,Result).
:- block relational_reflexive_closure2(-,?).
relational_reflexive_closure2(ARelation,Result) :- % print_term_summary(closure(ARelation,Result)),
    (closure_for_explicit_set(ARelation,Res)
      -> kernel_objects:equal_object(Res,Result)
      ; (expand_custom_set_to_list(ARelation,ERelation,_,relational_reflexive_closure2),
         is_full_relation(ERelation,WaitVar),
         compute_trans_closure(ERelation,Result1,WaitVar),
		% print(adding_id(ERelation,Result1)),nl,
		 rel_iterate0(ERelation,Result1,Result)
		%, print(done(Result)),nl
        )
    ).



/* The transitive (not reflexive) closure of a relation */

:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(1),int(2)),(int(2),int(6))],[(int(1),int(2)),(int(1),int(6)),(int(2),int(6))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(1),int(2)),(int(2),int(6)),(int(1),int(3))],[(int(1),int(2)),(int(1),int(3)),(int(1),int(6)),(int(2),int(6))]))).
:- assert_must_succeed(exhaustive_kernel_check(bsets_clp:relational_trans_closure([(int(6),int(7)),(int(1),int(2)),(int(2),int(6)),(int(1),int(3))],[(int(1),int(2)),(int(1),int(3)),(int(1),int(6)),(int(2),int(6)),(int(1),int(7)),(int(2),int(7)),(int(6),int(7))]))).
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4))]))).
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4)),(int(4),int(2))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4)),(int(4),int(2)),
                                             (int(1),int(2))]))).
:- assert_must_succeed((bsets_clp:relational_trans_closure([(int(1),int(4)),(int(4),int(2)),(int(2),int(3))],X),
                        kernel_objects:equal_object(X,[(int(1),int(4)),(int(4),int(2)),(int(2),int(3)),
                                             (int(4),int(3)),(int(1),int(2)),(int(1),int(3))]))).

% transitive closure for relations (closure1)
:- block relational_trans_closure(-,?).
relational_trans_closure(Relation,Result) :- % print_term_summary(closure1(Relation,Result)),
    try_expand_and_convert_to_avl_with_check(Relation,ARelation,relational_trans_closure),
    relational_trans_closure2(ARelation,Result).
:- block relational_trans_closure2(-,?).
relational_trans_closure2(ARelation,Result) :- %print(closure1(ARelation)),nl,
    (closure1_for_explicit_set(ARelation,Res)
      -> % print_term_summary(closure1_explicit(ARelation,Res)),
         kernel_objects:equal_object(Res,Result)
      ;  expand_custom_set_to_list(ARelation,ERelation,_,relational_trans_closure2),
         is_full_relation(ERelation,WaitVar),
         compute_trans_closure(ERelation,Result,WaitVar)
    ).

:- block compute_trans_closure(?,?,-).
compute_trans_closure(Relation,Result,_) :- % print(computing_trans_closure(Relation)),nl,
  compute_trans_closure(Relation,Result). %,print(result(Result)),nl.

compute_trans_closure(Relation,Result) :- %print_message(iter(Relation)),
    one_closure_iteration(Relation,Relation,Relation,Result1),
    ( equal_object(Result1,Relation), equal_object_optimized(Result,Result1,compute_trans_closure)
      ; % TO DO: use reification  instead of is_full_relation check above ??
      not_equal_object(Result1,Relation), compute_trans_closure(Result1,Result)
    ).

one_closure_iteration([],_,Res,Res).
one_closure_iteration([(X,Y)|T],ExpandedPreviousRel,PreviousRel,OutRel) :-
    add_tuples(ExpandedPreviousRel,X,Y,PreviousRel,IntRel),
    one_closure_iteration(T,ExpandedPreviousRel,IntRel,OutRel).


add_tuples([],_,_,OutRel,OutRel).
add_tuples([(X,Y)|T],OX,OY,InRel,OutRel) :-
   ( equal_object(Y,OX),add_element((X,OY),InRel,IntRel) ;
     not_equal_object(Y,OX), IntRel = InRel),
   add_tuples(T,OX,OY,IntRel,OutRel).


:- assert_must_succeed((is_full_relation(X,R),var(R),X=[],R==true)).
:- assert_must_succeed((is_full_relation(X,R),var(R),X=[(A,B)|T],var(R),A=int(1),var(R),B=A,var(R),T=[],R==true)).
:- block is_full_relation(-,?).
is_full_relation([],R) :- !,R=true.
is_full_relation([H|T],W) :- !, is_full_relation_aux(H,T,W).
is_full_relation(X,R) :-
    add_internal_error('Illegal Set for is_full_relation: ',is_full_relation(X,R)),fail.

:- block is_full_relation_aux(-,?,?).
is_full_relation_aux((X,Y),T,W) :- !, is_full_relation_aux2(X,Y,T,W).
is_full_relation_aux(X,T,W) :-
    add_internal_error('Illegal Set for is_full_relation: ',is_full_relation_aux(X,T,W)),fail.
:- block is_full_relation_aux2(-,?,?,?), is_full_relation_aux2(?,-,?,?).
is_full_relation_aux2(_X,_Y,T,W) :- is_full_relation(T,W).

/* ------------------ */
