% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(haskell_csp,[%clear_cspm_spec/0, % triggerred by clear specification event
                       parse_and_load_cspm_file/1, load_cspm_pl_file/1,
                       cspm_file_add_remove_unicode/3,

                       cspm_trans_enum/3,
                       animatable_process/1, animatable_process_cli/1,animatable_process_without_arguments/1,
                       channel/2, %cspPrintCompiled/2,
                       evaluate_argument/2, force_evaluate_argument/2, evaluate_int_argument/2,
                       force_evaluate_argument_for_member_check/2,
                       check_boolean_expression/1,

                       enumerate_channel_input_value/4,
                       enumerate_datatype_el/5,

                       is_a_datatype/2, csp_constructor/3,
                       %% csp_full_type_constructor/3,csp_full_type_constant/2, % exported by haskell_csp_analyzer.pl
                       dataTypeDef/2,
                       channel_type_list/2,
                       agent_compiled/3, % called in slicer_csp.pl ???
                       is_not_infinite_type/1,
                       get_symbol_span/2,
                       extract_span_from_event/4, extract_span_info/2,
                       get_internal_csp_representation/1,
                       get_csp_assertions_as_string/2,get_csp_assertions/1, get_csp_processes/1,
                       translate_csp_assertion/2, %translate_csp_assertions/2, %get_csp_assertions_enum/2,
                       parse_and_load_cspm_file_into_specific_pl_file/2,
			evaluate_type_list/2,
			%% get_formula_from_cspm_file/2,
      filter_formulas_from_pragmas/3,
			add_error_with_span/4,add_internal_error_with_span/4,
			evaluate_csp_expression/2,evaluate_csp_expression/3, parse_single_csp_expression/3,
			parse_single_csp_declaration/3,
			add_symbol_error/4,
			bindval/3, agent/3, agent_curry/3, subTypeDef/2, nameType/2,
			symbol/4, cspTransparent/1, cspPrint/1, valid_constant/1,
			parse_single_csp_expression_file/3,
                        %% perform_csp_self_check/0,
                        normalise_cspm_state/2,
                        check_compiled_term/1
                       ]).


:- use_module(probsrc(module_information)).
:- module_info(group,csp).
:- module_info(description,'Interpreter for CSP.').

:- use_module(library(lists)).
:- use_module(library(file_systems)).
%% :- use_module(library(process)).
%% :- use_module(library(system)).

/* the new CSP interpreter using the format produced by
   Marc's Haskell compiler fecsp/cspcomp */

/********* PROB modules *********/
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(typechecker)).
:- use_module(probsrc(kernel_objects),[top_level_dif/2,enumerate_basic_type/2]).
:- use_module(probsrc(tools),[safe_on_exception/3,string_concatenate/3,flatten/2,ajoin/2,split_atom/3]).
:- use_module(probsrc(preferences),[preference/2, set_preference/2, get_preference/2,init_preferences/0]).
:- use_module(probsrc(system_call),[system_call/4,system_call/5, get_writable_compiled_filename/3, read_all/2]).
:- use_module(probsrc(kernel_waitflags)).
:- use_module(probsrc(translate),[translate_cspm_state/2,translate_cspm_expression/2]).   
:- use_module(probsrc(specfile),[process_algebra_mode/0,csp_with_bz_mode/0,currently_opened_file/1]).
/******** ------------ *********/

/* Predicate for performing self check for all CSP modules */

:- public perform_general_csp_self_check/0.
perform_general_csp_self_check :-
   perform_csp_self_check,
   get_preference(use_clpfd_solver,Value),
   (Value=false ->
          NewValue = true
   ; NewValue = false),
   set_preference(use_clpfd_solver,NewValue),
   perform_csp_self_check,
   % reverting to the old use_clpfd_solver preference
   set_preference(use_clpfd_solver,Value).

perform_csp_self_check :-
   init_preferences,
   perform_self_check(csp_basic),
   perform_self_check(csp_tuples),
   perform_self_check(csp_sequences),
   perform_self_check(csp_sets),
   perform_self_check(haskell_csp),
   perform_self_check(haskell_csp_analyzer).

/* ------------------------------------- */


:- dynamic channel/2, bindval/3, agent/3, agent_curry/3, dataTypeDef/2, subTypeDef/2.
:- dynamic nameType/2, symbol/4, cspTransparent/1, cspPrint/1, assertRef/5. 
:- dynamic assertModelCheck/3, assertModelCheckExt/4, pragma/1.
:- dynamic assertLtl/4, assertCtl/4.

:- discontiguous channel/2, agent/3, bindval/3, assertModelCheck/3.

:- use_module(probcspsrc(haskell_csp_analyzer)).
:- use_module(probcspsrc(csp_tuples)).
:- use_module(probcspsrc(csp_sets)).
:- use_module(probcspsrc(csp_basic)).

/* ------------------------------------- */
/* Sample .pl Translation of a CSP file */

dataTypeDef('FRUIT',[constructor(apples),constructor(oranges),constructor(pears)]).

/* for testing is_not_infinite_type predicate */

dataTypeDef('Msg',[constructor('A'),constructor('B'),constructorC('con',dotTupleType(['SubMsg']))]).
dataTypeDef('Msg1',[constructor('A'),constructor('B')]).
dataTypeDef('SubMsg',[constructor('KA'),constructorC(conc,dotTupleType(['Msg'])),constructorC(ste,dotTupleType([builtin_call('Seq'('Msg'))]))]).
dataTypeDef('Inf',[constructorC(ste,dotTupleType([builtin_call('Seq'('FRUIT'))]))]).
/* for testing is_not_infinite_type predicate */

channel(left,type(dotTupleType(['FRUIT']))).
channel(right,type(dotTupleType(['FRUIT']))).
channel(mid,type(dotTupleType(['FRUIT']))).
bindval('SEND',prefix(span,[in(_x)],left,prefix(span,[out(_x)],mid,prefix(span,[],ack,val_of('SEND'))),span2),span).
bindval('REC',prefix(span,[in(_x)],mid,prefix(span,[out(_x)],right,prefix(span,[],ack,val_of('REC'))),span2),span).
bindval('SYSTEM',sharing(closure([mid,ack]),val_of('SEND'),val_of('REC'),span),span).
bindval('GEN1',prefix(span,[out(oranges)],left,prefix(span,[out(oranges)],right,val_of('GEN2')),span2),span).
bindval('GEN2',prefix(span,[out(pears)],left,prefix(span,[out(pears)],right,val_of('GEN1')),span2),span).
bindval('MAIN',sharing(closure([left,right]),val_of('SYSTEM'),val_of('GEN1'),span),span).
agent('P'(_),'prefix'('src_span'(8,8,8,9,87,1),['in'(_i2)],'c',
              'prefix'('src_span'(8,15,8,16,94,1),['out'(_i2)],'c',
           'stop'('src_span'(8,22,8,26,101,4)),'src_span'(8,19,8,21,97,10)),
              'src_span'(8,12,8,14,90,17)),'src_span'(8,8,8,26,87,18)).
bindval('P1',agent_call(span,'Q',[int(10)]),span).
agent('P2'(_n),agent_call(span,'Q',['-'(_n, int(10))]),span).
agent('P1',agent_call(span,'Q',[int(10)]),span).
agent('Q'(_n),ifte('=='(_n,int(0)),stop(span),prefix(span,[],'a',agent_call(span,'Q',['-'(_n,'int'(1))]),span),span,span,span),span).
symbol('SEND','SEND',src,'Process').
symbol('REC','REC',src(8,1,8,4,4,3),'Procsess').
symbol('SYSTEM','SYSTEM',src,'Process').
symbol('GEN1','GEN1',src,'Process').
symbol('GEN2','GEN2',src,'Process').
symbol('MAIN','MAIN',src(3,1,3,5,20,4),'Process').
assertRef('False','val_of'('MAIN',span),'Trace','val_of'('PROB_TEST_TRACE',span),span).
assertRef('True', 'val_of'('PROB_TEST_TRACE',span),'Trace', 'val_of'('MAIN',span),span).
assertModelCheckExt('False', sharing(closure([left,right]),val_of('SYSTEM',span),val_of('GEN1',span),span), 'Deterministic','F').
assertModelCheckExt('True', sharing(closure([left,right]),val_of('SYSTEM',span),val_of('GEN1',span),span), 'DeadlockFree','FD').
assertModelCheck('False','prefix'(span,[],'dotTuple'(['a','int'(1)]),'val_of'('P',span),span),'LivelockFree').
pragma('assert_ltl "GF(a)"').
pragma('ASSERT_LTL "GF(b)"').
pragma('ASSERT_LTL_TRUE "true"').

reset_for_selfcheck :- % TO DO: reset all facts and link with event_handling
  retractall(pragma(_)),
  assert(pragma('assert_ltl "GF(a)"')), assert(pragma('ASSERT_LTL "GF(b)"')), assert(pragma('ASSERT_LTL_TRUE "true"')).


% these predicates are generated by the csp analyzer, there are not the standard generated predicates from the translation
% process csp -> pl

/* ------------------------------------- */

/* ------------------------------------- */

animatable_process_without_arguments(X) :-
   symbol(RenamedX,X,_,_),
   (is_csp_process(RenamedX);is_possible_csp_process(RenamedX)).

animatable_process(X) :- 
   (symbol(RenamedX,X,_,_),(is_csp_process(RenamedX);is_possible_csp_process(RenamedX))) ; 
   (nonvar(X), X=agent_call(_Src,_F,_Args)) ; 
   (nonvar(X), X=agent_call_curry(_F,_Args)) ; 
   (agent(F,_,_),is_csp_process(F),animatable_process_with_arguments(F,proc(FName,N)),proc_with_arguments_to_string(FName,N,X)).

animatable_process_cli(X) :-
   symbol(RenamedX,X,_,_),
   (is_csp_process(RenamedX) ; is_possible_csp_process(RenamedX)).

animatable_process_with_arguments(X,proc(F,N)) :-
    functor(X,F,N),N>0.

check_compiled_term(Term)  :-
    nonvar(Term),!,
    Term=agent_call(_Src,F,Args),
    length(Args,N),
    ( (symbol(RenamedF,F,_,_),functor(P,RenamedF,N),is_csp_process(P)) -> 
         true
    ; translate_cspm_state(Term,Proc),
      add_error_fail(check_compiled_term, 'There is no such process in file: ',Proc)
    ).

/* ------------------------------------- */

/* Predicates for calling the CSPM Parser and loading the corresponding to the Csp File Prolog File */

parse_and_load_cspm_file(CSPFile) :-
   get_writable_compiled_filename(CSPFile,'.pl',PrologFile),
   print_message(informational,reading_cspm_file(CSPFile,PrologFile)),
   print(reading_cspm_file(CSPFile)),nl,
   get_cspm_parser_command(AbsParseCmd),
   print_message(informational,system_call(AbsParseCmd)),
   print(AbsParseCmd),nl,
   string_concatenate('--prologOut=', PrologFile, PrologFileOutArg),
   system_call(AbsParseCmd, ['translate',PrologFileOutArg, CSPFile], Text,JExit),
   print(exit(JExit)),nl,
   (JExit=exit(0)
    ->  load_cspm_pl_file(PrologFile,CSPFile)
    ;   add_error(parse_and_load_cspm_file,'Error while parsing CSP file: ',CSPFile),
	name(T,Text),
	add_error_fail(parse_and_load_cspm_file,'Std error: ',T)
   ).

/* Specific predicate for parsing and re-writting the Parse-Output into already existing prolog file.*/
% It will be needed by parsing a temporary files when the Csp Assertion Viewer is used.
parse_and_load_cspm_file_into_specific_pl_file(CSPFile, PrologFile) :- 
   get_cspm_parser_command(AbsParseCmd),
   string_concatenate('--prologOut=', PrologFile, PrologFileOutArg),
   system_call(AbsParseCmd, ['translate',PrologFileOutArg, CSPFile], Text,JExit),
   print(exit(JExit)),nl,
   (JExit=exit(0)
    ->  load_cspm_pl_file(PrologFile,CSPFile)
    ;   add_internal_error('Error while parsing CSP file: ',CSPFile),
        name(T,Text),
        add_error_fail(parse_and_load_cspm_file,'Std error: ',T)
   ).

/* Predicate for Translating specific Symbols and Operators in Csp into Unicode Symbols and vice versa.
   CSPM-Parser can handle Unicode-Symbols. */
% TranslArg must be only '--addUnicode=' or '--removeUnicode='
cspm_file_add_remove_unicode(CSPFile, CSPFile1, TranslArg) :-
   get_writable_compiled_filename(CSPFile,'.pl', PrologFile),
   print_message(informational,reading_cspm_file(CSPFile,_)),
   print(reading_cspm_file(CSPFile)),nl,
   get_cspm_parser_command(AbsParseUnicodeCmd),
   print_message(informational,system_call(AbsParseUnicodeCmd)),
   print(AbsParseUnicodeCmd),nl,
   string_concatenate(TranslArg,CSPFile1,TranslateArg),
   system_call(AbsParseUnicodeCmd, ['translate', TranslateArg, CSPFile], Text,JExit),
   print(exit(JExit)), nl,
   print(Text),
   (JExit=exit(0)
    ->  load_cspm_pl_file(PrologFile,CSPFile1)
    ;   add_error(parse_and_load_cspm_file,'Error while parsing CSP file: ',CSPFile),
	    name(T,Text),
	    add_error_fail(parse_and_load_cspm_file,'Std error: ',T)
   ).

parse_single_csp_declaration(Decl,CSPFile,Result) :-
   get_cspm_parser_command(Cmd),
   string_concatenate('--declarationToPrologTerm=', Decl, DeclarationArgument),
   system_call(Cmd, ['translate', DeclarationArgument, CSPFile],Text,ErrText,JExit),
   (JExit=exit(0)
       -> read_from_codes(Text,Result)
	; name(ErrT,ErrText),
          add_error_fail(parsing_csp_declaration,'Std error: ',ErrT)
   ).

parse_single_csp_expression(Context,Expr,Res) :- 
	( Context==eval -> Filename='no-file'
	; Context==ltl  -> 
              (csp_with_bz_mode
                 -> user: currently_opened_cspm_file_in_csp_and_b_mode(Filename)
                  ; currently_opened_file(Filename)
              )
	; otherwise -> add_internal_error('Internal Error: Unknown context argument: ', Context),fail
	),
	parse_single_csp_expression_file(Expr,Filename,Res).

:- use_module(library(codesio)).
parse_single_csp_expression_file(Expr,CSPFile,Res) :-
   get_cspm_parser_command(Cmd),
   string_concatenate('--expressionToPrologTerm=', Expr, ExpressionArgument),
   system_call(Cmd, ['translate', ExpressionArgument, CSPFile],Text,ErrText,JExit),
   (JExit=exit(0)
       -> read_from_codes(Text,Result),
          translate_expression(Result,Res)
	; name(ErrT,ErrText),
          add_error_fail(parsing_csp_expression,'Std error: ',ErrT)
   ).

evaluate_csp_expression(Expr,Result) :-
   ( parse_single_csp_expression(eval,Expr,ParsedExpr) ->
          evaluate_parsed_csp_expression(ParsedExpr,Expr,Result)
   ; add_error_fail(parsing_csp_expression, 'Error while parsing expression: ', Expr)
   ).

evaluate_parsed_csp_expression(ParsedExpr,Expr,Result) :-
   debug_println(9,parsed_csp_expression(ParsedExpr)),
   haskell_csp_analyzer:compile_body(ParsedExpr,'evaluate',[],[],CompiledExpr),
   ( evaluate_expression_with_timing(CompiledExpr,Res) ->
      ( translate_cspm_expression(Res, PrettyPrintedResult) -> 
         ajoin(PrettyPrintedResult,Result)
      ; add_error_fail(pretty_printing_evaluated_expression, 'Error while pretty-printing result: ', Res)
      )
   ; add_error_fail(evaluating_expression, 'Error while evaluating expression: ',Expr)
   ).

evaluate_csp_expression(Expr,CSPFile,Result) :-
   ( parse_single_csp_expression_file(Expr, CSPFile,ParsedExpr) ->
          evaluate_parsed_csp_expression(ParsedExpr,Expr,Result)
   ; add_error_fail(parsing_csp_expression, 'Error while parsing typed expression: ', Expr)
   ).

:- use_module(probsrc(tools),[cputime/1]).
evaluate_expression_with_timing(ParsedExpr,Res) :-
   tools: cputime(T1),
   evaluate_expression(ParsedExpr,Res),
   tools: cputime(T2),
   D is T2-T1,
   print('% Time to evaluate expression: '),print(D), print(' ms.%'),nl.

% TODO: Which are the expressions allowed to be written in events
translate_expression(listExp(rangeEnum(List)),list(List)).
translate_expression(setExp(rangeEnum(List)),setValue(List)).
translate_expression(dotTuple(List),dotTuple(Res)) :-
   maplist(translate_expression,List,Res).
translate_expression(X,X).

:- meta_predicate read_compiled_prolog_file(-,-,1).
/* Predicates for reading a Prolog file (usually generated after translating the appropriate Csp file with the CSPM-Tool)
   and filtering the asssertion declarations from it. Needed for the 'Check Csp Assertions' window in Tcl_Tk. */
read_compiled_prolog_file(CSPFile,Result,PredName) :-
   get_writable_compiled_filename(CSPFile,'.pl', PrologFile),
   debug_print(9,'opening Prolog file: '), debug_println(9,PrologFile),
   my_see(PrologFile),
   call(PredName,Result),
   !,seen,debug_println(9,done).

my_see(File) :-
   on_exception(error(existence_error(_,_),_),
   see(File),
   add_error_fail(my_see,'File does not exist: ',File)).

filter_assertion_declarations(Assertions) :- 
   read(Term), !,
   ((Term = end_of_file) -> (Assertions = [])
    ; (is_assertion(Term)
       -> (Assertions = [Term|T])
         ; (Assertions = T)
          ), filter_assertion_declarations(T)
   ).

:- assert_must_succeed((is_assertion(assertRef(_,_,_,_,_)),is_assertion(assertModelCheck(_,_,_)), is_assertion(assertModelCheckExt(_,_,_,_)))).

is_assertion(assertRef(_,_,_,_,_)).
is_assertion(assertModelCheck(_,_,_)).
is_assertion(assertModelCheckExt(_,_,_,_)).
is_assertion(assertLtl(_,_,_,_)).
is_assertion(assertCtl(_,_,_,_)).

filter_formulas_from_pragmas(Type,FlNList,FlFList) :-
   findall(Formulas,(pragma(S),get_formulas_from_pragma(Type,pragma(S),_Names,Formulas)),FList),
   findall(Names,(pragma(S),get_formulas_from_pragma(Type,pragma(S),Names,_Formulas)),NList),
   flatten(NList,FlNList),
   flatten(FList,FlFList).

get_formula_from_cspm_file(Type,Formula) :-
   filter_formulas_from_pragmas(Type,_Ns,[String|_]),
   (var(String) ->
          Formula = ''
   ; Formula=String
   ).

get_formulas_from_cspm_file(Type,String) :-
   filter_formulas_from_pragmas(Type,_Ns,Formulas),
   convert_string_list_to_string(Formulas,String).

string_concatenate_sep(Sep,Atom,R) :-
   string_concatenate(Atom,Sep,R).

convert_string_list_to_string(L,Res) :-
   maplist(string_concatenate_sep(';'),L,LRes),
   concatenate_string_list(LRes,Res).

get_formulas_from_pragma(Type,pragma(Str),Names,Formulas) :-
   atom_codes(Str,CodeList),
   (Type = ltl ->
       ltl_equations(Names,Formulas,CodeList,[])
   ;Type = ctl ->
       ctl_equations(Names,Formulas,CodeList,[])
   ; otherwise ->
       Names=[], Formulas=[]
   ).

%%%%%%%%%%%%%%%%%% Tests for the ASSERT_LTL Equations Parser below %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed((haskell_csp:reset_for_selfcheck)).
:- assert_must_succeed((haskell_csp: get_formulas_from_pragma(ltl,pragma('ASSERT_LTL_123=="F(e(a))" "Checking simple LTL property."'),
        Names,Formulas),Names==[123],Formulas==['F(e(a))'])).
:- assert_must_fail((haskell_csp: get_formulas_from_pragma(ltl,pragma('ASSERT_LTL_0123=="F(e(a))" "First digit in number must be nonzero."'),
        _Names,_Formulas))).
:- assert_must_succeed((haskell_csp: get_formulas_from_pragma(ltl,pragma('ASSERT_LTL=="F(e(a))"'),
        Names,Formulas),Names==[[95]],Formulas==['F(e(a))'])).
:- assert_must_succeed((haskell_csp: get_formulas_from_pragma(ltl,pragma('assert_ltl "F(e(a))" "Checking simple LTL property."'),
        Names,Formulas),Names==[[95]],Formulas==['F(e(a))'])).
:- assert_must_succeed((haskell_csp: get_formulas_from_pragma(ctl,pragma('assert_ctl "true" "Checking simple CTL property."'),
        Names,Formulas),Names==[[95]],Formulas==['true'])).
:- assert_must_succeed((haskell_csp: get_formulas_from_pragma(ltl,pragma('assert_ltl "F(e(a))" "Parsing multiple LTL equations."; ASSERT_LTL_TRUE "F(e(b))"; ASSERT_LTL_2=="F(e(c))"; ASSERT_LTL "F(e(d))"'),
        Names,Formulas),Names==[[95],[84,82,85,69],2,[95]],Formulas==['F(e(a))','F(e(b))','F(e(c))','F(e(d))'])).
:- assert_must_succeed((
             haskell_csp: get_formulas_from_cspm_file(ltl,S), S == 'GF(a);GF(b);true;')).
:- assert_must_succeed((
             haskell_csp: get_formula_from_cspm_file(ltl,S), S == 'GF(a)')).

%%%%%%%%%%%%%%%%%% Simple Parser for Parsing ASSERT_LTL Equations %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%% CTL assertions %%%%%%%%%%%%%%%
ctl_equations([N|Ns],[F|Fs]) --> ctl_equation(N,F),ctl_equations2(Ns,Fs).
ctl_equations2(Ns,Fs) --> ows,sep,ows,ctl_equations(Ns,Fs),{Fs\=[],Ns\=[]}.
ctl_equations2([],[]) --> ows.

ctl_equation(N,F) --> ows,assert_id_ctl(N), ows, eq,!, ows, string(F), ows, optional_string.

assert_id_ctl(N) --> ass_ctl,!,name(N).
assert_id_ctl(N) --> [97,115,115,101,114,116,95,99,116,108],{N="_"},!.
assert_id_ctl(N) --> "assert_ctl",{N="_"},!.
assert_id_ctl(N) --> [65,83,83,69,82,84,95,67,84,76],{N="_"},!.

ass_ctl --> [65,83,83,69,82,84,95,67,84,76,95],!.
ass_ctl --> "ASSERT_CTL_",!.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ltl_equations([N|Ns],[F|Fs]) --> ltl_equation(N,F),ltl_equations2(Ns,Fs).
ltl_equations2(Ns,Fs) --> ows,sep,ows,ltl_equations(Ns,Fs),{Fs\=[],Ns\=[]}.
ltl_equations2([],[]) --> ows.

ltl_equation(N,F) --> ows,assert_id_ltl(N), ows, eq,!, ows, string(F), ows, optional_string.

string(Atom) --> [34], stringcontent(String), [34], {name(Atom,String)}.

optional_string --> [34], stringcontent(_String), [34],!.
optional_string --> "",!.

stringcontent([C|Rest]) --> [C], {C \= 34}, !, stringcontent(Rest).
stringcontent([]) --> [].

assert_id_ltl(N) --> ass,!,name(N).
assert_id_ltl(N) --> [97,115,115,101,114,116,95,108,116,108],{N="_"},!.
assert_id_ltl(N) --> "assert_ltl",{N="_"},!.
assert_id_ltl(N) --> [65,83,83,69,82,84,95,76,84,76],{N="_"},!.
%assert_id(N) --> "ASSERT_LTL",{N="_"},!.

name(N) --> num(N),{number(N)},!.
name(N) --> ident(N),!.

ident([C|Cs]) --> [C],{member(C,"0123456789"); ("A" =< C,C =< "Z");("a" =< C,C =< "z");C="_"},ident(Cs),!.
ident([]) --> [],!.

num(Number) --> nonzerodigit(D), num2(Rest),{name(Number,[D|Rest])}.
num(Number) --> digit(D),{name(Number,[D])}.
num2([D|Rest]) --> digit(D),!,num2(Rest).
num2([]) --> [].

digit(D) --> [D], {member(D,"0123456789")},!.
nonzerodigit(ND) --> [ND], {member(ND,"123456789")},!.

sep --> [C],{member(C,";,")}.

ows --> [C], {member(C," \t\n\r")}, !, ows.
ows --> "".

eq --> [61,61],!.
eq --> "==",!.
eq --> [32],!.
eq --> "",!.

ass --> [65,83,83,69,82,84,95,76,84,76,95],!.
ass --> "ASSERT_LTL_",!.

/* ------------------------------------------------------------------------------------------------------------- */

/* Predicates for getting the CSPM-Tool command from the library and loading cspm Prolog files.*/ 
get_cspm_parser_command(AbsParseCmd) :-
   debug_println(9,getting_parser_cmd),
   on_exception(error(E,_),
     absolute_file_name(prob_lib(cspmf),AbsParseCmd,[access(exist),extensions(['.exe',''])]),
     add_error_fail(haskell_csp,'Could not find CSP-M Parser. Be sure to install cspmf in your ProB lib directory. ',E)),
   debug_println(9,parser(AbsParseCmd)).

get_cspm_tool_command(Cmd) :-
   get_preference(path_to_cspm,Cmd).
   
:- dynamic parseResult/5.
load_cspm_pl_file(PFile) :- load_cspm_pl_file(PFile,PFile).
load_cspm_pl_file(PrologFile,CSPFile) :- clear_cspm_spec,
   \+ file_exists(PrologFile),!,
   add_error(haskell_csp,'Could not generate Prolog encoding of CSP file: ',CSPFile),fail. 
load_cspm_pl_file(PrologFile,CSPFile) :-
   print(consulting(PrologFile)),nl,
   cia_consult_without_redefine_warning(PrologFile),
   /* TO DO: replace by safer load which checks that
    only correct facts are loaded !!!! */
   parseResult(ParseResult,ErrorMsg,StartLine,SC,Off),
   (ParseResult='ok' -> analyze ;  
     add_error(haskell_csp,'Error while parsing CSP-M: ',ParseResult:CSPFile),
     add_error(haskell_csp,'Error message: ',ErrorMsg,src_position(StartLine,SC,Off,1))
   ).

/* ------------------------------------------------------------------------------------------------------------ */

% A small pretty-printer for printing assert declarations
refinement_operator('Trace',' [T= ') :- !.
refinement_operator('Failure', ' [F= ') :- !.
refinement_operator('FailureDivergence',' [FD= ') :- !.
refinement_operator('RefusalTesting',' [R= ') :- !.
refinement_operator('RefusalTestingDiv',' [RD= ') :- !.
refinement_operator('RevivalTesting',' [V= ') :- !.
refinement_operator('RevivalTestingDiv',' [VD= ') :- !.
/* the clauses below will be called if we check the particular assertion direct from the command line and not from the .pl file */
refinement_operator('[T=',' [T= ') :- !.
refinement_operator('[F=', ' [F= ') :- !.
refinement_operator('[FD=',' [FD= ') :- !.
refinement_operator('[R=',' [R= ') :- !.
refinement_operator('[RD=',' [RD= ') :- !.
refinement_operator('[V=',' [V= ') :- !.
refinement_operator('[VD=',' [VD= ') :- !.
refinement_operator(A, _) :- add_error_fail(haskell_csp, 'Error while pretty-printing assert declaration. Invalid or unsupported refinement operator: ', A).

adding_not('True', 'not ') :- !.
adding_not('False', '') :- !.
adding_not(A,_) :- add_error_fail(haskell_csp,'Error while pretty-printing assert declaration. Unexpected argument infront the assertion declaration: ',A).

fdr_model('Deterministic', ' :[ deterministic ') :- !.
fdr_model('DeadlockFree', ' :[ deadlock free ') :- !.
fdr_model('LivelockFree', ' :[ livelock free ') :- !.
fdr_model(A,_) :- add_error_fail(haskell_csp, 'Error while pretty-printing assert declaration. Invalid or unsupported model inside of process assertion: ', A).

fdr_single_model('Deterministic', ' :[ deterministic [FD] ]') :- !.
fdr_single_model('DeadlockFree', ' :[ deadlock free [FD] ]') :- !.
fdr_single_model('LivelockFree', ' :[ livelock free ]') :- !.
fdr_single_model(A,_) :- add_error_fail(haskell_csp, 'Error while pretty-printing assert declaration. Invalid or unsupported model inside of process assertion: ', A).

fdr_ext('F', '[F] ]') :- !.
fdr_ext('FD', '[FD] ]') :- !.
fdr_ext(Ext, _) :- add_error_fail(haskell_csp, 'Not supported assertion type model: ',Ext).

:- assert_must_fail(proc_with_arguments_to_string('P',0,_)).
:- assert_must_fail(proc_with_arguments_to_string('P',-10,_)).
:- assert_must_succeed((proc_with_arguments_to_string('P',1,String), String=='P(-)')).
:- assert_must_succeed((proc_with_arguments_to_string('P',3,String), String=='P(-,-,-)')).

proc_with_arguments_to_string(Fun,N,FunRes) :-
   ( N>1 ->
     count_proc_arguments(N,R),
     concatenate_string_list([Fun,'('|R],FunRes)
   ; N =:= 1 ->
     string_concatenate(Fun,'(-)',FunRes)
   ).

count_proc_arguments(N,Res) :-
   count_proc_arguments(N,['-',')'],Res).

count_proc_arguments(N,L,Res) :-
   (N =:= 1 ->
       Res=L
   ; N>1 ->
       N1 is N-1,
       count_proc_arguments(N1,['-',','|L],Res)
   ).

:- assert_must_succeed((string_to_proc_term('P(-,-,-)',Term), Term = 'P'(_A,_B,_C))).
:- assert_must_succeed((string_to_proc_term('P(-)',Term), Term = 'P'(_X))).
:- assert_must_succeed((string_to_proc_term('P',Term), Term = 'P')).
:- assert_must_succeed((string_to_proc_term(p(1,2),Term), Term == p(1,2))).

string_to_proc_term(Proc,Term) :-
   atom(Proc),!,
   split_atom(Proc,['(',',',')'],[Fun|A]),
   length(A,N),
   gen_vars(N,Args),
   Term =.. [Fun|Args].
string_to_proc_term(X,X).

:- assert_must_succeed((string_to_proc_term('P(-,-,-)',Term), functor(Term,F,N), proc_with_arguments_to_string(F,N,Str), Str == 'P(-,-,-)')).

gen_vars(N,Vars) :- 
	( N>0 -> N1 is N-1, gen_vars(N1,T),Vars=[_X|T]
	; N =:= 0 -> Vars = []).

:- assert_must_succeed(string_term_with_args('Fun(1,2,-)')).
:- assert_must_fail(string_term_with_args('Fun(1,2,3)')).
:- assert_must_fail(string_term_with_args('Fun(X,_,_Y)')).
:- assert_must_fail(string_term_with_args(_Fun)).

string_term_with_args(StrTerm) :-
   atom(StrTerm),!,
   split_atom(StrTerm,['(',',',')'],[_Fun|A]),
   A\=[],member(-,A).

:- assert_must_succeed((haskell_csp:concatenate_string_list(['A','B','C'],'ABC'))).
:- assert_must_succeed((haskell_csp:concatenate_string_list([],''))).

concatenate_string_list(L,Res) :-
   concatenate_string_list(L,'',Res).

concatenate_string_list([],Res,Res).
concatenate_string_list([H|T],Str,Res) :-
   string_concatenate(Str,H,StrRes),
   concatenate_string_list(T,StrRes,Res).

translate_csp_assertions(L,Res) :-
   maplist(translate_csp_assertion_with_sep,L,LRes),
   concatenate_string_list(LRes,Res).

translate_csp_assertion_with_sep(Ass,AssStrWithSep) :-
   translate_with_sep(translate_csp_assertion,Ass,'$',AssStrWithSep).

% Sep must be an atom
translate_with_sep(TranslationPredicate,Arg,Sep,ArgStrSep) :-
   functor(Fun,TranslationPredicate,2),arg(1,Fun,Arg),arg(2,Fun,ArgStr),
   %Fun =.. [TranslationPredicate,Arg,ArgStr],
   call(Fun),
   string_concatenate(ArgStr,Sep,ArgStrSep).

translate_csp_assertion(AssertionExpr,ConcRes) :-
   /* assert not? P [Op= Q */
   (AssertionExpr = assertRef(N,P,Op,Q,_) ->
       refinement_operator(Op,Operator),
       translate_cspm_state(Q,Proc_2),
       L=[Operator,Proc_2]
   /* assert P |= LTL: "ltl-formula" */
   ;AssertionExpr = assertLtl(N,P,Formula,_) ->
       L = [' |= ', 'LTL: ', '\"', Formula, '\"']
   /* assert P |= CTL: "ctl-formula" */
   ;AssertionExpr = assertCtl(N,P,Formula,_) ->
       L = [' |= ', 'CTL: ', '\"', Formula, '\"']
   /* assert not? P :[Mod [E]] */
   ;AssertionExpr = assertModelCheckExt(N,P,Mod,E) ->
       fdr_model(Mod,Model),
       fdr_ext(E,Ext),
       L=[Model,Ext]
   /* assert not? P :[Mod] */
   ;AssertionExpr = assertModelCheck(N,P,Mod) ->
       fdr_single_model(Mod,Model),
       L=[Model]
   ; otherwise ->
       add_internal_error('Internal Error (CSP-M Parser): Unexpected assertion predicate: ',
                           translate_csp_assertion(AssertionExpr,ConcRes))
   ),
   adding_not(N,Not),
   translate_cspm_state(P,Proc),
   concatenate_string_list([Not, Proc|L], ConcRes).

%%%%%%%%%%%%%%%%%%%%%%%%%%% Testing get_csp_processes/1, get_csp_process/2 and translate_csp_assertions/2 predicates %%%%%%%%%%%%%%%%%%%%%%%%%%%%
find_and_assert_all_csp_processes_from_arity_zero :-
	findall(P,(bindval(P,Body,_),definite_cspm_process_expression(Body),assert(is_csp_process(P))),_L).

clear_all_is_csp_process_facts :- retractall(is_csp_process(_)).

:- assert_must_succeed((find_and_assert_all_csp_processes_from_arity_zero,get_csp_processes(Res),clear_all_is_csp_process_facts,
	Res=='P(-)$Q(-)$SEND$REC$SYSTEM$GEN1$GEN2$MAIN$')).
:- assert_must_succeed((findall(assertRef(A,B,C,D,S),haskell_csp: assertRef(A,B,C,D,S),L),haskell_csp: translate_csp_assertions(L,String),
           String == 'CSP: MAIN [T= CSP: PROB_TEST_TRACE$not CSP: PROB_TEST_TRACE [T= CSP: MAIN$')).
:- assert_must_succeed((findall(assertModelCheckExt(A,B,C,D),haskell_csp: assertModelCheckExt(A,B,C,D),L),
 findall(assertModelCheck(A,B,C),haskell_csp:assertModelCheck(A,B,C),L1,L),haskell_csp: translate_csp_assertions(L1,String),
 String == 'CSP: a.1->P :[ livelock free ]$CSP: SYSTEM [|{|left,right|}|] GEN1 :[ deterministic [F] ]$not CSP: SYSTEM [|{|left,right|}|] GEN1 :[ deadlock free [FD] ]$')).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* Tcl_Tk cannot handle complicated prolog strings, that's why we have to convert the from the Csp prolog file filtered assertions
   into simple string. Assertions will be separated with the '$' symbol. */
get_csp_assertions_as_string(CSPFile,R) :- 
	read_compiled_prolog_file(CSPFile,A,filter_assertion_declarations),
	translate_csp_assertions(A,R),debug_print(9,'string_for_tcl: '),debug_println(9,R).

/* The same as above for the processes, these will be separate by '$' as well. Needed for the process comboboxes in the 
   'Check Csp Assertions' window in Tcl_Tk. */
get_csp_processes(Res) :- 
   nl,print(getting_csp_processes),nl,
   findall(bindval(P,_,_),bindval(P,_,_),L),
   findall(agent(F,Body,_),agent(F,Body,_),L1,L),
   get_processes(L1,R1),
   concatenate_string_list(R1,Res1),
   string_concatenate(Res1,'',Res),
   print(Res),nl,nl.   

get_processes([],[]).
get_processes([bindval(A,_,_)|T],R) :- !,
   ( animatable_process(A) -> R=[A,'$'|Res] ; R=Res),
   get_processes(T,Res).
get_processes([agent(Fun,Body,_)|T],R) :- !,
   ( (definite_cspm_process_expression(Body),animatable_process_with_arguments(Fun,proc(F,N)))
	-> proc_with_arguments_to_string(F,N,FunRes),R = [FunRes, '$'|Res]
   ;       R=Res
   ),
   get_processes(T,Res).

/* ------------------------------------------------------------------------------------------------------------------------- */

clear_cspm_spec :-  % nl,print('RESET CSPM'),nl,nl,
   %reset_profiler,
   retractall(parseResult(_,_,_,_,_)),
   retractall(channel(_,_)),
   retractall(bindval(_,_,_)),
   retractall(cspTransparent(_)),
   retractall(agent(_,_,_)),
   retractall(dataTypeDef(_,_)),
   retractall(subTypeDef(_,_)),
   retractall(nameType(_,_)),
   retractall(assertRef(_,_,_,_,_)),
   % retractall(assertTauTrio(_,_,_,_,_,_)),
   retractall(assertModelCheckExt(_,_,_,_)),
   retractall(assertModelCheck(_,_,_)),
   retractall(channel_type_list(_,_)),
   retractall(csp_full_type_constructor(_,_,_)),
   retractall(pragma(_)).
   %retractall(csp_full_expanded_type(_,_)),
   %retractall(expanded_setFromTo(_,_,_)).

:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- register_event_listener(clear_specification,clear_cspm_spec,
                           'Clearing any CSPM Specification.').

is_a_datatype(DT,L) :-
   (dataTypeDef(DT,L) ; subTypeDef(DT,L)).

/* Tests for the is_not_infinite_type/1 predicate (see .pl translation sample on lines 68-106 ) */

:- assert_must_fail((assert(csp_full_type_constructor(_,'SubMsg',[dataType('Msg')])),
                     assert(csp_full_type_constructor(_,'Msg',[dataType('SubMsg')])),
                     haskell_csp:is_not_infinite_type(closure([tuple([receive])])))).
:- assert_must_succeed((assert(csp_full_type_constructor(_,'FRUIT',[constructor(apples),construcor(oranges),constructor(pears)])),
                       haskell_csp:is_not_infinite_type(closure([tuple([left])])))).
:- assert_must_fail((assert(csp_full_type_constructor(_,'FRUIT',[constructor(apples),construcor(oranges),constructor(pears)])),
                    assert(csp_full_type_constructor(_,'SubMsg',[dataType('Msg')])),
                    assert(csp_full_type_constructor(_,'Msg',[dataType('SubMsg')])),
                    haskell_csp:is_not_infinite_type(closure([tuple([left]),tuple([mid]),tuple([right]),tuple([receive])])))).
:- assert_must_fail((assert(csp_full_type_constructor(ste,'Inf',['Seq'(dataType('FRUIT'))])),
                    assert(csp_full_type_constructor(_,'FRUIT',[constructor(apples),construcor(oranges),constructor(pears)])),
                    haskell_csp:is_not_infinite_type(closure([tuple([infinite])])))).
:- assert_must_succeed((assert(csp_full_type_constructor(_,'FRUIT',[constructor(apples),construcor(oranges),constructor(pears)])),
                       haskell_csp:is_not_infinite_type(closure([tuple([left]),tuple([mid]),tuple([right])])))).
:- assert_must_fail((assert(csp_full_type_constructor(_,'FRUIT',[constructor(apples),construcor(oranges),constructor(pears)])),
                    assert(csp_full_type_constructor(_,'SubMsg',[dataType('Msg')])),
                    assert(csp_full_type_constructor(_,'Msg',[dataType('SubMsg')])),
                    assert(csp_full_type_constructor(ste,'Inf',['Seq'(dataType('FRUIT'))])),
                    haskell_csp:is_not_infinite_type(closure([tuple([left]),tuple([mid]),tuple([right]),tuple([infinite]),tuple([receive])])))).

:- assert_must_fail(haskell_csp:is_not_infinite_type(setFrom(1))).
:- assert_must_fail(haskell_csp:check_channel_input_value('dotTupleType'([setValue([int(1),int(2)]), setFrom(1)]),2)).
:- assert_must_succeed((assert(name_type('S', setFromTo(1,3))), haskell_csp: check_channel_input_value('S',2))).

/* is_not_infinite_type/1 */
/* check if the Event set is from infinite type */
/* needed for the refinement checker, when generating counter-examples, TODO: re-implement predicate */
is_not_infinite_type(X) :- var(X),!,fail.
% at first we are only interesting of checking the whole closure for infinite types
is_not_infinite_type(closure(X)) :- !,
  closure_has_not_infinite_set(X).
%% TODO: Cover all other cases for this predicate (like setValue, setFrom, ...)


closure_has_not_infinite_set(L) :-
   maplist(single_element_closure_has_not_infinite_set,L).

% the closure is basically the set of all channels in a specification
single_element_closure_has_not_infinite_set(tuple([Ch|_List])) :- 
    channel_type_list(Ch,ChannelTypeList),!,
    check_list_for_infinite_set(ChannelTypeList).
single_element_closure_has_not_infinite_set(_X).

check_list_for_infinite_set([]).
check_list_for_infinite_set([Type|L]) :-
    check_channel_input_value(Type,2),!,
    check_list_for_infinite_set(L).

check_channel_input_value(intType,_Rec) :- !,fail.
check_channel_input_value(boolType,_Rec) :- !.
check_channel_input_value(setFromTo(_L,_U),_Rec) :- !.
check_channel_input_value(setFrom(_L),_Rec) :- !,fail. % endless set
% we got a dot tuple of types, we need to check for infinity all types in the tuple.
check_channel_input_value('dotTupleType'(List),Rec):-!,
   l_check_channel_input_value(List,Rec).
check_channel_input_value('typeTuple'(List),Rec) :- !,
   l_check_channel_input_value(List,Rec). % the same as last clausel
check_channel_input_value(setValue(_L),_Rec) :- !.
check_channel_input_value(dataType(DT),Rec) :-
   is_a_datatype(DT,_),!,
   check_datatype_el(DT,Rec).
check_channel_input_value('Seq'(setValue([])),_Rec) :- !. % Seq({}) is the empty sequence <>
check_channel_input_value('Seq'(_Type),_Rec) :- !, fail. % e.g. Seq({a,b,c}) indicator for endless datatype (there are endless numbers of sequences)
check_channel_input_value(X,Rec) :-
   name_type(X,Type),!,
   check_channel_input_value(Type,Rec). 
check_channel_input_value(_X,_Rec).

check_datatype_el(DT,MaxRec) :- 
    (MaxRec > 0 -> M1 is MaxRec -1,
         (csp_constructor(_Cons,DT,ArgSubTypes) -> /* Subtypes recursion. */l_check_channel_input_value(ArgSubTypes,M1)
                                   ; /* constructors from finite type (there is no recursion)*/ true)).
        /* ending recursion of datatypes (recursion doesn't mean infinite type) */

check_channel_input_value_rec(Rec,HType) :-
   check_channel_input_value(HType,Rec).
l_check_channel_input_value(L,Rec) :-
   maplist(check_channel_input_value_rec(Rec),L).

/*
l_check_channel_input_value([],_Rec).
l_check_channel_input_value([HType|T],Rec) :-
   (check_channel_input_value(HType,Rec) ->
       l_check_channel_input_value(T,Rec)).
*/

/* end of the implementation of is_not_infinite_type/1 predicate */

get_csp_assertions(CSP_Assertions) :-
   findall(Ass, (is_assertion(Ass),Ass), CSP_Assertions),
   debug_println(9,CSP_Assertions).

/*
get_csp_assertions_enum(CSPFile, Result) :- 
   read_compiled_prolog_file(CSPFile,CSP_Assertions,filter_assertion_declarations),
   enumerate_assertions(CSP_Assertions,Result).

:- assert_must_succeed(haskell_csp:enumerate_assertions([a,b,c,d,e,f],[(a,0),(b,1),(c,2),(d,3),(e,4),(f,5)])).
:- assert_must_succeed(haskell_csp:enumerate_assertions([],[])).
enumerate_assertions(L,R) :- enumerate_assertions_acc(L,R,0).
enumerate_assertions_acc([],[],_).
enumerate_assertions_acc([H|T],[(H,Acc)|R],Acc) :- New_Acc is Acc +1, enumerate_assertions_acc(T,R,New_Acc).
*/

cia_consult_without_redefine_warning(File) :-
    (file_exists(File) -> true ; (add_error(haskell_csp,'CSP-M PL File does not exist: ',File),fail)),
    print(consult_without_redefine(File)),nl,
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off), 
    prolog_flag(discontiguous_warnings, Old3, off), 
    (load_files(File,[compilation_mode(assert_all)]) %consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    prolog_flag(discontiguous_warnings, _, Old3),
    OK=true.

/* ------------------------------------- */


/* THE CSP OPERATIONAL SEMANTICS */


/* ------------------------------------- */
%
%trans_pre_compute(X,Results) :-
%   findall(res(X,A,X1,Residue),call_residue(cspm_trans(X,A,X1),Residue), Results).
%   
%
%trans_cont(Results,X,A,X1) :-
%   member(res(X,A,X1,Residue),Results),
%   % print(trans_cont(X,A,X1,Residue)),nl,
%   call(Residue).
   
/* ------------------------------------- */

%cspm_trans(E,_,_) :- print(cspm_trans(E)),nl,fail.

:- assert_must_succeed(( haskell_csp:functor_dif(X,f), X=g(_) )).
:- assert_must_fail(( haskell_csp:functor_dif(X,f), X=f(_) )).

:- block functor_dif(-,?).
functor_dif(X,FY) :- functor(X,FX,_), FX\=FY.  

:- assert_pre(haskell_csp:cspm_trans(E,_,_,_), nonvar(E)).
:- assert_post(haskell_csp:cspm_trans(_,A,NE,_), (nonvar(A),nonvar(NE))).
:- block cspm_trans(-,?,?,?).
% TO DO: comment out for efficiency:
/* ------ */
%cspm_trans(X,_,_) :-  debug_println(9,cspm_trans(X)), var(X),
%  add_error(csp_interpreter,'### Variable process expression in cspm_trans !!',X),  fail.
cspm_trans(X,_,_,_) :-  undefined_process_construct(X),
  (get_preference(cspm_animate_all_processes,true),string_term_with_args(X) ->
       add_message(cspm_trans,'Warning: Tried to expand a process with non-ground arguments: ',X),fail
  ;add_error_fail(csp_interpreter,'### Undefined Construct in cspm_trans !! ',X)
  ).
/* ------  */

cspm_trans(stop(_),_,_,_) :- fail.
cspm_trans(omega,_,_,_) :- fail.

cspm_trans(skip(SrcSpan),tick(SrcSpan),omega,_).
%cspm_trans(skip(SrcSpan),tick(SrcSpan),stop(SrcSpan),_).

%cspm_trans(cspPrint(X),io(V1,print,no_loc_info_available),cspPrint(X)) :- evaluate_argument(X,V1).

cspm_trans('CHAOS'(SrcSpan,_Set),tau(chaos_stop(SrcSpan)),stop(SrcSpan),_).
cspm_trans('CHAOS'(SrcSpan,Set),io(V1,Ch,SrcSpan),'CHAOS'(SrcSpan,Set),_WF) :-   /* or should we add an intermediate tau ?? */
  % print(evaluate_argument(Set,S)),nl,
   evaluate_argument(Set,S), 
  % print(chaos(S)),nl,
   ~~mnf(haskell_csp:expand_channel_pattern_expression(S,ECList,SrcSpan)),
 %  print(setup_channel_skeleton(io(V1,Ch,SrcSpan))),nl,
   setup_channel_skeleton(io(V1,Ch,SrcSpan)),
 %  print(chaos(io(V1,Ch),ECList)),nl,
   hidden(io(V1,Ch,SrcSpan),ECList).
   
%   print(hidden(io(V1,Ch,SrcSpan),ECList)),nl.
   
/* external choice */
cspm_trans('[]'(X,Y,Span),NA,Res,WF) :- cspm_trans(X,A,X1,WF),
    shift_span_for_left_branch(Span,LSpan),
    merge_span_into_event(A,LSpan,NA),
    (  top_level_dif(A,tau(_)), Res=X1
     ;
       A=tau(_), tl_normalise('[]'(X1,Y,Span),Res) ).
cspm_trans('[]'(X,Y,Span),NA,Res,WF) :- cspm_trans(Y,A,Y1,WF),
    shift_span_for_right_branch(Span,RSpan),
    merge_span_into_event(A,RSpan,NA),
    ( top_level_dif(A,tau(_)), Res=Y1
     ;
      A=tau(_), tl_normalise('[]'(X,Y1,Span),Res) ).

/* internal choice */
cspm_trans('|~|'(X,_Y,SrcSpan),tau(int_choice_left(SrcSpan,LSpan)),X,_WF) :-
   shift_span_for_left_branch(SrcSpan,LSpan).
cspm_trans('|~|'(_X,Y,SrcSpan),tau(int_choice_right(SrcSpan,RSpan)),Y,_WF) :-
   shift_span_for_right_branch(SrcSpan,RSpan).

/*  ->  PREFIX */
/* Note: set restrictions on values are encoded as inGuard Expressions */
cspm_trans(prefix(SPAN1,Values,ChannelExpr,CSP,SPAN2), io(EV,Channel,SPAN), NormCSP,WF) :- 
   evaluate_channel_outputs(Values,ChannelExpr,EV,Channel,SPAN,WF),
   unify_spans(SPAN1,SPAN2,SPAN),
   full_normalise_csp_process(CSP,NormCSP). % ,print(prefix(io(EV,Channel))),nl.


/* interleave */
%cspm_trans('|||'(X,Y,Span),A,R,WF) :- cspm_trans(sharing(setValue([]),X,Y),A,R).
/* Firing rules for the |||-operator:

         P -a> P'
----------------------------
  P ||| Q -a> P' ||| Q

         Q -a> Q'
----------------------------
  P ||| Q -a> P ||| Q'


         P -tick> P'
----------------------------
  P ||| Q -tau> Omega ||| Q


         Q -tick> Q'
----------------------------
  P ||| Q -tau> P ||| Omega

*/

cspm_trans('|||'(omega,omega,Span),tick(Span),omega,_).
cspm_trans('|||'(X,Y,ISpan),A,Res,WF) :- 
     cspm_trans(X, ActionX, X1,WF),
     shift_span_for_left_branch(ISpan,LSpan),
      (  ActionX=tick(_), X2=omega
       ; functor_dif(ActionX,tick), X2=X1
      ),
      merge_span_into_event(ActionX,LSpan,MA),
	  (ActionX=tick(_) -> A=tau(MA); A=MA),
      tl_normalise('|||'(X2,Y,ISpan),Res).
cspm_trans('|||'(X,Y,ISpan),A,Res,WF) :- 
     cspm_trans(Y, ActionY, Y1,WF),
     shift_span_for_right_branch(ISpan,RSpan),
      (  ActionY=tick(_), Y2=omega
       ; functor_dif(ActionY,tick), Y2=Y1
      ),
      merge_span_into_event(ActionY,RSpan,MA), 
	  (ActionY=tick(_) -> A=tau(MA); A=MA),
      tl_normalise('|||'(X,Y2,ISpan),Res).

/* sequential composition */
cspm_trans(';'(P,Q,SeqSpan),AX,Res,WF) :- 
  cspm_trans(P,A,P1,WF),
  (  A=tick(_),
     merge_span_into_event(A,SeqSpan,MA),
     AX = tau(MA),
     full_normalise_csp_process(Q,Res)  /* is this required ?? */
   ;  functor_dif(A,tick), AX=A, 
      tl_normalise(';'(P1,Q,SeqSpan),Res)
  ).

  
/* Sharing:  [| a |] */
cspm_trans(sharing(CList,X,Y,SrcSpan),A,Res,WF) :-
   evaluate_argument(CList,EvCList),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCList,ECList,SrcSpan)),
   cspm_trans(esharing(ECList,X,Y,SrcSpan),A,Res,WF).
cspm_trans(esharing(_,omega,omega,SrcSpan), tick(SrcSpan), omega ,_WF).
cspm_trans(esharing(CList,X,Y,SrcSpan), Action, Result,WF) :- 
     cspm_trans(X, ActionX, X1,WF),
     (  (ActionX=tick(TS),
         shift_span_for_left_branch(SrcSpan,LSpan),
         merge_span_into_event(tau(tick(TS)),LSpan,Action),
         Result = esharing(CList,omega,Y,SrcSpan))
      ; (functor_dif(ActionX,tick),
         shift_span_for_left_branch(SrcSpan,LSpan),
         merge_span_into_event(ActionX,LSpan,Action),
         not_hidden(ActionX,CList), /* covers tau */
         Result = esharing(CList,X1,Y,SrcSpan))
      ; (ActionX=io(V1,Ch,Sp1), Action = io(V,Ch,Span),
         %((transferReq,[int(3),ac1,ac2,true])=(Ch,V1) -> trace ; true),
         Result = esharing(CList,X1,Y1,SrcSpan),
         unify_spans(Sp1,SrcSpan,SP1b),
         hidden(Action,CList),
         unify_values(V1,V2,V,Ch,SP1b),  % try and also incorporate Sp2 Span, without compromising speed 
         % print(call_cspm_transY_B1(Y,io(V2,Ch),Y1)),nl,
         cspm_trans(Y, io(V2,Ch,Sp2), Y1,WF),   /* <----  Y will be recomputed for every solution of X !!! */
         unify_spans(SP1b,Sp2,Span3),
         Span = span_info(sharing,Span3)
         )
     ).
cspm_trans(esharing(CList,X,Y,SrcSpan), Action2, Result,WF) :- 
    shift_span_for_right_branch(SrcSpan,RSpan),
    % print(esharing_call_cspm_transY_B2(Y,ActionY,Y1)),nl,
     cspm_trans(Y,ActionY,Y1,WF),  /* this gets computed a second time; is there a way to avoid this? */
     %print(esharing_cspm_transY_B2(Y,ActionY)),nl,
     (   ActionY = tick(TS), Action=tau(tick(TS)), Result=esharing(CList,X,omega,SrcSpan)
      ;  functor_dif(ActionY,tick), ActionY = Action,
         %print(try_non_sharing(ActionY)),nl,
         not_hidden(ActionY,CList), /* covers tau */
         Result = esharing(CList,X,Y1,SrcSpan)
     ),
    merge_span_into_event(Action,RSpan,Action2).


/* Linked Parallel  [ l <-> r,.. ] */
cspm_trans(lParallel(LinkList,X,Y,Span), Action,Result,WF) :-   
  evaluate_link_list(LinkList,EvLinkRenameList),
  %print(evlinklist(LinkList,EvLinkRenameList)),nl, %
  cspm_trans(elinkParallel(EvLinkRenameList,X,Y,Span),Action,Result,WF).
cspm_trans(elinkParallel(_,omega,omega,Span), tick(Span), omega ,_WF).
cspm_trans(elinkParallel(EvLinkRenameList,X,Y,LinkSpan),Action,Result,WF) :-
   cspm_trans(X,AX,X2,WF),  % print(elink_x(X,AX,X2,EvLinkRenameList)),nl, print('State: '),print(X),nl,
   (   AX = tick(TS), 
       shift_span_for_left_branch(LinkSpan,LSpan),
       merge_span_into_event(tau(tick(TS)),LSpan,Action),
       Result=elinkParallel(EvLinkRenameList,omega,Y,LinkSpan)
      ; 
       functor_dif(AX,tick),
	    ( shift_span_for_left_branch(LinkSpan,LSpan),
              merge_span_into_event(AX,LSpan,Action),
	      Result = elinkParallel(EvLinkRenameList,X2,Y,LinkSpan),
	      not_renamed(AX,EvLinkRenameList)
	      % ,print(not_renamed(AX,EvLinkRenameList)),nl %%
		; 
              Action=tau(link(AX,io(V,Ch,Span))),
	      Result = elinkParallel(EvLinkRenameList,X2,Y2,LinkSpan),
	      force_rename_action(AX,EvLinkRenameList,io(V1,Ch,Sp0)),
	      unify_spans(LinkSpan,Sp0,Sp1),
	      %% print(renamed(AX,io(V1,Ch),EvLinkRenameList)),nl, %%
	      unify_values(V1,V2,V,Ch,Sp1),   % try and also incorporate Sp2 Span, without compromising speed 
	      % MAYBE TO DO ?: instantiate V2 as much as possible ! --> in()/dot issues
	      %% print(cspm_trans(Y,io(V2,Ch,Sp2),Y2)),nl, %%
	      cspm_trans(Y,io(V2,Ch,Sp2),Y2,WF),   /* <----  Y will be recomputed for every solution of X !!! */
	      unify_spans(Sp1,Sp2,Span3),
	      %% print(unify(Sp1,Sp2,Span3)),nl, %%
              Span = span_info(sharing,Span3)
	   )
   ).
cspm_trans(elinkParallel(EvLinkRenameList,X,Y,LinkSpan),Action2,
           elinkParallel(EvLinkRenameList,X,YR,LinkSpan),WF) :-
  shift_span_for_right_branch(LinkSpan,RSpan),
  rev_rename_list(EvLinkRenameList,RevList), 
  cspm_trans(Y,ActionY,Y2,WF),   /* TO DO: try and avoid recomputation of cspm_trans(Y) ?! */
  ( ActionY=tick(TS), Action=tau(tick(TS)), YR=omega
   ;
    functor_dif(ActionY,tick),ActionY=Action,YR=Y2,not_renamed(Action,RevList)
  ),
  merge_span_into_event(Action,RSpan,Action2).

cspm_trans(aParallel(CListX,X,CListY,Y,SrcSpan),A,NewState,WF) :-  %print(expanding_aparx),nl,
   evaluate_argument(CListX,EvCListX),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCListX,ECListX,SrcSpan)), %print(expanding_apary),nl,
   evaluate_argument(CListY,EvCListY),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCListY,ECListY,SrcSpan)),
   Expanded = eaParallel(ECListX,X,ECListY,Y,SrcSpan),
   %we have expanded the channel synchronisation sets; avoid recomputing them every time
   % TO DO: investigate whether it also makes sense to precompute intersection
   cspm_trans(Expanded,A,NewState,WF).
cspm_trans(eaParallel(_,omega,_,omega,SrcSpan), tick(SrcSpan), omega ,_WF).
cspm_trans(eaParallel(ECListX,X,ECListY,Y,SrcSpan),A,eaParallel(ECListX,X2,ECListY,Y1,SrcSpan),WF) :- 
   %trans_pre_compute(Y,Results),
   cspm_trans(X, AX, X1,WF),
   % print(apar_x(AX, from(X))),nl,
   (  X2=X1,Y1=Y,
      hidden_or_tau(AX,ECListX),
      not_hidden(AX,ECListY),
      shift_span_for_left_branch(SrcSpan,LSpan),
      merge_span_into_event(AX,LSpan,A)
     ;
      AX=tick(_TS), X2=omega,Y1=Y,
      shift_span_for_left_branch(SrcSpan,LSpan),
      merge_span_into_event(tau(AX),LSpan,A)
     ;
       AX=io(V1,Ch,Sp1), X2=X1, A = io(V,Ch,Span),
       % nl,print(try_apar_sync(V1,Ch)),nl,
       hidden(io(V1,Ch,Sp1),ECListX), % print(left_ok),nl, % do V1 instead of V ?
       hidden(io(V,Ch,Span),ECListY), % print(right_ok),nl, % V2 instead of V 
       % Synchronisation possible
       unify_spans(Sp1,SrcSpan,SP1b),
       unify_values(V1,V2,V,Ch,SP1b),  % try and also incorporate Sp2 Span, without compromising speed 
        %print(trying(aparY(io(V2,Ch,Sp2)))),nl, %
       cspm_trans(Y, io(V2,Ch,Sp2), Y1,WF),   /* <----  Y will be recomputed for every solution of X !!! */
      % trans_cont(Results,Y, io(V2,Ch,Sp2), Y1),
       % print(done(aparY(io(V2,Ch,Sp2)))),nl, %
       unify_spans(SP1b,Sp2,Span3),
       Span = span_info(sharing,Span3)
   ).
cspm_trans(eaParallel(ECListX,X,ECListY,Y,SrcSpan),AS,eaParallel(ECListX,X,ECListY,Y2,SrcSpan),WF) :-  
   cspm_trans(Y, AY, Y1,WF),  /* To do: try and avoid recomputation of cspm_trans(Y) ?! */
   (  AY=A, Y2=Y1,
      hidden_or_tau(A,ECListY), not_hidden(A,ECListX)
    ;
      AY=tick(_TS), A = tau(AY), Y2=omega
   ),
   shift_span_for_right_branch(SrcSpan,RSpan),
   merge_span_into_event(A,RSpan,AS).
   
/* only used for outside xtl_interface ; not used internally anymore */
cspm_trans(val_of(X,Span),A,NewExpr,WF) :- (symbol(RenamedX,X,_,_) -> cspm_trans(agent_call(Span,RenamedX,[]),A,NewExpr,WF); 
   cspm_trans(X,A,NewExpr,WF)).
cspm_trans(agent_call(Span,F,Par),NA,NNewExpr,WF) :- 
     unfold_function_call_once(F,Par,Value,Span), 
    ~~pp_cll(cspm_trans(Value,A,NewExpr,WF)),
     full_normalise_csp_process(NewExpr,NNewExpr),
   %  print(merge_agcall(A,Span)),nl,
     merge_span_into_event(A,Span,NA).

% Par = [[x],[y]]
cspm_trans(agent_call_curry(F,Par),NA,NNewExpr,WF) :-
     Span=no_loc_info_available,
     unfold_function_call_curry_once(F,Par,Value,Span),
    ~~pp_cll(cspm_trans(Value,A,NewExpr,WF)),
     full_normalise_csp_process(NewExpr,NNewExpr),
   %  print(merge_agcall(A,Span)),nl,
     merge_span_into_event(A,Span,NA).

cspm_trans(builtin_call(X),Action,NewExpr,WF) :- % builtin_call is normally removed by precompilation; but can remain, e.g., for assertions
   cspm_trans(X,Action,NewExpr,WF).
cspm_trans(head(X),Action,NewExpr,WF) :-  force_evaluate_argument(X,EX),head_list(EX,Result),
   cspm_trans(Result,Action,NewExpr,WF).

/* hiding '\\' */

cspm_trans('\\'(Expr,CList,Span), A, Res ,WF) :- 
   evaluate_argument(CList,EvCList),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCList,ECList,no_loc_info_available)),
  % print(ehide(Expr,ECList,Span)),nl,
   cspm_trans(ehide(Expr,ECList,Span),A,Res,WF).

cspm_trans(ehide(Expr,CList,Span), A, Res ,WF) :-
%    (nonvar(A) -> ((A=io(_,_,_);A=tick(_)) -> ActionX=A ; A=tau(hide(ActionX)))
%      ; true),
	cspm_trans(Expr,ActionX,X,WF),
	cspm_hide_action(ActionX,X, CList,Span, A,Res).


cspm_trans(exception(CList,X,Y,Span), A, Res ,WF) :- 
   evaluate_argument(CList,EvCList),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCList,ECList,no_loc_info_available)),
   %print(exception(ECList,X,Y)),nl,
   cspm_trans(eexception(ECList,X,Y,Span),A,Res,WF).

cspm_trans(eexception(CList,X,Y,Span),A,Res,WF) :- % expanded version of exception
  cspm_trans(X, ActionX, X1,WF),
	(  functor_dif(ActionX,tick), ActionX=A,
	   Res = eexception(CList,X1,Y,Span),
	   not_hidden(A,CList)
	 ; merge_span_into_event(ActionX,Span,A),
       Res = Y,
       hidden(ActionX,CList)
	 ; A=tick(TS), ActionX=tick(TS), Res=omega % is this correct ??
	).
   
  
/* timeout */
cspm_trans('[>'(P,Q,SrcSpan),AS,Res,WF) :- cspm_trans(P,A,P1,WF),
   (  A=tau(_), Res='[>'(P1,Q,SrcSpan)
    ; 
     top_level_dif(A,tau(_)), % a transition which can be removed by the tau below
     Res=P1 ),
   shift_span_for_left_branch(SrcSpan,LSpan),
   merge_span_into_event(A,LSpan,AS).
cspm_trans('[>'(_P,Q,SrcSpan),tau(timeout(SrcSpan)),Q,_WF).


/* interrupt */

cspm_trans('/\\'(X,Y,Span),NA,Res,WF) :-
    cspm_trans(X,A,X1,WF), %%print(interrupt_left(X,A,X1)),nl,
    shift_span_for_left_branch(Span,LSpan),
    merge_span_into_event(A,LSpan,NA),
    (   A=tick(_), Res=X1
     ; functor_dif(A,tick), tl_normalise('/\\'(X1,Y,Span),Res)
    ).
cspm_trans('/\\'(X,Y,Span),NA,Res,WF) :-
    cspm_trans(Y,A,Y1,WF),
    shift_span_for_right_branch(Span,RSpan),
    merge_span_into_event(A,RSpan,NA),
    (  top_level_dif(A,tau(_)), Res=Y1
     ; A=tau(_), tl_normalise('/\\'(X,Y1,Span),Res)
    ).
    
/* renaming [[ old <- new,... | Generators ]] */
cspm_trans(procRenamingComp(X,GeneratorList,RenameList),RA,Res,WF) :-
   % print(procCompRenaming(GeneratorList)),nl, %%
   % warning: does not have the rangeEnum wrapper expected
   expand_set_comprehension(rangeEnum(RenameList),GeneratorList,setValue(ExpandedRenames)),
   %print(exp(ExpandedRenames)),nl,
   cspm_trans(procRenaming(ExpandedRenames,X,no_loc_info_available),RA,Res,WF).

/* renaming [[ old <- new,... ]] */
cspm_trans(procRenaming(RenameList,X,Span),RA,Res,WF) :-  %print(procRename(RenameList)),nl,
  evaluate_rename_list(RenameList,ERenameList),
  cspm_trans(eprocRenaming(ERenameList,X,Span),RA,Res,WF).
cspm_trans(eprocRenaming(RenameList,X,Span),RAS,Res,WF) :- 
   %print('RENAME: '), translate:print_cspm_state(X),nl,
   cspm_trans(X,A,X2,WF),
   tl_normalise(eprocRenaming(RenameList,X2,Span),Res),
   %print(try_rename(A,RenameList,RA,Span)),nl, %%
   % it's still possible that A is non-ground term, i.e. there are a non-ground variables (c?x?y) in the sub-process,
   % which have to be unified before applying renaming on them.
   (ground(A) -> true; enumerate_action(A)),
   rename_action(A,RenameList,RA),
   %%% print(done_rename_action(A,RenameList,RA)),nl,
   merge_span_into_event(RA,Span,RAS).
   % print(merge_span(RA,Span,RAS)),nl,
   
/* if then else */
cspm_trans(ifte(Test,Then,Else,S1,S2,S3),A,X1,WF) :- 
   %nl,print(checking_if_test(Test,Then,Else)),nl,
   ~~mnf(haskell_csp:evaluate_boolean_expression(Test,Res)),
    % print(if_evaluated_boolean_expression(Test,Res)),nl,nl,
   cspm_if_trans(Res,Then,Else,A,X1,S1,S2,S3,WF).

cspm_trans('&'(Test,Then),A,X1,WF) :- S=no_loc_info_available,
   cspm_trans(ifte(Test,Then,stop(S),S,S,S),A,X1,WF).


cspm_trans(repChoice(GeneratorList,Body,Span), Action,X1,WF) :- 
  %% nl,print(repChoice(GeneratorList,Body)),nl,
   replicate_expand_set_comprehension([Body],GeneratorList,setValue(Bodies)),
 %% print(bodies(Bodies)),nl,
 convert_into_choice(Bodies,Span,CHOICE),
%% print(choice(CHOICE)),nl,
   cspm_trans(CHOICE,Action,X1,WF).
   
cspm_trans(repInternalChoice(GeneratorList,Body,Span), tau(rep_int_choice(Span)),Res,_WF) :- 
    % print(repInternalChoice(GeneratorList,Span)),nl,
   replicate_expand_set_comprehension([Body],GeneratorList,Bodies),
   % print(bodies(Bodies)),nl,nl,
   (is_empty_set(Bodies,true)
    -> add_error(cspm_trans,'Empty set for replicated internal choice: ',Bodies,Span),fail
     ; is_member_set(Res,Bodies)).
           
cspm_trans(repInterleave(GeneratorList,Body,Span), A, X,WF) :- 
  %% tools:print_bt_message(repInterleave(GeneratorList)), %%
   % multiplicity is relevant for |||, so add variables to Body below in case Variables are not used
   extract_variables_from_generator_list(GeneratorList,Variables), 
   replicate_expand_set_comprehension([na_tuple([Body|Variables])],GeneratorList,setValue(Bodies)),
   convert_into_interleave(Bodies,Span,INTLV),
  %% tools:print_bt_message(intlv(INTLV)), %%
   cspm_trans(INTLV,A,X,WF). %, tools:print_bt_message(sol(A,X)).
           

cspm_trans(repSequence(GeneratorList,Body,Span),A,X,WF) :- 
   % print(repSequence(GeneratorList)),nl, %%
   expand_listcomprehension(rangeEnum([Body]),GeneratorList,list(Bodies)),
   % print(bodies(Bodies)),nl, %%
   convert_into_seqcomp(Bodies,Span,SEQ), %print(sequential(SEQ)),nl,
   cspm_trans(SEQ,A,X,WF).


cspm_trans(procRepAParallel(GeneratorList,pair(SyncSet,Body),Span), A, X,WF) :- 
   %print(procRepAParallel(GeneratorList,SyncSet)),nl,
   % multiplicity is relevant for sharing (it makes a difference if 1 or 2 copies are present), so add variables to Body below in case Variables are not used
   extract_variables_from_generator_list(GeneratorList,Variables), 
   %print(extractedVars(Variables)),nl,
   replicate_expand_set_comprehension([list([SyncSet,Body,Variables])],GeneratorList,setValue(Bodies)),
   % print(bodies_repA(Bodies)),nl,
   convert_into_eaParallel(Bodies,Span,APAR,_ALPH),   % nl,print(eapar(APAR)),nl,
   cspm_trans(APAR,A,X,WF).


cspm_trans(procRepLinkParallel(SyncSet,GeneratorList,Body,Span), A, X,WF) :-
  % print(procRepLinkParallel(GeneratorList)),nl,
   expand_listcomprehension(rangeEnum([Body]),GeneratorList,list(Bodies)), 
  % print(bodies(Bodies)),nl,
   convert_into_linkParallel(Bodies,SyncSet,Span,APAR,Body), %print(linkpar(APAR)),nl,
   %when((ground(APAR)),
   cspm_trans(APAR,A,X,WF).
    
cspm_trans(procRepSharing(SyncSet,GeneratorList,Body,Span), A, X,WF) :-
  % print(procRepSharing(GeneratorList)),nl,
   % multiplicity is relevant for sharing, so add variables to Body below in case Variables are not used
   extract_variables_from_generator_list(GeneratorList,Variables), 
   replicate_expand_set_comprehension([na_tuple([Body|Variables])],GeneratorList,setValue(Bodies)),
 %  print(bodies(Bodies)),nl,
   convert_into_Sharing(Bodies,SyncSet,Span,APAR), %print(repshare(APAR)),nl,
   cspm_trans(APAR,A,X,WF).

% ----------

cspm_hide_action(ActionX,X, CList,Span, ActionAfterHiding,ResultingCSPExpr) :-
	(  functor_dif(ActionX,tick), ActionX=ActionAfterHiding,
	   tl_normalise(ehide(X,CList,Span),ResultingCSPExpr),
	   not_hidden(ActionAfterHiding,CList)
	 ; merge_span_into_event(ActionX,Span,NActionX),
       ActionAfterHiding=tau(hide(NActionX)),
       tl_normalise(ehide(X,CList,Span),ResultingCSPExpr),
       hidden(ActionX,CList)
	 ; ActionAfterHiding=tick(TS), ActionX=tick(TS), ResultingCSPExpr=omega
	).

:- block cspm_if_trans(-,?,?,?, ?,?,?,?,?).
cspm_if_trans(true,Then,_Else,AS,X1,_S1,S2,_S3,WF) :- !, %print(ifte_true(Then)),nl,
   full_normalise_csp_process(Then,NThen), 
   cspm_trans(NThen,A,X1,WF),
   merge_span_into_event(A,S2,AS).
cspm_if_trans(false,_Then,Else,AS,X1,_S1,_S2,S3,WF) :- !, %print(ifte_false(Else)),nl,
   full_normalise_csp_process(Else,NElse), cspm_trans(NElse,A,X1,WF),
   merge_span_into_event(A,S3,AS).
cspm_if_trans(Other,Then,Else,AS,X1,S1,S2,S3,WF) :- 
   add_internal_error('Internal Error: Not a boolean value inside if-then-else: ',
                      cspm_if_trans(Other,Then,Else,AS,X1,S1,S2,S3,WF)),
   fail.
 
:- block convert_into_interleave(-,-,?).
convert_into_interleave([],Span,skip(Span)).
convert_into_interleave([na_tuple([H|_])|T],Span,R) :- convert_into_interleave3(T,H,Span,R).
:- block convert_into_interleave3(-,?,?,?).
convert_into_interleave3([],X,_Span,X).
convert_into_interleave3([na_tuple([H2|_])|T],H1,Span,'|||'(H1,Rest,Span)) :-
      convert_into_interleave3(T,H2,Span,Rest).
  
  
:- block convert_into_choice(-,-,?).
convert_into_choice([],Span,stop(Span)).
convert_into_choice([H|T],Span,R) :- convert_into_choice3(T,H,Span,R).
:- block convert_into_choice3(-,?,?,?).
convert_into_choice3([],X,_Span,X).
convert_into_choice3([H2|T],H1,Span,'[]'(H1,Rest,Span)) :- convert_into_choice3(T,H2,Span,Rest).

:- block convert_into_seqcomp(-,-,?).
convert_into_seqcomp(Bodies,Span,Res) :- convert_into_seqcomp2(Bodies,Span,Res).
convert_into_seqcomp2([],Span,skip(Span)) :- print(convert_into_seqcomp2([],Span,skip(Span))),nl.
convert_into_seqcomp2([H|T],Span,R) :- convert_into_seqcomp3(T,H,Span,R).
:- block convert_into_seqcomp3(-,?,?,?).
convert_into_seqcomp3([],X,_Span,X).
convert_into_seqcomp3([H2|T],H1,Span,';'(H1,Rest,Span)) :-
   convert_into_seqcomp2([H2|T],Span,Rest).


:- block convert_into_eaParallel(-,-,?,?).
convert_into_eaParallel([],Span,skip(Span),setValue([])).
convert_into_eaParallel([H|T],Span,APAR,ALPH) :- 
   when(nonvar(T),convert_into_eaParallel3([H|T],Span,APAR,ALPH)).
convert_into_eaParallel3([list([Sync,X,_])],_Span,X,ESync) :- !, evaluate_argument(Sync,ESync).
convert_into_eaParallel3([list([CListX,X,_]),L2|T], Span,AlphPar, Alphabet ) :-
   when(nonvar(T),(convert_into_eaParallel3([L2|T],Span, Rest,RestAlphabet),
      evaluate_argument(CListX,ECListX),
      evaluate_argument(union(RestAlphabet,ECListX),Alphabet),
      /* put more complicated process onto the left: better efficiency for cspm_trans */
      % TO DO: we could think about splitting the list in two 
      AlphPar  = aParallel(RestAlphabet,Rest,ECListX,X,Span))). % only aParallel



:- block convert_into_Sharing(-,?,-,?).
convert_into_Sharing([],_,Span,skip(Span)).
convert_into_Sharing([na_tuple([H|_])|T],SyncSet,Span,R) :- convert_into_Sharing3([H|T],SyncSet,Span,R).
:- block convert_into_Sharing3(-,?,?,?).
convert_into_Sharing3([X],_Sync,_Span,Res) :- !, Res=X.
convert_into_Sharing3([X|T],CListX,Span,sharing(CListX,Rest,X,Span) ) :-
   /* put more complicated process onto the left: better efficiency for cspm_trans */
   convert_into_Sharing(T,CListX,Span,Rest).
   
:- block convert_into_linkParallel(-,?,-,?,?).
convert_into_linkParallel([],_,Span,skip(Span),Body) :-
   add_error(haskell_csp,'Empty Set in Replicated Linked Parallel: ',Body).
convert_into_linkParallel([H|T],SyncSet,Span,APAR,Body) :-
      convert_into_linkParallel3([H|T],SyncSet,Span,APAR,Body).
:- block convert_into_linkParallel3(-,?,?,?,?).
convert_into_linkParallel3([X],_Sync,_Span,Res,_) :- !, Res=X.
convert_into_linkParallel3([X|T],CListX,Span,lParallel(CListX,Rest,X,Span),Body ) :-
   /* put more complicated process onto the left: better efficiency for cspm_trans */
   convert_into_linkParallel(T,CListX,Span,Rest,Body).

/* ------------------------------------- */

/* top_level_normalise */
/* Just check the top-level of the csp expression to see if it can
   be simplified/normalised */
tl_normalise('[]'(X,Y,Span),R) :- !,
  (X=Y -> R=X; (is_stop(X,_) -> R=Y ; (is_stop(Y,_) -> R=X ; R='[]'(X,Y,Span)))).  /* TO DO: also sort X,Y !? */
tl_normalise('|||'(X,Y,Span),R) :- !,
  ((is_stop(X,S),is_stop(Y,_)) % in case we use ==/2 then the 'X==stop(S),Y=stop(_)' test will always fail (see user's manual of sicstus)
    -> R=stop(S)
    ;  ((X==omega;is_skip(X)) -> R=Y    /* law 6.16 on page 142 of Roscoe's book */
           ; ((Y==omega;is_skip(Y)) -> R=X  /* same law, for RHS */
                  ; full_normalise_csp_process(X,XR),
                    full_normalise_csp_process(Y,YR),
                    ((preference(cspm_detailed_animation,true);X @=<Y)
                            -> R='|||'(XR,YR,Span)
                             ; R='|||'(YR,XR,Span))  %% Warning: this can invalidate left-right Span information
              )
       )
  ).  % we could replace skip|||P by P
 % (X==stop -> R=Y ; (Y==stop -> R=X ; R='|||'(X,Y))). P ||| stop  not equal to stop (because of omega)
tl_normalise(ehide(X,CList,Span),R) :- !,
  (is_stop(X,S) -> R=stop(S) ; try_tl_normalise(X,NX),R=ehide(NX,CList,Span)). 
tl_normalise(eprocRenaming(RL,X,Span),R) :- !,
  (is_stop(X,S) -> R=stop(S) ; try_tl_normalise(X,NX),R=eprocRenaming(RL,NX,Span)). 
tl_normalise(';'(X,Y,SPAN),R) :- !,
  (is_skip(X) -> R=Y  /* Law on page 140 of Roscoe */
    ; (is_stop(X,S) -> R=stop(S) ; R = ';'(X,Y,SPAN) )).  /* we could also check Y for == skip */
tl_normalise('/\\'(X,Y,Span),R) :- !,
  (is_stop(X,_) -> R=Y ; (is_stop(Y,_) -> R=X ; R='/\\'(X,Y,Span))). 
tl_normalise(ifte(A,B,C,_,_,_),R) :- !,  % print(norm_ifte(A)),nl,
   ~~mnf(haskell_csp:evaluate_boolean_expression(A,ARes)),
   % print(res(ARes)),nl,
   tl_normalise_ifte(ARes,B,C,R).

is_stop(X,S) :- nonvar(X),X=stop(S).

:- block tl_normalise_ifte(-,?,?,?).
tl_normalise_ifte(true,B,_C,R) :- !,full_normalise_csp_process(B,R).
tl_normalise_ifte(false,_,C,R) :- !, full_normalise_csp_process(C,R).
% evaluate_boolean_expression/2 will always throw an exception if A is not a boolean test
% probably this case will never occur
% tl_normalise_ifte(A,B,C,_R) :- add_error(tl_normalise,'Not a boolean test: ',A:B:C).

try_tl_normalise(A,R) :- (tl_normalise(A,B) -> R=B ; R=A).

is_skip(X) :- nonvar(X), X=skip(_).
% TO DO: add rules esharing(_,stop(S),stop(_),_)  -> stop(S),     

% P [] STOP = P
:- assert_must_succeed((
    P='[]'(skip(_),stop(span1),span2),
    haskell_csp: full_normalise_csp_process(P,NormP),
    NormP=skip(_))).
% P [] P = P
:- assert_must_succeed((
    P='[]'(skip(span1),skip(span1),span3),
    haskell_csp: full_normalise_csp_process(P,NormP),
    NormP=skip(span1))).
% P ||| STOP = STOP
:- assert_must_succeed((
    P='|||'(skip(_),stop(span1),span2),
    haskell_csp: full_normalise_csp_process(P,NormP),
    NormP=stop(span1))).
% SKIP; (P [] STOP) = P
:- assert_must_succeed((
    P=';'(skip(_),'[]'(skip(span1),stop(span3),span4),span2),
    haskell_csp: full_normalise_csp_process(P,NormP),
    NormP='[]'(skip(span1),stop(span3),span4))).
% P; SKIP = P
:- assert_must_succeed((
    P=';'(stop(_),skip(span1),span2),
    haskell_csp: full_normalise_csp_process(P,NormP),
    NormP=stop(_))).

/* ------------------------------------- */
%full_normalise_csp_process(P,_) :- print(full_normalise_csp_process(P)),nl,fail.
/* TO DO: add other procRep,... intermediate operators */
full_normalise_csp_process(P,NormP) :- full_normalise_csp_process(P,NormP,outer).
:- block full_normalise_csp_process(-,?,?).
full_normalise_csp_process('|||'(X,Y,Span),Res,_) :- !,
   %full_normalise_csp_process(X,NX,inner), full_normalise_csp_process(Y,NY,inner),
   full_normalise_csp_process(X,NX,outer), full_normalise_csp_process(Y,NY,outer),
   tl_normalise('|||'(NX,NY,Span),Res).
% both clausels below will be probably never executed. procRenamingComp/3 and procRepAParallel/3 as 
% replicated process expressions will be translated to agent_calls in the compiling/analyzing stage
% of loading the respective CSP model. DEAD CODE:
/*full_normalise_csp_process(procRenamingComp(X,GeneratorList,RenameList),Res,_) :- !,
   %print(normalise_procCompRenaming(GeneratorList)),nl,
   replicate_expand_set_comprehension(RenameList,GeneratorList,setValue(ExpandedRenames)),
   %print(exp(ExpandedRenames)),nl,
   Res = procRenaming(ExpandedRenames,X,no_loc_info_available).
full_normalise_csp_process(procRepAParallel(GeneratorList,pair(SyncSet,Body),Span), Res,_) :- !, 
  %print(procRepAParallel(GeneratorList,SyncSet)),nl,
   extract_variables_from_generator_list(GeneratorList,Variables), 
   replicate_expand_set_comprehension([list([SyncSet,Body,Variables])],GeneratorList,setValue(Bodies)),
  %print(bodies(Bodies)),nl,
   convert_into_eaParallel(Bodies,Span,APAR,_ALPH), %print(apar(APAR,_ALPH)),nl,
   full_normalise_csp_process(APAR,Res,inner). */
full_normalise_csp_process(repInterleave(GeneratorList,Body,Span),Res,_) :- !,
   extract_variables_from_generator_list(GeneratorList,Variables), 
   replicate_expand_set_comprehension([na_tuple([Body|Variables])],GeneratorList,setValue(Bodies)),
   convert_into_interleave(Bodies,Span,Res).
full_normalise_csp_process(sharing(CList,X,Y,SrcSpan),Res,_) :- !, %trace,
   evaluate_argument(CList,EvCList),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCList,ECList,SrcSpan)),
   full_normalise_csp_process(X,NX,outer),
   full_normalise_csp_process(Y,NY,outer),
   Res = esharing(ECList,NX,NY,SrcSpan).
full_normalise_csp_process(aParallel(CListX,X,CListY,Y,SrcSpan),Res,_) :- !,
   evaluate_argument(CListX,EvCListX),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCListX,ECListX,SrcSpan)),
   evaluate_argument(CListY,EvCListY),
   ~~mnf(haskell_csp:expand_channel_pattern_expression(EvCListY,ECListY,SrcSpan)),
  full_normalise_csp_process(X,NX,outer),
  full_normalise_csp_process(Y,NY,outer),
    %print(eaParallel(ECListX,NX,ECListY,NY,SrcSpan)),nl,
   Res = eaParallel(ECListX,NX,ECListY,NY,SrcSpan).
 /* Unfold if Par ground and unfolded value ground ?? */
full_normalise_csp_process(agent_call(Span,F,Par),Res,outer) :- % only allow outer agent_calls to be unfolded to avoid infinite expansion for e.g. P(1) = P(1) ||| P(1)
   preference(cspm_detailed_animation,false),
  %% fail, %% fail: agent_calls are not expanded here, which means better source location feedback in animator
   %% WE COULD MAKE THIS AN OPTION IN THE PREFERENCES; I am not sure what the performance impact is
   %% for scheduler this leads to a substantial increase in the number of states
   !,
   ((ground(Par),unfold_function_call_once(F,Par,Value,Span),
    (ground(Value) ; 
     agent_can_be_unfolded(Value)
				  )
    )
     -> full_normalise_csp_process(Value,Res,outer)
     ; Res = agent_call(Span,F,Par)
   ).
full_normalise_csp_process(val_of(RenamedX,Span),NormState,_) :- !,
    (symbol(RenamedX,_X,_,_) ->
        full_normalise_csp_process(agent_call(Span,RenamedX,[]),NormState,_)
    ; otherwise ->
        % does this occur in some case (see coverage info)
        full_normalise_csp_process(RenamedX,NormState,_)
    ).
full_normalise_csp_process(ifte(Test,Then,Else,S1,S2,S3),R,_) :- 
    ground(Test),!,
    evaluate_boolean_expression(Test,Res),
    (Res == true ->
      full_normalise_csp_process(Then,R)
    ; Res == false ->
      full_normalise_csp_process(Else,R)
    ; otherwise -> 
       add_internal_error('Internal Error: Not a boolean value inside if-then-else: ',
                      full_normalise_csp_process(ifte(Test,Then,Else,S1,S2,S3))),fail
    ).
full_normalise_csp_process(ifte(Test,Then,Else,S1,S2,S3),R,_) :- !,
    R =  ifte(Test,Then,Else,S1,S2,S3). % test could be false or true; top-level test could be evaluated ?!
full_normalise_csp_process('&'(Test,Then),R,_) :- !, R = '&'(Test,Then). % test could be false; don't unfold agent_calls inside Then
full_normalise_csp_process(';'(P,Q,SPAN),R,Scope) :- !,
    full_normalise_csp_process(P,NP,Scope),
    tl_normalise(';'(NP,Q,SPAN),R).
full_normalise_csp_process(prefix(SPAN,Values,ChannelExpr,CSP,SPAN2),R,_) :- !,
    (empty_inGuard_set(Values) -> 
        R=stop(SPAN) % (STOP-step) ?x:{} -> P = STOP (law 1.15)
    ; otherwise ->
        R = prefix(SPAN,Values,ChannelExpr,CSP,SPAN2)
    ).
full_normalise_csp_process(head(List),R,_) :- !, R=head(List). % should we evaluate ??
full_normalise_csp_process(agent_call(Span,A,Par),R,_) :- !, R=agent_call(Span,A,Par).
% DEAD CODE!!! agent_call_curry/2 will be translated into nested agent_call in the precompiling phase.
%full_normalise_csp_process(agent_call_curry(A,Par),R,_) :- !, print(agent_call_curry(A,Par)),nl,R=agent_call_curry(A,Par). 
full_normalise_csp_process(ehide(X,CList,Span),R,_) :- !,
    full_normalise_csp_process(X,NX,inner),
    tl_normalise(ehide(NX,CList,Span),R).
full_normalise_csp_process(eprocRenaming(RL,X,Span),R,_) :- !,
    full_normalise_csp_process(X,NX,inner),
    tl_normalise(eprocRenaming(RL,NX,Span),R).
full_normalise_csp_process(Proc,Res,_) :-
    binary_top_level_process_operator_possible_to_be_normalised(Proc,F,X,Y,Span),!,
    full_normalise_csp_process(X,NX,inner),
    full_normalise_csp_process(Y,NY,inner),
    functor(NTerm,F,3),arg(1,NTerm,NX),arg(2,NTerm,NY),arg(3,NTerm,Span),
    tl_normalise(NTerm,Res).
full_normalise_csp_process(Proc,Res,_) :- definite_cspm_process_construct(Proc,Spans,Rest,RecCSP), !,
    functor(Proc,F,N), functor(Res,F,N),
    maplist(full_normalise_csp_process_inner(inner),RecCSP,NormRecCSP),
    definite_cspm_process_construct(Res,Spans,Rest,NormRecCSP).
full_normalise_csp_process(P,R,_) :- !,
    add_warning(full_normalise_csp_process,'Could not normalise: ',P),R=P.

full_normalise_csp_process_inner(ProcLabel,Proc,NProc) :-
   full_normalise_csp_process(Proc,NProc,ProcLabel).

binary_top_level_process_operator_possible_to_be_normalised(ProcessTerm,F,X,Y,Span) :-
    functor(ProcessTerm,F,3),
    memberchk(F,['|||','[]','/\\']),
    arg(1,ProcessTerm,X),arg(2,ProcessTerm,Y),arg(3,ProcessTerm,Span).   

empty_inGuard_set(Values) :-
   member(inGuard(_X,Set),Values),
   is_empty_set(Set,true),!.

normalise_cspm_state(State,NormState) :-
      full_normalise_csp_process(State,NormState).

%% Probably DEAD CODE !!! (wait for the next coverage release to prove this assumption)
:- assert_must_succeed((agent_can_be_unfolded(X), X=procRenamingComp(_,_,_))).
:- assert_must_succeed((agent_can_be_unfolded(X), X=aParallel(_,_,_,_,_))).
agent_can_be_unfolded(procRenamingComp(_,_,_)).
agent_can_be_unfolded(aParallel(_,_,_,_,_)).
agent_can_be_unfolded(repInterleave(_,_,_)).

/* ------------------------------------- */


/* EVALUATING Agent Call arguments */

evaluate_agent_call_parameters(AgentName,Par,_) :- var(AgentName),!,
   add_internal_error('Internal Error: Trying to call VARIABLE function: ',agent_call(AgentName,Par)),fail.
evaluate_agent_call_parameters(AgentName,Par,_) :- var(Par),!,
   add_internal_error('Internal Error: Trying to call VARIABLE parameters: ',agent_call(AgentName,Par)),fail.
%evaluate_agent_call_parameters(agent_call(AName,[]),Par,EAgentCall) :- atomic(AName),!,
%    /* then AName must represent a lambda function which has been compiled ?!!? */
%    l_evaluate_arguments(Par,EPar),
%    EAgentCall =.. [AName|EPar].
evaluate_agent_call_parameters(_Call,Par,EPar) :- Par==[],!, EPar=[]. % nothing to evaluate
evaluate_agent_call_parameters(AgentNameExpr,Par,_EAgentCall) :- \+ atomic(AgentNameExpr),!,
   add_error(haskell_csp,'Trying to call non-atomic function: ',agent_call(AgentNameExpr,Par),AgentNameExpr),fail.
   %print(non_atomic_funcall(AgentNameExpr)),nl,
   % ~~mnf(haskell_csp:evaluate_argument(AgentNameExpr,AName)), print(function(AName)),nl,
   % l_evaluate_arguments(Par,EPar),  EAgentCall = apply(AName,EPar).
evaluate_agent_call_parameters(AName,Par,EPar) :-
    length(Par,Arity),
    (agent_parameters(AName,Arity,LazyStrictList)
      -> true
      ;  add_error(evaluate_agent_call_parameters,'Unknown function call: ',(AName/Arity))
    ),
    l_evaluate_arguments(Par,LazyStrictList,EPar).
  %  EAgentCall =.. [AName|EPar].


l_evaluate_arguments([],[],[]).
l_evaluate_arguments([A|T],[Lazy|TL],[EA|ET]) :-
     %% print(eval_arg(A,Lazy)),nl, %%
    (Lazy=lazy -> ~~mnf(haskell_csp:evaluate_argument(A,EA))
               ;  ~~mnf(haskell_csp:force_evaluate_argument(A,EA))
    ),
     %% print(arg_res(A,EA)),nl, %%
    l_evaluate_arguments(T,TL,ET).
    

evaluate_lambda_arguments([],[],[]).
evaluate_lambda_arguments([A|T],[FormalPar|FT],[EA|ET]) :-
  %%  print(eval_arg(A,Lazy)),nl,
    (var(FormalPar) -> A=EA % will be copied into lambda body and evaluated then ~~mnf(haskell_csp:evaluate_argument(A,EA))
               ;  ~~mnf(haskell_csp:force_evaluate_argument(A,EA)) % it is a pattern; we need to evaluate argument
    ),
  %%  print(arg_res(A,EA)),nl,
    evaluate_lambda_arguments(T,FT,ET).
/* ------------------------------------- */

/* EVALUATING Channel arguments */

setup_skel(_,in(_)).

setup_channel_skeleton(io(V,Ch,_)) :-
   channel_type_list(Ch,ChannelTypes),
   maplist(setup_skel,ChannelTypes,V).

/* ------------------------------------- */

/* enumerate_action(A) : if part of the inputs of A are not determined: enumerate 
   possible values */
   
cspm_trans_enum(E,A,E2) :- 
  init_wait_flags(WF),
  cspm_trans(E,A,E2,WF), 
  %  tools:print_bt_message(grounding_waitflags(A)), %%
  ground_wait_flags(WF), 
  % TO DO: we could also add enumerate_action to the WF store: in principle sets inGuards should be smaller
  %    tools:print_bt_message(enumerate_action(E,A)), %%
  enumerate_action(A)
 % ,tools:print_bt_message(finished_enumerate_action(A)),nl %%
  .


 /* TO DO: add enumeration to hide + removed values by rename; use tau(V,Ch) ?? */
 
enumerate_action(X) :- X \= tick(_), X\=tau(_), X\= io(_,_,_), !,
   add_internal_error('Internal error: Unknown CSP event: ', enumerate_action(X)).
enumerate_action(tick(_)).
enumerate_action(tau(X)) :- (nonvar(X) -> enumerate_tau_argument(X) ; true).
enumerate_action(io(Vals,Channel,Span)) :-
  if((channel_type_list(Channel,ChannelTypes)),
     (enume_ch_vals(ChannelTypes,Vals,Channel,Span,io(Vals,Channel,Span))),
     add_error(enumerate_action,'Channel not declared: ',Channel,Span)).
 
enumerate_tau_argument(io(V,C,S)) :- !,enumerate_action(io(V,C,S)).
enumerate_tau_argument(hide(X)) :- !,enumerate_action(X).
enumerate_tau_argument(link(_AX,AY)) :- !,enumerate_action(AY).
enumerate_tau_argument(_).

:- use_module(probsrc(translate),[translate_event/2]).
%% enume_ch_vals(T,V,C,_Span,FullEvent) :- print(enume_ch_vals(T,V,C)),nl,fail. %%
enume_ch_vals([],[],_,_Span,_) :- !.
enume_ch_vals([],Vals,_Channel,Span,FullEvent) :- !,
   (Vals = [in(V)]
      -> empty_tuple(V)  /* the empty dotTuple */
      ;  translate_event(FullEvent,ES),
         add_error(enumerate_action,'Too many output values on channel: ',ES:Vals, Span)
   ).
enume_ch_vals([Type|RestTypes],Vals,Channel,Span,FullEvent) :-  % T \= [],
  enume_ch_vals2(Vals,Type,RestTypes,Channel,Span,FullEvent).
  
enume_ch_vals2(X,Type,_RestTypes,_Channel,Span,FullEvent) :- % print(ecv2(X,Type,RestTypes,Channel)),nl, 
   var(X),!, 
   translate_event(FullEvent,ES),
   add_error(haskell_csp,'Variable value list on channel: ',ES:X:Type,Span),fail.
enume_ch_vals2([],Type,RestTypes,_Channel,Span,FullEvent) :- !, 
   translate_event(FullEvent,ES),
   add_error(enumerate_action,'Too few output values on channel: ',ES:[Type:RestTypes], Span).
enume_ch_vals2([in(V)|TAIL],Type,RestTypes,Channel,Span,FullEvent) :- nonvar(V),
    V=tuple(X),nonvar(X),X=[H|T], \+ csp_constructor(H,_,_), 
    !, % sometimes an in contains dot tuples;  test above not very elegant; can we simplify this ??
    (T==[] -> 
	enume_ch_vals2([in(H)|TAIL],Type,RestTypes,Channel,Span,FullEvent)
        ; enume_ch_vals2([in(H),in(tuple(T))|TAIL],Type,RestTypes,Channel,Span,FullEvent)).
enume_ch_vals2([in(V)|TAIL],Type,RestTypes,Channel,Span,_FullEvent) :- 
    TAIL==[],!,  /* only treat it as a tail_in(X) if it is definitely at the end of the list
                    see example LetTestChannel -> TEST(2) process for complication if we do
                    not check that TAIL==[] */
    (RestTypes=[]
       -> enumerate_channel_input_value(Type,V,Channel,Span)
       ;  enumerate_channel_input_value('dotTupleType'([Type|RestTypes]),V,Channel,Span)
    ).
enume_ch_vals2([IODVal|TV],Type,RestTypes,Channel,Span,FullEvent) :- !,
    enumerate_channel_value_type(IODVal,Type,Channel,Span),
    enume_ch_vals(RestTypes,TV,Channel,Span,FullEvent).
 
    
enumerate_channel_value_type(IOD,Type,Channel,Span) :-
   %print(enumerate_channel_value_type(IOD,Type,Channel,Span)),nl,
   (IOD=in(HX) %; IOD=dot(HX))
     -> enumerate_channel_input_value(Type,HX,Channel,Span)
     ;  (ground(IOD)
          -> check_channel_input_output_value(IOD,Type,Channel,Span,'Type error in channel input: ')
          ;  enumerate_channel_input_value(Type,IOD,Channel,Span)
        )
    ).

check_channel_input_output_value(X,Type,Ch,Span,ErrMsg) :-
   when(ground(X), ((peel_in(X,PX),is_member_set_alsoPat(PX,Type)) -> true
                     ; add_error(check_channel_output_value,ErrMsg,(Ch:(val(X):type(Type))),Span))).

/* peel input values which may lurk inside datavalues, especially records */
peel_in(in(X),R) :- !, R=X.
peel_in(record(F,L),R) :- !,
   R=record(F,PL),
   maplist(peel_in,L,PL).
peel_in(X,X).

:- assert_must_succeed(( haskell_csp:enumerate_channel_input_value(setFromTo(1,3),R,c,unknown),
  R==int(2) )).

enumerate_channel_input_value(Type,Val,Ch,Span) :- 
   MAXRECURSIONDEPTH = 2,
   enumerate_channel_input_value(Type,Val,Ch,MAXRECURSIONDEPTH,Span).


enumerate_channel_input_value(Type,Val,Ch,MaxRec,Span) :- 
   (ground(Val)
       -> check_channel_input_output_value(Val,Type,Ch,Span,'Type error in channel output: ')
       ;  enumerate_channel_input_value1(Type,Val,Ch,MaxRec,Span)
   ).

enumerate_channel_input_value1(Type,Val,Ch,MaxRec,Span) :-
       ((nonvar(Val),Val=in(VX))  /* can happen when question marks occur inside records, ... */
          ->  enumerate_channel_input_value1_1(Type,VX,Ch,MaxRec,Span)
          ;   enumerate_channel_input_value1_1(Type,Val,Ch,MaxRec,Span)
       ).

enumerate_channel_input_value1_1(Type,Val,Ch,MaxRec,Span) :-
	( (nonvar(Val),Val=alsoPat(X,Y)) ->
		enumerate_channel_input_value2(Type,X,Ch,MaxRec,Span),
		%enumerate_channel_input_value2(Type,Y,Ch,MaxRec,Span),
		unify_also_patterns(X,Y,_R)
	; (nonvar(Val),Val=appendPat(X,FunHead)) -> 
		agent_compiled(FunHead,_Value,_SRCSPAN),
        enumerate_channel_input_value2(Type,X,Ch,MaxRec,Span)
	; enumerate_channel_input_value2(Type,Val,Ch,MaxRec,Span)
	).

%enumerate_channel_input_value2(T,V,Ch,_MaxRec) :- print(enumerate_channel_input_value(T,V,Ch)),nl,fail.
:- use_module(probsrc(refinement_checker),[ignore_infinite_datatypes/0]).
:- assert_must_succeed((findall(Res, enumerate_channel_input_value2('Set'(setValue([int(1),int(2)])),Res,_Ch,_MaxRec,_Span),L), 
                       L == [setValue([int(1),int(2)]),setValue([int(1)]),setValue([int(2)]),setValue([])])).

enumerate_channel_input_value2(intType,X,Ch,_MaxRec,_Span) :- !,
   kernel_objects:enumerate_basic_type(X,integer,trigger_true(Ch)). % NOTE : this uses MAXINT preference value!
enumerate_channel_input_value2(boolType,X,_Ch,_MaxRec,_Span) :- !,(X=true ; X=false).
enumerate_channel_input_value2(setFromTo(Low,Up),Res,Ch,_MaxRec,Span) :- !, 
      set_enumeration_result(Res,int(X),Ch,setFromTo(Low,Up),Span),
      enumerate_csp_int(X,Low,Up).
      /*(preference(use_clpfd_solver,true),clpfd_interface: fd_var(Res) ->
          clpfd_interface: csp_clpfd_labeling([ffc,enum],[Res])
      ; set_enumeration_result(Res,int(X),Ch,setFromTo(Low,Up),Span),
        enumerate_csp_int(X,Low,Up)
      ).*/
enumerate_channel_input_value2(setFrom(Low),V,Ch,_MaxRec,Span) :- !,
      add_symbol_span(Ch,Span,CSpan),
      add_error(enumerate_channel_input_value,'Cannot enumerate infinite Type: ',(Ch,(setFrom(Low),V)),CSpan).
enumerate_channel_input_value2('Set'(X),Res,_Ch,_MaxRec,_Span) :- !, /* Probably DEAD CODE */
   enum_subset(X,Res).
enumerate_channel_input_value2('dotTupleType'(List),Res,Ch,MaxRec,Span) :- !,
  set_enumeration_tuple_result(Res,RR,Ch,'dotTupleType'(List),Span),
  l_enumerate_channel_input_value2(List,RR,Ch,dot_tuple,MaxRec,Span).
enumerate_channel_input_value2('typeTuple'(List),Res,Ch,MaxRec,Span) :- !, 
  set_enumeration_result(Res,na_tuple(RR),Ch,'typeTuple'(List),Span),
  % print(enum_tuple('typeTuple'(List),Res,Ch)),nl,
  l_enumerate_channel_input_value2(List,RR,Ch,type_tuple,MaxRec,Span).
enumerate_channel_input_value2(setValue(L),R,_Ch,_MaxRec,_Span) :- !,member(R,L).  /* R should be list skeleton at least */
enumerate_channel_input_value2(dataType(DT),C,Channel,MaxRec,Span) :- is_a_datatype(DT,_), !,
     enumerate_datatype_el(DT,C,Channel,MaxRec,Span).
% TO DO: add  all other types supported by is_member_set
enumerate_channel_input_value2('Seq'(setValue([])),V,_Ch,_MaxRec,_Span) :- !, V= list([]).
enumerate_channel_input_value2('Seq'(Type),V,Ch,_MaxRec,Span) :- !,
      (refinement_checker:ignore_infinite_datatypes -> V=sequence % needed for the Trace Debugger in Tcl/Tk  
      ; add_symbol_span(Ch,Span,CSpan),
        add_error(enumerate_channel_input_value,'Cannot enumerate infinite Seq Type: ',(Ch,('Seq'(Type),V)),CSpan)).
enumerate_channel_input_value2(agent_call(_Src, Name,Patterns),Val,Ch,_MaxRec,Span) :- !, 
    unfold_function_call_once(Name,Patterns,Res,Span),
    evaluate_argument(Res,Res1),
    enumerate_channel_input_value(Res1,Val,Ch,Span).
enumerate_channel_input_value2(Type,V,Ch,_MaxRec,Span) :-
      add_symbol_span(Ch,Span,CSpan),
      add_error(enumerate_channel_input_value,'Illegal type for channel: ',(Ch,(Type,V)),CSpan).

enumerate_datatype_el(DT,C,Channel,_MaxRec,Span) :- 
    if(csp_constant(C,DT),true,
       (nonvar(C),C \= record(_,_),
        add_error(enumerate_datatype_el,'Type error for channel input variable: ', Channel:(C,type(DT)),Span),
        fail)).
enumerate_datatype_el(DT,C,Channel,MaxRec,Span) :-
    ((MaxRec>0 ; nonvar(C))
        ->((nonvar(C) -> M1 = MaxRec ; M1 is MaxRec-1),
          C=record(Cons,Fields),
          csp_constructor(Cons,DT,ArgSubTypes),
         % print(recurse_enum_rec(Cons,DT,ArgSubTypes,Fields)),nl,
          l_enumerate_channel_input_value2(ArgSubTypes,Fields,Channel,dot_tuple,M1,Span))
        ; (refinement_checker:ignore_infinite_datatypes -> 
             print('### WARNING: Maximum Recursion Depth reached for enumeration at '),
       translate: print_span(Span)
           ; print('### Maximum Recursion Depth reached for enumeration at '),
             translate:print_span(Span),
             fail)
      ).

% ensure that ResVar is a tuple and flatten the tuple
set_enumeration_tuple_result(ResVar,ResultTerm,Channel,ExpectedType,Span) :-
  ((tuple(TL)=ResVar;dotTuple(TL)=ResVar) ->
      %print(flattening(ResVar)),nl,
      flatten_tuple_value(TL,ResultTerm) /* ok, no type error */ 
    ; add_symbol_span(Channel,Span,CSpan),
      add_error(set_enumeration_tuple_result,'Type error for instantiated channel input variable: ',Channel:(ResVar,type(ExpectedType)),CSpan),
      fail).

%%%%%%%%%%%%%%%%%%%%% UNIT TESTS FOR flatten_tuple_value/2, flatten_inner_tuple_value/2 and flatten2/3 %%%%%%%%%%%%%%%%%%%%%%%%%%
:- assert_must_succeed((haskell_csp: flatten_tuple_value(X,R), R == X)).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([],R), R == [])).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([X,tuple([int(1),int(2)])],R), R == [X,int(1),int(2)])).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([tuple([]),tuple([int(1),int(2)]),X,Y,tuple([int(10)])],R), R == [int(1),int(2),X,Y,int(10)])).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([X,tuple([int(1),Y,int(2),Z])],R), R == [X,int(1),Y,int(2),Z])).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([X,tuple([int(1),Y,int(2),Z]),int(3)],R), R == [X,int(1),Y,int(2),Z,int(3)])).
:- assert_must_succeed((haskell_csp: flatten_tuple_value([],[]))).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

flatten_tuple_value(X,R) :- var(X),!, R=X. % this can happen if we have a single in(X) representing ch?x 
flatten_tuple_value([],[]).
flatten_tuple_value([H|T],R) :- flatten2(H,T,R).

flatten_inner_tuple_value(X,R) :- var(X),!, print(variable_inner_tuple(X)),nl, R=[X].
flatten_inner_tuple_value([],[]).
flatten_inner_tuple_value([H|T],R) :- flatten2(H,T,R).

flatten2(VAR,T,Res) :- var(VAR),!, Res=[VAR|TR],flatten_inner_tuple_value(T,TR).
flatten2(tuple(LH),T,R) :- !, flatten_inner_tuple_value(LH,FLH),
   append(FLH,TR,R), flatten_inner_tuple_value(T,TR).
flatten2(H,T,[H|TR]) :- flatten_inner_tuple_value(T,TR).

set_enumeration_result(ResVar,ResultTerm,Channel,ExpectedType,Span) :-
  (ResVar=ResultTerm -> true /* ok, no type error */ 
    ; 
      add_symbol_span(Channel,Span,CSpan),
      add_error(set_enumeration_result,'Type error for instantiated channel input variable: ',Channel:(val(ResVar),type(ExpectedType)),CSpan),
      fail).
   
l_enumerate_channel_input_value2([],R,Ch,_SType,_MaxRec,Span) :- !,
   (var(R) -> R=[]
    ; R=[] -> true
    ; add_error(l_enumerate_channel_input_value,'Too many arguments on channel: ',(Ch,R),Span),fail).
l_enumerate_channel_input_value2([HType|T],[HVal|TV],Ch,SType,MaxRec,Span) :- !,
      ((TV==[],SType=dot_tuple) -> % enume the rest of the channel types to HVal (only for dotTuple types)
          enume_ch_vals([HType|T],[in(HVal)],Ch,Span,io([],Ch)) % TO DO: full event
      ; otherwise ->
          enumerate_channel_input_value(HType,HVal,Ch,MaxRec,Span), %%  <-- add Span
          l_enumerate_channel_input_value2(T,TV,Ch,SType,MaxRec,Span)
      ).
l_enumerate_channel_input_value2([HType|T],[],Ch,_SType,_MaxRec,Span) :- !,
      add_error(l_enumerate_channel_input_value,'Not enough arguments on channel: ',(Ch,[HType|T]),Span),fail.
l_enumerate_channel_input_value2(X,Y,Ch,_SType,MR,Span) :-
      add_internal_error('Illegal arguments: ',l_enumerate_channel_input_value2(X,Y,Ch,MR),Span),fail.
    
get_symbol_span(S,SPAN) :- atomic(S),symbol(S,_,SPAN,_),!.
get_symbol_span(S,SPAN) :-  nonvar(S),functor(S,FS,_),symbol(FS,_,SPAN,_),!.
get_symbol_span(_S,no_loc_info_available).


add_symbol_error(Src,Msg,Term,Ch) :-
   get_symbol_span(Ch,Span),
   add_error(Src,Msg,Term,Span).

add_symbol_span(Ch,Span,OutSpan) :- get_symbol_span(Ch,SSpan),
   \+ (SSpan=no_loc_info_available),!, unify_spans(Span,SSpan,OutSpan).
add_symbol_span(_,S,S).

% just like above, but can also be called with ordinary spans
add_error_with_span(Src,Msg,Term,Span) :-
   (Span=symbol(S) -> get_symbol_span(S,Span2) ; Span2=Span),
   add_error(Src,Msg,Term,Span2).

add_internal_error_with_span(Src,Msg,Term,Span) :-
   (Span=symbol(S) -> get_symbol_span(S,Span2) ; Span2 = Span),
   add_internal_error(Src,Msg,Term,Span2).

/* ------------------------------------- */
/*           BOOLEAN  EXPRESSIONS        */
/* ------------------------------------- */

%%%%%%%% UNIT TESTS FOR TESTING check_boolean_expressions/1,cond_check_boolean_expression/2,ifte_check_boolean_expression/3 %%%%%%%

:- assert_must_succeed((haskell_csp: check_boolean_expression(true))).
:- assert_must_fail((haskell_csp: check_boolean_expression(X), X = false)).
:- assert_must_succeed((haskell_csp: check_boolean_expression(bool_not(false)))).
:- assert_must_fail((haskell_csp: check_boolean_expression(X), X = false)).
:- assert_must_succeed((haskell_csp: check_boolean_expression(bool_or(true,_X)))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(bool_not('<='(int(1),int(0)))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(bool_not('<='(int(1),int(10)))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(bool_and('<='(int(1),int(10)), true)))).
:- assert_must_fail((haskell_csp: check_boolean_expression(bool_and(false,'<='(int(1),int(10)))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(bool_or('<='(int(1),int(0)), true)))).
:- assert_must_fail((haskell_csp: check_boolean_expression(bool_or(false,'<='(int(1),int(0)))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression('<='(int(10),int(X))), X = 30)).
:- assert_must_succeed((haskell_csp: check_boolean_expression('>='(int(10),int(X))), X = 3)).
:- assert_must_succeed((haskell_csp: check_boolean_expression('<='(int(10),int(X))), X = 10)).
:- assert_must_succeed((haskell_csp: check_boolean_expression('>='(int(10),int(X))), X = 10)).
:- assert_must_succeed((haskell_csp: check_boolean_expression('<'(int(10),int(X))),  X = 30)).
:- assert_must_succeed((haskell_csp: check_boolean_expression('>'(int(10),int(X))),  X = 3)).
:- assert_must_fail((haskell_csp: check_boolean_expression('<'(int(10),int(X))), X = 10)).
:- assert_must_fail((haskell_csp: check_boolean_expression('>'(int(10),int(X))), X = 10)).
:- assert_must_succeed((haskell_csp: check_boolean_expression(ifte('<='(int(3),int(20)),'>='(int(2),int(1)),'<='(int(3),int(1)),span,span,span)))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(ifte('<='(int(3),int(2)),'>='(int(2),int(1)),'<='(int(3),int(10)),span,span,span)))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(ifte('<='(int(3),int(2)),'>='(int(2),int(1)),'<='(int(3),int(10)),span,span,span)))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(ifte(X,'>='(int(2),int(1)),'<='(int(3),int(1)),span,span,span)), 
         X = head(listExp(rangeEnum(['<='(int(2),int(3)),'>'(int(2),int(3))]))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(ifte(X,'>='(int(2),int(1)),'<='(int(3),int(1)),span,span,span)), 
         X = head(listExp(rangeEnum(['<='(int(2),int(1)),'>'(int(2),int(3))]))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(head(listExp(rangeEnum(['<'(int(2),int(3)),'>'(int(2),int(3))])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(head(listExp(rangeEnum([true,false])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(head(listExp(rangeEnum([false,false])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(head(listExp(rangeEnum([false,true])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression((head(listExp(rangeEnum([null(listExp(rangeEnum([false,true])))]))))))).
:- assert_must_fail((haskell_csp: check_boolean_expression((head(listExp(rangeEnum([empty(setValue([int(1),int(2)]))]))))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(empty(setExp(rangeEnum([])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(head(listExp(rangeEnum([bool_not(false)])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(head(listExp(rangeEnum([bool_or(true,false)])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(head(listExp(rangeEnum(['<='(int(2),int(3)),'>'(int(2),int(3))])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(head(listExp(rangeEnum(['<='(int(2),int(1)),'>'(int(2),int(3))])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(null(listExp(rangeEnum([])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(null(listExp(rangeEnum([int(1),int(2),int(3)])))))).
:- assert_must_succeed((haskell_csp: check_boolean_expression(empty(setExp(rangeEnum([])))))).
:- assert_must_fail((haskell_csp: check_boolean_expression(empty(setExp(rangeEnum([int(1),int(2),int(3)])))))).
:- assert_must_succeed((haskell_csp: is_boolean_expression(true), haskell_csp: is_boolean_expression(false))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- block check_boolean_expression(-).

%check_boolean_expression(E) :- %print(check_boolean_expression(E)),nl,
%   var(E),add_error(check_boolean_expression,'Variable boolean expression:',E),fail.
check_boolean_expression(true) :- !.
check_boolean_expression(false) :- !, fail.
check_boolean_expression(bool_not(BExpr1)) :- !,
   evaluate_boolean_expression(BExpr1,false). /* provide custom check_not_... predicate ?? */
check_boolean_expression(bool_and(BExpr1,BExpr2)) :- !,
   check_boolean_expression(BExpr1),
   check_boolean_expression(BExpr2).  
check_boolean_expression(bool_or(BExpr1,BExpr2)) :- !,
   evaluate_boolean_expression(BExpr1,R1),
   cond_check_boolean_expression(R1,BExpr2).  
check_boolean_expression(ifte(Test,Then,Else,_,_,_)) :- !,  
   evaluate_boolean_expression(Test,TestRes), %print(check_bool_ifte(Test,TestRes)),nl,
   ifte_check_boolean_expression(TestRes,Then,Else). 
check_boolean_expression(head(X)) :- !, force_evaluate_argument(X,EX),head_list(EX,HResult),
      check_boolean_expression(HResult). 
check_boolean_expression(null(Arg1)) :- !, force_evaluate_argument(Arg1,EX), is_null_list(EX,true).
check_boolean_expression(empty(Arg1)) :- !, force_evaluate_argument(Arg1,EX), is_empty_set(EX,true).
check_boolean_expression(agent_call(Span,F,Par)) :- !, /* in this context we need the result of the function call */
   unfold_function_call_once(F,Par,Body,Span),
   %when(nonvar(Body),
   check_boolean_expression(Body).
check_boolean_expression('member'(Arg1,Arg2)) :-  !,  % after bool_binary_op: there is a special rule for member there
  force_evaluate_argument(Arg1,EX),
  force_evaluate_argument_for_member_check(Arg2,EY), %print(lazy_member(EX,EY)),nl,
 % causes 10 % slowdown on basin_olderog_bank, but code is probably required similar to evaluate_boolean_expression below
 %  when(ground(EX), (ground(EY) -> is_member_set(EX,EY) 
 %		   ; /* exploring lazy expression */ force_evaluate_argument(EY,REY),is_member_set(EX,REY))). 
  when((ground(EX),ground(EY)), is_member_set(EX,EY)).
check_boolean_expression(BExpr) :- %covering all boolean expression with relational operators instead of '==' and '!='%
  relational_binary_op(BExpr,Arg1,Arg2,EX,EY,Call),!,
  evaluate_int_argument(Arg1,EX),
  evaluate_int_argument(Arg2,EY),
  Call.
check_boolean_expression(BExpr) :-
  bool_binary_op(BExpr,Arg1,Arg2,EX,EY,Call), !,
  %print(bool_binary_op(Arg1,Arg2,Call)),nl, %
  force_evaluate_argument(Arg1,EX),
  force_evaluate_argument(Arg2,EY),
  when((ground(EX),ground(EY)), Call).
check_boolean_expression(X) :- 
   add_error(haskell_csp,'Not a boolean expression: ',X),
   fail.


:- block cond_check_boolean_expression(-,?).
cond_check_boolean_expression(true,_).
cond_check_boolean_expression(false,BExpr2) :- check_boolean_expression(BExpr2).

:- block ifte_check_boolean_expression(-,?,?).
ifte_check_boolean_expression(true,Then,_) :- !, check_boolean_expression(Then).
ifte_check_boolean_expression(false,_,Else) :- !, check_boolean_expression(Else).
% evaluate_boolean_expression/3 will raise an error if the boolean expression in the ifte-condition cannot be evaluated to true or false => Internal Error
ifte_check_boolean_expression(Other,Then,Else) :- add_internal_error(/*ifte_check_boolean_expression,*/ 
  'Internal Error: Not a boolean inside if-then-else test: ',
   ifte_check_boolean_expression(Other,Then,Else)),fail.

:- block evaluate_boolean_expression(-,?).
%evaluate_boolean_expression(E,_) :- %print(evaluate_boolean_expression(E)),nl,
%   var(E),add_error(evaluate_boolean_expression,'Variable boolean expression:',E),fail.
evaluate_boolean_expression(true,R) :- !, R=true.
evaluate_boolean_expression(false,R) :- !, R=false.
evaluate_boolean_expression(bool_not(BExpr1),Res) :- !,
   evaluate_boolean_expression(BExpr1,R1),
   negate(R1,Res).
evaluate_boolean_expression(bool_and(BExpr1,BExpr2),Res) :- !,
   evaluate_boolean_expression(BExpr1,R1),
   cond_evalf_boolean_expression(R1,BExpr2,Res).  
evaluate_boolean_expression(bool_or(BExpr1,BExpr2),Res) :- !,
   evaluate_boolean_expression(BExpr1,R1),
   cond_evalt_boolean_expression(R1,BExpr2,Res).
  % when(ground(R1),
  %   (R1=true -> Res = true
  %      ;  evaluate_boolean_expression(BExpr2,Res))).
evaluate_boolean_expression('<'(Arg1,Arg2),Res) :-  !,
  evaluate_int_argument(Arg1,EX),evaluate_int_argument(Arg2,EY),safe_less_than(EX,EY,Res).
evaluate_boolean_expression('>'(Arg1,Arg2),Res) :-  !,
  evaluate_int_argument(Arg1,EX),evaluate_int_argument(Arg2,EY),safe_less_than(EY,EX,Res).
evaluate_boolean_expression('<='(Arg1,Arg2),Res) :-  !,
  evaluate_int_argument(Arg1,EX),evaluate_int_argument(Arg2,EY),safe_less_than_equal(EX,EY,Res).
evaluate_boolean_expression('>='(Arg1,Arg2),Res) :-  !,
  evaluate_int_argument(Arg1,EX),evaluate_int_argument(Arg2,EY),safe_less_than_equal(EY,EX,Res).
evaluate_boolean_expression(ifte(Test,Then,Else,_,_,_),R) :- !, 
   evaluate_boolean_expression(Test,TestRes),
   ifte_eval_boolean_expression(TestRes,Then,Else,R).
evaluate_boolean_expression(head(X),R) :- !, force_evaluate_argument(X,EX),head_list(EX,HResult),
      evaluate_boolean_expression(HResult,R). 
evaluate_boolean_expression(agent_call(Span,F,Par),R) :- !, /* in this context we need the result of the function call */
   unfold_function_call_once(F,Par,Body,Span),
   %when(nonvar(Body),
   evaluate_boolean_expression(Body,R).
evaluate_boolean_expression(prolog_constraint(C),Res) :- !,  %% print(prolog_constraint(C)),nl,
    (call(C) -> Res = true ; Res = false).
evaluate_boolean_expression(null(Arg1),Res) :- !, force_evaluate_argument(Arg1,EX), is_null_list(EX,Res).
evaluate_boolean_expression(empty(Arg1),Res) :- !, force_evaluate_argument(Arg1,EX), is_empty_set(EX,Res).
evaluate_boolean_expression(BExpr,Res) :-
  bool_binary_op(BExpr,Arg1,Arg2,EX,EY,Call), !, % print(binop(BExpr,Arg1,Arg2)),nl,
  force_evaluate_argument(Arg1,EX), 
  force_evaluate_argument(Arg2,EY), %print(ebe(Call)),nl,
  when((ground(EX),ground(EY)), 
     (Call -> Res = true ; Res = false)).
evaluate_boolean_expression('member'(Arg1,Arg2),Res) :-  !, % after bool_binary_op: there is a special rule for member there
  force_evaluate_argument(Arg1,EX),
  force_evaluate_argument_for_member_check(Arg2,EY), %print(lazy_member(EX,EY)),nl,
  when(ground(EX), 
      (ground(EY) -> (is_member_set(EX,EY) -> Res = true ; Res = false) 
       ; /* exploring lazy expression */ force_evaluate_argument(EY,REY),(is_member_set(EX,REY) -> Res = true ; Res = false))).
evaluate_boolean_expression(builtin_call(X),Res) :- !, evaluate_boolean_expression(X,Res).
evaluate_boolean_expression(X,_R) :- 
   add_error(haskell_csp,'Not a boolean expression: ',X),
   fail.


:- block negate(-,?).
negate(true,false).
negate(false,true).

:- block cond_evalt_boolean_expression(-,?,?).
cond_evalt_boolean_expression(true,_,true).
cond_evalt_boolean_expression(false,BExpr2,Res) :- evaluate_boolean_expression(BExpr2,Res).

:- block cond_evalf_boolean_expression(-,?,?).
cond_evalf_boolean_expression(false,_,false).
cond_evalf_boolean_expression(true,BExpr2,Res) :- evaluate_boolean_expression(BExpr2,Res).

:- block ifte_eval_boolean_expression(-,?,?,?).
ifte_eval_boolean_expression(true,Then,_,Res) :- !, evaluate_boolean_expression(Then,Res).
ifte_eval_boolean_expression(false,_,Else,Res) :- !, evaluate_boolean_expression(Else,Res).
ifte_eval_boolean_expression(Other,Then,Else,Res) :- /* evaluate_boolean_expression/2 will catch the error if there 
                   is no boolean expression test, thus internal_error must be generated for this clause. */
	add_internal_error('Not a boolean inside if-then-else boolean expression test: ',
                            ifte_eval_boolean_expression(Other,Then,Else,Res)),fail.

bool_binary_int_op('<'(X,Y),X,Y,EX,EY,'<'(EX,EY)).
bool_binary_int_op('>'(X,Y),X,Y,EX,EY,'>'(EX,EY)).
bool_binary_int_op('>='(X,Y),X,Y,EX,EY,'>='(EX,EY)).
bool_binary_int_op('<='(X,Y),X,Y,EX,EY,'=<'(EX,EY)).

relational_binary_op('<'(X,Y),X,Y,EX,EY,safe_less_than(EX,EY)).
relational_binary_op('>'(X,Y),X,Y,EX,EY,safe_less_than(EY,EX)).
relational_binary_op('<='(X,Y),X,Y,EX,EY,safe_less_than_equal(EX,EY)).
relational_binary_op('>='(X,Y),X,Y,EX,EY,safe_less_than_equal(EY,EX)).

bool_binary_op('elem'(X,Y),X,Y,EX,EY,is_elem_list(EX,EY)).
%bool_binary_op('member'(X,Y),X,Y,EX,EY,is_member_set(EX,EY)) :- (X=setValue(_) -> false;true).
bool_binary_op('member'(X,'Set'(Y)),X,Y,EX,EY,is_subset_of(EX,EY)) :-
    (X\=setValue(_) -> false; \+ground(X) -> false; otherwise -> true).
bool_binary_op('=='(X,Y),X,Y,EX,EY,equal_element(EX,EY)).
bool_binary_op('!='(X,Y),X,Y,EX,EY,not_equal_element(EX,EY)).

bool_binary_op_symbolic('member'(X,Y),X,Y,EX,EY,is_member_set(EX,EY)). % dealt with separately

unary_op(null(X),X,EX,is_null_list(EX)).
unary_op(empty(X),X,EX,is_empty_set(EX,true)).

is_boolean_expression(true).
is_boolean_expression(false).
is_boolean_expression(bool_not(_)).
is_boolean_expression(bool_and(_,_)).
is_boolean_expression(bool_or(_,_)).
is_boolean_expression(X) :-
   (unary_op(X,_,_,_) ; bool_binary_op(X,_,_,_,_,_) ;  bool_binary_op_symbolic(X,_,_,_,_,_) ; bool_binary_int_op(X,_,_,_,_,_)).


/* ------------------------------------- */
/*                SEQUENCES              */
/* ------------------------------------- */

:- use_module(probcspsrc(csp_sequences)).

/* ------------------------------------- */
/*            VALUE  EXPRESSIONS         */
/* ------------------------------------- */

:- type channelio +--> (atomic ; dotTuple(list(atomic))).

:- type channel_set +--> (setValue(list(channelio)) ; closure(list(channelio))).

:- type cspm_expr +-->  (type(cspm_arith_expr) ; type(channel_set) ; atomic ; agent_call(ground,nonvar,list(any)) ).

:- type cspm_arith_expr +--> (int(integer) ; 
                              '+'(cspm_arith_expr,cspm_arith_expr) ;
                              '-'(cspm_arith_expr,cspm_arith_expr) ;
                              '*'(cspm_arith_expr,cspm_arith_expr) ;
                              '/'(cspm_arith_expr,cspm_arith_expr) ;
                              '%'(cspm_arith_expr,cspm_arith_expr) ).
 
% forcing arguments in the context of a member check; will be used as second arg to is_member_set
% This means that certain things do not need to be computed (Union for example)
:- block force_evaluate_argument_for_member_check(-,?).
force_evaluate_argument_for_member_check(agent_call(Span,F,Par),R) :- !,
   %print(unfold(F,Par,Body)),nl,
   unfold_function_call_once(F,Par,Body,Span),
   force_evaluate_argument_for_member_check(Body,R).
% keep the following symbolic:
force_evaluate_argument_for_member_check(setExp(rangeEnum(T),G),R) :- nonvar(T),T=[H|TT],TT==[], var(H), !, 
   % if the tuples T have more than one element, we currently cannot check membership symbolically
   R=setExp(rangeEnum(T),G). % TO DO expand for rangeClosed,...
force_evaluate_argument_for_member_check('Union'(L),R) :- !, 
   force_evaluate_argument_for_member_check_into_list(L,LL),  %print(evalUnion(L,LL)),nl,
   R='Union'(LL).
force_evaluate_argument_for_member_check('Inter'(L),R) :- !,
   force_evaluate_argument_for_member_check_into_list(L,LL), R='Inter'(LL).
force_evaluate_argument_for_member_check('Seq'(L),R) :- !, R='Seq'(L).
 
force_evaluate_argument_for_member_check(X,EX) :-   force_evaluate_argument(X,EX).

% force evaluating a set into a list of elements; do not worry about checking duplicates in list
:- block force_evaluate_argument_for_member_check_into_list(-,?).
%force_evaluate_argument_for_member_check_into_list(A,R) :- print(force_evaluate_argument_for_member_check_into_list(A,R)),nl,fail.
force_evaluate_argument_for_member_check_into_list(agent_call(Span,F,Par),R) :- !, 
   /* in this context we need the result of the function call */
   unfold_function_call_once(F,Par,Body,Span),
   force_evaluate_argument_for_member_check_into_list(Body,R).
force_evaluate_argument_for_member_check_into_list(setExp(rangeEnum(L)),R) :- !, 
   force_evaluate_list_for_member_check(L,LL),R=setExp(rangeEnum(LL)).
force_evaluate_argument_for_member_check_into_list(X,EX) :- force_evaluate_argument(X,EX).

:- block force_evaluate_list_for_member_check(-,?).
force_evaluate_list_for_member_check([],R) :- !,R=[].
force_evaluate_list_for_member_check([H|T],R) :- !, R=[EH|ET],
     force_evaluate_argument_for_member_check(H,EH),
     force_evaluate_list_for_member_check(T,ET).
force_evaluate_list_for_member_check(X,R) :-
    add_internal_error('Iternal Error: Could not evaluate list: ',force_evaluate_list_for_member_check(X,R)),R=X.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%% UNIT TESTS FOR force_evaluate_argument/2, evaluate_argument/2 and evaluate_expression/2 %%%%%%%%%%%%%%%%%%%%%%%
:- assert_must_succeed((haskell_csp: force_evaluate_argument(listFromTo(1,3),R), R == list([int(1),int(2),int(3)]))).
:- assert_must_succeed((haskell_csp: evaluate_argument(listFromTo(1,3),R), R == list([int(1),int(2),int(3)]))).
:- assert_must_succeed((haskell_csp: evaluate_argument(listFrom(1),R), R == listFrom(1))).
:- assert_must_succeed((haskell_csp: evaluate_argument(setFromTo(1,3),R), R == setFromTo(1,3))).
:- assert_must_succeed((haskell_csp: evaluate_argument(list([int(1),int(2)]),R), R == list([int(1),int(2)]))).
:- assert_must_succeed((haskell_csp: force_evaluate_argument(listPat([int(1),int(2)]),R), R == list([int(1),int(2)]))).
:- assert_must_succeed((haskell_csp: evaluate_argument(ifte('<='(int(3),int(20)),'>='(int(2),int(1)),'<='(int(3),int(1)),span,span,span),R) ,R == true)).
:- assert_must_succeed((
       haskell_csp: evaluate_argument(ifte('<='(int(3),int(2)),'>='(int(2),int(1)),'<='(int(3),int(1)),span,span,span),R), R ==false)).
:- assert_must_succeed((haskell_csp: evaluate_argument(ifte('<='(int(3),int(2)),int(1),int(2),span,span,span),R), R == int(2))).
:- assert_must_succeed((haskell_csp: evaluate_argument(dataType('FRUIT'),R), R == dataType('FRUIT'))).
:- assert_must_succeed((haskell_csp: evaluate_argument(concat(listExp(rangeEnum([listFromTo(1,3),listExp(rangeEnum([int(4)]))]))),R),
                      R == list([int(1),int(2),int(3),int(4)]))).
:- assert_must_succeed((haskell_csp: evaluate_argument(listPat([int(1),int(2)]),R), R == list([int(1),int(2)]))).
:- assert_must_succeed((haskell_csp: evaluate_argument(tuplePat([int(1),int(2)]),R), R == na_tuple([int(1),int(2)]))).
:- assert_must_succeed((haskell_csp: evaluate_expression(setExp(rangeEnum([int(1),int(2)])),R), R == setValue([int(1),int(2)]))).
:- assert_must_succeed((haskell_csp: evaluate_expression(setFromTo(1,3),R), R == setValue([int(1),int(2),int(3)]))).
:- assert_must_succeed((haskell_csp: evaluate_expression(stop(src),R), R == 'Typed expression is a process construct.')).
:- assert_must_succeed((haskell_csp: evaluate_expression(left,R), R == 'Typed expression is a channel name.')).
:- assert_must_succeed((haskell_csp: evaluate_expression('Msg1',R), R == setValue(['A','B']))).
:- assert_must_succeed((haskell_csp: evaluate_expression('Msg1',R), R == setValue(['A','B']))).
:- assert_must_succeed((haskell_csp: evaluate_expression(builtin_call(set(listExp(rangeEnum([int(1),int(1)])))),R), R == setValue([int(1)]))).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

evaluate_expression(X,R) :- var(X),!,
    add_internal_error('Internal Error: Variable expression to evaluate: ',evaluate_expression(X,R)),
    fail.
evaluate_expression('Events',R) :- !,
    findall(Channel,is_a_channel_name(Channel),CL),
    evaluate_list(CL,CLL), closure_expand(CLL,R).
evaluate_expression(closure(X),R) :- !,
    evaluate_list(X,XL), closure_expand(XL,R).
evaluate_expression(builtin_call(X),R) :- !, evaluate_expression(X,R).
evaluate_expression(setExp(X),R) :- !, evaluate_set_expression(X,EX),expand_symbolic_set(EX,R,evaluate_expression).
evaluate_expression(setFromTo(Low,Up),R) :- !, expand_symbolic_set(setFromTo(Low,Up),R,evaluate_expression).
evaluate_expression(val_of(X,Src),R) :- !, evaluate_argument2(val_of(X,Src),R).
evaluate_expression(DT,R) :- is_a_datatype(DT,_),!,
   expand_symbolic_set(dataType(DT),R,eval).
evaluate_expression(Proc,R) :- definite_cspm_process_expression(Proc),!, R='Typed expression is a process construct.'.
evaluate_expression(Ch,R) :- is_a_channel_name(Ch),!, R='Typed expression is a channel name.'.
evaluate_expression(E,R) :- force_evaluate_argument(E,R).

:- block force_evaluate_argument(-,?).
force_evaluate_argument(Num,R) :- number(Num),!,R=int(Num).
force_evaluate_argument(agent_call(Span,F,Par),R) :- !, /* in this context we need the result of the function call */
   unfold_function_call_once(F,Par,Body,Span),
   force_evaluate_argument(Body,R).
force_evaluate_argument(setExp(RangeExpr,Generators),R) :- !, 
    expand_set_comprehension(RangeExpr,Generators,R).
%force_evaluate_argument(comprehensionGenerator(Var,Set),R) :- !,
%  R = comprehensionGenerator(Var,ESet),
%  force_evaluate_argument(Set,ESet).
force_evaluate_argument(ifte(Test,Then,Else,_S1,_S2,_S3),R) :- !,
   evaluate_boolean_expression(Test,TestRes),  % print(force_ifte(Test,TestRes)),nl,
   ifte_force_evaluate_argument(TestRes,Then,Else,R).
force_evaluate_argument(list(X),R) :- !, %print(eval_list(X)),nl,
    force_evaluate_list(X,EX),  % use force_evaluate_list_skel ??; perf. problem in misusing_lists
    R=list(EX).
force_evaluate_argument(listPat(X),R) :- !, %print(eval_list(X)),nl,
    force_evaluate_list(X,EX),  % use force_evaluate_list_skel ??; perf. problem in misusing_lists
    R=list(EX). 
force_evaluate_argument(listFromTo(X,Y),Res) :- !, expand_sequence(listFromTo(X,Y),Res).
force_evaluate_argument(listExp(S),ERes) :- !, evaluate_list_expression(S,Res), expand_sequence(Res,ERes).
%force_evaluate_argument(setValue(X),R) :- !, force_evaluate_set(X,R), print(fes(X,R)),nl.
   % Maybe we should ensure that setValues are always force evaluated: we only need to
   % take care when using set operations for replicated operators
force_evaluate_argument(Destr,Result) :- destructor(Destr),!, % we may need to further evaluate result
     evaluate_argument2(Destr,EX), % do ordinary evaluation
     force_evaluate_argument(EX,Result).
force_evaluate_argument(E,R) :- evaluate_argument2(E,R).  /* could there be lazy constructs lurking deep inside E ??? */

destructor(head(_)).
destructor(na_tuple_projection(_,_,_)).
destructor(tuple_projection(_,_,_)).
%destructor(record(_,_)).
% records ?


:- block ifte_force_evaluate_argument(-,?,?,?).
ifte_force_evaluate_argument(true,Then,_,Res) :- !, force_evaluate_argument(Then,Res).
ifte_force_evaluate_argument(false,_,Else,Res) :- !, force_evaluate_argument(Else,Res).
/* See comment for the error clause of the ifte_evaluate_argument/4 predicate. Same argument as below 
   for assuming the error clause as caused by internal call only. */
ifte_force_evaluate_argument(Other,Then,Else,Res) :- 
   add_internal_error('Internal Error: Not a boolean inside if-then-else test: ',
                      ifte_force_evaluate_argument(Other,Then,Else,Res) ),fail.

:- block ifte_evaluate_argument(-,?,?,?).
ifte_evaluate_argument(true,Then,_,Res) :- !, evaluate_argument(Then,Res).
ifte_evaluate_argument(false,_,Else,Res) :- !, evaluate_argument(Else,Res).
/* evaluate_boolean_expression/2 predicate is called before ifte_evaluate_argument/4.
   evaluate_boolean_expression/2 provides as result either true or false, that's why only internal call can 
   raise error for this predicate. */
ifte_evaluate_argument(Other,Then,Else,Res) :-
   add_internal_error('Internal Error: Not a boolean inside if-then-else test: ',
                      ifte_evaluate_argument(Other,Then,Else,Res)),fail.

%evaluate_argument(E,R) :- print(evaluate_argument(E,R)),nl,
 %(nonvar(R) -> true ; when(nonvar(R),(print(inst_eval(E,R)),nl,trace))), evaluate_argument1(E,R).
:- block evaluate_argument(-,?).
evaluate_argument(Expr,Res) :- evaluate_argument2(Expr,Res).

evaluate_argument2(ifte(Test,Then,Else,_S1,_S2,_S3),R) :- !,
   evaluate_boolean_expression(Test,TestRes),  %print(ifte(Test,TestRes)),nl,
   ifte_evaluate_argument(TestRes,Then,Else,R).
  % when(ground(TestRes),(TestRes=true -> evaluate_argument(Then,R) ; evaluate_argument(Else,R))). 
evaluate_argument2(int(X),R) :- !,R=int(X).
% case should be added because of the clpfd-library
evaluate_argument2(Num,R) :- number(Num),!,R=int(Num).
evaluate_argument2(true,R) :- !,R=true. 
evaluate_argument2(false,R) :- !,R=false. 
evaluate_argument2(negate(X),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),
   safe_is(Res,-(EX)).
evaluate_argument2('+'(X,Y),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   safe_is(Res,EX+EY).
evaluate_argument2('-'(X,Y),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   safe_is(Res,EX-EY).
evaluate_argument2('*'(X,Y),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   safe_is(Res,EX * EY).
evaluate_argument2('/'(X,Y),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   safe_is(Res,EX // EY).
evaluate_argument2('%'(X,Y),Result) :- !, Result = int(Res),
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   safe_is(Res,EX mod EY).
evaluate_argument2(length(X),Result) :- !,  force_evaluate_argument(X,EX),length_list(EX,Result).
evaluate_argument2('#'(X),Result) :- !,  force_evaluate_argument(X,EX),length_list(EX,Result).
evaluate_argument2(head(X),Result) :- !,  force_evaluate_argument(X,EX),
   head_list(EX,Result).
evaluate_argument2(tail(X),Result) :- !,  force_evaluate_argument(X,EX),tail_list(EX,Result).
evaluate_argument2('^'(X,Y),Result) :- !,  
    force_evaluate_argument(X,EX),force_evaluate_argument(Y,EY),
    %print('^'(X,EX,Y,EY)),nl,
    append_list(EX,EY,Result). %, print(append_list(EX,EY,Result)),nl.
evaluate_argument2(concat(X),Result) :- !,  force_evaluate_argument(X,EX),
    concat_lists(EX,Result).
evaluate_argument2(listFromTo(X,Y),Res) :- !, 
    %evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
    expand_sequence(listFromTo(X,Y),Res).
evaluate_argument2(listFrom(X),Res) :- !, Res = listFrom(X). % already evaluated
evaluate_argument2(listExp(S),Res) :- !, evaluate_list_expression(S,Res).
evaluate_argument2(listExp(RangeExpr,Generators),R) :- !, %print(listExp(Generators,RangeExpr)),nl,
    expand_listcomprehension(RangeExpr,Generators,R). %, print(res(R)),nl.
%evaluate_argument2(listComp(Generators,Tuples),R) :- !, %force_evaluate_list(Generators,EGen),
%    expand_sequence(listComp(Generators,Tuples),R).
evaluate_argument2(set(X),Res) :- !, evaluate_argument2(seq_to_set(X),Res).
evaluate_argument2(seq(X),Res) :- !, evaluate_argument2(set_to_seq(X),Res).
evaluate_argument2(set_to_seq(X),Res) :- !,
    force_evaluate_argument(X,EX),
    convert_set_to_seq(EX,Res).
evaluate_argument2(seq_to_set(X),Res) :-  !,
    force_evaluate_argument(X,EX),
    convert_seq_to_set(EX,Res).
evaluate_argument2(setFromTo(X,Y),Res) :- !, Res = setFromTo(X,Y). % already computed set expression
evaluate_argument2(setExp(S),Res) :- !, evaluate_set_expression(S,Res).
evaluate_argument2(setExp(RangeExpr,Generators),R) :- !, %print(setExp(Generators,RangeExpr)),nl,
    expand_set_comprehension(RangeExpr,Generators,R).
evaluate_argument2(setValue(X),R) :- !, R=setValue(X).
evaluate_argument2(singleSetPatValue(X,Span),R) :- !, force_evaluate_argument(X,EX),
  %print(singleSetPatValue(X,EX,R)),nl,
  singleSetElement(EX,R,Span).
evaluate_argument2('Seq'(X),R) :- !, evaluate_argument(X,EX),R='Seq'(EX).
evaluate_argument2(intType,R) :- !, R=intType.
evaluate_argument2(boolType,R) :- !, R=boolType.
%evaluate_argument2(setEnum(X),R) :- !, evaluate_list(X,XL), evaluate_set(XL,R). 

evaluate_argument2(closureComp(Generators,Tuples),R) :- !, 
    % print(closureComp(Generators,Tuples)),nl,
     expand_symbolic_set(setExp(rangeEnum(Tuples),Generators),setValue(List),evaluate_argument_closureComp),
    % print(closureset(List)),nl,
     evaluate_argument2(closure(List),R). %print(doneclosureComp(R)),nl.
evaluate_argument2(closure(X),R) :- !, %print(eval(closure(X))),nl,
     evaluate_list(X,XL),   % print(evaluate_closure(XL,R)),nl,
      evaluate_closure(XL,R).
      %mnf(haskell_csp:expand_channel_pattern_expression(closure(X),R,no_loc_info_available)).
evaluate_argument2('Events',R) :- !,
    findall(Channel,is_a_channel_name(Channel),CL),
    evaluate_argument2(closure(CL),R).
evaluate_argument2(union(X,Y),Result) :- !,  force_evaluate_argument(X,EX), force_evaluate_argument(Y,EY),
    union_set(EX,EY,Result).
evaluate_argument2('Union'(X),Result) :- !,  force_evaluate_argument(X,EX), 
   big_union(EX,Result).
evaluate_argument2('Inter'(X),Result) :- !,  force_evaluate_argument(X,EX),
    big_inter(EX,Result).
evaluate_argument2(diff(X,Y),Result) :- !,  force_evaluate_argument(X,EX), force_evaluate_argument(Y,EY),
    %print(diff_set(X,EX,Y,EY)),nl,
    diff_set(EX,EY,Result).
evaluate_argument2(inter(X,Y),Result) :- !,  force_evaluate_argument(X,EX), force_evaluate_argument(Y,EY),
    inter_set(EX,EY,Result).
evaluate_argument2(card(X),Result) :- !,  evaluate_argument(X,EX),
    cardinality(EX,Result).
evaluate_argument2('Set'(X),Result) :- !, %print(subsets(X)),nl,
    evaluate_argument(X,EX),
    %print(calling_subsets(EX)),nl,
    subsets(EX,Result). %, print(subset_result(Result)),nl.
evaluate_argument2(agent_call(Span,F,Par),R) :- !,
  evaluate_agent_call(F,Par,R,Span).
evaluate_argument2(agent_call_curry(F,Par),R) :- !,
  Span=no_loc_info_available,
  unfold_function_call_curry_once(F,Par,Value,Span),
  evaluate_argument(Value,R).
evaluate_argument2(val_of(VAR,Span),R) :- !,  /* this can only happen inside channel type expressions;
  other val_of's are compiled away by haskell_csp_analyzer */
       unfold_function_call_once(VAR,[],Body,Span),
       evaluate_argument(Body,R).
evaluate_argument2([],R) :- !,R=[]. 
evaluate_argument2([H|T],R) :- !,R=[H|T].  /* list of channel outputs */ 
evaluate_argument2(builtin_call(X),R) :- !,  evaluate_argument(X,R).
evaluate_argument2(dotTuple(X),R) :- !, evaluate_dot_tuple(X,R).
evaluate_argument2(dotTupleType(TupleArgs),R) :- !, 
   evaluate_list(TupleArgs,ETupleArgs), R = dotTupleType(ETupleArgs).
evaluate_argument2(dotpat(X),R) :- !, evaluate_dot_tuple(X,R).
evaluate_argument2(tuple(F),R) :- !, R=tuple(F).
evaluate_argument2(alsoPat(X,Y),R) :- !,
    evaluate_argument2(X,EX),
    evaluate_argument2(Y,EY),
    R=alsoPat(EX,EY).
evaluate_argument2(lambda(A,B),R) :- !, %print(eval(lambda(A,B))),nl,
    R=lambda(A,B),
    (ground(A) -> true ;
      mynumbervars(A)). /* numbervars to avoid nonvar states for ProB (slows down hashing,...) */
evaluate_argument2(tuplePat(TupleArgs),R) :- !,  /* a non-associative tuple */
   R = na_tuple(EA),force_evaluate_list(TupleArgs,EA).
evaluate_argument2(tupleExp(TupleArgs),R) :- !,  /* a non-associative tuple; same as tuplePat but not in pattern position */
   R = na_tuple(EA),force_evaluate_list(TupleArgs,EA).
evaluate_argument2(typeTuple(TupleArgs),R) :- !,
   evaluate_list(TupleArgs,ETupleArgs), R = typeTuple(ETupleArgs).
evaluate_argument2(na_tuple(F),R) :- !, R=na_tuple(F). /* non-associative tuple */
evaluate_argument2(na_tuple_projection(Nr,Tuple,Span),R) :- !,
    evaluate_int_argument(Nr,ENr,Span),
    force_evaluate_argument(Tuple,ETuple),
    na_tuple_projection(ENr,ETuple,R,Span).
evaluate_argument2(list_projection(Nr,Len,List,Span),R) :- !,
    evaluate_int_argument(Nr,ENr,Span),
    evaluate_int_argument(Len,ELen,Span),
    force_evaluate_argument(List,EList),
    list_projection(ENr,ELen,EList,R,Span).
evaluate_argument2(tuple_projection(Nr,Tuple,Span),R) :- !,
    evaluate_int_argument(Nr,ENr,Span),
    force_evaluate_argument(Tuple,ETuple),
    tuple_projection(ENr,ETuple,R,Span).
evaluate_argument2(record(C,F),R) :- !, R=record(C,F).
evaluate_argument2(list(X),R) :- !, %print(eval_list(X)),nl,
    %evaluate_list(X,EX),
    R=list(X).    % TO DO: check if we also need listPat
evaluate_argument2(listPat(X),R) :- !,
    evaluate_list(X,EX), R=list(EX).
evaluate_argument2(rename(X,Y),R) :- !, 
    ~~mnf(haskell_csp:evaluate_rename_channel_expression(X,RX)),  /* should we evaluate ??? */
    ~~mnf(haskell_csp:evaluate_rename_channel_expression(Y,RY)),
   R=rename(RX,RY).  
evaluate_argument2(link(X,Y),R) :- !, 
    ~~mnf(haskell_csp:evaluate_rename_channel_expression(X,RX)),  /* should we evaluate ??? */
    ~~mnf(haskell_csp:evaluate_rename_channel_expression(Y,RY)),
   R=link(RX,RY).
evaluate_argument2(dataType(DT),R) :- !, R=dataType(DT).
evaluate_argument2(Channel,R) :-
    atomic(Channel),is_a_channel_name(Channel),!,R=tuple([Channel]).
evaluate_argument2(DT,R) :- name_type(DT,DT_Def), !, R=DT_Def.
evaluate_argument2(DT,R) :- dataTypeDef(DT,_), !, R=dataType(DT).
evaluate_argument2(DT,R) :- subTypeDef(DT,_), !, R=dataType(DT).
evaluate_argument2(X,R) :- valid_constant(X),!,R=X.
evaluate_argument2(X,R) :- agent_functor(X,_),!,R=X.  /* when a function name is passed  */
evaluate_argument2(X,R) :-
  is_csp_constructor(X),!,R=X.  /* FDR does not allow this as a valid value; but can be fed into closure */
evaluate_argument2(X,R) :- 
  (undefined_process_construct(X) -> 
    (is_boolean_expression(X) -> evaluate_boolean_expression(X,R) 
    /* we have a boolean value; we could also decide not to evaluate the argument and put R=X*/
        ; add_error(evaluate_argument2,'Cannot evaluate expression: ',X), R=X
     )
     ; R=X  /* we have a process description which is simply kept as is */
  ).

% setExp(.)
evaluate_set_expression(rangeEnum(List),Res) :- !, %evaluate_list(List,XL), 
   evaluate_set(List,Res).%evaluate_set(List,Res,evaluate_argument).
evaluate_set_expression(rangeClosed(X,Y),Res) :- !,
    evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   ((number(X),number(Y),X>Y) -> Res= setValue([]) ; Res = setFromTo(EX,EY)).
evaluate_set_expression(rangeOpen(X),Res) :- !,Res = setFrom(EX),
    evaluate_int_argument(X,EX).
evaluate_set_expression(A,Res) :-
    add_internal_error('Internal Error: Unknown setExp: ',evaluate_set_expression(A,Res)),Res=[].

% listExp(.)
evaluate_list_expression(rangeEnum(List),Res) :- !, Res=list(LR), evaluate_list(List,LR).
evaluate_list_expression(rangeClosed(X,Y),Res) :- !, % print(evaluate_rangeClosed(X,Y)),nl,
   evaluate_int_argument(X,EX),evaluate_int_argument(Y,EY),
   ((number(X),number(Y),X>Y) -> Res= list([]) ; Res = listFromTo(EX,EY)).
evaluate_list_expression(rangeOpen(X),Res) :- !,Res = listFrom(EX),
   evaluate_int_argument(X,EX).
evaluate_list_expression(A,Res) :- /* CSP-M Parser cannot generate list expression, 
			which is not of the type rangeEnum/1, rangeClosed/2, rangeOpen/1. 
			That's why we can assume that in this case we have internal error. */
	add_internal_error('Internal Error: Unknown listExp: ',ifte_eval_boolean_expression(A)),Res=[].



evaluate_function_name(F,EF) :-
   (atomic(F) -> EF=F ; force_evaluate_argument(F,EF)).

evaluate_agent_call(F,Par,Res,Span) :-  % print(eval_agent_call(F,Par)),nl,
  evaluate_function_name(F,EF),evaluate_agent_call2(EF,Par,Res,Span).

evaluate_agent_call2(lambda(Args,Body),Par,R,Span) :- !,
   unfold_function_call_once2(lambda(Args,Body),Par,NBody,Span),
   evaluate_argument(NBody,R).
evaluate_agent_call2(F,Par,R,Span) :- !,
  (atomic(F) -> true
   ; add_error(evaluate_argument,'Illegal function name for function application: ',agent_call(F,Par),Span),fail),
  X =.. [F|Par],
  ((is_csp_process(X) ; is_possible_csp_process(X)) % ATTENTION: the parameter in is_possible_csp_process/1 can be a function call, which can be missunderstood as process
		       % a reimplementation of the is_possible_csp_process/1 is needed.
       -> R=agent_call(Span,F,Par) %%%, print(keeping(X)),nl)
      /* keep agent_calls lazily; they are processes. If we unfold then the variables will be
         propagated into the term; causing problems by instantiation (see e.g. Repeat(Proc) = Proc ; Repeat(Proc)) */
         /* QUESTION: what about is_possible_csp_process ??? */
   ;  unfold_function_call_once2(F,Par,Body,Span),
       evaluate_argument(Body,R)              /* TO DO: avoid re-checking when(nonvar ... for recursive calls ?*/
       % force_evaluate_argument(Body,R)
        ).

% --------------------------------------------------------
%                       integer expressions
% --------------------------------------------------------

% --------------------------------------------------------

evaluate_int_argument(A,Res) :- evaluate_int_argument(A,Res,no_loc_info_available).

:- block evaluate_int_argument(-,?,?).

evaluate_int_argument(int(X),R,_) :- !, R=X.
evaluate_int_argument(agent_call(Span,F,Par),R,Span) :- !, /* in this context we need the result of the function call */
   unfold_function_call_once(F,Par,Body,Span),
   %print(unfolded(F,Par,Body)),nl,
   evaluate_int_argument(Body,R,Span).
evaluate_int_argument(Expr,IntRes,Span) :-
  force_evaluate_argument(Expr,Res), % because we need an integer, i.e., agent_calls need to be unfolded
  (Res = int(IntRes) -> true
   ; (Res \= int(_), add_error(haskell_csp,'Expression does not evaluate to integer: ','='(Expr,Res),Span),
      fail)
  ).

:- block evaluate_list(-,?).
evaluate_list([],R) :- !,R=[].
evaluate_list([H|T],R) :- !, R=[EH|ET],evaluate_argument(H,EH), evaluate_list(T,ET).
evaluate_list(X,R) :- add_internal_error('Internal Error: Could not evaluate: ',evaluate_list(X,R)),R=X.

:- block evaluate_type_list(-,?).
evaluate_type_list([],R) :- !,R=[].
evaluate_type_list([H|T],R) :- !, R=[EH|ET],
   translate_to_type_expr(H,TypeH),
   evaluate_argument(TypeH,EH),
   evaluate_type_list(T,ET).
evaluate_type_list(X,R) :-
   add_internal_error('Internal Error: Could not evaluate: ',evaluate_type_list(X,R)),R=X.

translate_to_type_expr(tupleExp(TupleArgs),R) :- !,
   R = typeTuple(TupleArgsRes),
   compile_tuple_args(TupleArgs,TupleArgsRes).
translate_to_type_expr(dotTuple(TupleArgs),R) :- !,
   R = dotTupleType(TupleArgsRes),
   compile_tuple_args(TupleArgs,TupleArgsRes).
translate_to_type_expr(X,CX) :-
   haskell_csp_analyzer:compile_body(X,'compile_tuple_args',[],[],CX).

compile_tuple_args([],R) :- !,R=[].
compile_tuple_args([H|T],R) :- !,
   R=[CH|RH],
   haskell_csp_analyzer:compile_body(H,'compile_tuple_args',[],[],CH),
   compile_tuple_args(T,RH).

% -----------------------------------------------------




force_evaluate_list(L,_EL) :- %print(force_evaluate_list(L,_EL)),nl,
   var(L),!, add_error(haskell_csp,'Trying to force evaluation of variable: ',L),fail.
force_evaluate_list([],R) :- !,R=[].
force_evaluate_list([H|T],R) :- !, R=[EH|ET],
    force_evaluate_argument(H,EH), %force_evaluate_argument(H,EH),
    force_evaluate_list(T,ET).
force_evaluate_list(X,R) :-
    add_internal_error('Could not (force) evaluate list: ',force_evaluate_list(X,R)),R=X.

unfold_function_call_curry_once(F,[],F,_Span).
unfold_function_call_curry_once(F,[Args|As],Value,Span) :-
     unfold_function_call_once(F,Args,Val,Span),!,
     % Val is a lambda-function
     unfold_function_call_curry_once(Val,As,Value,Span).

unfold_function_call_once(F,Par,Res,Span) :-
   evaluate_function_name(F,EF),
   unfold_function_call_once2(EF,Par,Res,Span).
   
unfold_function_call_once2(lambda(Args,Body),Parameters,R,Span) :- !,
    %print(unfold_lambda(Args,Parameters,Body,R)),nl,
%   varnumbers(lambda(Args,Body),lambda(CA,CB)),  /* lift the frozen parameter-vars back to real-variables; we should only lift variables in Args ! */
   lift(Args,CA,Body,CB),
    % print(lifted_lambda(CA,CB)),nl,
   evaluate_lambda_arguments(Parameters,CA,EvalParas),
   (CA=EvalParas -> R = CB
     ; add_error(cspm_trans,'Could not unify lambda args: ', lambda(CA,Parameters,body(Body)),Span),fail).
unfold_function_call_once2(FunName,Parameters,R,Span) :-  %print(unfold(FunName,Parameters)),nl,
  evaluate_agent_call_parameters(FunName,Parameters,EvalParas),
   %%print(call_agent_compiled(FunName,Parameters,EX)),nl,
  call_agent_compiled(FunName,Parameters,EvalParas,R,Span).

:- block call_agent_compiled(-,?,?,?,?),call_agent_compiled(?,?,-,?,?).
call_agent_compiled(FunCall,_,[],Res,Span) :- !, % no parameters
  call_agent_compiled2(FunCall,Res,Span).
call_agent_compiled(FunName,_Parameters,EvalParas,Res,Span) :-
  EX =.. [FunName|EvalParas],
  (ground(EvalParas) -> call_agent_compiled2(EX,Res,Span) 
     ; %print(not_ground_call(FunName,Parameters,EvalParas)),nl, %typechecker:debug_non_ground_term(EX),
       is_nonvar_list_skel(EvalParas,OK),  % was when(ground(EX),...)
       call_agent_compiled1(EX,Res,OK,Span)
  ).
:- block is_nonvar_list_skel(-,?).
%is_nonvar_list_skel([],true). % clause not coverable, [] is always ground (see ground(EvalParas) -> .. in the body of last clause of call_agent_compiled/5)
is_nonvar_list_skel([H|T],R) :- is_nonvar_list_skel(T,H,R).
:- block is_nonvar_list_skel(-,?,?),is_nonvar_list_skel(?,-,?).
is_nonvar_list_skel([],_,true).
is_nonvar_list_skel([H|T],_,R) :- is_nonvar_list_skel(T,H,R).

:- block call_agent_compiled1(?,?,-,?).
call_agent_compiled1(EX,Res,_,Span) :- call_agent_compiled2(EX,Res,Span).

call_agent_compiled2(EX,Res,Span) :-
   (agent_compiled(EX,Value,_SRCSPAN) 
        ->  %print(unfolded(Value,Res,EX)),nl,
            %(EX='OZS@__25'(_,_) -> trace ; true),
            Res=Value  /* local cut imposes functional interpetation */
      ;     functor(EX,F,N), functor(Skel,F,N),
            (is_defined_agent(Skel) ->  add_cspm_error(cspm_trans,'No matching case for function call: ',EX,Span)
                                   ;   add_cspm_error(cspm_trans,'Trying to apply undefined process/function: ',EX,Span)),
            fail
     ).

safe_is(X,Expr) :- when(ground(Expr), X is Expr).

add_cspm_error(Source,Msg,CSPE,Span) :-
  (translate_cspm_state(CSPE,Transl) -> true ; Transl=CSPE),
  add_error(Source,Msg,Transl,Span).

/* ------------------------------------- */
/*     Renaming related predicates       */
/* ------------------------------------- */


 :- assert_must_succeed((
      retractall(csp_full_type_constructor(_,_,_)),
      haskell_csp:rename_action(io([int(1)],left,unknown(test)),[rename(right,left),rename(left,mid)],Res),
      Res== io([int(1)],mid,unknown(test)) )).
:- assert_must_succeed(( 
      retractall(csp_full_type_constructor(_,_,_)),
      haskell_csp:rename_action(io([int(1)],ack,unknown(test)),[rename(right,left),rename(left,mid)],Res),
      Res== io([int(1)],ack,unknown(test)) )).
:- assert_must_succeed(( 
      retractall(csp_full_type_constructor(_,_,_)),
      haskell_csp:rename_action(io([int(1)],right,unknown(test)),[rename(ack,right),rename(right,dotTuple([mid,int(2)]))],Res),
      Res== io([int(2),int(1)],mid,unknown(test)) )).
 :- assert_must_succeed(( 
      retractall(csp_full_type_constructor(_,_,_)),
      haskell_csp:rename_action(io([int(1)],ack,unknown(test)),[rename(ack,left),rename(ack,right)],Res),
      Res== io([int(1)],left,unknown(test)) )).
 :- assert_must_succeed(( 
      retractall(csp_full_type_constructor(_,_,_)),
      haskell_csp:rename_action(io([int(1)],ack,unknown(test)),[rename(ack,left),rename(ack,right)],Res),
      Res== io([int(1)],right,unknown(test)) )).
 % :- assert_must_succeed(( 
 %     haskell_csp:haskell_csp:rename_action(io([int(0)],gen,unknown(test)),
 %  [rename(gen,gen1),rename(gen,dotTuple([gen2,int(1)])),
 %   rename(dotTuple([gen,int(0)]),dotTuple([gen2,int(9),int(9)])),
 %   rename(stop(no_loc_info_available),dotTuple([gen2,int(2)+int(3)]))],R),
  %   R == io([int(1),int(0)],gen2,unknown(test)) )).

:- block rename_action(-,?,-).
rename_action(tick(TS),_,tick(TS)).
rename_action(tau(X),_,tau(X)).
rename_action(io(V,Ch,Span),RenamePatternList,Res) :- Res=io(_,_,_), 
  match_pattern_list(RenamePatternList,V,Ch,Span,Res,false).

force_rename_action(io(V,Ch,Span),RenamePatternList,Res) :-  /* as above; but renaming must occur */
   match_pattern_list(RenamePatternList,V,Ch,Span,Res,true). % true: means MatchOccured: not allowed to not rename
   
%not_renamed(tick(_),_). % case cannot occur, functor_dif(AX,tick) is always called before the not_renamed/2 predicate 
not_renamed(tau(_X),_).
not_renamed(io(V,Ch,_Span),RenamePatternList) :- 
    not_covered_action(RenamePatternList,V,Ch).

not_covered_action([],_,_).
not_covered_action([rename(ChPat,_NewCh)|T],V,Ch) :- 
    (ChPat= tuple([Ch2|TPat2]) %dotTuple([Ch2|T2])
      -> (dif(Ch,Ch2) ; Ch=Ch2,no_match_list(TPat2,V))%,print(nca_dif(V,T2)),nl 
      ;  dif(Ch,ChPat)
    ), not_covered_action(T,V,Ch).

  :- assert_must_succeed(( haskell_csp:no_match_list([int(1)],[int(2)]) )).
  :- assert_must_succeed(( haskell_csp:no_match_list([int(2)],[int(1),int(2)]) )).
  :- assert_must_fail(( haskell_csp:no_match_list([],[int(1),int(2)]) )).
  :- assert_must_fail(( haskell_csp:no_match_list([int(2),int(3)],[int(2),int(3)]) )).
%%  :- assert_must_fail(( haskell_csp:no_match_list([out(int(2)),out(int(3))],[int(2),int(3)]) )).
  :- assert_must_succeed(( haskell_csp:no_match_list(X,[int(2),int(1)]), 
                           X=[int(2),int(2)])).

% check if a rename pattern matches a value on a channel:
no_match_list(RenPattern,ChValue) :- % print(no_match_list(A,B)),nl,
   l_match_pattern_value(RenPattern,ChValue,match_false).

% check if a value matches an entry in the ChannelPatternList
match_list(ChannelPatternList,V,Ch,Span) :-
   match_pattern_list(ChannelPatternList,V,Ch,Span,true,true).

:- block match_pattern_value(-,?,?), match_pattern_value(?,-,?).
match_pattern_value(record(C,PatArgs),record(VC,ValArgs),Res) :- !,
   match_record(C,PatArgs,VC,ValArgs,Res).
match_pattern_value(tuple(L1),tuple(L2),Res) :- !,
   l_match_pattern_value(L1,L2,Res).
match_pattern_value(X,tuple([H|Rest]),Res) :- !,(H=X -> Res = match_true(Rest); Res=match_false).
match_pattern_value(X,dotTuple([H|Rest]),Res) :- !,(H=X -> Res = match_true(Rest); Res=match_false).
match_pattern_value(X,record([X|Rest]),Res) :- !, Res = match_true(Rest).
match_pattern_value(Pat,Val,Res) :-
   when(?=(Pat,Val),(unify_check(Pat,Val)
                    -> Res=match_true([]) ; Res=match_false)).

unify_check(int(V1),I2) :- !, 
   (I2=int(V2) -> V1=V2
    %; preference(use_clpfd_solver,true), number(I2) -> V1 = I2
    ; add_internal_error('Type error in unification of values: ',unify_check(int(V1),I2)),fail).
unify_check(true,I2) :- !, 
   (I2=true -> true ; I2=false -> fail 
    ; add_internal_error('Type error in unification of values: ',unify_check(true,I2)),fail).
unify_check(false,I2) :- !, 
   (I2=false -> true ; I2=true -> fail 
    ; add_internal_error('Type error in unification of values: ',unify_check(false,I2)),fail).
unify_check(V,V).



:- block match_record(-,?,?,?,?), match_record(?,?,-,?,?).
match_record(C1,_,C2,_,Res) :- C1\=C2,!, Res=match_false.
match_record(C,PatArgs,C,ValArgs,Res) :- l_match_pattern_value(PatArgs,ValArgs,Res).

:- block l_match_pattern_value(-,?,?).
l_match_pattern_value(vclosure,_T,R) :- !, R=match_true(vclosure).
l_match_pattern_value([],T,R) :- !, R=match_true(T).
l_match_pattern_value([H|T],Y,R) :- !, l_match_pattern_value1(Y,H,T,R).
l_match_pattern_value(I,V,R) :-
    add_internal_error('Could not evaluate: ',l_match_pattern_value(I,V,R)),fail.

:- block l_match_pattern_value1(-,?,?,?).
l_match_pattern_value1([],H,T,R) :- !, add_error(haskell_csp,'Too many values in rename pattern: ',[H|T]), R=match_false.
l_match_pattern_value1([IODVal|IT],Pat,TPat,R) :- !,
    get_value_alsoPat(IODVal,PV), 
    match_pattern_value(Pat,PV,Res),
    % print(matched(Pat,PV,Res,TPat,IT)),nl, %
    l_match_pattern_value2(Res,TPat,IT,R).

:- block l_match_pattern_value2(-,?,?,?).
l_match_pattern_value2(match_true(M),TailPat,TailValue,Res) :- 
   ((TailPat=[],TailValue==[]) -> Res=match_true(M) % avoid error message in conjoin
    ;  TailPat=vclosure -> Res = match_true([]) % ignore match M; 
       % there could be strange pattern match, but
       % it can happen that a partial record was provided (e.g.,  [|{|c_s.c_picks, in philosophers.csp)
    ; (is_list(M),M\=[]) -> % one of the list arguments was a tuple, we need to pattern match the rest of the tuple arguments
        append(TailValue,M,TailValue1),
        l_match_pattern_value(TailPat,TailValue1,Res)
    ; otherwise ->
        l_match_pattern_value(TailPat,TailValue,RT), % print(conjoin(RT,match_true(M),Res)),nl,
        conjoin_match(RT,match_true(M),Res)).
l_match_pattern_value2(match_false,_,_,match_false).

:- block get_value_alsoPat(-,?).
get_value_alsoPat(X,Res) :-
    get_value(X,R),
    (nonvar(R),R=alsoPat(V1,_V2) ->
        Res=V1
    ; otherwise ->
        Res=R
    ).

:- block conjoin_match(-,?,?), conjoin_match(?,-,?).
conjoin_match(match_false,_,match_false).
conjoin_match(match_true(_T),match_false,match_false).
conjoin_match(match_true(T1),match_true(T2),Res) :- %append(T1,T2,TT).
   (T2==[] -> Res=match_true(T1)  /* note T2: is the first match, T1 the later one */
    ; T2==vclosure -> Res=match_true(T1) /* vclosure can propagate to the next level up */
    ; conjoin_error(T2,T1,ResMatch), Res=match_true(ResMatch)).

% note: records (for example) are gradually enumerated; hence delay error message until we know the value of the inner match
% check if we have a pattern match like rec.a.?.b.? -> this cannot be handled by us nor by FDR
:- block conjoin_error(-,?,?).
conjoin_error([],T1,M) :- !, M=T1.
conjoin_error(vclosure,T1,M) :- !, M=T1.
conjoin_error(T2,T1,T1) :- add_error(conjoin_match,'Rename pattern is incomplete in the middle. Ignoring first match: ',T2:T1).

:- assert_must_succeed(( 
    haskell_csp:match_pattern_list([tuple([c,int(0)|vclosure]),tuple([d|vclosure]),tuple([a,int(0),int(0),int(2)|vclosure])],[in(X),in(dotTuple([Y,Z]))],a,span,true,true), 
    X=int(0), Y=int(0), Z=int(2))).
:- assert_must_succeed(( 
    haskell_csp:match_pattern_list([tuple([c,int(0)|vclosure]),tuple([d|vclosure]),tuple([a,int(0),int(0),int(2)|vclosure])],[in(X),in(Y)],a,span,true,true),
    X=int(0), Y=tuple([int(0),int(2)]))).
:- assert_must_succeed(( 
    haskell_csp:match_pattern_list([tuple([c,int(0)|vclosure]),tuple([d|vclosure]),tuple([a,int(0),int(0),int(2)]),tuple([d|vclosure])],[in(X),in(Y)],a,span,true,true),
    X=int(0), Y=tuple([int(0),int(2)]))).
:- assert_must_fail((
    haskell_csp:match_pattern_list([tuple([a,int(0),int(0),int(2)]),tuple([c,int(0)|vclosure]),tuple([d|vclosure])],[in(X),in(Y)],a,span,true,true),
    X=int(0), Y=tuple([int(0),int(3)]))).
:- assert_must_fail((
    haskell_csp:match_pattern_list([tuple([c,int(0)|vclosure]),tuple([d|vclosure]),tuple([a,int(0),int(0),int(2)])],[in(X),in(Y)],a,span,true,true),
    X=int(0), Y=tuple([int(0),int(3),int(2)]))).

% match_pattern_list(PatOrRenameList, Value,Channel,Span,  Rename/MatchResult, HasMatchAlreadyOccured)
match_pattern_list([],V,Ch,Span,io(V,Ch,SpanV),MatchOccured) :- %print(no_match(V,Ch,MatchOccured)),nl,
   set_span(SpanV,Span,match_pattern_list(Ch,V)),
   MatchOccured=false. % only allow unrenamed action if no match occured earlier
match_pattern_list([rename(ChPat,NewCh)|T],V,Ch,Span,Res,MatchOccured) :- !,
   % print(checking_pat_match(ChPat,Ch,V)),nl, %%
   match_pattern_value(ChPat,tuple([Ch|V]),MatchRes),
   rename_action3(MatchRes,T,NewCh,V,Ch,Span,Res,MatchOccured).
match_pattern_list([ChPat|T],V,Ch,Span,Res,MatchOccured) :-  % for aParallel
   % print(checking_pat_match(ChPat,Ch,V)),nl, %%
   match_pattern_value(ChPat,tuple([Ch|V]),MatchRes),
   pat_match_action3(MatchRes,T,V,Ch,Span,Res,MatchOccured).

:- block pat_match_action3(-, ?,?,?,?, ?,?).
pat_match_action3(match_true(Rest), _T, _V,_Ch,_Span, Res,_MatchOccured) :-
   (Rest=[] -> Res=true ; Rest=vclosure -> Res=true). % when doing a pat_match: we only allow full matches
%pat_match_action3(match_true(_),T,V,Ch,Span,Res,_) :- match_pattern_list(T,V,Ch,Span,Res,true).
pat_match_action3(match_false,T,V,Ch,Span,Res,MatchOccured) :- match_pattern_list(T,V,Ch,Span,Res,MatchOccured).

:- block rename_action3(-, ?,?, ?,?,?, ?,?).
rename_action3(match_true(Rest), _T,NewCh, _V,_Ch,Span, Res,_MatchOccured) :-
   r_compose(NewCh,Rest,Span,Res).
rename_action3(match_true(_),T,_NewCh,V,Ch,Span,Res,_) :- match_pattern_list(T,V,Ch,Span,Res,true).
rename_action3(match_false,T,_NewCh,V,Ch,Span,Res,MatchOccured) :- match_pattern_list(T,V,Ch,Span,Res,MatchOccured).

r_compose(NewCh,Rest,Span,io(VRes,ResultingChannel,VSpan)) :-
   set_span(VSpan,Span,r_compose(NewCh)),
    %% print(rename_match(NewCh,Rest)),nl, %%
   evaluate_argument(dotTuple([NewCh|Rest]), ResultTuple),
   (ResultTuple = tuple([TCh|TV])
     ->  ResultingChannel=TCh,VRes=TV
     ;   add_internal_error('Illegal result tuple: ', evaluate_argument(dotTuple([NewCh|Rest]), ResultTuple)),
         ResultingChannel=NewCh, VRes=Rest
   ).

 
evaluate_rename_list([],[]).
evaluate_rename_list([rename(X,Y)|T],[rename(EX,EY)|ET]) :-
   ~~mnf(haskell_csp:evaluate_rename_channel_expression(X,EX)),
   ~~mnf(haskell_csp:evaluate_rename_channel_expression(Y,EY)),
   evaluate_rename_list(T,ET).
   
evaluate_link_list('linkList'(L),R) :- !,evaluate_link_list2(L,R).
evaluate_link_list('linkListComp'(GeneratorList,LinkList),R) :- !,
   %print(procCompLinkList(GeneratorList,LinkList)),nl,
   % warning: LinkList does not have the rangeEnum wrapper expected
   expand_set_comprehension(rangeEnum(LinkList),GeneratorList,setValue(ExpandedLinks)),
   evaluate_link_list2(ExpandedLinks,R).
evaluate_link_list(LL,R) :-
   add_internal_error('Cannot evaluate: ',evaluate_link_list(LL,R)),fail.
 
evaluate_link_list2([],[]).
evaluate_link_list2([link(X,Y)|T],[rename(EX,EY)|ET]) :-  /* note: link becomes rename */
   ~~mnf(haskell_csp:evaluate_rename_channel_expression(X,EX)),
   ~~mnf(haskell_csp:evaluate_rename_channel_expression(Y,EY)),
   evaluate_link_list2(T,ET).
     
rev_rename_list([],[]).
rev_rename_list([rename(X,Y)|T],[rename(Y,X)|RT]) :- rev_rename_list(T,RT).


evaluate_rename_channel_expression(E,RE) :- evaluate_argument(E,RE).
%  (E=dotTuple([CH|V])
%    -> (RE=dotTuple([CH|EV]),
%        l_evaluate_arguments(V,EV))
%    ; RE=E
%  ).
 
%%%%%%%%% DEAD CODE %%%%%%%%%%%%
/*
get_renamelist_range([],[]).
get_renamelist_range([rename(_X,Y)|T],[Y|DT]) :- 
    get_renamelist_range(T,DT).
*/

/* ------------------------------------- */

/* hidden(Action, ChannelPatternList) */
/* check whether an Action is covered by a CSP Channel Pattern List */

hidden(tau(_),_) :- fail.
hidden(tick(_),_) :- fail.
hidden(io(V,Ch,Span),  ChannelPatternList) :- match_list(ChannelPatternList,V,Ch,Span).
     
/* same as hidden but allows tau; useful for alphab. parallel */
:- block hidden_or_tau(-,?).
hidden_or_tau(tau(_),_).
hidden_or_tau(tick(_),_) :- fail.
hidden_or_tau(io(V,Ch,Span),  ChannelPatternList) :- match_list(ChannelPatternList,V,Ch,Span).


expand_channel_pattern_expression(X,_,SrcSpan) :- 
    % nl,print(expand_channel_pattern_expression(X)),nl,
    var(X), !,
    add_error(haskell_csp,'VARIABLE in expand_channel_pattern_expression: ',X,SrcSpan),fail.
expand_channel_pattern_expression(agent_call(Span,F,Par),Res,SrcSpan) :- !, 
   unfold_function_call_once(F,Par,Body,Span), expand_channel_pattern_expression(Body,Res,SrcSpan).
expand_channel_pattern_expression(setValue(List),EList,SrcSpan) :- !,l_expand(List,EList,SrcSpan).
expand_channel_pattern_expression(closure(List),EList,_SrcSpan) :- !,
     %print(haskell_csp:l_cexpand(List,EList)),nl,
     l_cexpand(List,EList).
%% this two clauses are already computed by expanding the setValue(-) and closure(-) predicates above (DEAD CODE)
/* expand_channel_pattern_expression([],R,_SrcSpan) :- !, R=[].  */    /* already computed */
/*expand_channel_pattern_expression([tuple([C|T])|T2],R,SrcSpan) :- !,*/ /* already computed */
/*       (check_channel_value_is_complete(C,T,SrcSpan)
         -> R=[tuple([C|T])|R2] ; R=R2),
       expand_channel_pattern_expression(T2,R2,SrcSpan).  
*/
expand_channel_pattern_expression(Exp,_,SrcSpan) :-
    add_error(haskell_csp,'Unknown expand_channel_pattern_expression: ',Exp,SrcSpan),fail.

/* translate a list of channel pattern expr. into uniform format */
l_expand(X,R,SrcSpan) :- var(X),
    add_internal_error_with_span(haskell_csp,'Variable List for l_expand',X,SrcSpan), R=[].
l_expand(List,Res,SrcSpan) :-
    convlist(expand_1(SrcSpan),List,Res).

expand_1(SrcSpan,El,R) :-
   expand(El,R,SrcSpan),!.

/*
l_expand([],[],_).
l_expand([H|T],Res,SrcSpan) :-
   (expand(H,EH,SrcSpan)->Res=[EH|ET];Res=ET),
   l_expand(T,ET,SrcSpan).
*/

expand(tuple([Ch|List]),R,SrcSpan) :- !,R=tuple([Ch|List]),
  check_channel_value_is_complete(Ch,List,SrcSpan).
expand(Channel,R,SrcSpan) :- 
   (atomic(Channel) -> R=tuple([Channel])
     ; add_error(haskell_csp,'Non-atomic argument: ',expand(Channel,R,SrcSpan)),fail).

check_channel_value_is_complete(Ch,List,SrcSpan) :-
 (channel_type_list(Ch,ChannelTypes)
  ->
    (same_length(ChannelTypes,List) -> true 
        ;  length(ChannelTypes,CL), length(List,L),
           (CL>L -> add_error(check_channel_value,'Incomplete event: ','.'(Ch,List),SrcSpan)
                  ; add_error(check_channel_value,'Too many values for event: ','.'(Ch,List),SrcSpan)
           ), fail
    )
  ; add_error(haskell_csp,'Undefined channel: ','.'(Ch,List),SrcSpan),fail
 ).
   
   
/* translate a list of channel pattern expr. into uniform format and add closure at end of list */
l_cexpand(X,R) :- var(X), add_error(l_cexpand,'Variable List',X), R=[].
l_cexpand(L,EL) :- maplist(cl_expand,L,EL).

% what about pat(_,_)
cl_expand(tuple([Ch|List]),R) :- !, R = tuple([Ch|EL]),append_vclosure(List,EL). %, print(app_vcl(Ch,List,EL)),nl.
cl_expand(Channel,R) :-
   (atomic(Channel) -> R = tuple([Channel|vclosure])
                    ; add_error(haskell_csp, 'Non-atomic argument: ',cl_expand(Channel,R)),fail).


append_vclosure([record(X,Fields)],Res) :- incomplete_record(record(X,Fields),[_|_],_),!,
   Res = [record(X,EFields)|vclosure],
   append_vclosure(Fields,EFields).
append_vclosure([],vclosure).
append_vclosure([H|T],[H|VT]) :- append_vclosure(T,VT).


/* not_hidden(Action, ChannelPatternList) */
/* check whether an Action is *not* covered by a CSP Channel Pattern List */

/* not_hidden_test(io(tail_in(X),outf), 
   [pat([int(2)],outf),pat([int(3)],outf),pat([int(8)],outf)]) */

:- block not_hidden(-,?).
not_hidden(tau(_),_).
not_hidden(tick(_),_).
not_hidden(io(V,Ch,Span),ChannelPatternList) :- % print(not_hidden_test(V,Ch,ChannelPatternList)),nl, %
      %not_hidden_test(ChannelPatternList,V,Ch,Span).
      match_pattern_list(ChannelPatternList,V,Ch,Span,io(_,_,_),false).  % Result is true in case match occurs with list


/* ------------------------------------- */

:- block get_value(-,?).

get_value(in(V),R) :- !,R=V.
get_value(out(V),R) :- !, add_internal_error('Obsolete out: ',get_value(out(V),R)), R=V.
get_value(dot(V),R) :- !, add_internal_error('Obsolete dot: ',get_value(out(V),R)), R=V.
get_value(record(Constructor,Fields),R) :- !, R = record(Constructor,ValueFields),
   l_get_value(Fields,ValueFields).
get_value(V,V).

:- block l_get_value(-,?).
l_get_value(L,VL) :- maplist(get_value,L,VL).
 
valid_constant(X) :- atomic(X), (csp_constant(X,_) -> true ; fail).


/* ------------------------------------------ */
/*                SPAN Utilities             */
/* ------------------------------------------ */

%%%%%%%%%%%%%%%%%%%%% UNIT TESTS FOR THE SPAN UTILITIES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- assert_must_succeed((haskell_csp: extract_span_info(_S,Info),Info == [])).
:- assert_must_succeed((haskell_csp: extract_span_info(span_info(info,src_span(3,9,3,12,45,3)),R),R == [info])).
:- assert_must_succeed((haskell_csp: extract_span_info(src_position(3,9,3,12), R),R == [])).
:- assert_must_succeed((haskell_csp: extract_span_info(src_span(3,9,3,12,45,3), R),R == [])).
:- assert_must_succeed((haskell_csp: extract_span_info(unknown(_), R),R == [])).
:- assert_must_succeed((haskell_csp: extract_span_info(multi_span(3,9,3,12,45,3,span_info(info,src_span(3,9,3,12,45,3))), R),
       R == [info])).
:- assert_must_succeed((haskell_csp: extract_span_info(src_span_operator(span_info(info,
              src_span(94417,10,94421,21,1496266,98)),src_span(94421,2,94421,4,1496345,2)),R), R == [info])).
:- assert_must_succeed((haskell_csp: extract_span_info(no_loc_info_available,R), R == [])).

:- assert_must_succeed((haskell_csp: unify_spans(src_span(3,9,3,12,45,3),S,R),S== src_span(3,9,3,12,45,3), R == src_span(3,9,3,12,45,3))).
:- assert_must_succeed((haskell_csp: unify_spans(src_span(3,9,3,12,45,3),src_span(4,9,5,16,40,5),R), 
          R == multi_span(3,9,3,12,45,3,src_span(4,9,5,16,40,5)))).
:- assert_must_succeed((haskell_csp: unify_spans(src_span_operator(src_span(94417,10,94421,21,1496266,98),src_span(94421,2,94421,4,1496345,2)),SP2, R), SP2 == src_span(94421,2,94421,4,1496345,2), 
                        R == src_span_operator(src_span(94417,10,94421,21,1496266,98),src_span(94421,2,94421,4,1496345,2)))).
:- assert_must_succeed((haskell_csp: unify_spans(unknown(_),SP2,R), SP2 == R)).
:- assert_must_succeed((haskell_csp: unify_spans(no_loc_info_available,SP2,R), SP2 == R)).
:- assert_must_succeed((haskell_csp: unify_spans(src_span(3,9,3,12,12,3), no_loc_info_available,R), R == src_span(3,9,3,12,12,3))).
:- assert_must_succeed((haskell_csp: unify_spans(span_info(info,src_span(3,9,3,12,45,3)),src_span(3,9,3,12,45,3),R), 
       R == span_info(info,src_span(3,9,3,12,45,3)))).
:- assert_must_succeed((haskell_csp: unify_spans(src_position(3,9,3,12),src_span(4,9,4,15,45,6),R),
        R == multi_span(3,9,3,21,3,12,src_span(4,9,4,15,45,6)))).
:- assert_must_succeed((extract_span_from_event2(start_cspm_MAIN,SPAN, start_cspm_MAIN,_), SPAN == src(3,1,3,5,20,4))).
:- assert_must_succeed((extract_span_from_event2(start_cspm('REC'),SPAN,start_cspm('REC'),_), SPAN == src(8,1,8,4,4,3))).
:- assert_must_fail((extract_span_from_event(_X,_Span,_T,_NS))).
:- assert_must_succeed((extract_span_info(src_position(_,_,_,_),R1), R1 ==[], extract_span_info(unknown(_),R2), R2 ==[])).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% span can already be set:
% problem: rename_action in eprocRenaming has to instantiate Span; in other contexts hidden should not instantiate span
set_span(SpanVar,SpanValue,PP) :- nonvar(SpanVar),!,
  debug_println(9,span_already_set(PP,SpanValue)).
  % print('*** Span already set to: '),  print(SpanVar),nl,
  % print('*** tried to assign value: '),print(SpanValue),nl,print('*** Context: '), print(PP),nl.
set_span(S,S,_C). % :- print(setting_span(S,_C)),nl.

extract_span_info(Span,Info) :- var(Span),!,Info=[].
extract_span_info(span_info(Info,Span),[Info|T]) :- extract_span_info(Span,T).
extract_span_info(src_position(_,_,_,_),[]).
extract_span_info(src_span(_,_,_,_,_,_),[]).
extract_span_info(src_span_operator(WholeSP1,_JustOperatorSP2),T)  :- extract_span_info(WholeSP1,T).
extract_span_info(unknown(_),[]).
extract_span_info(multi_span(_,_,_,_,_,_,SP1),T) :- extract_span_info(SP1,T).
extract_span_info(no_loc_info_available,[]).

:- block unify_spans(-,?,?).
%unify_spans(S,_,R) :- !,R=S. % commenting this in gains 10 % performance, e.g., on crossing.csp benchmark
unify_spans(SP1,SP2,R) :- ( unify_spans1(SP1,SP2,R) -> true
  ; add_internal_error('Failed: ',unify_spans(SP1,SP2,R)),set_span(SP1,R,unify_spans)).
:- block unify_spans1(-,?,?).
unify_spans1(SP1,SP2,R) :- nonvar(R),!,
   debug_println(9,span_already_computed(R,unify_spans(SP1,SP2))).
unify_spans1(span_info(Info,SP1),SP2,R) :- !, R=span_info(Info,RS),
   unify_spans1(SP1,SP2,RS).
unify_spans1(src_position(StartLine,SC,OffsetByte,LenBytes),SP2,R) :- number(SC), number(LenBytes),!,
   EndLine = StartLine, EC is SC+LenBytes,
   gen_multi_span(SP2, StartLine,SC,EndLine,EC,OffsetByte,LenBytes, R).
unify_spans1(src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes),S,R) :- !,
   unify_spans2(src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes),S,R).
unify_spans1(src_span_operator(SP1,SP2),SP3,R) :- !, R=src_span_operator(SP1,SP23),
   unify_spans1(SP2,SP3,SP23).
unify_spans1(multi_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes,SP1),SP2,R) :- !,
   unify_spans1(SP1,SP2,TailS),
   gen_multi_span(TailS, StartLine,SC,EndLine,EC,OffsetByte,LenBytes, R).
unify_spans1(unknown(_),S,R) :- !, R=S.
unify_spans1(no_loc_info_available,S,R) :- !, R=S.
unify_spans1(S,_,S) :-  add_error(unify_spans1,'Unknown Span: ',S).


:- block unify_spans2(-,?,?).
unify_spans2(src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes),S,R) :- 
          S=src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes),!, R=S.
unify_spans2(src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes),SP2,R) :- !,
          gen_multi_span(SP2, StartLine,SC,EndLine,EC,OffsetByte,LenBytes, R).
   

:- block gen_multi_span(-, ?,?,?,?, ?,?,?).
gen_multi_span(no_loc_info_available, StartLine,SC,EndLine,EC, OffsetByte,LenBytes, R) :- !,
   R = src_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes).
gen_multi_span(Span2,StartLine,SC,EndLine,EC,OffsetByte,LenBytes,R) :-
   R =  multi_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes, Span2),!.
% repeated clause
/*gen_multi_span(Span2,StartLine,SC,EndLine,EC,OffsetByte,LenBytes,R) :-
   nl,print(gen_multi_span_failed),nl, %trace,
   R =  multi_span(StartLine,SC,EndLine,EC,OffsetByte,LenBytes, Span2),!.
*/

:- block merge_span_into_event(?,-,?).
merge_span_into_event(Event,no_loc_info_available,NewEvent) :- !, 
  NewEvent=Event.
merge_span_into_event(Event,AdditionalSpan,NewEvent) :-
   (extract_span_from_event(Event,SPAN,Templ,NewSPAN)
     -> Templ = NewEvent, unify_spans(SPAN,AdditionalSpan,NewSPAN)
     ;  (%add_error(merge_span_into_event,'Unknown Event: ',Event),
         % sometimes we cannot extract a source span, e.g., tau(repInternalChoice ...)
         NewEvent = Event)
   ).

%type_of_event(io(  or colour_of_event ...
extract_span_from_event(Event,_Span,_Template,_NewSpan) :- var(Event),
  print('### extract_span_from_event variable'),nl,!,fail.
extract_span_from_event(Event,Span,Template,NewSpan) :- process_algebra_mode,!,
    extract_span_from_event2(Event,Span,Template,NewSpan).

%:- block extract_span_from_event2(-,?,?,?).    
extract_span_from_event2(start_cspm_MAIN,SPAN,start_cspm_MAIN,_) :- !,get_symbol_span('MAIN',SPAN).
extract_span_from_event2(start_cspm(S),SPAN,start_cspm(S),_) :- !,get_symbol_span(S,SPAN).
extract_span_from_event2(io(Ev,Ch,SPAN),SPAN,io(Ev,Ch,NewSPAN),NewSPAN).
extract_span_from_event2(tau(link(Left,Right)),SPAN,Templ,NS) :- !, Templ = tau(link(Left,RTempl)),
    extract_span_from_event2(Right,SPAN,RTempl,NS). % the linking code already puts info into right branch
    %extract_span_from_event2(Left,RS,LTempl,RS),
    %unify_spans(LS,RS,NS).
extract_span_from_event2(tau(hide(Event)),SPAN,Templ,NS) :- !, Templ = tau(hide(ETempl)),
    extract_span_from_event2(Event,SPAN,ETempl,NS).
%extract_span_from_event2(tau(repInternalChoice(X)),no_loc_info_available,Templ,NS) :- !, Templ = tau(repInternalChoice(X)).  
extract_span_from_event2(tau(int_choice_left(S,LSpan)),SPAN,Templ,NS) :- !,
    Templ = tau(int_choice_left(S,NS)),
    SPAN = LSpan. 
extract_span_from_event2(tau(int_choice_right(S,RSpan)),SPAN,Templ,NS) :- !,
    Templ = tau(int_choice_right(S,NS)),
    SPAN = RSpan.
extract_span_from_event2(tau(Event),SPAN,tau(Templ),NS) :-  simple_tau_event(Event,Constructor,SPAN),!,
      functor(Templ,Constructor,1),
      arg(1,Templ,NS).
%%extract_span_from_event2(i(SPAN),SPAN,i(NS),NS). %% deprecated
extract_span_from_event2(tick(SPAN),SPAN,tick(NS),NS).

simple_tau_event(chaos_stop(Span),chaos_stop,Span).
simple_tau_event(timeout(Span),timeout,Span).
%simple_tau_event(int_choice_left(_,SrcSpan),int_choice_left,SrcSpan).
   %shift_span_for_left_branch(SrcSpan,LSpan).
%simple_tau_event(int_choice_right(_,SrcSpan),int_choice_right,SrcSpan).
   %shift_span_for_right_branch(SrcSpan,RSpan).
simple_tau_event(rep_int_choice(Span),rep_int_choice,Span).
simple_tau_event(tick(Span),tick,Span).

:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(src_span(3,1,_,_,4,_),R), 
      R == src_span(3,1,3,2,4,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(src_position(3,1,3,_),R), 
      R == src_position(3,1,3,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(src_span_operator(_SP,src_span(3,1,_,_,4,_)),R), 
      R == src_span(3,1,3,2,4,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(no_loc_info_available,R),
      R == no_loc_info_available)).
:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(unknown,R),
      R == unknown)).
:- assert_must_succeed((haskell_csp: shift_span_for_left_branch(src,R),
      R == src)).
:- block shift_span_for_left_branch(-,?).
shift_span_for_left_branch(src_span(StartLine,SC,_EndLine,_EC,Off,_LenBytes),R) :- !,
                          NEC is SC+1,
                          R=src_span(StartLine,SC,StartLine,NEC,Off,1).
shift_span_for_left_branch(src_position(StartLine,SC,Off,_LenBytes),R) :- !,
                          R = src_position(StartLine,SC,Off,1).
shift_span_for_left_branch(src_span_operator(_WholeSpan,JustOperator),R) :- !,
                          shift_span_for_left_branch(JustOperator,R).
shift_span_for_left_branch(no_loc_info_available,R) :- !, R=no_loc_info_available.
shift_span_for_left_branch(unknown,R) :- !, R=unknown.
shift_span_for_left_branch(X,X):- print(not_shifting_left(X)),nl.

:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(src_span(_,_,3,2,6,5),R), 
      R == src_span(3,1,3,2,10,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(src_position(3,1,3,4),R), 
      R == src_position(3,4,6,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(src_span_operator(_SP,src_span(_,_,3,2,6,5)),R), 
      R == src_span(3,1,3,2,10,1))).
:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(no_loc_info_available,R),
      R == no_loc_info_available)).
:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(unknown,R),
      R == unknown)).
:- assert_must_succeed((haskell_csp: shift_span_for_right_branch(src,R),
      R == src)).

:- block shift_span_for_right_branch(-,?).
shift_span_for_right_branch(src_span(_StartLine,_SC,EndLine,EC,Off,LenBytes),R) :- !,
                          NL is 1, NSC is EC-1, NO is Off+LenBytes-1, 
                          R=src_span(EndLine,NSC,EndLine,EC,NO,NL).
shift_span_for_right_branch(src_position(StartLine,SC,Off,LenBytes),R) :- !,
                          NL is  1, NSC is SC+LenBytes-1, NO is Off+LenBytes-1,
                          R = src_position(StartLine,NSC,NO,NL).
shift_span_for_right_branch(src_span_operator(_WholeSpan,JustOperator),R) :- !,
                          shift_span_for_right_branch(JustOperator,R).
shift_span_for_right_branch(no_loc_info_available,R) :- !, R=no_loc_info_available.
shift_span_for_right_branch(unknown,R) :- !, R=unknown.
shift_span_for_right_branch(X,X):- print(not_shifting_right(X)),nl.
