:- multifile generate/2.
:- multifile shrink/3.

:- use_module(library(lists),[is_list/1,maplist/2]).
:- use_module(library(random),[random_permutation/2]).

% mutation of a list
generate(mutation(L:list),Value) :- 
    flattened(L) ,
    random_permutation(L,Value).
% mutation of a list of lists
generate(mutation(ListOfLists:list),Value) :- 
    % mutate each list on its own
    maplist(random_permutation,ListOfLists,Temp) ,
    % also mutate the whole list
    random_permutation(Temp,Value).

% mutation of ProB value list sets
generate(mutation(L:prob_value_set),Value) :- 
    generate(mutation(L:list),Value). 

% mutation of ProB value avl sets
generate(mutation(avl_set(Set):prob_value_set),avl_set(Value)) :- 
    avl_to_list(Set,AvlList) , 
    random_permutation(AvlList,Permutation) , 
    % don't sort values during creation of avl 
    ord_list_to_avl(Permutation,Value).  
% mutate several avl sets and return a single one
generate(mutation(L:prob_value_set),avl_set(Value)) :- 
    maplist(avl_to_list,L,ListOfLists) ,
    flatten(ListOfLists,List) , 
    random_permutation(List,Permutation) ,
    ord_list_to_avl(Permutation,Value). 

% mutation of ProB value sequences
% value avl seq
generate(mutation(avl_set(Seq):prob_value_seq),avl_set(Value)) :- 
    avl_to_list(Seq,AVLList) , 
    % remove indices to generate new ones after mutation
    % to get well-definedness
    findall(Key,member(Key-_,AVLList),Temp) ,
    seq_to_list(Temp,SeqValues) , 
    random_permutation(SeqValues,Permutation) ,  
    gen_indexed_couple_list(1,Permutation,CoupleList) , 
    findall(Key-true,member(Key,CoupleList),ValueAVL) ,
    list_to_avl(ValueAVL,Value).
generate(mutation([H|ListOfSeq]:prob_value_seq),avl_set(Value)) :- 
    H = avl_set(_) , 
    findall(Tree,member(avl_set(Tree),[H|ListOfSeq]),Seq) ,
    maplist(avl_to_list,Seq,Temp) , 
    maplist(avl_list_to_list,Temp,Temp2) , 
    maplist(seq_to_list,Temp2,List) , 
    flatten(List,SeqValues) , 
    random_permutation(SeqValues,Permutation) , 
    gen_indexed_couple_list(1,Permutation,CoupleList) , 
    findall(Key-true,member(Key,CoupleList),ValueAVL) ,
    list_to_avl(ValueAVL,Value).

% value list seq
generate(mutation(Seq:prob_value_seq),Value) :- 
    flattened(Seq) ,  
    seq_to_list(Seq,SeqValues) , 
    random_permutation(SeqValues,Permutation) , 
    gen_indexed_couple_list(1,Permutation,Value).
generate(mutation(ListOfSeq:prob_value_seq),Value) :- 
    maplist(seq_to_list,ListOfSeq,Temp) , 
    flatten(Temp,SeqValues) , 
    random_permutation(SeqValues,Permutation) , 
    gen_indexed_couple_list(1,Permutation,Value).

% mutation of trees
% random permutation of a binary tree
generate(mutation(Tree:tree(_)),Value) :- 
    \+ is_list(Tree) , 
    tree_to_list(Tree,TreeList) , 
    random_permutation(TreeList,TreePerm) , 
    list_to_tree(TreePerm,Value). 
    
% random permutation of several binary trees given in a list
generate(mutation(List:tree(_)),Value) :- 
    maplist(tree_to_list,List,TreeList) , 
    flatten(TreeList,FlattenedList) , 
    random_permutation(FlattenedList,TreePerm) , 
    list_to_tree(TreePerm,Value).

shrink(mutation(_:Type),Value,Shrunken) :-
    shrink(Type,Value,Shrunken).

% flatten a nested list 
flatten([],[]) :- !.
flatten(H,[H]) :- \+is_list(H).
flatten([H|T],List) :- 
    flatten(H,L1) , 
    flatten(T,L2) , 
    append(L1,L2,List).

% tests if list is flattened
flattened(L) :- 
    flatten(L,L).

% convert all Key-Value pairs to Key
avl_list_to_list(AVLList,List) :- 
    findall(Key,member(Key-_,AVLList),List).

% convert (Index,Value) to Value
seq_to_list(Seq,List) :- 
    findall(Value,member((_,Value),Seq),List).